/*
**  Module    : ancles.c
**  Version   : $Revision: 3.3 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Luca Medico, Gabriele Favalessa
**		Zoltan Zsoka, Filippo Natale
**
**  Synopsis  : this module contains the 'main' function of ANCLES and
**              the functions used to read data from files.
**              Functions used to drive the actions of the simulator are
**              also included here.
**
**  Functions : void 	print_node_to_node_path(Path****, int    ,
**                                              int    , int    )
**              void    print_cgen_to_cgen_path(int    , int    , int    )
**		void   	get_input(char *, void *)
**              static 	void	fix_extension(char *, char *)
**              static 	void   	print_internal(void)
**              static 	double 	eval_load_conv_fact(void)
**              static 	void    set_arrival_rate(CallGen *)
**              static 	void   	require_call_opening(int    )
**              static	double 	peak_bw_request(int    )
**              static	double 	average_bw_request(int    , double, double *)
**              static	void   	close_call(Call *)
**              	void   	schedule_next_arrival(int, int    )
**		static 	void	update_off_fbw(int	)
**		static 	void	compute_aggregate_costs(void)
**              int    	main(int argc, char *argv[])
**
**  History   : 31.10.1995 - Revision                                   -AF-
**              01.06.1996 - Revision (updatings concerning list        -AF-
**                           management and CAC algorithms).
**              01-03-1997 - Revision: modified CAC2,CAC3; added CAC4   -MM-
**              14-10-1997 - Revision: moved the routing choice at the
**                           CG level                                   -MM-
**              25-01-1998 - Revision: added all the support to LOAD BALANCING
**                                     moved the path matrix to the CG level
**                                     we can now have different
**                                     init_debug_level
**                                     - MM - LM - GF
**              20-08-1999 - Revision: added the support to discrete time
**			     updating of the official free bw for best effort
**			     traffics and aggregate information computation
**				-FN-ZZ-
**
25-01-2001 - Cleaning...  - ZZ -
**
**				2001 - 2003 ASONability included - ZZ -
**				24-07-2003 - New features: IP over ASON links
*/

#define MAIN
#include "ancles.h"
#include <ctype.h>

/************************
**  Global variables   **
************************/
static double new_aggregate_costs[MAX_NODE];
		/* array used to store the new computed values */

static double *B_link;
static int *N_link;
		/* arrays used for the calculation of estimated
		   bottleneck values in the new load dependent alg. */

/************************
**  Function Pointers  **
************************/

void            (*init_CAC)(void);
                                /* Pointer to the function configuring  */
                                /* the entities involved by the         */
                                /* selected Call Admission Control      */
                                /* algorithm.                           */
Boolean         (*exe_CAC)(Link *, double, double, double);
                                /* Pointer to the function executing    */
                                /* the actions required by the selected */
                                /* Call Admission Control algorithm.    */

/**************************
**  Function Prototypes  **
**************************/

#ifdef SAVE_THRESHOLDS
static  void	save_link_threshold(void);
#endif
static  void    fix_extension(char *, char *);
static  void    print_internal(void);
static  double  eval_load_conv_fact(void);
static  void    set_arrival_rate(CallGen *);
static  void    require_call_opening(int   , int );
static  void    close_call(Call *);
/*static  void    schedule_next_arrival(int  , int );*/
static  double  peak_bw_request(int    );
static  double  average_bw_request(int    , double, double *);
static  void    change_state(int    );
static  void	 update_off_fbw(int	);
static  void	 update_alg(int	);
static  void	 compute_aggregate_costs(void);
static  void    calc_estim_bottleneck(void);
static  void    init_pac(void);
static  double  under_estim_bottleneck(int , int );
static  void    init_dyn(void);

/***************************
**  Function Definitions  **
***************************/
#ifdef SAVE_THRESHOLDS
/*
**  Function  : void save_link_treshold(void)
**  Return    : none.
**  Remarks   : generate the file .THS containing the link threshold.
*/
static void save_link_threshold(void)
{
   char name[100];
   FILE* file;

   strcpy(name, file_radix);
   strcat(name, ".ths");

   if ((file = fopen(name, "w")) != NULL) {
     register int i, j;

     fprintf(file, "# Link threshold:\n");
     for(i = 0; i < no_of_links; i++) {
       if (link_primary_load[i] != 0.0) {
	 fprintf(file, "%d", i);
	
	 for(j = 0; j < no_of_thresholds; j++)
	   fprintf(file, "\t%f", link_vect[i]->routing_threshold[j]);

       	 fprintf(file, "\t# %s\n", link_name(i));
       }

     }
     fclose(file);
   }
}
#endif

#ifdef NON_COMPILARLA
/* Questa funzione non serve piu'... perche' i percorsi sono cgen -> cgen */

/*
**  Function  : void print_node_to_node_path(Path****, int     ord,
**                                           int     i, int     j)
**  Return    : none.
**  Remarks   : provisional function used to check the paths provided by
**              the functions included in getpaths.c.
*/
void print_node_to_node_path(Path**** nn_path_mat, int     ord,
			     int     i, int     j)
{
 LinkList *ptr;
 Path *path_point;

 printf("\nPaths from node %s to node %s :\n", node_name(i), node_name(j));
 path_point = nn_path_mat[ord][i][j];
 while(path_point != NULL)
   {
    printf("\t# hops: %d    Length (km): %.1f   Path: ",
            path_point->no_of_hops, path_point->length);
    ptr = path_point->path;
    while(ptr != NULL)
      {
       printf("%s ", link_name(ptr->link_elem->id));
       ptr = ptr->next;
      }
    printf("\n");
    path_point = path_point->next;
   }
}
#endif


/*
**  Function  : void get_input(char *format, void *variable)
**  Return    : none.
**  Remarks   : this function is to be used instead of 'scanf'. 'format'
**              is a 'scanf' format; 'variable' is the pointer to the
**              variable in which the input is written.
*/
void get_input(char *format, void *variable)
{
 static char linebuffer[255];
 char *pin;

 fgets(linebuffer, 255, stdin);         /* A datum is taken from stdin  */
 pin = strrchr (linebuffer, '\n');      /* without flushing problems.   */
 if(pin != NULL) *pin = '\0';

 sscanf(linebuffer, format, variable);  /* Only the beginning of the    */
                                        /* line is read, letting the    */
                                        /* user to write all the garbage*/
                                        /* he prefers on the input line.*/
}


/*
**  Function : static void fix_extension(char *name, char *ext)
**  Return   : none.
**  Remarks  : the string 'name' is checked for the presence of a filename
**             extension (considering the case of a full pathname) and
**             the extension 'ext' is appended to 'name' if no dot ('.') is
**             found in the terminal part of the string.
**             Different platform directory representation are considered
**             (VMS, DOS, Unix).
*/
static void fix_extension(char *name, char *ext)
{
 char *char_ptr;

# ifdef VAXC
 char_ptr = strchr(name, ']');

# elif MSDOS
 char_ptr = strrchr(name, '\\');

# else
 char_ptr = strrchr(name, '/');

# endif
 if(char_ptr == NULL && !strchr(name, '.'))
    {
     strcat(name, ".");
     strcat(name, ext);
    }
 else if(char_ptr != NULL && !strchr(char_ptr, '.'))
    {
     strcat(name, ".");
     strcat(name, ext);
    }
}


/*
**  Function  : static void print_internal(void)
**  Return    : none.
**  Remarks   : some internal ANCLES information (mainly the compile time
**              internal limits) is printed here.
*/
static void print_internal(void)
{
# ifdef __DATE__
 printf("ANCLES Version %s - Compiled on %s at %s\n",
                VERSION, __DATE__, __TIME__);
# else
 printf("ANCLES Version %s\n", VERSION);

# endif

 printf("ANCLES Internal Limits\n");
 printf("Number of Nodes: %4d  - Number of Node Connections per Node: %4d\n",
                       MAX_NODE, MAX_PORT);
 printf("Number of Users: %4d  - Number of Users per Node           : %4d\n",
                       MAX_NET_USER, MAX_USER);
 printf("Number of Links: %4d  - Data Precision                     : %4.0E\n",
                       MAX_LINK, PRECISION);
}


/*
**  Function  : static double eval_load_conv_fact(void)
**  Return    : the value of the factor to be used to convert the
**              normalized load contributions of the source users into
**              absolute load contributions.
**  Remarks   : the returned value is obtained by referring to the sum of
**              the capacities of the link connecting the source users to
**              their respective network access nodes as the whole network
**              capacity. I.E. The sum of the access link (UNI link).
*/
static double eval_load_conv_fact(void)
{
 register int i;
 double   cap_acc;

 cap_acc = 0.0;
 for(i = 0; i < no_of_cgens; i++)
    if((callgen_vect[i]->type != CG_TYPE10)
        && (callgen_vect[i]->type != CG_TYPE11)
        && (callgen_vect[i]->type != CG_TYPE12))
       cap_acc += capacity / callgen_vect[i]->ref_user->out_link->timescale;

 return cap_acc;
}


/*
**  Function  : static void set_arrival_rate(CallGen *cgen_ptr)
**  Return    : none.
**  Remarks   : the call generation rate of the source pointed by
**              'cgen_ptr' is evaluated according to the traffic
**              contribution of the user the call generator refers to.
*/
static void set_arrival_rate(CallGen *cgen_ptr)
{
 double average_bw;     /* Average Bandwidth requirement of the call    */
                        /* generator.                                   */
 double max_bw;
 double cgload;         /* Load of the call generator.                  */
 CGType1 *ci1;
 CGType2 *ci2;
 CGType3 *ci3;
 CGType4 *ci4;
 CGType5 *ci5;
 CGType6 *ci6;
 CGType7 *ci7;
 CGType9 *ci9;

 cgload = load_conv_fact * load * source_prob[cgen_ptr->ref_user->id];

 switch(cgen_ptr->type)
    {
     case CG_TYPE1 :
/* The Average Bandwidth of the generated calls is evaluated first.     */
        ci1 = (CGType1 *)cgen_ptr->info;
	average_bw = ( ci1->max_bit_rate + ci1->min_bit_rate ) * 0.5;
/* Then the arrival rate for the call generator is evaluated.           */
        ci1->arrival_rate = cgload / (ci1->mean_hold_time * average_bw);
        break;

     case CG_TYPE2 :
        ci2 = (CGType2 *)cgen_ptr->info;
/* The Average Bandwidth of the generated calls is evaluated first.     */
        average_bw = ( ci2->max_peak_rate + ci2->min_peak_rate ) * 0.5 *
           (ci2->mean_on_time /(ci2->mean_on_time + ci2->mean_off_time));
/* Then the arrival rate for the call generator is evaluated.           */
        ci2->arrival_rate = cgload / (ci2->mean_hold_time * average_bw);
        break;

     case CG_TYPE3 :
        ci3 = (CGType3 *)cgen_ptr->info;
/* The Average Bandwidth of the generated calls is evaluated first.     */
        average_bw = (ci3->max_bit_rate_avg + ci3->min_bit_rate_avg) * 0.5;
/* Then the arrival rate for the call generator is evaluated.           */
        ci3->arrival_rate = cgload / (ci3->mean_hold_time * average_bw);
        break;

     case CG_TYPE4 :
        ci4 = (CGType4 *)cgen_ptr->info;
/* The Average Bandwidth of the generated calls is evaluated first.     */
        average_bw = ci4->higher_bit_rate_avg * ci4->higher_bit_rate_prob +
            ci4->medium_bit_rate_avg * ci4->medium_bit_rate_prob +
            ci4->lower_bit_rate_avg * ci4->lower_bit_rate_prob;
/* Then the arrival rate for the call generator is evaluated.           */
        ci4->arrival_rate = cgload / (ci4->mean_hold_time * average_bw);
        break;

     case CG_TYPE5 :
        ci5 = (CGType5 *)cgen_ptr->info;
/* The Average Bandwidth of the generated calls is evaluated first.     */
        average_bw = ( ci5->max_max_rate + ci5->min_max_rate ) * 0.5;
/* The minimum acceptable bandwidth is evaluated.	         	*/
        ci5->min_acceptable_bw =
            ci5->min_accept_bw_thrs * average_bw;
/* The length of the longest call for the call generator is evaluated.	*/
        ci5->longest_call = ci5->mean_gen_traffic * lhstep_factor /
	    ci5->min_acceptable_bw;
/* The arrival rate for the call generator is evaluated.         	*/
        ci5->arrival_rate = cgload / ci5->mean_gen_traffic;
        break;

     case CG_TYPE6 :
/* The Average Bandwidth of the generated calls is evaluated first.     */
        ci6 = (CGType6 *)cgen_ptr->info;
        average_bw = ( ci6->max_no_of_waves + ci6->min_no_of_waves )
                     * 0.5 * wavelength_capacity;
        ci6->arrival_rate = cgload / (ci6->mean_hold_time * average_bw);
        break;

     case CG_TYPE7 :
/* The Average Bandwidth of the generated calls is evaluated first.     */
        ci7 = (CGType7 *)cgen_ptr->info;
	if(ci7->saturation > 0)
	    {
	    ci7->arrival_rate = 1.0 / (ci7->mean_hold_time * (1 - ci7->saturation) / ci7->saturation);
	    }
	else	// ci7->no_of_waves > 0 
	    {	
		max_bw = ci7->no_of_waves * wavelength_capacity;
		ci7->saturation = cgload / max_bw;
	    
/* In the ON-OFF mode can not be generated a traffic bigger than the
   rate of the calls. */
               if(ci7->saturation < 1.0)
        	ci7->arrival_rate = cgload / (ci7->mean_hold_time *
                               (max_bw - cgload));
               else
		{
    	        printf("Error:  The given load is too high and can not be\n\
realized by the %s CG_TYPE7 type user with %d wavelengths!\n",
                user_name(cgen_ptr->ref_user->id), ci7->no_of_waves);
                exit(1);
        	}
	    }	
	break;

     case CG_TYPE9 :
/* The Average Bandwidth of the generated calls is evaluated first.     */
        ci9 = (CGType9 *)cgen_ptr->info;
        average_bw = ( ci9->max_no_of_waves + ci9->min_no_of_waves )
                     * 0.5 * wavelength_capacity;
        ci9->arrival_rate = ( cgload * ci9->arr_rate_increment) / 
			((ci9->peakedness - 1.0) * average_bw);
        
/*		ci9->sch_at_end = FALSE;*/
/* first element of the ordered endtime list (of opened calls) is statically -1.0 */
        ci9->call_closures = dbl_list_alloc();
        ci9->call_closures->value = -1.0;
        ci9->call_closures->next = NULL;
        break;

     case CG_TYPE8 :
    	fprintf(stderr, "Invalid positive source probability for a call generator type 8!\n");
	exit(1);

     default:
      fprintf(stderr, "Invalid call generator type (%d) in function set_arrival_rate\n",
	      cgen_ptr->type);
      exit(1);
    }
}


/*
**  Function  : static void require_call_opening(int     source_cg, int     dest_cg)
**  Return    : none.
**  Remarks   : the arrival of the next call opening request from the call
**              generator identified by 'source_cg' is scheduled here.
**              The destination user is chosen according to the destination
**              cumulative probability matrix.
**              The peak bandwidth request is noticed to the network.
*/
static void require_call_opening(int     source_cg, int     dest_cg)
{
 double  peak_bandwidth;/* Peak bandwidth declared by the source call   */
                        /* generator.                                   */
 double  mean_bandwidth;/* Average bandwidth declared by the source     */
                        /* call generator.                              */
 double  minimum_bandwidth;
                        /* Minimum bandwidth declared by the source     */
                        /* call generator.                              */
 double  x;             /* Stochastic sample extracted from a uniform   */
                        /* distribution in order to fix the destination */
                        /* user.                                        */

 if((callgen_vect[source_cg]->type != CG_TYPE7) && (callgen_vect[source_cg]->type != CG_TYPE8))
  {
   x = uniform_0_1(&cg_seed_vect[source_cg]);
   for(dest_cg = 0; destination_prob[callgen_vect[source_cg]->ref_user->id]
	   [callgen_vect[dest_cg]->ref_user->id] <= x; dest_cg++);
  }
 peak_bandwidth = peak_bw_request(source_cg);
 mean_bandwidth = average_bw_request(source_cg, peak_bandwidth,
                                     &minimum_bandwidth);

 if(ipowdm_ena && (callgen_vect[source_cg]->type == CG_TYPE8))
 {
	 open_new_virt_link(callgen_vect[source_cg]->ref_user->connected_node,
		 callgen_vect[dest_cg]->ref_user->connected_node);
 }
 else
 {
	 /* if the virtual link strategy performs to open a new optical 
	 channel first and the IP routing begins only after */
	 if(ipowdm_ena && vl_open_algorithm == OCH_FIRST)
		 open_loglink = 1;
	 /* else IP routing is performed first */
	 else
	 {
		 open_loglink = 0;
		 exe_routing(source_cg, dest_cg, mean_bandwidth, peak_bandwidth,
		 minimum_bandwidth);
	 }

	 /* In the case of IPoWDM the story may not finish after a failed routing */
	 while(open_loglink)
	 {
		 /* perform the opening of a virtual link */
		 /* it is possible to reset open_loglink */
		 open_new_virt_link(callgen_vect[source_cg]->ref_user->connected_node,
			 callgen_vect[dest_cg]->ref_user->connected_node);
		 exe_routing(source_cg, dest_cg, mean_bandwidth, peak_bandwidth,
             minimum_bandwidth);
	 }
 }

  /* If the call generator is ON-OFF ASON or NEVER CLOSING ASON, do not place the next call,
     it will be inserted at the closing of this call. */
 if((callgen_vect[source_cg]->type != CG_TYPE7) &&
	(callgen_vect[source_cg]->type != CG_TYPE8))
	 schedule_next_arrival(source_cg, dest_cg);
}


/*
**  Function  : static double peak_bw_request(int cg_id)
**  Return    : the peak bandwidth required by the call.
**  Remarks   : the peak bandwidth is obtained through a probability
**              density function that is specific for each call generator
**              type.
*/
static double peak_bw_request(int cg_id)
{
 double lower_limit, upper_limit; /* Limits of the uniform density  */
                                  /* function.                      */
 CGType1 *ci1;
 CGType2 *ci2;
 CGType3 *ci3;
 CGType4 *ci4;
 CGType5 *ci5;
 CGType6 *ci6;
 CGType9 *ci9;

 switch(callgen_vect[cg_id]->type)
    {
     case CG_TYPE1 :
        ci1 = (CGType1 *)callgen_vect[cg_id]->info;
        lower_limit = ci1->min_bit_rate;
        upper_limit = ci1->max_bit_rate;
        break;

     case CG_TYPE2 :
        ci2 = (CGType2 *)callgen_vect[cg_id]->info;
        lower_limit = ci2->min_peak_rate;
        upper_limit = ci2->max_peak_rate;
        break;

     case CG_TYPE3 :
        ci3 = (CGType3 *)callgen_vect[cg_id]->info;
        lower_limit = ci3->max_bit_rate_avg - ci3->half_bw_interval;
        upper_limit = ci3->max_bit_rate_avg + ci3->half_bw_interval;
        break;

     case CG_TYPE4 :
        ci4 = (CGType4 *)callgen_vect[cg_id]->info;
        lower_limit = ci4->higher_bit_rate_avg - ci4->half_bw_interval;
        upper_limit = ci4->higher_bit_rate_avg + ci4->half_bw_interval;
        break;

     case CG_TYPE5 :
        ci5 = (CGType5 *)callgen_vect[cg_id]->info;
        lower_limit = ci5->min_max_rate;
        upper_limit = ci5->max_max_rate;
        break;

     case CG_TYPE6 :
        ci6 = (CGType6 *)callgen_vect[cg_id]->info;
        lower_limit = ci6->min_no_of_waves * wavelength_capacity;
        upper_limit = ci6->max_no_of_waves * wavelength_capacity;
        break;

     case CG_TYPE7 :
        lower_limit = wavelength_capacity;
        upper_limit = wavelength_capacity;
        break;

     case CG_TYPE8 :
        lower_limit = wavelength_capacity;
        upper_limit = wavelength_capacity;
        break;

     case CG_TYPE9 :
        ci9 = (CGType9 *)callgen_vect[cg_id]->info;
        lower_limit = ci9->min_no_of_waves * wavelength_capacity;
        upper_limit = ci9->max_no_of_waves * wavelength_capacity;
        break;

     case CG_TYPE10 :
	printf("\nError:  Function 'peak_bw_request' must not be called for");
	printf("\n        SINK call generators.");
	exit(1);
        break;

     case CG_TYPE11 :
	printf("\nError:  Function 'peak_bw_request' must not be called for");
	printf("\n        TCP_SINK call generators.");
	exit(1);
        break;

     case CG_TYPE12 :
	printf("\nError:  Function 'peak_bw_request' must not be called for");
	printf("\n        ASON_SINK call generators.");
	exit(1);
        break;

     default:
        fprintf(stderr, "ancles.c:peak_bw_request(): unknown call generator type\n");
	exit(1);
     }
  return (uniform(lower_limit, upper_limit, &cg_seed_vect[cg_id]));
}


/*
**  Function  : static double average_bw_request(int cg_id, double peak_rate,
**						 double *min_bw)
**  Return    : the average bandwidth required by the call.
**  Remarks   : the average bandwidth is obtained through a function that
**              is specific for each call generator type.
**              All the bandwidth parameters of the call other than the average
**              bandwidth are evaluated here.
*/
static double average_bw_request(int cg_id, double peak_rate,
				 double *min_bw)
{
 double lower_limit, upper_limit; /* Limits of the uniform density  */
                                  /* function.                      */
 double medium_vbr;               /* Value of the intermediate bit  */
                                  /* rate for type 4 calls.         */
 double average_bw;               /* the evaluated average bw       */

 CGType2 *ci2;
 CGType3 *ci3;
 CGType4 *ci4;

 switch(callgen_vect[cg_id]->type)
    {
     case CG_TYPE1 :
        *min_bw = peak_rate;
        average_bw = peak_rate;
        break;

     case CG_TYPE2 :
        *min_bw = 0.0;
	ci2 = (CGType2 *)callgen_vect[cg_id]->info;
        average_bw = peak_rate * ci2->mean_on_time /
              (ci2->mean_on_time + ci2->mean_off_time);
        break;

     case CG_TYPE3 :
	ci3 = (CGType3 *)callgen_vect[cg_id]->info;
        lower_limit = ci3->min_bit_rate_avg - ci3->half_bw_interval;
        upper_limit = ci3->min_bit_rate_avg + ci3->half_bw_interval;
        *min_bw = uniform(lower_limit, upper_limit, &cg_seed_vect[cg_id]);
        average_bw = (peak_rate + *min_bw) * 0.5;
        break;

     case CG_TYPE4 :
	ci4 = (CGType4 *)callgen_vect[cg_id]->info;
        lower_limit = ci4->lower_bit_rate_avg - ci4->half_bw_interval;
        upper_limit = ci4->lower_bit_rate_avg + ci4->half_bw_interval;
        *min_bw = uniform(lower_limit, upper_limit, &cg_seed_vect[cg_id]);

        lower_limit = ci4->medium_bit_rate_avg - ci4->half_bw_interval;
        upper_limit = ci4->medium_bit_rate_avg + ci4->half_bw_interval;
        medium_vbr = uniform(lower_limit, upper_limit, &cg_seed_vect[cg_id]);

/* The average bandwidth of the requested connection is returned.       */
        average_bw = peak_rate * ci4->higher_bit_rate_prob +
           medium_vbr * ci4->medium_bit_rate_prob +
           (*min_bw) * ci4->lower_bit_rate_prob;
        break;

     case CG_TYPE5:
        *min_bw = (UNDEFINED_BW);
	average_bw = (UNDEFINED_BW);
        break;

     case CG_TYPE6 :
        *min_bw = peak_rate;
        average_bw = peak_rate;
        break;

     case CG_TYPE7 :
        *min_bw = peak_rate;
        average_bw = peak_rate;
        break;

     case CG_TYPE8 :
        *min_bw = peak_rate;
        average_bw = peak_rate;
        break;

     case CG_TYPE9 :
        *min_bw = peak_rate;
        average_bw = peak_rate;
        break;

     case CG_TYPE10 :
	printf("\nError:  Function 'average_bw_request' must not be called ");
	printf("for\n        SINK call generators.");
	exit(1);
        break;

     case CG_TYPE11 :
	printf("\nError:  Function 'average_bw_request' must not be called ");
	printf("for\n        TCP_SINK call generators.");
	exit(1);
        break;

     case CG_TYPE12 :
	printf("\nError:  Function 'average_bw_request' must not be called ");
	printf("for\n        ASON_SINK call generators.");
	exit(1);
        break;

     default:
        fprintf(stderr, "ancles.c:average_bw_request(): unknown call generator type\n");
	exit(1);
    }
  return (average_bw);
}


/*
**  Function  : static void close_call(Call *del_call)
**  Return    : none.
**  Remarks   : the dynamic memory concerning the call just released
**              (the record in the general Call list and the records in the
**              lists associated to the links of the path) is released.
**              All the parameters whose value changes due to the call
**              closure are updated here.
**		If a best effort call still has data to transfer, put a
**		new close-event in the event list.
*/
static void close_call(Call *del_call)
{
 LinkList *path;        /* Path followed by the closing call.           */
 Link     *link_ptr;    /* Pointer to the links of the path.            */
 double   tmp_ref_bw;   /* Identifier of the band of the closing call   */
#ifdef NEW_MEASURE
 double   loc_mean, loc_peak;
 CallType1* ci1;
 CallType2* ci2;
 CallType3* ci3;
 CallType4* ci4;
#endif
 int    i, sel_wave;
 unsigned char free_mask;
 int    new_peak;

 Boolean  protmayclose = FALSE;
 Boolean  closed_prot_chan = FALSE, foundpeak;
 Path	*protpath; /* protection path */
 int	max_col;  /* maximum calls on protected links and nodes*/
 Intlist	*op_sec_ptr, *prev_os_ptr; /* wavelength to free in op. sec.*/
 ProtCh	*protchan, *prev_protchan;  /* prot. channel deiscriptor pointers */
 int	start_node;			/* starting point of the link */
 Boolean	not_first_node; /* indicates if start_node is 

/* Variables used for replacing the close event */
 LinkList  *this_link;
 CallType5 *temp_call = NULL;
 CallList *call_list_ptr;
/* at least a small timeslot is needed before we close a virtual link */
 double  res_time; 
 Event  *new_close_ev;

 CGType9 *ci9;
 Dbllist *deldbl;

 CallType8 *ct8;
 Boolean empty; /* virtual link(pair) is empty and closeable */
 int fwdlid, bwdlid;
/* char	output_file[80];*/

 double htime;
 int srcuser, dstuser;
 int hopdep;
 
 if(del_call->call_type == CALL_TYPE5)
  temp_call = (CallType5 *)(del_call->info);

/* If the U call didn't already transfer all the data -> replace the
   close events. */
 if(temp_call && !time_based_ct5 && !interrupt_happened &&
    (res_time = (temp_call->tr_left / temp_call->curr_br) -
    (ancles_time - temp_call->last_updt_time)) > DBL_CMP_THS)
  {
  /* if the end is near -> allow replacing on all changes*/
   if(res_time < EPSILON)
    {
     del_call->end_time = ancles_time + res_time;
     del_call->repl = TRUE;
    }
  /* else the the closing time is computed with the max bw value */
   else
    del_call->end_time = ancles_time + (res_time * temp_call->curr_br /
      temp_call->mbr_value);

  /* Inserting event of call closing */

   new_close_ev = event_alloc();
   new_close_ev->ev_type = CLOSING_CALL;
   new_close_ev->ev_dscrpt.deleting_call = del_call;
   new_close_ev->ev_time = del_call->end_time;
   del_call->my_close_event = new_close_ev;
   insert_event(new_close_ev);

  /* The call in the call-list and Ucall-list-on-link have to be replaced */
   for(this_link = del_call->path_ptr->path; this_link != NULL;
       this_link = this_link->next)
    {
     link_ptr = this_link->link_elem;
  /* Find the right element in the Ucall-list of the link  */
     call_list_ptr = search_Ucall_list_on_link(link_ptr, del_call, ancles_time);

     unrel_dequeue_Ucall_list_from_link(call_list_ptr, link_ptr, ancles_time);
     insert_Ucall_list_on_link(call_list_ptr, link_ptr);
    }
  /* Removing the call from the call-list and put it in its new place */
   unrel_dequeue_call(del_call, ancles_time);
   insert_call(del_call);
  }
 else
  {
   path = del_call->path_ptr->path;
   protpath = del_call->prot_path_ptr;

   /* the destruction of a virtual link is performed */
   if(ipowdm_ena && del_call->call_type == CALL_TYPE8)
   {
	   ct8 = (CallType8 *)del_call->info;
	   close_virt_link(ct8 -> fwd_link);
	   /* if backward link exists */
	   if(duplex_llc)
		   close_virt_link(ct8 -> bwd_link);

	   /* holding time statistics of virtual links */
	htime = ancles_time - del_call->start_time;
	srcuser = callgen_vect[del_call->source_id]->ref_user->id;
	dstuser = callgen_vect[del_call->dest_id]->ref_user->id;

	accu_vlink_lifetime[srcuser][dstuser] += htime;
	sqraccu_vlink_lifetime[srcuser][dstuser] += htime * htime;
	closed_vlinks[srcuser][dstuser]++;

   	accu_vlink_timehop[srcuser][dstuser] += 
			htime * del_call->path_ptr->no_of_hops;

	/* May be later: accu_vlink_number, accu_vlink_hopnum */
	

	   /* This detailed statistics is no more used !
	   strcpy(output_file, file_radix);
	   strcat(output_file, ".cg8");
	   if((CG8FILE = fopen(output_file, "a")) == NULL)
	   {
		   printf("\nError:  the file %s cannot be opened.\n", output_file);
		   exit(1);
	   }
	   fprintf(CG8FILE, "C\t%s\t%s\t%.6f\t%.6f\n",
		   user_name(callgen_vect[del_call->source_id]->ref_user->id),
		   user_name(callgen_vect[del_call->dest_id]->ref_user->id),
		   ancles_time, del_call->start_time);
	   fclose(CG8FILE);*/
   }
   
   not_first_node = FALSE;

   /* The closure of the ASON call changes the aviable waves on the links. */
   if(del_call->call_type == CALL_TYPE6 || del_call->call_type == CALL_TYPE8)
    {
	   /* dedicated protection is used in 1+1 */
	   protchan = del_call->dedic_prot_ch;

	   while(path != NULL)
	   {
		   link_ptr = path->link_elem;
		   sel_wave = link_ptr->calls_on->next->used_wave;
		   free_mask = 0x80;
		   for(i = 0; i < (sel_wave-1)%8; i++)
			   free_mask >>= 1;
		   if(link_ptr->is_node_node)
		   {
			   if( !link_ptr->free_fibers[sel_wave-1] )		
				   /* Reset the flag of the used wavelength if needed */
				   link_ptr->free_waves[(sel_wave-1)/8] |= free_mask;
			   
			   /* Update the number of available fibers */
			   link_ptr->free_fibers[sel_wave-1]++;

			   if(path->direction)
				   start_node = link_ptr->dest;
			   else
				   start_node = link_ptr->source;

			   /* If the path was protected, a protection channel may be closed. */
			   if(protpath)
			   {
				   if(protpath->protected_links[link_ptr->nnid]==
					   protpath->num_of_prot_chs ||
					   protpath->protected_nodes[start_node]==
					   protpath->num_of_prot_chs ||
			       /* if 1+1 routing was used */
					   protchan)
					   protmayclose = TRUE;
				   /* Number of calls on this link protected by a channel of this
				   path will be decreased. */
				   protpath->protected_links[link_ptr->nnid]--;

				   /* Number of calls on the start node of the link 
				     protected by a channel of this path will be decreased. */
				   /* Except of the first node of the whole path */
				   if(not_first_node)
					   protpath->protected_nodes[start_node]--;
				   not_first_node = TRUE;
			   }
		   }/* if */
		   link_ptr->used_waves--;
		   dequeue_call_list_from_link(link_ptr->calls_on->next, link_ptr);
		   
		   /* Change the peak value if needed: */
		   /* the call with the maximal wavelength order on this link is closed */
		   if(link_ptr->peak_wave == sel_wave)
		   {
		   /* Looking for the new peak wave (it is not
			   greater than the old one) */
			   new_peak = sel_wave;
			   foundpeak = FALSE;
			   while( !foundpeak && new_peak)
			   {
				   if( link_ptr->free_fibers[new_peak-1] <
					   link_ptr->fiber_num )
					   foundpeak = TRUE;
				   else
					   new_peak--;
			   }
			   link_ptr->peak_wave = new_peak;
		   }
		   path = path->next;
	   } /* while */

	   /* If a protection channel may be closed */
	   if(protmayclose)
	   {
		   max_col = 0;
		   for(i = 0; i < no_of_nnlinks; i++)
		   {
			   if(protpath->protected_links[i] > max_col)
				   max_col = protpath->protected_links[i];
		   }

		   for(i = 0; i < no_of_nodes; i++)
		   {
			   if(protpath->protected_nodes[i] > max_col)
				   max_col = protpath->protected_nodes[i];
		   }

		/* A protection channel has to be closed -> close it on each
		   traversed node-node links */
		   if(max_col < protpath->num_of_prot_chs ||
			 /* if 1+1 routing was used */
			   protchan)
		   {
			   closed_prot_chan = TRUE;
			   protpath->num_of_prot_chs = max_col;
			   path = protpath->path->next;
		/* We have information about the wavelengths which were used
		for the protection channel in each opaque section. */

			   /* If dedicated protection channel has to be closed */
			   if(protchan)
				   op_sec_ptr = protchan->opsec_waves;
			   /* Else close the latest open protection channel
			      (first in list) */
			   else
				   op_sec_ptr = protpath->prot_channels->opsec_waves;
			   
			   sel_wave = op_sec_ptr->value;
			   while(path->next != NULL)
			   {
				   link_ptr = path->link_elem;
				   free_mask = 0x80;
				   for(i = 0; i < (sel_wave-1)%8; i++)
					   free_mask >>= 1;
				   if(link_ptr->is_node_node)
				   {
					   if(!link_ptr->free_fibers[sel_wave-1])
						   /* Reset the flag of the used wavelength if needed */
						   link_ptr->free_waves[(sel_wave-1)/8] |= free_mask;
					   /* Update the number of free fibers by this wavelength */
					   link_ptr->free_fibers[sel_wave-1]++;
				   }/* if */
				   link_ptr->used_waves--;

				   /* Change the peak value if needed */
				   /* We closed the call with the maximal wavelength order on this link */
				   if(link_ptr->peak_wave == sel_wave)
				   {
					/* Looking for the new peak wave (it is not
					   greater than the old one) */
					   new_peak = sel_wave;
					   foundpeak = FALSE;
					   while( !foundpeak && new_peak)
					   {
						   if( link_ptr->free_fibers[new_peak-1] <
							   link_ptr->fiber_num )
							   foundpeak = TRUE;
						   else
							   new_peak--;
					   }
					   
					   link_ptr->peak_wave = new_peak;
				   }
				   
				   /* If the opaque section is finished, take the next 
				    and release the current descriptor */
				   if(path->endopsec || !path->next->next)
				   {
					   prev_os_ptr = op_sec_ptr;
					   op_sec_ptr = op_sec_ptr->next;
					   int_list_release(prev_os_ptr);
					   if(op_sec_ptr)
						   sel_wave = op_sec_ptr->value;
				   }
				   path = path->next;
			   }/* while */

		/* Dedicated channel will be released. */
			   if(protchan && protchan != protpath->prot_channels)
			   {
				   for(prev_protchan = protpath->prot_channels;
				    prev_protchan->next != protchan;
				    prev_protchan = prev_protchan->next);
				   prev_protchan->next = protchan->next;
			   }
			   else
			   {
		/* The latest inserted channel will be released. */
				   protchan = protpath->prot_channels;
				   protpath->prot_channels = protchan->next;
			   }
			   prot_ch_release(protchan);
		   }
	   }
    } /* if */

/* Updating the statistical variables involved by the closure of the    */
/* call pointed by 'del_call'.                                          */
   closed_call_stat_update(del_call, closed_prot_chan);

/* If a path through the network exists, the updatings above are        */
/* applied to each link of the path.                                    */
   if(del_call->call_type != CALL_TYPE6 && del_call->call_type != CALL_TYPE8)
    {
      res_time = vl_close_hold_time + DBL_CMP_THS;
	  while(path != NULL)
      {
       link_ptr = path->link_elem;
       if(del_call->call_type != CALL_TYPE5)
	     tmp_ref_bw = link_ptr->calls_on->next->ref_bw;
       else
	     tmp_ref_bw =
	                link_Ucall_stat[link_ptr->id]->U_calls_on->next->ref_bw;

	    link_ptr->free_band += tmp_ref_bw;

	    switch (selected_cac)
	     {
	      case CALL_ADM_CTRL1 :
		  if(del_call->call_type == CALL_TYPE5)
		     dequeue_Ucall_list_from_link(
			link_Ucall_stat[link_ptr->id]->U_calls_on->next,
			                                              link_ptr);
		  else
		     dequeue_call_list_from_link(
			link_ptr->calls_on->next, link_ptr);
#ifndef NEW_MEASURE
		update_Ucall_close(del_call, link_ptr);
#endif
	 	 break;

	      case CALL_ADM_CTRL2 :
	          update_samples_vect(link_ptr, FALSE);
	          backoff_ths[link_ptr->id] += tmp_ref_bw;
		  dequeue_call_list_from_link(link_ptr->calls_on->next,
		                                                      link_ptr);
		  break;

	      case CALL_ADM_CTRL3 :
	          update_samples_vect(link_ptr, FALSE);
	          backoff_ths[link_ptr->id] += tmp_ref_bw;
/* The Mean and Variance accumulators are decreased by the              */
/* contributions of the released connection.                            */
	          updt_var_mean(link_ptr->calls_on->next->call_elem,
                       &link_mean[link_ptr->id], &link_variance[link_ptr->id],
                       FALSE);
		  dequeue_call_list_from_link(link_ptr->calls_on->next,
		                                                      link_ptr);
		  break;

	      case CALL_ADM_CTRL4 :
		  if(del_call->call_type == CALL_TYPE5)
		     dequeue_Ucall_list_from_link(
			link_Ucall_stat[link_ptr->id]->U_calls_on->next,
			                                              link_ptr);
		  else
		     dequeue_call_list_from_link(
			link_ptr->calls_on->next, link_ptr);
#ifndef NEW_MEASURE
		update_Ucall_close(del_call, link_ptr);
#endif
	 	  break;
	     } /* case */

		   /* if a virtual link even became empty its type8 call closing event can be 
		  timed to a finite point (use it for the predefined links only if enabled) */
		if(ipowdm_ena && link_ptr->linked_call &&
			(link_ptr->linked_call->start_time > 2*DBL_MIN ||
			cl_predef_vl) &&
			vl_close_algorithm == EMPTY_CLOSE &&
			link_ptr->calls_on->next == link_ptr->calls_on &&
			link_Ucall_stat[link_ptr->id]->U_calls_on->next ==
			link_Ucall_stat[link_ptr->id]->U_calls_on)
		{
			empty = FALSE;
			if(!duplex_llc)
				empty = TRUE;
			else
			{
				fwdlid = ((CallType8 *)link_ptr->linked_call->info)->
					fwd_link->id;
				bwdlid = ((CallType8 *)link_ptr->linked_call->info)->
					bwd_link->id;
			/* in the case of duplex optical connections both virtual links 
			has to be empty to perform the closure */
				if(link_vect[fwdlid]->calls_on->next ==
					link_vect[fwdlid]->calls_on &&
					link_Ucall_stat[fwdlid]->U_calls_on->next ==
					link_Ucall_stat[fwdlid]->U_calls_on &&
					link_vect[bwdlid]->calls_on->next ==
					link_vect[bwdlid]->calls_on &&
					link_Ucall_stat[bwdlid]->U_calls_on->next ==
					link_Ucall_stat[bwdlid]->U_calls_on)
					empty = TRUE;
			}

			if(empty)
			{
			/* the timing should be different for each virtual link to close 
			if there are more links to close */
				res_time += DBL_CMP_THS;
			/* if the length of virtual link shall affect the before closing
				holding time, we set the factor linearly depending on the hopnumber */
				if(vlht_hopdep)
					hopdep = link_ptr->linked_call->path_ptr->no_of_hops - 1;
				else
					hopdep = 1;

				reinsert_call_and_event(link_ptr->linked_call, ancles_time + res_time * 
					hopdep);
			}
		}

       path = path->next;
      } /* while */

#ifdef NEW_MEASURE
   switch (del_call->call_type)
     {
      case CALL_TYPE1 :
		     ci1 = (CallType1 *)(del_call->info);
		     loc_mean = ci1->cbr_value;
	        loc_peak = ci1->cbr_value;
	        break;

      case CALL_TYPE2 :
	        ci2 = (CallType2 *)(del_call->info);
	        loc_mean = ci2->peak_rate * ci2->mean_on_time /
	                     (ci2->mean_on_time + ci2->mean_off_time);
	        loc_peak = ci2->peak_rate;
	        break;

      case CALL_TYPE3 :
	        ci3 = (CallType3 *)(del_call->info);
	        loc_mean = (ci3->min_vbr_value + ci3->max_vbr_value) / 2.0;
	        loc_peak = ci3->max_vbr_value;
	        break;

      case CALL_TYPE4 :
	        ci4 = (CallType4 *)(del_call->info);
	        loc_mean = ci4->higher_bit_rate_prob * ci4->higher_bit_rate_value +
	                   ci4->medium_bit_rate_prob * ci4->medium_bit_rate_value +
	                   ci4->lower_bit_rate_prob * ci4->lower_bit_rate_value;
	        loc_peak = ci4->higher_bit_rate_value;
	        break;

      case CALL_TYPE5 :
	        loc_mean = UNDEFINED_BW;
	        loc_peak = temp_call->mbr_value;
	        break;

      case CALL_TYPE6 :
      case CALL_TYPE8 :
	        loc_mean = 0.0;
	        loc_peak = 0.0;
	        break;

      default:
	        printf("ERROR: unknown call type %d in function close_call\n",
		                      del_call->call_type);
	        exit(1);
     } /* case */
    update_bottleneck_measure(del_call, loc_mean, loc_peak, CLOSE_CALL);
#endif
    }

	dequeue_call(head_of_call_list->next);

/* If the ason call generator is ON-OFF, place the next call. */
   if(callgen_vect[del_call->source_id]->type == CG_TYPE7)
        schedule_next_arrival(del_call->source_id, del_call->dest_id);
   
/* The intensity is decremented in the case of 
		Pascal distributed arrivals */
   if(callgen_vect[del_call->source_id]->type == CG_TYPE9)
   {
	   ci9 = (CGType9 *)callgen_vect[del_call->source_id]->info;
	   ci9->arrival_rate -= ci9->arr_rate_increment;

	   deldbl = ci9 -> call_closures->next;
	   ci9 -> call_closures->next = deldbl->next;
	   dbl_list_release(deldbl);


	   schedule_next_arrival(del_call->source_id, del_call->dest_id);

	   /*if(ci9->sch_at_end)
	   {
		   ci9 ->sch_at_end = FALSE; /* lehet, hogy nem is kell ! /
		   schedule_next_arrival(del_call->source_id, del_call->dest_id);
	   }
	   else
		   printf("megis kell az sch_at_end\n");*/
   }

  } /* if interrupted */
}


/*
**  Function  : void schedule_next_arrival(int     cgen_ident, int	dest_cgen_ident)
**  Return    : none.
**  Remarks   : the next arrival instant is scheduled in the Event List for
**              a call request coming from the call generator identified by
**              'cgen_ident' to the call generator identified by 'dest_cgen_ident'.
*/
void schedule_next_arrival(int     source_cgen_ident, int	dest_cgen_ident)
{
 Event    *next_arrival_event;  /* Event describing the next arrival     */
                                /* from the same call generator.         */
 TimeUnit arrival_interval;     /* Time interval between two arrivals    */
                                /* from the same call generator.         */
 TimeUnit mean_interval;        /* Mean value of the probability density */
                                /* function describing the interarrival  */
                                /* time for call generator 'source_cgen_ident'. */
 double nextend = 1.0E20;

 switch(callgen_vect[source_cgen_ident]->type)
    {
     case CG_TYPE1 :
        mean_interval =
            1.0 / ((CGType1 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE2 :
        mean_interval =
            1.0 / ((CGType2 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE3 :
        mean_interval =
            1.0 / ((CGType3 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE4 :
        mean_interval =
            1.0 / ((CGType4 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE5 :
        mean_interval =
            1.0 / ((CGType5 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE6 :
        mean_interval =
            1.0 / ((CGType6 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE7 :
        mean_interval =
            1.0 / ((CGType7 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
        break;

     case CG_TYPE8 :
	  break;

     case CG_TYPE9 :
        mean_interval =
            1.0 / ((CGType9 *)(callgen_vect[source_cgen_ident]->info))->arrival_rate;
		if(((CGType9 *)(callgen_vect[source_cgen_ident]->info))->
			call_closures->next)
			nextend = ((CGType9 *)(callgen_vect[source_cgen_ident]->info))->
			call_closures->next->value;

        break;

     case CG_TYPE10 :
	printf("\nError:  Function 'schedule_next_arrival' must not be ");
	printf("called for\n        SINK call generators.");
	exit(1);

     case CG_TYPE11 :
	printf("\nError:  Function 'schedule_next_arrival' must not be ");
	printf("called for\n        TCP_SINK call generators.");
	exit(1);

     case CG_TYPE12 :
	printf("\nError:  Function 'schedule_next_arrival' must not be ");
	printf("called for\n        ASON_SINK call generators.");
	exit(1);

     default:
        fprintf(stderr, "ancles.c:schedule_next_arrival(): unknown call generator type\n");
	exit(1);
    }
 if(callgen_vect[source_cgen_ident]->is_low)
    /* reduce the arrival rate */
    mean_interval /= callgen_vect[source_cgen_ident]->load_reduction_factor;

 if(callgen_vect[source_cgen_ident]->type == CG_TYPE8)
 {
	 arrival_interval = uniform_0_1(&sh_seed);
	 fprintf(CG8FILE, "%s\t%s\t%g\n",
		 user_name(callgen_vect[source_cgen_ident]->ref_user->id),
		 user_name(callgen_vect[dest_cgen_ident]->ref_user->id),
		 arrival_interval);
	 arrival_interval = DBL_MIN + DBL_MIN * arrival_interval;
 }
 else arrival_interval = (TimeUnit)negexp(mean_interval, &cg_seed_vect[source_cgen_ident]);

 /* Pascal generator generates arrival only if it would come
 before its next call closure event */
 if(callgen_vect[source_cgen_ident]->type == CG_TYPE9 && 
	 ancles_time + arrival_interval > nextend)
 {
	 /* the next call closure has to perform a (possible) scheduling of arrival *
	 ((CGType9 *)(callgen_vect[source_cgen_ident]->info))->sch_at_end = TRUE;*/
	 return;
 }

 next_arrival_event = event_alloc();
 next_arrival_event->ev_time = ancles_time + arrival_interval;
 next_arrival_event->ev_type = OPENING_CALL_REQUEST;
 next_arrival_event->ev_dscrpt.source_cgen_id = source_cgen_ident;

 if(callgen_vect[source_cgen_ident]->type == CG_TYPE7 ||
	 callgen_vect[source_cgen_ident]->type == CG_TYPE8)
	next_arrival_event->ev_dscrpt_2.dest_cgen_id = dest_cgen_ident;
 else next_arrival_event->ev_dscrpt_2.dest_cgen_id = source_cgen_ident;
 insert_event(next_arrival_event);
}



/*
**  Function  : static void change_state(int     cgen_ident)
**  Return    : none.
**  Remarks   : the function changes load state for the named source, and
**              schedules the next state change.
*/
static void change_state(int     cgen_ident)
{

 Event   *next_change_event;
 TimeUnit changetime;
 double mean_interval;

 if (callgen_vect[cgen_ident]->is_low)
    {
      callgen_vect[cgen_ident]->is_low = FALSE;
      mean_interval = callgen_vect[cgen_ident]->normal_load_duration;
    }
 else
   {
     callgen_vect[cgen_ident]->is_low = TRUE;
     mean_interval = callgen_vect[cgen_ident]->low_load_duration;
   }

 changetime = (TimeUnit)negexp(mean_interval, &cg_seed_vect[cgen_ident]);
 next_change_event = event_alloc();
 next_change_event->ev_type = LOAD_STATE_CHANGE;
 next_change_event->ev_time = ancles_time + changetime;
 next_change_event->ev_dscrpt.source_cgen_id = cgen_ident;
 insert_event(next_change_event);
}


/*
**  Function  : static void update_off_fbw(int     link_ident)
**  Return    : none.
**  Remarks   : updates the official free bandwidth and the number of
**		calls (class U) on the link referred by "link_ident"
**              and schedules the next updating event of this link
*/
static void update_off_fbw(int     link_ident)
{
 Event    *next_updating_event; /* Event, which describes the next 	*/
                                /* update of the same link.		*/

/* Updating the official bandwidth and the number of U class calls	*/
 link_Ucall_stat[link_ident]->official_free_Uband =
  link_Ucall_stat[link_ident]->free_Uband;
 link_Ucall_stat[link_ident]->official_no_of_Ucall =(int)floor(0.5 +
  uniform((1.0 - ((double) bw_measure_uncertainty / 100.0))*
  link_Ucall_stat[link_ident]->no_of_Ucall,
  (1.0 + ((double) bw_measure_uncertainty / 100.0))*
  link_Ucall_stat[link_ident]->no_of_Ucall, &sh_seed));

 if(link_Ucall_stat[link_ident]->unused_free_band -
   capacity * cong_thrsh / link_vect[link_ident]->timescale < DBL_CMP_THS)
  link_Ucall_stat[link_ident]->official_cong = TRUE;
 else
  link_Ucall_stat[link_ident]->official_cong = FALSE;

/* Creating, initializing and scheduling a new updating event	*/
 next_updating_event = event_alloc();
 next_updating_event->ev_type = OFFICIAL_FBW_UPDATE;
 next_updating_event->ev_dscrpt.link_id = link_ident;
 next_updating_event->ev_time =
   ancles_time + link_vect[link_ident]->bw_update_p;
/* Put the new updating event in the event-queue	*/
 insert_event(next_updating_event);
}


/*
**  Function  : void init_dyn()
**  Return    : none.
**  Remarks   : initialize some link indexed vectors used in the
**		calculation of the (estimated) bottleneck bandwidth.
**		The values in the array are set so as if the shortest
**		path algorithm was used in the network.
*/
void init_dyn()
{
 register int i;
 int src, dst;
 SD_pair *curr_sd_pair;

 if (N_link == NULL)
  {
   N_link = (int *)MMmalloc(sizeof(int) * no_of_links, "init_dyn");
   B_link = (double *)MMmalloc(sizeof(double) * no_of_links, "init_dyn");
  }

 for (i=0;i<no_of_links;i++)
  {
   if (link_vect[i]->bw_update_p > 0.0)
    B_link[i]=link_Ucall_stat[i]->official_free_Uband;
   else
    B_link[i]=link_Ucall_stat[i]->free_Uband;

   N_link[i]=0;

   /* skip user-node and node-user links by leaving N_link=0 */
   if (link_Ucall_stat[i]->is_node_node)
    {
   /* for every connection which has the shortest path passing this link */
     for(curr_sd_pair = link_Ucall_stat[i]->sp_on; curr_sd_pair != NULL;
	 curr_sd_pair = curr_sd_pair->next)
      {
       src = curr_sd_pair->src;
       dst = curr_sd_pair->dst;
       N_link[i]+=pair_routing_info[src][dst].actual_callnum;
      }
    }
  }
}

/*
**  Function  : void calc_estim_bottleneck ()
**  Return    : none.
**  Remarks   : calculates the bottleneck bandwidth for every call,
**		assuming the use of SP routing. The values will be
**		registered for every src-dst pairs (if at least one is
**		existing between them).
*/
void calc_estim_bottleneck()
{
 int l, bottleneck;
 int src, dst;
 double min_bw;
 Boolean  more_calls=TRUE;
 LinkList *curr_link;
 SD_pair *curr_sd_pair;

/*Initialization */
 init_dyn();
 for(src=0; src<no_of_cgens; src++)
   for(dst=0; dst<no_of_cgens; dst++)
      pair_routing_info[src][dst].bn_bw_ready = FALSE;

 while (more_calls)
   {
      min_bw = capacity+1;
      /* look for bottleneck link over all links, except of  */
      /* user-node and node-user links and for those links   */
      /* carrying no U calls                                 */
      bottleneck = -1;
      for(l=0; l < no_of_links; l++)
	{
	   if(N_link[l])
	    {
	     if (B_link[l]/(double)N_link[l] < min_bw)
	       {		
		  min_bw = B_link[l]/(double)N_link[l];
		  bottleneck = l;
	       }
	    }
	}

      /* if no suitable U calls were found in the previous cycle, */
      /* exit the while cycle                                     */
      if (min_bw > capacity)
	more_calls = FALSE;
      else	
	{
      /* for all the cgen-pairs using this link (on shortest path) */
	 for(curr_sd_pair = link_Ucall_stat[bottleneck]->sp_on;
	     curr_sd_pair != NULL; curr_sd_pair = curr_sd_pair->next)
	  {
	   src = curr_sd_pair->src;
	   dst = curr_sd_pair->dst;
      /* if the bw is not determined yet */
	   if(pair_routing_info[src][dst].bn_bw_ready != TRUE)
	    {
	      if(min_bw < ((CGType5 *)(callgen_vect[src]->info))->
	       max_max_rate)
	       pair_routing_info[src][dst].Bw_SP = min_bw;
	      else
	       pair_routing_info[src][dst].Bw_SP =
		((CGType5 *)(callgen_vect[src]->info))->max_max_rate;

	      pair_routing_info[src][dst].bn_bw_ready = TRUE;
      /* reduce the traffic on every link of the shortest path */
	     for(curr_link = pair_routing_info[src][dst].spath
	        ->path->next; curr_link->next != NULL; curr_link =
		curr_link->next)
	       {
		N_link[curr_link->link_elem->id] -=
		 pair_routing_info[src][dst].actual_callnum;
		B_link[curr_link->link_elem->id] -=  pair_routing_info[src][dst].Bw_SP *
		 pair_routing_info[src][dst].actual_callnum;
	       }
	    }
	  }
	}
     }
}

/*
**  Function  : double under_estim_bottleneck(int insrc, int indst)
**  Return    : an underestimation of the bottleneck bandwidth
**  Remarks   : the simpliest method is used to estimate the bw for
**		the incoming generator-pair, assuming the use of SP
**		routing.
*/
double under_estim_bottleneck(int insrc, int indst)
{
 double min_bw = capacity+1, actlink_bw = capacity+2;
 LinkList *curr_link;

 init_dyn();
 /* find the strictest link on the shortest path */
 for(curr_link = pair_routing_info[insrc][indst].spath
    ->path->next; curr_link->next != NULL; curr_link =
    curr_link->next)
   {
    if(N_link[curr_link->link_elem->id])
     actlink_bw = B_link[curr_link->link_elem->id] /
      (double)N_link[curr_link->link_elem->id];
  /* if there are no calls on the link -> bn bw is the link itself */
    else
     actlink_bw = B_link[curr_link->link_elem->id];

    if(actlink_bw < min_bw)
      min_bw = actlink_bw;
   }

 if(min_bw < ((CGType5 *)(callgen_vect[insrc]->info))->
    max_max_rate)
  return min_bw;

 return ((CGType5 *)(callgen_vect[insrc]->info))->max_max_rate;

}

/*
**  Function  : static void update_alg(int     source_cgen_id)
**  Return    : none.
**  Remarks   : updates the used algorithms (SP or MD) based on the
**		estimation of the bottleneck bandwidths. The decision is
**		based on filtered values (exponential filtering) with
**		histeresys and threshold.
*/
static void update_alg(int     source_cgen_id)
{
 Event    *next_updating_event; /* Event, which describes the next 	*/
                                /* update of the same link.		*/
 int	j;
 R_info  *act_pri;

/* estimating the bandwidth on the shortest path for the connections
   starting from the given source */
 calc_estim_bottleneck();

 for(j = 0; j < no_of_cgens; j++)
  {
   act_pri = &pair_routing_info[source_cgen_id][j];
   if(act_pri->pac)
     {
    /* validate only the assigned values of real bottleneck-estimation
       for the connections from the given source, if not assigned (i.e.,
       all the links used by the shortest path, were empty), set bw
       to the underestimated value */
      if(!act_pri->bn_bw_ready)
       {
	act_pri->Bw_SP = under_estim_bottleneck(source_cgen_id, j);
       }

      bottleneck_of_MD(source_cgen_id, j);

      act_pri->mu_SP =
	(act_pri->mu_SP * sample_deviation) +
	(act_pri->Bw_SP * (1 - sample_deviation));

      act_pri->mu_MD =
	(act_pri->mu_MD * sample_deviation +
	act_pri->Bw_MD * (1 - sample_deviation));

    /* Choice of the algorithm -> with threshold and histeresis. */
      if(act_pri->mu_SP +
	    act_pri->mu_MD > 0.0)
	{
	 if(2*(act_pri->mu_SP -
	    act_pri->mu_MD) /
	   (act_pri->mu_SP +
	    act_pri->mu_MD) > alg_decision_thrsh + histeresys_thrsh)
	  act_pri->sprou = TRUE;

	 if(2*(act_pri->mu_SP -
	    act_pri->mu_MD) /
	   (act_pri->mu_SP +
	    act_pri->mu_MD) < alg_decision_thrsh - histeresys_thrsh)
	  act_pri->sprou = FALSE;
	 }
      else
	  act_pri->sprou = FALSE;

/*      if(act_pri->sprou)
	RSPstate[source_cgen_id][j]++;
      else
	MDstate[source_cgen_id][j]++;
      act_pri->k++; */

     }
  }

/* Creating, initializing and scheduling a new updating event	*/
 next_updating_event = event_alloc();
 next_updating_event->ev_type = UPDATE_ALG;
 next_updating_event->ev_dscrpt.source_cgen_id = source_cgen_id;
 next_updating_event->ev_time = ancles_time + ac_period;
/* Put the new updating event in the event-queue	*/
 insert_event(next_updating_event);
}


/*
**  Function  : void init_pac ()
**  Return    : none.
**  Remarks   : Initialize the pair routing infos for every s-d pair. If
**		the generator uses the PAC method, updating events will
**		be inserted too.	*/
void init_pac()
{
 double		prev_prob;
 LinkList	*curr_link;
 SD_pair	*actual_sd, *prev_sd;
 R_info		*act_pri;
 Event          *current_event;
 int		i, j, pac_present = 0;

 for(i = 0; i < no_of_cgens; i++)
  {
   for(j = 0; j < no_of_cgens; j++)
    {
     if(!j)
      prev_prob = 0.0;
     else
      prev_prob = destination_prob[callgen_vect[i]->ref_user->id]
	 [callgen_vect[j-1]->ref_user->id];

     act_pri = &pair_routing_info[i][j];

     if(callgen_vect[i]->type != CG_TYPE10 &&
		 callgen_vect[i]->type != CG_TYPE11 &&
		 callgen_vect[i]->type != CG_TYPE12)
	   act_pri->spath = path_mat[(&callgen_vect[i]->
          choosen_routing)->path_order_type][i][j];
	 else
	   act_pri->spath = NULL;


     act_pri->Bw_SP = 0.0;
     act_pri->Bw_MD = 0.0;
     act_pri->mu_SP = 0.0;
     act_pri->mu_MD = 0.0;

     act_pri->sprou = FALSE;
     act_pri->bn_bw_ready = FALSE;
     act_pri->pac = FALSE;

/*     act_pri->k = 0;*/
     act_pri->actual_callnum = 0;

/* We have to take care only about the non zero traffic and */
/* generators using PAC routing. */
     if(destination_prob[callgen_vect[i]->ref_user->id]
	 [callgen_vect[j]->ref_user->id] - prev_prob > 0.0 &&
	 callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA9)
      {
/* pac is enabled only if the real links are used (it does not 
   work with virtual links!) */
		 if(ipowdm_ena)
		 {
			 printf("Error:  If IP over WDM is used the routing algorithm\n\
periodical_ac_routing not allowed!\n");
			 exit(1);
		 }
		 pac_present++;
		 act_pri->pac = TRUE;
      }

/* The list with the sd pairs the shortest path of which contains the link */
/* will be constructed here. */
     if(act_pri->spath)
      {
       for(curr_link = act_pri->spath->path->next; curr_link->next != NULL;
	   curr_link = curr_link->next)
	{
	  prev_sd = link_Ucall_stat[curr_link->link_elem->id]->sp_on;
	  if(prev_sd)
	   {
	    for(actual_sd =
	      link_Ucall_stat[curr_link->link_elem->id]->sp_on->next;
	      actual_sd != NULL; actual_sd = actual_sd->next)
	      prev_sd = actual_sd;

	    actual_sd = (SD_pair *)(MMmalloc(sizeof(SD_pair),
			   "init_pac - set sp_on field of links"));
	    prev_sd->next = actual_sd;
	   }
	  else
	   {
	    actual_sd = (SD_pair *)(MMmalloc(sizeof(SD_pair),
			   "init_pac - set sp_on field of links"));
	    link_Ucall_stat[curr_link->link_elem->id]->sp_on = actual_sd;
	   }
	  actual_sd->src = i;
	  actual_sd->dst = j;
	  actual_sd->next = NULL;
	}
      }
/*     RSPstate[i][j]=0;
     MDstate[i][j]=0;*/

    } /*for j */
  }/*for i*/

/* One event of type UPDATE_ALG is put in the Event List for update the */
/* choice of the algorithm for every Ucall generators (pac method) */
 if(pac_present && ac_period != 0.0)
  {
   for(i = 0; i < no_of_cgens; i++)
    {
      if(source_prob[callgen_vect[i]->ref_user->id] > 0.0 &&
	 callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA9)
        {
	  current_event = event_alloc();
	  current_event->ev_type = UPDATE_ALG;
	  current_event->ev_dscrpt.source_cgen_id = i;
	  current_event->ev_time = ancles_time +
	   uniform(0.0, ac_period, &sh_seed);
	/* Put the updating event in the event-queue			*/
	  insert_event(current_event);
        }
     }
  }
}
/*
**  Function  : static void compute_aggregate_costs(void)
**  Return    : none.
**  Remarks   : updates the aggregate costs of all the network nodes
**		(this cost will be used in the dlcp routings)
**              and schedules the next updating event
*/
static void compute_aggregate_costs(void)
{
 int curr_node;
 register int i;
 register int j;
 double temp_var;
 double aggregate_cost = 0.0;
 Event *next_recomputation;

 for(i=0;i<no_of_nodes;i++)
    {
     for(j=0;j<node_vect[i]->no_of_ports;j++)
        {
         curr_node = node_vect[i]->connected_node[j];
         temp_var = (node_vect[i]->outqueue[j]->free_band>0)? 1.0/node_vect[i]->
                                                    outqueue[j]->free_band:
                                                    MAX_COST;
         aggregate_cost += temp_var + node_vect[curr_node]->beta*
            		    node_vect[curr_node]->aggregate_cost;
        }
     aggregate_cost /= node_vect[i]->no_of_ports;
     new_aggregate_costs[i] = aggregate_cost;
     aggregate_cost = 0.0;
    }
 /* copy new computed values in the node structures */
 for(i=0;i<no_of_nodes;i++)
     node_vect[i]->aggregate_cost = new_aggregate_costs[i];

 next_recomputation = event_alloc();
 next_recomputation->ev_type = COMPUTE_AGGREGATE_COSTS;
 next_recomputation->ev_time = ancles_time + recomputation_slot_duration;
 insert_event(next_recomputation);
}


/*
**  Function  : void create_virtual_topology(int)
**  Return    : none.
**  Remarks   : creates a predefined virtual topology
**				according to the input parameter
*/
void create_virtual_topology(int topo)
{
	int i, j, k, l, m, n, vlconn;
	int nodeconn[MAX_NODE][MAX_NODE]; /* node-node connection */
	int intocnodes[MAX_NODE]; /* nodes to put in the spantree */

	int indep[MAX_NODE][MAX_NODE]; /* independent nodes in the actual spantree fragments */

	int nodecg8[MAX_NODE]; /* a genarator 8 in the node */
	int nodecg12[MAX_NODE]; /* a genarator 12 in the node */

	/* arrays to describe the distances (with random factor) of node-pairs */
	int candidate_from[MAX_NODE*MAX_NODE];
	int	candidate_to[MAX_NODE*MAX_NODE];
	double candidate_dist_rnd[MAX_NODE*MAX_NODE];

	/* arrays and variables used to order the ebove arrays (via index-references) */
	int	not_found_cand[MAX_NODE*MAX_NODE];
	int ord_cand[MAX_NODE*MAX_NODE];
	double min_dr;
	int candid_num, min_cand;

	if(topo == NONE) return;

	for(i=0; i< no_of_nodes; i++)
		for(j=0; j< no_of_nodes; j++)
			nodeconn[i][j] = 0;

	switch(topo)
	{
	case PHYSICAL:
/* Molto problematico!!! Meglio non usare ancora... */
		
		/* require one virtual link for each optical link */
		for(i=0; i< no_of_nnlinks; i++)
		{
			nodeconn[link_vect[nnlink_vect[i]]->source]
			[link_vect[nnlink_vect[i]]->dest]++;
			nodeconn[link_vect[nnlink_vect[i]]->dest]
			[link_vect[nnlink_vect[i]]->source]++;
		}
		break;
	case SPANTREE:
/* Molto problematico!!! */

		/* find a spanning tree */
		/* in this function we need to reach each cg8/cg3 generators,
			and only links among this subset of nodes can be created,
			since in the normal OXC-s may no splitting/adding/
			wavelength converting functions are available */
		for(i=0; i< no_of_nodes; i++)
			intocnodes[i] = 0;

		for(i=0; i< no_of_nodes; i++)
			for(j=0; j< no_of_nodes; j++)
				if(i==j)
					indep[i][j] = 0;
				else
					indep[i][j] = 1;

		/* only the nodes containing a data-level generator or sink
			are required to be inserted in the tree */
		for(i = 0; i < no_of_cgens; i++)
		{
			/* for each data-level generator */ 
			if(callgen_vect[i]->type<=CG_TYPE5) 
			{
				k = callgen_vect[i]->ref_user->connected_node;
				/* nonzero data level traffic source */
				if(source_prob[callgen_vect[i]->ref_user->id]>0.0)
					intocnodes[k] = 7;
				for(j = 0; j < no_of_cgens; j++)
				{
					/* seek the nonzero traffic sinks */
					if(traff_pattern[callgen_vect[i]->ref_user->id]
						[callgen_vect[j]->ref_user->id] > 0.0)
						intocnodes[callgen_vect[j]->ref_user->connected_node] = 7;
				}
			}
		}

		/* Find the assigned generators for the interesting nodes */
		for(i = 0; i < no_of_cgens; i++)
		{
			if(callgen_vect[i]->type==CG_TYPE8 &&
				intocnodes[callgen_vect[i]->ref_user->connected_node]>4)
			{
				intocnodes[callgen_vect[i]->ref_user->connected_node]-=4;
				nodecg8[callgen_vect[i]->ref_user->connected_node]=i;
			}
			if(callgen_vect[i]->type==CG_TYPE12 &&
				(intocnodes[callgen_vect[i]->ref_user->connected_node]%4)>2)
			{
				intocnodes[callgen_vect[i]->ref_user->connected_node]-=2;
				nodecg12[callgen_vect[i]->ref_user->connected_node]=i;
			}
		}

		/* Create an ordered linked list of the possible links 
			between the interesting nodes */

		k = 0;
		for(i=0; i< no_of_nodes; i++)
		{
			if(intocnodes[i])
			{
				if(intocnodes[i]>1)
				{
					printf("Inconsistent node description in node %s!\n",node_name(i));
					exit(0);
				}
				for(j=i+1; j< no_of_nodes; j++)
				{
					if(intocnodes[j])
					{
						/* inordered arrays, containing information of candidates */
						candidate_from[k] = i;
						candidate_to[k] = j;
						candidate_dist_rnd[k] = path_mat[IN_ORDER_OF_HOP_NUMBER]
							[nodecg8[i]][nodecg12[j]]->no_of_hops + 
						/* randomise the place among the equidistatnt candidates */
							uniform_0_1(&sh_seed);
						not_found_cand[k]=1;
						k++;
					}
				}
				
			}
		}

		candid_num = k;

		/* ordering the array with a helper array */
		for(i=0; i < candid_num; i++)
		{
			min_dr = no_of_nodes+1;
			for(j=0; j < candid_num; j++)
			{
				/* find the next element of ordered array */
				if(not_found_cand[j] && candidate_dist_rnd[j] < min_dr)
				{
					min_cand = j;
					min_dr = candidate_dist_rnd[j];
				}
			}
			/* insert in the ordered array */
			ord_cand[i]=min_cand;
			not_found_cand[min_cand] = 0;
		}

		/* find minimal spanning tree */
		for(i=0; i < candid_num; i++)
		{
			
			k = candidate_from[ord_cand[i]];
			l = candidate_to[ord_cand[i]];
			/* if no circle would establish */
			if(indep[k][l])
			{
				nodeconn[k][l]++;
				nodeconn[l][k]++;
				/* nodes in the spanning tree */
				for(m=0; m<no_of_nodes; m++)
					for(n=0; n<no_of_nodes; n++)
						if(m!=n && !indep[k][m] && !indep[l][n])
						{
						/* update the dependencies */
							indep[m][n] = 0;
							indep[n][m] = 0;
						}
			}
		}
		break;
	case FULLMESH:
		for(i = 0; i < no_of_cgens; i++)
		{
			/* for each data-level generator */ 
			if(callgen_vect[i]->type<=CG_TYPE5)
			{
				k = callgen_vect[i]->ref_user->connected_node;
				for(j = 0; j < no_of_cgens; j++)
				{
					/* for the nonzero traffics */
					if(traff_pattern[callgen_vect[i]->ref_user->id]
						[callgen_vect[j]->ref_user->id] > 0.0)
						nodeconn[k][callgen_vect[j]->ref_user->connected_node]++;
				}
			}
		}
		break;
	}

	/* knowing the node-node connections we can identify and create
	   virtual links, allowing only the above calculated number per node-node */
	for(i = 0; i < no_of_cgens; i++)
	{
		if(callgen_vect[i]->type==CG_TYPE8)
		{
			k = callgen_vect[i]->ref_user->connected_node;
			for(j = 0; j < no_of_cgens; j++)
			{
				if(callgen_vect[j]->type==CG_TYPE12)
				{
					l = callgen_vect[j]->ref_user->connected_node;
					if(nodeconn[k][l])
					{
						vlconn = nodeconn[k][l];
	/* avoid duplication because of multiplied cg8 connected to the same node */
						nodeconn[k][l] = 0;
	/* if duplex optical channels are established, from only one direction we
	   we need to open the virtual links */
						if(!duplex_llc || (k < l))
						{
							for(; vlconn>0; vlconn--)
							/* now simple create the links */
								schedule_next_arrival(i,j);
						}
					}
				}
			}
		}
	}
}


/*
**  Function  : int main(int argc, char *argv[])
**  Return    : 0 if no errors occurred during the program execution.
**  Remarks   : none.
*/
int main(int argc, char *argv[])
{
 int            i,j;
 char           *s;

 Event          *current_event;

 char           output_file[80],
                trf_filename[80];

 extern void initialize_parameters(void);
 extern FILE *yyin;
 extern Boolean autolink;

 Boolean got_filename,
         got_trffile,
         got_outfile,
         max_time_set,
         load_defined;

/* The variables used for the statistical analysis are defined and      */
/* initialized here. The meaning and definition of each variable can be */
/* found in the "Statistics User Manual".                               */

 void    (*f_vect1[MAX_STAT_VAR])();    /* Pointers to the functions    */
                                        /* running the actions needed   */
                                        /* when the initial transient   */
                                        /* is over.                     */
 void    (*f_vect2[MAX_STAT_VAR])();    /* Pointers to the functions    */
                                        /* running the actions needed   */
                                        /* when the stationary analysis */
                                        /* is over.                     */
 Boolean significant[MAX_STAT_VAR];     /* Array specifying the type of */
                                        /* the observed variables (TRUE */
                                        /* for significant variables,   */
                                        /* FALSE otherwise).            */
 int     index_of_stat_vars;            /* Index used to scan the vector*/
                                        /* of the observed variables.   */
 Boolean dlcp_used = FALSE;		/* Are there any dlcp-users in the	*/
 					/* network?			*/
 Boolean ewl_used = FALSE;		/* Are there any ewl-users in the	*/
 					/* network?			*/
 Boolean opaque_present = FALSE;	/* Are there any opaque nodes	*/
 					/* in the network?		*/
 int	od_routing;

/* Variables for actual values in cycle bodies */
 Link_ucall_stat  *actual_lUs;
 CallGen   *actual_cv;

/* The heading of ANCLES is printed here.       */
 printf("\n      ANCLES - ATM Networks Call LEvel Simulator - ");
 printf("V. %s\n\n Copyright (C) 1995/96 - Politecnico di Torino ", VERSION);
 printf("- Dipartimento di Elettronica\n");
 printf(" ASON features: Budapest UTE - Department Of Telecommunication");
 printf(" - 2001\n");
 printf(" IP over ASON features: Budapest UTE - University of Trento");
 printf(" - 2003\n");
 printf(" Copyright Limited for Academic Applications Only");
 printf(" \n\n");

/* Initialize the flags                                                 */
 UCall_present = FALSE;
 nonASON_Call_present = FALSE;

/* Initialization of the variables referring to the network entities.   */
 no_of_type1_cg = 0;
 no_of_type2_cg = 0;
 no_of_type3_cg = 0;
 no_of_type4_cg = 0;
 no_of_type5_cg = 0;
 no_of_type6_cg = 0;
 no_of_type7_cg = 0;
 no_of_type8_cg = 0;
 no_of_type9_cg = 0;

/* Initialization of the absolute simulation time.                      */
 ancles_time = 0.0;

/* Initialization of the variables involved by the ANCLES command line. */
 init_debug_level = 0;   /* by default: no debug report */
 got_filename = FALSE;
 got_outfile  = FALSE;
 max_time_set = FALSE;
 load_defined = FALSE;
 got_trffile  = FALSE;
 sh_seed      = DEFAULT_SEED;
 od_routing   = -1;


no_of_free_linklist=0;


/* Parsing of the Command line options.                                 */

 while(--argc > 0)
   {
    if((*++argv)[0] == '-')
      {
       s = argv[0] + 1;
       switch(*s)
         {
          case 'h' :         /* help */
             printf("Usage: ancles [options] filename\n");
             printf("Mandatory parameters (if missing, they are ");
             printf("required interactively):\n");
             printf("  -load=<value>    Offered relative load\n");
             printf("  -out=<filename>  Output files name ");
             printf("(without extension)\n");
             printf("  -time=<value>    Maximum simulation time\n");
             printf("Optional parameters:\n");
             printf("  -auto            Automatic declaration ");
             printf("of missing links \n");
             printf("  -h               Shows this message\n");
             printf("  -i=opt           Enables the screen echo during ");
             printf("the network\n");
             printf("                   initialization ");
             printf("(for debugging phase)\n");
	     printf("                   where 'opt' can be any combination of:\n");
             printf("                   n = ND-related output\n");
             printf("                   t = TRAFFIC-related output\n");
             printf("                   s = STATISTICS-related output\n");
             printf("                   r = ROUTING-related output\n");
             printf("                   m = PATH_DISTRIBUTION-related output\n");
             printf("                   a = all output\n");
	     printf("  -seed=<value>    Pseudo-random generator seed\n");
             printf("  -trf=<filename>  TRF (Traffic Relation File) ");
             printf("file name\n");
             printf("  -r               Reminds predefined parameter ");
             printf("names and exits\n");
             printf("  -z               Prints internal ANCLES ");
             printf("limits and exits\n");
             exit(0);

          case 'a' :                    /* The automatic declaration of */
             if(strcmp(s, "auto") == 0) /* missing links is enabled.    */
               {
                autolink = TRUE;
                printf("Autolink enabled\n");
               }
             else
                printf("Warning: Invalid option -%s ignored\n",s);
             break;

          case 'o' :                            /* Declaration of the   */
             if(strncmp(s, "out=", 4) == 0)     /* name radix for the   */
               {                                /* output files.        */
                s += 4;
                strcpy (file_radix, s) ;
                printf("File Radix %s\n", file_radix);
                got_outfile = TRUE;
               }
             else
               printf("Warning: Invalid option -%s ignored\n", s);
             break;

          case 'i' :                    /* The printing of debug        */
	    ++s;                       /* Skip the '=' sign */
	    while (*++s)
	      switch (*s)
		{
		case 'n':
		case 'N':
		  init_debug_level |= ND_DEBUG_LEVEL;
		  break;

		case 't':
		case 'T':
		  init_debug_level |= ND_TRAFFIC_DEBUG_LEVEL;
		  break;

		case 's':
		case 'S':
		  init_debug_level |= STATISTICAL_DEBUG_LEVEL;
		  break;

		case 'r':
		case 'R':
		  init_debug_level |= ROUTING_DEBUG_LEVEL;
		  break;

		case 'm':
		case 'M':
		  init_debug_level |= DEBUG_MEAN_PATH_LENGTH;
		  break;
		
		case 'a':
		case 'A':
		  init_debug_level = ALL_DEBUG_LEVEL;
		  break;

		default:
		  printf("Warning: Invalid flag `%c' for option -i ignored\n", *s);
		  break;
		}

	    if (init_debug_level)
	      printf("Initialization Trace enabled\n"); /* is enabled.   */
	    break;

          case 's' :                         /* The shared seed for the */
             if(strncmp(s, "seed=", 5) == 0) /* pseudo-random generators*/
               {                             /* is set by the user.     */
                s += 5;
                sscanf(s, "%ld", &sh_seed);
                if(sh_seed <= 0)
                  {
                   sh_seed = DEFAULT_SEED;
                   printf("Warning:  Non positive seed ignored\n");
                   printf("          The seed value is : %ld\n", sh_seed);
                  }
               }
             else if(strncmp(s, "selected_routing=", 17)==0)  /* Overdefine the */
               {                           /* routing algorithm of the nd file. */
                s+=17;
                sscanf(s, "%d", &od_routing);
                if(od_routing < 0)
                  {
                   od_routing = -1;
                   printf("Warning:  Invalid routing algorithm ignored\n");
                  }
               }
             else
               {
                printf("Warning: Invalid option -%s ignored\n",s);
                printf("         The seed value is : %ld\n", sh_seed);
               }
             break;

          case 'z' :                    /* Some internal ANCLES         */
             print_internal();          /* information (ANCLES limits)   */
             exit(0);                   /* is printed on the standard   */
                                        /* output.                      */
          case 'r' :                    /* All the predefined parameter */
             remind_parameter_names();  /* names are printed on the     */
             exit(0);                   /* standard output.             */

          case 'l' :                            /* Setting the offered  */
             if(strncmp(s, "load=", 5) == 0)    /* relative load.       */
               {
                s += 5;
                sscanf(s, "%lf", &load);
                printf("Offered Relative Load %f\n", load);
                load_defined = TRUE;
               }
             else
                printf("Warning: Invalid option -%s ignored\n", s);
             break;

          case 't' :                            /* Setting the maximum  */
             if(strncmp(s, "time=", 5) == 0)    /* simulation time.     */
               {
                s += 5;
                sscanf(s, "%lf", &max_time);
                printf("Max Simulation Time = %f\n", max_time);
                max_time_set = TRUE;
                break;
               }
             else if(strncmp(s, "trf=", 4)==0)  /* The use of the named */
               {                                /* TRF file is enabled. */
                s+=4;
                strcpy(trf_filename, s);
                fix_extension(trf_filename, "trf");
                got_trffile = TRUE;
                break;
               }
             else
                printf("Warning: Invalid option -%s ignored\n", s);
             break;
	
          default:
             printf("Warning: Invalid option -%s ignored\n", s);
         }
     }
   else
     {                                  /* No flag: input file name.    */
      s = argv[0] ;
      strcpy (descr_file, s);
      fix_extension(descr_file, "nd");
      got_filename = TRUE;
      printf("Input file %s\n", descr_file);
     }
  }

 if(!got_filename) /* No file name provided: it is interactively asked. */
   {
    printf("Network description file [.nd] : ");
    get_input("%s", descr_file);
    fix_extension(descr_file, "nd");
    printf("%s\n", descr_file);
   }

/* Opening the network description file.                                */
 if((yyin = fopen(descr_file, "r")) == NULL)    /* The given name does  */
    {                                           /* not correspond to an */
     printf("File %s not found\n", descr_file); /* existing file: the   */
     exit(1);                                   /* program execution is */
    }                                           /* immediately ended.   */


/* Initializing the parameters dealing with the parsing of the network  */
/* description file.                                                    */
 initialize_parameters();

/* Parsing the network description file.                                */
 if(yyparse() == 1)
   {
    fclose(yyin);
    printf("Error:  Invalid network description file\n");
    exit(1);
   }
 else
    fclose(yyin);

/* Initializing the seeds used by the call generators.                  */
 for(i = 0; i < no_of_cgens; i++)
        cg_seed_vect[i] = rnd32(sh_seed + i);

/* Allocate and initialize some arrays used in the statistics */
  init_statmem();

/* The relative load offered to the network is interactively asked      */
/* if not previously provided.                                          */
 if(!load_defined)
   {
    printf("Offered Relative Load : ");
    get_input("%lf", &load);
    printf("%f\n", load);
   }

/* Reading the Traffic Relations File (.TRF) (if previously provided).  */
 if(got_trffile)
   {
    printf("Reading TRF file: %s\n",trf_filename);
    if(!read_traff_relations(trf_filename))
      {
       printf("Error:  Invalid or not existing traffic relation file\n");
       exit(1);
      }
   }
 else
   {                                       /* If a TRF file has not been*/
    printf("Creating uniform traffic.\n"); /* provided, a traffic matrix*/
    create_uniform_traffic();              /* in which the load is      */
   }                                       /* equally divided among the */
                                           /* users is created.         */

/* The name of the output files (without extension) is asked here if    */
/* not previously provided.                                             */
 if(!got_outfile)
   {
    printf("Output files name (without extension) : ");
    get_input("%s", file_radix);
    printf("%s\n", file_radix);
   }
 strcpy(output_file, file_radix);
 strcat(output_file, ".str");
 if((fp_statistics = fopen(output_file, "w")) == NULL)
   {
    printf("\nError:  the file %s cannot be opened.\n", output_file);
    exit(1);
   }

/* Analysis of the settings for the global statistical variables.	*/
 index_of_stat_vars = 0;

 for(i = 0; i < 3; i++)
    global_stat_vars[i].number = NOT_USED_ID;
 for(i = 0; i < MAX_LINK; i++)
    link_stat_vars[i].number = NOT_USED_ID;
 for(i = 0; i < MAX_LINK; i++)
    ason_link_stat_vars[i].number = NOT_USED_ID;
 for(i = 0; i < MAX_NET_USER; i++)
    user_stat_vars[i].number = NOT_USED_ID;
 for(i = 0; i < MAX_NET_USER; i++)
    classU_user_stat_vars[i].number = NOT_USED_ID;

/* Checking for users not eligible for statistical analysis: 		*/
/* Class U Users, SINK, TCP_SINK and ASON_SINK Users and Idle Users.		*/
 for(i = 0; i < no_of_cgens; i++)
   {
    if (callgen_vect[i]->type == CG_TYPE5)
		/* allow printing stats for U call       */
		UCall_present = TRUE;                           

    if (callgen_vect[i]->type < CG_TYPE6)
		nonASON_Call_present = TRUE;                           
    
	if( (user_stat_vars[callgen_vect[i]->ref_user->id].info != NOT_ANALYSED)
	&& (callgen_vect[i]->type != CG_TYPE8)
	&& ((callgen_vect[i]->type == CG_TYPE5) || 	/* Class U	*/
	    (callgen_vect[i]->type == CG_TYPE10) || 	/* SINK		*/
	    (callgen_vect[i]->type == CG_TYPE11) || 	/* TCP_SINK	*/
	    (callgen_vect[i]->type == CG_TYPE12) || 	/* ASON_SINK	*/
            (source_prob[callgen_vect[i]->ref_user->id] <= 0.0)) )
	{						/* Idle	User	*/
	 --no_stat_vars;
	 if(user_stat_vars[callgen_vect[i]->ref_user->id].info == SIGNIFICANT)
		--no_sign_vars;
         user_stat_vars[callgen_vect[i]->ref_user->id].info = NOT_ANALYSED;
        }
   }
/* Checking for links not eligible for statistical analysis:		*/
/* User -> Node links and Node -> User links.				*/
 for(i = 0; i < no_of_users; i++)
    {
     if(link_stat_vars[user_vect[i]->in_link->id].info != NOT_ANALYSED)
	{
	 --no_stat_vars;
         if(link_stat_vars[user_vect[i]->in_link->id].info == SIGNIFICANT)
		--no_sign_vars;
	 link_stat_vars[user_vect[i]->in_link->id].info = NOT_ANALYSED;
	}
     if(link_stat_vars[user_vect[i]->out_link->id].info != NOT_ANALYSED)
	{
	 --no_stat_vars;
         if(link_stat_vars[user_vect[i]->out_link->id].info == SIGNIFICANT)
		--no_sign_vars;
	 link_stat_vars[user_vect[i]->out_link->id].info = NOT_ANALYSED;
	}
    }

/* Only the node-node, duplex-optic links can be statistically */
/* analysed from the point of view of the used wavelengths 	*/
 for(i = 0; i < MAX_LINK; i++)
    {
     if(ason_link_stat_vars[i].info != NOT_ANALYSED &&
        !link_vect[i]->duplex_optic)
	{
	 --no_stat_vars;
         if(ason_link_stat_vars[i].info == SIGNIFICANT)
		--no_sign_vars;
	 ason_link_stat_vars[i].info = NOT_ANALYSED;
	}
    }
 for(i = 0; i < no_of_users; i++)
    {
     if(ason_link_stat_vars[user_vect[i]->in_link->id].info != NOT_ANALYSED)
	{
	 --no_stat_vars;
         if(ason_link_stat_vars[user_vect[i]->in_link->id].info == SIGNIFICANT)
		--no_sign_vars;
	 ason_link_stat_vars[user_vect[i]->in_link->id].info = NOT_ANALYSED;
	}
     if(ason_link_stat_vars[user_vect[i]->out_link->id].info != NOT_ANALYSED)
	{
	 --no_stat_vars;
         if(ason_link_stat_vars[user_vect[i]->out_link->id].info == SIGNIFICANT)
		--no_sign_vars;
	 ason_link_stat_vars[user_vect[i]->out_link->id].info = NOT_ANALYSED;
	}
    }

 /* Class U */
 for(i = 0; i < no_of_cgens; i++)
   {
     if (classU_user_stat_vars[callgen_vect[i]->ref_user->id].info != NOT_ANALYSED
	 && (callgen_vect[i]->type != CG_TYPE5
	     || source_prob[callgen_vect[i]->ref_user->id] <= 0.0))
       {
	 --no_stat_vars;
	 if(classU_user_stat_vars[callgen_vect[i]->ref_user->id].info == SIGNIFICANT)
	   --no_sign_vars;
	
	 classU_user_stat_vars[callgen_vect[i]->ref_user->id].info = NOT_ANALYSED;
       }
   }

 if(no_stat_vars > MAX_STAT_VAR)
   {
    printf("\nError:  The number of analysed processes exceeds\n");
    printf("        the maximum permitted value.\n");
    printf("        Reduce the number of processes under statistical ");
    printf ("analysis.\n");
    exit(1);
   }
 if(no_sign_vars > MAX_SIGNIF_VAR)
   {
    printf("\nError:  The number of analysed significant processes exceeds\n");
    printf("        the maximum permitted value.\n");
    printf("        Reduce the number of significant processes under ");
    printf ("analysis.\n");
    exit(1);
   }
 if(no_sign_vars < 1)
   {
    printf("\nError:  There must be at least one significant process.\n");
    printf("        Declare a stochastic process under analysis ");
    printf("as significant.\n");
    exit(1);
   }

/* Setting the status of 'total carried load'.	*/
 if (global_stat_vars[0].info != NOT_ANALYSED)
   {
     global_stat_vars[0].number = index_of_stat_vars;
     names_stat_vars[index_of_stat_vars] =
       (char *)(MMmalloc(sizeof(char[50]), "main - total carried load"));

     sprintf(names_stat_vars[index_of_stat_vars], "total carried load");
     if(global_stat_vars[0].info == NOT_SIGNIFICANT)
       {
	 significant[index_of_stat_vars] = FALSE;
	 f_vect1[index_of_stat_vars]     = NULL;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
     else    /* global_stat_vars[0].info == SIGNIFICANT                  */
       {
	 significant[index_of_stat_vars] = TRUE;
	 f_vect1[index_of_stat_vars]     = reset;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
   }

/* Setting the status of 'total call blocking probability'.	*/
 if (global_stat_vars[1].info != NOT_ANALYSED)
   {
     global_stat_vars[1].number = index_of_stat_vars;
     names_stat_vars[index_of_stat_vars] =
       (char *)(MMmalloc(sizeof(char[50]),
			 "main - total call blocking prob."));

     sprintf(names_stat_vars[index_of_stat_vars], "total blocking probability");
     if(global_stat_vars[1].info == NOT_SIGNIFICANT)
       {
	 significant[index_of_stat_vars] = FALSE;
	 f_vect1[index_of_stat_vars]     = NULL;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
     else    /* global_stat_vars[1].info == SIGNIFICANT                  */
       {
	 significant[index_of_stat_vars] = TRUE;
	 f_vect1[index_of_stat_vars]     = reset;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
   }

/* Setting the status of 'used wavelengths'.	*/
 if (global_stat_vars[2].info != NOT_ANALYSED)
   {
     global_stat_vars[2].number = index_of_stat_vars;
     names_stat_vars[index_of_stat_vars] =
       (char *)(MMmalloc(sizeof(char[50]),
			 "main - used wavelengths"));

     sprintf(names_stat_vars[index_of_stat_vars], "used wavelengths");
     if(global_stat_vars[2].info == NOT_SIGNIFICANT)
       {
	 significant[index_of_stat_vars] = FALSE;
	 f_vect1[index_of_stat_vars]     = NULL;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
     else    /* global_stat_vars[2].info == SIGNIFICANT                  */
       {
	 significant[index_of_stat_vars] = TRUE;
	 f_vect1[index_of_stat_vars]     = reset;
	 f_vect2[index_of_stat_vars++]   = NULL;
       }
   }

/* Analysis of the settings for the user statistical variables            */
 for (i = 0; i < no_of_users; i++)
   {
     if(user_stat_vars[i].info != NOT_ANALYSED)
       {
	 user_stat_vars[i].number = index_of_stat_vars;
	 names_stat_vars[index_of_stat_vars] =
	   (char *)(MMmalloc(sizeof(char[50]),
			     "main - user statistical an. var."));

	 sprintf(names_stat_vars[index_of_stat_vars],
		 "blocking probability for user ");
	 strcat(names_stat_vars[index_of_stat_vars], user_name(i));
	 if(user_stat_vars[i].info == NOT_SIGNIFICANT)
           {
	     significant[index_of_stat_vars] = FALSE;
	     f_vect1[index_of_stat_vars]     = NULL;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
	 else    /* user_stat_vars[i].info == SIGNIFICANT                  */
           {
	     significant[index_of_stat_vars] = TRUE;
	     f_vect1[index_of_stat_vars]     = reset;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
       }
   }

/* Analysis of the settings for the link statistical variables.	*/
 for (i = 0; i < no_of_links; i++)
   {
     if(link_stat_vars[i].info != NOT_ANALYSED)
       {
	 link_stat_vars[i].number = index_of_stat_vars;
	 names_stat_vars[index_of_stat_vars] =
	   (char *)(MMmalloc(sizeof(char[50]),
			     "main - link statistical an. var."));
	 sprintf(names_stat_vars[index_of_stat_vars],
		 "carried load for link ");
	 strcat(names_stat_vars[index_of_stat_vars], link_name(i));
	 if(link_stat_vars[i].info == NOT_SIGNIFICANT)
           {
	     significant[index_of_stat_vars] = FALSE;
	     f_vect1[index_of_stat_vars]     = NULL;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
	 else    /* link_stat_vars[i].info == SIGNIFICANT                  */
           {
	     significant[index_of_stat_vars] = TRUE;
	     f_vect1[index_of_stat_vars]     = reset;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
       }
   }

/* Analysis of the settings for the ason link statistical variables.*/
 for (i = 0; i < no_of_links; i++)
   {
     if(ason_link_stat_vars[i].info != NOT_ANALYSED)
       {
	 ason_link_stat_vars[i].number = index_of_stat_vars;
	 names_stat_vars[index_of_stat_vars] =
	   (char *)(MMmalloc(sizeof(char[50]),
			     "main - ason link statistical an. var."));
	 sprintf(names_stat_vars[index_of_stat_vars],
		 "used wavelengths for link ");
	 strcat(names_stat_vars[index_of_stat_vars], link_name(i));
	 if(ason_link_stat_vars[i].info == NOT_SIGNIFICANT)
           {
	     significant[index_of_stat_vars] = FALSE;
	     f_vect1[index_of_stat_vars]     = NULL;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
	 else    /* ason_link_stat_vars[i].info == SIGNIFICANT                  */
           {
	     significant[index_of_stat_vars] = TRUE;
	     f_vect1[index_of_stat_vars]     = reset;
	     f_vect2[index_of_stat_vars++]   = NULL;
           }
       }
   }

/* Class U */
 for (i = 0; i < no_of_users; i++)
   {
     if (classU_user_stat_vars[i].info != NOT_ANALYSED)
       {
	 classU_user_stat_vars[i].number = index_of_stat_vars;
	 names_stat_vars[index_of_stat_vars] =
	   (char *)(MMmalloc(sizeof(char[50]),
			     "main - user statistical an. var."));

	 sprintf(names_stat_vars[index_of_stat_vars],
		 "assigned bw for class U user ");
	 strcat(names_stat_vars[index_of_stat_vars], user_name(i));
	 if(classU_user_stat_vars[i].info == NOT_SIGNIFICANT)
	   {
	     significant[index_of_stat_vars] = FALSE;
	     f_vect1[index_of_stat_vars]     = NULL;
	     f_vect2[index_of_stat_vars]   = NULL;
	   }
	 else    /* user_stat_vars[i].info == SIGNIFICANT                  */
	   {
	     significant[index_of_stat_vars] = TRUE;
	     f_vect1[index_of_stat_vars]     = reset;
	     f_vect2[index_of_stat_vars]   = NULL;
	   }

	 ++index_of_stat_vars;
       }
   }

/* The names of the variables that are object of complete statistical   */
/* analysis (mean value estimation with its confidence interval and     */
/* variance estimation) are printed out if the -i switch was included   */
/* in the ANCLES command line.                                          */
 if(init_debug_level & STATISTICAL_DEBUG_LEVEL)
        {
         fprintf(stderr, "\n --- ANCLES STATISTICAL VARIABLES --- \n");
         for(i = 0; i < no_stat_vars; i++)
		{
                 fprintf(stderr, "\n%s", names_stat_vars[i]);
		 if(significant[i] == TRUE)
			fprintf(stderr, " (significant)");
		}
         fprintf(stderr, "\n\n");
        }

 /* Initializing the entities required by the statistical module.        */
 printf("Initializing the entities required by the statistical module...\n");
 init_statistics(no_stat_vars, significant, names_stat_vars,
                 f_vect1, f_vect2, stop_simulating, fp_statistics);

/* Initializing the statistical parameters.                             */
 printf("Initializing the statistical parameters...\n");
 init_variables_stat();

/* Initializing the indicator of the first time execution for the       */
/* 'reset()' of ANCSTAT.C. This function is executed every time a       */
/* statistical variable exits its initial transient.                    */
 first_time = TRUE;

/* Initializing the indicators of the closure of the initial transient  */
/* for the users' and links' statistical variables.                     */
 for(i = 0; i < 3; i++)
     global_trans_closed[i] = FALSE;
 for(i = 0; i < MAX_NET_USER; i++)
     user_trans_closed[i] = FALSE;
 for(i = 0; i < MAX_LINK; i++)
     link_trans_closed[i] = FALSE;
 for(i = 0; i < MAX_LINK; i++)
     ason_link_trans_closed[i] = FALSE;
 for(i = 0; i < MAX_NET_USER; i++)
     classU_user_trans_closed[i] = FALSE;

/* Releasing the memory areas used by the parser.                       */
 release_memory();

/* The maximum simulation time is interactively asked if not previously */
/* provided.                                                            */
 if(!max_time_set)
   {
    printf("Max Simulation Time : ");
    get_input("%lf", &max_time);
    printf("%f\n", max_time);
   }

/* The factor to be used in order to obtain the call arrival rate       */
/* provided by each call generator on the basis of the Offered Relative */
/* Load is evaluated here.                                              */
 load_conv_fact = eval_load_conv_fact();

/* The 'arrival_rate' field of each call generator is evaluated here,   */
/* on the basis of the selected Relative Offered Load.                  */
/* If a user has been assigned a null load contribution in the .TRF     */
/* file, no arrival rate is evaluated for its call generator.           */
 for(i = 0; i < no_of_cgens; i++)
         if( (source_prob[callgen_vect[i]->ref_user->id] > 0.0) &&
	     (callgen_vect[i]->type != CG_TYPE10) &&
	     (callgen_vect[i]->type != CG_TYPE11) &&
	     (callgen_vect[i]->type != CG_TYPE12) )
                	set_arrival_rate(callgen_vect[i]);

/* The functions pointed by 'init_CAC' and 'exe_CAC' are deduced from   */
/* the Call Admission Control algorithm selection.                      */
 switch(selected_cac)
    {
     case CALL_ADM_CTRL1 :
        init_CAC = init_CAC1;
        exe_CAC = exe_CAC1;
        break;

     case CALL_ADM_CTRL2 :
        init_CAC = init_CAC2;
        exe_CAC = exe_CAC2;
        break;

     case CALL_ADM_CTRL3 :
        init_CAC = init_CAC3;
        exe_CAC = exe_CAC3;
        break;

     case CALL_ADM_CTRL4 :
        init_CAC = init_CAC4;
        exe_CAC = exe_CAC4;
        break;
    }

/* Initializing the Event, Call and CallList lists.                     */
 init_event_list();
 init_call_list();
 init_call_list_on_link();

/* The entities involved by the selected routing algorithm are  */
/* initialized here.                                            */
 if(od_routing != -1)
 {
	 for(i = 0; i < no_of_cgens; i++)
	 	callgen_vect[i]->choosen_routing.selected_routing = od_routing;

	 printf("Overdefinition of routing algorithm given in the\n");
	 printf("\tnetwork description file. Default selected routing: %d\n",od_routing);
 }
 printf("Initializing the entities involved by the selected routing algorithm...\n");
 init_routing();


/* The entities involved by the selected CAC algorithm are      */
/* initialized here.                                            */
 printf("Initializing the entities involved by the selected CAC algorithm...\n");
  init_CAC();

#ifdef SAVE_THRESHOLDS
 save_link_threshold();
#endif

/* The Loads caused on the node-to-node network Links by primary path  	*/
/* assignment and the consequent Trunk Reservation Thresholds are       */
/* printed out if the -i switch was included in the ANCLES command line */
 if(init_debug_level & ROUTING_DEBUG_LEVEL)
   {
    fprintf(stderr, "The number of sets is %d\n", no_of_thresholds);
    for(i = 0; i < no_of_links; i++)
      {
       fprintf(stderr, "\nLink %d\t- Load : %.3f Mbps\t",
                  i, link_primary_load[i]);
       fprintf(stderr, "Thresholds : {");
       for(j = 0; j < (no_of_thresholds-1); j++)
          fprintf(stderr, "%.3f, ", link_vect[i]->routing_threshold[j]);
       fprintf(stderr, "%.3f } Mbps", link_vect[i]->routing_threshold[j]);
      }
    fprintf(stderr, "\n");
   }

/* Initializing the interrupt flag.					*/
 interrupt_happened = FALSE;	

 init_bott_staticvars();

/* Initializing the realized events counter.                            */
 real_events_cnt = 0;

/* Initializing the accumulator for the number of events currently      */
/* scheduled.                                                           */
 sched_events_acc = 0;
 sched_events_cnt = 0;
 no_of_FES_operations = 0;

/* If needed, open the file for the permanent calls opening times */
 CG8FILE = NULL;
 if(no_of_type8_cg)
 {
	 strcpy(output_file, file_radix);
	 strcat(output_file, ".cg8");
	 if((CG8FILE = fopen(output_file, "w")) == NULL)
	 {
		 printf("\nError:  the file %s cannot be opened.\n", output_file);
		 exit(1);
	 }
	 fprintf(CG8FILE, "# Predefined permanent calls scheduling times\n");
	 fprintf(CG8FILE, "# Real value: DBL_MIN + DBL_MIN * Factor\n");
	 fprintf(CG8FILE, "# DBL_MIN = %g\n", DBL_MIN);
	 fprintf(CG8FILE, "# <From\tTo\tFactor>\n");
 }
 
 
/* One event of type OPENING_CALL_REQUEST is put in the Event List for  */
/* each call generator with positive load contribution in order to      */
/* start the simulation.                                                */

 for(i = 0; i < no_of_cgens; i++)
	{
	 actual_cv = callgen_vect[i];
	 if(source_prob[actual_cv->ref_user->id] > 0.0 &&
		 (actual_cv->type != CG_TYPE7) &&
		 (actual_cv->type != CG_TYPE8) &&
		 (actual_cv->type != CG_TYPE10) &&
		 (actual_cv->type != CG_TYPE11) &&
		 (actual_cv->type != CG_TYPE12) )
		 {
			 schedule_next_arrival(i, i);
			 if (actual_cv->is_nonstationary)
			 {
				 /* set the initial generator state to low if it has to start  */
				 /* in high state and viceversa (change_state changes the state*/
				 /* when it is called)                                         */
				 if (uniform_0_1(&cg_seed_vect[i]) *
					 (actual_cv->normal_load_duration +
					 actual_cv->low_load_duration) >
					 actual_cv->normal_load_duration)
					 actual_cv->is_low = FALSE;
				 else
					 actual_cv->is_low = TRUE;
				 change_state(i);
			 }
			 else
				 actual_cv->is_low = FALSE;
		 }
		 else if((actual_cv->type == CG_TYPE7 && 
			 source_prob[actual_cv->ref_user->id] > 0.0) || 
			 (actual_cv->type == CG_TYPE8))
		 {
			 int prev = 0;
			 double k;
			 int now;
			 
			 if (actual_cv->is_nonstationary)
			 {
				 /* set the initial generator state to low if it has to start  */
				 /* in high state and viceversa (change_state changes the state*/
				 /* when it is called)                                         */
				 if (uniform_0_1(&cg_seed_vect[i]) *
					 (actual_cv->normal_load_duration +
					 actual_cv->low_load_duration) >
					 actual_cv->normal_load_duration)
					 actual_cv->is_low = FALSE;
				 else
					 actual_cv->is_low = TRUE;
				 change_state(i);
			 }
			 else
				 actual_cv->is_low = FALSE;
			 
			 if(actual_cv->type == CG_TYPE7)
				 now = ((CGType7 *)actual_cv->info)->no_of_waves;
			 else
				 now = ((CGType8 *)actual_cv->info)->no_of_waves;
			 if(now > 0)
			 {
				 for(j = 0; j < no_of_cgens; j++)
				 {
					 if((k = destination_prob[callgen_vect[i]->ref_user->id]
						 [callgen_vect[j]->ref_user->id] * now - prev) > 0.5)
					 {
						 prev += (int)floor(k + 0.5);
						 for(; k > 0.5; k -= 1.0)
						 {  
							 schedule_next_arrival(i,j); 
						 }
						 /* if(fabs(k) > PRECISION)
						 {
						 fprintf(stderr, "Error: the given traffic relations \
							 for call generator %d can not be realized with %d wavelengths.\n", i, now);
						 exit(1);
						 }*/
					 }

				/* If the waves can not be distributed accirding to the
				   pattern -> redefine them! 
					 prev += (int)floor((traff_pattern[callgen_vect[i]->ref_user->id]
					 [callgen_vect[j]->ref_user->id]/
					 source_prob[callgen_vect[i]->ref_user->id]) *
						 now + 0.5);
				 }
				 if(prev != now)
				 printf("The traffic of call generator %d  could be transported\
					 over %d wavelengths instead of %d.\n", i, prev, now);

				 if(actual_cv->type == CG_TYPE7)
					 ((CGType7 *)actual_cv->info)->no_of_waves = prev;
				 else
					 ((CGType8 *)actual_cv->info)->no_of_waves = prev;

				 for(j = 0; j < no_of_cgens; j++)
				 {
					 for(k = floor((traff_pattern[callgen_vect[i]->ref_user->id]
					 [callgen_vect[j]->ref_user->id]/
					 source_prob[callgen_vect[i]->ref_user->id]) *
						 now + 0.5); k > 0.0; k -= 1.0)
						 schedule_next_arrival(i,j); */
				 }


			 }
			 else
			 {
				 double actual_capacity = ((CGType7 *)actual_cv->info)->saturation * 
					 wavelength_capacity;
				 int n;
				 
				 for(j = 0; j < no_of_cgens; ++j)
				 {
					 now = (int)ceil(load_conv_fact * load * traff_pattern
						 [callgen_vect[i]->ref_user->id]
						 [callgen_vect[j]->ref_user->id] / 
						 actual_capacity);
					 ((CGType7 *)actual_cv->info)->no_of_waves += now;
					 for(n = 0; n < now; ++n)
						 schedule_next_arrival(i,j);
				 }
				 if(actual_cv->type == CG_TYPE7)
					printf("The traffic of call generator %d  could be transported\
					 over %d wavelengths.\n", i, ((CGType7 *)actual_cv->info)->no_of_waves);
				 else;
			 }				
		 }
	}


 if(CG8FILE)
 {
/* Create predefined virtual topology if required */
     create_virtual_topology(precreav_topology);
 
/* this type of statistics is not used
 fprintf(CG8FILE, "\n#Open and close permanent calls\n\
#<Open/Close/Refused\tFrom\tTo\tTime\tPath no/Start time/ - >\n");*/
	 fclose(CG8FILE);
 }

/* If DLCP is used compute the aggregate costs for dlcp algorithm	*/
/* queue an event of type  after <dlcp_recomputation_slot> secs		*/
/* If equal wavelength 1+1 protection is used, no opaque nodes are  */
/* enabled! */

 for(i = 0; i < no_of_cgens; i++)
   {
   /* If there is at least one user uses dlcp routing, the aggregation
      functionality must work */
    if(callgen_vect[i]->choosen_routing.selected_routing >= ROUTING_RA200)
       dlcp_used = TRUE;
   /* If there is at least one user uses ewl routing */
    if(callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA74 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA75 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA78 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA79 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA124 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA125 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA128 ||
		callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA129 )
       ewl_used = TRUE;

	/* ason routings are not allowed for normal CG1-CG5 generators! */
	if( callgen_vect[i]->choosen_routing.selected_routing >= ROUTING_RA50 &&
		callgen_vect[i]->choosen_routing.selected_routing <= ROUTING_RA200 &&
		callgen_vect[i]->type < CG_TYPE6)
	{
		printf("Error:  The routing of generator %s is not compatible\
			to its type!\n", cgen_name(i));
		exit(1);
	}
   }
 if(dlcp_used)
   compute_aggregate_costs();

 for(i = 0; i < no_of_nodes; i++)
 {
	 if(node_vect[i]->opaque)
		 opaque_present = TRUE;
 }

 if(ewl_used && opaque_present)
 {
	 printf("Error:  If equal wavelength routing is used\n\
no opaque nodes are enabled!\n");
	 exit(1);
 }

/* The CAC2 and CAC3 are not allowed yet if IP over WDM is used */
/* the theory neither */
 if(ipowdm_ena && (selected_cac == CALL_ADM_CTRL2 ||
	 selected_cac == CALL_ADM_CTRL3 || theo_analysis))
 {
	 if(theo_analysis)
		 printf("Error:  If IP over WDM is used no theoretical analysis \
is enabled!\n");
	 else
		 printf("Error:  If IP over WDM is used CAC2 and CAC3 are not allowed!\n");
	 exit(1);
 }



/* One event of type OFFICIAL_FBW_UPDATE is put in the Event List for   */
/* each link with a duration of a randomly choosen value beetween 0 and */
/* the update period time of the link. If the update period time is 0.0	*/
/* don't put any event in the list.					*/
 for(i = 0; i < no_of_links; i++)
	{
	  actual_lUs = link_Ucall_stat[i];
	  if(actual_lUs->is_node_node)
	   {
      /* Initializing the official bandwidth and the number
        of U class calls						*/
            actual_lUs->official_free_Uband =
             actual_lUs->free_Uband;
            actual_lUs->official_no_of_Ucall =
             actual_lUs->no_of_Ucall;
            actual_lUs->official_cong = FALSE;

     /* Creating, initializing and scheduling the first updating event	*/
	    if (link_vect[i]->bw_update_p > 0.0)
	    {
	     current_event = event_alloc();
	     current_event->ev_type = OFFICIAL_FBW_UPDATE;
	     current_event->ev_dscrpt.link_id = i;
	     current_event->ev_time = ancles_time + 1.0 +
	      uniform(0.0, link_vect[i]->bw_update_p, &sh_seed);
      /* Put the updating event in the event-queue			*/
	     insert_event(current_event);
	    }
	   }
      /* Initialization the link properties: the passing shortest paths	*/
	  actual_lUs->sp_on = NULL;
	}

/* Initialization of the Periodic Algorithm Choice routing	*/
 init_pac();

/* Set the value of sample deviation in the transient phase */
 sample_deviation = 0.9;



/* Initializing the data structure used to store network state 		*/
/* snapshot.								*/

 if (max_state_photos > 0)
	init_photo_struct();

/* Install the SIGNAL trace */
normal = TRUE;                          /* Standard exit option.             */
#ifdef unix
   signal(SIGUSR1, print_all_actual_stat);
                                        /* if a SIGUSR1 is received, print   */
					/* a brief report of the status of   */
					/* the simulation and continue       */

   signal(SIGUSR2, print_actual_stat);
                                        /* if a SIGUSR2 is received, print   */
					/* a long report of the status of    */
					/* the simulation and continue       */


   signal(SIGINT, write_stat_and_exit); /* If a SIGINT (ctrl+c) is received, */
					/* write all the file and then exit. */

   signal(SIGQUIT, write_stat_and_exit);/* If a SIGQUIT  is received,        */
					/* write all the file and then exit. */

   signal(SIGTERM, write_stat_and_exit);/* If a SIGTERM  is received,        */
					/* write all the file and then exit. */

#endif

/* Obtaining the CPU time elapsed at the beginning of the main          */
/* simulation cycle.                                                    */
 start_CPU_time = clock();
 printf("Starting simulation...\n");
/* Main simulation cycle.       */
 while(ancles_time <= max_time)
   {
	current_event = head_of_event_list->next;
    ancles_time = current_event->ev_time;
	if(ancles_time > max_time)
		continue;

/*if(ancles_time > 2897.2)
real_events_cnt += 0;*/

	++real_events_cnt;
    switch(current_event->ev_type)
       {
       case OPENING_CALL_REQUEST :
	 require_call_opening(current_event->ev_dscrpt.source_cgen_id, current_event->ev_dscrpt_2.dest_cgen_id);
	 break;
	
       case CLOSING_CALL :
	 close_call(current_event->ev_dscrpt.deleting_call);
	 break;
	
       case LOAD_STATE_CHANGE :
	 change_state(current_event->ev_dscrpt.source_cgen_id);
         break;

       case OFFICIAL_FBW_UPDATE :
	 update_off_fbw(current_event->ev_dscrpt.link_id);
         break;

       case UPDATE_ALG :
	 update_alg(current_event->ev_dscrpt.source_cgen_id);
         break;

       case COMPUTE_AGGREGATE_COSTS :
         compute_aggregate_costs();
         break;

       }
    dequeue_event(current_event);
   }

/* The simulation run is stopped when 'ancles_time' exceeds 'max_time', */
/* apart from the current state of the statistical analysis.            */
 stop_simulating();
 free_statistic_arrays();
 return 0;
} /* main ... */
