/*
**  Module    : ancles.h
**  Version   : $Revision: 3.6 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico, Claudio Casetti
**              Filippo Natale, Zoltan Zsoka
**
**  Synopsis  : this module collects declarations and definitions of items
**              referred to by more than one module of ANCLES.
**
**  History   : 19-SEP-1995 - Creation of the module.
**              01-JUN-1996 - Revision                  - AF -
**              01-MAR-1997 - Revision			- MM -
**              14-OCT-1997 - Revision: moved the routing choice to the
**                            CG level.                 - MM -
**              05-DEC-1997 - Revision: added all the stuff about the load
**                            balancing                 - MM - GF - LM -
**              15-JAN-1999 - Revision: added support for dynamic routing and
**                            done some tiding up for static routing -MM-CC-
**              20-AUG-1999 - Revision: adding support for discrete updating
**                            of the available bw for best effort calls
**				and new algorithms (MSF and DLCP) -ZZ-FN-
**		31-AUG-1999 - Revision: Ucalls uses data_amount instead of
**				duration, interrupt		- ZZ -
**		01-FEB-2001 - Added elements for support of th new routing
**				algorithms  - ZZ -
**		01-MAR-2001 - Export the parts of the module into submodules:
**				ancdef.h : macro definitions
**				ancstr.h : type definitions
**				ancvar.h : variable declarations - ZZ -
**
**              01-APR-23 - Insert the features for ASON:
**                          ASON call generators and duplex fibres - ZZ -
*/


/*  Initial condition for the compilation of the module. */
# ifndef        ANCLES_H
#define         ANCLES_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#ifdef unix
 #include <signal.h>
 #include <unistd.h>
#endif

# ifndef __STRICT_ANSI__
#define  __STRICT_ANSI__
# endif

#ifdef __STRICT_ANSI__
 int strcasecmp(const char* a, const char* b);
 char* strdup(const char* s);
#endif

#include "stathead.h"
#include "ancdef.h"
#include "ancstr.h"
#include "ancvar.h"


/***************************
**  Functions Prototypes  **
***************************/

/*  Module : ancles.c   */
void    get_input(char *, void *);
void    init_variables(void);
void    print_cgen_to_cgen_path(int    , int    , int    );
void	schedule_next_arrival(int,	int);

/*  Module : random.c   */
long    rnd32(long);
double  uniform(double, double, long *);
double  uniform_0_1(long *);
double  negexp(double, long *);
int     poisson(double, long *);
int     geometric0(double, long *);
int     geometric1(double, long *);
int     geometric_trunc1(double, int, long *);
int     trunc_exp(double, long, long *);


/*  Module : getpaths.c */
void      find_all_paths(void);
void      find_all_pairs(void);
void	  free_pathlist(Path* );
Boolean equ_double(double, double, double);
Path * ***alloc_path_mat(int     size, int     num);
Pair * ***alloc_pair_mat(int     size, int     num);
void	  substitute_link(int);
int compose_ip_paths( int, int, Path*, Link*, Path*);
Path*	free_linkpath(Path*);

/*  Module : routing.c  */
void    init_routing(void);
Call*   exe_routing(int    , int    , double, double, double);
void	bottleneck_of_MD(int , int);
Path*	iterate_path_list(Path*, int*, Routing*, Path*);
Pair*	iterate_pair_list(Pair*, int*, Routing*, Pair*);

/*  Module : asonrou.c  */
Path*	ason_routing(int, int, Routing*, int*, Path*);
Path*	ason_prot_routing(int, int, Routing*, int*, Pair*);


/*  Module : load_balancing.c */
void	balance_paths(int    , int    );

/*  Module : calladco.c */
double  eval_equivalent_bw(double, double, double, double);
double  eval_modified_bw(double, double);
void    updt_var_mean(Call *, double *, double *, Boolean);
int     update_samples_vect(Link *, Boolean);
void    init_CAC1(void);
void    init_CAC2(void);
void    init_CAC3(void);
void    init_CAC4(void);
/*void    init_CAC5(void);*/
void    init_Ucall(void);
Boolean exe_CAC1(Link *, double, double, double);
Boolean exe_CAC2(Link *, double, double, double);
Boolean exe_CAC3(Link *, double, double, double);
Boolean exe_CAC4(Link *, double, double, double);
/*Boolean exe_CAC5(Link *, double, double, double);*/
double  update_Ucall_open(Link *, double, double);
void    update_Ucall_close(Call *, Link *);

/*  Module : lists.c    */
void    insert_event(Event *);
void    dequeue_event(Event *);
void    dequeue_far_event(Event *);
void    unrel_dequeue_event(Event *, double);
void    insert_call(Call *);
void    dequeue_call(Call *);
void    unrel_dequeue_call(Call *, double);
void    insert_call_list_on_link(CallList *, Link *);
void    dequeue_call_list_from_link(CallList *, Link *);
void    unrel_dequeue_call_list_from_link(CallList *, Link *, double);
void    init_event_list(void);
void    init_call_list(void);
void    init_call_list_on_link(void);
void    init_Ucall_list_on_link(void);
void    insert_Ucall_list_on_link(CallList *, Link *);
void    dequeue_Ucall_list_from_link(CallList *, Link *);
void    unrel_dequeue_Ucall_list_from_link(CallList *, Link *, double);
void	init_call_list_on_new_link(int);
void	exchange_call_list_on_links(int , int);
void	init_Ucall_list_on_new_link(int);
void	exchange_Ucall_list_on_links(int , int);
CallList  *search_Ucall_list_on_link(Link *, Call *, double);
CallList  *search_call_list_on_link(Link *, Call *, double);
void	reinsert_call_and_event(Call* , double);

/*  Module : freelist.c */
void * 	MMmalloc(size_t, const char *);
LinkList  *linklist_alloc(void);
void      linklist_release(LinkList *);
void      whole_linklist_release(void);
Path      *path_alloc(void);
void      path_release(Path *);
void      whole_pathlist_release(void);
Stack     *stack_alloc(void);
void      stack_release(Stack *);
void      whole_stack_release(void);
Call      *call_alloc(void);
void      call_release(Call *);
CallList  *call_list_alloc(void);
void      call_list_release(CallList *);
CallType1 *call_type1_alloc(void);
void      call_type1_release(CallType1 *);
CallType2 *call_type2_alloc(void);
void      call_type2_release(CallType2 *);
CallType3 *call_type3_alloc(void);
void      call_type3_release(CallType3 *);
CallType4 *call_type4_alloc(void);
void      call_type4_release(CallType4 *);
CallType5 *call_type5_alloc(void);
void      call_type5_release(CallType5 *);
CallType6 *call_type6_alloc(void);
void      call_type6_release(CallType6 *);
CallType8 *call_type8_alloc(void);
void      call_type8_release(CallType8 *);
ProtCh	  *prot_ch_alloc(void);
void      prot_ch_release(ProtCh *);
Intlist   *int_list_alloc(void);
void      int_list_release(Intlist *);
VCset     *vcset_alloc(void);
void      vcset_release(VCset *);
void      whole_vcset_release(void);
Event     *event_alloc(void);
void      event_release(Event *);
Dbllist   *dbl_list_alloc(void);
void      dbl_list_release(Dbllist *);


/*  Module : ancstat.c  */
void    generate_CBR_report(FILE * out_fp);
void    generate_CDL_report(FILE * out_fp);
void    generate_CL_report(FILE * out_fp);
void    generate_CG8_report(FILE * out_fp);
void    generate_AWL_report(FILE * out_fp);
void    generate_AWH_report(FILE * out_fp);
void    generate_ABW_report(FILE * out_fp);
void    generate_LWH_report(FILE * out_fp);
void    generate_WLC_report(FILE * out_fp);
void    generate_ROU_report(FILE * out_fp);
void    generate_UBR_report(FILE * out_fp);
void    generate_IBR_report(FILE * out_fp);
void    generate_ULH_report(FILE * out_fp);
void    generate_ILH_report(FILE * out_fp);
void    generate_4CC_report(FILE * out_fp);
void    generate_UCL_report(FILE * out_fp);
void    generate_HST_report(FILE * out_fp);
void    generate_WVL_report(FILE * out_fp);
void    generate_STR_report();
void    accepted_call_stat_update(Call *, Boolean, Boolean );
void    refused_call_stat_update(int    , int    );
void    closed_call_stat_update(Call *, Boolean );
void    reset(void);
void    stop_simulating(void);
void    init_statmem(void);
void    free_statistic_arrays(void);

/*  Module : photo.c    */
void    init_photo_struct(void);
void	manage_photo_list(int    );
void    rename_photo_files(void);

/*  Module : signal.c   */
void    print_actual_stat(int);
void    print_all_actual_stat(int);
void    write_stat_and_exit(int);

/*  Module : bottleneck.c */
void	init_bott_staticvars(void);
void    update_bottleneck_measure(Call*, double, double, int);

/***************************
**  ND Modules Functions  **
***************************/

/*  Module : nd_ancles.c        */
void    init_variables_stat(void);

/*  Module : nd_init.c          */
void    initialize_parameters(void);

/*  Module : nd_service.c       */
const char    *node_name(int);
const char    *link_name(int);
const char    *user_name(int);
const char    *cgen_name(int);
const char    *route_name(int);
void    release_memory(void);

/*  Module : nd_traffic.c       */
Boolean read_traff_relations(char *);
void    create_uniform_traffic(void);
void    remind_parameter_names(void);

/*  Module : nd_yacc.y          */
int     yyparse();

/*********************************
**  Theoretic Models Functions  **
**********************************/

/*  Module   theo_complex.c		*/
double	fbinom(int, int);
double	factorial(int);
void  createbinom(int);
void  createfacto(int);
double	theo_multiindep_complex(int, double *, int *);
double	theo_multimost_complex(int, double *, int *);
double	theo_bpp_complex(int, double *, int *);
double  theo_lisomani_complex(int);


/*********************************
**  IP over WDM Functions       **
**********************************/
/*  Module   iplinks.c			*/
void open_new_virt_link(int , int );
void close_virt_link(Link* );


# endif /* ANCLES_H */
