/*
**  Module    : asonrou.c
**  Version   : $Revision: 1.0$
**  Date      : $Date: 2002/06/11 15:15:50 $
**  Authors   : Zoltan Zsoka
**
**  Synopsis  : this module contains the functions dealing with the
**              available routing algorithms for ason networks.
**
**  Functions : 
**				Path* ason_routing(int, int, Routing*, int*, Path*)
**				Path* ason_prot_routing(int, int, Routing*, int*, Pair*)
**
**  History   : 11-JUN-2002 - Creation of the module.   -ZZ-
*/

# ifndef        ANCLES_H
#include        "ancles.h"
# endif

/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : Path* ason_routing(int cgen_src, int cgen_dest,
**			Routing* cg_routing, int* curr_path_no,  Path* primary_path)
**  Return    : Pointer to the chosen path
**  Remarks   : With this function we can choose handle only one route,
**              for only one wavelength capacity.
**				If there are opaque nodes, the choice of the wavelength is
**				performed separated on the opaque sections.
**              The allocation of the first aviable wavelength is included
**              in this algorithm.
*/
Path* ason_routing(int cgen_src, int cgen_dest, Routing* cg_routing,
			int* curr_path_no, Path* primary_path)
{
  Path* path;			/* Current path. */
  Path* sel_path = NULL;	/* Selected path. */
  LinkList* link;		/* Used to scan the path links.  */
  LinkList* os_first_link; /* first link in the current opaque section */
  int i, j, k, sel_wave[MAX_NODE];
  int lowest_sw[MAX_NODE];
  int no_freewaves, min_no_freewaves, max_min_no_freewaves = 0;
  int no_diff_freewaves; /* number of different free waves */
  int sel_path_no = (*curr_path_no);
  int opsecnum, max_oswave, lowest_max = 8 * max_wvoct_all_links + 1;
  unsigned char wave_mask;          /* The actual mask of the aviable waves */
  unsigned char alloc_mask;         /* The mask used for allocate wave */
  unsigned char wave_mask_cp;       /* Copy of wavemask */
  Boolean more_paths = TRUE, path_is_ok, path_found = FALSE;
  Boolean endofos; /* end of opaque section */

  int free_chans[8];
  int actualwave = 0, ch_links, max_ch_links = 0;
  int no_checked_links;
  double rand_0_1;

  path = primary_path;

  do
  {
    no_checked_links = 0;
    switch (cg_routing->selected_routing)
     {
      case ROUTING_RA50:	/* Only primary path routing */
      case ROUTING_RA51:	/* Shortest lowest path routing */
      case ROUTING_RA53:	/* Alternate shortest path routing */
       {
		max_oswave = 0;
		opsecnum = 0;
		link = path->path;
		path_is_ok = TRUE;
		/* The opaque sections are handled sequentially */
		while(link != NULL && path_is_ok)
		{
			/* The node-to-node links are checked for the free wave */
			os_first_link = link;
			sel_wave[opsecnum] = -1;
			/* Reset the maximum of checked links for the next opsec */
			max_ch_links = 0; 
			for(i = 0; i < max_wvoct_all_links && sel_wave[opsecnum] == -1; i++)
			{
				wave_mask = 0xFF;
				endofos = FALSE;
				link = os_first_link;
				ch_links = 0;
				while(!endofos && link != NULL)
				{
					/* Comparing the free_wave mask of every node-node link */
					if(link->link_elem->is_node_node)
					{
						/* Count the links checked */
						if(wave_mask)
							ch_links++;							  
						wave_mask &= link->link_elem->free_waves[i];
						if(link->endopsec)
							endofos = TRUE;
					}
					link = link->next;
				}
				if( wave_mask )
				{
					/* The numberring of the wavelengths starts with 1 */
					sel_wave[opsecnum] = (8 * i) + 1;
					/* shifting  while the first 1 is found */
					while(! (wave_mask & 0x80))
					{
						sel_wave[opsecnum]++;
						wave_mask <<= 1;
					}
				}
				if(ch_links > max_ch_links)
					max_ch_links = ch_links;
			} /* for */
			if(sel_wave[opsecnum] == -1)
				path_is_ok = FALSE;
			if(sel_wave[opsecnum] > max_oswave)
				max_oswave = sel_wave[opsecnum];
			opsecnum ++;
			no_checked_links += max_ch_links;
		} /* while */

		/* Links are put in the union of paths */
		for(i = 0, link = path->path->next; i < no_checked_links; i++)
		{
			in_path_union[link->link_elem->id] ++;
			link = link->next;
		}

        if(cg_routing->selected_routing == ROUTING_RA50 ||
	        cg_routing->selected_routing == ROUTING_RA53)
		{
			more_paths = FALSE;
			if(path_is_ok)
			{
				sel_path = path;
				sel_path_no = (*curr_path_no);
				for(i=0; i<opsecnum; i++)
				   lowest_sw[i] = sel_wave[i];
				path_found = TRUE;
			}
			else
			{
	/* The alternate shortest path searchs till a route is found */
			 if(cg_routing->selected_routing == ROUTING_RA53)
			 {
				path = iterate_path_list(path, curr_path_no,
					cg_routing,	primary_path);
				if(path != NULL)
					more_paths = TRUE;
			 }
			}

		}
		else
		{
			/* Looking for the lowest free wavelength */
			if(path_is_ok && max_oswave < lowest_max)
			{
				sel_path = path;
				sel_path_no = (*curr_path_no);
				for(i=0; i<opsecnum; i++)
				   lowest_sw[i] = sel_wave[i];
				lowest_max = max_oswave;
				path_found = TRUE;
			}
            path = iterate_path_list(path, curr_path_no, cg_routing,
				primary_path);
            if(path == NULL)
				more_paths = FALSE;
		}
	   }
		break;

      case ROUTING_RA52:	/* Shortest widest path routing */
       {
	  /* The node-to-node links are checked for the free waves */
		min_no_freewaves = 8 * max_wvoct_all_links + 1;
		opsecnum = 0;
		link = path->path;
		path_is_ok = TRUE;
		/* The opaque sections are handled sequentially */
		while(link != NULL && path_is_ok)
		{
			/* The node-to-node links are checked for the free wave */
			os_first_link = link;
			sel_wave[opsecnum] = -1;
			no_freewaves = 0;
			/* Reset the maximum of checked links for the next opsec */
			max_ch_links = 0; 
			for(i = 0; i < max_wvoct_all_links; i++)
			{
				wave_mask = 0xFF;
				for(k=0; k<8; k++)
					free_chans[k] = MAX_FIBER_NUM;
				endofos = FALSE;
				link = os_first_link;
				ch_links = 0;
				while(!endofos && link != NULL)
				{
					/* Comparing the free_wave mask of every node-node link */
					if(link->link_elem->is_node_node)
					{
						if(wave_mask)
							ch_links++;							  
						wave_mask &= link->link_elem->free_waves[i];
						/* count the real number of free channels */
						for(k = 0; k < 8; k++)
						{
							if(8*i + k < link->link_elem->max_waves)
							{
								if(free_chans[k] > 
									 link->link_elem->free_fibers[8*i+k])
									free_chans[k] =
									 link->link_elem->free_fibers[8*i+k];
							}
						}
						if(link->endopsec)
							endofos = TRUE;
					}
					link = link->next;
				}
				if( wave_mask )
				{
					/* Set the wavelength which is used if this path is choosen */
					if(sel_wave[opsecnum] == -1)
					{
						/* The numberring of the wavelengths starts with 1 */
						sel_wave[opsecnum] = (8 * i) + 1;
						/* shifting  while the first 1 is found */
						wave_mask_cp = wave_mask;
						while(! (wave_mask_cp & 0x80))
						{
							sel_wave[opsecnum]++;
							wave_mask_cp <<= 1;
						}
					}
					/* Increment the number of free waves on the link */
					k = 0;
					while(wave_mask)
					{
						if(wave_mask & 0x80)
							no_freewaves += free_chans[k];
						k++;
						wave_mask <<= 1;
					}
					
				} /*if*/
				if(ch_links > max_ch_links)
					max_ch_links = ch_links;
			} /*for*/
			if(sel_wave[opsecnum] == -1)
				path_is_ok = FALSE;
			if(min_no_freewaves > no_freewaves)
				min_no_freewaves = no_freewaves;
			opsecnum ++;
			no_checked_links += max_ch_links;
		} /* while */

		/* Links are put in the union of paths */
		for(i = 0, link = path->path->next; i < no_checked_links; i++)
		{
			in_path_union[link->link_elem->id] ++;
			link = link->next;
		}
		
		/* Looking for the widest (and shortest) path */
		if(min_no_freewaves > max_min_no_freewaves)
		{
			sel_path = path;
			sel_path_no = (*curr_path_no);
			max_min_no_freewaves = min_no_freewaves;
			for(i=0; i<opsecnum; i++)
				lowest_sw[i] = sel_wave[i];
			path_found = TRUE;
		}
		
		path = iterate_path_list(path, curr_path_no, cg_routing,
			primary_path);
		if(path == NULL)
			more_paths = FALSE;
	   }
	   break;

	   /* Implemented by Papp Denes */
	  case ROUTING_RA62:		/*  Random widest shortest path routing */
	  case ROUTING_RA60:		/*  FSP with random wavelength assignment*/
	  case ROUTING_RA63:		/*  ASP with random wavelength assignment*/
	  case ROUTING_RA90:		/*  60, but random wavelength by channel */
	  case ROUTING_RA92:		/*  62, but random wavelength by channel */
	  case ROUTING_RA93:		/*  63, but random wavelength by channel */
		  {
			  /* The node-to-node links are checked for the free waves */
			  min_no_freewaves = 8 * max_wvoct_all_links + 1;
			  opsecnum = 0;
			  link = path->path;
			  path_is_ok = TRUE;
			  /* The opaque sections are handled sequentially */
			  while(link != NULL && path_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave[opsecnum] = -1;
				  no_diff_freewaves = 0;
				  no_freewaves = 0;
				  /* Reset the maximum of checked links for the next opsec */
				  max_ch_links = 0; 
				  for(i = 0; i < max_wvoct_all_links; i++)
				  {
					  wave_mask = 0xFF;
					  for(k=0; k<8; k++)
						  free_chans[k] = MAX_FIBER_NUM;
					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link_Ucall_stat[link->link_elem->id]->is_node_node)
						  {
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k < link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }
							  if(link->endopsec)
								  endofos = TRUE;
						  }
						  link = link->next;
						  
					  }
					  if( wave_mask )
					  {
						  actualwave = 8*i + 1;
						  k = 0;
						  while(wave_mask)
						  {
							  if(wave_mask & 0x80)
							  {
								  if (cg_routing->selected_routing == ROUTING_RA90 ||
									  cg_routing->selected_routing == ROUTING_RA92 ||
									  cg_routing->selected_routing == ROUTING_RA93)
								  {
									  
								  /* Increment the number of free waves on the link
									  and note the wavelength number */
									  for (j = 0;j<free_chans[k];j++)
									  {
										  tfree[no_freewaves] = actualwave;
										  no_freewaves++;
									  }/*for*/
									  
									  
									   /* The number of the available channels can be 
									   different from the number of different free 
									  waves (one wave -> more fibers -> more channels) */
								  }
								  
								  else 
								  {
								  /* Increment the number of free waves on the link
									  and note the wavelength number */
									  tfree[no_diff_freewaves] = actualwave;
									  /* The number of the available channels can be 
									  different from the number of different free 
									  waves (one wave -> more fibers -> more channels) */
									  no_freewaves += free_chans[k];
								  }
								  no_diff_freewaves++;
							  }
							  k++;
							  wave_mask <<= 1;
							  actualwave++;
						  }
						  
					  } /*if*/
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  } /*for*/
				  
				  /* choose a random element of the free wavelength */
				  if(no_diff_freewaves != 0)
				  {
					  rand_0_1 = uniform_0_1(&sh_seed);
					  
					  /* to avoid the owerflow of tfree */
					  if(rand_0_1 == 1) rand_0_1 = 0;
					  
					  if(cg_routing->selected_routing == ROUTING_RA90 ||
						 cg_routing->selected_routing == ROUTING_RA92 ||
						 cg_routing->selected_routing == ROUTING_RA93)
						 sel_wave[opsecnum] = 
							tfree[(int)floor(no_freewaves*rand_0_1)];
					  else
						  sel_wave[opsecnum] =
							tfree[(int)floor(no_diff_freewaves*rand_0_1)];
				  }
				  else
				  {
					  /* no waves are available on the link */
					  sel_wave[opsecnum] = -1;
					  path_is_ok = FALSE;
				  }
				  if(min_no_freewaves > no_freewaves)
					  min_no_freewaves = no_freewaves;
				  opsecnum ++;
				  no_checked_links += max_ch_links;
			  } /* while */
			  
			  /* Links are put in the union of paths */
			  for(i = 0, link = path->path->next; i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  
			  if(cg_routing->selected_routing == ROUTING_RA62 ||
				 cg_routing->selected_routing == ROUTING_RA92)
			  {
				  /* Looking for the widest (and shortest) path */
				  if(min_no_freewaves > max_min_no_freewaves)
				  {
					  sel_path = path;
					  sel_path_no = (*curr_path_no);
					  max_min_no_freewaves = min_no_freewaves;
					  for(i=0; i<opsecnum; i++)
						  lowest_sw[i] = sel_wave[i];
					  path_found = TRUE;
				  }
				  
				  path = iterate_path_list(path, curr_path_no, cg_routing,
					  primary_path);
				  if(path == NULL)
					  more_paths = FALSE;
				  
			  } /* if routing_RA62 or routing_RA92*/
			  
			  if(cg_routing->selected_routing == ROUTING_RA60 ||
				  cg_routing->selected_routing == ROUTING_RA63 ||
				  cg_routing->selected_routing == ROUTING_RA90 ||
				  cg_routing->selected_routing == ROUTING_RA93)
			  {
				  more_paths = FALSE;
				  if(path_is_ok)
				  {
					  sel_path = path;
					  sel_path_no = (*curr_path_no);
					  for(i=0; i<opsecnum; i++)
						  lowest_sw[i] = sel_wave[i];
					  path_found = TRUE;
				  }
				  else
				  {
					  /* The alternate shortest path searchs till a route is found */
					  if(cg_routing->selected_routing == ROUTING_RA63 ||
						  cg_routing->selected_routing == ROUTING_RA93)
					  {
						  path = iterate_path_list(path, curr_path_no,
							  cg_routing,	primary_path);
						  if(path != NULL)
							  more_paths = TRUE;
					  }
				  }
			  }
		  } /* case RA60, RA62, RA63, RA90, RA92, RA93 */	  
		  break;

      default:
		  {
			  /* Something went wrong  */
			  printf("ERROR: unknown routing algorithm: %d \n",
				  cg_routing->selected_routing);
			  exit(1);
		  }
		  break;
     } /* switch */
	} while (more_paths);

	if(path_found) /* In this case we found a free wavelength */
	{   /* on the links of the sel_path. Now we allocate it on the links. */
		opsecnum = 0;
		link = sel_path->path;
		j=0;
		/* The opaque sections are handled sequentially */
		while(link != NULL)
		{
			/* We allocate the wavelength on the node-to-node
			links of the opaque section*/
			os_first_link = link;
			endofos = FALSE;
			alloc_mask = 0x80;
			for(i = 0; i < (lowest_sw[opsecnum]-1)%8; i++)
				alloc_mask >>= 1;
			while(!endofos && link != NULL)
			{
				if(link->link_elem->is_node_node)
				{
					/* if the last fiber is allocated */
					if( !(--link->link_elem->free_fibers[lowest_sw[opsecnum]-1]) )
						link->link_elem->free_waves[(lowest_sw[opsecnum]-1)/8] 
						&= (~alloc_mask);
					
					/* Statistics of the wavelength opened for the new call */
					calls_on_link_wave[link->link_elem->id]
						[lowest_sw[opsecnum]]++;
					
					if(link->endopsec)
						endofos = TRUE;
				}
				if(lowest_sw[opsecnum] > link->link_elem->peak_wave)
					link->link_elem->peak_wave = lowest_sw[opsecnum];
				link->link_elem->used_waves++;
				u_wave[j++] = lowest_sw[opsecnum];
				link = link->next;
			}
			opsecnum++;
		} /* while */
	}
	(*curr_path_no) = sel_path_no;
	return sel_path;

}


/*
**  Function  : Path* ason_prot_routing(int cgen_src, int cgen_dest,
**			Routing* cg_routing, int* curr_pair_no,  Pair* primary_pair)
**  Return    : Pointer to the chosen working path
**  Remarks   : With this function we can choose a pair of routes,
**              for only one wavelength capacity.
**				If there are opaque nodes, the choice of the wavelength is
**				performed separated on the opaque sections.
**              The allocation of the first aviable wavelength is included
**              in this algorithm.
*/
Path* ason_prot_routing(int cgen_src, int cgen_dest, Routing* cg_routing,
			int* curr_pair_no, Pair* primary_pair)
{
  Pair* pair;			/* Current pair of paths. */

  Pair* sel_pair = NULL;	/* Selected pair. */
  LinkList* link;		/* Used to scan the path links.  */
  LinkList* os_first_link; /* first link in the current opaque section */
  int i, j, sel_wave1[MAX_NODE], sel_wave2[MAX_NODE];
  int lowest_sw1[MAX_NODE];
  int lowest_sw2[MAX_NODE];
/*  int no_freewaves, min_no_freewaves, max_min_no_freewaves = 0;*/
  int sel_pair_no = (*curr_pair_no);
  int opsecnum1, opsecnum2, max_prot;
  int lowest_max = 8 * max_wvoct_all_links + 1; /* max_oswave, */
  unsigned char wave_mask;          /* The actual mask of the aviable waves */
  unsigned char alloc_mask;         /* The mask used for allocate wave */
  Boolean more_pairs = TRUE, pair_is_ok, pair_found = FALSE, curr_new_prot_chan;
  Boolean endofos; /* end of opaque section */
  ProtCh	*new_prot_ch;
  Intlist	*new_intlist, *prev_intlist;
  int	start_node;			/* starting point of the link */
  Boolean	not_first_node; /* indicates if start_node is 
								not the first node of the path */
  int   min_new_hops = 2 * no_of_nodes; /* less needed number of channel "hops" */
  double   min_new_length = 2 * MAX_PLENGTH; /* less needed total channel length */

  Routing	virt_routing;
  Path*	sel_path;
  Path*	be_path;
  int curr_path_no;

  int k;
  int no_freewaves,min_no_freewaves;
  int no_diff_freewaves; /* number of different free waves */
  int free_chans[8];
  int actualwave = 0, ch_links, max_ch_links = 0;
  int   no_checked_links;
  double rand_0_1;
  Boolean path_is_ok;

  pair = primary_pair;
  if(!pair)
  {
	  /* No pair exists */
	  used_prot_path = NULL;
	  return NULL;
  }

  do
	{
    switch (cg_routing->selected_routing)
     {
      case ROUTING_RA70:	/* Fixed shortest 1+1 pair routing */
      case ROUTING_RA73:	/* Alternate shortest 1+1 pair routing */
      case ROUTING_RA76:	/* Best effort fixed shortest 1+1 pair routing */
      case ROUTING_RA77:	/* Best effort alternate shortest 1+1 pair routing */
		{
			  /*	max_oswave = 0;*/
			  
			  /* looking for free wavelengths on the working path */
			  opsecnum2 = 0;
			  link = pair->path2->path;
			  pair_is_ok = TRUE;
			  no_checked_links = 0;
			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave2[opsecnum2] = -1;
				  /* Reset the maximum of checked links for the next opsec */
				  max_ch_links = 0; 
				  for(i = 0; i < max_wvoct_all_links && sel_wave2[opsecnum2] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  /* Count the links checked */
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  if(link->endopsec)
								  endofos = TRUE;
						  }					  
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  /* The numberring of the wavelengths starts with 1 */
						  sel_wave2[opsecnum2] = (8 * i) + 1;
						  /* shifting  while the first 1 is found */
						  while(! (wave_mask & 0x80))
						  {
							  sel_wave2[opsecnum2]++;
							  wave_mask <<= 1;
						  }
					  }
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  }/* for */
				  if(sel_wave2[opsecnum2] == -1)
					  pair_is_ok = FALSE;
					  /*			if(sel_wave[opsecnum] > max_oswave)
				  max_oswave = sel_wave[opsecnum];*/
				  opsecnum2 ++;
				  no_checked_links += max_ch_links;
			  } /* while */

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  /* looking for free wavelengths on the protective path */
			  opsecnum1 = 0;
			  link = pair->path1->path;
			  no_checked_links = 0;
			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave1[opsecnum1] = -1;
				  max_ch_links = 0; 
				  for(i = 0; i < max_wvoct_all_links && sel_wave1[opsecnum1] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  if(link->endopsec)
								  endofos = TRUE;
						  }
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  /* The numberring of the wavelengths starts with 1 */
						  sel_wave1[opsecnum1] = (8 * i) + 1;
						  /* shifting  while the first 1 is found */
						  while(! (wave_mask & 0x80))
						  {
							  sel_wave1[opsecnum1]++;
							  wave_mask <<= 1;
						  }
					  }
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  }/* for */
				  if(sel_wave1[opsecnum1] == -1)
					  pair_is_ok = FALSE;
					  /*			if(sel_wave[opsecnum] > max_oswave)
				  max_oswave = sel_wave[opsecnum];*/
				  opsecnum1 ++;
				  no_checked_links += max_ch_links;
			  } /* while */

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next;
				i< no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  more_pairs = FALSE;
			  if(pair_is_ok)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  for(i=0; i<opsecnum1; i++)
					  lowest_sw1[i] = sel_wave1[i];
				  for(i=0; i<opsecnum2; i++)
					  lowest_sw2[i] = sel_wave2[i];
				  pair_found = TRUE;
				  new_prot_chan = TRUE;
				  dedicated_protchan = TRUE;
			  }
			  else
			  {
				  /* The alternate shortest path searchs till a route is found */
				  if(cg_routing->selected_routing == ROUTING_RA73 ||
					 cg_routing->selected_routing == ROUTING_RA77)
				  {
					  pair = iterate_pair_list(pair, curr_pair_no,
						  cg_routing,	primary_pair);
					  if(pair != NULL)
						  more_pairs = TRUE;
				  }
			  }
		}
		break;

/* Routing algorithms with equal wavelength on the working and protective path. */
      case ROUTING_RA74:	/* Fixed shortest 1+1 pair with ewl routing */
      case ROUTING_RA75:	/* Alternate shortest 1+1 pair with ewl routing */
      case ROUTING_RA78:	/* Best effort fixed shortest 1+1 pair with ewl routing */
      case ROUTING_RA79:	/* Best effort alternate shortest 1+1 pair with ewl routing */
		  {
			  /* Do not care about opaque nodes! */
			  
			  /* The node-to-node links are checked for the free wave */
			  sel_wave2[0] = -1;
			  max_ch_links = 0; 
			  no_checked_links = 0;
			  for(i = 0; i < max_wvoct_all_links && sel_wave2[0] == -1; i++)
			  {
				  wave_mask = 0xFF;
				  /* looking for free wavelengths on the working path */
				  link = pair->path2->path;
				  ch_links = 0;
				  while(link != NULL && wave_mask)
				  {
					  /* Comparing the free_wave mask of every node-node link */
					  if(link->link_elem->is_node_node)
					  {
						  /* Count the links checked */
						  ch_links++;							  
						  wave_mask &= link->link_elem->free_waves[i];
					  }
					  link = link->next;
				  }
				  /* looking for free wavelengths on the protective path */
				  link = pair->path1->path;
				  while(link != NULL && wave_mask)
				  {
					  /* Comparing the free_wave mask of every node-node link */
					  if(link->link_elem->is_node_node)
					  {
						  /* Count the links checked */
						  ch_links++;							  
						  wave_mask &= link->link_elem->free_waves[i];
					  }
					  link = link->next;
				  }
				  
				  if( wave_mask )
				  {
					  /* The numberring of the wavelengths starts with 1 */
					  sel_wave2[0] = (8 * i) + 1;
					  /* shifting  while the first 1 is found */
					  while(! (wave_mask & 0x80))
					  {
						  sel_wave2[0]++;
						  wave_mask <<= 1;
					  }
				  }
				  if(ch_links > max_ch_links)
					  max_ch_links = ch_links;
			  }
			  no_checked_links += max_ch_links;

			  if(no_checked_links > pair->path2->no_of_hops)
			  {
				  max_ch_links = pair->path2->no_of_hops;
				  no_checked_links -= pair->path2->no_of_hops;
			  }
			  else
			  {
				  max_ch_links = no_checked_links;
				  no_checked_links = 0;
			  }
			  
			  /* Working links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next; 
				i < max_ch_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  /* Protective links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  more_pairs = FALSE;
			  if(sel_wave2[0] != -1)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  lowest_sw1[0] = sel_wave2[0];
				  lowest_sw2[0] = sel_wave2[0];
				  pair_found = TRUE;
				  new_prot_chan = TRUE;
				  dedicated_protchan = TRUE;
			  }
			  else
			  {
				  /* The alternate shortest path searchs till a route is found */
				  if(cg_routing->selected_routing == ROUTING_RA75 ||
					  cg_routing->selected_routing == ROUTING_RA79)
				  {
					  pair = iterate_pair_list(pair, curr_pair_no,
						  cg_routing,	primary_pair);
					  if(pair != NULL)
						  more_pairs = TRUE;
				  }
			  }
		  }
		  break;

/* no sense!     case ROUTING_RA80:	/* fixed shortest shared pair routing */
      case ROUTING_RA83:	/* Alternate shortest shared pair routing */
							/* shortest means the preference of the pair */
							/* which allocates the less new wl resources */
      case ROUTING_RA87:	/* Best effort alternate shortest 1+1 pair routing */
		{
			  /* looking for free wavelengths on the working path */

			  /* we need a temporary list of the wavelengths, because it is
			     not sure that this path (and this pair) will be chosen */
			  opsecnum2 = 0;
			  link = pair->path2->path;
			  /* If the working path is shorter than the path found
			     as "shortest" (less resource needed) */
			  if( ( cg_routing->path_order_type == IN_ORDER_OF_HOP_NUMBER &&
				  pair->path2->no_of_hops < min_new_hops) ||
				  ( cg_routing->path_order_type == IN_ORDER_OF_LENGTH &&
				  pair->path2->length < min_new_length ))
				  pair_is_ok = TRUE;
			  else
				  pair_is_ok = FALSE;

			  curr_new_prot_chan = FALSE;
			  max_prot = 0;
			  not_first_node = FALSE;
			  no_checked_links = 0;

			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave2[opsecnum2] = -1;
				  /* Reset the maximum of checked links for the next opsec */
				  max_ch_links = 0; 
				  for(i = 0; i < max_wvoct_all_links && 
					  sel_wave2[opsecnum2] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  /* Count the links checked */
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  if(link->endopsec)
								  endofos = TRUE;
							  if(pair->path1->
								  protected_links[link->link_elem->nnid]+1 > max_prot)
								  max_prot = 
								  pair->path1->protected_links[link->link_elem->nnid]+1;

							  /* the starting point of the link */
							  if(link->direction)
								  start_node = link->link_elem->dest;
							  else
								  start_node = link->link_elem->source;

							  if(not_first_node && pair->path1->
								  protected_nodes[start_node]+1 > max_prot)
								  max_prot = 
								  pair->path1->protected_nodes[start_node]+1;
							  not_first_node = TRUE;

						  }					  
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  /* The numberring of the wavelengths starts with 1 */
						  sel_wave2[opsecnum2] = (8 * i) + 1;
						  /* shifting  while the first 1 is found */
						  while(! (wave_mask & 0x80))
						  {
							  sel_wave2[opsecnum2]++;
							  wave_mask <<= 1;
						  }
					  }
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  }/*for*/
				  if(sel_wave2[opsecnum2] == -1)
					  pair_is_ok = FALSE;
					  /*			if(sel_wave[opsecnum] > max_oswave)
				  max_oswave = sel_wave[opsecnum];*/
				  opsecnum2 ++;
				  no_checked_links += max_ch_links;
			  } /* while */

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next; 
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }

			  if(pair_is_ok)
			  {
				  /* If new protection channel has to be opened */
				  if(max_prot > pair->path1->num_of_prot_chs)
					  curr_new_prot_chan = TRUE;

				  /* Check if this pair could be the "shortest" */
				  if( curr_new_prot_chan &&
					 !(cg_routing->path_order_type == IN_ORDER_OF_HOP_NUMBER &&
					  pair->path1->no_of_hops + 
					  pair->path2->no_of_hops < min_new_hops) &&
					 !(cg_routing->path_order_type == IN_ORDER_OF_LENGTH &&
					  pair->path1->length + 
					  pair->path2->length < min_new_length))
				 /* if it would be longer, don't choose this path */
					  pair_is_ok = FALSE;
					  
			  /*the case of only working path was already checked */
			  }

			  /* looking for free wavelengths on the protective path */
			  opsecnum1 = 0;
			  link = pair->path1->path;
			  no_checked_links = 0;
			  /* If needed, the opaque sections are handled sequentially */
			  if(curr_new_prot_chan)
			  {		  
				  while(link != NULL && pair_is_ok)
				  {
					  /* The node-to-node links are checked for the free wave */
					  os_first_link = link;
					  sel_wave1[opsecnum1] = -1;
					  max_ch_links = 0; 
					  for(i = 0; i < max_wvoct_all_links && sel_wave1[opsecnum1] == -1; i++)
					  {
						  wave_mask = 0xFF;
						  endofos = FALSE;
						  link = os_first_link;
						  ch_links = 0;
						  while(!endofos && link != NULL)
						  {
							  /* Comparing the free_wave mask of every node-node link */
							  if(link->link_elem->is_node_node)
							  {
								  /* Count the links checked */
								  if(wave_mask)
									  ch_links++;							  
								  wave_mask &= link->link_elem->free_waves[i];
								  if(link->endopsec)
									  endofos = TRUE;
							  }
							  link = link->next;
						  }
						  if( wave_mask )
						  {
							  /* The numberring of the wavelengths starts with 1 */
							  sel_wave1[opsecnum1] = (8 * i) + 1;
							  /* shifting  while the first 1 is found */
							  while(! (wave_mask & 0x80))
							  {
								  sel_wave1[opsecnum1]++;
								  wave_mask <<= 1;
							  }
						  }
						  if(ch_links > max_ch_links)
							  max_ch_links = ch_links;
					  }
					  if(sel_wave1[opsecnum1] == -1)
						  pair_is_ok = FALSE;
						  /*			if(sel_wave[opsecnum] > max_oswave)
					  max_oswave = sel_wave[opsecnum];*/
					  opsecnum1 ++;
					  no_checked_links += max_ch_links;
				  } /* while */
			  }

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next; 
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  /* "Shorter" path found */
			  if(pair_is_ok)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  for(i=0; i<opsecnum2; i++)
					  lowest_sw2[i] = sel_wave2[i];

				  new_prot_chan = curr_new_prot_chan;

				  /* New channel has to be open */
				  if(curr_new_prot_chan)
				  {
					  for(i=0; i<opsecnum1; i++)
						  lowest_sw1[i] = sel_wave1[i];
					  min_new_hops = pair->path1->no_of_hops + 
								  pair->path2->no_of_hops;
					  min_new_length = pair->path1->length + 
								  pair->path2->length;
				  }
				  else
				  {
					  min_new_hops = pair->path1->no_of_hops;
					  min_new_length = pair->path1->length;
				  }

				  pair_found = TRUE;
			  }

			  more_pairs = FALSE;
			  pair = iterate_pair_list(pair, curr_pair_no,
				  cg_routing,	primary_pair);
			  if(pair != NULL)
				  more_pairs = TRUE;
		}
		break;

		case ROUTING_RA120:	/* Fixed shortest 1+1 pair routing +random*/
		case ROUTING_RA123:	/* Alternate shortest 1+1 pair routing +random*/
		case ROUTING_RA126:	/* Best effort fixed shortest 1+1 pair routing +random*/
		case ROUTING_RA127:	/* Best effort alternate shortest 1+1 pair routing +random*/
		{
		  /*	max_oswave = 0;*/
			  
			  /* looking for free wavelengths on the working path */
			  
			min_no_freewaves = 8 * max_wvoct_all_links + 1;
			
			opsecnum2 = 0;
			  link = pair->path2->path;
			  pair_is_ok = TRUE;
			  no_checked_links = 0;
			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave2[opsecnum2] = -1;
				  no_diff_freewaves = 0;
				  no_freewaves = 0;
				  max_ch_links = 0; 

				  for(i = 0; i < max_wvoct_all_links && sel_wave2[opsecnum2] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  
					  for(k=0; k<8; k++)
						  free_chans[k] = MAX_FIBER_NUM;

					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  
							  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k < link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }

							  if(link->endopsec)
								  endofos = TRUE;
						  }					  
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  actualwave = 8*i + 1;
						  k = 0;
						  while(wave_mask)
						  {
							  if(wave_mask & 0x80)
							  {
							  if (cg_routing->selected_routing == ROUTING_RA120 ||
								  cg_routing->selected_routing == ROUTING_RA123 ||
								  cg_routing->selected_routing == ROUTING_RA126 ||
								  cg_routing->selected_routing == ROUTING_RA127)
								{
							  
							  /* Increment the number of free waves on the link
								  and note the wavelength number */
								  for (j = 0;j<free_chans[k];j++)
									{
										tfree[no_freewaves] = actualwave;
										no_freewaves++;
									}/*for*/

								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
								}
							  
							  else 
								{
								/* Increment the number of free waves on the link
								  and note the wavelength number */
								  tfree[no_diff_freewaves] = actualwave;
								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
								  no_freewaves += free_chans[k];
								}
							  
								no_diff_freewaves++;
							  }
							  k++;
							  wave_mask <<= 1;
							  actualwave++;
						  }
						  
					  } /*if*/
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  } /*for*/
				  
				  /* choose a random element of the free wavelength */
				  if (no_diff_freewaves != 0)
				  {
					  rand_0_1 = uniform_0_1(&sh_seed);
					  
					  /* to avoid the owerflow of tfree */
					  if (rand_0_1 == 1) rand_0_1 = 0;
					 
					  
						if (cg_routing->selected_routing == ROUTING_RA120 ||
							cg_routing->selected_routing == ROUTING_RA123 ||
							cg_routing->selected_routing == ROUTING_RA126 ||
							cg_routing->selected_routing == ROUTING_RA127)
						{
							sel_wave2[opsecnum2] = 
								tfree[(int)floor(no_freewaves*rand_0_1)];
						}
						else
							sel_wave2[opsecnum2] = 
								tfree[(int)floor(no_diff_freewaves*rand_0_1)];
				  }
				  else
				  {
					  /* no waves are available on the link */
					  sel_wave2[opsecnum2] = -1;
					  pair_is_ok = FALSE;
				  }
				  
				  if(min_no_freewaves > no_freewaves)
					  min_no_freewaves = no_freewaves;
				  opsecnum2 ++;
				  no_checked_links += max_ch_links;
			  } /* while */
					  
			  
			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  /* looking for free wavelengths on the protective path */
			  
			  opsecnum1 = 0;
			  link = pair->path1->path;
			  pair_is_ok = TRUE;
			  no_checked_links = 0;
			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave1[opsecnum1] = -1;
				  no_diff_freewaves = 0;
				  no_freewaves = 0;
				  max_ch_links = 0; 

				  for(i = 0; i < max_wvoct_all_links && sel_wave1[opsecnum1] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  
					  for(k=0; k<8; k++)
						  free_chans[k] = MAX_FIBER_NUM;

					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  
							  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k < link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }

							  if(link->endopsec)
								  endofos = TRUE;
						  }					  
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  actualwave = 8*i + 1;
						  k = 0;
						  while(wave_mask)
						  {
							  if(wave_mask & 0x80)
							  {
							  if (cg_routing->selected_routing == ROUTING_RA120 ||
								  cg_routing->selected_routing == ROUTING_RA123 ||
								  cg_routing->selected_routing == ROUTING_RA126 ||
								  cg_routing->selected_routing == ROUTING_RA127)
								{
							  
							  /* Increment the number of free waves on the link
								  and note the wavelength number */
								  for (j = 0;j<free_chans[k];j++)
									{
										tfree[no_freewaves] = actualwave;
										no_freewaves++;
									}/*for*/

								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
								}
							  
							  else 
								{
								/* Increment the number of free waves on the link
								  and note the wavelength number */
								  tfree[no_diff_freewaves] = actualwave;
								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
								  no_freewaves += free_chans[k];
								}
							  
								no_diff_freewaves++;
							  }
							  k++;
							  wave_mask <<= 1;
							  actualwave++;
						  }
						  
					  } /*if*/
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  } /*for*/
				  
				  /* choose a random element of the free wavelength */
				  if (no_diff_freewaves != 0)
				  {
					  rand_0_1 = uniform_0_1(&sh_seed);
					  
					  /* to avoid the owerflow of tfree */
					  if (rand_0_1 == 1) rand_0_1 = 0;
					 
					  
						if (cg_routing->selected_routing == ROUTING_RA120 ||
							cg_routing->selected_routing == ROUTING_RA123 ||
							cg_routing->selected_routing == ROUTING_RA126 ||
							cg_routing->selected_routing == ROUTING_RA127)
						{
							sel_wave1[opsecnum1] = 
								tfree[(int)floor(no_freewaves*rand_0_1)];
						}
						else
							sel_wave1[opsecnum1] = 
								tfree[(int)floor(no_diff_freewaves*rand_0_1)];
				  }
				  else
				  {
					  /* no waves are available on the link */
					  sel_wave1[opsecnum1] = -1;
					  pair_is_ok = FALSE;
				  }
				  
				  if(min_no_freewaves > no_freewaves)
					  min_no_freewaves = no_freewaves;
				  opsecnum1 ++;
				  no_checked_links += max_ch_links;
			  } /* while */
			  
			  
			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  more_pairs = FALSE;
			  if(pair_is_ok)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  for(i=0; i<opsecnum1; i++)
					  lowest_sw1[i] = sel_wave1[i];
				  for(i=0; i<opsecnum2; i++)
					  lowest_sw2[i] = sel_wave2[i];
				  pair_found = TRUE;
				  new_prot_chan = TRUE;
				  dedicated_protchan = TRUE;
			  }
			  else
			  {
				  /* The alternate shortest path searchs till a route is found */
				  if(cg_routing->selected_routing == ROUTING_RA123 ||
					 cg_routing->selected_routing == ROUTING_RA127)
				  {
					  pair = iterate_pair_list(pair, curr_pair_no,
						  cg_routing,	primary_pair);
					  if(pair != NULL)
						  more_pairs = TRUE;
				  }
			  }
		}
		break;

/* Routing algorithms with equal wavelength on the working and protective 
		path+random wavelength assignment. */
	  case ROUTING_RA124:	/* Fixed shortest 1+1 pair with ewl routing */
      case ROUTING_RA125:	/* Alternate shortest 1+1 pair with ewl routing */
      case ROUTING_RA128:	/* Best effort fixed shortest 1+1 pair with ewl routing */
      case ROUTING_RA129:	/* Best effort alternate shortest 1+1 pair with ewl routing */
		  {
			  /* Do not care about opaque nodes! */
			  
			  min_no_freewaves = 8 * max_wvoct_all_links + 1;

			  /* The node-to-node links are checked for the free wave */
			  		  
			  sel_wave2[0] = -1;
			  no_diff_freewaves = 0;
			  no_freewaves = 0;
			  max_ch_links = 0; 
			  no_checked_links = 0;
			  
			  for(i = 0; i < max_wvoct_all_links && sel_wave2[0] == -1; i++)
			  {
				  wave_mask = 0xFF;
				  for(k=0; k<8; k++)
						  free_chans[k] = MAX_FIBER_NUM;

				  /* looking for free wavelengths on the working path */
				  link = pair->path2->path;
				  ch_links = 0;
				  while(link != NULL && wave_mask)
				  {
					  /* Comparing the free_wave mask of every node-node link */
					  if(link->link_elem->is_node_node)
					  {
						  ch_links++;							  
						  wave_mask &= link->link_elem->free_waves[i];
						  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k < link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }
					  }
					  link = link->next;
				  }
				  /* looking for free wavelengths on the protective path */
				  link = pair->path1->path;
				  while(link != NULL && wave_mask)
				  {
					  /* Comparing the free_wave mask of every node-node link */
					  if(link->link_elem->is_node_node)
					  {
						  ch_links++;							  
						  wave_mask &= link->link_elem->free_waves[i];
						  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k <= link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }
					  }
					  link = link->next;
				  }
				  
				  if( wave_mask )
				  {
					
					actualwave = 8*i + 1;
						  k = 0;
						  while(wave_mask)
						  {
							  if(wave_mask & 0x80)
							  {
							  						  
							  /* Increment the number of free waves on the link
								  and note the wavelength number */
								  for (j = 0;j<free_chans[k];j++)
									{
										tfree[no_freewaves] = actualwave;
										no_freewaves++;
									}/*for*/

								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
															  
							  no_diff_freewaves++;
							  }
							  k++;
							  wave_mask <<= 1;
							  actualwave++;
						  }
						  
				  } /*if*/
				  if(ch_links > max_ch_links)
					  max_ch_links = ch_links;
			  } /*for*/

			  no_checked_links += max_ch_links;

			  /* choose a random element of the free wavelength */
			  if (no_diff_freewaves != 0)
			  {
				  rand_0_1 = uniform_0_1(&sh_seed);
				  
				  /* to avoid the owerflow of tfree */
				  if (rand_0_1 == 1) rand_0_1 = 0;
				  
				  
				  sel_wave2[0] = 
					  tfree[(int)floor(no_freewaves*rand_0_1)];
				  
			  }
			  else
			  {
				  /* no waves are available on the link */
				  sel_wave2[0] = -1;
				  path_is_ok = FALSE;
			  }
			  
			  if(min_no_freewaves > no_freewaves)
				  min_no_freewaves = no_freewaves;
			  

			  if(no_checked_links > pair->path2->no_of_hops)
			  {
				  max_ch_links = pair->path2->no_of_hops;
				  no_checked_links -= pair->path2->no_of_hops;
			  }
			  else
			  {
				  max_ch_links = no_checked_links;
				  no_checked_links = 0;
			  }

			  /* Working links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next;
				i < max_ch_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  /* Protective links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }
			  
			  more_pairs = FALSE;
			  if(sel_wave2[0] != -1)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  lowest_sw1[0] = sel_wave2[0];
				  lowest_sw2[0] = sel_wave2[0];
				  pair_found = TRUE;
				  new_prot_chan = TRUE;
				  dedicated_protchan = TRUE;
			  }
			  else
			  {
				  /* The alternate shortest path searchs till a route is found */
				  if(cg_routing->selected_routing == ROUTING_RA125 ||
					  cg_routing->selected_routing == ROUTING_RA129)
				  {
					  pair = iterate_pair_list(pair, curr_pair_no,
						  cg_routing,	primary_pair);
					  if(pair != NULL)
						  more_pairs = TRUE;
				  }
			  }
		  }
		  break;

		case ROUTING_RA133:	/* Alternate shortest shared pair routing + random 
			wavelength assignment shortest means the preference of the pair
			which allocates the less new wl resources */
		case ROUTING_RA137:	/* Best effort alternate shortest 1+1 pair routing + 
			random wavelength assignment */
		{
			  /* looking for free wavelengths on the working path */

			  /* we need a temporary list of the wavelengths, because it is
			     not sure that this path (and this pair) will be chosen */
			  
			min_no_freewaves = 8 * max_wvoct_all_links + 1;
			opsecnum2 = 0;
			  link = pair->path2->path;
			  /* If the working path is shorter than the path found
			     as "shortest" (less resource needed) */
			  if( ( cg_routing->path_order_type == IN_ORDER_OF_HOP_NUMBER &&
				  pair->path2->no_of_hops < min_new_hops) ||
				  ( cg_routing->path_order_type == IN_ORDER_OF_LENGTH &&
				  pair->path2->length < min_new_length ))
				  pair_is_ok = TRUE;
			  else
				  pair_is_ok = FALSE;

			  curr_new_prot_chan = FALSE;
			  max_prot = 0;
			  not_first_node = FALSE;
			  no_checked_links = 0;

			  /* The opaque sections are handled sequentially */
			  while(link != NULL && pair_is_ok)
			  {
				  /* The node-to-node links are checked for the free wave */
				  os_first_link = link;
				  sel_wave2[opsecnum2] = -1;
				  no_diff_freewaves = 0;
				  no_freewaves = 0;
				  /* Reset the maximum of checked links for the next opsec */
				  max_ch_links = 0; 
				  for(i = 0; i < max_wvoct_all_links && 
					  sel_wave2[opsecnum2] == -1; i++)
				  {
					  wave_mask = 0xFF;
					  
					  for(k=0; k<8; k++)
						  free_chans[k] = MAX_FIBER_NUM;
					  
					  endofos = FALSE;
					  link = os_first_link;
					  ch_links = 0;
					  while(!endofos && link != NULL)
					  {
						  /* Comparing the free_wave mask of every node-node link */
						  if(link->link_elem->is_node_node)
						  {
							  /* Count the links checked */
							  if(wave_mask)
								  ch_links++;							  
							  wave_mask &= link->link_elem->free_waves[i];
							  for(k = 0; k < 8; k++)
							  {
								  if(8*i + k < link->link_elem->max_waves)
								  {
									  if(free_chans[k] > 
										  link->link_elem->free_fibers[8*i+k])
										  free_chans[k] =
										  link->link_elem->free_fibers[8*i+k];
								  }
							  }
							  if(link->endopsec)
								  endofos = TRUE;
							  if(pair->path1->
								  protected_links[link->link_elem->nnid]+1 > max_prot)
								  max_prot = 
								  pair->path1->protected_links[link->link_elem->nnid]+1;

							  /* the starting point of the link */
							  if(link->direction)
								  start_node = link->link_elem->dest;
							  else
								  start_node = link->link_elem->source;

							  if(not_first_node && pair->path1->
								  protected_nodes[start_node]+1 > max_prot)
								  max_prot = 
								  pair->path1->protected_nodes[start_node]+1;
							  not_first_node = TRUE;

						  }					  
						  link = link->next;
					  }
					  if( wave_mask )
					  {
						  actualwave = 8*i + 1;
						  k = 0;
						  while(wave_mask)
						  {
							  if(wave_mask & 0x80)
							  {
							    /* Increment the number of free waves on the link
								  and note the wavelength number */
								  for (j = 0;j<free_chans[k];j++)
									{
										tfree[no_freewaves] = actualwave;
										no_freewaves++;
									}/*for*/

								  
							  /* The number of the available channels can be 
							      different from the number of different free 
								  waves (one wave -> more fibers -> more channels) */
								no_diff_freewaves++;
							  }
							  k++;
							  wave_mask <<= 1;
							  actualwave++;
						  }
						  
					  } /*if*/
					  if(ch_links > max_ch_links)
						  max_ch_links = ch_links;
				  } /*for*/
				  
				  /* choose a random element of the free wavelength */
				  if (no_diff_freewaves != 0)
				  {
					  rand_0_1 = uniform_0_1(&sh_seed);
					  
					  /* to avoid the owerflow of tfree */
					  if (rand_0_1 == 1) rand_0_1 = 0;
					 
					  
						sel_wave2[opsecnum2] = 
						tfree[(int)floor(no_freewaves*rand_0_1)];
						
				  }
				  else
				  {
					  /* no waves are available on the link */
					  sel_wave2[opsecnum2] = -1;
					  path_is_ok = FALSE;
				  }
				  
				  if(min_no_freewaves > no_freewaves)
					  min_no_freewaves = no_freewaves;
				  opsecnum2 ++;
				  no_checked_links += max_ch_links;
			  } /* while */

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path2->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }

			  if(pair_is_ok)
			  {
				  /* If new protection channel has to be opened */
				  if(max_prot > pair->path1->num_of_prot_chs)
					  curr_new_prot_chan = TRUE;

				  /* Check if this pair could be the "shortest" */
				  if( curr_new_prot_chan &&
					 !(cg_routing->path_order_type == IN_ORDER_OF_HOP_NUMBER &&
					  pair->path1->no_of_hops + 
					  pair->path2->no_of_hops < min_new_hops) &&
					 !(cg_routing->path_order_type == IN_ORDER_OF_LENGTH &&
					  pair->path1->length + 
					  pair->path2->length < min_new_length))
				 /* if it would be longer, don't choose this path */
					  pair_is_ok = FALSE;
					  
			  /*the case of only working path was already checked */
			  }

			  /* looking for free wavelengths on the protective path */
			  opsecnum1 = 0;
			  link = pair->path1->path;
			  no_checked_links = 0;
			  /* If needed, the opaque sections are handled sequentially */
			  if(curr_new_prot_chan)
			  {		  
				  while(link != NULL && pair_is_ok)
				  {
					  /* The node-to-node links are checked for the free wave */
					  os_first_link = link;
					  sel_wave1[opsecnum1] = -1;
					  max_ch_links = 0; 
					  for(i = 0; i < max_wvoct_all_links && sel_wave1[opsecnum1] == -1; i++)
					  {
						  wave_mask = 0xFF;
						  endofos = FALSE;
						  link = os_first_link;
						  ch_links = 0;
						  while(!endofos && link != NULL)
						  {
							  /* Comparing the free_wave mask of every node-node link */
							  if(link->link_elem->is_node_node)
							  {
								  /* Count the links checked */
								  if(wave_mask)
									  ch_links++;							  
								  wave_mask &= link->link_elem->free_waves[i];
								  if(link->endopsec)
									  endofos = TRUE;
							  }
							  link = link->next;
						  }
						  if( wave_mask )
						  {
							  /* The numberring of the wavelengths starts with 1 */
							  sel_wave1[opsecnum1] = (8 * i) + 1;
							  /* shifting  while the first 1 is found */
							  while(! (wave_mask & 0x80))
							  {
								  sel_wave1[opsecnum1]++;
								  wave_mask <<= 1;
							  }
						  }
						  if(ch_links > max_ch_links)
							  max_ch_links = ch_links;
					  }/*for*/
					  if(sel_wave1[opsecnum1] == -1)
						  pair_is_ok = FALSE;
						  /*			if(sel_wave[opsecnum] > max_oswave)
					  max_oswave = sel_wave[opsecnum];*/
					  opsecnum1 ++;
					  no_checked_links += max_ch_links;
				  } /* while */
			  }

			  /* Links are put in the union of paths */
			  for(i = 0, link = pair->path1->path->next;
				i < no_checked_links; i++)
			  {
				  in_path_union[link->link_elem->id] ++;
				  link = link->next;
			  }

			  /* "Shorter" path found */
			  if(pair_is_ok)
			  {
				  sel_pair = pair;
				  sel_pair_no = (*curr_pair_no);
				  for(i=0; i<opsecnum2; i++)
					  lowest_sw2[i] = sel_wave2[i];

				  new_prot_chan = curr_new_prot_chan;

				  /* New channel has to be open */
				  if(curr_new_prot_chan)
				  {
					  for(i=0; i<opsecnum1; i++)
						  lowest_sw1[i] = sel_wave1[i];
					  min_new_hops = pair->path1->no_of_hops + 
								  pair->path2->no_of_hops;
					  min_new_length = pair->path1->length + 
								  pair->path2->length;
				  }
				  else
				  {
					  min_new_hops = pair->path1->no_of_hops;
					  min_new_length = pair->path1->length;
				  }

				  pair_found = TRUE;
			  }

			  more_pairs = FALSE;
			  pair = iterate_pair_list(pair, curr_pair_no,
				  cg_routing,	primary_pair);
			  if(pair != NULL)
				  more_pairs = TRUE;
		}
		break;      

		default:
			{
				/* Something went wrong  */
				printf("ERROR: unknown routing algorithm: %d \n",
					cg_routing->selected_routing);
				exit(1);
			}
			break;
     } /* switch */
	} while (more_pairs);

	if(pair_found) /* In this case we found a free wavelength */
	{   /* on the links of the working path of the sel_pair. 
		Now we allocate it on the links. */
		opsecnum2 = 0;
		link = sel_pair->path2->path;
		j=0;
		not_first_node = FALSE;

		/* The opaque sections are handled sequentially */
		while(link != NULL)
		{
			/* We allocate the wavelength on the node-to-node
			links of the opaque section*/
			os_first_link = link;
			endofos = FALSE;
			alloc_mask = 0x80;
			for(i = 0; i < (lowest_sw2[opsecnum2]-1)%8; i++)
				alloc_mask >>= 1;
			while(!endofos && link != NULL)
			{
				if(link->link_elem->is_node_node)
				{
					/* the starting point of the link */
					if(link->direction)
						start_node = link->link_elem->dest;
					else
						start_node = link->link_elem->source;

					if( !(--link->link_elem->free_fibers[lowest_sw2[opsecnum2]-1]) )
						link->link_elem->free_waves[(lowest_sw2[opsecnum2]-1)/8] 
						&= (~alloc_mask);

					/* Statistics of the wavelength opened for the new call */
					calls_on_link_wave[link->link_elem->id]
						[lowest_sw2[opsecnum2]]++;
					
					if(link->endopsec)
						endofos = TRUE;
					/* update the protection of link and its starting node */
					sel_pair->path1->protected_links[link->link_elem->nnid]++;
					if(not_first_node)
						sel_pair->path1->protected_nodes[start_node]++;
					not_first_node = TRUE;
				}
				if(lowest_sw2[opsecnum2] > link->link_elem->peak_wave)
					link->link_elem->peak_wave = lowest_sw2[opsecnum2];
				link->link_elem->used_waves++;
				u_wave[j++] = lowest_sw2[opsecnum2];
				link = link->next;
			}
			opsecnum2++;
		} /* while */

		/* We have to allocate a new channel on the protective path. */
		if(new_prot_chan)
		{
			opsecnum1 = 0;
			link = sel_pair->path1->path;
			j=0;
			/* Create a new protection channel */
			new_prot_ch = prot_ch_alloc();
			prev_intlist = NULL;

			/* The opaque sections are handled sequentially */
			while(link != NULL)
			{
			/* We allocate the wavelength on the node-to-node
				links of the opaque section*/
				os_first_link = link;
				endofos = FALSE;
				alloc_mask = 0x80;
				for(i = 0; i < (lowest_sw1[opsecnum1]-1)%8; i++)
					alloc_mask >>= 1;
				while(!endofos && link != NULL)
				{
					if(link->link_elem->is_node_node)
					{
						if( !(--link->link_elem->free_fibers[lowest_sw1[opsecnum1]-1]) )
							link->link_elem->free_waves[(lowest_sw1[opsecnum1]-1)/8] 
							&= (~alloc_mask);

						/* Statistics of the wavelength opened for the new call */
						calls_on_link_wave[link->link_elem->id]
							[lowest_sw1[opsecnum1]]++;
						
						if(link->endopsec)
							endofos = TRUE;
					}
					if(lowest_sw1[opsecnum1] > link->link_elem->peak_wave)
						link->link_elem->peak_wave = lowest_sw1[opsecnum1];
					link->link_elem->used_waves++;
					link = link->next;
				}
				/* link a new wavelength of transparent section
				at the end of the list */
				new_intlist = int_list_alloc();
				new_intlist->value = lowest_sw1[opsecnum1];
				new_intlist->next = NULL;
				if(!prev_intlist)
					new_prot_ch->opsec_waves = new_intlist;
				else
					prev_intlist->next = new_intlist;
				prev_intlist = new_intlist;

				opsecnum1++;

			} /* while */
			/* insert the new channel description */
			new_prot_ch->next = sel_pair->path1->prot_channels;
			sel_pair->path1->prot_channels = new_prot_ch;
			sel_pair->path1->num_of_prot_chs++;
		}


	}
	else
		/* Best effort routing algorithms -> we simple call the single path
	routing algorithms with a virtual routing structure */
	{
		if( cg_routing->selected_routing == ROUTING_RA76 ||
			cg_routing->selected_routing == ROUTING_RA77 ||
			cg_routing->selected_routing == ROUTING_RA78 ||
			cg_routing->selected_routing == ROUTING_RA79 ||
			cg_routing->selected_routing == ROUTING_RA87 ||
			cg_routing->selected_routing == ROUTING_RA126 ||
			cg_routing->selected_routing == ROUTING_RA127 ||
			cg_routing->selected_routing == ROUTING_RA128 ||
			cg_routing->selected_routing == ROUTING_RA129 ||
			cg_routing->selected_routing == ROUTING_RA137)
		{
			virt_routing.selected_range_type.type =
				cg_routing->selected_range_type.type;
			virt_routing.selected_range_type.value.dval =
				cg_routing->selected_range_type.value.dval;
			virt_routing.selected_range_type.value.ival =
				cg_routing->selected_range_type.value.ival;
			virt_routing.path_order_type =
				cg_routing->path_order_type;
			virt_routing.use_hop_by_hop =
				cg_routing->use_hop_by_hop;
			virt_routing.use_crankback =
				cg_routing->use_crankback;
			virt_routing.primary_threshold =
				cg_routing->primary_threshold;
			virt_routing.alternate_threshold =
				cg_routing->alternate_threshold;

			switch (cg_routing->selected_routing)
			{
			case ROUTING_RA76:	/* Best effort fixed shortest 1+1 pair routing */
			case ROUTING_RA78:	/* Best effort fixed shortest 1+1 pair with ewl routing */
				{
					virt_routing.selected_routing = ROUTING_RA50;
				}
				break;
			case ROUTING_RA77:	/* Best effort alternate shortest 1+1 pair routing */
			case ROUTING_RA79:	/* Best effort alternate shortest 1+1 pair with ewl routing */
			case ROUTING_RA87:	/* Best effort alternate shortest shared pair routing */
				{
					virt_routing.selected_routing = ROUTING_RA53;
				}
				break;
			case ROUTING_RA126:	/* Best effort fixed shortest 1+1 pair routing + random */
			case ROUTING_RA128:	/* Best effort fixed shortest 1+1 pair with ewl routing + random */
				{
					virt_routing.selected_routing = ROUTING_RA90;
				}
				break;
			
			case ROUTING_RA127:	/* Best effort alternate shortest 1+1 pair routing + random */
			case ROUTING_RA129:	/* Best effort alternate shortest 1+1 pair with ewl routing + random */
			case ROUTING_RA137:	/* Best effort alternate shortest shared pair routing + random */
				{
					virt_routing.selected_routing = ROUTING_RA93;
				}
				break;
			
			default:
				;
			}
			
			be_path = path_mat[virt_routing.path_order_type]
				[cgen_src][cgen_dest];
			
			curr_path_no = 1;
			
			sel_path = ason_routing(cgen_src, cgen_dest, &virt_routing,
				&curr_path_no, be_path);
			(*curr_pair_no) = curr_path_no;
			if(sel_path)
				protection_be = TRUE;
			return sel_path;
		}
	}

	(*curr_pair_no) = sel_pair_no;
	if(sel_pair)
	{
		used_prot_path = sel_pair->path1;
		return sel_pair->path2;

	}
	else
	{
		return NULL;
	}
}
