/*
**  Module    : calladco.c
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1999/01/14 14:59:00 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains the functions dealing with the
**              available Call Admission Control algorithms.
**
**  Functions : double  eval_equivalent_bw(double, double, double, double)
**              double  eval_modified_bw(double, double)
**              double  eval_call_bw_sample(CallList *, double, int)
**              double  eval_real_time_bw(CallList *, double, Boolean*, int)
**              void    init_CAC1(void)
**              Boolean exe_CAC1(Link *, double, double, double)
**              void    init_CAC2(void)
**              Boolean exe_CAC2(Link *, double, double, double)
**              void    init_CAC3(void)
**              Boolean exe_CAC3(Link *, double, double, double)
**              void    init_CAC4(void)
**              Boolean exe_CAC4(Link *, double, double, double)
**              void    init_Ucall(void)
**              void    update_Ucall_close(Call *, Link *)
**              double  update_Ucall_open(Link *, double , double)
**
**  History   : 19-OCT-1995 - Creation of the module.   -AF-
**              01-JUN-1996 - Revision.                 -AF-
**              01-MAR-1997 - Revision			-MM-

31-01-2001 Cleaning... -ZZ-
*/
 
# ifndef        ANCLES_H
#include        "ancles.h"
# endif
 
/***********************
**  Global Variables  **
***********************/
 
static TimeUnit last_req_time[MAX_LINK];/* Absolute Simulation Time of  */
                                        /* the last received connection */
                                        /* request.                     */
static int  next_sample_index[MAX_LINK];
                                        /* Location of the next item to */
                                        /* be inserted into the circular*/
                                        /* buffer collecting the load   */
                                        /* measures.                    */
static double aux_freq;

static double ** P_11;	/* This matrix stores the probability of remaining */
			/* in the ON state for a VBR_ON_OFF call generator */
			/* it is calculated once, because this probability */
			/* depends only on the link speed and the statistic*/
			/* parameter of the call generators                */
			/* Its rows refer to existing call_generators,     */
			/* while its columns refer to the links            */ 
static double ** P_22;	/* Ditto, but this matrix stores the prob. of 	   */
			/* remaining in the OFF state.		 	   */

/***************************
**  Function Prototypes   **
***************************/
 
double  eval_call_bw_sample(CallList *, double, int );
double  eval_real_time_bw(CallList *, double, Boolean *, int );
double  eval_gauss_sample(long *, double, double);
 
/***************************
**  Function Definitions  **
***************************/
 
/*
**  Function  : double eval_equivalent_bw(double mean, double peak,
**                                        double ref_cap, double p_loss)
**  Return    : the equivalent bandwidth of a call.
**  Remarks   : for the evaluation of the equivalent bandwidth the values
**              of the following four parameters have to be specified:
**              - mean    : it is the average bit rate of the call;
**              - peak    : it is the peak rate of the call;
**              - ref_cap : it is the capacity of the link over wich the
**                          call has been routed;
**              - p_loss  : it is the target cell loss probability.
**
**              The equivalent bandwidth is evaluated on the basis of the
**              flow chart appearing in "Multi-Rate Models for Dimensioning
**              and Performance Evaluation of ATM Networks," COST 242
**              Interim Report, June 1994, page 15.
**              The equivalent bandwidth of a call generator is also
**              evaluated through the same algorithm. The capacity of the link
**              connecting the corresponding user to its network access
**              node is used as the capacity required by the algorithm.
*/
double eval_equivalent_bw(double mean, double peak, double ref_cap,
                                                    double p_loss)
{
 double x, z, a, eqbw; 

 if (mean != UNDEFINED_BW)
   {
     x = mean / peak;
     z = -2.0 * log10(p_loss);
     a = 1.0 + z / 100.0;
     z = z / (ref_cap / peak);
     if(z < 1.0 || equ_double(z, 1.0, DBL_CMP_THS))
       {
         if(x < (1.0 / (3.0 * z)) 
	    || equ_double(x, 1.0 / (3.0 * z), DBL_CMP_THS))
            eqbw = (a * mean * (1.0 + 3.0 * z * (1.0 - x)));
         else
            eqbw = (a * peak);
       }
    else
       {
         if(x < (1.0 / (3.0 * z * z)) 
	    || equ_double(x, 1.0 / (3.0 * z * z), DBL_CMP_THS))
           eqbw = (a * mean * (1.0 + 3.0 * z * z * (1.0 - x)));
        else
           eqbw = (a * peak);
       }
     return eqbw;
  }
  else /* chiamata di classe U: il picco e' la sua BW equival. */
    return peak;
}

/*
**  Function  : double eval_modified_bw(double mean, double peak)
**  Return    : the modified bandwidth of a call.
**  Remarks   : for the evaluation of the modified bandwidth the values
**              of the following four parameters have to be specified:
**              - mean    : it is the average bit rate of the call;
**              - peak    : it is the peak rate of the call;
**
*/
double eval_modified_bw(double mean, double peak)
{
    return(mean * ( 1.0 + Gamma * (1.0 - mean/peak)));
}
 
 
/*
**  Function  : void updt_var_mean(Call *call_ptr, double *acc_mean,
**                                 Boolean add)
**  Return    : none.
**  Remarks   : the mean of the load contribution provided by 'call_ptr' is
**              added to the accumulator pointed by 'acc_mean' if 'add' is
**              TRUE whereas it is subtracted if 'add' is FALSE.
**              The same rule applies to the variance of the load
**              contribution provided by the call pointed by 'call_ptr'.
*/
void updt_var_mean(Call *call_ptr, double *acc_mean, double *acc_var,
                     Boolean add)
{
 double loc_mean, loc_var;
 CallType1 *ci1;
 CallType2 *ci2;
 CallType3 *ci3;
 CallType4 *ci4;
 
 switch (call_ptr->call_type)
   {
    case CALL_TYPE1 :
        ci1 = (CallType1 *)(call_ptr->info);
	loc_mean = ci1->cbr_value;
        loc_var = 0.0;
        break;
 
    case CALL_TYPE2 :
	ci2 = (CallType2 *)(call_ptr->info);
        loc_mean = ci2->peak_rate * ci2->mean_on_time /
                   ( ci2->mean_on_time + ci2->mean_off_time );
        loc_var = loc_mean * (ci2->peak_rate - loc_mean);
        break;
 
    case CALL_TYPE3 :
	ci3 = (CallType3 *)(call_ptr->info);
        loc_mean = ( ci3->min_vbr_value + ci3->max_vbr_value ) / 2.0;
        loc_var = pow( ci3->max_vbr_value - ci3->min_vbr_value, 2.0) / 12.0;
        break;
 
    case CALL_TYPE4 :
	ci4 = (CallType4 *)(call_ptr->info);
        loc_mean = ci4->higher_bit_rate_prob * ci4->higher_bit_rate_value +
                   ci4->medium_bit_rate_prob * ci4->medium_bit_rate_value +
                   ci4->lower_bit_rate_prob * ci4->lower_bit_rate_value;
        loc_var =  pow( ci4->higher_bit_rate_value - loc_mean, 2.0) *
                   ci4->higher_bit_rate_prob + 
		   pow( ci4->medium_bit_rate_value - loc_mean, 2.0) *
                   ci4->medium_bit_rate_prob +
                   pow( ci4->lower_bit_rate_value - loc_mean, 2.0) *
                   ci4->lower_bit_rate_prob;
        break;

   default:
     printf("ERROR: unknown call type %d in function updt_var_mean\n",
	    call_ptr->call_type);
     exit(1);
   }
 if(add == TRUE)
        {
         *acc_mean += loc_mean;
         *acc_var += loc_var;
        }
 else
        {
         *acc_mean -= loc_mean;
         *acc_var -= loc_var;
        }
}
 
 
/*
**  Function  : double eval_call_bw_sample(CallList *in_call, double link_bw,
**					   int link_id)
**  Return    : a load sample for the call pointed by 'in_call'.
**  Remarks   : the slow version of the CAC algorithm based on real time
**              traffic measurements (CAC 2) requires to identify the load
**              contribution of each connection supported by the link during
**              the slots composing the measurement frame.
**              The load contribution of the connection pointed by
**              'in_call' is evaluated here, taking into account that the
**              contribution value derives from a cell counting.
**              If the ratio between the instantaneous rate of the call and
**              the link bandwidth is under a fixed threshold the load
**              contribution is issued directly (without referring to the
**              integer number of cells accumulated during the slot),
**              because reliability problems could affect the measure in
**              this case.
*/
double eval_call_bw_sample(CallList *in_call, double link_bw, int link_id)
{
 double ref_rate;
 double dummy;
 Boolean under_ths;
 int no_cells;
 
 ref_rate = eval_real_time_bw(in_call, link_bw, &under_ths, link_id);
 
/* The value of 'under_ths' discriminates the way to evaluate the       */
/* reference bit rate for the call: if it is FALSE the rate is directly */
/* issued, whereas the conversion of the number of cells into           */
/* the transmission rate is made if it is TRUE.                         */
 if(under_ths == TRUE)
        return (ref_rate);
 no_cells = (int)floor(dummy = (ref_rate * meas_wind_size / link_bw));
 if(uniform_0_1(&sh_seed) < (dummy - no_cells))
        no_cells++;
 return (no_cells * link_bw / meas_wind_size);
}
 
 
/*
**  Function  : double eval_real_time_bw(CallList *call_point, double cap, 
**		Boolean *und_ths, int link_id)
**  Return    : the instantaneous value of the bandwidth currently used by
**              the call pointed by 'call_point'.
**  Remarks   : the instantaneous bandwidth of the call is obtained
**              according to the call type, as requested by CAC 2.
**              The issued value holds constant along the whole
**              measurement slot. The value takes into account the
**              measurement error contribution if a non zero value was
**              assigned to 'meas_rel_error' before starting the simulation.
**              The reference rate actually provided by the call is
**              compared with the link bandwidth in order to decide whether
**              the load contribution has to be expressed in terms of
**              accumulated cells or not. If the rate between the former
**              and the latter is such that the number of cells coming from
**              the call are less than 3.0% of the cells composing a slot,
**              the load contribution will be issued without passing
**              through the cell counting.
*/
double eval_real_time_bw(CallList *call_point, double cap, Boolean *und_ths,
			int link_id)
{
 double ref_rate;       /* Reference bit rate used to evaluate the call */
                        /* load contribution.                           */
 double err_contr;      /* Load contribution given by the error in the  */
                        /* real time traffic measurement.               */
 long *loc_seed;        /* Pointer to the seed for pseudo-random        */
                        /* sequences referring to the generator of the  */
                        /* call currently considered.                   */
 
 
 loc_seed = &cg_seed_vect[call_point->call_elem->source_id];
 
 switch (call_point->call_elem->call_type)
    {
     case CALL_TYPE1 :
        ref_rate = ((CallType1 *)(call_point->call_elem->info))->cbr_value;
 
/* Comparing the reference call rate with the link bandwidth.   */
        if(ref_rate < (cap * 3.0 / meas_wind_size))
                *und_ths = TRUE;
 
/* The measurement error contribution is added if 'meas_rel_error'      */
/* has been assigned a positive value.                                  */
        if(meas_rel_error > 0.0)
           {
            ref_rate += meas_rel_error *
                        uniform(-ref_rate, ref_rate, loc_seed);
            if(ref_rate < 0.0)  return 0.0;
           }
        return ref_rate;
 
     case CALL_TYPE2 :
       {
        CallType2 * call2; /* Auxiliary pointer to avoid multiple indirection*/
        call2 = (CallType2 *)(call_point->call_elem->info);
        ref_rate = call2->peak_rate;
 
/* Comparing the reference call rate with the link bandwidth.   */
        if(ref_rate < (cap * 3.0 / meas_wind_size))
                *und_ths = TRUE;
                
	if(call2->on_status)  /* the state is ON */
	  {
           /* I determine the new state through a random variable */
	   if ( uniform_0_1(loc_seed) > P_11[call2->cgen_id][link_id])
	      /* state changed to OFF */         
	      call2->on_status = FALSE;
	   else
	      /* state remains ON */
	   ;
	  }
	else /* the state is OFF */
	  {
           /* I determine the new state through a random variable */
	   if ( uniform_0_1(loc_seed) > P_22[call2->cgen_id][link_id])
	      call2->on_status = TRUE;
	      /* state changed to ON */
	   else
	      /* state remains OFF*/
	   ;
	  }
	if( call2->on_status )
	  {
/* The measurement error contribution is added if 'meas_rel_error'      */
/* has been assigned a positive value.                                  */
 	   if(meas_rel_error > 0.0)
	     {
              double on_prob = call2->mean_on_time / 
		(call2->mean_on_time + call2->mean_off_time);
	      err_contr = meas_rel_error * on_prob *
		     uniform(-ref_rate, ref_rate, loc_seed);
	      if(err_contr < 0.0)
	      err_contr *= 2.0;
	      return (ref_rate + err_contr);
	     }
	   return ref_rate;
	  }
 	else
	  {
/* The measurement error contribution is added if 'meas_rel_error'      */
/* has been assigned a positive value.                                  */
	   if(meas_rel_error > 0.0)
	     {	
              double on_prob = call2->mean_on_time / 
		(call2->mean_on_time + call2->mean_off_time);
	      err_contr = meas_rel_error * on_prob *
		     uniform(-ref_rate, ref_rate, loc_seed);
	      if(err_contr > 0.0)
 		 return err_contr;
	     }
	   return 0.0;
	  }
      }
    case CALL_TYPE3 :
      {
        CallType3 * call3;	/* Aux pointer to avoid multiple indirection */
	call3 = ((CallType3 *)(call_point->call_elem->info));
        ref_rate = (uniform(call3->min_vbr_value, call3->max_vbr_value,
                 	loc_seed));
 
/* Comparing the reference call rate with the link bandwidth.   */
        if(ref_rate < (cap * 3.0 / meas_wind_size))
                *und_ths = TRUE;
 
/* The measurement error contribution is added if 'meas_rel_error'      */
/* has been assigned a positive value.                                  */
        if(meas_rel_error > 0.0)
                {
                 ref_rate += (meas_rel_error *
                                uniform(-ref_rate, ref_rate, loc_seed));
                 if(ref_rate < 0.0)     return 0.0;
                }
        return ref_rate;
      } 
     case CALL_TYPE4 :
      {
        CallType4 * call4;	/* Aux pointer to avoid multiple indirection */
               /* The 'x' variable is used when a connection of*/
               /* type 4 is investigated. It is used to        */
               /* identify the bandwidth currently required    */
               /* by the connection, among a discrete          */
               /* collection of permitted values.              */
        double x = uniform_0_1(loc_seed);
     	call4 = ((CallType4 *)(call_point->call_elem->info));
        if(x > call4->lower_bit_rate_prob )
           {
            if(x > (call4->medium_bit_rate_prob + call4->lower_bit_rate_prob) )
                ref_rate = call4->higher_bit_rate_value;
            else
                ref_rate = call4->medium_bit_rate_value;
           }
        else
           ref_rate = call4->lower_bit_rate_value;
 
/* Comparing the reference call rate with the link bandwidth.   */
        if(ref_rate < (cap * 3.0 / meas_wind_size))
                *und_ths = TRUE;
 
/* The measurement error contribution is added if 'meas_rel_error'      */
/* has been assigned a positive value.                                  */
        if(meas_rel_error > 0.0)
           {
            ref_rate += (meas_rel_error *
                        uniform(-ref_rate, ref_rate, loc_seed));
            if(ref_rate < 0.0)  return 0.0;
           }
        return ref_rate;
      }

    default:
      printf("ERROR: unknown call type %d in function eval_real_time_bw\n",
	     call_point->call_elem->call_type);
      exit(1);
    }
}
 
 
/*
**  Function  : double eval_gauss_sample(long *seed, double mean,
**                                       double variance)
**  Return    : the load measured over the link on the basis of the
**              Gaussian approximation (CAC 3).
**  Remarks   : the sum of 12 independent variables uniformly distributed
**              between -0.5 and 0.5 is used to obtain a Normally
**              distributed variable.
*/
double eval_gauss_sample(long *seed, double mean, double variance)
{
 double acc_x = 0.0;
 int i = 0;
 
 for(; i < 12; i++)
         acc_x += uniform(-0.5, 0.5, seed);
 
 return (acc_x * sqrt(variance) + mean);
}
 
 
/*
**  Function  : int update_samples_vect(Link *link_pointer, Boolean need)
**  Return    : the number of new samples to be gathered.
**  Remarks   : this function evaluates the new load samples needed to
**              update the load measure referring to 'link_pointer'.
**              'now_and_forever' is used to start the updating of the array
**              when the direct measurement of the link load is really needed.
*/
int update_samples_vect(Link *link_pointer, Boolean now_and_forever)
{
 static Boolean must_update = FALSE;

 if(now_and_forever)
     must_update = TRUE;
 
 if(must_update)
    {
     double link_band = capacity / link_pointer->timescale;
     int max_samples;
     int link_id = link_pointer->id;
     register int i, kk;
     CallList *call_ptr;

     if((max_samples = (int)floor(aux_freq /
			(double)link_pointer->timescale /
                       (ancles_time - last_req_time[link_id]) )) >
                       no_load_samples)
         max_samples = no_load_samples;
 
     if(selected_cac == CALL_ADM_CTRL3)
        {
         for(i = 0; i < max_samples; i++)
           {
            if((kk = (i + next_sample_index[link_id])) >= no_load_samples)
                kk -= no_load_samples;
            load_samples_vect[link_id][kk] = eval_gauss_sample(&sh_seed,
                                             link_mean[link_id],
                                             link_variance[link_id]);
           }
        }
     else               /* CALL_ADM_CTRL2  */
        {

         for(i = 0; i < max_samples; i++)
           {
            if((kk = (i + next_sample_index[link_id])) >= no_load_samples)
                kk -= no_load_samples;
            load_samples_vect[link_id][kk] = 0.0;
            for(call_ptr = link_pointer->calls_on->next;
	        call_ptr != link_pointer->calls_on; call_ptr = call_ptr->next)
               {
                load_samples_vect[link_id][kk] +=
                           eval_call_bw_sample(call_ptr, link_band, link_id);
               }
           }   
        }
 
     if((next_sample_index[link_id] = (next_sample_index[link_id] + i)) >=
         no_load_samples)
          next_sample_index[link_id] -= no_load_samples;
 
     if(max_samples > 0)
          last_req_time[link_id] = ancles_time;
 
     return max_samples;
    }
 else
    return 0;
}
 
 
/*
**  Function  : void init_CAC1(void)
**  Return    : none.
**  Remarks   : this function configures all the entities involved by the
**              CAC algorithm based on the Equivalent Bandwidth
**              Allocation.
*/
void init_CAC1(void)
{
  init_Ucall();
}
 
 
/*
**  Function  : Boolean exe_CAC1(Link *link_pointer, double mean_band,
**                               double peak_band, double link_threshold)
**  Return    : TRUE if the call request can be accepted, FALSE otherwise.
**  Remarks   : this function drives all the actions required by the CAC
**              algorithm based on the Equivalent Bandwidth Allocation.
**              Its parameters are:
**                - 'link_pointer': the link under examination;
**                - 'mean_band': the declared average bandwidth of the
**                  source generator;
**                - 'peak_band': the declared peak bandwidth of the source
**                  generator;
**                - 'link_threshold': is the protection factor to use; zero if
**                  we are checking a primary path
*/
Boolean exe_CAC1(Link *link_pointer, double mean_band, double peak_band,
                 double link_threshold)
{
 double equivalent_bw;  /* Equivalent Bandwidth of the incoming         */
                        /* connection request.                          */
 equivalent_bw = eval_equivalent_bw(mean_band, peak_band,
                 capacity / link_pointer->timescale, cell_loss_rate);
 if((link_pointer->free_band - link_threshold) < equivalent_bw)
                return FALSE;
 return TRUE;
}
 
 
/*
**  Function  : void init_CAC2(void)
**  Return    : none.
**  Remarks   : this function configures all the entities involved by the
**              CAC algorithm based on the Peak Rate Allocation over Real
**              Time Traffic Estimations (slow version).
*/
void init_CAC2(void)
{
 int i, j;
 CGType2 *ci2;
 double delta_t, stationary, frequency;
 
 P_11 = (double **)(MMmalloc (no_of_cgens * sizeof(double*),
                              "init_CAC2 - P_11"));
 P_22 = (double **)(MMmalloc (no_of_cgens * sizeof(double*),
                              "init_CAC2 - P_22"));

 for (i=0; i< no_of_cgens; i++)
   {
    P_11[i] = (double *)(MMmalloc(no_of_links * sizeof(double),
                                  "calladco.c - P_11"));
    P_22[i] = (double *)(MMmalloc(no_of_links * sizeof(double),
                                  "calladco.c - P_22"));
   }
 aux_freq = capacity * 1000000.0 / (meas_wind_size * 53.0 * 8.0);
 for(i = 0; i < no_of_cgens; i++)
    for(j = 0; j < no_of_links; j++)
      {
       if( callgen_vect[i]->type == CG_TYPE2 )
	 { 
	  delta_t = meas_wind_size * 53.0 * 8.0 / capacity / 1000000.0 *
			link_vect[j]->timescale;

	  ci2 = (CGType2 *)(callgen_vect[i]->info);
	  stationary = ci2->mean_off_time /
	 	  ( ci2->mean_on_time + ci2->mean_off_time );

	  frequency = - ( ci2->mean_on_time + ci2->mean_off_time )/
	   	  ( ci2->mean_on_time * ci2->mean_off_time );

	  P_11[i][j] = (1.0 - stationary) + stationary *
						exp(frequency * delta_t);   
	  P_22[i][j] = stationary + (1.0 - stationary) *
						exp(frequency * delta_t);   
	 }
       else
	 {
	  P_11[i][j] = 0.0;
	  P_22[i][j] = 0.0;
	 }
      }

 for(i = 0; i < no_of_links; i++)
   {
    backoff_ths[i] = capacity / link_vect[i]->timescale;
    next_sample_index[i] = 0;
    last_req_time[i] = 0.0;
    load_samples_vect[i] = (double *)MMmalloc(no_load_samples * sizeof(double),
                                              "init_CAC2 - load_samples_vect");
    for(j = 0; j < no_load_samples; j++)
        load_samples_vect[i][j] = 0.0;
   }
}
 
 
/*
**  Function  : Boolean exe_CAC2(Link *link_pointer, double mean_band,
**                               double peak_band, double link_threshold)
**  Return    : TRUE if the call request can be accepted, FALSE otherwise.
**  Remarks   : this function drives all the actions required by the CAC
**              algorithm based on the Peak Rate Allocation over Real
**              Time Traffic Estimations (slow version).
**              Its parameters are:
**                - 'link_pointer': the link under examination;
**                - 'mean_band': the declared average bandwidth of the
**                  source generator;
**                - 'peak_band': the declared peak bandwidth of the source
**                  generator;
**                - 'link_threshold': the protection factor to use; zero if
**                  we are checking a primary path.
*/
Boolean exe_CAC2(Link *link_pointer, double mean_band, double peak_band,
                 double link_threshold)
{
 register int i;
 
 double avg_link_load;  /* Mean value of the link load.                 */
 double var_link_load;  /* Variance of the link load.                   */
 double link_band;      /* Total bandwidth of the link.                 */
 double loc_prot_fact;  /* Local value of the protection factor applied */
                        /* to the link capacity.                        */
 
 double acc_acc_samples = 0.0;          /* Accumulator for the load     */
                                        /* samples.                     */
 double acc_sq_acc_samples = 0.0;       /* Accumulator for the squares  */
                                        /* of the load samples.         */
 int link_id = link_pointer->id;    /* Index identifying the link   */
                                        /* under examination.           */
 
 if(peak_band > backoff_ths[link_id])
        return FALSE;

/* first try to accept the call with the peak CAC */
   
 if(peak_band < (link_pointer->free_band - link_threshold)
    || equ_double(peak_band, link_pointer->free_band - link_threshold, 
		  DBL_CMP_THS))
    {
       update_samples_vect(link_pointer, FALSE);
       return TRUE;
    }
/* then try to use the measure of the current load */
   
 link_band = capacity / link_pointer->timescale;
 update_samples_vect(link_pointer, TRUE);
 
 if(no_load_samples == 1)
    {
     avg_link_load = load_samples_vect[link_id][0];
     var_link_load = 0.0;
    }
 else
    {
     for(i = 0; i < no_load_samples; i++)
        {
         acc_acc_samples += load_samples_vect[link_id][i];
         acc_sq_acc_samples += load_samples_vect[link_id][i] *
                               load_samples_vect[link_id][i];
        }
     avg_link_load = acc_acc_samples / no_load_samples;
     var_link_load = (acc_sq_acc_samples -
                      no_load_samples * avg_link_load * avg_link_load) /
                     (no_load_samples - 1.0);
    }
 
 if(var_link_load < 0.0)
	var_link_load = 0.0;
 
 loc_prot_fact = prot_fact;
 
 if(peak_band >
     (loc_prot_fact * link_band - avg_link_load -
            k_star * sqrt(var_link_load) - link_threshold))
     {
/* Here 'peak_band' is not greater than 'backoff_ths[link_id]'. */
	backoff_ths[link_id] = peak_band - PRECISION;
	return FALSE;
     }
 return TRUE;
}
 
 
/*
**  Function  : void init_CAC3(void)
**  Return    : none.
**  Remarks   : this function configures all the entities involved by the
**              CAC algorithm based on the Modified Peak Rate Allocation
**              over Real Time Traffic Estimations.
*/
void init_CAC3(void)
{
 int i, j;
 
 aux_freq = capacity * 1000000.0 / (meas_wind_size * 53.0 * 8.0);
 link_mean = (double *)MMmalloc(no_of_links * sizeof(double), "init_CAC3");
 link_variance = (double *)MMmalloc(no_of_links * sizeof(double), "init_CAC3");
 for(i = 0; i < no_of_links; i++)
   {
    link_mean[i] = 0.0;
    link_variance[i] = 0.0;
    backoff_ths[i] = capacity / link_vect[i]->timescale;
    next_sample_index[i] = 0;
    last_req_time[i] = 0.0;
    load_samples_vect[i] = (double *)MMmalloc(no_load_samples * sizeof(double),
                                              "init_CAC3");
    for(j = 0; j < no_load_samples; j++)
        load_samples_vect[i][j] = 0.0;
   }
}
 
 
/*
**  Function  : Boolean exe_CAC3(Link *link_pointer, double mean_band,
**                               double peak_band, double link_threshold)
**  Return    : TRUE if the call request can be accepted, FALSE otherwise.
**  Remarks   : this function drives all the actions required by the CAC
**              algorithm based on the Modified Peak Rate Allocation over Real
**              Time Traffic Estimations.
**              Its parameters are:
**                - 'link_pointer': the link under examination;
**                - 'mean_band': the declared average bandwidth of the
**                  source generator;
**                - 'peak_band': the declared peak bandwidth of the source
**                  generator;
**                - 'link_threshold': the protection factor to be used; 
**                  zero if we are checking a primary path.
*/
Boolean exe_CAC3(Link *link_pointer, double mean_band, double peak_band,
                 double link_threshold)
{
 register int i;
 
 double avg_link_load;  /* Mean value of the link load.                 */
 double var_link_load;  /* Variance of the link load.                   */
 double link_band;      /* Total bandwidth of the link.                 */
 double loc_prot_fact;  /* Local value of the protection factor applied */
                        /* to the link capacity.                        */
 
 double acc_acc_samples = 0.0;          /* Accumulator for the load     */
                                        /* samples.                     */
 double acc_sq_acc_samples = 0.0;       /* Accumulator for the squares  */
                                        /* of the load samples.         */
 int link_id = link_pointer->id;    /* Index identifying the link   */
                                        /* under examination.           */
 
 if(peak_band > backoff_ths[link_id])
        return FALSE;

/* first try to use the peak allocation algorithm */

 if(peak_band < (link_pointer->free_band - link_threshold)
    || equ_double(peak_band, link_pointer->free_band - link_threshold,
		  DBL_CMP_THS))
   {
      update_samples_vect(link_pointer, FALSE);
      return TRUE;
   }
 
/* else use the measured load */
   
 link_band = capacity / link_pointer->timescale;
 update_samples_vect(link_pointer, TRUE);
 
 if(no_load_samples == 1)
    {
     avg_link_load = load_samples_vect[link_id][0];
     var_link_load = 0.0;
    }
 else
    {
     for(i = 0; i < no_load_samples; i++)
        {
         acc_acc_samples += load_samples_vect[link_id][i];
         acc_sq_acc_samples += load_samples_vect[link_id][i] *
                               load_samples_vect[link_id][i];
        }
     avg_link_load = acc_acc_samples / no_load_samples;
     var_link_load = (acc_sq_acc_samples -
                      no_load_samples * avg_link_load * avg_link_load) /
                     (no_load_samples - 1.0);
    }
 
 if(var_link_load < 0.0)
	var_link_load = 0.0;
 
 loc_prot_fact = prot_fact;
 
 if(peak_band >
       (loc_prot_fact * link_band - avg_link_load -
              k_star * sqrt(var_link_load) - link_threshold))
    {
/* Here 'peak_band' is not greater than 'backoff_ths[link_id]'. */
       backoff_ths[link_id] = peak_band - PRECISION;
       return FALSE;
    }
 return TRUE;
}
 
 
/*
**  Function  : void init_CAC4(void)
**  Return    : none.
**  Remarks   : this function configures all the entities involved by the
**              CAC4 algorithm 
*/
void init_CAC4(void)
{
  init_Ucall();
}

 
/*
**  Function  : Boolean exe_CAC4(Link *link_pointer, double mean_band,
**                               double peak_band, double link_threshold)
**  Return    : TRUE if the call request can be accepted, FALSE otherwise.
**  Remarks   : this function drives all the actions required by the CAC
**              algorithm whose functioning is described below:
**
**              - if the incoming call belongs to either Class 1 or Class 2,
**                it is accepted when the total Peak Cell Rate (PCR) of
**                Class 1 and Class 2 is less than 'alpha' * 'capacity'.
**              - if the incoming call belongs to Class U, it is accepted
**                when the total PRC of Class U is less than
**                'beta' * ('capacity' - total PCR of DBR calls of class 1/2 -
**                total SCR of SBR call of class 1/2).
**              The parameters of the function are:
**                      - 'link_pointer': pointer to the link under
**                                        examination;
**                      - 'mean_band'   : mean bandwidth of the call;
**                      - 'peak_band'   : peak bandwidth of the call;
**                      - 'link_threshold': the protection factor to be used.
**                                          zero if we are checking a primary 
**                                          path
**   NOTE: THE link_threshold IS USED ONLY BY THE CLASS 1/2 CALL!
*/
Boolean exe_CAC4(Link *link_pointer, double mean_band, double peak_band,
                 double link_threshold)
{     
 Link_ucall_stat	*this_link = link_Ucall_stat[link_pointer->id];
 double 		link_band  = capacity / link_pointer->timescale;
 Boolean		response; 
                       
 if(mean_band != UNDEFINED_BW)
   {                    /* The call does not belong to Class U. */
     double mod_bw;

     mod_bw = eval_modified_bw(mean_band, peak_band);
     if((mod_bw + this_link->acc_PCR_1_2) 
	< (alpha * link_band - link_threshold)
	|| equ_double(mod_bw + this_link->acc_PCR_1_2, 
		      alpha * link_band - link_threshold, DBL_CMP_THS))
        response = TRUE;    /* The call is accepted.                */
     else
        response = FALSE;   /* The call is refused.                 */
   }
 else                   /* The call belongs to Class U.         */
   {
    if((peak_band + this_link->acc_PCR_U) < (beta * this_link->free_Uband)
       || equ_double(peak_band + this_link->acc_PCR_U, 
		     beta * this_link->free_Uband, DBL_CMP_THS))
        response = TRUE;    /* The call is accepted.                */
    else
        response = FALSE;   /* The call is refused.                 */
   }
 return response;
}

/*
**  Function  : void init_Ucall(void)
**  Return    : none.
**  Remarks   : this function configures all the entities involved by the
**              CAC algorithm that is applied when users of Class U are
**              present.
*/
void init_Ucall(void)
{
 int 		i;
 Link_ucall_stat 	*this;
 
 for(i = 0; i < no_of_links; i++)
   {
    link_Ucall_stat[i] = (Link_ucall_stat *)MMmalloc(sizeof(Link_ucall_stat), 
                          "init_Ucall");
    this = link_Ucall_stat[i];
    this->start_time = 0.0;
    this->last_updt_time = 0.0;
    this->free_Uband = capacity / link_vect[i]->timescale;
    this->fb_integ = 0.0;
    this->last_unus_time = 0.0;
    this->unused_free_band = capacity / link_vect[i]->timescale;
    this->un_fb_integ = 0.0;
    this->no_of_Ucall = 0;
    this->no_of_Ucall_integ = 0.0;
    this->acc_PCR_1_2 = 0.0;
    this->acc_PCR_U = 0.0; 
    this->U_PCR_int = 0.0;
    this->last_Uupdt_time = 0.0;
    this->is_node_node = TRUE;
   }
/* Variables used for general statistics are reset */
 for(i = 0; i < no_of_cgens; i++)
   {
    gen_finUload[i] = 0.0;
    gen_finUtras[i] = 0.0;
    gen_intUload[i] = 0.0;
    gen_intUtras[i] = 0.0;
    classed_prop_intUtras[i][0] = 0.0;
    classed_prop_intUtras[i][1] = 0.0;
    classed_prop_intUtras[i][2] = 0.0;
    classed_prop_intUtras[i][3] = 0.0;
    gen_Uload_start_time[i] = 0.0;
   }
 for(i=0; i<no_of_users; i++)
   {
    link_Ucall_stat[user_vect[i]->out_link->id]->is_node_node=FALSE;
    link_Ucall_stat[user_vect[i]->in_link->id]->is_node_node=FALSE;
   }
 init_Ucall_list_on_link();
}

static void restrict_Ucalls_on_link(CallList *U_calls_on, double avg_link_band)  
{
  CallList *call_ptr;    /* Pointer to a call currently routed over the  */
			 /* link.                                        */

  /* For each call of class U in the link the occurrence of bandwidth     */
  /* restriction is checked here.                                         */
  for(call_ptr = U_calls_on->next;
      call_ptr != U_calls_on; call_ptr = call_ptr->next)
    {
      double   free_band;    /* Band released by a call of Class U when it*/
                             /* is restricted.                            */

      free_band = ((CallType5 *)(call_ptr->call_elem->info))->curr_br -
	avg_link_band;

      if(free_band > 0.0)    /* The bandwidth assigned to the call   */
	{                    /* must be restricted.                  */
	  LinkList *path_ptr; /* Pointer to the path where a call is routed. */
	  CallType5 *temp = ((CallType5 *)(call_ptr->call_elem->info));
 
	  /* Updating the bandwidth allocated to the call and its
             related variables.  */
	  temp->br_integ +=
	    (ancles_time - temp->last_updt_time) * temp->curr_br;
	  temp->last_updt_time = ancles_time;
	  temp->curr_br = avg_link_band;
 
	  /* Updating the unused free bandwidth for each link over
	  which the restricted call is routed. */

	  for(path_ptr = call_ptr->call_elem->path_ptr->path;
	      path_ptr != NULL; path_ptr = path_ptr->next)
	    {
	      Link_ucall_stat *temp_link;
	      temp_link = link_Ucall_stat[path_ptr->link_elem->id];

	      temp_link->un_fb_integ += temp_link->unused_free_band *
		(ancles_time - temp_link->last_unus_time);
	      temp_link->last_unus_time = ancles_time;
	      temp_link->unused_free_band += free_band;
	    }
	}
    }
} 

/*
**  Function  : double update_Ucall_open(Link *link_pointer, double mean_band,
**                                      double peak_band)
**  Return    : the bandwidth available for calls of Class U in the link
**              pointed by 'link_pointer'.
**  Remarks   : this function drives all the actions required by the CAC
**              algorithm when a new call is reaching a link.
**              The incoming call has not yet been included in the
**              Ucall_list.
**              The function parameters are:
**                - 'link_pointer' : pointer to the link under examination;
**                - 'mean_band'    : the declared mean bandwidth of the call;
**                - 'peak_band'    : the declared peak bandwidth of the call;
*/
double update_Ucall_open(Link *link_pointer, double mean_band, 
			 double peak_band)
{
 double   avg_link_band = 0.0;
                        /* Mean value of the portion of link bandwidth  */
                        /* that is available for calls of type U.       */
 Link_ucall_stat *this_link = link_Ucall_stat[link_pointer->id];
                        /* Pointer to the record collecting the link    */
                        /* information needed by calls of Class U.      */

 if(mean_band != UNDEFINED_BW)  /* The call does not belong to Class U. */
  {
   this_link->acc_PCR_1_2 += eval_modified_bw(mean_band, peak_band);
            
/* Updating all the variables concerning the free bandwidth.    */
   this_link->fb_integ += (this_link->free_Uband *
                           (ancles_time - this_link->last_updt_time));
   this_link->last_updt_time = ancles_time;
   this_link->free_Uband -= mean_band;
 
/* Updating the variables concerning the unused bandwidth.      */
   this_link->un_fb_integ += (this_link->unused_free_band *
                              (ancles_time - this_link->last_unus_time));
   this_link->last_unus_time = ancles_time;
   this_link->unused_free_band -= mean_band;
 
/* The link bandwidth to be assigned to each call of Class U is         */
/* evaluated here, according to the state of the link.                  */
   if(this_link->no_of_Ucall > 0)
     {
      avg_link_band = this_link->free_Uband / this_link->no_of_Ucall;
      
      restrict_Ucalls_on_link(this_link->U_calls_on, avg_link_band);
     }
  }
 else   /* The incoming call belongs to class U.        */
  {
   this_link->U_PCR_int += this_link->acc_PCR_U * 
			   (ancles_time - this_link->last_Uupdt_time);
   /* The current call is not included. */
   this_link->no_of_Ucall_integ += this_link->no_of_Ucall * 
			   (ancles_time - this_link->last_Uupdt_time);
   this_link->last_Uupdt_time = ancles_time;
   this_link->acc_PCR_U += peak_band;
 
/* Updating the link bandwidth to be assigned to calls of Class U.      */
   avg_link_band = this_link->free_Uband / (this_link->no_of_Ucall + 1);
 
   restrict_Ucalls_on_link(this_link->U_calls_on, avg_link_band);
  }
 
 return(avg_link_band);
}

static void update_Ucalls_on_link(CallList* U_calls_on, double prev_min_rate_on_link)  
{
  CallList  *this_call;  /* Pointer to the call in the list of Class U	*/
			/* calls that is associated to the link.	*/
  double av_br;

  /* Updating the variables regarding the statistics about the calls of */
  /* Class U.								*/
  for(this_call = U_calls_on->next;
      this_call != U_calls_on; this_call = this_call->next)
    {
      CallType5 *temp_call = ((CallType5 *)(this_call->call_elem->info));

      if(temp_call->curr_br > prev_min_rate_on_link
	 || equ_double(temp_call->curr_br, prev_min_rate_on_link, DBL_CMP_THS))
	{
	  LinkList  *path_ptr;   /* Pointer to the path over which the
	                            call to be closed was routed.  */
	  double    min_rate;    /* Minimum bit rate in the link,
	                            after the call has been closed.  */
	  double    free_band;   /* Difference between the old and the
	                            new bit rate.  */
   
	  /* Updating the bandwidth currently used by the call in each
	  link composing its path.  */
	  min_rate = this_call->ref_bw;	/* Bandwidth currently	*/
	                                /* used by the call.	*/
 
	  for(path_ptr = this_call->call_elem->path_ptr->path;
	      path_ptr != NULL; path_ptr = path_ptr->next)
	    {
	      Link_ucall_stat *temp_link;

	      temp_link = link_Ucall_stat[path_ptr->link_elem->id];
	      av_br = temp_link->free_Uband / temp_link->no_of_Ucall;
	      if(min_rate > av_br)
		min_rate = av_br;
	    }

	  /* Updating the statistics about the call. */
	  temp_call->br_integ +=
	    (ancles_time - temp_call->last_updt_time) * temp_call->curr_br;
	  temp_call->last_updt_time = ancles_time;
	  free_band = temp_call->curr_br - min_rate;
	  temp_call->curr_br = min_rate;
 
	  /* Updating the variables regarding the Unused Free Bandwidth. */
 
	  for(path_ptr = this_call->call_elem->path_ptr->path;
	      path_ptr != NULL; path_ptr = path_ptr->next)
	    {
	      Link_ucall_stat *temp_link;

	      temp_link = link_Ucall_stat[path_ptr->link_elem->id];
	      temp_link->un_fb_integ += temp_link->unused_free_band *
		(ancles_time - temp_link->last_unus_time);
	      temp_link->last_unus_time = ancles_time;
	      temp_link->unused_free_band += free_band;
	    }
	}
    }
} 

/*
**  Function  : void update_Ucall_close(Call *call_ptr, Link *link_ptr)
**  Return    : None
**  Remarks   : this function drives all the actions required by the CAC 
**              algorithm when a U call is closed.
**              Every time a call closure occurs, the bandwidth assigned to
**		each call of Class U must be reconsidered.
*/
void update_Ucall_close(Call *call_ptr, Link *link_ptr)
{
 Link_ucall_stat *this_stat = link_Ucall_stat[link_ptr->id];
                        /* Record to be processed.			*/
 double    loc_mean;    /* Mean bandwidth currently used by the call	*/
			/* being closed.				*/
 double    loc_peak;    /* Peak bandwidth of the call being closed.	*/

 CallType1 *ci1;
 CallType2 *ci2;
 CallType3 *ci3;
 CallType4 *ci4;
 CallType5 *ci5;

 switch (call_ptr->call_type)
   {
    case CALL_TYPE1 :
	ci1 = (CallType1 *)(call_ptr->info);
	loc_mean = ci1->cbr_value;
        loc_peak = ci1->cbr_value;
        break;
 
    case CALL_TYPE2 :
	ci2 = (CallType2 *)(call_ptr->info);
        loc_mean = ci2->peak_rate * ci2->mean_on_time /
                   ( ci2->mean_on_time + ci2->mean_off_time );
        loc_peak = ci2->peak_rate;
        break;
 
    case CALL_TYPE3 :
	ci3 = (CallType3 *)(call_ptr->info);
        loc_mean = ( ci3->min_vbr_value + ci3->max_vbr_value ) / 2.0;
        loc_peak = ci3->max_vbr_value;
        break;
 
    case CALL_TYPE4 :
	ci4 = (CallType4 *)(call_ptr->info);
        loc_mean = ci4->higher_bit_rate_prob * ci4->higher_bit_rate_value +
                   ci4->medium_bit_rate_prob * ci4->medium_bit_rate_value +
                   ci4->lower_bit_rate_prob * ci4->lower_bit_rate_value;
        loc_peak = ci4->higher_bit_rate_value;
        break;
 
    case CALL_TYPE5 :
	ci5 = (CallType5 *)(call_ptr->info);
        loc_mean = UNDEFINED_BW;
        loc_peak = ci5->mbr_value;
        break;

    default:
        printf("ERROR: unknown call type %d in function update_Ucall_close\n",
	       call_ptr->call_type);
	exit(1);
   }
 
 if(loc_mean != UNDEFINED_BW)
   {                   	/* The call does not belong to Class U.	*/
    double prev_free_Uband;

    this_stat->acc_PCR_1_2 -= eval_modified_bw(loc_mean, loc_peak);
      
/* Updating the variables concerning the free bandwidth.	*/
    prev_free_Uband = this_stat->free_Uband;
    this_stat->fb_integ += this_stat->free_Uband *
                           (ancles_time - this_stat->last_updt_time);
    this_stat->last_updt_time = ancles_time;
    this_stat->free_Uband += loc_mean;
 
/* Updating the variables concerning the unused free bandwidth.	*/
    this_stat->un_fb_integ += this_stat->unused_free_band *
                              (ancles_time - this_stat->last_unus_time);
    this_stat->last_unus_time = ancles_time;
    this_stat->unused_free_band += loc_mean;
 
    if(this_stat->no_of_Ucall != 0)
      update_Ucalls_on_link(this_stat->U_calls_on, 
		 prev_free_Uband / this_stat->no_of_Ucall);
   }
 else
   {		/* The call belongs to Class U.	*/
    double prev_min_rate_on_link;
                        /* Minimum bit rate in the link, before closing	*/
			/* the call.					*/
    this_stat->U_PCR_int += this_stat->acc_PCR_U * 
			    (ancles_time - this_stat->last_Uupdt_time);
    /* The current call is included. */
    this_stat->no_of_Ucall_integ += this_stat->no_of_Ucall * 
			   (ancles_time - this_stat->last_Uupdt_time);
    this_stat->last_Uupdt_time = ancles_time;
    this_stat->acc_PCR_U -= loc_peak;
 
/* Updating the variables regarding the Unused Free Bandwidth.	*/
    this_stat->un_fb_integ += this_stat->unused_free_band *
                              (ancles_time - this_stat->last_unus_time);
    this_stat->last_unus_time = ancles_time;
    this_stat->unused_free_band += ci5->curr_br;
 
    prev_min_rate_on_link = this_stat->free_Uband / this_stat->no_of_Ucall;
    this_stat->no_of_Ucall--;
    
    update_Ucalls_on_link(this_stat->U_calls_on, prev_min_rate_on_link);
   }
}
