/*
**  Module    : freelist.c
**  Version   : $Revision: 3.1 $
**  Date      : $Date: 1998/07/28 09:36:09 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico.
**
**  Synopsis  : this module contains the functions dealing with the free
**              lists used by ANCLES. 
**              Free lists are used instead of the standard C functions dealing 
**              with dynamic allocation and releasing of memory in order to 
**              make these actions faster and safer.
**              The free lists are organized as LIFO queues (stack) with a 
**              variable pointing to their top record.
**              Each time the stack is found empty a fixed number of 
**              consecutive memory locations are requested to the system.
**
**  Functions : LinkList  *linklist_alloc(void)
**              void      linklist_release(LinkList *)
**              void      whole_linklist_release(void)
**              Path      *path_alloc(void)
**              void      path_release(Path *)                               
**              Stack     *stack_alloc(void)
**              void      stack_release(Stack *)
**              void      whole_stack_release(void)
**              CallList  call_list_alloc(void)
**              void      call_list_release(CallList *)
**              Call      *call_alloc(void)
**              void      call_release(Call *)
**              VCset     *vcset_alloc(void)
**              void      vcset_release(VCset *)
**              void      whole_vcset_release(void)
**              Event     *event_alloc(void)
**              void      event_release(Event *rel_event)
**              CallType1 *call_type1_alloc(void)
**              void      call_type1_release(CallType1 *)
**              CallType2 *call_type2_alloc(void)
**              void      call_type2_release(CallType2 *)
**              CallType3 *call_type3_alloc(void)
**              void      call_type3_release(CallType3 *)
**              CallType4 *call_type4_alloc(void)
**              void      call_type4_release(CallType4 *)
**              CallType5 *call_type5_alloc(void)
**              void      call_type5_release(CallType5 *)
**              CallType6 *call_type6_alloc(void)
**              void      call_type6_release(CallType6 *)
**              CallType6 *call_type8_alloc(void)
**              void      call_type8_release(CallType6 *)
**				ProtCh	  *prot_ch_alloc(void)
**				void	  prot_ch_release(ProtCh *)
**				Intlist	  *int_list_alloc(void)
**				void	  int_list_release(Intlist *)
**				Dbllist	  *dbl_list_alloc(void)
**				void	  dbl_list_release(Dntlist *)
**
**  History   : 19-SEP-1995 - Reorganization of the module.     -AF-
**              01-JUN-1996 - Revision.                         -AF-
**              01-MAR-1997 - Revision				-MM-
*/
 
#include "ancles.h"
 
/* define this if you have memory trouble... */

  #define MEMORY_DEBUG

/***********************
**  Global Variables  **
***********************/
 
static Stack     *top_stack_flist = NULL; /* Pointer to the top of         */
                                          /* the 'Stack' free list.        */
static LinkList  *top_link_flist  = NULL; /* Pointer to the top of         */
                                          /* the 'LinkList' free list.     */
static Path      *top_path_flist  = NULL; /* Pointer to the top of         */
                                          /* the 'Path' free list.         */
static VCset     *top_vcset_flist = NULL; /* Pointer to the top of         */
                                          /* the 'VCset' free list.        */
static Event     *top_event_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'Event' free list.     */
static CallList  *top_calll_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallList' free list.  */
static Call      *top_Call_flist  = NULL; /* Pointer to the top element    */
                                          /* on the 'Call' free list.      */
static CallType1 *top_call1_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType1' free list. */
static CallType2 *top_call2_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType2' free list. */
static CallType3 *top_call3_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType3' free list. */
static CallType4 *top_call4_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType4' free list. */
static CallType5 *top_call5_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType5' free list. */
static CallType6 *top_call6_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType6' free list. */
static CallType8 *top_call8_flist = NULL; /* Pointer to the top element    */
                                          /* on the 'CallType8' free list. */
static ProtCh	 *top_protch_flist = NULL;/* Pointer to the top element    */
                                          /* on the 'ProtCh' free list.	   */
static Intlist	 *top_intlist_flist = NULL;/* Pointer to the top element   */
                                          /* on the 'Intlist' free list.   */
static Dbllist	 *top_dbllist_flist = NULL;/* Pointer to the top element   */
                                          /* on the 'Dbllist' free list.   */
# ifdef MEMORY_DEBUG
  static long n_stack=0;
  static long n_linklist=0;
  static long n_path=0;
  static long n_vcset=0;
  static long n_event=0;
  static long n_calllist=0;
  static long n_call=0;
  static long n_calltype1=0;
  static long n_calltype2=0;
  static long n_calltype3=0;
  static long n_calltype4=0;
  static long n_calltype5=0;
  static long n_calltype6=0;
  static long n_calltype8=0;
  static long n_protch=0;
  static long n_intlist=0;
  static long n_dbllist=0;
# endif
/***************************
**  Function Definitions  **
***************************/
 
/*
**  Function  : void *MMmalloc(size_t size, const char * function_name)
**  Return    : the pointer to the memory of the new allocated block
**  Remarks   : MMmalloc() allocates size bytes and returns a pointer to the
**              allocated memory.  The memory is not cleared.
**              If an error occours, an error is printed, including the
**              function name of the calling function.
*/
void *MMmalloc(size_t size, const char * f_name)
{
   void *temp_pointer;
   if((temp_pointer = malloc(size)) == NULL)
     {
/* If problems arise from the memory allocation, an error message is    */
/* printed before exiting the program execution.                        */
      printf("\nError:  Memory allocation error in ANCLES function ");
      printf("%s\n", f_name);
# ifdef MEMORY_DEBUG
  printf("Stato della memoria: \n");
  printf("Numero elementi presenti di tipo stack:\t%ld\n",n_stack);
  printf("Numero elementi presenti di tipo linklist:\t%ld\n",n_linklist);
  printf("Numero elementi presenti di tipo path:\t%ld\n",n_path);
  printf("Numero elementi presenti di tipo vcset:\t%ld\n",n_vcset);
  printf("Numero elementi presenti di tipo event:\t%ld\n",n_event);
  printf("Numero elementi presenti di tipo calllist:\t%ld\n",n_calllist);
  printf("Numero elementi presenti di tipo call:\t%ld\n",n_call);
  printf("Numero elementi presenti di tipo calltype1:\t%ld\n",n_calltype1);
  printf("Numero elementi presenti di tipo calltype2:\t%ld\n",n_calltype2);
  printf("Numero elementi presenti di tipo calltype3:\t%ld\n",n_calltype3);
  printf("Numero elementi presenti di tipo calltype4:\t%ld\n",n_calltype4);
  printf("Numero elementi presenti di tipo calltype5:\t%ld\n",n_calltype5);
  printf("Numero elementi presenti di tipo calltype6:\t%ld\n",n_calltype6);
  printf("Numero elementi presenti di tipo calltype8:\t%ld\n",n_calltype8);
  printf("Numero elementi presenti di tipo protch:\t%ld\n",n_protch);
  printf("Numero elementi presenti di tipo intlist:\t%ld\n",n_intlist);
# endif
      exit(1);
     }

   return temp_pointer;
}

/*
**  Function  : LinkList *linklist_alloc(void)
**  Return    : the pointer to the LinkList record just summoned.
**  Remarks   : the pointer to the record on the top of the 'LinkList' free
**              list is returned if different than NULL (the value of
**              'top_link_flist' is changed in this case). If the 'LinkList'
**              free list is empty, a new memory area is dinamically
**              allocated.
*/
LinkList *linklist_alloc(void)
{
 LinkList *new_linklist;
 
 if(top_link_flist == NULL)     /* The LinkList stack is empty.         */
    {
# ifdef MEMORY_DEBUG
  n_linklist++;
# endif
     return ((LinkList *)MMmalloc(sizeof(LinkList), "linklist_alloc"));
    }
 else                           /* The LinkList stack is not empty.     */
    {
     new_linklist = top_link_flist;
     top_link_flist = new_linklist->next;
     new_linklist->next = NULL;
no_of_free_linklist--;
     return new_linklist;
    }
}
 
 
/*
**  Function  : void linklist_release(LinkList *rel_linklist)
**  Return    : none.
**  Remarks   : the 'LinkList' record pointed by 'rel_linklist' is
**              placed on the top of the LinkList free list (the value of
**              'top_link_flist' is changed).
*/
void linklist_release(LinkList *rel_linklist)
{
 rel_linklist->next = top_link_flist;
 top_link_flist = rel_linklist;
 no_of_free_linklist++;
}
 
 
/*
**  Function  : void whole_linklist_release(void)
**  Return    : none.
**  Remarks   : this function is called when all the available paths
**              between each pair of nodes have been found; the memory areas
**              previously assigned to the 'LinkList' records are returned to
**              the system.
*/
void whole_linklist_release(void)
{
 LinkList *ptr;
 
 while(top_link_flist != NULL)
        {
# ifdef MEMORY_DEBUG
  n_linklist--;
# endif
         ptr = top_link_flist;
         top_link_flist = top_link_flist->next;
no_of_free_linklist--;
         free(ptr);
        }
}
 
 
/*
**  Function  : Path *path_alloc(void)
**  Return    : the pointer to the 'Path' record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Path' free
**              list is returned if different than NULL (the value of
**              'top_path_flist' is changed in this case). If the 'Path'
**              free list is empty, a new memory area is dinamically
**              allocated. The 'next' field is used in order to link the
**              free list elements.
*/
Path *path_alloc(void)
{
 Path *new_path;
 
 if(top_path_flist == NULL)     /* The Path free list is empty.         */
    {
# ifdef MEMORY_DEBUG
	 n_path++;
# endif

	 new_path = (Path *)MMmalloc(sizeof(Path), "path alloc");
	 new_path->protected_links = 
		 (int *)MMmalloc((MAX_LINK - no_of_links + no_of_nnlinks) * 
		 sizeof(int), "path alloc");
	 new_path->protected_nodes = 
		 (int *)MMmalloc(no_of_nodes * sizeof(int),"path alloc");
    }
 else                           /* The Path free list is not empty.     */
    {
     new_path = top_path_flist;
     top_path_flist = new_path->next;
	}
 new_path->next = NULL;
 return new_path;
}
 
 
/*
**  Function  : void path_release(Path *rel_path)
**  Return    : none.
**  Remarks   : the 'Path' record pointed by 'rel_path' is
**              placed on the top of the 'Path' free list (the value
**              of 'top_path_flist' is changed).
*/
void path_release(Path *rel_path)
{
 rel_path->next = top_path_flist;
 top_path_flist = rel_path;
}
 
/*
**  Function  : void whole_pathlist_release(void)
**  Return    : none.
**  Remarks   : this function is called when all the available paths
**              between each pair of nodes have been found; the memory areas
**              previously assigned to the 'LinkList' records are returned to
**              the system.
*/
void whole_pathlist_release(void)
{
 Path *ptr;
 
 while(top_path_flist != NULL)
        {
# ifdef MEMORY_DEBUG
  n_path--;
# endif
         ptr = top_path_flist;
         top_path_flist = top_path_flist->next;

		 free(ptr->protected_links);
		 free(ptr->protected_nodes);
         free(ptr);
        }
}
 
 
/*
**  Function  : Stack *stack_alloc(void)
**  Return    : the pointer to the 'Stack' record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Stack' free
**              list is returned if different than NULL (the value of
**              'top_stack_flist' is changed in this case). If the 'Stack'
**              free list is empty, a new memory area is dinamically
**              allocated.
*/
Stack *stack_alloc(void)
{
 Stack *new_stack;
 
 if(top_stack_flist == NULL)    /* The Stack free list is empty.        */
    {
# ifdef MEMORY_DEBUG
  n_stack++;
# endif
     return ((Stack *)MMmalloc(sizeof(Stack), "stack_alloc"));
    }
 else                           /* The Stack free list is not empty.    */
    {
     new_stack = top_stack_flist;
     top_stack_flist = new_stack->next;
     new_stack->next = NULL;
     return new_stack;
    }
}
 
 
/*
**  Function  : void stack_release(Stack *rel_stack)
**  Return    : none.
**  Remarks   : the 'Stack' record pointed by 'rel_stack' is
**              placed on the top of the 'Stack' free list (the value of
**              'top_stack_flist' is changed).
*/
void stack_release(Stack *rel_stack)
{
 rel_stack->next = top_stack_flist;
 top_stack_flist = rel_stack;
}
 
 
/*
**  Function  : void whole_stack_release(void)
**  Return    : none.
**  Remarks   : this function is called when all the available paths
**              between each pair of nodes have been found; the memory areas
**              previously assigned to the 'Stack' records are returned to
**              the system.
*/
void whole_stack_release(void)
{
 Stack *ptr;
 
 while(top_stack_flist != NULL)
        {
# ifdef MEMORY_DEBUG
  n_stack--;
# endif
         ptr = top_stack_flist;
         top_stack_flist = top_stack_flist->next;
         free(ptr);
        }
}
 

/*
**  Function  : VCset *vcset_alloc(void)
**  Return    : the pointer to the 'VCset' record just summoned.
**  Remarks   : the pointer to the record on the top of the 'VCset' free
**              list is returned if different than NULL (the value of
**              'top_vcset_flist' is changed in this case). If the 'VCset'
**              free list is empty, a new memory area is dinamically
**              allocated.
*/
VCset *vcset_alloc(void)
{
 VCset *new_vcset;
 
 if(top_vcset_flist == NULL)    /* The VCset free list is empty.        */
    {
# ifdef MEMORY_DEBUG
  n_vcset++;
# endif
     return ((VCset *)MMmalloc(sizeof(VCset), "vcset_alloc"));
    }
 else                           /* The VCset free list is not empty.    */
    {
     new_vcset = top_vcset_flist;
     top_vcset_flist = new_vcset->next;
     new_vcset->next = NULL;
     return new_vcset;
    }
}
 
 
/*
**  Function  : void vcset_release(VCset *rel_vcset)
**  Return    : none.
**  Remarks   : the 'VCset' record pointed by 'rel_vcset' is
**              placed on the top of the 'VCset' free list (the value of
**              'top_vcset_flist' is changed).
*/
void vcset_release(VCset *rel_vcset)
{
 rel_vcset->next = top_vcset_flist;
 top_vcset_flist = rel_vcset;
}
 
 
/*
**  Function  : void whole_vcset_release(void)
**  Return    : none.
**  Remarks   : this function is called when the translation of ANCLES 
**              connections in CLASS users has been completed; the memory 
**              areas previously assigned to the 'VCset' records are returned 
**              to the system.
*/
void whole_vcset_release(void)
{
 VCset *ptr;
 
 while(top_vcset_flist != NULL)
        {
# ifdef MEMORY_DEBUG
  n_vcset--;
# endif
         ptr = top_vcset_flist;
         top_vcset_flist = top_vcset_flist->next;
         free(ptr);
        }
}

 
/*
**  Function  : Event *event_alloc(void)
**  Return    : the pointer to the 'Event' type record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Event' free
**              list is returned if different than NULL (the value of
**              'top_event_flist' is changed in this case). If the 'Event'
**              free list is empty, a new memory area is dinamically allocated.
*/
Event *event_alloc(void)
{
 Event *new_event;
 
 if(top_event_flist == NULL)    /* The Event free list is empty.        */
    {
# ifdef MEMORY_DEBUG
  n_event++;
# endif
     return ((Event *)MMmalloc(sizeof(Event), "event_alloc"));
    }
 else                           /* The Event free list is not empty.    */
    {
     new_event = top_event_flist;
     top_event_flist = new_event->next;
     new_event->next = NULL;
     return new_event;
    }
}
 
 
/*
**  Function  : void event_release(Event *rel_event)
**  Return    : none.
**  Remarks   : the Event record pointed by 'rel_event' is
**              placed on the top of the Event stack (the value of
**              'top_event_flist' is changed).
*/
void event_release(Event *rel_event)
{
 rel_event->next = top_event_flist;
 rel_event->pred = NULL;
 top_event_flist = rel_event;
}
 
 
/*
**  Function  : CallList *call_list_alloc(void)
**  Return    : the pointer to the CallList record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallList' free
**              list is returned if different than NULL (the value of
**              'top_calll_flist' is changed in this case). If the
**              'CallList' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallList *call_list_alloc(void)
{
 CallList *new_calllist;
 
 if(top_calll_flist == NULL)    /* The CallList free list is empty.     */
    {
# ifdef MEMORY_DEBUG
  n_calllist++;
# endif
     return ((CallList *)MMmalloc(sizeof(CallList), "call_list_alloc"));
    }
 else                           /* The CallList free list is not empty. */
    {
     new_calllist = top_calll_flist;
     top_calll_flist = new_calllist->next;
     new_calllist->next = NULL;
     return new_calllist;
    }
}
 
 
/*
**  Function  : void call_list_release(CallList *rel_call_list)
**  Return    : none.
**  Remarks   : the CallList record pointed by 'rel_call_list' is
**              placed on the top of the CallList stack (the value of
**              'top_calll_flist' is changed).
*/
void call_list_release(CallList *rel_call_list)
{
 rel_call_list->next = top_calll_flist;
 rel_call_list->pred = NULL;
 top_calll_flist = rel_call_list;
}
 
 
/*
**  Function  : Call *call_alloc(void)
**  Return    : the pointer to the Call record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Call' free
**              list is returned if different than NULL (the value of
**              'top_Call_flist' is changed in this case). If the
**              'Call' free list is empty, a new memory area is dinamically
**              allocated.
*/
Call *call_alloc(void)
{
 Call *new_call;
 
 if(top_Call_flist == NULL)     /* The Call free list is empty.         */
    {
# ifdef MEMORY_DEBUG
  n_call++;
# endif
     return ((Call *)MMmalloc(sizeof(Call), "call_alloc"));
    }
 else                           /* The Call free list is not empty.     */
    {
     new_call = top_Call_flist;
     top_Call_flist = new_call->next;
     new_call->next = NULL;
     return new_call;
    }
}
 
 
/*
**  Function  : void call_release(Call *rel_call)
**  Return    : none.
**  Remarks   : the Call record pointed by 'rel_call' is
**              placed on the top of the Call stack (the value of
**              'top_Call_flist' is changed).
*/
void call_release(Call *rel_call)
{
 rel_call->next = top_Call_flist;
 rel_call->pred = NULL;
 top_Call_flist = rel_call;
}
 
 
/*
**  Function  : CallType1 *call_type1_alloc(void)
**  Return    : the pointer to the CallType1 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType1' free
**              list is returned if different than NULL (the value of
**              'top_call1_flist' is changed in this case). If the
**              'CallType1' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType1 *call_type1_alloc(void)
{
 CallType1 *new_call_type1;
 
 if(top_call1_flist == NULL)    /* The CallType1 free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_calltype1++;
# endif
     return ((CallType1 *)MMmalloc(sizeof(CallType1), "call_type1_alloc"));
    }
 else                           /* The CallType1 free list is not empty.*/
    {
     new_call_type1 = top_call1_flist;
     top_call1_flist = new_call_type1->next;
     new_call_type1->next = NULL;
     return new_call_type1;
    }
}
 
 
/*
**  Function  : void call_type1_release(CallType1 *rel_call_type1)
**  Return    : none.
**  Remarks   : the CallType1 record pointed by 'rel_call_type1' is
**              placed on the top of the CallType1 stack (the value of
**              'top_call1_flist' is changed).
*/
void call_type1_release(CallType1 *rel_call_type1)
{
 rel_call_type1->next = top_call1_flist;
 top_call1_flist = rel_call_type1;
}
 
 
/*
**  Function  : CallType2 *call_type2_alloc(void)
**  Return    : the pointer to the CallType2 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType2' free
**              list is returned if different than NULL (the value of
**              'top_call2_flist' is changed in this case). If the
**              'CallType2' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType2 *call_type2_alloc(void)
{
 CallType2 *new_call_type2;
 
 if(top_call2_flist == NULL)     /* The CallType2 free list is empty.   */
    {
# ifdef MEMORY_DEBUG
  n_calltype2++;
# endif
     return ((CallType2 *)MMmalloc(sizeof(CallType2), "call_type2_alloc"));
    }
 else                           /* The CallType2 free list is not empty.*/
    {
     new_call_type2 = top_call2_flist;
     top_call2_flist = new_call_type2->next;
     new_call_type2->next = NULL;
     return new_call_type2;
    }
}
 
 
/*
**  Function  : void call_type2_release(CallType2 *rel_call_type2)
**  Return    : none.
**  Remarks   : the CallType2 record pointed by 'rel_call_type2' is
**              placed on the top of the CallType2 stack (the value of
**              'top_call2_flist' is changed).
*/
void call_type2_release(CallType2 *rel_call_type2)
{
 rel_call_type2->next = top_call2_flist;
 top_call2_flist = rel_call_type2;
}
 
 
/*
**  Function  : CallType3 *call_type3_alloc(void)
**  Return    : the pointer to the CallType3 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType3' free
**              list is returned if different than NULL (the value of
**              'top_call3_flist' is changed in this case). If the
**              'CallType3' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType3 *call_type3_alloc(void)
{
 CallType3 *new_call_type3;
 
 if(top_call3_flist == NULL)     /* The CallType3 free list is empty.   */
    {
# ifdef MEMORY_DEBUG
  n_calltype3++;
# endif
     return ((CallType3 *)MMmalloc(sizeof(CallType3), "call_type3_alloc"));
    }
 else                           /* The CallType3 free list is not empty.*/
    {
     new_call_type3 = top_call3_flist;
     top_call3_flist = new_call_type3->next;
     new_call_type3->next = NULL;
     return new_call_type3;
    }
}
 
 
/*
**  Function  : void call_type3_release(CallType3 *rel_call_type3)
**  Return    : none.
**  Remarks   : the CallType3 record pointed by 'rel_call_type3' is
**              placed on the top of the CallType3 stack (the value of
**              'top_call3_flist' is changed).
*/
void call_type3_release(CallType3 *rel_call_type3)
{
 rel_call_type3->next = top_call3_flist;
 top_call3_flist = rel_call_type3;
}
 
 
/*
**  Function  : CallType4 *call_type4_alloc(void)
**  Return    : the pointer to the CallType4 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType4' free
**              list is returned if different than NULL (the value of
**              'top_call4_flist' is changed in this case). If the
**              'CallType4' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType4 *call_type4_alloc(void)
{
 CallType4 *new_call_type4;
 
 if(top_call4_flist == NULL)     /* The CallType4 free list is empty.   */
    {
# ifdef MEMORY_DEBUG
  n_calltype4++;
# endif
     return ((CallType4 *)MMmalloc(sizeof(CallType4), "call_type4_alloc"));
    }
 else                           /* The CallType4 free list is not empty.*/
    {
     new_call_type4 = top_call4_flist;
     top_call4_flist = new_call_type4->next;
     new_call_type4->next = NULL;
     return new_call_type4;
    }
}
 
 
/*
**  Function  : void call_type4_release(CallType4 *rel_call_type4)
**  Return    : none.
**  Remarks   : the CallType4 record pointed by 'rel_call_type4' is
**              placed on the top of the CallType4 stack (the value of
**              'top_call4_flist' is changed).
*/
void call_type4_release(CallType4 *rel_call_type4)
{
 rel_call_type4->next = top_call4_flist;
 top_call4_flist = rel_call_type4;
}


/*
**  Function  : CallType5 *call_type5_alloc(void)
**  Return    : the pointer to the CallType5 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType5' free
**              list is returned if different than NULL (the value of
**              'top_call5_flist' is changed in this case). If the
**              'CallType5' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType5 *call_type5_alloc(void)
{
 CallType5 *new_call_type5;
 
 if(top_call5_flist == NULL)     /* The CallType5 free list is empty.   */
    {
# ifdef MEMORY_DEBUG
  n_calltype5++;
# endif
     return ((CallType5 *)MMmalloc(sizeof(CallType5), "call_type5_alloc"));
    }
 else                           /* The CallType5 free list is not empty.*/
    {
     new_call_type5 = top_call5_flist;
     top_call5_flist = new_call_type5->next;
     new_call_type5->next = NULL;
     return new_call_type5;
    }
}
 
 
/*
**  Function  : void call_type5_release(CallType5 *rel_call_type5)
**  Return    : none.
**  Remarks   : the CallType5 record pointed by 'rel_call_type5' is
**              placed on the top of the CallType5 stack (the value of
**              'top_call5_flist' is changed).
*/
void call_type5_release(CallType5 *rel_call_type5)
{
 rel_call_type5->next = top_call5_flist;
 top_call5_flist = rel_call_type5;
}

/*
**  Function  : CallType6 *call_type6_alloc(void)
**  Return    : the pointer to the CallType6 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType6' free
**              list is returned if different than NULL (the value of
**              'top_call6_flist' is changed in this case). If the
**              'CallType6' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType6 *call_type6_alloc(void)
{
 CallType6 *new_call_type6;
 
 if(top_call6_flist == NULL)    /* The CallType6 free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_calltype6++;
# endif
     return ((CallType6 *)MMmalloc(sizeof(CallType6), "call_type6_alloc"));
    }
 else                           /* The CallType6 free list is not empty.*/
    {
     new_call_type6 = top_call6_flist;
     top_call6_flist = new_call_type6->next;
     new_call_type6->next = NULL;
     return new_call_type6;
    }
}
 
 
/*
**  Function  : void call_type6_release(CallType6 *rel_call_type6)
**  Return    : none.
**  Remarks   : the CallType6 record pointed by 'rel_call_type6' is
**              placed on the top of the CallType1 stack (the value of
**              'top_call6_flist' is changed).
*/
void call_type6_release(CallType6 *rel_call_type6)
{
 rel_call_type6->next = top_call6_flist;
 top_call6_flist = rel_call_type6;
}

/*
**  Function  : CallType8 *call_type8_alloc(void)
**  Return    : the pointer to the CallType8 record just summoned.
**  Remarks   : the pointer to the record on the top of the 'CallType8' free
**              list is returned if different than NULL (the value of
**              'top_call8_flist' is changed in this case). If the
**              'CallType8' free list is empty, a new memory area is
**              dinamically allocated.
*/
CallType8 *call_type8_alloc(void)
{
 CallType8 *new_call_type8;
 
 if(top_call8_flist == NULL)    /* The CallType6 free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_calltype8++;
# endif
     return ((CallType8 *)MMmalloc(sizeof(CallType8), "call_type8_alloc"));
    }
 else                           /* The CallType6 free list is not empty.*/
    {
     new_call_type8 = top_call8_flist;
     top_call8_flist = new_call_type8->next;
     new_call_type8->fwd_link = NULL;
     new_call_type8->bwd_link = NULL;
     new_call_type8->next = NULL;
     return new_call_type8;
    }
}
 
 
/*
**  Function  : void call_type8_release(CallType8 *rel_call_type8)
**  Return    : none.
**  Remarks   : the CallType8 record pointed by 'rel_call_type8' is
**              placed on the top of the CallType8 stack (the value of
**              'top_call8_flist' is changed).
*/
void call_type8_release(CallType8 *rel_call_type8)
{
 rel_call_type8->next = top_call8_flist;
 top_call8_flist = rel_call_type8;
}


/*
**  Function  : ProtCh *prot_ch_alloc(void)
**  Return    : the pointer to the ProtCh record just summoned.
**  Remarks   : the pointer to the record on the top of the 'ProtCh' free
**              list is returned if different than NULL (the value of
**              'top_protch_flist' is changed in this case). If the
**              'ProtCh' free list is empty, a new memory area is
**              dinamically allocated.
*/
ProtCh *prot_ch_alloc(void)
{
 ProtCh *new_protch;
 
 if(top_protch_flist == NULL)    /* The ProtCh free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_protch++;
# endif
     return ((ProtCh *)MMmalloc(sizeof(ProtCh), "prot_ch_alloc"));
    }
 else                           /* The ProtCh free list is not empty.*/
    {
     new_protch = top_protch_flist;
     top_protch_flist = new_protch->next;
     new_protch->next = NULL;
     return new_protch;
    }
}
 
 
/*
**  Function  : void prot_ch_release(ProtCh *rel_protch)
**  Return    : none.
**  Remarks   : the ProtCh record pointed by 'rel_protch' is
**              placed on the top of the ProtCh stack (the value of
**              'top_protch_flist' is changed).
*/
void prot_ch_release(ProtCh *rel_protch)
{
 rel_protch->next = top_protch_flist;
 top_protch_flist = rel_protch;
}

/*
**  Function  : Intlist *int_list_alloc(void)
**  Return    : the pointer to the Intlist record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Intlist' free
**              list is returned if different than NULL (the value of
**              'top_intlist_flist' is changed in this case). If the
**              'Intlist' free list is empty, a new memory area is
**              dinamically allocated.
*/
Intlist *int_list_alloc(void)
{
 Intlist *new_intlist;
 
 if(top_intlist_flist == NULL)    /* The Intlist free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_intlist++;
# endif
     return ((Intlist *)MMmalloc(sizeof(Intlist), "int_list_alloc"));
    }
 else                           /* The Intlist free list is not empty.*/
    {
     new_intlist = top_intlist_flist;
     top_intlist_flist = new_intlist->next;
     new_intlist->next = NULL;
     return new_intlist;
    }
}
 
 
/*
**  Function  : void int_list_release(Intlist *rel_intlist)
**  Return    : none.
**  Remarks   : the Intlist record pointed by 'rel_intlist' is
**              placed on the top of the Inlist stack (the value of
**              'top_intlist_flist' is changed).
*/
void int_list_release(Intlist *rel_intlist)
{
 rel_intlist->next = top_intlist_flist;
 top_intlist_flist = rel_intlist;
}

/*
**  Function  : Intlist *dbl_list_alloc(void)
**  Return    : the pointer to the Dbllist record just summoned.
**  Remarks   : the pointer to the record on the top of the 'Dbllist' free
**              list is returned if different than NULL (the value of
**              'top_dbllist_flist' is changed in this case). If the
**              'Dbllist' free list is empty, a new memory area is
**              dinamically allocated.
*/
Dbllist *dbl_list_alloc(void)
{
 Dbllist *new_dbllist;
 
 if(top_dbllist_flist == NULL)    /* The Intlist free list is empty.    */
    {
# ifdef MEMORY_DEBUG
  n_dbllist++;
# endif
     return ((Dbllist *)MMmalloc(sizeof(Dbllist), "dbl_list_alloc"));
    }
 else                           /* The Intlist free list is not empty.*/
    {
     new_dbllist = top_dbllist_flist;
     top_dbllist_flist = new_dbllist->next;
     new_dbllist->next = NULL;
     return new_dbllist;
    }
}
 
 
/*
**  Function  : void dbl_list_release(Intlist *rel_dbllist)
**  Return    : none.
**  Remarks   : the Dbllist record pointed by 'rel_dbllist' is
**              placed on the top of the Inlist stack (the value of
**              'top_dbllist_flist' is changed).
*/
void dbl_list_release(Dbllist *rel_dbllist)
{
 rel_dbllist->next = top_dbllist_flist;
 top_dbllist_flist = rel_dbllist;
}

