/*
**  Module    : getpaths.c
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano,
**              Luca Medico, Gabriele Favalessa
**
**  Synopsis  : this module collects the functions regarding the static
**              issues of the routing algorithm. These issues are
**              independent of the selected routing algorithm.
**              All the available paths between each pair of nodes are
**              first detected. Then, the paths are ordered according to 
**              the criterion previously selected between two available 
**              options: Number of Hops or Total Path Length, where the 
**              meaning of 'length' may be either "physical length" of the 
**              path or "total cost" of the path, defined on the basis of a 
**              suitable cost function.
**              Since the ordering criterion has been selected, multiple
**              options are available for bounding the number of usable
**              paths between each pair of nodes. Only the set of permitted
**              paths is kept when the path detection is over.
**              A linked list of path records is thus created for each
**              ordered pair of users. All the linked lists are pointed
**              by the elements of an N x N matrix, where N is the number
**              of network users.
**
**  Global Functions :  void    find_all_paths(void)
**                      Boolean equ_double(double, double, double)
**                      static  void open_new_path(Path****, int     , int    ,
**                                                 	int    , int     *)
**                      static  Boolean there_is_loop(int    )
**                      static void  find_all_cgen_to_cgen_paths(Path****);
**                      static void  find_node_node_paths(Path****, int    ,
**                                                        int    , int    );
**                      Path**** alloc_path_mat(int    , int    );
**                      static void  free_path_mat(Path**** , int     ,
**                                                             int     num);
**						static void trunc_src_pathlist(Path** path, int src);
**						static void trunc_glob_pathlist(Path** path);
**
**						static Path*** create_cgen_cgen_routes( void);
**
**
**  History   : 21-SEP-1995 - Reorganization of the module.     -AF-
**              01-JUN-1996 - Revision.                         -AF-
**              01-MAR-1997 - Revision				-MM-
**              14-OCT-1997 - Ported the routing information at the
**                            CG level                          -MM-
**              23-JAN-1998 - Changed the path_matrix: now the path are stored
**                            from user to user (not node to node) -LM-GF-MM-
**

31-01-2001 Cleaning... -ZZ-
**              30-APR-2001 - Introduce the support of duplex optic links -ZZ-
**				07-EP-2001 - Independent path pairs for the multipath routing -ZZ-
**				   DEC-2001 - Store only the needed paths (consider constraints) -ZZ-
**				11-APR-2002 - Link substitution for section protection
*/

#include "ancles.h"

/***************************
**  Constant Definitions  **
***************************/

/******************************
**  Static Global Variables  **
******************************/

/* Auxiliary structure used to find all paths */
struct stack_elem {
  int     node_id, port_number, direction;
  Boolean	endopsec;	/* The end-node of the link can be the end
							of an opaque section. */
};

static struct stack_elem* stack_bottom, * stack_top;

/**********************************
**  Static Functions Prototypes  **
**********************************/

/*static void free_path_mat(Path ****nn_mat, int     size, int     num);*/
static int cmp_path_length(Path *a, Path *b);
static int cmp_path_no_of_hops(Path *a, Path *b);
static Path *clone_path_list(Path *path);
static Path *clone_pathlink_list(Path *path);
static LinkList *clone_linklist(LinkList *orig);
static Path *build_cgen_to_cgen_path_list(Path ***node_to_node_path_mat,
					  int     s, int     d);
static Path *find_node_node_paths(int     start_node, int     end_node);
static Path *sort_path_list(Path *unsorted, int (*cmp)(Path *, Path *));
static Path* shuffle_equivalent_paths(Path* list);
static Path *create_new_path(double length, int no_of_hops,
							 int no_of_opcros);
static Boolean there_is_loop(int     node);
static Pair* pairs_length_ordering(Pair *list);
static void trunc_src_pathlist(Path** path, int src);
static void trunc_glob_pathlist(Path** path, int max_more_hop,
								int min_hopnum);
static Path*** create_cgen_cgen_routes( void);
static void free_linkpathlist(Path* path);


/*
**  Function  : Path * ***alloc_path_mat(int     size, int     num)
**  Return    : the the new 3D-matrix of pointer.
**  Remarks   : this function allocate a 3D-matrix of Path pointer:
**              Path *path_mat[num][size][size]
*/
Path * ***alloc_path_mat(int     size, int     num)
{
   int ord, row, column;
   Path**** nn_mat;

   nn_mat = (Path****)MMmalloc(num * sizeof(Path***), "alloc_path_mat");

   for (ord = 0; ord < num; ++ord)
   {
      nn_mat[ord] = (Path***)MMmalloc(size * sizeof(Path**), "alloc_path_mat");

      for (row = 0; row < size; ++row)
      {
         nn_mat[ord][row] = (Path**)MMmalloc(size * sizeof(Path*), 
					     "alloc_path_mat");

         for (column = 0; column < size; ++column)
            nn_mat[ord][row][column] = NULL;
      }
   }
   return nn_mat;
}

/*
**  Function  : Pair * ***alloc_pair_mat(int     size, int     num)
**  Return    : the the new 3D-matrix of pointer.
**  Remarks   : this function allocate a 3D-matrix of Pair pointer:
**              Pair *pair_mat[num][size][size]
*/
Pair * ***alloc_pair_mat(int     size, int     num)
{
   int ord, row, column;
   Pair**** nn_mat;

   nn_mat = (Pair****)MMmalloc(num * sizeof(Pair***), "alloc_pair_mat");

   for (ord = 0; ord < num; ++ord)
   {
      nn_mat[ord] = (Pair***)MMmalloc(size * sizeof(Pair**), "alloc_pair_mat");

      for (row = 0; row < size; ++row)
      {
         nn_mat[ord][row] = (Pair**)MMmalloc(size * sizeof(Pair*), 
					     "alloc_pair_mat");

         for (column = 0; column < size; ++column)
            nn_mat[ord][row][column] = NULL;
      }
   }
   return nn_mat;
}


/* Allocate a SIZE per SIZE matrix. */
static Path*** alloc_bidim_path_mat(int     size)
{
   int row, column;
   Path*** nn_mat;

   nn_mat = (Path***)MMmalloc(size * sizeof(Path**), "alloc_path_mat");

   for (row = 0; row < size; ++row)
     {
       nn_mat[row] = (Path**)MMmalloc(size * sizeof(Path*), 
				      "alloc_path_mat");
       
       for (column = 0; column < size; ++column)
	 nn_mat[row][column] = NULL;
     }

   return nn_mat;
}

/* Free the list of paths and all the paths it contains,
   but not the linkpointers!  */
void free_pathlist(Path* path)
{
	Path* prev;
	
	while(path)
	 {
		prev = path;
		path = path->next;
		path_release(prev);
	}
}

/* Free the list of paths and all the paths it contains
   and the link_pointers too. */
static void free_linkpathlist(Path* path)
{
	LinkList* link = NULL;
	LinkList* next;
	Path* prev;
	
	while(path)
	{
		link = path->path;
		
		while (link)
		{
			
			next = link->next;
			linklist_release(link);
			link = next;
		}
		
		prev = path;
		path = path->next;
		path_release(prev);
	}
}

/* Free the path and its link_pointers too.
   Returns the next path */
Path* free_linkpath(Path* path)
{
	LinkList* link = NULL;
	LinkList* next;
	Path*	ret;

	if(path)
	{
		link = path->path;
		
		while (link)
		{
			
			next = link->next;
			linklist_release(link);
			link = next;
		}
		
		ret = path->next;
		path_release(path);
		return ret;
	}
	else
		return path;
}

/* Free a SIZE per SIZE matrix and all the paths it contains. */
static void free_bidim_path_mat(Path*** nn_mat, int     size)
{
   int row, column;

   for (row = 0; row < size; ++row)
     {
       for (column = 0; column < size; ++column)
         {
	   if (nn_mat[row][column])
	     {
               LinkList* link = NULL;

	       if (nn_mat[row][column])
		 link = nn_mat[row][column]->path;

               while (link)
		 {
		   LinkList* next;

		   next = link->next;
		   linklist_release(link);
		   link = next;
		 }

               path_release(nn_mat[row][column]);
	     }
         }

       free(nn_mat[row]);
     }

   free(nn_mat);
}

/*
**  Function  : void free_path_mat(Path **** nn_mat, int     size, int     num)
**  Return    : none
**  Remarks   : this function free a 3D-matrix of Path pointer and all the
**              paths stored in it:
**              free (Path *path_mat[num][size][size])
*/
static void free_path_mat(Path**** nn_mat, int     size, int     num)
{
   int ord, row, column;

   for (ord = 0; ord < num; ++ord)
   {
      for (row = 0; row < size; ++row)
      {
         for (column = 0; column < size; ++column)
         {
            if (nn_mat[ord][row][column])
            {
               LinkList* link = NULL;

	       if (nn_mat[ord][row][column])
		 link = nn_mat[ord][row][column]->path;

               while (link)
               {
                  LinkList* next;

                  next = link->next;
                  linklist_release(link);
                  link = next;
               }

               path_release(nn_mat[ord][row][column]);
            }
         }

         free(nn_mat[ord][row]);
      }

      free(nn_mat[ord]);
   }

   free(nn_mat);
}

 
/**********************************
**  Global Function Definitions  **
**********************************/

/* Print to stderr the average number of hops and the average length
   of all paths connecting node S to node D. */
static void print_average_path_length(Path* path, int     s, int     d)
{
  if (path)
    {
      int count = 0, primary_hops, sum_hops = 0, max_hops, i;
      int distrib[MAX_NODE] = { 0 };

      primary_hops = path->no_of_hops;

      ++distrib[primary_hops];
      max_hops = primary_hops;

      for (path = path->next ; path; path = path->next)
	{
	  sum_hops += path->no_of_hops;
	  ++count;

	  ++distrib[path->no_of_hops];

	  if (path->no_of_hops > max_hops)
	    max_hops = path->no_of_hops;
	}

      fprintf(stderr, "%.8s <-> %.8s\t%.3g\t%.3g\t%2d  :", 
	      node_name(s),
	      node_name(d),
	      (sum_hops + primary_hops) / (count + 1.0),
	      (double)sum_hops / count,
	      count + 1);

      for (i = 1; i <= max_hops; ++i)
	fprintf(stderr, " %2d", distrib[i]);

      fprintf(stderr, "\n");
    }
}
 
/*
**  Function  : void find_all_paths(void)
**  Return    : none.
**  Remarks   : this function detects all the available paths between the
**              users. We look for all the paths, for all the user-user
**              traffic relations.
*/
void find_all_paths(void)
{
  Path*** nn_mat;
  int     s, d,lid;
  Path*   actpath;
  Path*   actroute;
  Path*** cgroute_mat;

  /* Temporary matrix for node to node path lists. */
  nn_mat = alloc_bidim_path_mat(no_of_nodes);

  /* if virtual links are used, node-node pathmatrix has to be created */
  if(ipowdm_ena)
     virt_nn_mat = alloc_bidim_path_mat(no_of_nodes);

  stack_bottom = MMmalloc(no_of_nodes * sizeof(struct stack_elem), "find_all_paths");

  for (s = 0; s < no_of_nodes; ++s)
  {
	  for (d = 0; d < no_of_nodes; ++d)
	  {
		  if (s != d)
			  nn_mat[s][d] = find_node_node_paths(s, d);
		  else
			  nn_mat[s][d] = NULL;

		  if(ipowdm_ena)
		  /* the dynamic node-node path matrix is empty at the beginning */
		    virt_nn_mat[s][d] = NULL;
	  }
  }

  /* create the list of paths which evite the substituted link */
  subslinkid = -1;
  subspaths = NULL;
  for(lid = 0; lid < no_of_links; lid++)
	  if(link_vect[lid]->substituted)
	  {
		  /* Only one substituable link is considered */
		  if(subslinkid != -1)
		  {
			  printf("Error:  Only one link can be substituted by \
section protecion!\n");
			  exit(1);
		  }

		  subslinkid = lid;
		  subspaths = clone_path_list(nn_mat[link_vect[lid]->
			  source][link_vect[lid]->dest]);
		  /* reallocate linklists to not loose them at the free 
		      of the nn_mat matrix */
		  for(actpath = subspaths; actpath != NULL;
		  actpath = actpath->next)
			  actpath->path = clone_linklist(actpath->path);
	  }

  free(stack_bottom);

  /* Sort every path list in the matrix.  Sorting the node to node
     matrix instead of the CG to CG path matrix saves some time (the
     first matrix is usually smaller) without altering the order
     obtained since all the CG to CG paths are simply 2 hops longer
     than user to user paths. */
  for (s = 0; s < no_of_nodes; ++s)
    for (d = 0; d < no_of_nodes; ++d)
      nn_mat[s][d] = sort_path_list(nn_mat[s][d], cmp_path_no_of_hops);

  /* Build a CG to CG path matrix (whose paths include user->node and
     node->user links) starting from the node to node path matrix.*/
  /* If there are predefined routes, create CG to CG routes */
  for (s = 0; s < no_of_cgens; ++s)
	{
		if(callgen_vect[s]->type != CG_TYPE10 &&
			callgen_vect[s]->type != CG_TYPE11 &&
			callgen_vect[s]->type != CG_TYPE12 &&
		    /* in case of virtual topology for IP generators, the 
			list of paths is empty first */
			(callgen_vect[s]->type > CG_TYPE5 ||
			 !ipowdm_ena))
		{
			if(callgen_vect[s]->choosen_routing.path_order_type ==
				IN_ORDER_OF_HOP_NUMBER ||
				callgen_vect[s]->choosen_routing.path_order_type ==
				BALANCED_IN_ORDER_OF_HOP_NUMBER)
			{
				for (d = 0; d < no_of_cgens; ++d)
					path_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = 
					shuffle_equivalent_paths(build_cgen_to_cgen_path_list(nn_mat, s, d));
			}
			else
			{
				for (d = 0; d < no_of_cgens; ++d)
					path_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = 
					build_cgen_to_cgen_path_list(nn_mat, s, d);
			}

		}
		else
			for (d = 0; d < no_of_cgens; ++d)
				path_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = NULL;
				
	}

  if (init_debug_level & DEBUG_MEAN_PATH_LENGTH)
    for (s = 0; s < no_of_nodes; ++s)
      for (d = s; d < no_of_nodes; ++d)
	print_average_path_length(nn_mat[s][d], s, d);
	  
  /* If there are predefined routes, create CG to CG routes */
  cgroute_mat = create_cgen_cgen_routes();

  /* Free the temporary matrix. */
  free_bidim_path_mat(nn_mat, no_of_nodes);  

  /* initialize the no_times_this_path_is_utilized vector */
  max_cgen_cgen_no_of_path = 0;
  /* find the maximum number of paths */
  for(s = 0; s < no_of_cgens; s++)
	  for(d = 0; d < no_of_cgens; d++)
	  {
		  int count = 0;
		  Path * this_path;
		  if (path_mat[0][s][d] == NULL)
			  continue;

		  /* don't forget the predefined routes (still not inserted in the list) */
		  for(actroute = cgroute_mat[s][d]; actroute; 
			actroute = actroute->next)
				count++;
	  
		  for(this_path = path_mat[0][s][d];
			this_path != NULL; this_path=this_path->next)
				count++;
		  if(count > max_cgen_cgen_no_of_path)
			  max_cgen_cgen_no_of_path = count;
	  }
  /* now allocate the vector */
  no_times_this_path_is_utilized =
	  (int *)MMmalloc(sizeof(int)*max_cgen_cgen_no_of_path, "find_all_paths");
  for(s=0; s < max_cgen_cgen_no_of_path; s++)
	  no_times_this_path_is_utilized[s]=0;

	  /* Build the CG to CG path matrix ordered by length by cloning and
     then reordering the path lists according to the choosen ordering */
  for (s = 0; s < no_of_cgens; ++s)
	  for (d = 0; d < no_of_cgens; ++d)
      {
		  Path* list;

		  if(path_mat[IN_ORDER_OF_HOP_NUMBER][s][d])
			  hopdist[callgen_vect[s]->ref_user->id]
			  [callgen_vect[d]->ref_user->id] = 
			  path_mat[IN_ORDER_OF_HOP_NUMBER][s][d]->no_of_hops;
		  else
			  hopdist[callgen_vect[s]->ref_user->id]
			  [callgen_vect[d]->ref_user->id] = -1;
		  
		  if(callgen_vect[s]->choosen_routing.path_order_type ==
			  IN_ORDER_OF_LENGTH ||
			 callgen_vect[s]->choosen_routing.path_order_type ==
			  BALANCED_IN_ORDER_OF_LENGTH)
		  {
			  list = clone_path_list(path_mat[IN_ORDER_OF_HOP_NUMBER][s][d]);
			  
			  if(list)
			  {
				  list = sort_path_list(list, cmp_path_length);
				  list = shuffle_equivalent_paths(list);
			  }

    /* if length based ordering is used, the hop number based matrices are no
	   more necessary, neither the paths excluded by constraints */
			  free_pathlist(path_mat[IN_ORDER_OF_HOP_NUMBER][s][d]);
			  path_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = NULL;

	/* insert the predefined routes at the beginning of the pathlist
			  (also the cloning of the linklists is needed) */
			  actroute = clone_pathlink_list(cgroute_mat[s][d]);
			  if(actroute)
			  {
				  for(actpath = actroute; actpath->next;
					actpath = actpath->next);
				  actpath->next = list;
				  list = actroute;
			  }
			  
			  trunc_src_pathlist(&list, s);		  
		  }
		  else
    /* if hop number based ordering is used, the length based matrices are no
	   more necessary, neither the paths excluded by constraints */
		  {
			  list = NULL;
  
	/* insert the predefined routes at the beginning of the pathlist
			  (also the cloning of the linklists is needed) */
			  actroute = clone_pathlink_list(cgroute_mat[s][d]);
			  if(actroute)
			  {
				  for(actpath = actroute; actpath->next;
					actpath = actpath->next);
				  actpath->next = 
					  path_mat[IN_ORDER_OF_HOP_NUMBER][s][d];
				  path_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = actroute;
			  }
			  
			  trunc_src_pathlist(&path_mat[IN_ORDER_OF_HOP_NUMBER][s][d], s);
		  }
 
		  path_mat[IN_ORDER_OF_LENGTH][s][d] = list;

      }


  /* Free the temporary matrix. */
  free_bidim_path_mat(cgroute_mat, no_of_cgens);  

  whole_linklist_release();

/*printf("Size: %d\n", sizeof(LinkList));
printf("tfree: %x length:%d\n", &tfree, (8 * MAX_WVOCTETS * MAX_FIBER_NUM + 1)*sizeof(int));
printf("utso: %x\n", &tfree[8 * MAX_WVOCTETS * MAX_FIBER_NUM + 1]);
for(actpath=path_mat[IN_ORDER_OF_LENGTH][12][9]; actpath; actpath = actpath->next)
{
	LinkList* link;

for(link=actpath->path; link; link = link->next)
{
	printf("%s ", link_name(link->link_elem->id));
}
printf("; ");
}
* exit(0);*/
}


/*
**  Function  : void find_pathpairs(void)
**  Return    : none.
**  Remarks   : this function detects all the available pairs of 
**				independent paths between the users whith a more path
**				needing routing function.
*/
void find_all_pairs(void)
{
	Pair*	actpair;
	Pair*	npair;
	Path*	path1;
	Path*	path2;
	int     s, d, i, nphops;
	double	nplength;
	LinkList	*link, *last_link;
	Boolean independent, path1links[MAX_LINK], found, path1nodes[MAX_NODE];
	
	for (s = 0; s < no_of_cgens; ++s)
	{
		for (d = 0; d < no_of_cgens; ++d)
		{
			pair_mat[IN_ORDER_OF_LENGTH][s][d] = NULL;
			pair_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = NULL;
			if (
				callgen_vect[s]->type != CG_TYPE10 &&
				callgen_vect[s]->type != CG_TYPE11 &&
				callgen_vect[s]->type != CG_TYPE12 &&
				s != d && 
				(callgen_vect[s]->choosen_routing.selected_routing >= 70 &&
				 callgen_vect[s]->choosen_routing.selected_routing < 90 || 
				 callgen_vect[s]->choosen_routing.selected_routing >= 120 &&
				 callgen_vect[s]->choosen_routing.selected_routing < 140)
				)
			{
				for(path1 = path_mat[callgen_vect[s]->
					choosen_routing.path_order_type][s][d];
				path1 != NULL; path1 = path1->next)
				{
					for(i = 0; i < no_of_links; i++)
					/* this array shows which links are enabled in the
					other path */
					path1links[i] = TRUE;
					
					for(i = 0; i < no_of_nodes; i++)
					/* this array shows which nodes are enabled in the
					other path */
					path1nodes[i] = TRUE;
					
					/* zero length paths */
					last_link = path1->path;
					for(link = path1->path->next; link->next!=NULL;
					link = link->next)
					{
						/* mark the links and nodes of the first path */
						path1links[link->link_elem->id] = FALSE;
						path1nodes[link->link_elem->source] = FALSE;
						path1nodes[link->link_elem->dest] = FALSE;
						last_link = link;
					}
					
					/* Start and endpoint of the route */		
					path1nodes[last_link->next->link_elem->source] = TRUE;
					path1nodes[path1->path->link_elem->dest] = TRUE;
					
					for(path2 = path1->next; path2 != NULL;
					path2 = path2->next)
					{
						independent = TRUE;
						for(link = path2->path->next; link->next!=NULL;
						link = link->next)
						{
							/* check the common links and nodes */
							independent = independent &&
								(path1links[link->link_elem->id]) &&
								(path1nodes[link->link_elem->source]) &&
								(path1nodes[link->link_elem->dest]);
						}
						
						if(independent)
						{
						/* if independent pair found, insert it in the 
							linked list ordered by the sum of hop number */
							npair = (Pair *)MMmalloc(sizeof(Pair),
								"find_all_pairs");
							npair->path1 = path1;
							npair->path2 = path2;
							nphops = path1->no_of_hops + path2->no_of_hops;
							nplength = path1->length + path2->length;
							npair->next = NULL;
							
							switch(callgen_vect[s]->
								choosen_routing.path_order_type)
							{
								/* ordering the pairs by the sum of hop numbers */
							case IN_ORDER_OF_HOP_NUMBER:
								{
									
									/* insert the first pair */
									if(!pair_mat[IN_ORDER_OF_HOP_NUMBER][s][d])
										pair_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = npair;
									else
									{
										actpair = pair_mat[IN_ORDER_OF_HOP_NUMBER][s][d];
										/* insert to the first place */
										if(actpair->path1->no_of_hops + 
											actpair->path2->no_of_hops > nphops)
										{
											npair->next = actpair;
											pair_mat[IN_ORDER_OF_HOP_NUMBER][s][d] = npair;
										}
										else
										{
											found = FALSE;
											while(actpair->next && !found)
											{
												/* right insertion place is found */
												if(actpair->next->path1->no_of_hops + 
													actpair->next->path2->no_of_hops > nphops)
													found = TRUE;
												else
													actpair = actpair->next;
											}
											/* insert the new pair */
											npair->next = actpair->next;
											actpair->next = npair;
											
										}
									}
								}
								break;
								
								
								/* ordering the pairs by the sum of pathlengths */
							case IN_ORDER_OF_LENGTH:
								{
									
									/* insert the first pair */
									if(!pair_mat[IN_ORDER_OF_LENGTH][s][d])
										pair_mat[IN_ORDER_OF_LENGTH][s][d] = npair;
									else
									{
										actpair = pair_mat[IN_ORDER_OF_LENGTH][s][d];
										/* insert to the first place */
										if(actpair->path1->length + 
											actpair->path2->length > nplength)
										{
											npair->next = actpair;
											pair_mat[IN_ORDER_OF_LENGTH][s][d] = npair;
										}
										else
										{
											found = FALSE;
											while(actpair->next && !found)
											{
												/* right insertion place is found */
												if(actpair->next->path1->length + 
													actpair->next->path2->length > nplength)
													found = TRUE;
												else
													actpair = actpair->next;
											}
											/* insert the new pair */
											npair->next = actpair->next;
											actpair->next = npair;
											
										}
									}
								}
								break;
								
							default:
								printf("\nError:  Generator %s has a non allowed type,", cgen_name(s));
								printf("\n        of path ordering (no load balancing");
								printf("\n        in protection is available.");
								exit(1);
								
							}
						} /* indep */
					} /* path2 */
				} /* path1 */
			} /* generator && s!=d && routing */
		} /* d */
	} /* s */
}


/* Create an ordered linked list of pathpairs. The order function
	is the sum of the lengths of the two paths. Allocates the 
	neccessary memory too. */
static Pair* pairs_length_ordering(Pair *list)
{
	Pair* npair;
	Pair* actpair;
	Pair* newlist = NULL;
	double nplength;
	Boolean found;
	
	for(; list != NULL; list = list->next)
	{
		npair = (Pair *)MMmalloc(sizeof(Pair),
			"pairs_length_odering");
		npair->path1 = list->path1;
		npair->path2 = list->path2;
		nplength = list->path1->length + list->path2->length;
		npair->next = NULL;
		
		/* insert the first pair */
		if(!newlist)
			newlist = npair;
		else
		{
			actpair = newlist;
			/* insert to the first place */
			if(actpair->path1->length + 
				actpair->path2->length > nplength)
			{
				npair->next = actpair;
				newlist = npair;
			}
			else
			{
				found = FALSE;
				while(actpair->next && !found)
				{
					/* right insertion place is found */
					if(actpair->next->path1->length + 
						actpair->next->path2->length > nplength)
						found = TRUE;
					actpair = actpair->next;
				}
				/* insert the new pair */
				npair->next = actpair->next;
				actpair->next = npair;
			}
		}
	}
	return newlist;
}

/* Shuffle the elements of LIST. Return the head of the shuffled
   list. If TAIL is not NULL place the tail of the shuffled list in
   *TAIL. */
static Path* shuffle_path_list(Path* list, Path** tail)
{
  Path* shuffled_list = NULL;
  Path* path, * prev_path;
  int count = 0, path_number;

  if (tail)
    *tail = NULL;

  /* Count the paths in the original list. */
  for (path = list; path; path = path->next)
    ++count;

  /* Iterate until the original list is empty. */
  for(; count; count--)
    {
      /* Select at random one path in the original list. */
      path_number = (int)floor((double)count * uniform_0_1(&sh_seed));

      /* Fix the upper extreme problem (we should use a uniform
	 distibution in [0,1) not in [0,1]).  */
      if (path_number == count)
	--path_number;

      for (path = list, prev_path = NULL;
	   path_number--;
	   prev_path = path, path = path->next)
	;

      /* Remove the selected path from the original list. */
      if (prev_path)
	prev_path->next = path->next;
      else
	list = path->next;

      /* Put the removed path at the head of the shuffled list. */
      path->next = shuffled_list;
      shuffled_list = path;

      if (tail && *tail == NULL)
	*tail = path;
    }

  return shuffled_list;
}

/* Shuffle the groups of paths contained in LIST that have the same
   length and number of hops. */
static Path* shuffle_equivalent_paths(Path* list)
{
  Path* path, * old_last_eq_path = NULL;
  Path* first_eq_path, * last_eq_path;

  path = list;
  while (path)
    {
      first_eq_path = path;

      while (path 
	     && equ_double(path->length, first_eq_path->length, DBL_CMP_THS)
	     && path->no_of_hops == first_eq_path->no_of_hops)
	{
	  last_eq_path = path;
	  path = path->next;
	}

      /* The group of equivalent paths starts with FIRST_EQ_PATH and
	 ends with LAST_EQ_PATH. OLD_LAST_EQ_PATH is the path
	 preceding the group and PATH is the path following the
	 group. */
      if (first_eq_path != last_eq_path)
	{
	  last_eq_path->next = NULL;

	  if (old_last_eq_path)
	    old_last_eq_path->next = shuffle_path_list(first_eq_path, &last_eq_path);
	  else
	    list = shuffle_path_list(first_eq_path, &last_eq_path);
	
	  last_eq_path->next = path;
	}

      old_last_eq_path = last_eq_path;	
    }

  return list;
}

static int cmp_path_length(Path* a, Path* b)
{
  if (equ_double(a->length, b->length, DBL_CMP_THS))
    return a->no_of_hops - b->no_of_hops;
  else if (a->length > b->length)
    return 1;
  else
    return -1;
}

static int cmp_path_no_of_hops(Path* a, Path* b)
{
  if (a->no_of_hops == b->no_of_hops)
    {
      if (equ_double(a->length, b->length, DBL_CMP_THS))
	return 0;
      else if (a->length > b->length)
	return 1;
      else
	return -1;
    }

  return a->no_of_hops - b->no_of_hops;
}

/* The original and the cloned path list share the same link list
   structures (the link lists are not reallocated).  */
static Path* clone_path_list(Path* path)
{
   int i;
   Path* path_head = NULL, * last_path = NULL;
   
   for(; path != NULL; path = path->next)
   {
      if (last_path != NULL)
       {
         last_path->next = path_alloc();
         last_path = last_path->next;
       }
      else
       {
         last_path = path_alloc();
         path_head = last_path;
       }

      last_path->next = NULL;
      last_path->path = path->path;
      last_path->length = path->length;
      last_path->no_of_hops = path->no_of_hops;
      last_path->no_of_opaque_crs = path->no_of_opaque_crs;

	  last_path->num_of_prot_chs = path->num_of_prot_chs;
	  last_path->prot_channels = path->prot_channels;
	  for(i = 0; i<no_of_nnlinks; i++)
		  last_path->protected_links[i] = path->protected_links[i];
	  for(i = 0; i<no_of_nodes; i++)
		  last_path->protected_nodes[i] = path->protected_nodes[i];

   }

   return path_head;
}

/* The original and the cloned path list do not share the same link list
   structures (the link lists are actually reallocated).  */
static Path* clone_pathlink_list(Path* path)
{
   Path* path_head = NULL;

   path_head = clone_path_list(path);

   for(path = path_head; path != NULL; path = path->next)
      path->path = clone_linklist(path->path);

   return path_head;
}

/*
**  Function  : LinkList* clone_linklist(LinkList* orig)
**  Return    : the cloned linklist.
**  Remarks   : auxiliary funztion used to clone the LinkList.
*/
static LinkList* clone_linklist(LinkList* orig)
{
  LinkList* new = NULL, * last = NULL;

  for(; orig != NULL; orig = orig->next)
    {
      if (last != NULL)
       {
	  last->next = linklist_alloc();
	  last = last->next;
       }
      else
       {
	 last = linklist_alloc();
	 new = last;
       }

      last->next = NULL;
      last->link_elem = orig->link_elem;
      last->direction = orig->direction;
      last->endopsec = orig->endopsec;

    }

  return new;
}

/*
**  Function  : void trunc_src_pathlist(Path** path, int src)
**  Return    : None
**  Remarks   : Releases those paths from the list, which
**				do not accord to the constraints of the source
*/
static void trunc_src_pathlist(Path** path, int src)
{
	int i = 0;  /* number of examined paths */
	Boolean accepted = TRUE;
	Path* actual_path;
	Path* prev_path;
	
	if( (*path) == NULL )
		return;
	
	actual_path = (*path);
	/* Still the paths are right */
	while(accepted && actual_path != NULL)
	{
		i++;   
		/* Control the constraint */
		switch(callgen_vect[src]->
			choosen_routing.selected_range_type.type)
		{
		case ALL_PATHS_NUMBER:
			/* Go throught remaining paths*/
			break;
		case ONLY_PRIMARY_PATH:
			if (i > 1)
				accepted = FALSE;
			break;
		case FIXED_PATH_NUMBER:
			if (i > callgen_vect[src]->choosen_routing.
				selected_range_type.value.ival )
				accepted = FALSE;
			break;
		case MAX_HOP_PATH_NUMBER:
			if (actual_path->no_of_hops >
				callgen_vect[src]->choosen_routing.
				selected_range_type.value.ival )
				accepted = FALSE;
			break;
		case MAX_HOP_MORE_PATH_NUMBER:
			if (actual_path->no_of_hops >
				callgen_vect[src]->choosen_routing.
				selected_range_type.value.ival + 
				(*path)->no_of_hops)
				accepted = FALSE;
			break;
		case MAX_LENGTH_PATH_NUMBER:
			if (actual_path->length >
				callgen_vect[src]->choosen_routing.
				selected_range_type.value.dval)
				accepted = FALSE;
			break;
		case MAX_LENGTH_MORE_PATH_NUMBER:
			if (actual_path->length >
				callgen_vect[src]->choosen_routing.
				selected_range_type.value.dval +
				(*path)->length)
				accepted = FALSE;
			break;
		default:
			break;
		}
		/* take the next path if the actual was still acceptable */
		if(accepted)
		{
			prev_path = actual_path;
			actual_path = actual_path->next;
		}
		
	}
	
	/* release is needed only if there are non-acceptable paths */
	if(!accepted)
	{
		free_linkpathlist(actual_path);
		/* if the first path (path*) is accepted */
		if( i > 1)
			prev_path->next = NULL;
		else
			(*path) = NULL;
	}
	return;
}

/*
**  Function  : void trunc_glob_pathlist(Path** path, int max_more_hop,
**								 int min_hopnum)
**  Return    : None
**  Remarks   : Releases those paths from the list, which
**				do not accord to the global relative constraint
*/
static void trunc_glob_pathlist(Path** path, int max_more_hop,
								int min_hopnum)
{
	Path* actual_path;
	Path* prev_path = NULL;
	Boolean min_found = FALSE;
	
	if( (*path) == NULL )
		return;
	
	actual_path = (*path);
	/* check all the paths before one with the minimal hopnum
	   (after this we had the right constraint in the precheck */
	while(!min_found && actual_path != NULL)
	{
		if(actual_path->no_of_hops == min_hopnum)
			min_found = TRUE;

		if(actual_path->no_of_hops > max_more_hop + min_hopnum)
		{
			/* remove the first from the actual list */
			if(prev_path==NULL)
				(*path) = actual_path->next;
			else
				prev_path->next = actual_path->next;

			actual_path = free_linkpath(actual_path);
		}
		else
		{
			prev_path = actual_path;
			actual_path = actual_path->next;
		}
	}

	return;
}

/* Return a list containing all the paths connecting the source call
   generator S to the destination call generator D.  If the source is
   a sink or is a transmitter but does not generate traffic the list
   is empty.  

   This function works by extending the existing node to node paths to
   include the user to node and node to user links.  If the call
   generators are connected to the same node the path connecting the
   two contains only the user to node and node to user links. */
static Path* build_cgen_to_cgen_path_list(Path*** node_to_node_path_mat,
		int     s, int     d)
{
	Path* path_list = NULL;
	double prev_dest_prob = 0.0;
	
	int i;
	User* su; /* source user */
	User* du; /* destination user */
	
	su = callgen_vect[s]->ref_user;
	du = callgen_vect[d]->ref_user;
	
	/* for the determination of the destination probability */
	if(du->id)
		prev_dest_prob = destination_prob[su->id][du->id-1];

	if ( ( (callgen_vect[s]->type != CG_TYPE10  )
		&& (callgen_vect[s]->type != CG_TYPE11  )
		&& (callgen_vect[s]->type != CG_TYPE12  )
		&& (traff_pattern[su->id][du->id] > 0.0 ) ) || 
		(callgen_vect[s]->type == CG_TYPE8 && 
		 (ipowdm_ena || /* if virtual links are used, no need of predefined
						never_closing_ason calls */
		 destination_prob[su->id][du->id] - prev_dest_prob > 0.0)) )
	{
		if ( (su->connected_node == du->connected_node)
			&& ( s != d) )
			/* two user connected by the same node*/
		{
			/* net->user link and user->net link */
			LinkList* nu_link, * un_link;
			/* node -> user link */
			nu_link = linklist_alloc();
			nu_link->link_elem = du->in_link;
			nu_link->direction = 0;
			nu_link->endopsec = FALSE;
			nu_link->next = NULL;
			/* user -> node link */
			un_link = linklist_alloc();
			un_link->link_elem = su->out_link;
			un_link->direction = 0;
			un_link->endopsec = FALSE;
			un_link->next = nu_link;
			
			path_list = path_alloc();
			path_list->path = un_link;
			path_list->next = NULL;
			path_list->no_of_hops = 0;
			path_list->length = 0.0;
			path_list->no_of_opaque_crs = 0;
			/*nu_link->link_elem->link_len +
			un_link->link_elem->link_len;*/
			
			path_list->num_of_prot_chs = 0;
			path_list->prot_channels = NULL;
			for(i = 0; i<no_of_nnlinks; i++)
				path_list->protected_links[i] = 0;
			for(i = 0; i<no_of_nodes; i++)
				path_list->protected_nodes[i] = 0;
			
		}
		else /* two user not connected by the same node */
		{
			/* path from node to node in the net */
			Path * nn_path;
			
			Path * last_path = NULL;
			/* the two user are not conncted to the same node */
			nn_path = node_to_node_path_mat[su->connected_node][du->connected_node];
			
			for ( ;nn_path; nn_path = nn_path->next)
			{
				/* net->user link and user->net link */
				LinkList* un_link, * nu_link;
				
				if (last_path)
				{
					/* the path list is already created */
					last_path->next = path_alloc();
					last_path = last_path->next;
				}
				else
					/* we must create the path list the first time */
					path_list = last_path = path_alloc();
				
				un_link = linklist_alloc();
				un_link->link_elem = su->out_link;
				un_link->direction = 0;
				un_link->endopsec = FALSE;
				un_link->next = clone_linklist(nn_path->path);
				
				for (nu_link = un_link; nu_link->next; nu_link = nu_link->next);
				
				nu_link->next = linklist_alloc();
				nu_link = nu_link->next;
				
				nu_link->link_elem = du->in_link;
				nu_link->direction = 0;
				nu_link->endopsec = FALSE;
				nu_link->next = NULL;
				
				last_path->path = un_link;
				last_path->next = NULL;
				/* not add the lenght of the in e out link 
				last_path->length = nn_path->length +
				un_link->link_elem->link_len +
				nu_link->link_elem->link_len; */
				last_path->length = nn_path->length;
				/* Too long paths are not allowed ! */
				if(last_path->length > MAX_PLENGTH)
				{
					printf("Error:  The network contains too long paths.\n");
					exit(1);
				}
				
				/* not add the user->net and net->user link to the no of hop */
				/* last_path->no_of_hops = nn_path->no_of_hops + 2; */
				last_path->no_of_hops = nn_path->no_of_hops;
				
				last_path->no_of_opaque_crs = nn_path->no_of_opaque_crs;
				
				last_path->num_of_prot_chs = nn_path->num_of_prot_chs;
				last_path->prot_channels = nn_path->prot_channels;
				
				for(i = 0; i<no_of_nnlinks; i++)
					last_path->protected_links[i] = nn_path->protected_links[i];
				
				for(i = 0; i<no_of_nodes; i++)
					last_path->protected_nodes[i] = nn_path->protected_nodes[i];
			}
			
		}
	}
	return path_list;
}

/* Return a list containing all the paths connecting the node
   START_NODE to the node END_NODE. */
static Path* find_node_node_paths(int     start_node, int     end_node)
{
  Path* path_list = NULL;

  int      curr_no_of_hops;	/* Number of hops forming the       */
				/* current path.                    */
  double   curr_length;		/* Total length of the current path. */

  int      curr_no_of_opcrs;	/* Number of opaque sections forming the */
				/* current path.                    */
  int     node_id;    /* Identifier of the node on the top of the stack. */

  int     port_id;    /* Current output port identifier for the node on
  			the top of the stack. */

  int     new_end;    /* The possible new endpoint of the actual path. */

  int		dir = 0;		/* Direction of the link in the path 
						(useful for duplex optics) */
  int	  curr_minhop;

  if(node_vect[start_node]->unreachable || 
	  node_vect[end_node]->unreachable)
	  return path_list;

  curr_no_of_hops = 0;
  curr_length = 0.0;
  curr_no_of_opcrs = 0;
  curr_minhop = MAX_NODE;

  stack_top = stack_bottom;
  stack_top->node_id = start_node;
  stack_top->port_number = -1;

  do
    {
      node_id = stack_top->node_id;

      if (node_id == end_node)
	{
	  Path* new_path;
	  if(node_vect[end_node]->opaque)
	    new_path = create_new_path(curr_length, curr_no_of_hops,
		  curr_no_of_opcrs-1);
	  else
	    new_path = create_new_path(curr_length, curr_no_of_hops,
		  curr_no_of_opcrs);

	  /* The minimal number of hop is derived adaptively */
	  if(curr_no_of_hops < curr_minhop)
		  curr_minhop = curr_no_of_hops;

	  /* Insert the new path at the beginning of the list. */
	  new_path->next = path_list;
	  path_list = new_path;

	  /* Move one step back: remove the destination node from the
             stack... */
	  --stack_top;
	  if(stack_top->endopsec)
		  --curr_no_of_opcrs;

	  /* ...and decrease length and number of hops. */
	  --curr_no_of_hops;
	  curr_length -=
	    node_vect[stack_top->node_id]
	    ->outqueue[stack_top->port_number]->link_len;
	}
      else
	{
	  /* The output port counter is incremented to check a new
             output link.  */
	  port_id = ++(stack_top->port_number);

	  if (port_id == node_vect[node_id]->no_of_ports)
	  {
	      /* No more output ports are available for this node,
                 remove it from the stack.  */
	      --stack_top;

	     if (stack_top >= stack_bottom)
		 {
		  if(stack_top->endopsec)
		    --curr_no_of_opcrs;
		  --curr_no_of_hops;
		  curr_length -=
		    node_vect[stack_top->node_id]
		    ->outqueue[stack_top->port_number]->link_len;
		 }
	  }
	  else
	  {
	      /* Pointer to the currently checked link.  */
	      Link    *link_ptr;

	      link_ptr = node_vect[node_id]->outqueue[port_id];

	      /* temporaneo per accorciare la ricerca dei percorsi */
#ifdef XXX_MAX_NO_OF_HOPS
	      if (curr_no_of_hops <= XXX_MAX_NO_OF_HOPS)
#endif
#ifdef XXX_MAX_LENGTH
		if (((curr_length + link_ptr->link_len) < XXX_MAX_LENGTH) ||
		    (equ_double(curr_length + link_ptr->link_len, XXX_MAX_LENGTH,
				DBL_CMP_THS)))
#endif

		/* Precheck of the global hop constraints */
	    if(curr_no_of_hops < xxx_max_no_of_hops &&
		   curr_no_of_hops < curr_minhop + xxx_more_no_of_hops)
		{

				
		/* The possible new endpoint of the path is calculated */
        /* If the link is a duplex fibre, there can be differences... */
        if(link_ptr->dest == node_id)
		{
          new_end = link_ptr->source;
		  dir = 1;
		}
        else
		{
          new_end = link_ptr->dest;
		  dir = 0;
		}
		  /* If no loop is closed by the current link... */
		  /* And the new node is reachable	*/
		  if (! there_is_loop(new_end)
			  && !node_vect[new_end]->unreachable)
		    {
		      /* Push the node into the stack. */
		      stack_top->direction = dir;
			  stack_top->endopsec = node_vect[new_end]->opaque;
			  if(stack_top->endopsec)
				  ++curr_no_of_opcrs;
		      ++stack_top;
		      stack_top->node_id = new_end;
			  stack_top->port_number = -1;

		      ++curr_no_of_hops;
		      curr_length += link_ptr->link_len;
		    }
		} /* precheck */
	  }
	}
    }
	while(stack_top >= stack_bottom);

/* if there is a global constraint for the paths control it*/
	trunc_glob_pathlist(&path_list, xxx_more_no_of_hops, curr_minhop);
	return path_list;
}

/* Sort a path list using the criterion defined by the comparing
   function 'cmp'. */
static Path* sort_path_list(Path* unsorted, int (*cmp)(Path*, Path*))
{
  Path* sorted = NULL;
  Path* unsorted_head; /* the current head of the unsorted part*/
  Path* prev_ins_point;/* pointer to the place where to insert*/
  Path* ins_point;     /* the new element                     */

  while (unsorted != NULL)
    {
      /* Remove the head from the unsorted list.  */
      unsorted_head = unsorted;
      unsorted = unsorted->next;

      /* Scan the sorted list to find the insertion point.  */
      for (prev_ins_point = NULL, ins_point = sorted;
	   /* scan the list until the insertion point is smaller then the */
	   /* unsorted head AND the insertion point is not at the end     */
	   (ins_point) && (cmp(ins_point, unsorted_head) < 0);
	   /* move to the next element */
	   prev_ins_point = ins_point, ins_point = ins_point->next);

      /* Insert the head of the unsorted list in the sorted list.  */
      if (prev_ins_point)
	{
	  unsorted_head->next = prev_ins_point->next;
	  prev_ins_point->next = unsorted_head;
	}
      else
	{
	  unsorted_head->next = sorted;
	  sorted = unsorted_head;
	}
    }

  return sorted;
}


/*
**  Function  : void substitute_link(int sl_id)
**  Return    : None
**  Remarks   : Substitute the sl_id link on each path of the
**				network, with the help of the list of paths
**				able to evite that link (listed in the global
**				subspaths node-node path list)
**				Reset path to null if independent section can 
**				not be inserted
*/
void substitute_link(int sl_id)
{
	int s, d, i;
	int	pathlinks[MAX_LINK];
		/* vectors for the indication of links in the path */
	Path	*actual_path, *prev_path, *del_path;
		/* pointers to handle the pathlists */
	Path	*spath; /* actual substitution-path */
	Boolean	found;	/* substitution was successful */
	LinkList	*link; /* actual link in the path */
	LinkList	*subslinkls; /* link to substitute */
	LinkList	*antesubsls; /* anterior link of that to substitute */
	LinkList	*llsclone;   /* linklist clone of substitution-path */


	/* modify the CG to CG path matrix */
	for (s = 0; s < no_of_cgens; ++s)
	{
		if(callgen_vect[s]->type != CG_TYPE10 &&
			callgen_vect[s]->type != CG_TYPE11 &&
			callgen_vect[s]->type != CG_TYPE12)
		{
			/* order the substitution paths */
			if(callgen_vect[s]->choosen_routing.
				path_order_type == IN_ORDER_OF_HOP_NUMBER ||
				callgen_vect[s]->choosen_routing.
				path_order_type == BALANCED_IN_ORDER_OF_HOP_NUMBER)
				subspaths = sort_path_list(subspaths,
				cmp_path_no_of_hops);
			else
				subspaths = sort_path_list(subspaths,
				cmp_path_length);

			for (d = 0; d < no_of_cgens; ++d)
			{
				prev_path = NULL;
				actual_path = path_mat[callgen_vect[s]->
					choosen_routing.path_order_type][s][d];
				/* each path will be rechecked if it contains 
				   the substituable link */
				while(actual_path != NULL)
				{
					antesubsls = NULL;
					subslinkls = NULL;
					for(link = actual_path->path; link != NULL &&
						subslinkls == NULL; link = link->next)
					{
						if(link->link_elem->id == sl_id)
							subslinkls = link;
						else
							antesubsls = link;
					}
					/* do nothing if path does not contain that link */
					if(subslinkls == NULL)
					{
						prev_path = actual_path;
						actual_path = actual_path->next;
					}
					else
					{
						for(i = 0; i < no_of_links; i++) 
							pathlinks[i] = 0;
						
						for(link = actual_path->path; link != NULL;
						link = link->next)
							pathlinks[link->link_elem->id] = 1;
						
						spath = subspaths;
						found = FALSE;
						
						/* find the "shortest" independent path */
						while(!found && spath!=NULL)
						{
							found = TRUE;
							for(link = spath->path; link != NULL && found;
							link = link->next)
								/* check if links are independent */
								if(pathlinks[link->link_elem->id])
									found = FALSE;
							
							if(!found)
								spath = spath->next;
						}
						
						/* if the substitution was not successfull */
						if(!found)
						{
							/* cut out the touched path */
							if(prev_path != NULL)
								prev_path->next = actual_path->next;
							else
								path_mat[callgen_vect[s]->choosen_routing.
								path_order_type][s][d] = actual_path->next;
							
							del_path = actual_path;
							actual_path = actual_path->next;
							/* delete the path if the link is not substituable */
							path_release(del_path);
						}
						else
						{
							/* substitute the link by the found path */
							llsclone = clone_linklist(spath->path);
							for(link = llsclone; link->next != NULL;
							link = link->next);
							if(antesubsls)
							{
						/* if we do not substitute the first link on the path */
								antesubsls->next = llsclone;
							}
							else
							{
								actual_path->path = llsclone;
							}
							link->next = subslinkls->next;

							/* recalculate the path properties */

							if(subslinkls->endopsec)
								actual_path->no_of_opaque_crs--;
							actual_path->no_of_hops += spath->no_of_hops - 1;
							actual_path->length += 
								spath->length - subslinkls->link_elem->link_len;
							actual_path->no_of_opaque_crs += spath->no_of_opaque_crs ;
							
							linklist_release(subslinkls);
							prev_path = actual_path;
							actual_path = actual_path->next;
						}
					} /* else */
				} /* while */
			} /* for d */
		} /* if */
	} /* for s */
}


/*
**  Function  : Path*** create_cgen_cgen_routes()
**  Return    : Creates a bidimensional matrix of paths that contains
**				the cgen to cgen versions of the predefined routes
**  Remarks   : The matrix pointer
*/
Path*** create_cgen_cgen_routes()
{
	Path*	a_path, * nn_path;
	Path*** nn_routes;
	Path*** cc_routes;
	Path * last_path;
	int s, d, i;

	double prev_dest_prob = 0.0;
	
	User* su; /* source user */
	User* du; /* destination user */
	LinkList* un_link, * nu_link;

	/* Create the node-node matrix */
	nn_routes = alloc_bidim_path_mat(no_of_nodes);

	/* Fill the nn matrix based on the routes */
	for(i = 0; i< no_of_routes; i++)
	{
		a_path = clone_pathlink_list(route_vect[i]->routepath);
		a_path->next =
			nn_routes[route_vect[i]->from][route_vect[i]->to];
		nn_routes[route_vect[i]->from][route_vect[i]->to] = a_path;
	}

	/* Allocate the cgen-cgen matrix */
	cc_routes = alloc_bidim_path_mat(no_of_cgens);

	/* Fill the cc matrix based on the nn matrix */
	for(s = 0; s < no_of_cgens; s++)
		for(d = 0; d < no_of_cgens; d++)
		{
			
			su = callgen_vect[s]->ref_user;
			du = callgen_vect[d]->ref_user;
			
			/* for the determination of the destination probability */
			if(du->id)
				prev_dest_prob = destination_prob[su->id][du->id-1];
			
				/* If the s-d pair performs traffic and route from node to node
			is defined */
			if ( (s != d) &&
				(nn_path = nn_routes[su->connected_node]
				[du->connected_node])
				&& (callgen_vect[s]->type != CG_TYPE10)
				&& (callgen_vect[s]->type != CG_TYPE11)
				&& (callgen_vect[s]->type != CG_TYPE12)
				&& ((traff_pattern[su->id][du->id] > 0.0) || 
				(callgen_vect[s]->type == CG_TYPE8 && 
				destination_prob[su->id][du->id] - prev_dest_prob > 0.0)) )
			{
				last_path = NULL;
				
				for ( ;nn_path; nn_path = nn_path->next)
				{
					/* Add net->user link and user->net link */
					
					if (last_path)
					{
						/* the path list is already created */
						last_path->next = path_alloc();
						last_path = last_path->next;
					}
					else
						/* we must create the path list the first time */
						cc_routes[s][d] = last_path = path_alloc();
					
					un_link = linklist_alloc();
					un_link->link_elem = su->out_link;
					un_link->direction = 0;
					un_link->endopsec = FALSE;
					un_link->next = clone_linklist(nn_path->path);
					
					for (nu_link = un_link; nu_link->next; nu_link = nu_link->next);
					
					nu_link->next = linklist_alloc();
					nu_link = nu_link->next;
					
					nu_link->link_elem = du->in_link;
					nu_link->direction = 0;
					nu_link->endopsec = FALSE;
					nu_link->next = NULL;
					
					last_path->path = un_link;
					last_path->next = NULL;
					/* not add the lenght of the in e out link 
					last_path->length = nn_path->length +
					un_link->link_elem->link_len +
					nu_link->link_elem->link_len; */
					last_path->length = nn_path->length;
					/* Too long paths are not allowed ! */
					if(last_path->length > MAX_PLENGTH)
					{
						printf("Error:  The network contains too long paths.\n");
						exit(1);
					}
					
					/* not add the user->net and net->user link to the no of hop */
					/* last_path->no_of_hops = nn_path->no_of_hops + 2; */
					last_path->no_of_hops = nn_path->no_of_hops;
					
					last_path->no_of_opaque_crs = nn_path->no_of_opaque_crs;
					
					last_path->num_of_prot_chs = nn_path->num_of_prot_chs;
					last_path->prot_channels = nn_path->prot_channels;
					
					for(i = 0; i<no_of_nnlinks; i++)
						last_path->protected_links[i] = nn_path->protected_links[i];
					
					for(i = 0; i<no_of_nodes; i++)
						last_path->protected_nodes[i] = nn_path->protected_nodes[i];
				} /* for */
				
			} /* if */
		} /* for */
		
	/* Free the temporary matrix */
	free_bidim_path_mat(nn_routes, no_of_nodes);

	return cc_routes;
}


/*
**  Function  : Boolean equ_double(double a, double b, double ths)
**  Return    : TRUE if fabs((a - b) / b) < ths, FALSE otherwise.
**  Remarks   : none.
*/
Boolean equ_double(double a, double b, double ths)
{
  if (b == 0.0)
    return fabs(a) < ths ? TRUE : FALSE;
  else
    return (fabs(a - b) < (ths * b)) ? TRUE : FALSE;
}


/* Store the data pertaining the current path (number of hops, length
   number of crossed opaque nodes and link list) in a new structure. 
   The link list is obtained from the elements of the stack. */
static Path* create_new_path(double length, int no_of_hops, 
							 int no_of_opcros)
{
  Path* new_path;
  struct stack_elem* stack_ptr;
  LinkList* last;
  int i;

  /* Check if the stack is empty. */
  if (stack_top < stack_bottom)
    {
      fprintf(stderr, "Error:  Path detection error in ANCLES function ");
      fprintf(stderr, "'create_new_path'\n");
      exit(1);
    }

  new_path = path_alloc();
  new_path->length = length;
  new_path->no_of_hops = no_of_hops;
  new_path->no_of_opaque_crs = no_of_opcros;

  new_path->num_of_prot_chs = 0;
  new_path->prot_channels = NULL;
  for(i = 0; i<no_of_nnlinks; i++)
	  new_path->protected_links[i] = 0;
  for(i = 0; i<no_of_nodes; i++)
	  new_path->protected_nodes[i] = 0;

  /* Build the link list from the elements of the stack. */
  stack_ptr = stack_bottom;
  last = new_path->path = linklist_alloc();
  last->link_elem = node_vect[stack_ptr->node_id]->outqueue[stack_ptr->port_number];
  last->next = NULL;
/* if duplex optic is used, its direction has to be shown */
  last->direction = stack_ptr->direction;
  last->endopsec = stack_ptr->endopsec;

  ++stack_ptr;
  for ( ; stack_ptr < stack_top; ++stack_ptr)
    {
      last->next = linklist_alloc();
      last = last->next;
	/* if duplex optic is used, its direction has to be shown */
      last->direction = stack_ptr->direction;
      last->endopsec = stack_ptr->endopsec;
	  last->link_elem = node_vect[stack_ptr->node_id]->outqueue[stack_ptr->port_number];
      last->next = NULL;
    }

  return new_path;
}


/*
**  Function  : static Boolean there_is_loop(int     node)
**  Return    : TRUE if a loop involving 'node' is detected in the current
**              path, FALSE otherwise.
**  Remarks   : the check for the presence of loops is based on the
**              comparison of the incoming node identifier with the
**              identifiers already in the stack.
*/
static Boolean there_is_loop(int     node)
{
  struct stack_elem* ptr;

  for (ptr = stack_bottom; ptr <= stack_top; ++ptr)
    if (ptr->node_id == node)
      return TRUE;

  return FALSE;
}

/*
**  Function  : int compose_ip_paths( int, int, Path*, Link*, Path*)
**  Return    : number of created paths (included node-node)
**  Remarks   : Compose node-node and higher level cgen-cgen
**              paths and insert them to the right place in the list
*/
int compose_ip_paths( int s, int d, Path* path1,
					 Link* new_link, Path* path2)
{
	int	i, hopnum = 1;
	double length = new_link->link_len;
	int su, du, no_of_nnpaths = 0;
	int uupath_accepted, no_of_uupaths = 0;
	Path	*new_path, *last_path, *act_path;
	LinkList *link, *nlink, *un_link, *nu_link;

	if(path1)
	{
	/* length and number of hops */
		hopnum += path1->no_of_hops;
		length += path1->length;
	}
	if(path2)
	{
	/* length and number of hops */
		hopnum += path2->no_of_hops;
		length += path2->length;
	}

	/* Check and return if the path would be too long */
#ifdef XXX_MAX_NO_OF_HOPS
	if (hopnum > XXX_MAX_NO_OF_HOPS)
		return 0;
#endif
#ifdef XXX_MAX_LENGTH
	if (length > XXX_MAX_LENGTH)
		return 0;
#endif
    if(hopnum > xxx_max_no_of_hops || length > MAX_PLENGTH)
		return 0;

	/* using here a check of nnpaths number based on hopnum ordered list */
	for(act_path = virt_nn_mat[s][d]; act_path; act_path = act_path->next)
	{
		if(act_path->no_of_hops <= hopnum)
			no_of_nnpaths++;
	}
    if(no_of_nnpaths >= maxvirtnnpaths)
		return 0;

	nlink = linklist_alloc();
	nlink->link_elem = new_link;
	nlink->direction = 0;
	nlink->endopsec = FALSE;
	nlink->next = NULL;

	/* allocate path */
	new_path = path_alloc();

	/* composition */
	if(path1)
	{
		new_path->path = clone_linklist(path1->path);
		/* attach new link after the clone of path1 */
		for(link = new_path->path; link->next; link = link->next);
		link->next = nlink;
	}
	else
		new_path->path = nlink;

	if(path2)
		nlink->next = clone_linklist(path2->path);

	/* set path attributes */
	new_path->no_of_hops = hopnum;
	new_path->length = length;
	new_path->no_of_opaque_crs = 0;
	new_path->num_of_prot_chs = 0;
	new_path->prot_channels = NULL;
	for(i = 0; i<no_of_nnlinks; i++)
		new_path->protected_links[i] = 0;
	for(i = 0; i<no_of_nodes; i++)
		new_path->protected_nodes[i] = 0;

	/* insert it at the beginning of the list of node-node paths*/
	new_path->next = virt_nn_mat[s][d];
	virt_nn_mat[s][d] = new_path;

	no_of_nnpaths = 1;

	/* create all cgen-cgen paths that uses this node-node path */
	for(su = 0; su < node_vect[s]->no_of_users; su++)
	{
		for(du = 0; du < node_vect[d]->no_of_users; du++)
		{
			if(user_vect[node_vect[s]->connected_user[su]]->
				generator->type < CG_TYPE6 &&
				user_vect[node_vect[d]->connected_user[du]]->
				generator->type > CG_TYPE9 &&
				traff_pattern[node_vect[s]->connected_user[su]]
				[node_vect[d]->connected_user[du]] > 0.0)
			{
				/* user-user path */
				new_path = path_alloc();
				uupath_accepted = 0;

				/* put the new path to its right place in the ordered list 
				according to the ordering princips */
				if(user_vect[node_vect[s]->connected_user[su]]->generator->
					choosen_routing.path_order_type == IN_ORDER_OF_HOP_NUMBER ||
					user_vect[node_vect[s]->connected_user[su]]->generator->
					choosen_routing.path_order_type == 
					BALANCED_IN_ORDER_OF_HOP_NUMBER)
				{
					/* hop based */
					last_path = path_mat[IN_ORDER_OF_HOP_NUMBER]
						[user_vect[node_vect[s]->connected_user[su]]
						->generator->id]
						[user_vect[node_vect[d]->connected_user[du]]
						->generator->id];
					if(!last_path || last_path->no_of_hops >= hopnum)
						/* insert into the empty list or
						as first element of the list*/
					{
						new_path->next = last_path;
						path_mat[IN_ORDER_OF_HOP_NUMBER]
							[user_vect[node_vect[s]->connected_user[su]]
							->generator->id]
							[user_vect[node_vect[d]->connected_user[du]]
							->generator->id] = new_path;
						uupath_accepted++;
					/* the first routing table elemenet was changed */
						rt_sp_changed[node_vect[s]->connected_user[su]]							
							[node_vect[d]->connected_user[du]]++;
					}
					else
					{
						/* find the first path which is not shorter */
						for(act_path = last_path->next, i = 0;
						act_path && act_path->no_of_hops < hopnum && i < maxvirtpaths; 
						act_path = act_path->next, i++)
							last_path = last_path->next;

						/* allow to insert path only before a given order 
							number (max_no_of_virt_routes) of paths */
						if(i<maxvirtpaths)
						{
							last_path->next = new_path;
							new_path->next = act_path;
							uupath_accepted++;
						}
					}
				}
				else
				{
					/* length based */
					last_path = path_mat[IN_ORDER_OF_LENGTH]
						[user_vect[node_vect[s]->connected_user[su]]
						->generator->id]
						[user_vect[node_vect[d]->connected_user[du]]
						->generator->id];
					if(!last_path || last_path->length >= length)
					{
						/* insert into the empty list or
						as first element of the list*/
						path_mat[IN_ORDER_OF_LENGTH]
							[user_vect[node_vect[s]->connected_user[su]]
							->generator->id]
							[user_vect[node_vect[d]->connected_user[du]]
							->generator->id] = new_path;
						uupath_accepted++;
					}
					else
					{
						/* find the first path which is not shorter */
						for(act_path = last_path->next, i = 0;
						act_path && act_path->length < length && i<maxvirtpaths; 
						act_path = act_path->next, i++)
							last_path = last_path->next;

						/* allow to insert path only before a given order 
							number (max_no_of_virt_routes) of paths */
						if(i<maxvirtpaths)
						{
							last_path->next = new_path;
							new_path->next = act_path;
							uupath_accepted++;
						}
					}
				}
				/* really create the path only if it was accepted */
				if(uupath_accepted)
				{
				/* user-node */
					un_link = linklist_alloc();
					un_link->link_elem = user_vect[node_vect[s]->
						connected_user[su]]->out_link;
					un_link->direction = 0;
					un_link->endopsec = FALSE;
					/* node-node */
					un_link->next = clone_linklist(virt_nn_mat[s][d]->path);
					for (nu_link = un_link; nu_link->next; nu_link = nu_link->next);
					/* node-user */
					nu_link->next = linklist_alloc();
					nu_link = nu_link->next;
					nu_link->link_elem = user_vect[node_vect[d]->
						connected_user[du]]->in_link;
					nu_link->direction = 0;
					nu_link->endopsec = FALSE;
					nu_link->next = NULL;
					
					new_path->path = un_link;
					new_path->length = length;
					new_path->no_of_hops = hopnum;
					
					new_path->no_of_opaque_crs = 0;
					new_path->num_of_prot_chs = 0;
					new_path->prot_channels = NULL;
					for(i = 0; i<no_of_nnlinks; i++)
						new_path->protected_links[i] = 0;
					for(i = 0; i<no_of_nodes; i++)
						new_path->protected_nodes[i] = 0;
					no_of_uupaths++;
				}
				else
					path_release(new_path);
			}
		}
	}
	return no_of_nnpaths + no_of_uupaths;
}
