/*
**  Module: iplinks.c
**  Version:
**  Date: 
**  Authors   : Zoltan Zsoka
**  Synopsis  : this module contains the functions dealing with the
**              management of the virtual links introduced in the 
**				version GMPLS of ancles. A virtual link seems to be
**				a real link from the IP level point of view and a type8
**				from the phisical (optical) layer point of view
**  Functions : void open_new_virt_link
**				Link* create_virtual_link
**				void close_virt_link
**				void destroy_virt_link
**  History   : 01-AUG-2003 - Creation of the module.           -ZZ-
*/

#include "ancles.h"

static Link* create_virtual_link(Call *, int);
static void destroy_virtual_link(Link *);

void open_new_virt_link(int from, int to)
{
	int u, vl_from_cg = -1;
	int vl_to_cg = -1;
	Call* call_ptr;
/*	char	output_file[80];*/

	/* If there is no more place in the linkvector */
	if(no_of_links + no_virt_links + 1 == MAX_LINK)
	{
		printf("More than %d virtual links can not be opened!\n",
			MAX_LINK - no_of_links);
		exit(1);
	}

	/* enable multiplied open of logical links ???
    open_loglink = 0; */

    switch(vl_open_algorithm)
    {
	/* open a new, direct link between the endpoints */
	case DIRECT_VLINK :
	case OCH_FIRST :
	case EHC_GROOM :
		/* find the generators for the type8 call */
	      for(u = 0; u < node_vect[from]->no_of_users; u++)
		  {
			 
			  if(user_vect[node_vect[from]->connected_user[u]]->
				  generator->type == CG_TYPE8)
				  vl_from_cg = user_vect[node_vect[from]->		
					connected_user[u]]->generator->id;
		  }
	      for(u = 0; u < node_vect[to]->no_of_users; u++)
		  {
			  if(user_vect[node_vect[to]->connected_user[u]]->
				  generator->type == CG_TYPE12)
				  vl_to_cg = user_vect[node_vect[to]->		
					connected_user[u]]->generator->id;
		  }
	    break;
    }

	/* no convenient CG8 generator is found */
    if(vl_from_cg<0 || vl_to_cg <0)
		return;
	/* virtual links statistics */
	/* we don't use this detailed statistics 
	strcpy(output_file, file_radix);
	strcat(output_file, ".cg8");
	if((CG8FILE = fopen(output_file, "a")) == NULL)
	{
		printf("\nError:  the file %s cannot be opened.\n", output_file);
		exit(1);
	}*/

	if(call_ptr = exe_routing(vl_from_cg, vl_to_cg,
		wavelength_capacity, wavelength_capacity, wavelength_capacity))
	{
		/* crate forward (and backward) virtual link */
		((CallType8 *)(call_ptr->info))->fwd_link =
			create_virtual_link(call_ptr, 0);
		if(duplex_llc)
		{
			((CallType8 *)(call_ptr->info))->bwd_link =
				create_virtual_link(call_ptr, 1);
		}
		else
			((CallType8 *)(call_ptr->info))->bwd_link =
			NULL;

	/* we don't use this detailed statistics 
		fprintf(CG8FILE, "O\t%s\t%s\t%.6f\t%d\n",
			user_name(callgen_vect[vl_from_cg]->ref_user->id),
			user_name(callgen_vect[vl_to_cg]->ref_user->id),
			ancles_time, call_ptr->curr_path_no);*/
	}
	/* we don't use this detailed statistics 
	else
		fprintf(CG8FILE, "R\t%s\t%s\t%.6f\n",
		user_name(callgen_vect[vl_from_cg]->ref_user->id),
		user_name(callgen_vect[vl_to_cg]->ref_user->id),
		ancles_time);

	fclose(CG8FILE);*/
}

Link* create_virtual_link(Call* my_call, int backward)
{
    Link* new_link;
    Link_ucall_stat* new_lus;
	LinkList* link;
	int j, s, d, path1nodes[MAX_NODE];
	int startnode, endnode;
	Boolean new_alloc = FALSE;
	Path* path1;
	Path* path2;
	Event* next_updating_event;
	Boolean touched;

	/* the direction of the link is set according to the input */
	if(backward)
	{
		startnode = callgen_vect[my_call->dest_id]->ref_user->connected_node;
		endnode = callgen_vect[my_call->source_id]->ref_user->connected_node;
	}
	else
	{
		startnode = callgen_vect[my_call->source_id]->ref_user->connected_node;
		endnode = callgen_vect[my_call->dest_id]->ref_user->connected_node;
	}
	
    /* put the first invalid element to valid or allocate a new one
	in the vector of link pointers and Link_ucall_stat pointers */
    new_link = link_vect[no_of_links + no_virt_links];
    new_lus = link_Ucall_stat[no_of_links + no_virt_links];
    if(!new_link)
	{
		new_link = (Link *)MMmalloc(sizeof(Link), "create_virtual_link");
		link_vect[no_of_links + no_virt_links] = new_link;
		new_lus = (Link_ucall_stat *)MMmalloc(sizeof(Link_ucall_stat),
	    "create_virtual_link");
		link_Ucall_stat[no_of_links + no_virt_links] = new_lus;
		new_alloc = TRUE;
	}
    new_link -> id = no_of_links + no_virt_links;
    no_virt_links++;

    /* set attributes */
	new_link->source = startnode;
	new_link->dest = endnode;
	new_link->timescale = capacity/wavelength_capacity;
	new_link->free_band = wavelength_capacity;
	new_link->bw_update_p = virtlink_bup;

	new_link->linked_call = my_call;

    new_link->routing_threshold[0] = 0.0;
    new_link->routing_threshold[1] = SIBAL_THRESHOLD;
	

	new_link->duplex_optic = FALSE;
	new_link->substituted = FALSE;
	new_link->max_waves = 0;
	new_link->used_waves = 0;
	new_link->peak_wave = 0;
    for(j = 0; j < MAX_WVOCTETS; j++)
        new_link->free_waves[j] = 0x00; /* not aviable waves */
	new_link->free_fibers = NULL;
	new_link->fiber_num = 0;
	new_link->is_node_node = TRUE;

	new_link->nnid = no_of_nnlinks;
	nnlink_vect[no_of_nnlinks++] = new_link->id;

	new_link->link_len = 0.0;
	for(link = my_call->path_ptr->path; link; link = link->next)
		new_link->link_len += link->link_elem->link_len;
	 
	new_lus->start_time = ancles_time;
	new_lus->last_updt_time = ancles_time;
	new_lus->free_Uband = wavelength_capacity;
	new_lus->official_free_Uband = wavelength_capacity;

	new_lus->fb_integ = 0.0;
	new_lus->last_unus_time = ancles_time;
	new_lus->unused_free_band = wavelength_capacity;
	new_lus->un_fb_integ = 0.0;
	new_lus->official_cong = FALSE;
	new_lus->no_of_Ucall = 0;
	new_lus->no_of_Ucall_integ = 0.0;
	new_lus->official_no_of_Ucall = 0.0;
	new_lus->estim_no_of_Ucall = 0.0;
	new_lus->is_node_node = TRUE;
	new_lus->acc_PCR_1_2 = 0.0;
	new_lus->acc_PCR_U = 0.0;
	new_lus->U_PCR_int = 0.0;
	new_lus->last_Uupdt_time = ancles_time;
	new_lus->sp_on = NULL;

	/* if a brand new link was even allocated */
	if(new_alloc)
	{
		init_call_list_on_new_link(new_link->id);
		init_Ucall_list_on_new_link(new_link->id);
	}

    /* insert official bw update event */
	if(virtlink_bup > 0)
	{
	/* Creating, initializing and scheduling a new updating event	*/
		next_updating_event = event_alloc();
		next_updating_event->ev_type = OFFICIAL_FBW_UPDATE;
		next_updating_event->ev_dscrpt.link_id = new_link->id;
		next_updating_event->ev_time =
			ancles_time + uniform_0_1(&sh_seed)*new_link->bw_update_p;
		/* Put the new updating event in the event-queue	*/
		insert_event(next_updating_event);
	}

    /* refresh path list among nodes and higher level endpoints */
	for(s = 0; s < no_of_nodes; s++)
	{
		if(s == startnode)
		{
			for(d=0; d<no_of_nodes;d++)
			{
				if(d == endnode && s!=d)
			/* Compose the node-node and all the ip level paths from three parts */
			/* Consists of only the new link */
					compose_ip_paths( s, d, NULL, new_link, NULL);
				else
				{
					/* Second path */
					for(path2 = virt_nn_mat[endnode][d]; path2; path2 = path2->next)
					{
						touched = FALSE;
						for(link = path2->path; !touched && link; link = link->next)
						{
							if(link->link_elem->source == s ||
								link->link_elem->dest == s)
								touched = TRUE;
						}
						/* Take only the independent paths */
						if(!touched)
			/* Compose the node-node and all the ip level paths from three parts */
							compose_ip_paths( s, d, NULL, new_link, path2);
					}
				}
			}
		}
		else
		{
			for(path1 = virt_nn_mat[s][startnode]; path1; 
			path1 = path1->next)
			{
				/* Inicialize the list of nodes touched by the path1 */
				for(j=0; j<no_of_nodes; j++)
					path1nodes[j] = 0;
				/* Registry of touched nodes */
				for(link = path1->path; link; link = link->next)
				{
					path1nodes[link->link_elem->source]++;
					path1nodes[link->link_elem->dest]++;
				}
				for(d=0; d<no_of_nodes;d++)
				{
					if(d == endnode && !path1nodes[d])
			/* Compose the node-node and all the ip level paths from three parts */
						compose_ip_paths( s, d, path1, new_link, NULL);
					else
					{
						/* Second path */
						for(path2 = virt_nn_mat[endnode][d]; path2; path2 = path2->next)
						{
							touched = FALSE;
							for(link = path2->path; !touched && link; link = link->next)
							{
								if(path1nodes[link->link_elem->source] ||
									path1nodes[link->link_elem->dest])
									touched = TRUE;
							}
							/* Take only the independent paths */
							if(!touched)
			/* Compose the node-node and all the ip level paths from three parts */
								compose_ip_paths( s, d, path1, new_link, path2);
						}
					}
				}
			}
		}
	}
	return new_link;
}

void close_virt_link(Link* del_link)
{
	int no_of_calls = 0;
	CallList *curr_clist;

	/* we need to remove the open calls from the link before the close */
	/* not Ucalls */
    for(curr_clist = del_link->calls_on->next; curr_clist != 
		del_link->calls_on; curr_clist = curr_clist->next)
	{
	/* the current version supports only closing empty links */
		no_of_calls++;
	}
	/* Ucalls */
    for(curr_clist = link_Ucall_stat[del_link->id]->U_calls_on->next; curr_clist != 
		link_Ucall_stat[del_link->id]->U_calls_on; curr_clist = curr_clist->next)
	{
	/* the current version supports only closing empty links */
		no_of_calls++;
	}

	destroy_virtual_link(del_link);
}


void destroy_virtual_link(Link* del_link)
{
    int s, d, found;
    LinkList *curr_link;
    Path *path, *prev_path;
    Link_ucall_stat* del_lus;
    Event* curr_event;

    int linkid = del_link -> id;
    
	if(virtlink_bup > 0)
	{
	/* the next official bandwidth update event need to 
	be moved from the event list */
		for(curr_event=head_of_event_list->next, found=0;
		!found; curr_event = curr_event->next)
			if(curr_event->ev_type == OFFICIAL_FBW_UPDATE &&
				curr_event->ev_dscrpt.link_id == linkid)
			{
				found++;
				dequeue_far_event(curr_event);
			}
	}
    
    /* release the link and the Ucall_stat 
	(actually we only need to bring it to the first invalid place */
    no_virt_links--;

	if(linkid < no_of_links + no_virt_links)
	{
		link_vect[linkid] = link_vect[no_of_links + no_virt_links];
		link_vect[no_of_links + no_virt_links] = del_link;
		/* exchange the vectors dealing with the calls on the link */
		exchange_call_list_on_links(linkid, no_of_links + no_virt_links);

		link_vect[linkid]->id = linkid;
		link_vect[no_of_links + no_virt_links]->id = 
			no_of_links + no_virt_links;
		
		del_lus = link_Ucall_stat[linkid];
		link_Ucall_stat[linkid] = link_Ucall_stat[no_of_links + no_virt_links];
		link_Ucall_stat[no_of_links + no_virt_links] = del_lus;
		/* change the vectors dealing with the Ucalls on the link */
		exchange_Ucall_list_on_links(linkid, no_of_links + no_virt_links);
	}
	/* update the node-node id information */
	link_vect[linkid]->nnid = del_link->nnid;
	no_of_nnlinks--;
    
    /* delete from the matrix all the paths that 
	contain the releasable link, */
    for(s = 0; s < no_of_cgens; s++)
    {
        if(callgen_vect[s]->type <CG_TYPE6)
        {
			for(d = 0; d < no_of_cgens; d++)
			{
				if(callgen_vect[d]->type == CG_TYPE10 ||
					callgen_vect[d]->type == CG_TYPE11)
				{
					prev_path = NULL;
					path = path_mat[callgen_vect[s]->choosen_routing.
						path_order_type][s][d];
					while(path)
					{
						for(curr_link = path->path; curr_link && 
							curr_link->link_elem!=del_link;
						curr_link = curr_link->next);
						if(curr_link)
						{
							/* if the path contains the link */
							if(prev_path)
								/* not the first path in the list */
								prev_path->next = path->next;
							else
							{
								path_mat[callgen_vect[s]->choosen_routing.
									path_order_type][s][d] = path->next;
								rt_sp_changed[callgen_vect[s]->ref_user->id]
									[callgen_vect[d]->ref_user->id]++;
							}
							path = free_linkpath(path);
						}
						else
						{
							prev_path = path;
							path = path->next;
						}
					}
				}
			}
		}
	}
    /* delete from the node-node matrix too */
    for(s = 0; s < no_of_nodes; s++)
    {
        for(d = 0; d < no_of_nodes; d++)
		{
			prev_path = NULL;
			path = virt_nn_mat[s][d];
			while(path)
			{
				for(curr_link = path->path; curr_link && 
					curr_link->link_elem!=del_link; curr_link = curr_link->next);
				if(curr_link)
				{
					/* if the path contains the link */
					if(prev_path)
						/* not the first path in the list */
						prev_path->next = path->next;
					else
						virt_nn_mat[s][d] = path->next;
					
					path = free_linkpath(path);
				}
				else
				{
					prev_path = path;
					path = path->next;
				}
			}
		}
    }
}
