/*
**  Module    : lists.c
**  Version   : $Revision: 3.2 $
**  Date      : $Data$
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains the functions dealing with the
**              management of the Event List, the general Call List and the
**              Lists collecting the calls supported by the network links.
**              The management criterion is the same for all the lists that
**              are referred to by the functions of this module: they are
**              double linked circular lists, with an empty record acting
**              as head-end, and the records inside are ordered according
**              to the value of a temporal key (the realization time for an
**              event, the closure time for a call).
**              Each list has a companion array of pointers, whose items
**              address the entry points of the list: each time a new
**              record is getting the list, its temporal key is compared
**              with the values associated to the entry points (flags),
**              so that the destination sublist is singled out at once.
**              The insertion in the sublist is based on its linear
**              scanning. Each time two pointers refer to the same record
**              in the list the flag assignment is repeated, in
**              order to keep the size of the sublists as uniform as
**              possible.
**		There are some new and modified functions for the support
**		of the data amount (and not duration) based traffic of the
**		U-class type calls.
**
**  Functions : void insert_event(Event *)
**              void dequeue_event(Event *)
**              void dequeue_far_event(Event *)
**              void unrel_dequeue_event(Event *, double)
**              void insert_call(Call *)
**              void dequeue_call(Call *)
**              void unrel_dequeue_call(Call *, double)
**              void insert_call_list_on_link(CallList *, Link *)
**              void dequeue_call_list_from_link(CallList *, Link*)
**              void unrel_dequeue_call_list_from_link(CallList *,
**              void insert_Ucall_list_on_link(CallList *, Link *)
**              void dequeue_Ucall_list_from_link(CallList *, Link *)
**              void unrel_dequeue_Ucall_list_from_link(CallList *,
**			Link *, double)
**              void init_event_list(void)
**              void init_call_list(void)
**              void init_call_list_on_link(void)
**              void init_call_list_on_new_link(void)
**				void exchange_call_list_on_links(int , int)
**              void init_Ucall_list_on_link(void)
**              void init_Ucall_list_on_new_link(void)
**				void exchange_Ucall_list_on_links(int , int)
**              CallList *search_Ucall_list_on_link(Link *, Call *,
**			double)
**              CallList *search_call_list_on_link(Link *, Call *,
**			double)
**				void reinsert_call_and_event(Call* , double)
**
**  History   : 22-OCT-1995 - Creation of the module.           -GP-
**              01-JUN-1996 - Modification of the algorithm
**                            controlling the linked lists.     -MM-
**              01-MAR-1997 - Revision				-MM-
**		31-AUG-1999 - Revision	: adding dequeue functions
**				without releasing the items	-ZZ-

30-01-2001 Cleaning -ZZ-
*/

# ifndef        ANCLES_H
#include        "ancles.h"
# endif

/***************************
**  Constant definitions  **
***************************/

# ifndef START_FLAG             /* Initial value of the number of flags */
#define  START_FLAG     4       /* used to speed up the insertion of new*/
# endif                         /* records in the lists; it must be 2^n */

# ifndef MAX_STEP               /* Maximum number of elements in a      */
#define  MAX_STEP       50      /* single sublist.                      */
# endif


/***********************
**  Global Variables  **
***********************/

Event 	**event_flags;          /* Array of pointers to the access      */
                                /* points in the event list.            */
int num_event_flag;		/* Number of flags currently used in    */
                                /* the event list.                      */

Call    **call_flags;           /* Array of pointers to the access      */
                                /* points in the call list.             */
int num_call_flag;          	/* Number of flags currently used in    */
                                /* the call list.                       */
long    call_count;             /* Counter for the number of calls in   */
                                /* the call list.                       */

CallList ***link_flags;         /* Array of pointers to the access      */
                                /* points to the call lists associated  */
                                /* to the links.                        */
int *num_link_flag;         	/* Array containing the number of flags */
                                /* currently used for each link call    */
                                /* list.                                */
long    *link_call_count;       /* Array of counters for the number of  */
                                /* calls supported by each link.        */

CallList ***Ulink_flags;        /* Array of pointers to the access      */
                                /* points to the Class U call lists 	*/
				/* associated to the links.		*/
int *Unum_link_flag;        	/* Array containing the number of flags */
                                /* currently used for each Class U call */
                                /* list.                                */
long    *Ulink_call_count;      /* Array of counters for the number of  */
                                /* Class U calls supported by each link.*/


/**************************
**  Function Prototypes  **
**************************/


/***************************
**  Function Definitions  **
***************************/


/*
**  Function  : void insert_event(Event *event)
**  Return    : none
**  Remarks   : 'event' points to the event that must be inserted into the
**              Event List. The Event List is a double linked circular list
**              with an empty head. 'head_of_event_list' is the pointer
**              to the empty head that is initialized with an 'ev_time' field
**              equal to the minimum value of a 'double'.
**              The event insertion exploits the presence of flags
**              addressing partial sublists inside the Event List in order
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void insert_event(Event *event)
{
 Event          *cur;
 TimeUnit       sched_time;
 long           half_of_list = sched_events_cnt >> 1;
 int		flag_number = 0;

 ++sched_events_cnt;
 sched_events_acc += sched_events_cnt;

/* Provisional solution against overflow on sched_events_acc.   */
 if(sched_events_acc <= 0)
        {
         sched_events_acc -= sched_events_cnt;
         mean_sum_of_FES_events +=
                              (double)sched_events_acc / no_of_FES_operations;
         no_of_sum++;

         no_of_FES_operations = 1;
         sched_events_acc = sched_events_cnt;
        }

 ++no_of_FES_operations;

/* The test for insertion starts from the event with the highest        */
/* 'ev_time' field.                                                     */
 sched_time = event->ev_time;
 cur = head_of_event_list->pred;

/* The first step towards insertion consists in the dichotomous search  */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming event is lower than the one of   */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the event is placed at the end of the Event List.          */
 if(cur->ev_time > sched_time)
   {
    flag_number = num_event_flag >> 1;
    half_of_list = num_event_flag >> 2;

    while(half_of_list)
      {
       if(event_flags[flag_number]->ev_time > sched_time)
          flag_number += half_of_list;
       else
         flag_number -=half_of_list;

       half_of_list >>= 1;
      }
    cur = event_flags[flag_number];

/* The linear search through the sublist is started here.               */
    while(cur->ev_time < sched_time)
       cur = cur->next;

    while(cur->ev_time > sched_time)
       cur = cur->pred;
   }

/* Insertion of the Event record.                                       */
 event->pred = cur;
 event->next = cur->next;
 event->next->pred = event;
 cur->next = event;
}


/*
**  Function  : void dequeue_event(Event *event)
**  Return    : none
**  Remarks   : this function removes the realized event from the circular
**              list it belongs to by connecting its predecessor with its
**              successor.
**              The array of companion flags is updated every time the first
**              sublist gets empty.
*/
void dequeue_event(Event *event)
{
 Event          *cur;
 register int   i, j;
 long           step;

 --sched_events_cnt;

 event->pred->next = event->next;
 event->next->pred = event->pred;

/* Each time an event is drawn from the Event List the state of the     */
/* pointers used as companion flags is checked: if at least two flags   */
/* are pointing to the outgoing event the set of flags needs to be      */
/* rearranged over the Event List.                                      */
 if((event_flags[num_event_flag - 1] == event) ||
    (event_flags[0] == event_flags[1]))
   {
    step = sched_events_cnt / (num_event_flag + 1);

    if(step > MAX_STEP)
      {
       free(event_flags);
       num_event_flag <<= 1;
       event_flags = (Event **)MMmalloc((sizeof(Event *)) * num_event_flag,
                                            "dequeue_event");
       step = sched_events_cnt / (num_event_flag + 1);
      }
    cur = head_of_event_list;
    for(j = 0; j < num_event_flag; j++)
      {
       for(i = 0; i < step; i++)
            cur = cur->pred;
       event_flags[j] = cur;
      }
   }
 event_release(event);
}

/*
**  Function  : void unrel_dequeue_event(Event *event, double sched_time)
**  Return    : none
**  Remarks   : the same functionality as the normal dequeue, but doesn't
**		release the event. sched_time is used for the searching in
**		array of companion flags
*/
void unrel_dequeue_event(Event *event, double sched_time)
{
 Event          *cur;
 register int   i, j;
 long           step;

 int        	flag_number;
 int        	half_of_list;
 int		found;

  --sched_events_cnt;

 event->pred->next = event->next;
 event->next->pred = event->pred;

/* Each time an event is drawn from the Event List but it isn't sure,	*/
/* that he was the first in the time order, the state of the        	*/
/* pointers used as companion flags is checked: if there is any flag	*/
/* pointing to the outgoing call the set of flags needs to be		*/
/* rearranged over the Event List.                                      */

/* Searching for the element in the flag list 				*/
 if(event_flags[0] == event)
  found = 1;
 else
  {
   found = 0;
   flag_number = num_event_flag >> 1;
   half_of_list = num_event_flag >> 2;

   while(half_of_list)
    {
     if(event_flags[flag_number] == event)
      half_of_list = 0;
     else
      {
       if(event_flags[flag_number]->ev_time > sched_time)
         flag_number += half_of_list;
       else
         flag_number -= half_of_list;
       half_of_list >>= 1;
      }
    }

   if(event_flags[flag_number] == event)
    found = 1;
  }


 if( found ||
    (event_flags[0] == event_flags[1]))
   {
    step = sched_events_cnt / (num_event_flag + 1);

    if(step > MAX_STEP)
      {
       free(event_flags);
       num_event_flag <<= 1;
       event_flags = (Event **)MMmalloc((sizeof(Event *)) * num_event_flag,
                                            "dequeue_event");
       step = sched_events_cnt / (num_event_flag + 1);
      }
    cur = head_of_event_list;
    for(j = 0; j < num_event_flag; j++)
      {
       for(i = 0; i < step; i++)
            cur = cur->pred;
       event_flags[j] = cur;
      }
   }
}

/*
**  Function  : void dequeue_far_event(Event *event)
**  Return    : none
**  Remarks   : the same functionality as the normal dequeue, but updates
**		the companion flags if it's necessary even if the event is not
**		surely the first in the event list.
*/
void dequeue_far_event(Event *event)
{
 Event          *cur;
 register int   i, j;
 long           step;

 int        	flag_number;
 int        	half_of_list;
 int		found;
 double		sched_time;

 sched_time = event->ev_time;

  --sched_events_cnt;

 event->pred->next = event->next;
 event->next->pred = event->pred;

/* Each time an event is drawn from the Event List but it isn't sure,	*/
/* that he was the first in the time order, the state of the        	*/
/* pointers used as companion flags is checked: if there is a  flag	*/
/* pointing to the outgoing call the set of flags needs to be		*/
/* rearranged over the Event List.                                      */

/* Searching for the element in the flag list 				*/
 if(event_flags[0] == event)
  found = 1;
 else
  {
   found = 0;
   flag_number = num_event_flag >> 1;
   half_of_list = num_event_flag >> 2;

   while(half_of_list)
    {
     if(event_flags[flag_number] == event)
      half_of_list = 0;
     else
      {
       if(event_flags[flag_number]->ev_time > sched_time)
         flag_number += half_of_list;
       else
         flag_number -= half_of_list;
       half_of_list >>= 1;
      }
    }

   if(event_flags[flag_number] == event)
    found = 1;
  }


 if( found ||
    (event_flags[0] == event_flags[1]))
   {
    step = sched_events_cnt / (num_event_flag + 1);

    if(step > MAX_STEP)
      {
       free(event_flags);
       num_event_flag <<= 1;
       event_flags = (Event **)MMmalloc((sizeof(Event *)) * num_event_flag,
                                            "dequeue_event");
       step = sched_events_cnt / (num_event_flag + 1);
      }
    cur = head_of_event_list;
    for(j = 0; j < num_event_flag; j++)
      {
       for(i = 0; i < step; i++)
            cur = cur->pred;
       event_flags[j] = cur;
      }
   }
 event_release(event);
}


/*
**  Function  : void insert_call(Call *call)
**  Return    : none
**  Remarks   : 'call' points to the call that must be inserted into the
**              general Call List. The Call List is a double linked circular
**              list with an empty head. 'head_of_call_list' is the pointer
**              to the empty head that is initialized with an 'end_time' field
**              equal to the minimum value of a 'double'.
**              The call insertion exploits the presence of flags
**              addressing partial sublists inside the Call List in order
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void insert_call(Call *call)
{
 Call           *cur;
 double         sched_time;
 int        flag_number;
 int        half_of_list;

 call_count++;

/* The test for insertion starts from the call with the highest         */
/* 'end_time' field.                                                    */
 sched_time = call->end_time;
 cur = head_of_call_list->pred;

/* The first step towards insertion consists in the dichotomous search  */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming call is lower than the one of    */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the call is placed at the end of the general Call List.    */
 if(cur->end_time > sched_time)
   {
    flag_number = num_call_flag >> 1;
    half_of_list = num_call_flag >> 2;

    while(half_of_list)
      {
       if(call_flags[flag_number]->end_time > sched_time)
          flag_number += half_of_list;
       else
          flag_number -=half_of_list;

       half_of_list >>= 1;
      }

    cur = call_flags[flag_number];

/* The linear search through the sublist is started here.               */
    while (cur->end_time < sched_time)
       cur = cur->next;
    while (cur->end_time > sched_time)
       cur = cur->pred;
   }

/* Insertion of the Call record.                                        */
 call->pred = cur;
 call->next = cur->next;
 call->next->pred = call;
 cur->next = call;
}


/*
**  Function  : void dequeue_call(Call *call)
**  Return    : none.
**  Remarks   : this function removes the released call from the circular
**              list it belongs to by connecting its predecessor with its
**              successor.
**              The array of companion flags is updated every time the first
**              sublist gets empty.
*/
void dequeue_call(Call *call)
{
 register int   i, j;
 Call           *cur;
 long           step;

 --call_count;

 call->pred->next = call->next;
 call->next->pred = call->pred;
 switch(call->call_type)
    {
     case CALL_TYPE1 :
      call_type1_release( (CallType1 *)(call->info) );
      break;

     case CALL_TYPE2 :
      call_type2_release( (CallType2 *)(call->info) );
      break;

     case CALL_TYPE3 :
      call_type3_release( (CallType3 *)(call->info) );
      break;

     case CALL_TYPE4 :
      call_type4_release( (CallType4 *)(call->info) );
      break;

     case CALL_TYPE5 :
      call_type5_release( (CallType5 *)(call->info) );
      break;

      case CALL_TYPE6 :
      call_type6_release( (CallType6 *)(call->info) );
      break;
      
	  case CALL_TYPE8 :
      call_type8_release( (CallType8 *)(call->info) );
      break;
 }

/* Each time a call is drawn from the Call List the state of the        */
/* pointers used as companion flags is checked: if at least two flags   */
/* are pointing to the outgoing call the set of flags needs to be       */
/* rearranged over the Call List.                                       */
 if((call_flags[num_call_flag - 1] == call) ||
    (call_flags[0] == call_flags[1]))
   {
    step = call_count / (num_call_flag + 1);
    if(step > MAX_STEP)
      {
       free(call_flags);
       num_call_flag <<= 1;

       call_flags = (Call **)MMmalloc((sizeof(Call *)) * num_call_flag,
                                      "dequeue call");
       step = call_count / (num_call_flag+1);
      }
    cur = head_of_call_list;
    for(j = 0; j < num_call_flag; j++)
      {
       for(i = 0; i < step; i++)
          cur = cur->pred;
       call_flags[j] = cur;
      }
   }
 call_release(call);
}


/*
**  Function  : void unrel_dequeue_call(Call *call, double sched_time)
**  Return    : none.
**  Remarks   : the same functionality as the normal dequeue, but doesn't
**		release the call. sched_time is used for the search in the
**		array of companion flags
*/
void unrel_dequeue_call(Call *call, double sched_time)
{
 register int   i, j;
 Call           *cur;
 long           step;

 int        	flag_number;
 int        	half_of_list;
 int		found;

 --call_count;

 call->pred->next = call->next;
 call->next->pred = call->pred;

/* Each time a call is drawn from the Call List, but it isn't sure,	*/
/* that he was the first in the time order, the state of the        	*/
/* pointers used as companion flags is checked: if there is a  flag	*/
/* pointing to the outgoing call the set of flags needs to be		*/
/* rearranged over the Call List.                                       */


/* Searching for the element in the flag list 				*/
 if(call_flags[0] == call)
  found = 1;
 else
  {
   found = 0;
   flag_number = num_call_flag >> 1;
   half_of_list = num_call_flag >> 2;

   while(half_of_list)
    {
     if(call_flags[flag_number] == call)
      half_of_list = 0;
     else
      {
       if(call_flags[flag_number]->end_time > sched_time)
         flag_number += half_of_list;
       else
         flag_number -= half_of_list;
       half_of_list >>= 1;
      }
   }

   if(call_flags[flag_number] == call)
    found = 1;
  }

 if( found ||
    (call_flags[0] == call_flags[1]))
   {
    step = call_count / (num_call_flag + 1);
    if(step > MAX_STEP)
      {
       free(call_flags);
       num_call_flag <<= 1;

       call_flags = (Call **)MMmalloc((sizeof(Call *)) * num_call_flag,
                                      "dequeue call");
       step = call_count / (num_call_flag+1);
      }
    cur = head_of_call_list;
    for(j = 0; j < num_call_flag; j++)
      {
       for(i = 0; i < step; i++)
          cur = cur->pred;
       call_flags[j] = cur;
      }
   }
}

/* Inserts the call and its closing event to the given time */
void reinsert_call_and_event(Call* call, double schedtime)
{
	double oldtime = call->end_time;
	LinkList* link_ptr;
	CallList*	call_list_ptr;

	/* Put the call to its new place */
	unrel_dequeue_call(call, oldtime);
	call->end_time = schedtime;
	insert_call(call);

	/* Put the call close event to its new place */
	unrel_dequeue_event(call->my_close_event, oldtime);
	call->my_close_event->ev_time = schedtime;
	insert_event(call->my_close_event);

	/* Put all the call_lists on the links to their new place */
	for(link_ptr = call->path_ptr->path; 
	link_ptr; link_ptr = link_ptr->next)
	{
		if(call->call_type != CALL_TYPE5)
		{
			/* find call_list on link */
			call_list_ptr = search_call_list_on_link(link_ptr->link_elem,
				call, oldtime);
			/* and put on the new place */
			unrel_dequeue_call_list_from_link(call_list_ptr, 
				link_ptr->link_elem, oldtime);
			insert_call_list_on_link(call_list_ptr, link_ptr->link_elem);
		}
		else
		{
			/* find Ucall_list on link */
			call_list_ptr = search_Ucall_list_on_link(link_ptr->link_elem,
				call, oldtime);
			/* and put on the new place */
			unrel_dequeue_Ucall_list_from_link(call_list_ptr,
				link_ptr->link_elem, oldtime);
			insert_Ucall_list_on_link(call_list_ptr,
				link_ptr->link_elem);
		}
	}
}

/*
**  Function  : void insert_call_list_on_link(CallList *call_list,
**                                            Link *link_ptr)
**  Return    : none
**  Remarks   : 'call_list' points to the CallList record that must be
**              inserted into the list of the calls currently supported by
**              'link_ptr'. The List is a double linked circular list with
**              an empty head. 'link_ptr->calls_on' is the pointer to the
**              empty head that is initialized with an 'end_time' field
**              equal to the minimum value of a 'double'.
**              The record insertion exploits the presence of flags
**              addressing partial sublists inside the Link Call List in order
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void insert_call_list_on_link(CallList *call_list, Link *link_ptr)
{
 CallList       *cur;
 double         sched_time;
 int        half_of_list;
 int        flag_number;
 int        id = link_ptr->id;

 link_call_count[id]++;

/* The test for insertion starts from the call_list record referring to */
/* the call with the highest 'end_time' field.                          */
 sched_time = call_list->call_elem->end_time;
 cur = link_ptr->calls_on->pred;

/* The first step towards insertion consists in the dichotomous search  */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming call is lower than the one of    */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the call is placed at the end of the link Call List.       */
 if(cur->call_elem->end_time > sched_time)
   {
    flag_number = num_link_flag[id] >> 1;
    half_of_list = num_link_flag[id] >> 2;

    while(half_of_list)
      {
       if(link_flags[id][flag_number]->call_elem->end_time > sched_time)
          flag_number += half_of_list;
       else
          flag_number -=half_of_list;

       half_of_list >>= 1;
      }

    cur = link_flags[id][flag_number];

/* The linear search through the sublist is started here.               */
    while(cur->call_elem->end_time < sched_time)
       cur = cur->next;
    while(cur->call_elem->end_time > sched_time)
       cur = cur->pred;
   }

/* Insertion of the CallList record.                                    */
 call_list->pred = cur;
 call_list->next = cur->next;
 call_list->next->pred = call_list;
 cur->next = call_list;

/* Provisional criterion to manage the snapshot sequence. 		*/
 if(
/* Snapshots are still requested.			 */
     (max_state_photos > 0) && 	
/* The initial transient is over.			 */
     (first_time == FALSE) &&	
/* The relative load threshold has been exceeded.	 */
     (link_carried_load[link_ptr->id][TOTAL_LINK_LOAD] >	
      rel_load_ths * capacity / link_ptr->timescale) &&
/* At least one call is currently active on the link 	 */
/* (maybe redundant test).				 */
     (link_ptr->calls_on->next != link_ptr->calls_on) &&
/* The minimum congestion duration is beeing exceeded.	 */
     (link_ptr->calls_on->next->call_elem->end_time >
      (ancles_time + cong_duration)))

     {
/* If we are using Ucalls check also if there is 		 */
/* At least one Ucall is currently active on the link 	 */
     if (( UCall_present) && (min_ucall_present))
       {
/* TEMPORANEAMENTE CAMBIATO PER CAPIRE QUALCOSA:
        if (link_Ucall_stat[link_ptr->id]->no_of_Ucall == min_ucall_present)
           manage_photo_list(link_ptr->id);*/
/* The original: */
        if (link_Ucall_stat[link_ptr->id]->no_of_Ucall >= min_ucall_present++)
           /* next photo will be taken if there is one more call... */
           manage_photo_list(link_ptr->id);

       }
     else
        manage_photo_list(link_ptr->id);
     }
}


/*
**  Function  : void dequeue_call_list_from_link(CallList *call,
**                                               Link *link_ptr)
**  Return    : none
**  Remarks   : this function removes the released call from the circular
**              list it belongs to by connecting its predecessor with its
**              successor.
**              The array of companion flags is updated every time the first
**              sublist gets empty.
*/
void dequeue_call_list_from_link(CallList *call, Link *link_ptr)
{
 register int   i, j;
 CallList       *cur;
 long           step;
 int 		id = link_ptr->id;

 link_call_count[id]--;

 call->pred->next = call->next;
 call->next->pred = call->pred;

/* Each time a CallList record is drawn from a link Call List the state */
/* of the pointers used as companion flags is checked: if at least two  */
/* flags are pointing to the outgoing CallList record the set of flags  */
/* needs to be rearranged over the link Call List.                      */
 if((link_flags[id][num_link_flag[id] - 1] == call) ||
    (link_flags[id][0] == link_flags[id][1]))
    {
     step = link_call_count[id] / (num_link_flag[id] + 1);
     if(step > MAX_STEP)
       {
        free(link_flags[id]);
        num_link_flag[id] <<= 1;

        link_flags[id] =
           (CallList **)MMmalloc((sizeof(CallList *)) * num_link_flag[id],
           			 "dequeue_call_list_from_link");
        step = link_call_count[id] / (num_link_flag[id]+1);
       }
     cur = link_ptr->calls_on->pred;
     for(j = 0; j < num_link_flag[id]; j++)
       {
        for(i = 0; i < step; i++)
           cur = cur->pred;
        link_flags[id][j] = cur;
       }
    }
 call_list_release(call);
}

/*
**  Function  : void unrel_dequeue_call_list_from_link(CallList *call,
**                                    Link *link_ptr, double sched_time)
**  Return    : none
**  Remarks   : this function removes the call from the circular
**              list it belongs to by connecting its predecessor with its
**              successor, but it does not release the call_list
**              The array of companion flags is updated every time the first
**              sublist gets empty.
*/
void unrel_dequeue_call_list_from_link(CallList *call, Link *link_ptr,
									   double sched_time)
{
 register int   i, j;
 CallList       *cur;
 long           step;
 int 		id = link_ptr->id;

  int        	flag_number;
 int        	half_of_list;
 int		found;

 link_call_count[id]--;

 call->pred->next = call->next;
 call->next->pred = call->pred;

/* Each time a CallList record is drawn from a link Call List, */
/* but it isn't sure, that he was the first in the time order, the state*/
/* of the pointers used as companion flags is checked: if there is a  */
/* flag pointing to the outgoing CallList record the set of flags  */
/* needs to be rearranged over the link Call List.                      */
/* Searching for the element in the flag list 				*/
 if(link_flags[id][0] == call)
  found=1;
 else
  {
   found = 0;
   flag_number = num_link_flag[id] >> 1;
   half_of_list = num_link_flag[id] >> 2;

   while(half_of_list)
    {
     if(link_flags[id][flag_number] == call)
      half_of_list = 0;
     else
      {
       if(link_flags[id][flag_number]->call_elem->end_time > sched_time)
         flag_number += half_of_list;
       else
         flag_number -= half_of_list;
       half_of_list >>= 1;
      }
    }

   if(link_flags[id][flag_number] == call)
    found = 1;
  }
 if( found ||
    (link_flags[id][0] == link_flags[id][1]))
 {
     step = link_call_count[id] / (num_link_flag[id] + 1);
     if(step > MAX_STEP)
       {
        free(link_flags[id]);
        num_link_flag[id] <<= 1;

        link_flags[id] =
           (CallList **)MMmalloc((sizeof(CallList *)) * num_link_flag[id],
                                  "unrel_dequeue_call_list_from_link");
        step = link_call_count[id] / (num_link_flag[id]+1);
       }
     cur = link_ptr->calls_on->pred;
     for(j = 0; j < num_link_flag[id]; j++)
       {
        for(i = 0; i < step; i++)
           cur = cur->pred;
        link_flags[id][j] = cur;
       }
 }
}


/*
**  Function  : void init_event_list(void)
**  Return    : none
**  Remarks   : this function initializes the Event list allocating the head
**              element: 'head_of_event_list' and setting its 'ev_time' field
**              to the minimum value of a 'double' (DBL_MIN) as required by
**              the insertion algorithm.
**              The array of pointers supporting the Event List management
**              is also initialized.
*/
void init_event_list(void)
{
 int        i;

 head_of_event_list = event_alloc();
 head_of_event_list->ev_time = DBL_MIN;
 head_of_event_list->next = head_of_event_list;
 head_of_event_list->pred = head_of_event_list;

 num_event_flag = START_FLAG;
 event_flags = (Event **)MMmalloc((sizeof(Event *)) * num_event_flag,
                                  "init_event_list");

 for(i = 0; i < num_event_flag; i++)
     event_flags[i] = head_of_event_list;
}


/*
**  Function  : void init_call_list(void)
**  Return    : none
**  Remarks   : this function initializes the Call list allocating the head
**              element: 'head_of_call_list' and setting its 'end_time' field
**              to the minimum value of a 'double' (DBL_MIN) as required by
**              the insertion algorithm.
**              The array of pointers supporting the general Call List
**              management is also initialized.
*/
void init_call_list(void)
{
 int        i;

 head_of_call_list = call_alloc();
 head_of_call_list->end_time = DBL_MIN;
 head_of_call_list->pred = head_of_call_list;
 head_of_call_list->next = head_of_call_list;

 num_call_flag = START_FLAG;
 call_flags = (Call **)MMmalloc((sizeof(Call *)) * num_call_flag,
                                "init_call_list");
 for(i = 0; i < num_call_flag; i++)
    call_flags[i] = head_of_call_list;
}


/*
**  Function  : void init_call_list_on_link(void)
**  Return    : none
**  Remarks   : this function initializes the call list for each link
**              allocating the head element ('calls_on') and setting its
**              'call_elem' field to point to 'head_of_call_list' whose
**              'end_time' field contains the minimum value of a 'double',
**              as required by the insertion algorithm.
**              The arrays of pointers supporting the link Call Lists
**              management are also initialized.
*/
void init_call_list_on_link(void)
{
 register int   i, j;

 link_flags = (CallList ***)MMmalloc(sizeof(CallList **) * MAX_LINK /*no_of_links*/,
                                     "init_call_list_on_link");

 num_link_flag = (int *)MMmalloc(sizeof(int) * MAX_LINK /*no_of_links*/,
                                     "init_call_list_on_link");
 link_call_count = (long *)MMmalloc(sizeof(long) * MAX_LINK /*no_of_links*/,
                                     "init_call_list_on_link");
 for(i = 0; i < no_of_links; i++)
   {
    link_vect[i]->calls_on = call_list_alloc();
    link_vect[i]->calls_on->call_elem = head_of_call_list;
    link_vect[i]->calls_on->pred = link_vect[i]->calls_on;
    link_vect[i]->calls_on->next = link_vect[i]->calls_on;

    num_link_flag[i] = START_FLAG;
    link_call_count[i] = 0;
    link_flags[i] =
       (CallList **)MMmalloc((sizeof(CallList *)) * num_link_flag[i],
                             "init_call_list_on_link");
    for(j = 0; j < num_link_flag[i]; j++)
       link_flags[i][j] = link_vect[i]->calls_on;
   }
}

void init_call_list_on_new_link(int id)
{
	Link* new_link=link_vect[id];
	int j;

	new_link->calls_on = call_list_alloc();
	new_link->calls_on->call_elem = head_of_call_list;
	new_link->calls_on->pred = new_link->calls_on;
	new_link->calls_on->next = new_link->calls_on;
	num_link_flag[new_link->id] = START_FLAG;
	link_call_count[id] = 0;
	link_flags[id] =
		(CallList **)MMmalloc((sizeof(CallList *)) * 
		num_link_flag[id],"init_call_list_on_new_link");
	for(j = 0; j < num_link_flag[id]; j++)
		link_flags[id][j] = new_link->calls_on;
}

void exchange_call_list_on_links(int a, int b)
{
	int tempint;
	CallList** tempcl;
	
	tempint = num_link_flag[a];
	num_link_flag[a] = num_link_flag[b];
	num_link_flag[b] = tempint;
	
	tempint = link_call_count[a];
	link_call_count[a] = link_call_count[b];
	link_call_count[b] = tempint;
	
	tempcl = link_flags[a];
	link_flags[a] = link_flags[b];
	link_flags[b] = tempcl;
}


/*
**  Function  : void init_Ucall_list_on_link(void)
**  Return    : none
**  Remarks   : this function initializes the Class U call list for each link.
**              It allocates the head element ('calls_on') and sets its
**              'call_elem' field to point to 'head_of_call_list', whose
**              'end_time' field contains the minimum value of a 'double',
**              as required by the insertion algorithm.
**              The arrays of pointers supporting the Class U link call Lists
**              management are also initialized.
*/
void init_Ucall_list_on_link(void)
{
 register int    i, j;

 Ulink_flags = (CallList ***)MMmalloc(sizeof(CallList **) * MAX_LINK /*no_of_links*/,
                                      "init_Ucall_list_on_link");

 Unum_link_flag = (int *)MMmalloc(sizeof(int) * MAX_LINK /*no_of_links*/,
                                      "init_Ucall_list_on_link");
 Ulink_call_count = (long *)MMmalloc(sizeof(long) * MAX_LINK /*no_of_links*/,
                                      "init_Ucall_list_on_link");
 for(i = 0; i < no_of_links; i++)
   {
    link_Ucall_stat[i]->U_calls_on = call_list_alloc();
    link_Ucall_stat[i]->U_calls_on->call_elem = head_of_call_list;
    link_Ucall_stat[i]->U_calls_on->pred = link_Ucall_stat[i]->U_calls_on;
    link_Ucall_stat[i]->U_calls_on->next = link_Ucall_stat[i]->U_calls_on;

    Unum_link_flag[i] = START_FLAG;
    Ulink_call_count[i] = 0;
    Ulink_flags[i] =
       (CallList **)MMmalloc((sizeof(CallList *)) * Unum_link_flag[i],
       					"init_Ucall_list_on_link");
    for(j = 0; j < Unum_link_flag[i]; j++)
	Ulink_flags[i][j] = link_Ucall_stat[i]->U_calls_on;
   }
}

void init_Ucall_list_on_new_link(int id)
{
	Link_ucall_stat* new_lus=link_Ucall_stat[id];
	int j;
	
	new_lus->U_calls_on = call_list_alloc();
	new_lus->U_calls_on->call_elem = head_of_call_list;
	new_lus->U_calls_on->pred = new_lus->U_calls_on;
	new_lus->U_calls_on->next = new_lus->U_calls_on;
	Unum_link_flag[id] = START_FLAG;
	Ulink_call_count[id] = 0;
	Ulink_flags[id] =
		(CallList **)MMmalloc((sizeof(CallList *)) * 
		Unum_link_flag[id],"init_Ucall_list_on_new_link");
	for(j = 0; j < Unum_link_flag[id]; j++)
		Ulink_flags[id][j] = 
		new_lus->U_calls_on;
}

void exchange_Ucall_list_on_links(int a, int b)
{
	int tempint;
	CallList** tempcl;
	
	tempint = Unum_link_flag[a];
	Unum_link_flag[a] = Unum_link_flag[b];
	Unum_link_flag[b] = tempint;
	
	tempint = Ulink_call_count[a];
	Ulink_call_count[a] = Ulink_call_count[b];
	Ulink_call_count[b] = tempint;
	
	tempcl = Ulink_flags[a];
	Ulink_flags[a] = Ulink_flags[b];
	Ulink_flags[b] = tempcl;
}

/*
**  Function  : void insert_Ucall_list_on_link(CallList *call_list,
**                                             Link *link_ptr)
**  Return    : none
**  Remarks   : 'call_list' points to the CallList record that must be
**              inserted into the list of the Class U calls that are
**		currently supported by 'link_ptr'.
**		The List is a double linked circular list with an empty head.
**		'link_ptr->calls_on' is the pointer to the empty head that
**		is initialized with an 'end_time' field equal to the
**		minimum value of a 'double'.
**              The record insertion exploits the presence of flags
**              addressing partial sublists inside the Link UCall List in order
**              to get the right location faster.
**              The search inside the flag set is dichotomous.
*/
void insert_Ucall_list_on_link(CallList *call_list, Link *link_ptr)
{
 CallList       *cur;
 double         sched_time;
 int		half_of_list;
 int		flag_number;
 int		id = link_ptr->id;

 Ulink_call_count[id]++;

/* The test for insertion starts from the call_list record referring to */
/* the call with the highest 'end_time' field.                          */
 sched_time = call_list->call_elem->end_time;
 cur = link_Ucall_stat[link_ptr->id]->U_calls_on->pred;

/* The first step towards insertion consists in the dichotomous search  */
/* of the target sublist, through the available set of companion flags. */
/* If the temporal key of the incoming call is lower than the one of    */
/* the last item in the list, the dichotomous search is started,        */
/* otherwise the call is placed at the end of the link Call List.       */
 if(cur->call_elem->end_time > sched_time)
   {
    flag_number = Unum_link_flag[id] >> 1;
    half_of_list = Unum_link_flag[id] >> 2;

    while(half_of_list)
      {
       if(Ulink_flags[id][flag_number]->call_elem->end_time > sched_time)
          flag_number += half_of_list;
       else
          flag_number -=half_of_list;

       half_of_list >>= 1;
      }

    cur = Ulink_flags[id][flag_number];

/* The linear search through the sublist is started here.               */
    while(cur->call_elem->end_time < sched_time)
       cur = cur->next;
    while(cur->call_elem->end_time > sched_time)
       cur = cur->pred;
   }

/* Insertion of the CallList record.                                    */
 call_list->pred = cur;
 call_list->next = cur->next;
 call_list->next->pred = call_list;
 cur->next = call_list;
}


/*
**  Function  : void dequeue_Ucall_list_from_link(CallList *call,
**                                               Link *link_ptr)
**  Return    : none
**  Remarks   : this function removes the released Class U call from the
**		circular list it belongs to by connecting its predecessor
**		with its successor.
**              The array of companion flags is updated every time the first
**              sublist gets empty.
*/
void dequeue_Ucall_list_from_link(CallList *call, Link *link_ptr)
{
 register int   i, j;
 CallList       *cur;
 long           step;
 int		id = link_ptr->id;

 Ulink_call_count[id]--;

 call->pred->next = call->next;
 call->next->pred = call->pred;

/* Each time a CallList record is drawn from a Class U Call List the 	*/
/* state of the pointers used as companion flags is checked: if at 	*/
/* least two flags are pointing to the outgoing CallList record, the 	*/
/* set of flags needs to be rearranged over the Class U Call List.	*/
 if((Ulink_flags[id][Unum_link_flag[id] - 1] == call) ||
    (Ulink_flags[id][0] == Ulink_flags[id][1]))
    {
     step = Ulink_call_count[id] / (Unum_link_flag[id] + 1);
     if(step > MAX_STEP)
       {
        free(Ulink_flags[id]);
        Unum_link_flag[id] <<= 1;

        Ulink_flags[id] =
           (CallList **)MMmalloc((sizeof(CallList *)) * Unum_link_flag[id],
                                  "dequeue_Ucall_list_from_link");
        step = Ulink_call_count[id] / (Unum_link_flag[id]+1);
       }
     cur = link_Ucall_stat[id]->U_calls_on->pred;
     for(j = 0; j < Unum_link_flag[id]; j++)
       {
        for(i = 0; i < step; i++)
           cur = cur->pred;
        Ulink_flags[id][j] = cur;
       }
    }
 call_list_release(call);
}

/*
**  Function  : void unrel_dequeue_Ucall_list_from_link(CallList *call,
**                                              Link *link_ptr,
**						double sched_time)
**  Return    : none
**  Remarks   : the same functionality as the normal dequeue, but doesn't
**		release the calllist. sched_time is used for the search in the
**		array of companion flags
*/
void unrel_dequeue_Ucall_list_from_link(CallList *call, Link *link_ptr,
						double sched_time)
{
 register int   i, j;
 CallList       *cur;
 long           step;
 int        id = link_ptr->id;

 int        	flag_number;
 int        	half_of_list;
 int		found;

 Ulink_call_count[id]--;

 call->pred->next = call->next;
 call->next->pred = call->pred;

/* Each time a CallList record is drawn from a Class U Call List 	*/
/* but it isn't sure, that he was the first in the time order, the	*/
/* state of the pointers used as companion flags is checked: if there	*/
/* is a flag pointing to the outgoing call the set of flags needs to	*/
/* be rearranged over the Class U Call List.				*/

/* Searching for the element in the flag list 				*/
 if(Ulink_flags[id][0] == call)
  found=1;
 else
  {
   found = 0;
   flag_number = Unum_link_flag[id] >> 1;
   half_of_list = Unum_link_flag[id] >> 2;

   while(half_of_list)
    {
     if(Ulink_flags[id][flag_number] == call)
      half_of_list = 0;
     else
      {
       if(Ulink_flags[id][flag_number]->call_elem->end_time > sched_time)
         flag_number += half_of_list;
       else
         flag_number -= half_of_list;
       half_of_list >>= 1;
      }
    }

   if(Ulink_flags[id][flag_number] == call)
    found = 1;
  }
 if( found ||
    (Ulink_flags[id][0] == Ulink_flags[id][1]))
    {
     step = Ulink_call_count[id] / (Unum_link_flag[id] + 1);
     if(step > MAX_STEP)
       {
        free(Ulink_flags[id]);
        Unum_link_flag[id] <<= 1;

        Ulink_flags[id] =
           (CallList **)MMmalloc((sizeof(CallList *)) * Unum_link_flag[id],
                                  "dequeue_Ucall_list_from_link");
        step = Ulink_call_count[id] / (Unum_link_flag[id]+1);
       }
     cur = link_Ucall_stat[id]->U_calls_on->pred;
     for(j = 0; j < Unum_link_flag[id]; j++)
       {
        for(i = 0; i < step; i++)
           cur = cur->pred;
        Ulink_flags[id][j] = cur;
       }
    }
}

/*
**  Function  : CallList* search_Ucall_list_on_link(Link *link_ptr,
**					      Call *call, double sched_time)
**  Return    : The searched CallList
**  Remarks   : Finds the CallList on the Link refered by the 'link_ptr'
**              whose call_elem record is Call (for Ucalls only).
**		sched_time is used for the search in the
**		array of companion flags
*/
CallList *search_Ucall_list_on_link(Link *link_ptr, Call *call, double sched_time)
{

 CallList	*cur;
 int		half_of_list;
 int		flag_number;
 int		id = link_ptr->id;

/* Using the same method find a CallList, as in the insert function */
 cur = link_Ucall_stat[id]->U_calls_on->pred;

 flag_number = Unum_link_flag[id] >> 1;
 half_of_list = Unum_link_flag[id] >> 2;

 while(half_of_list)
  {
   if(Ulink_flags[id][flag_number]->call_elem->end_time > sched_time)
     flag_number += half_of_list;
   else
     flag_number -= half_of_list;

   half_of_list >>= 1;
  }

 cur = Ulink_flags[id][flag_number];

/* The linear search through the sublist is started here.               */
    if(cur->call_elem->end_time < sched_time)
     {
      while(cur->call_elem != call)
        cur = cur->next;
     }
    else
     {
      while(cur->call_elem != call)
        cur = cur->pred;
     }

    return cur;
}

/*
**  Function  : CallList* search_call_list_on_link(Link *link_ptr,
**					      Call *call, double sched_time)
**  Return    : The searched CallList
**  Remarks   : Finds the CallList on the Link refered by the 'link_ptr'
**              whose call_elem record is Call.
**		sched_time is used for the search in the
**		array of companion flags
*/
CallList *search_call_list_on_link(Link *link_ptr, Call *call, double sched_time)
{
 CallList       *cur;
 int        half_of_list;
 int        flag_number;
 int        id = link_ptr->id;

/* Using the same method find a CallList, as in the insert function */
 cur = link_ptr->calls_on->pred;

    flag_number = num_link_flag[id] >> 1;
    half_of_list = num_link_flag[id] >> 2;

    while(half_of_list)
      {
       if(link_flags[id][flag_number]->call_elem->end_time > sched_time)
          flag_number += half_of_list;
       else
          flag_number -=half_of_list;

       half_of_list >>= 1;
      }

    cur = link_flags[id][flag_number];

/* The linear search through the sublist is started here.               */
    if(cur->call_elem->end_time < sched_time)
     {
      while(cur->call_elem != call)
        cur = cur->next;
     }
    else
     {
      while(cur->call_elem != call)
        cur = cur->pred;
     }

    return cur;
}