/*
**  Module    : load_balancing.c
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Gabriele Favalessa, Luca Medico, Marco Mellia
**
**  Synopsis  : this module orders paths using load balancing
**
**  Functions : static void copy_in_sl_vect(double *, Path *)
**              static void sort_sl_vect(double *, int    )
**              static void free_sl_vect(double **, int    )
**              static double eval_max_static_load(Path *)
**              static double eval_static_contrib(int    , LinkList *)
**              static void update_static_load(Path *, int    , int    )
**              static void init_static_load(void)
**              static int cmp_num_eq_paths(struct gg_path *, struct gg_path *)
**              static int cmp_num_hops(struct gg_path *, struct gg_path *)
**              static int cmp_length(struct gg_path *, struct gg_path *)
**              static Boolean equivalent_paths(Path *, Path *, int    )
**              static int     count_eq_paths(Path *, int    )
**              static Path* clone_path_list(Path *)
**              static Path* remove_best_path(Path **, int    )
**              void balance_paths_equiv_first(int    )
**              void balance_paths_weight_first(int    )
**              static void assign_best_path(struct gg_path *, int     *,
**                                           int     *, int    , int     *)
**              void do_the_balance(int    , int    , struct gg_paths*)
**
*/

#include "ancles.h"

/* Define this if you  want ANCLES to report the pat_matix in a file */
/* Also the static load is reported */
/* #define DEBUG_BALANCING */

/*
** Global Structures
*/
static double* static_load = NULL;
                            /* Static load vector.  */
			    /* The vector index is the link number. */

struct gg_paths
{                                    
  struct gg_paths* next;
  int     src_cgen;     /* source call generator                        */
  int     dest_cgen;    /* destination call generator                   */
  int     num_eq_paths; /* number of equivalent paths to the first path */ 
  Path* path_list;      /* paths list from src_cgen to dest_cgen        */
};



/*
**  Function Prototypes
*/ 
static void copy_in_sl_vect(double *, Path *);
static void sort_sl_vect(double *, int    );
static void free_sl_vect(double **, int    );
static double eval_max_static_load(Path *);
static double eval_static_contrib(int    , LinkList *);
static void update_static_load(Path *, int    , int    );
static int cmp_num_eq_paths(struct gg_paths *, struct gg_paths *);
static int cmp_num_hops(struct gg_paths *, struct gg_paths *);
static int cmp_length(struct gg_paths *, struct gg_paths *);
static Boolean equivalent_paths(Path *, Path *, int    );
static int     count_eq_paths(Path *, int    );
static Path* clone_path_list(Path *);
static Path* remove_best_path(Path **, int    );
void balance_paths(int    , int    );
void do_the_balance(int    , int    , struct gg_paths*);

#ifdef DEBUG_BALANCING
/*
**  Function  : fstampa_cgen_to_cgen(FILE *ffpp, int     ord, int     i,
**                                                               int     j)
**  Return    : None.
**  Remarks   : Report in the file ffpp all the paths between source i and
**              destination j, from the matrix ord.
*/
static void fstampa_cgen_to_cgen(FILE *ffpp, int     ord, int     i, int     j)
{
 LinkList *ptr;
 Path *path_point;

 path_point = path_mat[ord][i][j];  
 if(path_point == NULL)
    return;
 fprintf(ffpp,"\nPaths from cgen %s to cgen %s :\n", cgen_name(i), cgen_name(j));
 /* for all the path */ 
 for(;path_point != NULL; path_point = path_point->next)
   {
    fprintf(ffpp,"# hops: %d    Length (km): %.1f   Path: ",
            path_point->no_of_hops, path_point->length);
    for(ptr = path_point->path; ptr != NULL; ptr = ptr->next)
       fprintf(ffpp,"%s ", link_name(ptr->link_elem->id));
    fprintf(ffpp,"\n");
   }
}

/*
**  Function  : save_path(int     ord)
**  Return    : None.
**  Remarks   : save the path matrix in the correct file.
*/
static void save_path(int     ord)
{
FILE *ffpp;
int     i,j;
char name[80];

strcpy(name, file_radix);  
switch(ord)
    {
    case IN_ORDER_OF_HOP_NUMBER : strcat(name, ".hop");
    				  break;

    case IN_ORDER_OF_LENGTH : strcat(name, ".len");
    			      break;

    case BALANCED_IN_ORDER_OF_HOP_NUMBER : strcat(name, ".bhop");
    					   break;

    case BALANCED_IN_ORDER_OF_LENGTH : strcat(name, ".blen");
    				       break;

    default : printf("error in function save_path\n");
              exit(1);
    }
 if((ffpp = fopen(name, "w")) == NULL)
   {
    printf("Warning:  The file %s cannot be opened.\n", name);
    printf("          Properly formatted data regarding the Path ");
    printf("connecting calls generators\n          will not be available.\n");
   }
 else
   {
    for(i=0;i<no_of_cgens;i++)
       for(j=0;j<no_of_cgens;j++)
         fstampa_cgen_to_cgen(ffpp,ord,i,j);
    fclose(ffpp); 
   }
}


/*
**  Function  : static void save_primary_number_call(int     ord)
**  Return    : None.
**  Remarks   : save the number of primary call routed on each link,
**              for each kind of call generators
*/
static void save_primary_number_call(int     ord)
{
register int i,j;
LinkList *list_ptr;
char name[80];
FILE *ffpp;
int     gcl1[MAX_LINK];
int     gcl2[MAX_LINK];
int     gcl3[MAX_LINK];
int     gcl4[MAX_LINK];
int     gclu[MAX_LINK];

for(i = 0; i < MAX_LINK; i++)
  {
  gcl1[i] = 0;
  gcl2[i] = 0;
  gcl3[i] = 0;
  gcl4[i] = 0;
  gclu[i] = 0;
  }
for(i = 0; i < no_of_cgens; i++)
  for(j = 0; j < no_of_cgens; j++)
    {
    if(path_mat[ord][i][j] == NULL)
      continue;
    for(list_ptr = path_mat[ord][i][j]->path; list_ptr != NULL;
        list_ptr = list_ptr->next)
      {
       switch(callgen_vect[i]->type)
        {
        case CG_TYPE1 : ++gcl1[list_ptr->link_elem->id];
		        break;
	case CG_TYPE2 : ++gcl2[list_ptr->link_elem->id];
    			break;
	case CG_TYPE3 : ++gcl3[list_ptr->link_elem->id];
    			break;
	case CG_TYPE4 : ++gcl4[list_ptr->link_elem->id];
    			break;
	case CG_TYPE5 : ++gclu[list_ptr->link_elem->id];
    			break;
	default       : printf("error calling save_primary_number_call\n");
	                exit(1);
        } 
      }
    } /* for */
strcpy(name, file_radix);
if (ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
  strcat(name, ".pri_num_cal_hop");
else
  strcat(name, ".pri_num_cal_len");
if((ffpp = fopen(name, "w")) == NULL)
  {
  printf("error opening %s file\n",name);
  exit(1);
  }
fprintf(ffpp,"#Number of static call for every links\n\n");
fprintf(ffpp,"#Link name\tCL1\tCL2\tCL3\tCL4\tCLU\n");
for(i = 0; i < no_of_links; i++)
  fprintf(ffpp,"%10s\t%d\t%d\t%d\t%d\t%d\n", link_name(i), gcl1[i], gcl2[i],
  					   gcl3[i], gcl4[i], gclu[i]);
fclose(ffpp);
}



/*
**  Function  : static void save_static_load(int     is_primary, int     ord)
**  Return    : None.
**  Remarks   : save the static load accumulated on each link.
*/
static void save_static_load(int     is_primary, int     ord)
{
FILE *out;
int     i;
char name[80];
static int numero_chiamate = 0;

numero_chiamate++;

strcpy(name, file_radix);
if(ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
    strcat(name, ".stat_load_hop");
else
    strcat(name, ".stat_load_len");
if( numero_chiamate == 1 )
   if((out = fopen(name, "w")) == NULL)
     {
     printf("error opening %s file\n",name);
     exit(1);
     }
else        
   if((out = fopen(name, "a")) == NULL)
     {
     printf("error re-opening %s file\n",name);
     exit(1);
     }

switch (numero_chiamate)
  case 1:
     if (selected_non_U_priority)
        fprintf(out, "#Carico statico primario per la CLASSE NON U\n");
     else
        fprintf(out, "#Carico statico primario complessivo\n");
     break;
  case 2:
     if (selected_non_U_priority)
        fprintf(out, "#Carico statico totale per la CLASSE NON U\n");
     else
        fprintf(out, "#Carico statico totale complessivo\n");
     break;
  case 3:
     fprintf(out, "#Carico statico totale per la CLASSE NON U\n");
     fprintf(out, "#E carico statico primario per la CLASSE U\n");
     break;
  case 4:
     fprintf(out, "#Carico statico totale per la CLASSE NON U\n");
     fprintf(out, "#E carico statico totale per la CLASSE U\n");
     break;
  
for(i = 0; i < no_of_links; ++i)
    fprintf(out, "link %s:\t%f\n\n", link_name(i), static_load[i]);
fclose(out);
}


#if 0
static void save_after_sort(struct gg_paths *paths, int     num_paths, 
                            int     ord)
{
FILE *ffpp;
register int i;
char name[80];

strcpy(name, file_radix);
if(ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
  strcat(name, ".after_sort_hop");
else
  strcat(name, ".after_sort_len");
if((ffpp = fopen(name, "w")) == NULL)
  {
  printf("error opening %s file\n",name);
  exit(1);
  }
for(i = 0; i < num_paths; i++)
  {
  fprintf(ffpp, "Path from %s to %s:\n", cgen_name(paths[i].src_cgen),
                                         cgen_name(paths[i].dest_cgen));
  fprintf(ffpp,"\n\n");
  }
fclose(ffpp);
}
#endif

#endif /* DEBUG_BALANCING */



/*
**  Function  : static void copy_in_sl_vect(double *sl, Path *path_ptr)
**  Return    : none.
**  Remarks   : copy static load of a path in sl_vect
**              Is used when two (or more) path have the same maximin static
**              load, to choice between that.
*/
static void copy_in_sl_vect(double *sl, Path *path_ptr)
{
 LinkList *path;
 int     jmax;
 int     j;

 path = path_ptr->path->next;
 jmax = path_ptr->no_of_hops;
 for(j = 0; j < jmax; j++)
	 {
	 sl[j] = static_load[path->link_elem->id];
	 path=path->next;
	 }
 /*  Put a terminator at the end. Needed later. */
 sl[jmax] = -1.0;
}



/*
**  Function  : static void sort_sl_vect(double *sl, int     jmax)
**  Return    : none.
**  Remarks   : the function order the vector in decreasing manner
*/
static void sort_sl_vect(double *sl, int     jmax)
{
int     j;
int     k;
int     kmax;
double tmp_load;
double max_load;

for(j = 0; j < jmax; j++)
  {
  max_load = -2.0;
  kmax = -1;
  for(k = j; k < jmax; k++)
    if(sl[k] > max_load) 
      {
      max_load = sl[k];
      kmax = k;
      }
  tmp_load = sl[j]; 
  sl[j] = sl[kmax];
  sl[kmax] = tmp_load;
  }
}



/*
**  Function  : static void free_sl_vect(double **sl_vect, int     imax)
**  Return    : none.
**  Remarks   : free memory allocated for sl_vect
**              note: sl_vect is the vector containg the sl vector...
*/
static void free_sl_vect(double **sl_vect, int     imax)
{
 int     i;

 for(i = 0; i < imax; i++)
   if(sl_vect[i] != NULL)
      {
      free(sl_vect[i]);
      sl_vect[i] = NULL;
      }
}			



/*
**  Function  : static double eval_max_static_load_in_path(Path *path_base)
**  Return    : maximum static load in path pointe by path_base.
**  Remarks   : scan path pointed by path_base and determine the maximum
**              static load in relative links.
*/
static double eval_max_static_load(Path *path_base)
{
LinkList *path_ptr;
double max_static_load;

max_static_load = 0.0;

/* do not consider the user -> node link, but the others */
for(path_ptr = path_base->path->next;
	path_ptr->next != NULL; path_ptr=path_ptr->next)
  {
   if(static_load[path_ptr->link_elem->id] > max_static_load)
        max_static_load = static_load[path_ptr->link_elem->id]; 
  }

return max_static_load;
}



/*
**  Function  : static double eval_static_contrib(int     cgid, 
**                                                LinkList *path_ptr)
**  Return    : static load contribution.
**  Remarks   : static load contibution is evaluated according the CAC choosen
*/
static double eval_static_contrib(int     i, LinkList *path_ptr)
{
  double peak;
  double mean;
  double static_load;
  CGType1 *cg1;
  CGType2 *cg2;
  CGType3 *cg3;
  CGType4 *cg4;
  CGType5 *cg5;
  CGType6 *cg6;
  CGType7 *cg7;
  CGType8 *cg8;
  CGType9 *cg9;

  switch(callgen_vect[i]->type)  
        {                           
         case CG_TYPE1 :        /* CBR source.                          */
            cg1 = (CGType1 *)(callgen_vect[i]->info);
	    peak = (cg1->min_bit_rate + cg1->max_bit_rate) / 2.0;
            mean = peak;
            break;

         case CG_TYPE2 :        /* On-Off VBR source.                   */
            cg2 = (CGType2 *)(callgen_vect[i]->info);
            peak = (cg2->min_peak_rate + cg2->max_peak_rate) / 2.0;
            mean = peak * cg2->mean_on_time /
                (cg2->mean_on_time + cg2->mean_off_time);
            break;

         case CG_TYPE3 :        /* Uniform VBR source.                  */
            cg3 = (CGType3 *)(callgen_vect[i]->info);
            peak = cg3->max_bit_rate_avg;
            mean = (cg3->min_bit_rate_avg + cg3->max_bit_rate_avg) / 2.0;
            break;

         case CG_TYPE4 :        /* Video VBR source.                    */
            cg4 = (CGType4 *)(callgen_vect[i]->info);
            peak = cg4->higher_bit_rate_avg;
            mean = cg4->higher_bit_rate_avg * cg4->higher_bit_rate_prob +
                cg4->medium_bit_rate_avg * cg4->medium_bit_rate_prob +
                cg4->lower_bit_rate_avg * cg4->lower_bit_rate_prob;
            break;
	    
	 case CG_TYPE5 :        /* Unspecified source.            */
            cg5 = (CGType5 *)(callgen_vect[i]->info);
            peak = (cg5->min_max_rate + cg5->max_max_rate) / 2.0;
	    mean = peak;
            break;

	 case CG_TYPE6 :        /* ASON source.                  */
            cg6 = (CGType6 *)(callgen_vect[i]->info);
            peak = wavelength_capacity * 
				(cg6->min_no_of_waves + cg6->max_no_of_waves) / 2.0;
	    mean = peak;
            break;

	 case CG_TYPE7 :        /* On-Off ASON source.    */
            cg7 = (CGType7 *)(callgen_vect[i]->info);
			peak = cg7->no_of_waves * wavelength_capacity;
		mean = peak;
            break;

	 case CG_TYPE8 :        /* Never closing ASON source.    */
            cg8 = (CGType8 *)(callgen_vect[i]->info);
            peak = wavelength_capacity * cg8->no_of_waves;
	    mean = peak;
            break;

	 case CG_TYPE9 :        /* Pascal ASON source.            */
            cg9 = (CGType9 *)(callgen_vect[i]->info);
            peak = wavelength_capacity * 
				(cg9->min_no_of_waves + cg9->max_no_of_waves) / 2.0;
	    mean = peak;
            break;

	 default :              /* Sink.                          */
	    printf("error in function eval_static_contrib.\n");
	    printf("call generator considered has not peak or");
	    printf("medium bit rate\n");
	    exit(1);
        }

  switch(selected_cac)
    {
    case CALL_ADM_CTRL1: 
      static_load = eval_equivalent_bw(mean,peak,
                                capacity/path_ptr->link_elem->timescale,
                                cell_loss_rate);
      break;
    case CALL_ADM_CTRL2: 
      static_load = peak;
      break;
    case CALL_ADM_CTRL3: 
      static_load = peak;
      break;
    case CALL_ADM_CTRL4: 
        static_load = eval_modified_bw(mean,peak);
      break;
    default:
      printf("ERROR: unknown CAC %d in function eval_static_contrib\n",
	     selected_cac);
      exit(1);
    }

/* The Class U calls are considered separately. */
 if( callgen_vect[i]->type == CG_TYPE5)
    static_load = peak/beta;

 return static_load;
}



/*
**  Function  : static void update_static_load(Path *path_base, 
**                                             int     s, int     d)
**  Return    : none.
**  Remarks   : update static load in path pointed by path_base, connecting
**              source generator s and destination generator d. 
*/
static void update_static_load(Path *path_base, int     s, int     d)
{
  int     s_user_id;
  int     d_user_id;
  LinkList *path_ptr;

  s_user_id=callgen_vect[s]->ref_user->id;
  d_user_id=callgen_vect[d]->ref_user->id;

  for(path_ptr = path_base->path; path_ptr != NULL;
      path_ptr = path_ptr->next)
    static_load[path_ptr->link_elem->id] += eval_static_contrib(s,path_ptr)*
        traff_pattern[s_user_id][d_user_id] * path_ptr->link_elem->timescale;
}


/* The following function are used to sort the path */


/*
**  Function  : static int cmp_num_eq_paths(struct gg_paths* a,struct gg_paths* b)
**  Return    : difference between number of equivalent paths.
**  Remarks   : none.
*/
static int cmp_num_eq_paths(struct gg_paths* a, struct gg_paths* b)
{
   return a->num_eq_paths - b->num_eq_paths;
}


/*
**  Function  : static int cmp_num_hops(struct gg_paths* a, struct gg_paths* b)
**  Return    : difference between number of hops.
**  Remarks   : none.
*/
static int cmp_num_hops(struct gg_paths* a, struct gg_paths* b)
{
   return a->path_list->no_of_hops - b->path_list->no_of_hops;
}



/*
**  Function  : static int cmp_length(struct gg_path* a, struct gg_path* b)
**  Return    : difference between length
**  Remarks   : return {-1, 0, 1} and not the differnce.
*/
static int cmp_length(struct gg_paths* a, struct gg_paths* b)
{
   if (equ_double(a->path_list->length, b->path_list->length, DBL_CMP_THS))
      return 0;
   else if (a->path_list->length > b->path_list->length)
      return 1;
   else
      return -1;
}

/*
**  Function  : static int cmp_num_eq_and_hop(struct gg_paths* a,
**					         struct gg_paths* b)
**  Return    : the difference between the number of equivalen paths, and, if 
**              they are the same, the difference between the number of hops.
**  Remarks   : none.
*/
static int cmp_num_eq_and_hop(struct gg_paths* a, struct gg_paths* b)
{
  int result;
   
  result = a->num_eq_paths - b->num_eq_paths;
 
  if (! result)
    result = a->path_list->no_of_hops - b->path_list->no_of_hops;

#ifdef DEBUG_BALANCING
/* tentativo per cercare di ristabilire un ordine topologico nel criterio */
/* di ordinamento */
  if (! result)
    result = a->src_cgen - b->src_cgen;

  if (! result)
    result = a->dest_cgen - b->dest_cgen;
#endif
 
  return result;
}
 
/*
**  Function  : static int cmp_num_eq_and_length(struct gg_paths* a,
**					         struct gg_paths* b)
**  Return    : the difference between the number of equivalen paths, and, if 
**              they are the same, the normalized difference between the length.
**  Remarks   : none.
*/
static int cmp_num_eq_and_length(struct gg_paths* a, struct gg_paths* b)
{
  int result;
 
  result = a->num_eq_paths - b->num_eq_paths;
 
  if (! result)
    {
      if (equ_double(a->path_list->length, b->path_list->length, DBL_CMP_THS))
	result = 0;
      else if (a->path_list->length > b->path_list->length)
	result = 1;
      else
	result = -1;
    }

#ifdef DEBUG_BALANCING
/* tentativo per cercare di ristabilire un ordine topologico nel criterio */
/* di ordinamento */
  if (! result)
    result = a->src_cgen - b->src_cgen;

  if (! result)
    result = a->dest_cgen - b->dest_cgen;
#endif
 
  return result;
}

/*
**  Function  : static Boolean equivalent_paths(Path* a, Path* b, int     ord)
**  Return    : TRUE if paths are equivalent 
**  Remarks   : 
*/
static Boolean equivalent_paths(Path* a, Path* b, int     ord)
{
  if (ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
    {
      return 
        abs(a->no_of_hops - b->no_of_hops) <= range_of_equiv_hop 

        ? TRUE : FALSE;
    }
  else 
    if (ord == BALANCED_IN_ORDER_OF_LENGTH)
    {
      return( 
        ( (fabs(a->length - b->length) < range_of_equiv_length )

        ||

          (equ_double (fabs(a->length - b->length),  range_of_equiv_length
		                                             , DBL_CMP_THS))
        )
        ? TRUE : FALSE);
    }
  printf("internal error in equivalent_path function\n");
  exit(1);
}



/*
**  Function  : static int     count_eq_paths(Path* path, int     ord)
**  Return    : number of equivalent paths.
**  Remarks   : number of equivalent paths is returned according with the range
**              of equivalence.
*/
static int     count_eq_paths(Path* path, int     ord)
{
   int     num_eq_paths = 0;
   Path* first = path;

   while (path)
   {
      if (! equivalent_paths(first, path, ord))
         break;

      ++num_eq_paths;

      path = path->next;
   }

   return num_eq_paths;
}



/*
**  Function  : static Path* clone_path_list(Path* path)
**  Return    : pointer to the path list
**  Remarks   : none.
*/
static Path* clone_path_list(Path* path)
{
   Path* path_head = NULL, * last_path = NULL;

   for(; path != NULL; path = path->next)
   {
      if (last_path == NULL)
       {
         last_path = path_alloc();
         path_head = last_path;
       }
      else
       {
         last_path->next = path_alloc();
         last_path = last_path->next;
       }

      last_path->next = NULL;
      last_path->path = path->path;
      last_path->length = path->length;
      last_path->no_of_hops = path->no_of_hops;
   }

   return path_head;
}



/*
**  Function  : static Path *remove_best_path(Path **list_head_ptr,
**                                            int     num_eq_paths)
**  Return    : the path with the minimum maximum static load.
**  Remarks   : This function is the core of the load balancing algoritmh.
**              It choose between num_eq_paths paths in the list pointed
**              by list_head_prt the right path to be assigned, looking
**              for the minimum maximum-statically-loaded path.
**              
*/
static Path* remove_best_path(Path** list_head_ptr, int     num_eq_paths)
{
  int     i;
  Path* path, * prev_path;
  Path * best_path, * prev_best_path = NULL;
  double min_max_static_load;

  best_path = NULL;

  min_max_static_load = DBL_MAX;

  /* Look for the path with the smaller maximum static load over its
     links.  */
   
  /* remarks: path is the currently checked path, while prev_path is the     */
  /* previous analized path. This is necessary to remove path from the list. */
  for (i = 0, prev_path = NULL, path = *list_head_ptr; 
       i < num_eq_paths; 
       i++, prev_path = path, path = path->next)
    {
      double load;
      load = eval_max_static_load(path);

      if (equ_double(load, min_max_static_load, DBL_CMP_THS))
        /* if there are more than one minimum loaded path, the best */
	/* is choosen in a second time                              */
	best_path = NULL;
      else if (load < min_max_static_load)
	{
	  min_max_static_load = load;
	  best_path = path;
	  prev_best_path = prev_path;
	}
    }

  if (! best_path)   /* More than one path was found.  */
    {
      /* the criterion of choice is more complicated... */ 
      int     j;
      double **sl_vect;	
      double min_static_load;

      /* build a vector to store num_equ_path vectors of static load */
      sl_vect = MMmalloc((num_eq_paths * sizeof(double *)), "remove best path");

      for (i = 0, path = *list_head_ptr; 
	   i < num_eq_paths; 
	   i++, path = path->next)
	{
	  if (equ_double(eval_max_static_load(path), min_max_static_load, DBL_CMP_THS))
	    {
              /* if the path is one of the equivalent, build a vector of the */
	      /* static load for each link of the path.                      */
	      sl_vect[i] = MMmalloc((path->no_of_hops + 1) * sizeof(double),
				    "remove best path");
	      /* fill the sl vect with the static load of the link...*/
	      copy_in_sl_vect(sl_vect[i], path);
	      /* sort the vector, to allow the comparison between sl_vect[i] */
	      /* and sl_vect[j]                                              */
	      sort_sl_vect(sl_vect[i], path->no_of_hops + 1);
	    } 
	  else
	    sl_vect[i] = NULL;
	}
      /* now check the vectors ... */
      /* for all the maximum load do [] until we have found the best ... */ 
      for (j = 1; ! best_path; ++j)
	{
	  min_static_load = DBL_MAX;
	   
          /* for all the equivalent path... */
	  for (i = 0, prev_path = NULL, path = *list_head_ptr; 
	       i < num_eq_paths; 
	       i++, prev_path = path, path = path->next)
	    {
	       /* if the current path is one of the possible equivalent... */
	       if (sl_vect[i] != NULL)
		{
		  if (sl_vect[i][j] < 0.0)
		    /* if we are at the end, the path are perfectly equal */
		    {
		      best_path = path;
		      prev_best_path = prev_path;
		      break;
		    }

		  /* if there is more than one minimum loaded path try again */
		  if (equ_double(sl_vect[i][j], min_static_load, DBL_CMP_THS))
		    best_path = NULL;
		  else
		    if (sl_vect[i][j] < min_static_load)
		     {
		      /* We have found the mimimum loaded path until now ...*/
		      /* free the other path struct used in sorting         */
		      free_sl_vect(sl_vect, i);
		      /* remember the best path until now...                */
		      best_path = path;
		      /* ... and the previous to allow the removing from the*/
		      /* list ...                                           */
		      prev_best_path = prev_path;
		      /* the new minimum ...                                */
		      min_static_load = sl_vect[i][j];
		     }
		    else
		     {
		     /* this path can't be the best any more ... remove it */
		     /* from the sorting structure                         */
		      free(sl_vect[i]);
		      sl_vect[i] = NULL;
		     }
		}
	    }
	}
/* free the struct used for sorting */
      free_sl_vect(sl_vect, num_eq_paths);
      free(sl_vect);
    }

/* return the best path found */
 if (prev_best_path)
    prev_best_path->next = best_path->next;
  else
    *list_head_ptr = best_path->next;

  best_path->next = NULL;

  return best_path;
}

/*
**  Function  : static struct gg_paths* assign_paths(struct gg_paths* gg_list,
**						     int     ord)
**              
**  Return    : the head of the gg_paths list
**              
**  Remarks   : Assign one path for every gg pair in the list gg_list.
**              
*/
static struct gg_paths* assign_paths(struct gg_paths* gg_list, int     ord)
{
  struct gg_paths* prev, * curr, * next;

  /* Assign one path for every gg pair in the list 'gg_list'.  */ 
  for (curr = gg_list; curr != NULL; curr = curr->next)
    {
      Path* best_path;

      best_path = remove_best_path(&curr->path_list, curr->num_eq_paths);

      /* Insert best_path at the end of the list in path_mat.  */
      if(! path_mat[ord][curr->src_cgen][curr->dest_cgen])
	path_mat[ord][curr->src_cgen][curr->dest_cgen] = best_path;
      else
	{
	  Path* last;

	  for (last = path_mat[ord][curr->src_cgen][curr->dest_cgen];
	       last->next;
	       last = last->next)
	    ;
	  
	  last->next = best_path;
	}
      /* Update the static load vector */
      update_static_load(best_path, curr->src_cgen, curr->dest_cgen);

      /* Update the number of equivalent paths. */
      if (! --curr->num_eq_paths)
	/* If the old group is empty, create the next one, if possible */
	curr->num_eq_paths = count_eq_paths(curr->path_list, ord);
    }
  
  /* Now rebuild the list                         */
  /* Remove elements without paths from the list. */
  for (prev = NULL, curr = gg_list; curr != NULL ; curr = next)
    {
      next = curr->next;
      
      if (! curr->num_eq_paths)
	{
	  if (prev)
	    prev->next = curr->next; /* 'prev' does not change. */
	  else /* The next node is the new head. */
	    gg_list = curr->next; /* 'prev' is still NULL. */

	  free(curr);
	}
      else
	prev = curr; /* 'prev' is the current path. */
    }
   
  /* The reconstructured list */
  return gg_list;
}

/*
**  Function  : static struct gg_paths* sort_gg_paths_list
**                             (struct gg_paths* unsorted,
**		                int (*cmp)(struct gg_paths*, struct gg_paths*))
**  Return    : The list sorted using the criterion defined by the function
**              'cmp'.
**  Remarks   : This function order the list 'usorted' using the 'cmp'
**              function as ordering criterion, which depends on the load
**              balancing selected.
**              
*/
static struct gg_paths* sort_gg_paths_list(struct gg_paths* unsorted, 
			      int (*cmp)(struct gg_paths*, struct gg_paths*))
{
  struct gg_paths* sorted = NULL;
  struct gg_paths* unsorted_head; /* the current head of the unsorted part*/
  struct gg_paths* prev_ins_point;/* pointer to the place where to insert*/
  struct gg_paths* ins_point;     /* the new element                     */

  while (unsorted != NULL)
    {
      /* Remove the head from the unsorted list.  */
      unsorted_head = unsorted;
      unsorted = unsorted->next;
      
      /* Scan the sorted list to find the insertion point.  */
      for (prev_ins_point = NULL, ins_point = sorted;
	   /* scan the list until the insertion point is smaller then the */
	   /* unsorted head AND the insertion point is not at the end     */
	   (ins_point) && (cmp(ins_point, unsorted_head) < 0); 
	   /* move to the next element */
	   prev_ins_point = ins_point, ins_point = ins_point->next);
	
      /* Insert the head of the unsorted list in the sorted list.  */
      if (prev_ins_point)
	{
	  unsorted_head->next = prev_ins_point->next;
	  prev_ins_point->next = unsorted_head;
	}
      else
	{
	  unsorted_head->next = sorted;
	  sorted = unsorted_head;
	}
    }

  return sorted;
}

/*
**  Function  : void balance_paths(int     ord, int     algorithm)
**              
**  Return    : None.
**              
**  Remarks   : The main function of the load balancing algoritmh:
**              build the matrix ord, using the ordering algoritmh 'algoritmh'
*/
void balance_paths(int     ord, int     algorithm)
{
  int     s, d;                    /* suorce id and destination id       */ 
  struct gg_paths* gg_list = NULL; /* the list containing all the paths  */
                                   /* between generators                 */
  int     i;
  
  if(static_load == NULL)
    {
     /* create the static load vector and fill it with zero.                */
     static_load = MMmalloc(sizeof(double) * no_of_links, "balance_paths");
     for (i = 0; i < no_of_links; ++i)
       static_load[i] = 0.0;
    }
  if(selected_non_U_priority)
    {
     /* balance all the non-U class paths */ 
     /* first copy the path_matrix into a list of paths.                     */
     for (s = 0; s < no_of_cgens; ++s)
       for(d = 0; d < no_of_cgens; ++d)
         {
	   if(path_mat[ord-2][s][d] &&
		   callgen_vect[s]->choosen_routing.path_order_type == ord)
	     {
	       struct gg_paths* new;
               if(callgen_vect[s]->type != CG_TYPE5)
                 {
	          new = MMmalloc(sizeof(struct gg_paths), "balance_paths");
	          new->src_cgen = s;
	          new->dest_cgen = d;
	          new->path_list = clone_path_list(path_mat[ord-2][s][d]);
	          new->num_eq_paths = count_eq_paths(new->path_list, ord);
	          new->next = gg_list;

			  /* remove the original (not balanced) pathlist */
			  free_pathlist(path_mat[ord-2][s][d]);
			  path_mat[ord-2][s][d]=NULL;

	          gg_list = new;
                 }
	     }
         }
     do_the_balance(ord, algorithm, gg_list);
     /* free the gg_list before rebuild it */
     gg_list = NULL;
     
     /* assign the primary load to the static_load vect before balancing the */
     /* class U */
     for (i = 0; i < no_of_links; ++i)
       static_load[i] = 0.0;
     for (s = 0; s < no_of_cgens; ++s)
       for(d = 0; d < no_of_cgens; ++d)
          if (path_mat[ord][s][d] != NULL)
	      update_static_load(path_mat[ord][s][d],s,d);

     /* now, balance the class U paths                                       */ 
     /* first copy the path_matrix into a list of paths.                     */
     for (s = 0; s < no_of_cgens; ++s)
       for(d = 0; d < no_of_cgens; ++d)
         {
	   if(path_mat[ord-2][s][d] &&
		   callgen_vect[s]->choosen_routing.path_order_type == ord)
	     {
	       struct gg_paths* new;
               if(callgen_vect[s]->type == CG_TYPE5)
                 {
	          new = MMmalloc(sizeof(struct gg_paths), "balance_paths");
	          new->src_cgen = s;
	          new->dest_cgen = d;
	          new->path_list = clone_path_list(path_mat[ord-2][s][d]);
	          new->num_eq_paths = count_eq_paths(new->path_list, ord);
	          new->next = gg_list;

			  /* remove the original (not balanced) pathlist */
			  free_pathlist(path_mat[ord-2][s][d]);
			  path_mat[ord-2][s][d]=NULL;

	          gg_list = new;
                 }
	     }
         }
     do_the_balance(ord, algorithm, gg_list);
     /* free the gg_list before rebuild it */
     gg_list = NULL;
    }
  else /* selected_non_U_priority == FALSE */
    {
     /* balance all the non-U class and the class-U paths together          */ 
     /* first copy the path_matrix into a list of paths.                     */
     for (s = 0; s < no_of_cgens; ++s)
       for(d = 0; d < no_of_cgens; ++d)
         {
	   if(path_mat[ord-2][s][d] &&
		   callgen_vect[s]->choosen_routing.path_order_type == ord)
	     {
	       struct gg_paths* new;
	       new = MMmalloc(sizeof(struct gg_paths), "balance_paths");
	       new->src_cgen = s;
	       new->dest_cgen = d;
	       new->path_list = clone_path_list(path_mat[ord-2][s][d]);
	       new->num_eq_paths = count_eq_paths(new->path_list, ord);
	       new->next = gg_list;

		   /* remove the original (not balanced) pathlist */
		   free_pathlist(path_mat[ord-2][s][d]);
		   path_mat[ord-2][s][d]=NULL;

		   gg_list = new;
	     }
         }
     do_the_balance(ord, algorithm, gg_list);
     /* free the gg_list before rebuild it */
     gg_list = NULL;
   }
  /* free the static load vector */
  free(static_load);
  static_load = NULL;
  whole_pathlist_release();
}

/*
**  Function  : void do_the_balance(int     ord, int     algorithm,
**                                  struct gg_paths*)
**              
**  Return    : None.
**              
**  Remarks   : sort the matrix ord, using the ordering algoritmh 'algoritmh'
*/
void do_the_balance(int     ord, int     algorithm,struct gg_paths* gg_list)
{
  /* Assign all the primary paths.  */
  if ((algorithm == FIRST_NO_OF_EQUIV_PATH)   || 
      (algorithm == MODIFIED_FIRST_NO_OF_HOP))
    {
      int (*ordering_function)(struct gg_paths*, struct gg_paths*);

      if (ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
	 ordering_function = cmp_num_eq_and_hop;
      else
	 ordering_function = cmp_num_eq_and_length;

      /* sort the list of path */
      gg_list = sort_gg_paths_list(gg_list, ordering_function);

      /* assign all the primary path */
      gg_list = assign_paths(gg_list, ord);

#ifdef DEBUG_BALANCING
      /* primary static load only in the case of primary first */
      save_static_load(TRUE, ord);
#endif 
    }

  /* assign all the remaining paths */
  while (gg_list != NULL)
    {
     if (algorithm == FIRST_NO_OF_EQUIV_PATH) 
	{
	   int (*ordering_function)(struct gg_paths*, struct gg_paths*);
	    
	   if (ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
	     ordering_function = cmp_num_eq_and_hop;
	   else
	     ordering_function = cmp_num_eq_and_length;

	   /* sort the list of path */
	   gg_list = sort_gg_paths_list(gg_list, ordering_function);

	   gg_list = assign_paths(gg_list, ord);
	}
      else
	{
	  /* FIRST_NO_OF_HOP */
	  struct gg_paths* prev, * curr, * tail;

	  int (*ordering_function)(struct gg_paths*, struct gg_paths*);

	  if (ord == BALANCED_IN_ORDER_OF_HOP_NUMBER)
	     ordering_function = cmp_num_hops;
	  else
	     ordering_function = cmp_length;
	   
	  /* sort the list of path */
	  gg_list = sort_gg_paths_list(gg_list, ordering_function);
	   
	  /* Split the list in two segments.  The first segment will contains*/
	  /* elements equivalent to the head, the second the remaining of the*/
	  /* list.                                                           */
	  /* find the path with the same number of hops.                     */
	  for (prev = gg_list, curr = gg_list->next; 
	       (curr) &&
	       (equivalent_paths(gg_list->path_list, curr->path_list, ord));
	       prev = curr, curr = curr->next)
	    ;

	  /* 'prev' points to the last element of the first segment.  Since */
	  /* this segment contains at least one element (equivalent to      */
	  /* itself) 'prev' is never NULL.                                  */
	  /* save the remaining part of the list                            */
	  tail = curr;
	  /* cut the list                                                   */
	  prev->next = NULL;

	  /* Now 'gg_list' points to the first segment, 'tail' to the second.  */

	  gg_list = sort_gg_paths_list(gg_list, cmp_num_eq_paths);
	  gg_list = assign_paths(gg_list, ord);

	  /* Join the two segments back in one list pointed by 'gg_list'. */
	  /* for semplicity, add the old part at the end of the list      */
	  if (gg_list)
	    {
	      for (curr = gg_list; curr->next; curr = curr->next)
		;

	      curr->next = tail; 
	    }
	  else
	    gg_list = tail;
	}
    }

#ifdef DEBUG_BALANCING
  /* save the order and static load */
  save_path(ord-2);
  save_path(ord);
  save_static_load(FALSE, ord);
  save_primary_number_call(ord);
#endif
}

