/*
**  Module    : nd.h
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1998/07/30 07:57:43 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module collects declarations and definitions of items
**              referred to by the modules of the ND parser.
**
**  History   : 01-SEP-1995 - Creation of the module.   - GP -
**              01-JUN-1996 - Revision.                 - AF -
**              01-MAR-1997 - Revision			- MM -
**              23-APR-2002 - Routes			- ZZ -
*/


#include "ancles.h"
#include <ctype.h>

struct nd_instance* global_instance_list;

# ifdef MAIN_ND
#define DECL_ND
# else
#define DECL_ND extern
# endif

typedef struct parsing_entity Parse;
typedef struct parse_node     PNode;
typedef struct parse_link     PLink;
typedef struct parse_user     PUser;
typedef struct parse_callgen  PCGen;
typedef struct parse_route    PRoute;
typedef union  ancles_struct   Ancles;
typedef struct { int ident; const char* name; } SymTable;

enum { T_NODE = 1, T_LINK, T_USER, T_CGEN, T_ROUTE };

#define ND_SIBAL_TH -0.9011971  /* Identifier for Sibal-Desimone        */
                                /* threshold evaluation method          */

/* All types are derived from the two base types (integer and
   real). To obtain the base type of type X use the macro
   BASE_TYPE(X).

   Deriving a new type is useful to limit the allowed values to a
   finite set of symbolic constants (one example is the NDT_CAC type
   for the 'selected_cac' parameter), or to allow the input of
   symbolic constants besides that of numeric constants (for example
   the NDT_FIXED_PATH_NUMBER type for the 'path_number' parameter that
   accepts an integer or the constants 'only_primary_path' and
   'all_paths_number' as input).

   To add a new parameter:

     1. If symbolic constants are needed create a new type in the enum
        'nd_parameter_type', deriving it from one of the base types.

     2. Add a new parameter in the 'parameter_table' array. Select the
        context in which the parameter can appear; if appropriate
        define default, minimum and maximum values for the parameter
        (and set the corresponding flags).

     3. If symbolic constants are needed add them in the
        'init_constant_table' array.

     4. Add the checks in the function check_values() if needed (for
	example if the parameter can conflict with the others).
	Simple range checking is done automatically when the right
	fields of the 'nd_parameter' structure are defined (see point
	2).

     5. In the module 'nd_ancles.c' add the code to store the
        parameter value in ancles variables or structures when parsing
        is done (remember that all parsing data is freed before the
        simulation is started).

*/
enum nd_parameter_type
{
  NDT_INTEGER = 0x01,
  NDT_REAL = 0x02,

  NDT_ROUTING_ALGORITHM = 0x10 | NDT_INTEGER,
  NDT_CAC = 0x20 | NDT_INTEGER,
  NDT_LOAD_BALANCING_ALGORITHM = 0x30 | NDT_INTEGER,
  NDT_PATH_ORDER = 0x40 | NDT_INTEGER,
  NDT_SELECTED_RANGE_TYPE = 0x50 | NDT_INTEGER,
  NDT_STATISTICAL = 0x60 | NDT_INTEGER,
  NDT_GENERATOR_TYPE = 0x70 | NDT_INTEGER,
  NDT_ROUTING_THRESHOLD = 0x80 | NDT_REAL,
  NDT_FIXED_PATH_NUMBER = 0x90 | NDT_INTEGER,
  NDT_BOOLEAN = 0xa0 | NDT_INTEGER,
  NDT_LB_HOP_RANGE = 0xb0 | NDT_INTEGER,
  NDT_LB_LENGTH_RANGE = 0xc0 | NDT_REAL,
  NDT_VL_OPEN_ALGORITHM = 0xd0 | NDT_INTEGER,
  NDT_VL_CLOSE_ALGORITHM = 0xe0 | NDT_INTEGER,
  NDT_PRECREATE_VTOPOLOGY = 0xf0 | NDT_INTEGER,
};

#define BASE_TYPE(x) ((x) & 0x0f)

enum nd_parameter_context
{
  NDC_GLOBAL = 0x000001,
  NDC_LINK = 0x000002,
  NDC_NODE = 0x000004,
  NDC_USER = 0x000008,
  NDC_ROUTE = 0x000010,
  NDC_GENERATOR_TYPE1 = 0x000100,
  NDC_GENERATOR_TYPE2 = 0x000200,
  NDC_GENERATOR_TYPE3 = 0x000400,
  NDC_GENERATOR_TYPE4 = 0x000800,
  NDC_GENERATOR_TYPE5 = 0x001000,
  NDC_GENERATOR_TYPE6 = 0x002000,
  NDC_GENERATOR_TYPE7 = 0x004000,
  NDC_GENERATOR_TYPE8 = 0x008000,
  NDC_GENERATOR_TYPE9 = 0x010000,
  NDC_GENERATOR_TYPE10 = 0x020000,
  NDC_GENERATOR_TYPE11 = 0x040000,
  NDC_GENERATOR_TYPE12 = 0x080000,

  /* A parameter whose context mask includes NDC_DUMMY is not pure
     global nor local, and thus does not generate a warning if
     undefined. This is a hack for the 'call_generator_type'
     parameter. */
  NDC_DUMMY = 0x800000,
};

/* To check if X is a generic generator context write

   if (X & NDC_GENERATOR_MASK)
     {
       ...
     }
*/
#define NDC_GENERATOR_MASK (NDC_GENERATOR_TYPE1 |\
			    NDC_GENERATOR_TYPE2 | NDC_GENERATOR_TYPE3 |\
			    NDC_GENERATOR_TYPE4 | NDC_GENERATOR_TYPE5 |\
			    NDC_GENERATOR_TYPE6 | NDC_GENERATOR_TYPE7 |\
			    NDC_GENERATOR_TYPE8 | NDC_GENERATOR_TYPE9 )


/* A pure global parameter can appear only in a global context. */
#define PURE_GLOBAL_PARAMETER(p) (((p)->context & NDC_GLOBAL)\
				  && ! ((p)->context & ~NDC_GLOBAL))

#define LOCAL_PARAMETER(p) (((p)->context & ~NDC_GLOBAL)\
			    && !((p)->context & NDC_DUMMY))

enum nd_parameter_flags
{
  /* Parameter can't appear in a demand statement. Currently used only
     by 'call_generator_type' parameter. */
  NDF_NO_DEMAND = 0x0001,

  /* Parameter has a default value stored in def_(real|integer)
     field. */
  NDF_DEFAULT = 0x0002,

  /* Check if parameter is greater or equal than the value specified
     in min_(real|integer) field. */
  NDF_CHECK_MIN = 0x0004,

  /* Check if parameter is lower or equal than the value specified
     in max_(real|integer) field. */
  NDF_CHECK_MAX = 0x0008,

  /* When parameter is out of range reset it to its default value. */
  NDF_RESET_TO_DEFAULT = 0x0010,

  /* Check if parameter is greater than the value specified in
     min_(real|integer) field. */
  NDF_CHECK_MIN_EXCLUDED = 0x0040,

  /* Check if parameter is lower than the value specified in
     max_(real|integer) field. */
  NDF_CHECK_MAX_EXCLUDED = 0x0080,

  /* Parameter accepts only symbolic constants in a finite set. */
  NDF_ENUMERATION = 0x0100,

  /* Parameter is a vector. */
  NDF_VECTOR = 0x0200,
};

/* Check both minimum and maximum. */
#define NDF_CHECK_RANGE (NDF_CHECK_MIN|NDF_CHECK_MAX)
#define NDF_CHECK_RANGE_EXCLUDED (NDF_CHECK_MIN_EXCLUDED|NDF_CHECK_MAX_EXCLUDED)

/* Size of freshly created vectors. */
#define ND_VECTOR_INIT_MAX_LEN 8

/* Used when an initializer for real vector parameters is needed. */
struct nd_real_vector
{
  long len;
  double element[ND_VECTOR_INIT_MAX_LEN];
};

union nd_value
{
  long integer;
  double real;
  struct nd_vector* vector;
};

struct nd_vector
{
  long len;
  long max_len;
  union nd_value element[ND_VECTOR_INIT_MAX_LEN];
};

/* Each couple integer-real should be a nd_value union, but unions
   cannot be statically initialized. */
struct nd_parameter
{
  const char* name;
  const char* description;
  enum nd_parameter_type type;
  long context;
  short flags;

  long def_integer;
  double def_real;

  long min_integer;
  double min_real;

  long max_integer;
  double max_real;

  struct nd_real_vector* def_real_vector;

  /* When this parameter is a vector, `size_parameter' is the integer
     parameter that define its size.  It can be NULL, in which case
     the size of the vector is not predefined. */
  const char* size_parameter;
};

/* Initializers for the previous structure. */
#define INTEGER_VALUE(i) i, 0.0
#define REAL_VALUE(r) 0, r
#define UNDEFINED_VALUE 0, 0.0

struct nd_constant
{
  const char* name;
  enum nd_parameter_type type;
  union nd_value value;
  struct nd_constant* next;
};

enum nd_instance_status
{
  NDS_UNDEFINED,
  NDS_ASSIGNED,
  NDS_DEMAND,
};

struct nd_instance
{
  struct nd_parameter* parameter;
  union nd_value value;
  enum nd_instance_status status;
  struct nd_instance* next;
};

/* Definitions used in the ND_TRAFFIC module.   */
#define NORMAL_LINE  1  /* The line contains a destination traffic      */
                        /* fraction.                                    */
#define SOURCE_LINE  2  /* The line contains a source traffic fraction. */
#define COMMENT_LINE 4  /* The line contains a remark.                  */
#define ERROR_LINE   8  /* The line contains no valid data.             */

union ancles_struct                 /* Union representing the possible  */
  {                                 /* ANCLES entities.                 */
   PNode  *node;
   PLink  *link;
   PUser  *user;
   PCGen  *gener;
   PRoute *route;
  };

struct parsing_entity   /* General structure for the lists of parsed    */
  {                     /* ANCLES entities.                             */
   Parse *next;
   int   id;            /* Unique identifier for the parsed entity.     */
   int   type;          /* Type of parsed entity.                       */
   int   decl_line;     /* Line at which the entity declaration started.*/
   struct nd_instance* instance_list;
   Ancles elem;         /* Pointer (in the union) to the entity         */
                        /* structure.                                   */
  };

struct parse_node       /* Structure for the Node Entity.               */
  {
   int   id;
   int   port_no;
   int   user_no;
   int   connected_user[MAX_USER];
   int   connected_node[MAX_PORT];
   int   from_user[MAX_USER];
   int   to_user[MAX_USER];
   int   inlink[MAX_PORT];
   int   outlink[MAX_PORT];
  };

struct parse_link       /* Structure for the Link Entity.               */
  {
   int       id;
   int       source;
   int       destination;
   Boolean   assigned;
  };

struct parse_user       /* Structure for the User Entity.               */
  {
   int       id;
   int       connected_node;
   int       link_id;
   int       link_from_node;
  };

struct parse_callgen    /* Structure for the Call Generator Entity.     */
  {
   int       id;
   int       type;
   int       refer_user;
  };

struct parse_route       /* Structure for the Route Entity.               */
  {
   int   id;
   int   from_node;
   int   to_node;
   int   contained_link[MAX_LINK];
   int	 link_no;
  };


DECL_ND SymTable node_symbol_table[MAX_NODE],
                 link_symbol_table[MAX_LINK],
                 user_symbol_table[MAX_NET_USER],
                 callgen_symbol_table[MAX_CALL_GEN],
                 route_symbol_table[MAX_ROUTE];

extern  int     linenum;

/*
**  Prototypes of the functions used in the modules forming the ND language.
**  The interface functions are prototyped in ANCLES.H
*/

/*  Module : nd_ancles.c                */
Link    *link_by_id(int);
User    *user_by_id(int);
void 	initialize_ANCLES_structures(void);
void    init_real_links(void);
void    init_real_users(void);
void    init_real_call_gens(void);
void    init_real_nodes(void);
void    init_real_routes(void);
void    init_variables(void);

/*  Module : nd_connect.c       */
Parse   *get_element_by_id(Parse *, int);
void    build_node_network(void);
void    add_users_to_nodes(void);
void    add_call_gens_to_users(void);

/*  Module : nd_init.c          */
Parse   *init_parsed_node(int);
Parse   *init_parsed_link(int);
Parse   *init_parsed_user(int);
Parse   *init_parsed_cgen(int);
Parse   *init_parsed_route(int);

/*  Module : nd_insert.c        */
double  get_extended_input(void);

/*  Module : nd_service.c       */
Boolean in_list_parse(Parse **, Parse *);
int     insert_symbol(SymTable *, char *, int);
struct nd_vector* create_vector(void);
struct nd_vector* insert_vector_element(struct nd_vector *, long, union nd_value );
struct nd_vector* clone_real_vector(const struct nd_real_vector* vect);
struct nd_vector* clone_vector(const struct nd_vector* vect);
void vector_integer_to_real(struct nd_vector* vector);
int     lookup_symbol(SymTable *, char *, int);
long int     set_cgen_context(enum cg_type);
void    add_connection(Parse *, int, int);
void    add_usage(Parse *, int, int);
void    add_referring(Parse *, int);
void    add_link_to_route(Parse *, int);
void    add_fn_to_route(Parse *, int);
void    add_tn_to_route(Parse *, int);

/*  Module nd_traffic.c         */
int     parse_line_traffic(char *, char *, double *);

/* Module nd_parameters.c */
void nd_print_debug(void);
struct nd_parameter* lookup_parameter_by_name(const char*);
struct nd_parameter* lookup_parameter_by_type(enum nd_parameter_type);
struct nd_constant* lookup_constant_by_name(const char*);
struct nd_instance* assign_parameter(struct nd_instance* , struct nd_parameter* ,
				     union nd_value , long );
struct nd_instance* create_default_instances(long );
Boolean assign_demand(struct nd_instance* , struct nd_parameter* );
struct nd_instance* lookup_instance_by_name(struct nd_instance* , const char* );
long integer_instance_value(struct nd_instance* , const char* );
double real_instance_value(struct nd_instance* , const char* );
struct nd_vector* real_vector_instance_value(struct nd_instance* , const char* );
Boolean check_values(void);
Boolean check_missing_parameters(void);
void ask_for_demand(void);
void init_constants(void);
void init_parameters(void);
void print_error(Parse*, const char*, ...);
void print_warning(Parse*, const char*, ...);
