/*
**  Module    : nd_connect.c
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : This module contains the functions needed to create and check
**              the actual connections among the ND-ANCLES entities during 
**              the initialization phase.
**
**  Functions : Parse    *get_element_by_id(Parse *,int);
**              void     build_node_network(void);
**              void     add_users_to_nodes(void);
**              void     add_call_gens_to_users(void);
**

01-02-2001 Cleaning... -ZZ-
*/
 
#define MAIN_ND
 
#include "nd.h"
 
extern Parse *global_node_list,
             *global_link_list,
             *global_user_list,
             *global_cgen_list,
             *global_route_list;

extern int pnode_num,
           plink_num,
           puser_num,
           pcgen_num,
		   proute_num;

Boolean autolink = FALSE;
 
 
/*
**  Function  : Parse *get_element_by_id(Parse *list, int ident)
**  Return    : the pointer to the 'Parse' element identified by 'ident',
**              NULL if no element is found.
**  Remarks   : none.
*/
Parse *get_element_by_id(Parse *list, int ident)
{
 while(list != NULL)
    {
     if(list->id == ident) break;
     list = list->next;
    }
 return list;
}
 
 
/*
**  Function  : void build_node_network(void)
**  Return    : none.
**  Remarks   : the node network is built. The node connections
**              are checked and the link identifiers in each node are set
**              for both the directions, creating undefined links if the
**              `autolink' flag is set.
*/
void build_node_network(void)
{
	Parse   *p, *head, *tmp;
	PNode   *knot, *tonk;
	PLink   *linkone, *linktwo;
	int     i, j, k;
	int     dest_node;
	int     link_knot, link_tonk;
	Boolean connection_decl, connection_done;
	
	if(init_debug_level & ND_DEBUG_LEVEL)
		printf("\n --- NODES CONNECTIONS ---\n\n");
	
	head = global_node_list;
	for(i = 0; i < pnode_num; i++)
	{
		knot = head->elem.node;
		if(knot->port_no > MAX_PORT)
		{
			printf("Error:  Too many ports declared for node %s \n",node_name(i));
			exit(1);
		}
		for(j = 0; j < knot->port_no; j++)
		{
			dest_node = knot->connected_node[j];
			link_knot = knot->outlink[j];
			p = get_element_by_id(global_node_list, dest_node);
			if(p == NULL)
			{
				printf("Error:  Node %s - Destination node %s not declared\n",
					node_name(i), node_name(dest_node));
				exit(1);
			}
			tmp = get_element_by_id(global_link_list, link_knot);
			if(tmp != NULL)
				linkone = tmp->elem.link;
			else
			{
				if(autolink)
				{
					if(init_debug_level & ND_DEBUG_LEVEL)
						printf("  Automatic declaration for link %s\n",
						link_name(link_knot));
					tmp = init_parsed_link(link_knot);
					++plink_num;
					in_list_parse(&global_link_list, tmp);
					linkone = tmp->elem.link;
				}
				else
				{
					printf("Error:  Node %s - Link %s is not declared\n",
						node_name(i), link_name(link_knot));
					exit(1);
				}
			}
			if(linkone->assigned) continue; /* If the current link has been */
			/* already assigned no further  */
			/* actions are needed.          */
			tonk = p->elem.node;
			connection_decl = FALSE;
			connection_done = FALSE;
			/* If the link is not duplex looking for a pair link,
			which is starting from the other node */  
			
			if(!integer_instance_value(tmp->instance_list, "duplex_optic"))
			{
				for(k = 0; k < tonk->port_no; k++)
				{
					if(tonk->connected_node[k] != i) continue;
					connection_decl = TRUE;
					link_tonk = tonk->outlink[k];
					if(link_knot == link_tonk)
					{
						printf(
							"Error:  Node %s and node %s are connected by the same link %s\n",
							node_name(dest_node), node_name(i), link_name(link_knot));
						exit(1);
					}
					tmp = get_element_by_id(global_link_list, link_tonk);
					if(tmp != NULL)
						linktwo = tmp->elem.link;
					else
					{
						if(autolink)
						{
							if(init_debug_level & ND_DEBUG_LEVEL)
								printf("  Automatic declaration for link %s\n",
								link_name(link_tonk));
							tmp = init_parsed_link(link_tonk);
							++plink_num;
							in_list_parse(&global_link_list, tmp);
							linktwo = tmp->elem.link;
						}
						else
						{
							printf("Error:  Node %s - Link %s is not declared\n",
								node_name(i), link_name(link_tonk));
							exit(1);
						}
					}
					if(linktwo->assigned) continue;
					
					/* Checking whether the two links have the same parameter values.       */
					/*  if(linkone->timescale.value != linktwo->timescale.value ||
					linkone->link_len.value  != linktwo->link_len.value)   continue;   */
					/* Probably the two links are not matching,  */
					/* so the next link is considered.           */
					
					/* The two links are not completely assigned: the whole assignment is   */
					/* done all the same even if, perhaps, some test is redundant, and the  */
					/* only assignment of source and destination could be enough.           */
					if(linkone->source != NOT_USED_ID && linkone->source != i)
					{
						printf(
							"Error:  Node %s - Link %s has yet been assigned to node %s\n",
							node_name(i),link_name(link_knot),node_name(linkone->source));
						exit(1);
					}
					else
						linkone->source = i;
					if(linkone->destination != NOT_USED_ID &&
						linkone->destination != dest_node)
					{
						printf(
							"Error:  Node %s - Link %s has yet been assigned to node %s\n",
							node_name(i), link_name(link_knot),
							node_name(linkone->destination));
						exit(1);
					}
					else
						linkone->destination = dest_node;
					if(linktwo->source != NOT_USED_ID && linktwo->source != dest_node)
					{
						printf(
							"Error:  Node %s - Link %s has yet been assigned to node %s\n",
							node_name(i), link_name(link_tonk),
							node_name(linktwo->source));
						exit(1);
					}
					else
						linktwo->source = dest_node;
					if(linktwo->destination != NOT_USED_ID && linktwo->destination != i)
					{
						printf(
							"Error:  Node %s - Link %s has yet been assigned to node %s\n",
							node_name(i), link_name(link_tonk),
							node_name(linktwo->destination));
						exit(1);
					}
					else
						linktwo->destination = i;
					
					connection_done = TRUE;
					if(init_debug_level & ND_DEBUG_LEVEL)
						printf("Node %s -> %s with link %s <-> %s \n",
						node_name(i), node_name(dest_node),
						link_name(link_knot), link_name(link_tonk));
					linkone->assigned = TRUE;
					linktwo->assigned = TRUE;
					knot->inlink[j] = link_tonk;
					tonk->inlink[k] = link_knot;
					break;
				}
			}
			else
			{
				/* The duplex optic links are referred from both side of the connection */
				connection_decl = TRUE;
				if(linkone->source != NOT_USED_ID && linkone->source != i)
				{
					printf(
						"Error:  Node %s - Link %s has yet been assigned to node %s\n",
						node_name(i),link_name(link_knot),node_name(linkone->source));
					exit(1);
				}
				else
					linkone->source = i;
				if(linkone->destination != NOT_USED_ID &&
					linkone->destination != dest_node)
				{
					printf(
						"Error:  Node %s - Link %s has yet been assigned to node %s\n",
						node_name(i), link_name(link_knot),
						node_name(linkone->destination));
					exit(1);
				}
				else
					linkone->destination = dest_node;
				linkone->assigned = TRUE;
				knot->inlink[j] = link_knot;
				/* Assign new connected node and link to the destination node */
				tonk->connected_node[tonk->port_no] = i;
				tonk->outlink[tonk->port_no] = link_knot;
				tonk->inlink[tonk->port_no++] = link_knot;
				connection_done = TRUE;
			}
			
			if(!connection_decl)
			{
				printf("Error:  Destination node %s is not connected to node %s\n",
					node_name(i), node_name(dest_node));
				exit(1);
			}
			if(!connection_done)
			{
				printf("Error:  Link %s cannot connect node %s and node %s\n",
					link_name(link_knot), node_name(i), node_name(dest_node));
				exit(1);
			}
			
		}	
		head = head->next;
   }
}

 
/*
**  Function  : void add_users_to_nodes(void)
**  Return    : none.
**  Remarks   : the connection with the corresponding node is initialized 
**              for each user, setting the link identifiers in both the
**              directions and creating undefined links if the `autolink'
**              flag is set.
*/
void add_users_to_nodes(void)
{
 Parse *p, *head, *tmp;
 PNode *knot;
 PUser *user;
 PLink *link, *link2;
 int k;
 int node_id, knil;
 Boolean connection_decl, connection_done;
 struct nd_instance* inst, * inst2;
 
 if(init_debug_level & ND_DEBUG_LEVEL)
    printf("\n --- USERS CONNECTIONS ---\n\n");
 
 for(head = global_user_list; head != NULL; head = head->next)
   {
    user = head->elem.user;
    node_id = user->connected_node;
    p = get_element_by_id(global_node_list, node_id);
    if(p == NULL)
      {
       printf("Error:  User %s - Destination node %s not declared\n",
                     user_name(user->id), node_name(node_id));
       exit(1);
      }
    knot = p->elem.node;
 
    tmp = get_element_by_id(global_link_list,user->link_id);
    if(tmp != NULL)
      {
       link = tmp->elem.link;
       inst = tmp->instance_list;
      }
    else
      {
       if(autolink)
         {
          if(init_debug_level & ND_DEBUG_LEVEL)
              printf("  Automatic declaration for link %s\n",
                 link_name(user->link_id));
          tmp = init_parsed_link(user->link_id);
          ++plink_num;
          in_list_parse(&global_link_list, tmp);
          link = tmp->elem.link;
	  inst = tmp->instance_list;
         }
       else
         {
          printf("Error:  User %s - Link %s is not declared\n",
                   user_name(user->id), link_name(user->link_id));
          exit(1);
         }
      }
    if(link->assigned)
      {
       printf("Error:  User %s - Link %s has yet been assigned\n",
                   user_name(user->id), link_name(user->link_id));
       exit(1);
      }
 
    connection_decl = FALSE;
    connection_done = FALSE;
    for(k = 0; k < knot->user_no; k++)
      {
       if(knot->connected_user[k] != user->id) continue;
       connection_decl = TRUE;
       knil = knot->to_user[k];
       if(knil == user->link_id)
        {
         printf(
          "Error:  User %s and node %s are connected by the same link %s\n",
                   user_name(user->id), node_name(node_id), link_name(knil));
         exit(1);
        }
       tmp = get_element_by_id(global_link_list,knil);
       if(tmp != NULL)
	 {
	   link2 = tmp->elem.link;
	   inst2 = tmp->instance_list;
	 }
       else
        {
         if(autolink)
          {
           if(init_debug_level & ND_DEBUG_LEVEL)
                printf("  Automatic declaration for link %s\n",
                   link_name(knil));
           tmp = init_parsed_link(knil);
           ++plink_num;
           in_list_parse(&global_link_list, tmp);
           link2 = tmp->elem.link;
	   inst2 = tmp->instance_list;
          }
         else
          {
           printf("Error:  Node %s - Link %s is not declared\n",
                     node_name(node_id), link_name(knil));
           exit(1);
          }
        }
       if(link2->assigned) continue;
 
/* Checking the relation between the values of the parameters related   */
/* to the two links.                                                    */
      if(real_instance_value(inst, "link_timescale") 
	 != real_instance_value(inst2, "link_timescale")
	 || real_instance_value(inst, "link_length") 
	 != real_instance_value(inst2, "link_length") )
                continue;  /* This action is similar to that done for   */
                           /* the node-node links but, perhaps, is not  */
                           /* so useful, because a user can be          */
                           /* connected to one node only.               */
 
      if(link->source != NOT_USED_ID && link->source != user->id)
        {
         printf("Error:  User %s - Link %s has yet been assigned to user %s\n",
                 user_name(user->id),
                 link_name(user->link_id), user_name(link->source));
         exit(1);
        }
      else
         link->source = user->id;
      if(link->destination != NOT_USED_ID && link->destination != node_id)
        {
         printf("Error:  User %s - Link %s has yet been assigned to node %s\n",
                 user_name(user->id),
                 link_name(user->link_id), node_name(link->destination));
         exit(1);
        }
      else
         link->destination = node_id;
      link->assigned  = TRUE;
 
      link2->source = node_id;
      link2->destination = user->id;
      link2->assigned  = TRUE;
 
      knot->from_user[k] = user->link_id;
      user->link_from_node = knil;
      connection_done = TRUE;
      if(init_debug_level & ND_DEBUG_LEVEL)
        printf("User %s -> Node %s with link %s <-> %s \n",
                user_name(user->id), node_name(node_id),
                link_name(user->link_id), link_name(knil));
      break;
     }
    if(!connection_decl)
      {
       printf("Error:  Destination user %s is not connected to node %s\n",
               user_name(user->id), node_name(node_id));
       exit(1);
      }
    if (!connection_done)
      {
       printf("Error:  Link %s cannot connect user %s and node %s\n",
               link_name(user->link_id), user_name(user->id),
               node_name(node_id));
       exit(1);
      }
   }
}
 
 
/*
**  Function  : void add_call_gens_to_users(void)
**  Return    : none.
**  Remarks   : this function checks whether each call generator refers to a
**              declared user and each user owns at least one call generator.
*/
void add_call_gens_to_users(void )
{
 Parse          *head,*tmp;
 PCGen          *callgen;
 Boolean        referred_user[MAX_NET_USER];
 int            i;
 
 for(head = global_cgen_list; head != NULL; head = head->next)
   {
    callgen = head->elem.gener;
    tmp = get_element_by_id(global_user_list, callgen->refer_user);
    if(tmp == NULL)
      {
       printf(
        "Error:  Call Generator %s - Reference User %s is not declared\n",
                cgen_name(callgen->id), user_name(callgen->refer_user));
       exit(1);
      }
   }
 
/* The vector of referred users is initialized here.    */
 for(i = 0; i < puser_num; i++)
        referred_user[i] = FALSE;
 
 for(head = global_cgen_list; head != NULL; head = head->next)
   {
    callgen = head->elem.gener;
    referred_user[callgen->refer_user] = TRUE;
   }
 
/* Testing if any user has no referring call generator. */
 if(pcgen_num == 0)
   {
    printf("Error: No Call Generators declared\n");
    exit(1);
   }
 if(pcgen_num == 1)
   {
    printf("Error: Only one Call Generator declared\n");
    exit(1);
   }
 for(i = 0; i < puser_num; i++)
   if(referred_user[i] == FALSE)
     printf ("\nWarning:  User %s has not any call generator\n\n",
           user_name(i));
}
