/*
   Module: nd_parameters.c
   Version: $Revision: 3.4 $
   Date: $Date: 1999/01/11 14:53:50 $
*/
/*Modifications:	DLCP algoritmi - Filippo Natale
			Updating link properties - Zoltan Zsoka
			MSF algoritmi - Zoltan Zsoka
			Ucall data amount - Zoltan Zsoka
	1999/08/
01-02-2001 Cleaning... -ZZ-
**              01-APR-23 - Insert the features for ASON:
**                          ASON call generators and duplex fibres - ZZ -
*/

#include "ancles.h"
#include "nd.h"

extern Parse* global_route_list;
extern Parse* global_cgen_list;
extern Parse* global_user_list;
extern Parse* global_link_list;
extern Parse* global_node_list;
extern int pnode_num, plink_num, puser_num, pcgen_num, proute_num;
extern struct nd_real_vector vector_list;

/* Local functions */

static Boolean check_cg_values(Parse *entity);
static Boolean check_path_values(Parse *entity);

static struct nd_instance*
assign_undefined_parameter(struct nd_instance *instance_list,
			   struct nd_parameter *parameter,
			   union nd_value value, long context);
static Boolean
check_instances_range(Parse* entity,
		      struct nd_instance* instance_list,
		      enum nd_parameter_context context);
static Boolean
check_unassigned_instances(Parse *node,
			   struct nd_instance *instance_list);
static struct nd_constant *
create_integer_constant(const char *name,
			enum nd_parameter_type type, long value);

static struct nd_constant *
create_real_constant(const char *name,
		     enum nd_parameter_type type, double value);

static void
ask_instances(struct nd_instance *instance, Parse *entity);

static void
assign_local_undefined_instances(struct nd_parameter *parameter,
				 union nd_value value);

static void
print_instances(struct nd_instance *list, long status);

static const char *
identify_entity(Parse *entity, Boolean verbose);

static void
print_value(struct nd_parameter*, union nd_value);

static const char*
get_constant_name(enum nd_parameter_type type,
		  union nd_value value);

/* Instances list fo global parameters. */
struct nd_instance* global_instance_list;

/* Default value for the 'routing_threshold[]' parameter. */
static struct nd_real_vector default_routing_threshold =
{
  2, { 0.0, ND_SIBAL_TH }
};

static struct nd_parameter parameter_table[] =
{
  /*
  ** Link parameters
  */
  { "link_timescale",
    "Timescale of the Link",
    NDT_REAL, NDC_GLOBAL | NDC_LINK, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  { "link_histo_steps",
    "Number of Intervals for the Link Load Histogram",
    NDT_INTEGER, NDC_GLOBAL | NDC_LINK, NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
  },

  { "link_length",
    "Length of the Link in Km",
    NDT_REAL, NDC_GLOBAL | NDC_LINK, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "capacity",
    "Base Link Capacity",
    NDT_REAL, NDC_GLOBAL, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "bw_update_period",
    "Official Bandwidth Updaing Time of the Link",
    NDT_REAL, NDC_GLOBAL | NDC_LINK, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  { "virtlink_bw_update_period",
    "Official Bandwidth Updaing Time of the Virtual Links",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  { "bw_measure_uncertainty",
    "Uncertainty of the Official Bandwidth Measure in Percent",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_RESET_TO_DEFAULT |
    NDF_CHECK_MIN | NDF_CHECK_MAX,
    REAL_VALUE(5.0),
    REAL_VALUE(0.0),
    REAL_VALUE(100.0),
  },


  { "link_is_substituted",
    "The Link is Substituted on the Paths",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_LINK,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  /* Parameters for the ASON support on links */
  { "duplex_optic",
    "The Link is a Duplex Fibre",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_LINK,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "fiber_capacity",
    "Wavelength capacity of a fiber on the link.",
    NDT_INTEGER, NDC_GLOBAL | NDC_LINK,
	NDF_DEFAULT | NDF_RESET_TO_DEFAULT |
	NDF_CHECK_MIN |	NDF_CHECK_MAX,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
    INTEGER_VALUE(8*MAX_WVOCTETS),
  },

  { "wavelength_capacity",
    "Bitrate of a Wavelength (capacity -> wavelength).",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN_EXCLUDED,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

/* IP over WDM (Gancles) support */
  { "enable_IPoWDM",
    "Enable the IP over WDM simulation.",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_LINK,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  /*
  ** Node parameters
  */
  { "alpha_node",
    "Weighting coefficient for aggregate cost",
    NDT_REAL, NDC_GLOBAL | NDC_NODE, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  { "beta_node",
    "Weighting coefficient for aggregate costs\n"
    "\t\t\t\t propagation",
    NDT_REAL, NDC_GLOBAL | NDC_NODE,NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.1),
    REAL_VALUE(0.0),
  },

  { "node_is_opaque",
    "Opaque functionality of node in ASON network\n",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_NODE,NDF_DEFAULT |
    NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "node_is_unreachable",
    "If true, all the links connected to the node will be cut off\n",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_NODE,NDF_DEFAULT |
    NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  /*
  ** Statistical parameters
  */
  { "accuracy",
    "Simulation Accuracy",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                          NDF_RESET_TO_DEFAULT,
    REAL_VALUE(0.02),
    REAL_VALUE(0.0),
  },

  { "confid_level",
    "Confidence Level",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(0.95),
    REAL_VALUE(0.0),
  },

  { "trans_accuracy",
    "Accuracy Level required to close the Transient",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(0.99),
    REAL_VALUE(0.0),
  },

  { "num_observation",
    "Maximum Number of Observations",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                             NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(10000000),
    INTEGER_VALUE(0),
  },

  { "min_trans_len",
    "Minimum Number of Observations during the\n"
    "\t\t\t\t Transient",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                             NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(5000),
    INTEGER_VALUE(0),
  },

  { "trans_batch_size",
    "Size of the Batch used to detect the Transient\n"
    "\t\t\t\t Duration",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                             NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(30),
    INTEGER_VALUE(0),
  },

  { "trans_seq_len",
    "Minimum Number of Batches during the Transient",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                             NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(3),
    INTEGER_VALUE(0),
  },

  { "stat_batch_size",
    "Initial Size of a Batch during the Stationary\n"
    "\t\t\t\t Analysis",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
                             NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(10000),
    INTEGER_VALUE(0),
  },

  { "total_blocking_probability",
    "Total Blocking Probability",
    NDT_STATISTICAL, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_SIGNIFICANT),
  },

  { "total_carried_load",
    "Total Carried Load",
    NDT_STATISTICAL, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(SIGNIFICANT),
  },

  { "used_wavelengths",
    "Used Wavelengths (ASONcles)",
    NDT_STATISTICAL, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_SIGNIFICANT),
  },

  { "user_blocking_probability",
    "Blocking Probability for a specific User",
    NDT_STATISTICAL, NDC_GLOBAL | NDC_USER, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_ANALYSED),
  },

  { "classU_user_assigned_bw",
    "Mean assigned bw per U call for a specific User",
    NDT_STATISTICAL, NDC_GLOBAL | NDC_USER, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_ANALYSED),
  },

  { "link_carried_load",
    "Load carried by a specific link",
    NDT_STATISTICAL, NDC_GLOBAL | NDC_LINK, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_ANALYSED),
  },

  { "link_used_wavelengths",
    "Used Wavelengths on a Link (ASONcles)",
    NDT_STATISTICAL, NDC_GLOBAL | NDC_LINK, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NOT_ANALYSED),
  },


  { "time_based_ct5",
    "The duration of CallType5 calls is based on time\n"
    "\t\t\t\tinstaed of the data amount.",
    NDT_BOOLEAN, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "general_load_red_factor",
    "Factor for the general reduction of load\n"
    "\t\t\t\tafter the transient phase",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
    NDF_CHECK_MAX | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  /*
  ** Interrupt statistics
  */

  { "interrupt_statistics",
    "Enables the statistics of the interrupted calls",
    NDT_BOOLEAN, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "lhstep_factor",
    "Factor for the calculation of the steplength\n"
    "\t\t\t\tof the length-histogram",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED |
    NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  /*
  ** Theoretical analysis
  */
  { "theo_analysis",
    "Enables the theoretical analysis",
    NDT_BOOLEAN, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  
  /*
  ** Listing the available node-node routes
  */
  { "list_routes",
    "Enables the listing of the routes",
    NDT_BOOLEAN, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  
  /*
  ** Routing parameters
  */
  { "load_balancing_algorithm",
    "Using balanced_primary_first or \n"
    "\t\t\t\t balance_shortest_path",
    NDT_LOAD_BALANCING_ALGORITHM, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FIRST_NO_OF_HOP),
  },

  { "non_u_priority",
    "If his flag is TRUE, balance first the non-U\n"
    "\t\t\t\t class",
    NDT_BOOLEAN, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "range_of_equiv_length",
    "The difference between length to consider\n"
    "\t\t\t\t two paths equivalent",
    NDT_LB_LENGTH_RANGE, NDC_GLOBAL,
    NDF_DEFAULT| NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  { "range_of_equiv_hop",
    "The difference in hops to consider two paths\n"
    "\t\t\t\t equivalent",
    NDT_LB_HOP_RANGE, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
  },

  /* RA200 */
  { "recomputation_slot_duration",
    "Time [secs] elapsed between \n"
    "\t\t\t\t two successive aggregate costs\n"
    "\t\t\t\t computations",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED,
    REAL_VALUE(60.0),
    REAL_VALUE(0.0),
  },

/****************************************************************/

/* These two parameters were copied from the file:
    /home7/mellia/Ancles/gabriele/nd_parameters.c !!!    */

  { "path_number_hop_range", "commento",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_MASK, NDF_DEFAULT,
    INTEGER_VALUE(INT_MAX),
  },

  { "path_number_length_range", "commento",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_MASK, NDF_DEFAULT,
    REAL_VALUE(DBL_MAX),
  },

/****************************************************************/

  /* Global constraints for paths ranges	*/

  { "xxx_max_no_of_hops",
    "Global constraint of absolute hop number in paths",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(MAX_NODE+1),
    INTEGER_VALUE(0),
  },

  { "xxx_more_no_of_hops",
    "Global constraint of relative hop number in paths",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(MAX_NODE+1),
    INTEGER_VALUE(0),
  },


  /* New Load-Dependent method Periodical Algorithm Choice */
  { "algorithm_choice_period",
    "Period length of the updating of algorithm choice",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  { "pac_sample_deviation",
    "Filternig factor of the old samples in the mean value\n"
    "\t\t\tin PAC routing",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_CHECK_MIN | NDF_CHECK_MAX_EXCLUDED,
    REAL_VALUE(0.99),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "alg_decision_threshold",
    "Threshold of decision in choice of algorithm SP/MD,\n"
    "\t\t\tif negative -> rather SP",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_CHECK_MIN_EXCLUDED | NDF_CHECK_MAX_EXCLUDED,
    REAL_VALUE(0.0),
    REAL_VALUE(-1.0),
    REAL_VALUE(1.0),
  },

  { "histeresys_threshold",
    "Threshold of histeresys in choice of algorithm SP/MD",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_CHECK_MIN | NDF_CHECK_MAX_EXCLUDED,
    REAL_VALUE(0.01),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "congestion_threshold",
    "Threshold of the congestion detection on a link",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT |
    NDF_CHECK_MIN | NDF_CHECK_MAX_EXCLUDED,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  /* RA6 */
  { "widest_bottleneck_range",
    "Difference in Mbps to consider two paths equivalent\n"
    "\t\t\t\t in the shortest-widest RA",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  /* RA5 */
  { "overload_threshold",
    "Threshold used by the load-dependent RA",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.75),
    REAL_VALUE(0.0),
  },

  /* RA4 */
  { "hop_weight",
    "Hop weight used by RA4 based on path cost\n"
    "\t\t\t\t evaluation",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  { "length_weight",
    "Length weight used by RA4 based on path cost\n"
    "\t\t\t\t evaluation",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  { "congestion_weight",
    "Congestion weight for RA4 path cost evaluation",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  /* RA2 */
  { "primary_threshold",
    "Index of primary link threshold in the\n"
    "\t\t\t\t routing_threshold[] vector",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_MIN | NDF_CHECK_MAX_EXCLUDED,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
    INTEGER_VALUE(0), /* Defined at run time */
  },

  { "alternate_threshold",
    "Index of alternate link threshold in the\n"
    "\t\t\t\t routing_threshold[] vector",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_MIN | NDF_CHECK_MAX_EXCLUDED,
    INTEGER_VALUE(1),
    INTEGER_VALUE(0),
    INTEGER_VALUE(0), /* Defined at run time */
  },

  { "no_of_threshold",
    "Number of possible thresholds",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_RANGE | NDF_NO_DEMAND,
    INTEGER_VALUE(2),
    INTEGER_VALUE(1),
    INTEGER_VALUE(MAX_THRESHOLD),
  },

  { "routing_threshold[]",
    "Vector defining how to use the link thresholds",
    NDT_ROUTING_THRESHOLD, NDC_GLOBAL | NDC_LINK,
    NDF_DEFAULT | NDF_VECTOR | NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
    UNDEFINED_VALUE,
    &default_routing_threshold,
    "no_of_threshold"
  },

  /*
  ** CAC parameters
  */
  { "selected_cac",
    "Selected Call Admission Control Algorithm",
    NDT_CAC, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(CALL_ADM_CTRL1),
  },

  /* CAC 1 */
  { "cell_loss_rate",
    "Cell Loss Rate for Equivalent Bandwidth\n"
    "\t\t\t\t evaluation",
    NDT_REAL, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_RANGE_EXCLUDED | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0e-9),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  /* CAC 2 and 3*/
  { "no_load_samples",
    "Number of Samples for real time traffic\n"
    "\t\t\t\t estimation",
    NDT_INTEGER, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_RANGE | NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(100),
    INTEGER_VALUE(1),
    INTEGER_VALUE(1000),
  },

  { "meas_wind_size",
    "Number of Link Slots in a Sample",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_RANGE | NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(100),
    INTEGER_VALUE(100),
    INTEGER_VALUE(1000),
  },

  { "k_star",
    "Coefficient weighing the load samples std.\n"
    "\t\t\t\t deviation",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_RANGE | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(8.5),
    REAL_VALUE(0.0),
    REAL_VALUE(1000.0),
  },

  { "prot_fact",
    "Protection Factor applied to the Link Capacity",
    NDT_REAL, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED | NDF_CHECK_MAX | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "meas_rel_error",
   "pdf parameter for the errors on the Load\n"
   "\t\t\t\t Measures",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_RANGE | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  /* CAC 4 */
  { "alpha",
   "Class 1/2 calls protection factor",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT,
    REAL_VALUE(1.0),
  },

  { "beta",
    "Class U calls protection factor",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT,
    REAL_VALUE(1.0),
  },

  { "gamma",
    "Non-U Class modification factor",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT,
    REAL_VALUE(1.0),
  },
		
  /*
  ** Photos parameters
  */
  { "rel_load_ths",
    "Relative Load Threshold used to take Network\n"
    "\t\t\t\t State Snapshots",
    NDT_REAL, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED | NDF_CHECK_MAX | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "cong_duration",
    "Congestion Duration used to take Network State\n"
    "\t\t\t\t Snapshots",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    REAL_VALUE(1.0),
    REAL_VALUE(0.0),
  },

  { "max_state_photos",
    "Maximum Number of allowed Network State Photos",
    NDT_INTEGER, NDC_GLOBAL,
    NDF_DEFAULT | NDF_CHECK_RANGE | NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
    INTEGER_VALUE(5),
  },

  { "min_Ucall_present",
    "Minimun Number of Ucall present on a link\n"
    "\t\t\t\t to take a Snapshots",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN | NDF_RESET_TO_DEFAULT,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
  },

  /*
  ** Call generator parameters
  */
  { "call_generator_type",
    "Type of the Call Generator (TYPE1, TYPE2, etc.)",
    NDT_GENERATOR_TYPE, NDC_GLOBAL | NDC_DUMMY, NDF_NO_DEMAND | NDF_ENUMERATION,
    UNDEFINED_VALUE,
  }, /* no def */

  { "is_nonstationary",
    "TRUE if CG generates call in a non-stationary\n"
    "\t\t\t\t fashion",
    NDT_BOOLEAN, NDC_GLOBAL | NDC_GENERATOR_MASK, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_ENUMERATION, INTEGER_VALUE(FALSE),
  },

  { "load_reduction_factor",
    "factor by which the load is to be reduced when\n"
    "\t\t\t\t in low-load state",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_RANGE_EXCLUDED,
    REAL_VALUE(0.75),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "normal_load_duration",
    "average duration of normal load state",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED,
    REAL_VALUE(43200),
    REAL_VALUE(0.0),
  },

  { "low_load_duration",
    "average duration of low load state",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_MIN_EXCLUDED,
    REAL_VALUE(43200),
    REAL_VALUE(0.0),
  },

  { "selected_routing",
    "Selected Routing Algorithm",
    NDT_ROUTING_ALGORITHM, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(ROUTING_RA1),
  },

  { "path_order_type",
    "by Balanced (or not) Hop Number or by\n"
    "\t\t\t\t Balanced (or not) Length",
    NDT_PATH_ORDER, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(IN_ORDER_OF_HOP_NUMBER),
  },

  { "selected_range_type",
    "select range type of considered path\n",
    NDT_SELECTED_RANGE_TYPE, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_ENUMERATION | NDF_DEFAULT,
    INTEGER_VALUE(FIXED_PATH_NUMBER),
  },

  { "path_number_range",
    "The maximum number of paths to be considered\n"
    "\t\t\t\t when fixed_path_number is selected",
    NDT_FIXED_PATH_NUMBER, NDC_GLOBAL | NDC_GENERATOR_MASK,
    NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(INT_MAX),
    INTEGER_VALUE(1),
  },

  { "path_hop_range",
    "The number of hops to be considered according\n"
    "\t\t\t\t to selected_range_type criterion",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_MASK, NDF_DEFAULT,
    INTEGER_VALUE(INT_MAX),
  },

  { "path_length_range",
    "The length of the paths to be considered \n"
    "\t\t\t\t according to selected_range_type criterion",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_MASK, NDF_DEFAULT,
    REAL_VALUE(DBL_MAX),
  },

  { "use_hop_by_hop",
    "Use hop by hop alternate routing.",
    NDT_BOOLEAN,NDC_GLOBAL | NDC_GENERATOR_MASK,NDF_DEFAULT|NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "use_crankback",
    "Use hop by hop alternate\n"
    "\t\t\t\t routing with crankback.",
    NDT_BOOLEAN,NDC_GLOBAL | NDC_GENERATOR_MASK,NDF_DEFAULT|NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },


  { "cg1_min_bit_rate",
    "Minimum Bit Rate for CBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE1, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg1_max_bit_rate",
    "Maximum Bit Rate for CBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE1, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg1_mean_hold_time",
    "Mean Holding Time for CBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE1, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg2_min_peak_rate",
    "Minimum Peak Rate for On-Off VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE2, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg2_max_peak_rate",
    "Maximum Peak Rate for On-Off VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE2, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg2_mean_on_time",
    "Mean On Time for On-Off VBR Sources",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE2, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg2_mean_off_time",
    "Mean Off Time for On-Off VBR Sources",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE2, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg2_mean_hold_time",
    "Mean Holding Time for On-Off VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE2, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },


  { "cg3_min_bit_rate_avg",
    "Average Minimum Bit Rate for Uniform VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE3, 0,
    UNDEFINED_VALUE,
  },

  { "cg3_max_bit_rate_avg",
    "Average Naximum Bit Rate for Uniform VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE3, 0,
    UNDEFINED_VALUE,
  },

  { "cg3_half_bw_interval",
    "Half Bandwidth of uniform pdf Bit Rate for\n"
    "\t\t\t\t Uniform VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE3, 0,
    UNDEFINED_VALUE,
  },

  { "cg3_mean_hold_time",
    "Mean Holding Time for Uniform VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE3, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },


  { "cg4_lower_bit_rate_avg",
    "Average Lower Bit Rate for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, 0,
    UNDEFINED_VALUE,
  },

  { "cg4_medium_bit_rate_avg",
    "Average Medium Bit Rate for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, 0,
    UNDEFINED_VALUE,
  },

  { "cg4_higher_bit_rate_avg",
    "Average Higher Bit Rate for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, 0,
    UNDEFINED_VALUE,
  },

  { "cg4_lower_bit_rate_prob",
    "Lower Bit Rate Probability for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg4_medium_bit_rate_prob",
    "Medium Bit Rate Probability for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg4_higher_bit_rate_prob",
    "Higher Bit Rate Probability for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg4_half_bw_interval",
    "Half Bandwidth of uniform pdf Bit Rate for\n"
    "\t\t\t\t Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, 0,
    UNDEFINED_VALUE,
  },

  { "cg4_mean_hold_time",
    "Mean Holding Time for Video VBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE4, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg5_min_max_rate",
    "Minimum MAX Bit Rate for UBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE5, NDF_CHECK_MIN,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg5_max_max_rate",
    "Maximum MAX Bit Rate for UBR Calls",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE5, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg5_mean_gen_traffic",
    "Mean Generated Traffic for UBR Calls in Mbits",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE5, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "min_accept_bw_thrs",
    "Threshold for Minimal Acceptable Bandwidth of \n"
    "\t\t\t\ta UBR Call (else Interrupt if enabled)",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE5, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN | NDF_CHECK_MAX,
    REAL_VALUE(0.1),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "cg6_min_req_waves",
    "Minimum Number of Requested Wavelengths for an ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE6, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(1),
    INTEGER_VALUE(1),
  },

  { "cg6_max_req_waves",
    "Maximum Number of Requested Wavelengths for an ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE6, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(1),
    INTEGER_VALUE(1),
  },

  { "cg6_mean_hold_time",
    "Mean Holding Time for ASON Calls.",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE6, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg7_saturation",
    "Saturation of ASON calls.",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE7, NDF_DEFAULT |
	NDF_CHECK_MIN |
    NDF_CHECK_MAX,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "cg7_req_waves",
    "Number of Requested Wavelengths for an On-Off ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE7, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
  },

  { "cg7_mean_hold_time",
    "Mean Holding Time for an On-Off ASON Call.",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE7, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },

  { "cg8_req_waves",
    "Number of Requested Wavelengths for a Never Closing ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE8, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(0),
    INTEGER_VALUE(0),
  },


  { "cg9_min_req_waves",
    "Minimum Number of Requested Wavelengths for a Pas-ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE9, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(1),
    INTEGER_VALUE(1),
  },

  { "cg9_max_req_waves",
    "Maximum Number of Requested Wavelengths for a Pas-ASON Call.",
    NDT_INTEGER, NDC_GLOBAL | NDC_GENERATOR_TYPE9, NDF_DEFAULT |
    NDF_CHECK_MIN,
    INTEGER_VALUE(1),
    INTEGER_VALUE(1),
  },

  { "cg9_mean_hold_time",
    "Mean Holding Time for Pascal ASON Calls.",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE9, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(0.0),
  },
	
  { "cg9_peakedness",
    "Peakedness of distribution for Pascal ASON Calls (Z>1).",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE9, NDF_CHECK_MIN_EXCLUDED,
    UNDEFINED_VALUE,
    REAL_VALUE(1.0),
  },
/* Manageing the virtual IP links */
  { "vl_open_algorithm",
    "Strategy of opening virtual IP links",
    NDT_VL_OPEN_ALGORITHM, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(DIRECT_VLINK),
  },

  { "vl_close_algorithm",
    "Strategy of closing virtual IP links",
    NDT_VL_CLOSE_ALGORITHM, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(EMPTY_CLOSE),
  },

  { "vl_close_hold_time",
    "Needed holding time of the empty state of link before close it",
    NDT_REAL, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    REAL_VALUE(0.0),
    REAL_VALUE(0.0),
  },

  { "max_no_of_virt_routes",
    "Global constraint of target number of virtual link based paths",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(10000),
    INTEGER_VALUE(0),
  },

  { "max_no_of_virt_nn_routes",
    "Global constraint of target number of virtual link based\n"
	"node-node paths",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT | NDF_CHECK_MIN,
    INTEGER_VALUE(10000),
    INTEGER_VALUE(0),
  },

  /* We may tune this functionality, but the current 
 solution os a global parameter */
  { "duplex_lowerlevel_conn",
    "The lower level connections are duplex",
    NDT_BOOLEAN, NDC_GLOBAL,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "min_vl_open_thrs",
    "Threshold for Ucall Bandwidth for open a new virtual link",
    NDT_REAL, NDC_GLOBAL | NDC_GENERATOR_TYPE5, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN | NDF_CHECK_MAX,
    REAL_VALUE(0.1),
    REAL_VALUE(0.0),
    REAL_VALUE(1.0),
  },

  { "max_hop_vl_open",
    "Maximal allowed hopnumber for not open a new virtual link",
    NDT_INTEGER, NDC_GLOBAL, NDF_DEFAULT |
    NDF_RESET_TO_DEFAULT | NDF_CHECK_MIN | NDF_CHECK_MAX,
    INTEGER_VALUE(MAX_NODE),
    INTEGER_VALUE(0),
    INTEGER_VALUE(MAX_NODE),
  },

  { "ena_close_predef_virtlinks",
    "Enable to close the predefined virtual links",
    NDT_BOOLEAN, NDC_GLOBAL,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "ena_hopdep_vl_hold_time",
    "Enable the dependence of hopnumber for vlclose_holdtime",
    NDT_BOOLEAN, NDC_GLOBAL,
    NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(FALSE),
  },

  { "precreate_virtual_topology",
    "Perform the creation of a virtual topology at the start",
    NDT_PRECREATE_VTOPOLOGY, NDC_GLOBAL, NDF_DEFAULT | NDF_ENUMERATION,
    INTEGER_VALUE(NONE),
  },

};
#define NUM_PARAMETERS (sizeof(parameter_table) / sizeof(struct nd_parameter))

/* Check the validity of all parameters. This includes range checking
   for all instances (global and local) plus some checks depending on
   the instance type (some CG parameters are mutually exclusive, some
   routing algorithms work only for class U call generators etc.).
   Return TRUE if all values are valid.

   Note that when this function is called global parameters are
   already stored in ancles variables, while local parameters still
   reside in instances lists. */
Boolean check_values(void)
{
  Boolean all_ok = TRUE;
  Parse* node;

  /* Dinamically set the upper bound. */
  lookup_parameter_by_name("primary_threshold")->max_integer = no_of_thresholds;
  lookup_parameter_by_name("alternate_threshold")->max_integer = no_of_thresholds;

  /* Check the range of global parameters. It's easier to consider
     them as list members, even if they are already stored in ancles
     variables. */
  if (check_instances_range(NULL, global_instance_list, NDC_GLOBAL))
    all_ok = FALSE;

  /*
  ** Call generator parameters
  */
  for (node = global_cgen_list; node; node = node->next)
    {
      /* Skip sinks */
      if (set_cgen_context(node->elem.gener->type) & NDC_GENERATOR_MASK)
	{
	  if (check_instances_range(node, node->instance_list,
				    set_cgen_context(node->elem.gener->type)))
	    all_ok = FALSE;

	  if (! check_cg_values(node))
	    all_ok = FALSE;

	  if (! check_path_values(node))
	    all_ok = FALSE;
	}
    }

  /*
  ** Link parameters
  */
  for (node = global_link_list; node; node = node->next)
   {
     struct nd_vector* routing_threshold;

     /* The upper bound for 'primary_threshold' and
	'alternate_threshold' parameters was set dinamically to
	'no_of_thresholds' and is checked automatically with the other
	parameters here. */
     if (check_instances_range(node, node->instance_list, NDC_LINK))
       all_ok = FALSE;

     routing_threshold =
       real_vector_instance_value(node->instance_list, "routing_threshold[]");

     /* Check the size of the thresholds vector. */
     if (routing_threshold->len != no_of_thresholds)
       {
	 all_ok = FALSE;

	 print_error(node,
		     "the number of assigned thresholds [%d] "
		     "is not equal to the choosen number of sets [%ld]\n",
		     routing_threshold->len, no_of_thresholds);
       }
     /* Check if there are any links with non-integer timescale */
     /* and CLASS snapshots are required                        */
     if( integer_instance_value(global_instance_list, "max_state_photos") > 0)
     {
        double real_timescale;
        real_timescale = real_instance_value(node->instance_list,
                        "link_timescale");
        if( floor(real_timescale) != real_timescale)
        {
           all_ok = FALSE;
           print_error(node,
		     "not-integer timescale [%f] selected while "
                     "max_state_photos parameter is greater than zero\n",
		     real_timescale);
        }
     }

   }

  /*
  ** Node parameters
  */
  for (node = global_node_list; node; node = node->next)
    if (check_instances_range(node, node->instance_list, NDC_NODE))
      all_ok = FALSE;

  /*
  ** User parameters
  */
  for (node = global_user_list; node; node = node->next)
    if (check_instances_range(node, node->instance_list, NDC_USER))
      all_ok = FALSE;

  /* The presence of a Type 5 Call Generator is checked here. */
  for (node = global_cgen_list; node; node = node->next)
    {
      if (node->elem.gener->type == CG_TYPE5
	  && selected_cac != CALL_ADM_CTRL4
	  && selected_cac != CALL_ADM_CTRL1)
	{
	  all_ok = FALSE;
	
	  print_error(NULL,
             "The presence of a Class U Call Generator requires either CAC1 or CAC4");

	  break;
	}
   }

  return all_ok;
}

/* Check the values defining the bit rate of call generator ENTITY. Return
   TRUE if all values are valid. */
static Boolean
check_cg_values(Parse* entity)
{
  struct nd_instance* il = entity->instance_list;
  Boolean all_ok = TRUE;

  switch (entity->elem.gener->type)
    {
    case CG_TYPE1:
      if (real_instance_value(il, "cg1_min_bit_rate")
	  > real_instance_value(il, "cg1_max_bit_rate"))
	{
	  print_error(entity, "Min bit rate is greater than max bit rate");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE2:
      if (real_instance_value(il, "cg2_min_peak_rate")
	  > real_instance_value(il, "cg2_max_peak_rate"))
	{
	  print_error(entity, "Min peak rate is greater than max peak rate");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE3:
      {
	double min_bit_rate_avg, max_bit_rate_avg, half_bw_interval;

	min_bit_rate_avg = real_instance_value(il, "cg3_min_bit_rate_avg");
	max_bit_rate_avg = real_instance_value(il, "cg3_max_bit_rate_avg");
	half_bw_interval = real_instance_value(il, "cg3_half_bw_interval");

	if (min_bit_rate_avg < half_bw_interval)
	  {
	    print_error(entity,"Min bit rate can result negative");
	    all_ok = FALSE;
	  }

	if (max_bit_rate_avg <= half_bw_interval)
	  {
	    print_error(entity, "Min bit rate can result non positive");
	    all_ok = FALSE;
	  }

	if (min_bit_rate_avg + half_bw_interval
	    > max_bit_rate_avg - half_bw_interval)
	  {
	    print_error(entity, "Min bit rate can result greater than max bit rate");
	    all_ok = FALSE;
	  }
      }
      break;

    case CG_TYPE4:
      {
	double lower_bit_rate_avg, medium_bit_rate_avg,
	  higher_bit_rate_avg, half_bw_interval;

	lower_bit_rate_avg = real_instance_value(il, "cg4_lower_bit_rate_avg");
	medium_bit_rate_avg = real_instance_value(il, "cg4_medium_bit_rate_avg");
	higher_bit_rate_avg = real_instance_value(il, "cg4_higher_bit_rate_avg");
	half_bw_interval = real_instance_value(il, "cg4_half_bw_interval");

	if (lower_bit_rate_avg < half_bw_interval)
	  {
	    print_error(entity, "Lower bit rate can result negative");
	    all_ok = FALSE;
	  }

	if (medium_bit_rate_avg <= half_bw_interval)
	  {
	    print_error(entity, "Medium bit rate can result non positive");
	    all_ok = FALSE;
	  }

	if (higher_bit_rate_avg <= half_bw_interval)
	  {
	    print_error(entity, "Higher bit rate can result non positive");
	    all_ok = FALSE;
	  }

	if (!( (lower_bit_rate_avg + half_bw_interval
		< medium_bit_rate_avg - half_bw_interval)
	       && (medium_bit_rate_avg + half_bw_interval
		   < higher_bit_rate_avg - half_bw_interval) ))
	  {
	    print_error(entity,
			"Invalid bit rate values or "
			"uniform pdf interval value too big");
	    all_ok = FALSE;
	  }

	if (fabs( real_instance_value(il, "cg4_lower_bit_rate_prob")
		  + real_instance_value(il, "cg4_medium_bit_rate_prob")
		  + real_instance_value(il, "cg4_higher_bit_rate_prob")
		  - 1.0)
	    > PRECISION)
	  {
	    print_error(entity, "Total bit rate probability must sum to 1.0");
	    all_ok = FALSE;
	  }
      }
      break;

    case CG_TYPE5:
      if (real_instance_value(il, "cg5_min_max_rate")
	  > real_instance_value(il, "cg5_max_max_rate"))
	{
	  print_error(entity, "Min max rate is greater than max max rate");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE6:
      if (integer_instance_value(il, "cg6_min_req_waves")
	  > integer_instance_value(il, "cg6_max_req_waves"))
	{
	  print_error(entity, "Min req waves is greater than max req waves");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE7:
      if ((integer_instance_value(il, "cg7_req_waves") == 0) == 
             (fabs(real_instance_value(il, "cg7_saturation")) < PRECISION))
	{
	  print_error(entity, "Exactly one of the req_waves and the saturation parameters has to be 0!");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE8:
/* because of virtual links the 0 wave load is possible! 
     if (!(integer_instance_value(il, "cg8_req_waves") > 0))
	{
	  print_error(entity, "Req waves is not greater than 0");
	  all_ok = FALSE;
	}*/
      break;

    case CG_TYPE9:
      if (integer_instance_value(il, "cg9_min_req_waves")
	  > integer_instance_value(il, "cg9_max_req_waves"))
	{
	  print_error(entity, "Min req waves is greater than max req waves");
	  all_ok = FALSE;
	}
      break;

    case CG_TYPE10 :
    case CG_TYPE11 :
    case CG_TYPE12 :
      break;
    }

  return all_ok;
}

/* Check the compatibility of parameters, owned by ENTITY, defining
   paths number and order. Return TRUE if all values are valid. */
static Boolean
check_path_values(Parse* entity)
{
  long number, order, routing, value;
  struct nd_instance* il = entity->instance_list;
  Boolean all_ok = TRUE;

  /* Extract from the instances list the values we are about to
     check. */
  number = integer_instance_value(il, "selected_range_type");
  value = integer_instance_value(il, "path_number_range");
  order = integer_instance_value(il, "path_order_type");
  routing = integer_instance_value(il, "selected_routing");

  /* The paths can be limited by the number of hops only if they are
     ordered by hop. (XXX: in balanced lists with range not equal to
     zero in fact the limiting algorithm doesn't work, since the
     ordering by hop is not strictly respected.) */
  if (number == MAX_HOP_MORE_PATH_NUMBER
      && (order == IN_ORDER_OF_LENGTH || order == BALANCED_IN_ORDER_OF_LENGTH))
    {
      print_error(entity,
		  "Path Order Type and Path Number Choice "
		  "have been assigned conflicting values");
      all_ok = FALSE;
    }

  if (number == MAX_HOP_PATH_NUMBER
      && (order == IN_ORDER_OF_LENGTH || order == BALANCED_IN_ORDER_OF_LENGTH))
    {
      print_error(entity,
		  "Path Order Type and Path Number Choice "
		  "have been assigned conflicting values");
      all_ok = FALSE;
    }

  /* The paths can be limited by the length only if they are ordered
     by length. */
  if (number == MAX_LENGTH_MORE_PATH_NUMBER
      && (order == IN_ORDER_OF_HOP_NUMBER
	  || order == BALANCED_IN_ORDER_OF_HOP_NUMBER))
    {
      print_error(entity,
		  "Path Order Type and Path Number Choice "
		  "have been assigned conflicting values");
      all_ok = FALSE;
    }

  if (number == MAX_LENGTH_PATH_NUMBER
      && (order == IN_ORDER_OF_HOP_NUMBER
	  || order == BALANCED_IN_ORDER_OF_HOP_NUMBER))
    {
      print_error(entity,
		  "Path Order Type and Path Number Choice "
		  "have been assigned conflicting values");
      all_ok = FALSE;
    }

  /* Routing algorithms other than RA1 and RA50 require more than one path in
     the list to work. */
  if (number == FIXED_PATH_NUMBER && value == 1
      && (routing == ROUTING_RA2 || routing == ROUTING_RA3 ||
	  routing == ROUTING_RA4 || routing == ROUTING_RA5 ||
	  routing == ROUTING_RA6 || routing == ROUTING_RA7 ||
	  routing == ROUTING_RA8 || routing == ROUTING_RA9 ||
	  routing == ROUTING_RA10 || routing == ROUTING_RA11 ||
	  routing == ROUTING_RA12 || routing == ROUTING_RA51 ||
	  routing == ROUTING_RA52 || routing == ROUTING_RA53 ||
	  routing == ROUTING_RA62 || routing == ROUTING_RA63 ||
	  routing == ROUTING_RA73 || routing == ROUTING_RA75 ||
	  routing == ROUTING_RA77 || routing == ROUTING_RA79 ||
	  routing == ROUTING_RA83 || routing == ROUTING_RA87 ||
	  routing == ROUTING_RA92 || routing == ROUTING_RA93 ||
	  routing == ROUTING_RA123 || routing == ROUTING_RA125 ||
	  routing == ROUTING_RA127 || routing == ROUTING_RA129 ||
	  routing == ROUTING_RA133 || routing == ROUTING_RA137 ||
	  routing == ROUTING_RA200 || routing == ROUTING_RA201 ))
    {
      print_warning(entity,
		  "The selected Routing Algorithm "
		  "needs a number of available paths greater than one");
    }

  return all_ok;
}

/*
** Initialization of constants used as keywords
** for all parameter declarations
**
** Remarks: if there is a synonym, the second declaration is
**          printed for the demand request.
*/

struct init_constant
{
  const char* name;
  enum nd_parameter_type type;
  long integer;
  double real;
}
init_constant_table[] =
{

  /* selected_routing parameters */
  { "routing_ra1", NDT_ROUTING_ALGORITHM, ROUTING_RA1, 0.0 },
  { "single_path_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA1, 0.0 },

  { "routing_ra2", NDT_ROUTING_ALGORITHM, ROUTING_RA2, 0.0 },
  { "controlled_alternate_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA2, 0.0 },

  { "routing_ra3", NDT_ROUTING_ALGORITHM, ROUTING_RA3, 0.0 },
  { "uncontrolled_alternate_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA3, 0.0 },

  { "routing_ra4", NDT_ROUTING_ALGORITHM, ROUTING_RA4, 0.0 },
  { "minimum_cost_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA4, 0.0 },

  { "routing_ra5", NDT_ROUTING_ALGORITHM, ROUTING_RA5, 0.0 },
  { "load_dependent_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA5, 0.0 },

  { "routing_ra6", NDT_ROUTING_ALGORITHM, ROUTING_RA6, 0.0 },
  { "shortest_widest_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA6, 0.0 },

  { "routing_ra7", NDT_ROUTING_ALGORITHM, ROUTING_RA7, 0.0 },
  { "minimum_distance_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA7, 0.0 },

  { "routing_ra8", NDT_ROUTING_ALGORITHM, ROUTING_RA8, 0.0 },
  { "multimetric_sequential_filtering", NDT_ROUTING_ALGORITHM, ROUTING_RA8, 0.0 },

  { "routing_ra9", NDT_ROUTING_ALGORITHM, ROUTING_RA9, 0.0 },
  { "periodical_ac_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA9, 0.0 },

  { "routing_ra10", NDT_ROUTING_ALGORITHM, ROUTING_RA10, 0.0 },
  { "mindist_cd_apriori_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA10, 0.0 },

  { "routing_ra11", NDT_ROUTING_ALGORITHM, ROUTING_RA11, 0.0 },
  { "mindist_cd_aposteori_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA11, 0.0 },

  { "routing_ra12", NDT_ROUTING_ALGORITHM, ROUTING_RA12, 0.0 },
  { "mincost_cd_apriori_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA12, 0.0 },

  { "routing_ra50", NDT_ROUTING_ALGORITHM, ROUTING_RA50, 0.0 },
  { "ASON_fixed_shortest_path_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA50, 0.0 },

  { "routing_ra51", NDT_ROUTING_ALGORITHM, ROUTING_RA51, 0.0 },
  { "ASON_shortest_lowest_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA51, 0.0 },

  { "routing_ra52", NDT_ROUTING_ALGORITHM, ROUTING_RA52, 0.0 },
  { "ASON_shortest_widest_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA52, 0.0 },

  { "routing_ra53", NDT_ROUTING_ALGORITHM, ROUTING_RA53, 0.0 },
  { "ASON_alternate_shortest_path_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA53, 0.0 },

  { "routing_ra60", NDT_ROUTING_ALGORITHM, ROUTING_RA60, 0.0 },
  { "ASON_fixed_shortest_path_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA60, 0.0 },

  { "routing_ra62", NDT_ROUTING_ALGORITHM, ROUTING_RA62, 0.0 },
  { "ASON_shortest_widest_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA62, 0.0 },

  { "routing_ra63", NDT_ROUTING_ALGORITHM, ROUTING_RA63, 0.0 },
  { "ASON_alternate_shortest_path_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA63, 0.0 },

  { "routing_ra70", NDT_ROUTING_ALGORITHM, ROUTING_RA70, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA70, 0.0 },

  { "routing_ra73", NDT_ROUTING_ALGORITHM, ROUTING_RA73, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA73, 0.0 },

  { "routing_ra74", NDT_ROUTING_ALGORITHM, ROUTING_RA74, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_ewl_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA74, 0.0 },

  { "routing_ra75", NDT_ROUTING_ALGORITHM, ROUTING_RA75, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_ewl_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA75, 0.0 },

  { "routing_ra76", NDT_ROUTING_ALGORITHM, ROUTING_RA76, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_pbe_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA76, 0.0 },

  { "routing_ra77", NDT_ROUTING_ALGORITHM, ROUTING_RA77, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_pbe_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA77, 0.0 },

  { "routing_ra78", NDT_ROUTING_ALGORITHM, ROUTING_RA78, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_ewl_pbe_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA78, 0.0 },

  { "routing_ra79", NDT_ROUTING_ALGORITHM, ROUTING_RA79, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_ewl_pbe_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA79, 0.0 },

  { "routing_ra80", NDT_ROUTING_ALGORITHM, ROUTING_RA80, 0.0 },
  { "ASONn:1_fixed_shortest_pathpair_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA80, 0.0 },

  { "routing_ra83", NDT_ROUTING_ALGORITHM, ROUTING_RA83, 0.0 },
  { "ASONn:1_alternate_shortest_pathpair_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA83, 0.0 },

  { "routing_ra87", NDT_ROUTING_ALGORITHM, ROUTING_RA87, 0.0 },
  { "ASONn:1_fixed_shortest_pathpair_pbe_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA87, 0.0 },

  { "routing_ra90", NDT_ROUTING_ALGORITHM, ROUTING_RA90, 0.0 },
  { "ASON_multifiber_fixed_shortest_path_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA90, 0.0 },
  
  { "routing_ra92", NDT_ROUTING_ALGORITHM, ROUTING_RA92, 0.0 },
  { "ASON_multifiber_shortest_widest_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA92, 0.0 },
  
  { "routing_ra93", NDT_ROUTING_ALGORITHM, ROUTING_RA93, 0.0 },
  { "ASON_multifiber_alternate_shortest_path_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA93, 0.0 },

  { "routing_ra120", NDT_ROUTING_ALGORITHM, ROUTING_RA120, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA120, 0.0 },

  { "routing_ra123", NDT_ROUTING_ALGORITHM, ROUTING_RA123, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA123, 0.0 },

  { "routing_ra124", NDT_ROUTING_ALGORITHM, ROUTING_RA124, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_ewl_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA124, 0.0 },

  { "routing_ra125", NDT_ROUTING_ALGORITHM, ROUTING_RA125, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_ewl_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA125, 0.0 },

  { "routing_ra126", NDT_ROUTING_ALGORITHM, ROUTING_RA126, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_pbe_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA126, 0.0 },

  { "routing_ra127", NDT_ROUTING_ALGORITHM, ROUTING_RA127, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_pbe_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA127, 0.0 },

  { "routing_ra128", NDT_ROUTING_ALGORITHM, ROUTING_RA128, 0.0 },
  { "ASON1+1_fixed_shortest_pathpair_ewl_pbe_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA128, 0.0 },

  { "routing_ra129", NDT_ROUTING_ALGORITHM, ROUTING_RA129, 0.0 },
  { "ASON1+1_alternate_shortest_pathpair_ewl_pbe_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA129, 0.0 },

  { "routing_ra133", NDT_ROUTING_ALGORITHM, ROUTING_RA133, 0.0 },
  { "ASONn:1_alternate_shortest_pathpair_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA133, 0.0 },

  { "routing_ra137", NDT_ROUTING_ALGORITHM, ROUTING_RA137, 0.0 },
  { "ASONn:1_fixed_shortest_pathpair_pbe_rndwa_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA137, 0.0 },
  
  { "routing_ra200", NDT_ROUTING_ALGORITHM, ROUTING_RA200, 0.0 },
  { "controlled_dlcp_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA200, 0.0 },

  { "routing_ra201", NDT_ROUTING_ALGORITHM, ROUTING_RA201, 0.0 },
  { "uncontrolled_dlcp_routing", NDT_ROUTING_ALGORITHM, ROUTING_RA201, 0.0 },

  /* path_hop_range parameters */

  { "only_primary_path", NDT_FIXED_PATH_NUMBER, 1, 0.0 },
  { "all_paths_number", NDT_FIXED_PATH_NUMBER, INT_MAX, 0.0 },

  /* selected_range_type parameters */

  { "fixed_path_number", NDT_SELECTED_RANGE_TYPE, FIXED_PATH_NUMBER, 0.0 },
  { "max_hop_path_number", NDT_SELECTED_RANGE_TYPE, MAX_HOP_PATH_NUMBER, 0.0 },
  { "max_hop_more_path_number", NDT_SELECTED_RANGE_TYPE, MAX_HOP_MORE_PATH_NUMBER, 0.0 },
  { "max_length_path_number", NDT_SELECTED_RANGE_TYPE, MAX_LENGTH_PATH_NUMBER, 0.0 },
  { "max_length_more_path_number", NDT_SELECTED_RANGE_TYPE, MAX_LENGTH_MORE_PATH_NUMBER, 0.0 },

  /* path_order_type parameters */

  { "paths_in_order_of_hop_number", NDT_PATH_ORDER, IN_ORDER_OF_HOP_NUMBER, 0.0 },
  { "paths_in_order_of_length", NDT_PATH_ORDER, IN_ORDER_OF_LENGTH, 0.0 },
  { "balanced_paths_in_order_of_hop_number", NDT_PATH_ORDER, BALANCED_IN_ORDER_OF_HOP_NUMBER, 0.0 },
  { "balanced_paths_in_order_of_length", NDT_PATH_ORDER, BALANCED_IN_ORDER_OF_LENGTH, 0.0 },

  /* load_balancing_algorithm parameters */

  { "first_no_of_equiv_path", NDT_LOAD_BALANCING_ALGORITHM, FIRST_NO_OF_EQUIV_PATH, 0.0 },
  { "balanced_primary_first", NDT_LOAD_BALANCING_ALGORITHM, FIRST_NO_OF_EQUIV_PATH, 0.0 },

  { "first_no_of_hop", NDT_LOAD_BALANCING_ALGORITHM, FIRST_NO_OF_HOP, 0.0 },
  { "balanced_shortest_path", NDT_LOAD_BALANCING_ALGORITHM, FIRST_NO_OF_HOP, 0.0 },

  { "modified_first_no_of_hop", NDT_LOAD_BALANCING_ALGORITHM, MODIFIED_FIRST_NO_OF_HOP, 0.0 },
  { "balance_primary_then_shortest", NDT_LOAD_BALANCING_ALGORITHM, MODIFIED_FIRST_NO_OF_HOP, 0.0 },

  /* range_of_equiv_hop parameter */

  { "all_hop", NDT_LB_HOP_RANGE, INT_MAX, 0.0 },

  /* range_of_equiv_length parameter */

  { "all_length", NDT_LB_LENGTH_RANGE, 0, DBL_MAX },

  /* selected_cac parameters */

  { "call_adm_ctrl1", NDT_CAC, CALL_ADM_CTRL1, 0.0 },
  { "equival_bw_alloc_cac", NDT_CAC, CALL_ADM_CTRL1, 0.0 },

  { "call_adm_ctrl2", NDT_CAC, CALL_ADM_CTRL2, 0.0 },
  { "real_time_traffic_estim_cac", NDT_CAC, CALL_ADM_CTRL2, 0.0 },

  { "call_adm_ctrl3", NDT_CAC, CALL_ADM_CTRL3, 0.0 },
  { "fast_traffic_estim_cac", NDT_CAC, CALL_ADM_CTRL3, 0.0 },

  { "call_adm_ctrl4", NDT_CAC, CALL_ADM_CTRL4, 0.0 },
  { "mixed_traffic_cac", NDT_CAC, CALL_ADM_CTRL4, 0.0 },

  { "call_adm_ctrl5", NDT_CAC, CALL_ADM_CTRL5, 0.0 },
  { "ason_traffic_simple_cac", NDT_CAC, CALL_ADM_CTRL5, 0.0 },

  /* call_generator_type parameters */

  { "type1", NDT_GENERATOR_TYPE, CG_TYPE1, 0.0 },
  { "cbr", NDT_GENERATOR_TYPE, CG_TYPE1, 0.0 },

  { "type2", NDT_GENERATOR_TYPE, CG_TYPE2, 0.0 },
  { "on_off_vbr", NDT_GENERATOR_TYPE, CG_TYPE2, 0.0 },

  { "type3", NDT_GENERATOR_TYPE, CG_TYPE3, 0.0 },
  { "uniform_vbr", NDT_GENERATOR_TYPE, CG_TYPE3, 0.0 },

  { "type4", NDT_GENERATOR_TYPE, CG_TYPE4, 0.0 },
  { "video_vbr", NDT_GENERATOR_TYPE, CG_TYPE4, 0.0 },

  { "type5", NDT_GENERATOR_TYPE, CG_TYPE5, 0.0 },
  { "unspecified", NDT_GENERATOR_TYPE, CG_TYPE5, 0.0 },

  { "type6", NDT_GENERATOR_TYPE, CG_TYPE6, 0.0 },
  { "ason", NDT_GENERATOR_TYPE, CG_TYPE6, 0.0 },

  { "type7", NDT_GENERATOR_TYPE, CG_TYPE7, 0.0 },
  { "on_off_ason", NDT_GENERATOR_TYPE, CG_TYPE7, 0.0 },

  { "type8", NDT_GENERATOR_TYPE, CG_TYPE8, 0.0 },
  { "never_closing_ason", NDT_GENERATOR_TYPE, CG_TYPE8, 0.0 },

  { "type9", NDT_GENERATOR_TYPE, CG_TYPE9, 0.0 },
  { "pascal_ason", NDT_GENERATOR_TYPE, CG_TYPE9, 0.0 },

  { "type10", NDT_GENERATOR_TYPE, CG_TYPE10, 0.0 },
  { "sink", NDT_GENERATOR_TYPE, CG_TYPE10, 0.0 },

  { "type11", NDT_GENERATOR_TYPE, CG_TYPE11, 0.0 },
  { "tcp_sink", NDT_GENERATOR_TYPE, CG_TYPE11, 0.0 },

  { "type12", NDT_GENERATOR_TYPE, CG_TYPE12, 0.0 },
  { "ason_sink", NDT_GENERATOR_TYPE, CG_TYPE12, 0.0 },

  /* statistical parameters */

  { "not_significant", NDT_STATISTICAL, NOT_SIGNIFICANT, 0.0 },
  { "significant", NDT_STATISTICAL, SIGNIFICANT, 0.0 },
  { "not_analysed", NDT_STATISTICAL, NOT_ANALYSED, 0.0 },

  /* routing_threshold parameter */

  { "sibal_threshold", NDT_ROUTING_THRESHOLD, 0, ND_SIBAL_TH },

  /* general parameters */

  { "true", NDT_BOOLEAN, TRUE, 0.0 },
  { "false", NDT_BOOLEAN, FALSE, 0.0 },

  /* constants for virtual links */
  { "direct_vlink", NDT_VL_OPEN_ALGORITHM, DIRECT_VLINK, 0.0 },
  { "och_first", NDT_VL_OPEN_ALGORITHM, OCH_FIRST, 0.0 },
  { "ehc_groom", NDT_VL_OPEN_ALGORITHM, EHC_GROOM, 0.0 },
  { "eff_vlink", NDT_VL_OPEN_ALGORITHM, EFF_VLINK, 0.0 },
  { "empty_close", NDT_VL_CLOSE_ALGORITHM, EMPTY_CLOSE, 0.0 },
  { "none", NDT_PRECREATE_VTOPOLOGY, NONE, 0.0 },
  { "physical", NDT_PRECREATE_VTOPOLOGY, PHYSICAL, 0.0 },
  { "spantree", NDT_PRECREATE_VTOPOLOGY, SPANTREE, 0.0 },
  { "fullmesh", NDT_PRECREATE_VTOPOLOGY, FULLMESH, 0.0 },
};
#define NUM_CONSTANTS (sizeof(init_constant_table) / sizeof(struct init_constant))

/* List of predefined constants, it's built from the above
   initialization table. */
static struct nd_constant* constant_list = NULL;

/* Search for a parameter by name (case insensitive). */
struct nd_parameter*
lookup_parameter_by_name(const char* name)
{
  int i;

  for (i = 0; i < NUM_PARAMETERS; ++i)
    if (strcasecmp(name, parameter_table[i].name) == 0)
      return &parameter_table[i];

  return NULL;
}

/* If there is exactly *one* parameter with the given type return that
   parameter, otherwise return NULL. */
struct nd_parameter* lookup_parameter_by_type(enum nd_parameter_type type)
{
  int i;
  struct nd_parameter* found = NULL;

  for (i = 0; i < NUM_PARAMETERS; ++i)
    if (parameter_table[i].type == type)
      {
       if (found == NULL)
	found = &parameter_table[i];
       else
	return NULL;
      }

  return found;
}

/* This function must be called once and before parsing is started. */
void init_parameters(void)
{
  /* Create the global instance list. This list is unique and holds
     one instance for every global or local parameter.  */
  global_instance_list = create_default_instances(NDC_GLOBAL);
}

/* Create a list containing one instance for each parameter that can
   appear in CONTEXT.

   In the global context the new instance has the default value or is
   undefined if no default is available. In a local context the new
   instance assume the value of the corresponding global parameter, if
   this is defined. */
struct nd_instance* create_default_instances(long context)
{
  struct nd_instance* instance_list = NULL;
  int i;

  /* For every parameter that can appear in CONTEXT. */
  for (i = 0; i < NUM_PARAMETERS; ++i)
    if (parameter_table[i].context & context)
      {
	struct nd_instance* instance;

	instance = MMmalloc(sizeof(*instance), "create_default_instances");
	instance->parameter = &parameter_table[i];
	instance->next = instance_list;
	instance_list = instance;

	if (context == NDC_GLOBAL)
	  {
	    /* If this parameter has a default value assign it to the
	       instance. */
	    if (instance->parameter->flags & NDF_DEFAULT)
	      {
		if (! (instance->parameter->flags & NDF_VECTOR))
		  {
		    if (BASE_TYPE(instance->parameter->type) == NDT_INTEGER)
		      instance->value.integer = instance->parameter->def_integer;
		    else if (BASE_TYPE(instance->parameter->type) == NDT_REAL)
		      instance->value.real = instance->parameter->def_real;
		  }
		else
		  {
		    if (BASE_TYPE(instance->parameter->type) == NDT_REAL)
		      instance->value.vector =
			clone_real_vector(instance->parameter->def_real_vector);
		    else
		      {
			fprintf(stderr, "Integer vectors are non implemented\n");
			exit(1);
		      }
		  }

		instance->status = NDS_ASSIGNED;
	      }
	    else
	      instance->status = NDS_UNDEFINED;
	  }
	else /* The context is local. */
	  {
	    struct nd_instance* global_instance;

	    global_instance = lookup_instance_by_name(global_instance_list,
						      parameter_table[i].name);

	    /* If an assigned instance of this parameter exists in the
	       global context then assign its value to the local
	       instance. */
	    if (global_instance->status == NDS_ASSIGNED)
	      {
		if (! (instance->parameter->flags & NDF_VECTOR))
		  instance->value = global_instance->value;
		else
		  instance->value.vector =
		    clone_vector(global_instance->value.vector);

		instance->status = NDS_ASSIGNED;
	      }
	    else
	      instance->status = NDS_UNDEFINED;
	  }
      }

  return instance_list;
}

/* If PARAMETER can appear in CONTEXT assign VALUE to its instance in
   INSTANCE_LIST.  Return the assigned instance or NULL if the context
   is invalid.  */
struct nd_instance* assign_parameter(struct nd_instance* instance_list,
		 struct nd_parameter* parameter,
		 union nd_value value,
		 long context)
{
  struct nd_instance* instance = NULL;

  /* If PARAMETER can appear in CONTEXT. */
  if (parameter->context & context)
    {
      instance = lookup_instance_by_name(instance_list, parameter->name);

      if (instance)
	{
	  /* Assign the constant value to the instance. */

	  if (instance->parameter->flags & NDF_VECTOR)
	    {
	      if (instance->value.vector != NULL)
		free(instance->value.vector);
	
	      /* Do not allocate a new vector. */
	      instance->value.vector = value.vector;
	    }
	  else
	    instance->value = value;

	  instance->status = NDS_ASSIGNED;
	}
      else
	fprintf(stderr, "Instance of parameter `%s' not found in context 0x%lx\n",
		parameter->name, context);
    }

  return instance;
}

/* Similar to 'assign_parameter()' but does not overwrite an already
   assigned instance.  Also differs from 'assign_parameter()' in that
   when assigning a vector a new one is allocated. */
static struct nd_instance* assign_undefined_parameter(
			   struct nd_instance* instance_list,
			   struct nd_parameter* parameter,
			   union nd_value value,
			   long context)
{
  struct nd_instance* instance = NULL;

  if (parameter->context & context)
    {
      instance = lookup_instance_by_name(instance_list, parameter->name);

      if (instance)
	{
	  if (instance->status == NDS_UNDEFINED)
	    {
	      if (instance->parameter->flags & NDF_VECTOR)
		/* Allocate a new vector. */
		instance->value.vector = clone_vector(value.vector);
	      else
		instance->value = value;

	      instance->status = NDS_ASSIGNED;
	    }
	}
      else
	fprintf(stderr, "Instance of parameter `%s' not found in context 0x%lx\n",
		parameter->name, context);
    }

  return instance;
}

/* Return TRUE if VALUE is in the valid range for PARAMETER.  */
static Boolean value_in_range(struct nd_parameter* parameter,
				union nd_value value)
{
  Boolean in_range = TRUE;

  /* Symbolic constants can violate the range. If VALUE is not a
     symbolic constant the check the range.  */
  if (! get_constant_name(parameter->type, value))
    {
      if (BASE_TYPE(parameter->type) == NDT_INTEGER)
	{
	  if (parameter->flags & NDF_CHECK_MIN
	      && value.integer < parameter->min_integer)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MIN_EXCLUDED
	      && value.integer <= parameter->min_integer)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MAX
	      && value.integer > parameter->max_integer)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MAX_EXCLUDED
	      && value.integer >= parameter->max_integer)
	    in_range = FALSE;
	}
      else if (BASE_TYPE(parameter->type) == NDT_REAL)
	{
	  if (parameter->flags & NDF_CHECK_MIN
	      && value.real < parameter->min_real)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MIN_EXCLUDED
	      && value.real <= parameter->min_real)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MAX
	      && value.real > parameter->max_real)
	    in_range = FALSE;

	  if (parameter->flags & NDF_CHECK_MAX_EXCLUDED
	      && value.real >= parameter->max_real)
	    in_range = FALSE;
	}
    }

  return in_range;
}

/* Check INSTANCE bounds. Return TRUE if the instance value is in the
   valid range. */
static Boolean
instance_value_in_range(struct nd_instance* instance)
{
  Boolean in_range = TRUE;

  if (! (instance->parameter->flags & NDF_VECTOR))
    /* Check scalar value bounds. */
    in_range = value_in_range(instance->parameter, instance->value);
  else
    {
      /* Check the bounds of every element in the vector. */
      int i;

      for (i = 0; i < instance->value.vector->len && in_range; ++i)
	in_range = value_in_range(instance->parameter,
				  instance->value.vector->element[i]);
    }

  return in_range;
}

/* Check the bounds of defined instances in INSTANCE_LIST for which
   CONTEXT is a valid context (i.e. pure global parameters in a global
   context and local parameters in a local context).  ENTITY is the
   owner of the instance list or NULL if INSTANCE_LIST is the global
   list.

   An instance out of range and with the NDF_RESET_TO_DEFAULT flag set
   is reset to its default value with a warning message.

   Return TRUE if there are instances out of range that can't be
   resetted to default. */
static Boolean
check_instances_range(Parse* entity, struct nd_instance* instance_list,
		      enum nd_parameter_context context)
{
  Boolean bad_values = FALSE;
  struct nd_instance* instance;

  /* For every defined instance in a valid context (i.e. skip
     instances of local parameters appearing in the global
     context). */
  for (instance = instance_list; instance; instance = instance->next)
    {
      if (((PURE_GLOBAL_PARAMETER(instance->parameter) && context == NDC_GLOBAL)
	   || (LOCAL_PARAMETER(instance->parameter) && context != NDC_GLOBAL))
	  && instance->status != NDS_UNDEFINED)
	{
	  if (! instance_value_in_range(instance))
	    {
	      struct nd_parameter* parameter;

	      parameter = instance->parameter;

	      if (parameter->flags & NDF_RESET_TO_DEFAULT)
		{
		  print_warning(entity, "parameter `%s' out of range"
                                        "setting its default value.\n",
				      parameter->name);

		  /* Reset instance to its default value. */
		  if (BASE_TYPE(parameter->type) == NDT_INTEGER)
		    instance->value.integer = parameter->def_integer;
		  else if (BASE_TYPE(parameter->type) == NDT_REAL)
		    instance->value.real = parameter->def_real;
		}
	      else
		{
		  print_error(entity, "parameter `%s' out of range\n",
				      parameter->name);

		  bad_values = TRUE;
		}
	    }
	}
    }

  return bad_values;
}

/* Check for undefined instances in list INSTANCE_LIST. NODE is the
   owner of the instance list or NULL if INSTANCE_LIST is the global
   list. Return TRUE if some parameter is missing.

   Every global parameter (in a global context) and every local
   parameter in a local context must be assigned. Local parameters in
   a global context need not to be assigned. */
static Boolean
check_unassigned_instances(Parse* node, struct nd_instance* instance_list)
{
  Boolean missing = FALSE;
  struct nd_instance* instance;

  /* For every defined instance in a valid context (i.e. skip
     instances of local parameters appearing in the global
     context). */
  for (instance = instance_list; instance; instance = instance->next)
    if ((PURE_GLOBAL_PARAMETER(instance->parameter)
	 || (LOCAL_PARAMETER(instance->parameter) && node))
	&& instance->status != NDS_ASSIGNED)
      {
	missing = TRUE;
	print_error(node, "Parameter `%s' has not been assigned",
			    instance->parameter->name);
      }

  return missing;
}

/* Check for missing parameters in all existing contexts. Return TRUE
   if all parameters are defined. */
Boolean
check_missing_parameters(void)
{
  Boolean all_ok = TRUE;
  Parse* node;

  if (check_unassigned_instances(NULL, global_instance_list))
    all_ok = FALSE;

  for (node = global_cgen_list; node; node = node->next)
    if (check_unassigned_instances(node, node->instance_list))
      all_ok = FALSE;

  for (node = global_link_list; node; node = node->next)
    if (check_unassigned_instances(node, node->instance_list))
      all_ok = FALSE;

  for (node = global_node_list; node; node = node->next)
    if (check_unassigned_instances(node, node->instance_list))
      all_ok = FALSE;

  for (node = global_user_list; node; node = node->next)
    if (check_unassigned_instances(node, node->instance_list))
      all_ok = FALSE;

  return all_ok;
}

/*****************
****************** Functions dealing with constants.
*****************/

struct nd_constant*
lookup_constant_by_name(const char* name)
{
  struct nd_constant* node;

  for (node = constant_list; node; node = node->next)
    if (strcasecmp(name, node->name) == 0)
      break;

  return node;
}

/* Used by 'init_constants' to create a constants list starting from
   the constant initialization table. */
static struct nd_constant*
create_integer_constant(const char* name, enum nd_parameter_type type,
			long value)
{
  struct nd_constant* new;

  new = MMmalloc(sizeof(*new), "create_integer_constant");
  new->name = name;
  new->type = type;
  new->value.integer = value;

  new->next = constant_list;
  constant_list = new;

  return new;
}

/* Used by 'init_constants' to create a constants list starting from
   the constant initialization table. */
static struct nd_constant*
create_real_constant(const char* name, enum nd_parameter_type type,
		     double value)
{
  struct nd_constant* new;

  new = MMmalloc(sizeof(*new), "create_integer_constant");
  new->name = name;
  new->type = type;
  new->value.real = value;

  new->next = constant_list;
  constant_list = new;

  return new;
}

/* Create a list of constants starting from the constants
   initialization table. Must be called before the parsing is
   started. */
void
init_constants(void)
{
  int i;

  for (i = 0; i < NUM_CONSTANTS; ++i)
    if (BASE_TYPE(init_constant_table[i].type) == NDT_INTEGER)
      create_integer_constant(init_constant_table[i].name,
			      init_constant_table[i].type,
			      init_constant_table[i].integer);
    else if (BASE_TYPE(init_constant_table[i].type) == NDT_REAL)
      create_real_constant(init_constant_table[i].name,
			   init_constant_table[i].type,
			   init_constant_table[i].real);
    else
      fprintf(stderr, "Unknown base type for constant `%s'\n",
	      init_constant_table[i].name);
}

/* Retrieve the constant name starting from its type and value. If no
constant exists with the given type and value then return NULL.  */
static const char*
get_constant_name(enum nd_parameter_type type, union nd_value value)
{
  struct nd_constant* node;

  for (node = constant_list; node; node = node->next)
    if ((node->type == type)
	&& ((BASE_TYPE(node->type) == NDT_INTEGER
	     && node->value.integer == value.integer)
	    || (BASE_TYPE(node->type) == NDT_REAL
		&& node->value.real == value.real)))
      break;

  return node ? node->name : NULL;
}

/*****************
****************** Functions for "demand" statements.
*****************/

/* Set the status of PARAMETER instance that appear in INSTANCE_LIST
   to NDS_DEMAND. Return FALSE if the instance was already in
   NDS_DEMAND state.

   Note that a previous 'default' (or 'parameter') statement for this
   parameter will be silently overwritten. */
Boolean
assign_demand(struct nd_instance* instance_list, struct nd_parameter* parameter)
{
  struct nd_instance* instance;

  instance = lookup_instance_by_name(instance_list, parameter->name);

  if (instance->status != NDS_DEMAND)
    {
      instance->status = NDS_DEMAND;
      return TRUE;
    }
  else
    return FALSE;
}

void
ask_for_demand(void)
{
  Parse* node;

  ask_instances(global_instance_list, NULL);

  for (node = global_cgen_list; node; node = node->next)
    ask_instances(node->instance_list, node);

  for (node = global_link_list; node; node = node->next)
    ask_instances(node->instance_list, node);

  for (node = global_node_list; node; node = node->next)
    ask_instances(node->instance_list, node);

  for (node = global_user_list; node; node = node->next)
    ask_instances(node->instance_list, node);
}

/* Store one line of user input in BUF and null terminate it. Return
   the pointer to the first non blank character. */
char*
get_input_line(char* buf, size_t buflen)
{
  char* s;

  if (fgets(buf, buflen, stdin) == NULL)
    *buf = '\0';

  s = strchr(buf, '\n');
  if (s) *s = '\0';

  while (isspace(*buf))
    ++buf;

  return buf;
}

enum get_input_result
{
  GIR_OK,
  GIR_ERROR,
  GIR_EMPTY_LINE
};

/* Get a value for PARAMETER from the user.  Return GIR_OK if the user
   input is valid, GIR_EMPY_LINE if the user entered an empy line and
   GIR_ERROR if the user input si not valid. */
enum get_input_result
get_input_value(struct nd_parameter* parameter, union nd_value* value)
{
  char buf[256];
  char* s;
  enum get_input_result result = GIR_ERROR;

  s = get_input_line(buf, 256);

  if (*s == '\0')
    return GIR_EMPTY_LINE;

  if (BASE_TYPE(parameter->type) == NDT_INTEGER)
    {
      long integer;
      char* end;

      errno = 0;
      integer = strtol(s, &end, 10);

      if (! errno && end != s && (isspace(*end) || *end == '\0'))
	{
	  /* Valid integer value. */

	  value->integer = integer;
	  result = GIR_OK;
	}
    }
  else if (BASE_TYPE(parameter->type) == NDT_REAL)
    {
      double real;
      char* end;

      errno = 0;
      real = strtod(s, &end);

      if (! errno && end != s && (isspace(*end) || *end == '\0'))
	{
	  /* Valid real input. */

	  value->real = real;
	  result = GIR_OK;
	}
    }

  if (result == GIR_ERROR)
    {
      /* If the user input is invalid as a number check if it's a
	 symbolic constant.  */
      if (parameter->type != BASE_TYPE(parameter->type) && isalpha(*s))
	{
	  struct nd_constant* constant;

	  constant = lookup_constant_by_name(s);

	  if (constant && constant->type == parameter->type)
	    {
	      *value = constant->value;
	      result = GIR_OK;
	    }
	  else
	    fprintf(stderr,
		    "Invalid input (unknown constant or bad constant type)\n");
	}
      else
	fprintf(stderr, "Invalid numeric input\n");
    }

  return result;
}

/* Ask the user for a value in a finite set for INSTANCE.
   ENTITY_NAME identify the instance owner; it can be NULL and it's
   used only in the prompt. */
void
get_instance_choice(struct nd_instance* instance, const char* entity_name)
{
  Boolean got_input = FALSE;
  struct nd_parameter* parameter = instance->parameter;

  do
    {
      union nd_value value;
      struct nd_constant* cnst;
      long last_value = -1;
      long min_value = INT_MAX;
      long max_value = INT_MIN;

      printf("\nChoice a value for the `%s' parameter", parameter->name);

      if (entity_name && *entity_name)
	printf(" in %s:\n", entity_name);
      else
	printf(":\n");

      /* The following code:

	  - Uses constants values as labels.

	  - Assumes that all values between min and max are valid.

	  - Assumes that constant with identical type and value are
	    contiguous in constant_list. */
      for (cnst = constant_list; cnst; cnst = cnst->next)
	if (cnst->type == parameter->type
	    && cnst->value.integer != last_value)
	  {
	    if (cnst->value.integer < min_value)
	      min_value = cnst->value.integer;

	    if (cnst->value.integer > max_value)
	      max_value = cnst->value.integer;

	    last_value = cnst->value.integer;
	    printf("(%ld) %s\n", cnst->value.integer, cnst->name);
	  }

      if (get_input_value(parameter, &value) == GIR_OK)
	{
	  if (value.integer >= min_value && value.integer <= max_value)
	    {
	      instance->value = value;
	      got_input = TRUE;
	    }
	  else
	    fprintf(stderr, "Enter a number between %ld and %ld\n",
		    min_value, max_value);
	}
    }
  while (! got_input
#ifdef unix
         && isatty(fileno(stdin)));
#else
         &&  FALSE);
#endif
  /* If the standard input is not interactive there is no user who can
     correct the input so it's better to stop now. */
  if (! got_input)
    exit(1);
}

/* Print the valid range for PARAMETER.  */
void
print_valid_range(struct nd_parameter* parameter)
{
  /* Print the valid range (if available). */
  if (parameter->flags & (NDF_CHECK_RANGE | NDF_CHECK_MIN_EXCLUDED
			  | NDF_CHECK_MAX_EXCLUDED))
    {
      printf("Valid range is ");
      if (parameter->flags & NDF_CHECK_MIN)
	printf("[%g, ",
	       BASE_TYPE(parameter->type) == NDT_INTEGER
	       ? parameter->min_integer : parameter->min_real);
      else if (parameter->flags & NDF_CHECK_MIN_EXCLUDED)
	printf("(%g, ",
	       BASE_TYPE(parameter->type) == NDT_INTEGER
	       ? parameter->min_integer : parameter->min_real);
      else
	printf("(-inf, ");

      if (parameter->flags & NDF_CHECK_MAX)
	printf("%g]\n",
	       BASE_TYPE(parameter->type) == NDT_INTEGER
	       ? parameter->max_integer : parameter->max_real);
      else if (parameter->flags & NDF_CHECK_MAX_EXCLUDED)
	printf("%g)\n",
	       BASE_TYPE(parameter->type) == NDT_INTEGER
	       ? parameter->max_integer : parameter->max_real);
      else
	printf("inf)\n");
    }
}

/* Ask the user for a value (real or integer) for INSTANCE.
   ENTITY_NAME identify the instance owner; it can be NULL and it's
   used only in the prompt. */
void
get_instance_value(struct nd_instance* instance, const char* entity_name)
{
  Boolean got_input = FALSE;
  struct nd_parameter* parameter = instance->parameter;

  do
    {
      union nd_value value;

      /* Print the request. */
      if (BASE_TYPE(parameter->type) == NDT_INTEGER)
	printf("\nEnter an integer value for the `%s' parameter",
	       parameter->name);
      else if (BASE_TYPE(parameter->type) == NDT_REAL)
	printf("\nEnter a real value for the `%s' parameter",
	       parameter->name);
      else
	{
	  fprintf(stderr,
		  "Parameter type 0x%x not handled by 'get_instance_value'\n",
		  parameter->type);
	  exit(1);
	}

      if (entity_name && *entity_name)
	printf(" in %s:\n", entity_name);
      else
	printf(":\n");

      print_valid_range(parameter);

      /* Print the constants set. */
      if (parameter->type != BASE_TYPE(parameter->type))
	{
	  struct nd_constant* cnst;
	  printf("You may also enter one of the following constants:\n");

	  for (cnst = constant_list; cnst; cnst = cnst->next)
	    if (cnst->type == parameter->type)
	      printf("  %s\n", cnst->name);
	}

      /* Get a value from the user. */
      if (get_input_value(parameter, &value) == GIR_OK)
	{
	  instance->value = value;

	  if (instance_value_in_range(instance))
	    got_input = TRUE;
	  else
	    fprintf(stderr, "Value not in range\n");
	}
    }
  while (! got_input
#ifdef unix
         && isatty(fileno(stdin)));
#else
         &&  FALSE);
#endif

  /* If the standard input is not interactive there is no user who can
     correct the input so it's better to stop now. */
  if (! got_input)
    exit(1);
}

/* Ask the user for a vector for INSTANCE, that belongs to entity
   called ENTITY_NAME (or is a global instance if ENTITY_NAME is
   NULL).  The size of the vector is either defined by another
   parameter or is choosen by the user at input time.  */
void
get_instance_vector_value(struct nd_instance* instance,
			  const char* entity_name)
{
  int j;
  struct nd_instance* size_inst;
  struct nd_parameter* parameter = instance->parameter;

  if (BASE_TYPE(parameter->type) == NDT_INTEGER)
    printf("\nEnter an integer vector for the `%s' parameter",
	   parameter->name);
  else if (BASE_TYPE(parameter->type) == NDT_REAL)
    printf("\nEnter a real vector for the `%s' parameter",
	   parameter->name);
  else
    {
      fprintf(stderr,
	      "Parameter type 0x%x not handled by "
	      "'get_instance_vector_value'\n",
	      parameter->type);
      exit(1);
    }

  if (entity_name && *entity_name)
    printf(" in %s,\n", entity_name);
  else
    printf(",\n");

  /* If possible get the size of this vector from another
     parameter. */
  if (parameter->size_parameter != NULL)
    size_inst =
      lookup_instance_by_name(global_instance_list,
			      parameter->size_parameter);
  else
    size_inst = NULL;

  /* At this point no automatic range checking has been performed so
     check here if the size parameter is valid. */
  if (size_inst && (! instance_value_in_range(size_inst)))
    {
      fprintf(stderr, "\nError: '%s' exceeding the maximum value\n",
	      size_inst->parameter->name);
      exit(1);
    }

  if (size_inst)
    printf("insert %ld value(s) on different lines:\n",
	   size_inst->value.integer);
  else
    printf("insert vector elements on different lines, terminate\n"
	   "with an empty line:\n");

  print_valid_range(parameter);

  /* Print the constants set. */
  if (parameter->type != BASE_TYPE(parameter->type))
    {
      struct nd_constant* cnst;
      printf("You may also enter one of the following constants:\n");

      for (cnst = constant_list; cnst; cnst = cnst->next)
	if (cnst->type == parameter->type)
	  printf("  %s\n", cnst->name);
    }

  if (instance->value.vector == NULL)
    instance->value.vector = create_vector();

  /* If the size of the vector is unknown (i.e. size_inst == NULL)
     exit when the user insert an empty line.  Otherwise exit when the
     user has entered all vector elements. */
  j = 0;
  while (1)
    {
      union nd_value value;
      enum get_input_result result;

      printf("[%d] = ", j);
      result = get_input_value(parameter, &value);

      if (result == GIR_OK)
	{
	  if (value_in_range(parameter, value))
	    {
	      /* Success: insert the value in the vector and increment
		 the index. */
	      instance->value.vector =
		insert_vector_element(instance->value.vector, j, value);

	      ++j;
	    }
	  else
	    printf("Value out of range\n");
	}
      else if (result == GIR_EMPTY_LINE && size_inst == NULL)
	break;

      if (size_inst && j >= size_inst->value.integer)
	break;
    }
}

/* Ask the user for the values of every instance in the list INSTANCE
   that has a status of NDS_DEMAND.  ENTITY is the owner of the
   instance list or NULL if INSTANCE is the global list. */
static void
ask_instances(struct nd_instance* instance, Parse* entity)
{
  for ( ; instance; instance = instance->next)
    if (instance->status == NDS_DEMAND)
      {
	if (instance->parameter->flags & NDF_ENUMERATION)
	  {
	    /* The parameter accepts only symbolic constants in a
	       finite set. */

	    get_instance_choice(instance,
				identify_entity(entity, FALSE));

	    instance->status = NDS_ASSIGNED;

	    /* Print the assigned value. */
	    fprintf(stderr, "%s = %s\n", instance->parameter->name,
		    get_constant_name(instance->parameter->type, instance->value));
	  }
	else
	  {
	    /* The parameter is a number or a vector (and perhaps
	       accepts symbolic constants). */

	    if (! (instance->parameter->flags & NDF_VECTOR))
	      get_instance_value(instance,
				 identify_entity(entity, FALSE));
	    else
	      get_instance_vector_value(instance,
					identify_entity(entity, FALSE));

	    instance->status = NDS_ASSIGNED;

	    /* Print the assigned value. */
	    printf("%s = ", instance->parameter->name);

	    if (! (instance->parameter->flags & NDF_VECTOR))
	      {
		print_value(instance->parameter, instance->value);
		printf("\n");
	      }
	    else
	      {
		int i;

		printf("{ ");

		for (i = 0; i < instance->value.vector->len; ++i)
		  {
		    if (i != 0) printf(", ");

		    print_value(instance->parameter, instance->value.vector->element[i]);
		  }

		printf(" }\n");
	      }
	  }

	/* If this is a global demand (i.e. entity == NULL) propagate
	   the value of this instance to all the undefined local
	   instances.  */
	if (entity == NULL && instance->status == NDS_ASSIGNED)
	    assign_local_undefined_instances(instance->parameter, instance->value);
      }
}

/* Assign VALUE to all instances of PARAMETER not yet defined in local
   contexts. */
static void
assign_local_undefined_instances(struct nd_parameter* parameter,
				 union nd_value value)
{
  Parse* node;

  if (parameter->context & NDC_GENERATOR_MASK)
    for (node = global_cgen_list; node; node = node->next)
      assign_undefined_parameter(node->instance_list, parameter, value,
				 set_cgen_context(node->elem.gener->type));

  if (parameter->context & NDC_LINK)
    for (node = global_link_list; node; node = node->next)
      assign_undefined_parameter(node->instance_list, parameter, value, NDC_LINK);

  if (parameter->context & NDC_NODE)
    for (node = global_node_list; node; node = node->next)
      assign_undefined_parameter(node->instance_list, parameter, value, NDC_NODE);

  if (parameter->context & NDC_USER)
    for (node = global_user_list; node; node = node->next)
      assign_undefined_parameter(node->instance_list, parameter, value, NDC_USER);
}

/*****************
****************** Functions dealing with single instances.
*****************/

/* Return the instance named NAME in list NODE, or NULL if it doesn't
   exist. */
struct nd_instance*
lookup_instance_by_name(struct nd_instance* node, const char* name)
{
  for ( ; node; node = node->next)
    if (strcasecmp(name, node->parameter->name) == 0)
      break;

  return node;
}

/* Return the value assumed by the real parameter NAME in the instance
   list LIST. */
double
real_instance_value(struct nd_instance* list, const char* name)
{
  struct nd_instance* instance;

  instance = lookup_instance_by_name(list, name);
  if (instance
      && instance->status == NDS_ASSIGNED
      && BASE_TYPE(instance->parameter->type) == NDT_REAL)
    return instance->value.real;
  else
    {
      fprintf(stderr, "Parameter `%s' does not exist or is not real\n", name);
      exit(1);
    }
}

/* Return the value assumed by the real vector parameter NAME in the
   instance list LIST. */
struct nd_vector*
real_vector_instance_value(struct nd_instance* list, const char* name)
{
  struct nd_instance* instance;

  instance = lookup_instance_by_name(list, name);
  if (instance
      && instance->status == NDS_ASSIGNED
      && BASE_TYPE(instance->parameter->type) == NDT_REAL
      && (instance->parameter->flags & NDF_VECTOR))
    return instance->value.vector;
  else
    {
      fprintf(stderr, "Parameter `%s' does not exist or is not a real vector\n", name);
      exit(1);
    }
}

/* Return the value assumed by the real vector parameter NAME in the
   instance list LIST. */
struct nd_vector*
integer_vector_instance_value(struct nd_instance* list, const char* name)
{
  struct nd_instance* instance;

  instance = lookup_instance_by_name(list, name);
  if (instance
      && instance->status == NDS_ASSIGNED
      && BASE_TYPE(instance->parameter->type) == NDT_INTEGER
      && (instance->parameter->flags & NDF_VECTOR))
    return instance->value.vector;
  else
    {
      fprintf(stderr, "Parameter `%s' does not exist or "
	      "is not an integer vector\n", name);
      exit(1);
    }
}

/* Return the value assumed by the integer parameter NAME in the
   instance list LIST. */
long
integer_instance_value(struct nd_instance* list, const char* name)
{
  struct nd_instance* instance;

  instance = lookup_instance_by_name(list, name);
  if (instance
      && instance->status == NDS_ASSIGNED
      && BASE_TYPE(instance->parameter->type) == NDT_INTEGER)
    return instance->value.integer;
  else
    {
      fprintf(stderr, "Parameter `%s' does not exist or is not integer\n", name);
      exit(1);
    }
}

/*****************
****************** Output functions for debug and error handling
*****************/

/*
**  Function  : void remind_parameter_names(void)
**  Return    : none.
**  Remarks   : the list of ANCLES parameters is printed on the standard
**              output.
*/
void
remind_parameter_names(void)
{
  int i;

  printf("ANCLES Parameters\n");
  for (i = 0; i < NUM_PARAMETERS; ++i)
    printf("%30s - %s\n", parameter_table[i].name, parameter_table[i].description);
}

/* Print the number of defined entities and the values of all global
   instances. */
void
nd_print_debug(void)
{
  printf("\n <<< INITIALIZATION TRACE >>>\n\n");

  printf(" --- PARSING DONE ---\n\n");
  printf("Read %2d nodes, %2d links, %2d users, %2d call generators\
and %2d routes\n",
	 pnode_num, plink_num, puser_num, pcgen_num, proute_num);

  printf("\n --- DEFAULT VALUES ---\n\n");

  print_instances(global_instance_list, NDS_ASSIGNED);

  printf("\n --- DEMAND VARIABLES ---\n\n");

  print_instances(global_instance_list, NDS_DEMAND);
}

/* Print VALUE as a number or, if possible, as a sybolic constant.  */
static void
print_value(struct nd_parameter* parameter, union nd_value value)
{
  const char* s;

  s = get_constant_name(parameter->type, value);

  if (s)
    printf("%s", s);
  else if (BASE_TYPE(parameter->type) == NDT_INTEGER)
    printf("%ld", value.integer);
  else if (BASE_TYPE(parameter->type) == NDT_REAL)
    printf("%g", value.real);
}

/* Print all instances in LIST that have the given STATUS.  For
   instances with status NDS_ASSIGNED print also the instance value.  */
static void
print_instances(struct nd_instance* list, long status)
{
  for ( ; list; list = list->next)
    if (list->status == status)
      {
	if (status == NDS_DEMAND)
	  printf("%s\n", list->parameter->name);
	else
	  {
	    printf("%s = ", list->parameter->name);

	    if (! (list->parameter->flags & NDF_VECTOR))
	      {
		print_value(list->parameter, list->value);
		printf("\n");
	      }
	    else
	      {
		int i;

		printf("{ ");

		for (i = 0; i < list->value.vector->len; ++i)
		  {
		    if (i != 0) printf(", ");

		    print_value(list->parameter, list->value.vector->element[i]);
		  }

		printf(" }\n");
	      }
	  }
      }
}

/* Return a pointer to a string containing the ENTITY type (user, link
   etc.), its name and (if VERBOSE is TRUE) its line number in the ND
   file. For example "node `London' defined at line 121". */
static const char*
identify_entity(Parse* entity, Boolean verbose)
{
  static char buf[128];

  if (entity)
    {
      const char* type_name;
      const char* entity_name;

      switch (entity->type)
	{
	case T_USER:
	  type_name = "user";
	  entity_name = user_name(entity->id);
	  break;
	case T_LINK:
	  type_name = "link";
	  entity_name = link_name(entity->id);
	  break;
	case T_NODE:
	  type_name = "node";
	  entity_name = node_name(entity->id);
	  break;
	case T_CGEN:
	  type_name = "call generator";
	  entity_name = cgen_name(entity->id);
	  break;
	default:
	  type_name = "unknown";
	  entity_name = "unknown";
	  break;
	}

      if (verbose)
	sprintf(buf, "%s `%s' defined at line %d",
		 type_name, entity_name, entity->decl_line);
      else
	sprintf(buf, "%s `%s'",
		 type_name, entity_name);
    }
  else
    *buf = '\0';

  return buf;
}

void
print_error(Parse* entity, const char* errfmt, ...)
{
  va_list ap;

  if (entity)
    fprintf(stderr, "Error in %s:\n  ", identify_entity(entity, TRUE));
  else
    fprintf(stderr, "Error: ");

  va_start(ap, errfmt);
  vfprintf(stderr, errfmt, ap);
  va_end(ap);

  fprintf(stderr, "\n");
}

void
print_warning(Parse* entity, const char* errfmt, ...)
{
  va_list ap;

  if (entity)
    fprintf(stderr, "Warning in %s:\n  ", identify_entity(entity, TRUE));
  else
    fprintf(stderr, "Warning: ");

  va_start(ap, errfmt);
  vfprintf(stderr, errfmt, ap);
  va_end(ap);

  fprintf(stderr, "\n");
}

/*
** Not ANSI library functions
*/
#ifdef __STRICT_ANSI__
char* strdup(const char* s)
{
  char* new = NULL;

  if (s)
    {
      new = malloc(strlen(s) + 1);

      if (new)
	strcpy(new, s);
    }

  return new;
}

int strcasecmp(const char* a, const char* b)
{
  while (toupper(*a) == toupper(*b))
    {
      if (*a == '\0')
	return 0;

      ++a;
      ++b;
    }

  return toupper(*a) - toupper(*b);
}
#endif
