/*
**  Module    : nd_service.c
**  Version   : $Revision: 3.1 $
**  Date      : $Date: 1998/07/28 09:51:21 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains the functions frequently used by the
**              other ND modules for checking the parsing context or
**              returning the symbolic name of the entities.
**
**  History:    01-APR-24 - Insert the features for ASON:
**                          ASON call generators and duplex fibres - ZZ -
**				02-APR-24 - Support of predefined routes - ZZ -
**
**  Functions : Boolean in_list_parse(Parse **head, Parse *elem);
**              int insert_symbol(SymTable *table, char *symbol, int dimension);
**              int lookup_symbol(SymTable *table, char *symbol, int dimension);
**              Boolean insert_real_vector(IntVect *list, double value);
**              char *node_name(int ident);
**              char *link_name(int ident);
**              char *user_name(int ident);
**              char *cgen_name(int ident);
**              long int  set_cgen_context(int user_type);
**              void add_connection(Parse *list, int node_id, int link_id);
**              void add_usage(Parse *list, int user_id, int link_id);
**              void add_referring(Parse *list, int user_id);
**				void add_link_to_route(Parse *, int, int);
**				void add_fn_to_route(Parse *, int);
**				void add_tn_to_route(Parse *, int);
**              void release_memory(void);
*/

#include "nd.h"
#include <string.h>

extern Parse *global_node_list,
             *global_link_list,
             *global_user_list,
             *global_cgen_list,
             *global_route_list;

extern int pnode_num,
           plink_num,
           puser_num,
           pcgen_num,
           proute_num;
static char empty_name[1] = "";


/*
**  Function  : Boolean in_list_parse(Parse **head, Parse *elem)
**  Return    : FALSE if the structure pointed by 'elem' is already
**              present in the list, TRUE otherwise.
**  Remarks   : the parsed element is added to the corresponding list.
**              The list is ordered by identifier and duplicated
**              identifiers are not allowed.
*/
Boolean in_list_parse(Parse **head, Parse *elem)
{
 Parse *p, *q;

 if(*head == NULL)
   {
    *head = elem;
    return TRUE;
   }
 if((*head)->next == NULL)
   {
    if((*head)->id < elem->id)
      {
       (*head)->next = elem;
       elem->next = NULL;
       return TRUE;
      }
    else if((*head)->id > elem->id)
      {
       elem->next = *head;
       *head = elem;
       return TRUE;
      }
    else return FALSE;
   }
 if((*head)->id > elem->id)
   {
    elem->next = *head;
    *head = elem;
    return TRUE;
   }
 else
   if((*head)->id == elem->id)
     return FALSE;
   /* Here 'elem' is surely greater than '*head'. */
 q = *head;
 p = q->next;
 while(p!=NULL)
   {
    if(p->id >= elem->id) break;
    q = p;
    p = p->next;
   }
 if(p != NULL && p->id == elem->id)
    return FALSE;
 q->next = elem;
 elem->next = p;
 return TRUE;
}


/*
**  Function  : int insert_symbol(SymTable *table, char *symbol, int dimension)
**  Return    : the index of the inserted element or 'dimension' if the
**              table is full.
**  Remarks   : a symbol table is scanned for inserting a symbol.
**              The entries are converted to lowercase.
*/
int insert_symbol(SymTable *table, char *symbol, int dimension)
{
  register int i;

  for(i = 0; i < dimension; i++)
    {
      if(table[i].name == NULL)
	{
	  /* This is the first empty entry. The entries are supposed to be        */
	  /* contigous so that the first empty space is the useful one.           */
	  table[i].name = symbol;
	  table[i].ident = i;
	  return i;
	}
      else if (strcasecmp(table[i].name, symbol) == 0)
	{
	  free(symbol);
	  return i;
	}
    }

  free(symbol);

  return dimension;
}


/*
**  Function  : int lookup_symbol(SymTable *table, char *symbol, int dimension)
**  Return    : the index of the symbol in the table or 'dimension' if the
**              symbol does not exist.
**  Remarks   : a symbol table is scanned for the required symbol.
**              The entries are converted to lowercase.
*/
int lookup_symbol(SymTable *table, char *symbol, int dimension)
{
  register int i;

  for(i = 0; i < dimension; i++)
    if (strcasecmp(table[i].name, symbol) == 0)
      return i;

  return dimension;
}

/* Create a vector.  A freshly created vector is empty and has room
   for `ND_VECTOR_INIT_MAX_LEN' elements.  */
struct nd_vector*
create_vector(void)
{
  struct nd_vector* v;

  v = MMmalloc(sizeof(struct nd_vector), "create_vector");
  v->len = 0;
  v->max_len = ND_VECTOR_INIT_MAX_LEN;

  return v;
}

/* Insert VALUE in VECTOR position INDEX.  Extend the vector when more
   space is needed.  */
struct nd_vector*
insert_vector_element(struct nd_vector* vector, long index,
		      union nd_value value)
{
  if (index >= vector->len)
    vector->len = index + 1;

  if (vector->len > vector->max_len)
    {
      vector->max_len *= 2;
      vector = realloc(vector, sizeof(struct nd_vector)
		     + sizeof(union nd_value)
		     * (vector->max_len - ND_VECTOR_INIT_MAX_LEN));

      if (vector == NULL)
	{
	  fprintf(stderr, "Error: memory exhausted in "
		  "function insert_vector_element\n");
	  exit(1);
	}
    }

  vector->element[index] = value;

  return vector;
}

/* Convert an integer vector to a real vector.  */
void
vector_integer_to_real(struct nd_vector* vector)
{
  long i;

  for (i = 0; i < vector->len; ++i)
    vector->element[i].real =
      (double)vector->element[i].integer;
}

/* Create a vector starting from a real vector. */
struct nd_vector*
clone_real_vector(const struct nd_real_vector* vect)
{
  struct nd_vector* new;
  union nd_value value;
  int i;

  new = create_vector();

  for (i = 0; i < vect->len; ++i)
    {
      value.real = vect->element[i];
      new = insert_vector_element(new, i, value);
    }

  return new;
}

struct nd_vector*
clone_vector(const struct nd_vector* vect)
{
  struct nd_vector* new;
  int i;

  new = create_vector();

  for (i = 0; i < vect->len; ++i)
    new = insert_vector_element(new, i, vect->element[i]);

  return new;
}

/*
**  Function  : char *node_name(int ident)
**  Return    : the pointer to a string containing the name of node 'ident'
**              or an empty string if the node 'ident' does not exist.
**  Remarks   : the returned string is the one contained in the symbol table
**              structure.
*/
const char *node_name(int ident)
{
 register int i;

 for(i = 0; i < MAX_NODE; i++)
        if(node_symbol_table[i].ident == ident)
                return node_symbol_table[i].name;
 return empty_name;
}


/*
**  Function  : char *link_name(int ident)
**  Return    : the pointer to a string containing the name of link
**              'ident' or an empty string if the link 'ident' does not
**              exist.
**  Remarks   : the returned string is the one contained in the symbol table
**              structure.
*/
const char *link_name(int ident)
{
 register int i;

 for(i = 0; i < MAX_LINK; i++)
        if(link_symbol_table[i].ident == ident)
                return link_symbol_table[i].name;
 return empty_name;
}


/*
**  Function  : char *user_name(int ident)
**  Return    : the pointer to a string containing the name of user
**              'ident' or an empty string if the user 'ident' does not
**              exist.
**  Remarks   : the returned string is the one contained in the symbol table
**              structure.
*/
const char *user_name(int ident)
{
 register int i;

 for(i = 0; i < MAX_NET_USER; i++)
        if(user_symbol_table[i].ident == ident)
                return user_symbol_table[i].name;
 return empty_name;
}


/*
**  Function  : char *cgen_name(int ident)
**  Return    : the pointer to a string containing the name of call
**              generator 'ident' or an empty string if the call
**              generator 'ident' does not exist.
**  Remarks   : the returned string is the one contained in the symbol table
**              structure.
*/
const char *cgen_name(int ident)
{
 register int i;

 for(i = 0; i < MAX_CALL_GEN; i++)
        if(callgen_symbol_table[i].ident == ident)
                return callgen_symbol_table[i].name;
 return empty_name;
}

/*
**  Function  : char *route_name(int ident)
**  Return    : the pointer to a string containing the name of predefined
**              route 'ident' or an empty string if the route 'ident' 
**				does not exist.
**  Remarks   : the returned string is the one contained in the symbol table
**              structure.
*/
const char *route_name(int ident)
{
 register int i;

 for(i = 0; i < MAX_ROUTE; i++)
        if(route_symbol_table[i].ident == ident)
                return route_symbol_table[i].name;
 return empty_name;
}


/*
**  Function  : long int set_cgen_context(int user_type)
**  Return    : a context identifier.
**  Remarks   : none.
*/
long int set_cgen_context(enum cg_type cgen_type)
{
 long int context;

 switch(cgen_type)
   {
    case CG_TYPE1 :
       context = NDC_GENERATOR_TYPE1;
       break;

    case CG_TYPE2 :
       context = NDC_GENERATOR_TYPE2;
       break;

    case CG_TYPE3 :
       context = NDC_GENERATOR_TYPE3;
       break;

    case CG_TYPE4 :
       context = NDC_GENERATOR_TYPE4;
       break;

    case CG_TYPE5 :
       context = NDC_GENERATOR_TYPE5;
       break;

    case CG_TYPE6 :
       context = NDC_GENERATOR_TYPE6;
       break;

    case CG_TYPE7 :
       context = NDC_GENERATOR_TYPE7;
       break;

    case CG_TYPE8 :
       context = NDC_GENERATOR_TYPE8;
       break;

    case CG_TYPE9 :
       context = NDC_GENERATOR_TYPE9;
       break;

    case CG_TYPE10 :
       context = NDC_GENERATOR_TYPE10;
       break;

    case CG_TYPE11 :
       context = NDC_GENERATOR_TYPE11;
       break;

    case CG_TYPE12 :
       context = NDC_GENERATOR_TYPE12;
       break;

/* When adding a new Call Generator Type, insert here a 'case CG_TYPEx:'*/
/* where 'x' is the new Call Generator Type.                            */
    default:
       fprintf(stderr,"nd_services.c:set_cgen_context():");
       fprintf(stderr," unknown call generator type %d\n", cgen_type);
       exit(1);
   }
 return context;
}

/*
**  Function  : void add_connection(Parse *list, int node_id, int link_id)
**  Return    : none.
**  Remarks   : information about the connected node and the connecting
**              link is recorded in a Parse structure (Node or User).
*/
void add_connection(Parse *list, int node_id, int link_id)
{
 PNode *tnode;
 PUser *tuser;

 switch(list->type)
   {
    case T_NODE :
      tnode = list->elem.node;
      if(tnode->port_no >= MAX_PORT)
        {
         printf("Error:  Port capacity exceeded for node %s\n",
                 node_name(tnode->id));
         exit(1);
        }
      tnode->connected_node[tnode->port_no] = node_id;
      tnode->outlink[tnode->port_no] = link_id;
      tnode->port_no++;
      break;

    case T_USER :
      tuser = list->elem.user;
      tuser->connected_node = node_id;
      tuser->link_id = link_id;
      break;
   }
}


/*
**  Function  : void add_usage(Parse *list, int user_id, int link_id)
**  Return    : none.
**  Remarks   : information about the connected user and the connecting
**              link is recorded in a Parse Node structure.
*/
void add_usage(Parse *list, int user_id, int link_id)
{
 PNode *tnode;

 tnode = list->elem.node;
 if(tnode->user_no >= MAX_USER)
   {
    printf("Error:  User capacity exceeded for node %s\n",
            node_name(tnode->id));
    exit(1);
   }
 tnode->connected_user[tnode->user_no] = user_id;
 tnode->to_user[tnode->user_no] = link_id;
 tnode->user_no++;
}


/*
**  Function  : void add_referring(Parse *list, int user_id)
**  Return    : none.
**  Remarks   : information about the reference user is recorded
**              in a Parse Call Generator structure.
*/
void add_referring(Parse *list, int user_id)
{
 PCGen *pgen;

 pgen = list->elem.gener;
 pgen->refer_user = user_id;
}


/*
**  Function  : void add_fn_to_route(Parse *list, int node_id)
**  Return    : none.
**  Remarks   : information about the starting node of the route
**              is recorded in a Parse Route structure.
*/
void add_fn_to_route(Parse *list, int node_id)
{
 PRoute *troute;

 troute = list->elem.route;
 troute->from_node = node_id;
}

/*
**  Function  : void add_tn_to_route(Parse *list, int node_id)
**  Return    : none.
**  Remarks   : information about the ending node of the route
**              is recorded in a Parse Route structure.
*/
void add_tn_to_route(Parse *list, int node_id)
{
 PRoute *troute;

 troute = list->elem.route;
 troute->to_node = node_id;
}

/*
**  Function  : void add_link_to_route(Parse *list, int link_id)
**  Return    : none.
**  Remarks   : information about a link of the route
**              is recorded in a Parse Route structure.
*/
void add_link_to_route(Parse *list, int link_id)
{
 PRoute *troute;

 troute = list->elem.route;
 if(troute->link_no >= MAX_LINK)
   {
    printf("Error:  Max number of links exceeded for route %s\n",
            route_name(troute->id));
    exit(1);
   }
 troute->contained_link[troute->link_no] = link_id;
 troute->link_no++;
}




static void
release_instance_list_memory(struct nd_instance* instance)
{
  struct nd_instance* next;

  for ( ; instance; instance = next)
    {
      next = instance->next;

      if (instance->parameter->flags & NDF_VECTOR)
	free(instance->value.vector);

      free(instance);
    }
}


static void
release_parse_list_memory(Parse* entity)
{
  Parse* next;

  for ( ; entity; entity = next)
    {
      next = entity->next;

      free(entity->elem.node); /* Type is not important. */
      release_instance_list_memory(entity->instance_list);
      free(entity);
    }
}

/*
**  Function  : void release_memory(void)
**  Return    : none.
**  Remarks   : all the memory allocated by the ND data structures
**              (parse and demand lists) is released.
*/
void release_memory(void)
{
  release_instance_list_memory(global_instance_list);

  release_parse_list_memory(global_node_list);
  global_node_list = NULL;

  release_parse_list_memory(global_link_list);
  global_link_list = NULL;

  release_parse_list_memory(global_user_list);
  global_user_list = NULL;

  release_parse_list_memory(global_cgen_list);
  global_cgen_list = NULL;

  release_parse_list_memory(global_route_list);
  global_route_list = NULL;

}
