/*
**  Module    : nd_traffic.c
**  Version   : $Revision: 3.1 $
**  Date      : $Date: 1998/07/28 09:51:21 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains the functions needed to manage
**              the Traffic Relation File and the internal traffic matrix.
**
**  Functions : int parse_line_traffic(char *, char *, double *)
**              Boolean read_traff_relations(char *)
**              void create_uniform_traffic(void)
**              void remind_parameter_names(void)
**
19.02.2001 Cleaning... ZZ
**              01-APR-23 - Insert the features for ASON:
**                          ASON call generators and duplex fibres - ZZ -
*/

#include "nd.h"

extern Parse *global_node_list,
             *global_link_list,
             *global_user_list,
             *global_cgen_list,
             *global_route_list;

extern int pnode_num,
           plink_num,
           puser_num,
           pcgen_num,
           proute_num;




/*
**  Function  : int parse_line_traffic(char *, char *, double *)
**  Return    : a code indicating the type of the string contained in
**              'buffer' or an error indicator.
**              The allowed code values are:
**
**              - COMMENT_LINE :  comment (starting with #)
**              - SOURCE_LINE  :  source definition (starting with @)
**              - NORMAL_LINE  :  destination definition
**              - ERROR_LINE   :  error in the line format
**
**              In 'string' is returned the user name (2nd field),
**              in 'value' is returned the load fraction (3rd field).
**  Remarks   : none.
*/
int parse_line_traffic(char *buffer, char *string, double *value)
{
 int read_params, status;

 if(buffer == NULL)
        return ERROR_LINE;
 switch(buffer[0])
   {
    case '#' :
         status = COMMENT_LINE;
         break;

    case '@' :
         buffer[0] = ' ';
         read_params = sscanf(buffer, "%s %lf", string, value);
         status = (read_params == 2 ? SOURCE_LINE : ERROR_LINE);
         break;

    default:
         read_params = sscanf(buffer, "%s %lf", string, value);
         status = (read_params == 2 ? NORMAL_LINE : ERROR_LINE);
         break;
   }
 return status;
}


/*
**  Function  : Boolean read_traff_relations(char *filename)
**  Return    : TRUE if the traffic relation file is correctly read.
**  Remarks   : a list of traffic relations for the users is read from a
**              suitable file (usually its extension is .TRF).
**              All the operations are done with the precision of PRECISION
**              (usually 1E-4). In order to use better precision, the
**              code must be modified.
*/
Boolean read_traff_relations(char *filename)
{
 register int i, j;
/* double temp_mat[MAX_NET_USER][MAX_NET_USER];*/
 double load_fraction, source_load;
 int    no_of_destinations[MAX_NET_USER];
 double load_of_destinations[MAX_NET_USER];
 char   line_buffer[256], name[256];

 FILE *fptr;
 int line_num, line_status, source_user_id, user_id, num_of_sources;
 Boolean hunt_source_line;

/* Only the non-zero values of the traffic matrix have to be assigned.  */
/* The reference way for describing the traffic contributions is        */
/* the following one:                                                   */
/*
    @ <source user name> <fraction of generated traffic>
        <destination user name> <fraction of received traffic>
        <destination user name> <fraction of received traffic>
         .....
        <destination user name> <fraction of received traffic>
    @ <source user name> <fraction of generated traffic>
         .....
                                                                        */
/* The basic conditions on the values assigned to the fractions of      */
/* generated and received traffic impose that the sum of the received   */
/* traffic over the destinations of each source user equals 1.0 and     */
/* that the sum of the generated traffic over all the sources           */
/* equals 1.0 .                                                         */
/* The '@' symbol should be placed in the first column.                 */
/* Comment lines have to begin with a '#' symbol.                       */

 source_user_id = MAX_NET_USER;
 for(i = 0; i < MAX_NET_USER; i++)
    for(j = 0; j < MAX_NET_USER; j++)
       {
        traff_pattern[i][j] = 0.0;
        destination_prob[i][j] = 0.0;
       }
 for(i = 0; i < MAX_NET_USER; i++)
    {
     no_of_destinations[i] = 0;
     load_of_destinations[i] = 0.0;
     source_prob[i] = 0.0;
    }

 if((fptr = fopen(filename, "r")) == NULL)
        return FALSE;

/* Step 1: it is first checked whether the file is formally correct     */
/* (the number of destinations for each source is checked).             */
 line_num = 0;
 num_of_sources = 0;
 hunt_source_line = TRUE;
 source_load = 0.0;
 while(fgets(line_buffer, 256, fptr) != NULL)
   {
    ++line_num;
    line_status = parse_line_traffic(line_buffer, name, &load_fraction);
    switch(line_status)
      {
       case COMMENT_LINE :
         continue;

       case ERROR_LINE :
         printf("Error in line %5d of file %s\n", line_num, filename);
         return FALSE;

       case SOURCE_LINE :
       case NORMAL_LINE :
         if((load_fraction < 0.0) || (load_fraction > 1.0))
            {
             printf("Error - Line %4d of file %s. Invalid load fraction\n",
                        line_num, filename);
             return FALSE;
            }
         if(line_status == SOURCE_LINE)
            {
             source_user_id = lookup_symbol(user_symbol_table, name,
                                no_of_users);
             if(source_user_id == no_of_users)
                {
                 printf("Error - Line %4d of file %s. User %s ",
                                line_num, filename, name);
                 printf("not found in the symbol table\n");
                 return FALSE;
                }
             hunt_source_line = FALSE;
             no_of_destinations[source_user_id] = 0;
             load_of_destinations[source_user_id] = 0.0;
             source_load += load_fraction;
             ++num_of_sources;
            }
         else
            {
             user_id = lookup_symbol(user_symbol_table, name, no_of_users);
             if(user_id == no_of_users)
                {
                 printf("Error - Line %4d of file %s. User %s ",
                                line_num, filename, name);
                 printf("not found in the symbol table\n");
                 return FALSE;
                }
             if(hunt_source_line)
                {
                 printf("Warning - Line %4d of file %s skipped. ",
                                line_num, filename);
                 printf("Missing source user name\n");
                 continue;
                }
             if(user_id == user_vect[source_user_id]->id)
                {
                 printf("Error - Line %4d of file %s. ", line_num, filename);
                 printf("Coincident Source and Destination\n");
                 return FALSE;
                }
             ++no_of_destinations[source_user_id];
             load_of_destinations[source_user_id] += load_fraction;
             if((load_of_destinations[source_user_id] - 1.0) > PRECISION)
                {
                 printf("Error - Line %4d of file %s. ",
                                line_num, filename);
                 printf("Total load fraction exceeding 1.0\n");
                 return FALSE;
                }
            }
         break;

      } /* switch ... */
   } /* while ... */

 /* The sum of the source probabilities should be 1.0 if not only 
    never_closing_ason generators are present */
 if((num_of_sources > no_of_type8_cg) &&
	 (fabs(source_load - 1.0) >= PRECISION ))
   {
    printf("Error:  The total source traffic must sum to 1.0\n");
    printf("        The actual total source traffic is %f\n",source_load);
    printf("        source_load - 1.0 = %g\n",source_load - 1.0);
    printf("        fabs(source_load - 1.0) = %g\n", fabs(source_load - 1.0));
    return FALSE;
   }
 if(num_of_sources < (no_of_users - no_of_type10_cg - no_of_type11_cg
                      - no_of_type12_cg))
   {
    printf("Error:  Missing declaration for %d user\n",
	no_of_users - no_of_type10_cg - no_of_type11_cg - no_of_type12_cg
- num_of_sources);
    return FALSE;
   }
 else if(num_of_sources > no_of_users)
   {
    printf("Error:  Too many source (user) declarations\n");
    return FALSE;
   }

/* Step 2: the structure of the file is safe now. The work can go on.   */
 rewind(fptr);
 line_num = 0;
 hunt_source_line = TRUE;
 while(fgets(line_buffer, 256, fptr) != NULL)
   {
    ++line_num;
    switch(parse_line_traffic(line_buffer, name, &load_fraction))
      {
       case COMMENT_LINE :
         continue;

       case ERROR_LINE :
         printf("Error in line %5d of file %s\n", line_num, filename);
         return FALSE;

       case SOURCE_LINE :
         source_user_id = lookup_symbol(user_symbol_table, name,
				        no_of_users);
         source_load = load_fraction;
         source_prob[source_user_id] = load_fraction;
         hunt_source_line = FALSE;
         if(no_of_destinations[source_user_id] == 0)
            for(i = 0; i < no_of_users; i++)
               {
                if(i == user_vect[source_user_id]->id)
                        continue;
                traff_pattern[source_user_id][i] =
                        load_fraction / (no_of_users - 1);
                destination_prob[source_user_id][i] =
                        1.0 / (no_of_users-1);
               }
         else if(fabs(load_of_destinations[source_user_id] - 1.0) >=
                        PRECISION)
            {
             printf("Error - Line %5d of file %s.\n", line_num, filename);
             printf("        The following destination traffic ");
             printf("fractions must sum to 1.0\n");
             return FALSE;
            }
         break;

       case NORMAL_LINE :
         if(hunt_source_line)
                continue;
         user_id = lookup_symbol(user_symbol_table, name, no_of_users);
         traff_pattern[source_user_id][user_id] =
                        source_load * load_fraction;
         destination_prob[source_user_id][user_id] = load_fraction;
         break;
      } /* switch ... */
   } /* while ... */
 fclose(fptr);

/* Step 3: the presence of duplicated declarations is checked (multiple */
/* declarations would alter the traffic fractions previously evaluated).*/
 source_load = 0.0;
 for(i = 0; i < no_of_users; i++)
        source_load += source_prob[i];
 if((num_of_sources > no_of_type8_cg) &&
	 (fabs(source_load - 1.0) >= PRECISION ))
   {
    printf("Error:  There must be a duplicated source in the TRF file\n");
    return FALSE;
   }

/* Step 4: the presence of invalid traffic relations is checked here	*/
/* (users belonging to different traffic classes cannot exchange data	*/
/* with each other).							*/
 for(i = 0; i < no_of_cgens; i++)
   {
    switch(callgen_vect[i]->type)
	{
	 case CG_TYPE1 :	
	 case CG_TYPE2 :	
	 case CG_TYPE3 :	
	 case CG_TYPE4 :
	    for(j = 0; j < no_of_cgens; j++)
		 if((callgen_vect[j]->type != CG_TYPE10) &&
		      (destination_prob[callgen_vect[i]->ref_user->id]
				       [callgen_vect[j]->ref_user->id] > 0.0))
		   {
		    printf("Error:  Traffic Class mismatch in the TRF file:");
		    printf("\n        User %s cannot send data to User %s\n",
		    		user_name(callgen_vect[i]->ref_user->id),
		    		user_name(callgen_vect[j]->ref_user->id));
		    return FALSE;
		   }
	    break;

	 case CG_TYPE5  :	
	    for(j = 0; j < no_of_cgens; j++)
		 if((callgen_vect[j]->type != CG_TYPE11) &&
		    (destination_prob[callgen_vect[i]->ref_user->id]
				     [callgen_vect[j]->ref_user->id] > 0.0))
		   {
		    printf("Error:  Traffic Class mysmatch in the TRF file:");
		    printf("\n        User %s cannot send data to User %s\n",
		    		user_name(callgen_vect[i]->ref_user->id),
		    		user_name(callgen_vect[j]->ref_user->id));
		    return FALSE;
		   }
	    break;

	 case CG_TYPE6  :	
	 case CG_TYPE7  :	
	 case CG_TYPE8  :	 
	 case CG_TYPE9  :	 
	    for(j = 0; j < no_of_cgens; j++)
		 if((callgen_vect[j]->type != CG_TYPE12) &&
		    (destination_prob[callgen_vect[i]->ref_user->id]
				     [callgen_vect[j]->ref_user->id] > 0.0))
		   {
		    printf("Error:  Traffic Class mysmatch in the TRF file:");
		    printf("\n        User %s cannot send data to User %s\n",
		    		user_name(callgen_vect[i]->ref_user->id),
		    		user_name(callgen_vect[j]->ref_user->id));
		    return FALSE;
		   }
	    break;

	 case CG_TYPE10 :	
	    for(j = 0; j < no_of_cgens; j++)
		 if(destination_prob[callgen_vect[i]->ref_user->id]
				    [callgen_vect[j]->ref_user->id] > 0.0)
		   {
		    printf("Error:  Traffic Class mismatch in the TRF file:");
		    printf("\n        User %s (SINK type) cannot send ",
		    		user_name(callgen_vect[i]->ref_user->id));
		    printf("data to any other User\n");
		    return FALSE;
		   }
	    break;

	 case CG_TYPE11 :	
	    for(j = 0; j < no_of_cgens; j++)
		 if(destination_prob[callgen_vect[i]->ref_user->id]
				    [callgen_vect[j]->ref_user->id] > 0.0)
		   {
		    printf("Error:  Traffic Class mismatch in the TRF file:");
		    printf("\n        User %s (TCP_SINK type) cannot send ",
		    		user_name(callgen_vect[i]->ref_user->id));
		    printf("data to any other User\n");
		    return FALSE;
		   }
	    break;

	 case CG_TYPE12 :	
	    for(j = 0; j < no_of_cgens; j++)
		 if(destination_prob[callgen_vect[i]->ref_user->id]
				    [callgen_vect[j]->ref_user->id] > 0.0)
		   {
		    printf("Error:  Traffic Class mismatch in the TRF file:");
		    printf("\n        User %s (ASON_SINK type) cannot send ",
		    		user_name(callgen_vect[i]->ref_user->id));
		    printf("data to any other User\n");
		    return FALSE;
		   }
	    break;
	}
    }


 /* The resulting Traffic Matrix is printed on the standard output      */
 /* stream if the -i option was part of the ANCLES command line.	*/

 if(init_debug_level & ND_TRAFFIC_DEBUG_LEVEL)
   {
    printf("\n >>> TRAFFIC MATRIX <<<\n\n");
    for(i = 0; i < no_of_users; i++)
        {
         for(j = 0; j < no_of_users; j++)
                printf("%6.4f ", destination_prob[i][j]);
         printf("\n");
        }
    printf("\n");
   }

/* Translate the destination_prob from user-user relation to            */
/*   cgen-cgen relation                                                 */
/* Why??? ZSZ *

 for(i = 0; i < no_of_cgens ; i++)
    for(j = 0; j < no_of_cgens; j++)
        temp_mat[i][j] = destination_prob[callgen_vect[i]->ref_user->id] 
                   [callgen_vect[j]->ref_user->id];

/* Evaluate the cumulative distribution *
 for(i = 0; i < no_of_cgens; i++)
   {
    destination_prob[i][0] = temp_mat[i][0];
    for(j = 1; j < no_of_cgens; j++)
       {
        temp_mat[i][j] += temp_mat[i][j - 1];
        destination_prob[i][j] = temp_mat[i][j];
       }
   }

 for(i = 0; i < no_of_cgens; i++)
   {
    if((destination_prob[i][no_of_cgens - 1] > 0.0) &&
       (fabs(destination_prob[i][no_of_cgens - 1] - 1.0) >= PRECISION))
      {
       printf("Error:  There must be a duplicated destination ");
       printf("in the TRF file\n");
       return FALSE;
      }
   }
*/
/* The resulting Source Probability Vector and Destination Probability  */
/* Matrix are printed on the standard output stream if the -i option    */
/* was part of the ANCLES command line.                                 *
 if(init_debug_level & ND_TRAFFIC_DEBUG_LEVEL)
   {
    printf(" >>> SOURCE PROBABILITY VECTOR <<<\n\n");
    for(j = 0; j < no_of_cgens; j++)
        printf("%6.4f ", source_prob[j]);
    printf("\n\n");
    printf(" >>> DESTINATION PROBABILITY MATRIX <<<\n\n");
    for(i = 0; i < no_of_cgens; i++)
        {
         for(j = 0; j < no_of_cgens; j++)
                printf("%6.4f ", destination_prob[i][j]);
         printf("\n");
        }
    printf("\n");
   } */

/* Use only the probabilities indiced by the _users_!!! ZSZ */

 /* Evaluate the cumulative distribution */
 for(i = 0; i < no_of_users; i++)
    for(j = 1; j < no_of_users; j++)
        destination_prob[i][j] += destination_prob[i][j - 1];

 for(i = 0; i < no_of_users; i++)
   {
    if((destination_prob[i][no_of_users - 1] > 0.0) &&
       (fabs(destination_prob[i][no_of_users - 1] - 1.0) >= PRECISION))
      {
       printf("Error:  There must be a duplicated destination ");
       printf("in the TRF file\n");
       return FALSE;
      }
   }

/* The resulting Source Probability Vector and Destination Probability  */
/* Matrix are printed on the standard output stream if the -i option    */
/* was part of the ANCLES command line.                                 */
 if(init_debug_level & ND_TRAFFIC_DEBUG_LEVEL)
   {
    printf(" >>> SOURCE PROBABILITY VECTOR <<<\n\n");
    for(j = 0; j < no_of_users; j++)
        printf("%6.4f ", source_prob[j]);
    printf("\n\n");
    printf(" >>> DESTINATION PROBABILITY MATRIX <<<\n\n");
    for(i = 0; i < no_of_users; i++)
        {
         for(j = 0; j < no_of_users; j++)
                printf("%6.4f ", destination_prob[i][j]);
         printf("\n");
        }
    printf("\n");
   }
 
 
 
 return TRUE;
}


/*
**  Function  : void create_uniform_traffic(void)
**  Return    : none.
**  Remarks   : a uniform traffic matrix is created when no TRF file is
**              defined. For every valid connection (1,2,3,4 -> 10,
**		5 -> 11 and 6->12) the same probability will be assigned
*/
void create_uniform_traffic(void)
{
 register int      i, j;
 int allcons = 0;  /* Number of all the valid connections in the network */

 for(i = 0; i < no_of_users; i++)
  {
   source_prob[i] = 0.0;
   for(j = 0; j < no_of_users; j++)
     {
      traff_pattern[i][j] = 0.0;
	  destination_prob[i][j] = 0.0;
     }
  }

 /* Counting the valid connections. */
 for(i = 0; i < no_of_cgens; i++)
   {
    switch(callgen_vect[i]->type)
	{
	 case CG_TYPE1 :	
	 case CG_TYPE2 :	
	 case CG_TYPE3 :	
	 case CG_TYPE4 :
	   for(j = 0; j < no_of_cgens; j++)
		if(callgen_vect[j]->type == CG_TYPE10 &&
		   callgen_vect[i]->ref_user->connected_node !=
		   callgen_vect[j]->ref_user->connected_node)
		 {
		 /* counting the connections starting from the source */
		  source_prob[callgen_vect[i]->ref_user->id] += 1.0;
		 /* Indicate if a connection (src-dst) exists */
		  traff_pattern[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  destination_prob[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  allcons++;
		 }
	   break;

	 case CG_TYPE5 :
	   for(j = 0; j < no_of_cgens; j++)
		if(callgen_vect[j]->type == CG_TYPE11 &&
		   callgen_vect[i]->ref_user->connected_node !=
		   callgen_vect[j]->ref_user->connected_node)
		 {
		 /* counting the connections starting from the source */
		  source_prob[callgen_vect[i]->ref_user->id] += 1.0;
		 /* Indicate if a connection (src-dst) exists */
		  traff_pattern[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  destination_prob[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  allcons++;
		 }
	   break;

	 case CG_TYPE6 :
	 case CG_TYPE9 :
	   for(j = 0; j < no_of_cgens; j++)
		if(callgen_vect[j]->type == CG_TYPE12 &&
		   callgen_vect[i]->ref_user->connected_node !=
		   callgen_vect[j]->ref_user->connected_node)
		 {
		 /* counting the connections starting from the source */
		  source_prob[callgen_vect[i]->ref_user->id] += 1.0;
		 /* Indicate if a connection (src-dst) exists */
		  traff_pattern[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  destination_prob[callgen_vect[i]->ref_user->id]
	              [callgen_vect[j]->ref_user->id] = 1.0;
		  allcons++;
		 }
	   break;

	 case CG_TYPE7:
	 case CG_TYPE8:
	     fprintf(stderr, "The network contains CG_TYPE7 or CG_TYPE8 type");
	     fprintf(stderr, " call generators,\n");
	     fprintf(stderr, "whose traffic must be given in a trf file.\n");
	     exit(1);

	 case CG_TYPE10 :	
	 case CG_TYPE11 :
	 case CG_TYPE12 :
	   break;
	 
	 default : 
	     fprintf(stderr, "Invalid call generator type in create_uniform_traffic()\n");
	     exit(1);
        }
      }
 /* Set the probabilities */
 if(allcons)
  {
   for(i = 0; i < no_of_cgens; i++)
     {
      if(source_prob[callgen_vect[i]->ref_user->id])
       {
	for(j = 0; j < no_of_cgens; j++)
	 {
	  traff_pattern[callgen_vect[i]->ref_user->id]
	               [callgen_vect[j]->ref_user->id] /= (double) allcons;

	  destination_prob[callgen_vect[i]->ref_user->id]
	        	  [callgen_vect[j]->ref_user->id] /=
	    source_prob[callgen_vect[i]->ref_user->id];
	 }
	source_prob[callgen_vect[i]->ref_user->id] /= (double) allcons;
       }
     }
  }

/* The destination matrix is transformed here into a cumulative one.    */
/* The resulting Traffic Matrix is printed on the standard output       */
/* stream if the -i option was part of the ANCLES command line.         */
 if(init_debug_level & ND_TRAFFIC_DEBUG_LEVEL)
   {
    printf("\n >>> TRAFFIC MATRIX <<<\n\n");
    for(i = 0; i < no_of_users; i++)
      {
       for(j = 0; j < no_of_users; j++)
                printf("%6.4f ", destination_prob[i][j]);
       printf("\n");
      }
    printf("\n");
   }

 for(i = 0; i < no_of_users; i++)
    for(j = 1; j < no_of_users; j++)
        destination_prob[i][j] += destination_prob[i][j - 1];

/* The resulting Source Probability Vector and Destination Probability  */
/* Matrix are printed on the standard output stream if the -i option    */
/* was part of the ANCLES command line.                                 */
 if(init_debug_level & ND_TRAFFIC_DEBUG_LEVEL)
   {
    printf(" >>> SOURCE PROBABILITY VECTOR <<<\n\n");
    for(j = 0; j < no_of_users; j++)
        printf("%6.4f ", source_prob[j]);
    printf("\n\n");
    printf(" >>> DESTINATION PROBABILITY MATRIX <<<\n\n");
    for(i = 0; i < no_of_users; i++)
      {
       for(j = 0; j < no_of_users; j++)
                printf("%6.4f ", destination_prob[i][j]);
       printf("\n");
      }
    printf("\n");
   }
}


