/*
**  Module    : nd_yacc.y
**  Version   : $Revision: 3.1 $
**  Date      : $Date: 1998/07/28 09:51:21 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains the yacc definitions of the ND
**              language. It should be compiled with
**                      yacc -d nd_yacc.y
**              in order to produce the header file to be used with the
**              module
**                      nd_lex.l
**
**	Changes   : support the explicitely given routes for routing
**							- 2002.04.23. Zoltan Zsoka
**
*/
 
/*
**  C language prologue: #include & variable definition
*/
%{
#include "nd.h"
 
extern char *yytext;
extern int yylex(void);
void yyerror(char *);

extern struct nd_instance* global_instance_list;
 
Parse *local_list = NULL,
      *global_node_list = NULL,
      *global_link_list = NULL,
      *global_user_list = NULL,
      *global_route_list = NULL,
      *global_cgen_list = NULL;

int pnode_num = 0,
    plink_num = 0,
    puser_num = 0,
    pcgen_num = 0,
    proute_num = 0;
enum nd_parameter_context context = NDC_GLOBAL;
int element_identifier;
static Boolean declared_connection;
static Boolean declared_reference;
static Boolean constant_is_vector;

%}
 
/*
**  YACC preliminary definitions: %union, %token and %type
*/
 
%union 
{
  int id;
  struct nd_constant constant;
  char* string;
}
 
%token DEFAULT_TOKEN
%token DEMAND_TOKEN
%token PARAMETER_TOKEN
%token NODE_TOKEN
%token USER_TOKEN
%token LINK_TOKEN
%token CALL_GENERATOR_TOKEN
%token ROUTE_TOKEN
%token REFERRING_TO_TOKEN
%token CONNECTED_TO_TOKEN
%token USED_BY_TOKEN
%token TYPE_TOKEN
%token THRU_TOKEN
%token WITH_TOKEN
%token ENDWITH_TOKEN
%token LINKLIST_TOKEN
%token FROM_NODE_TOKEN
%token TO_NODE_TOKEN
%token ASSIGN_TOKEN
%token ENDCLAUSE_TOKEN
%token SEPARATOR_TOKEN
%token NUMBER_TOKEN
%token STRING_TOKEN
%token BRACE_LEFT_TOKEN
%token BRACE_RIGHT_TOKEN
%token MINUS_TOKEN
 
%type <constant> scalar_constant constant NUMBER_TOKEN realvect_sequence

%type <id> node_symbolic_identifier user_symbolic_identifier
%type <id> link_symbolic_identifier call_gen_symbolic_identifier
%type <id> route_symbolic_identifier

%type <string> STRING_TOKEN

/*
**  YACC language definition
*/
%%
 
/*
**  program            
**      : global_declarations net_declarations
*/
program         
	: global_declarations net_declarations
           {
	     int i;
	     if(init_debug_level & ND_DEBUG_LEVEL)
	       nd_print_debug();

	     ask_for_demand();

	     if(!check_missing_parameters())
	       YYABORT;

	     /* Store global parameters in ancles global variables. */
	     init_variables();

	     if(!check_values())
	       YYABORT;

	     build_node_network();

	     if(init_debug_level & ND_DEBUG_LEVEL)
	       printf("Execution build_node_network()     OK\n");

	     add_users_to_nodes();

	     if(init_debug_level & ND_DEBUG_LEVEL)
	       printf("Execution add_users_to_nodes()     OK\n");

	     add_call_gens_to_users();
	     if(init_debug_level & ND_DEBUG_LEVEL)
	       printf("Execution add_call_gens_to_users() OK\n");

	     if(init_debug_level & ND_DEBUG_LEVEL)
	       {
                 printf("\n --- LINK IDENTIFIERS ---\n\n");
                 for(i = 0; i < plink_num; i++)
		   printf("Link %2d is %s\n", i, link_name(i));
	       }

	     /* Build the structures for nodes, links, users and call
                generators. */
	     initialize_ANCLES_structures();
	   }
	;       /* program */
 
net_declarations
	: net_clause
	| net_declarations net_clause
	;
 
net_clause
	: node_clause
	| link_clause
	| user_clause
	| call_generator_clause
	| route_clause
	;
 
global_declarations
	: /* empty */
	| global_declarations glob_decl_clause
        ;
 
glob_decl_clause
	: default_clause
	| demand_clause
	;
 
default_clause  
	: DEFAULT_TOKEN param_assign_list ENDCLAUSE_TOKEN
	;

/* Analogous to 'default_clause' but can only appear in a local
   context. */
parameter_clause
	: PARAMETER_TOKEN param_assign_list ENDCLAUSE_TOKEN
	;
 
param_assign_list
	: param_assignment
	| param_assign_list SEPARATOR_TOKEN param_assignment
	;
 
/* 
** param_assignment
**     : STRING_TOKEN ASSIGN_TOKEN constant
**     | constant
**     ;
*/
param_assignment
	: STRING_TOKEN ASSIGN_TOKEN constant
	   {
	     struct nd_parameter* parameter;
	     struct nd_instance* il;
	     Boolean incompatible_types;

	     /* Check if the parameter exists. */
	     parameter = lookup_parameter_by_name($1);
	  
	     if (parameter == NULL)
	       {
		 fprintf(stderr, "Line %3d: unknown parameter `%s'\n", linenum, $1);
		 YYABORT;
	       }

	     free($1);

	     incompatible_types = FALSE;

	     /* Check scalar to vector and vector to scalar
		assignments. */
	     if (((parameter->flags & NDF_VECTOR) && (! constant_is_vector))
		 || ((! (parameter->flags & NDF_VECTOR)) && constant_is_vector))
	       incompatible_types = TRUE;

	     if (incompatible_types)
	       {
		 fprintf(stderr, 
			 "Line %3d: scalar to vector assignment\n",
			 linenum);
		 YYABORT;
	       }

	     if (parameter->flags & NDF_ENUMERATION)
	       {
		 if (parameter->type != $3.type)
		   incompatible_types = TRUE;
	       }
	     else /* Parameter is numeric */
	       {
		 if ($3.type != BASE_TYPE($3.type))
		   {
		     if (parameter->type != $3.type)
		       incompatible_types = TRUE;
		   }
		 else /* Constant has one of the basic types */
		   {
		     /* Of the two assignments that mix basic types
			only the one that assign a real value to an
			integer parameter is forbidden.  The one that
			assign an integer value to a real parameter is
			performed with a conversion.  */
		     if ($3.type == NDT_REAL
			 && BASE_TYPE(parameter->type) == NDT_INTEGER)
		       incompatible_types = TRUE;
		     else if ($3.type == NDT_INTEGER
			      && BASE_TYPE(parameter->type) ==
			      NDT_REAL)
		       {
			 if (parameter->flags & NDF_VECTOR)
			   vector_integer_to_real($3.value.vector);
			 else
			   $3.value.real = (double)$3.value.integer;
		       }
		   }
	       }

	     if (incompatible_types)
	       {
		 fprintf(stderr, 
			 "Line %3d: incompatible types in parameter assignment\n",
			 linenum);
		 YYABORT;
	       }

	     /* Assign the constant value to the instance of this
		parameter. */
	     il = (context == NDC_GLOBAL) 
	       ? global_instance_list : local_list->instance_list;

	     /* NOTE: when assigning a vector assign_parameter()
                doesn't allocate a new one. */
	     if (! assign_parameter(il, parameter, $3.value, context))
	       {
		 fprintf(stderr, "Line %3d: invalid context\n", linenum);
		 YYABORT;
	       }
	   }

	| constant
	   {
	     struct nd_parameter* parameter;
	     struct nd_instance* il;

	     /* Check if this constant type is enough to go back to an
		unique parameter. */
	     parameter = lookup_parameter_by_type($1.type);
	     if (parameter == NULL)
	       {
		 fprintf(stderr, "Line %3d: lone constant `%s'\n", 
			 linenum, $1.name);
		 YYABORT;
	       }

	     /* Assign the constant value to the instance of the
		corresponding parameter. */
	     il = (context == NDC_GLOBAL) 
	       ? global_instance_list : local_list->instance_list;

	     /* NOTE: when assigning a vector assign_parameter()
                doesn't allocate a new one. */
	     if (! assign_parameter(il, parameter, $1.value, context))
	       {
		 fprintf(stderr, "Line %3d: invalid context\n", linenum);
		 YYABORT;
	       }
	   }
	;

constant
	: scalar_constant
	   {
	     $$ = $1;

	     constant_is_vector = FALSE;
	   }

        | MINUS_TOKEN scalar_constant
          {
	    $$ = $2;

	    if (BASE_TYPE($2.type) == NDT_REAL)
	      $$.value.real = - $$.value.real;
	    else if (BASE_TYPE($2.type) == NDT_INTEGER)
	      $$.value.integer = - $$.value.integer;
	  }
      
	| BRACE_LEFT_TOKEN realvect_sequence BRACE_RIGHT_TOKEN
          {
	    $$ = $2;

	    constant_is_vector = TRUE;
          }
        ;

scalar_constant
	: NUMBER_TOKEN

        | STRING_TOKEN
	   {
	     struct nd_constant* constant;
	     constant = lookup_constant_by_name($1);

	     if (constant)
	       $$ = *constant;
	     else
	       {
		 fprintf(stderr, "Line %3d: unknown constant `%s'\n", linenum, $1);
		 YYABORT;
	       }

	     free($1);
	   }
        ;

realvect_sequence
	: scalar_constant
	   {
	     $$.value.vector = create_vector();

	     $$.value.vector = 
	       insert_vector_element($$.value.vector, 0, $1.value);

	     $$.name = "real vector";
	     $$.type = $1.type;
	   }

	| realvect_sequence SEPARATOR_TOKEN scalar_constant
	   {
	     if ($3.type != $$.type)
	       {
		 if ($$.type == NDT_INTEGER
		     && BASE_TYPE($3.type) == NDT_INTEGER)
		   {
		     /* "Integer -> derived integer" conversion */
		     $$.type = $3.type;
		   }
		 else if ($$.type == NDT_INTEGER
			   && BASE_TYPE($3.type) == NDT_REAL)
		   {
		     /* "Integer -> real" and "Integer -> derived
                        real" conversion */
		     $$.type = $3.type;
		     vector_integer_to_real($$.value.vector);
		   }
		 else if ($$.type == NDT_REAL
			  && BASE_TYPE($3.type) == NDT_REAL)
		   {
		     /* "Real -> derived real" conversion */
		     $$.type = $3.type;
		   }
		 else if (BASE_TYPE($$.type) == NDT_REAL
			  && $3.type == NDT_INTEGER)
		   {
		     /* An integer value is always allowed in a real
                        vector. */
		     $3.value.real = (double)$3.value.integer;
		   }
		 else
		   {
		     fprintf(stderr, 
			     "Line %3d: incompatible elements in vector definition\n",
			     linenum);

		     YYABORT;
		   }
	       }
	     
	     $$.value.vector = 
	       insert_vector_element($$.value.vector, $$.value.vector->len, $3.value);
	   }
	;

demand_clause
	: DEMAND_TOKEN param_name_list ENDCLAUSE_TOKEN
        ;
 
 
param_name_list 
	: parameter_name
	| param_name_list SEPARATOR_TOKEN parameter_name
	;
 
/*
** parameter_name  
**     : STRING_TOKEN
*/
parameter_name  
	: STRING_TOKEN
           {
	     struct nd_parameter* parameter;

	     /* Check if the parameter exists. */
	     parameter = lookup_parameter_by_name($1);

	     if (parameter == NULL)
	       {
		 fprintf(stderr,
			 "Line %3d: undefined parameter `%s'\n", linenum, $1);
		 YYABORT;
	       }

	     free($1);

	     /* Check if this parameter can appear in a demand
                statement. */
	     if (parameter->flags & NDF_NO_DEMAND)
	       {
		 fprintf(stderr,
			 "Line %3d: Parameter `%s' cannot appear "
			 "in a demand statement\n",
			 linenum,
			 parameter->name);

		 YYABORT;
	       }

	     /* Put the instance of this parameter in NDS_DEMAND
		status. 

		A duplicated demand statement for the same parameter
		is an error in the global context, but is tolerated in
		local contexts. */
	     if (context == NDC_GLOBAL)
	       {
		 Boolean success = FALSE;

		 success = assign_demand(global_instance_list, parameter);

		 if ( !success)
		   {
		     fprintf(stderr,
			     "Warning:  Duplicated demand "
			     "in line %d for '%s'\n",
			     linenum, parameter->name);
		   }
	       }
	     else
	       assign_demand(local_list->instance_list, parameter);
	   }
	;
 

/*****************
****************** Node rules
*****************/

node_clause
	: NODE_TOKEN node_id
	  WITH_TOKEN clauses_list_node ENDWITH_TOKEN
	  ENDCLAUSE_TOKEN
           {
	     if (! in_list_parse(&global_node_list, local_list))
	       {
		 print_error(local_list, "duplicated node declaration");
		 YYABORT;
	       }
	   }
	;
 
node_id
	: node_symbolic_identifier
           {
	     context = NDC_NODE;
	     local_list = init_parsed_node($1);
	     ++pnode_num;
	   }
	;
 
node_symbolic_identifier
	: STRING_TOKEN
           {
	     element_identifier = insert_symbol(node_symbol_table, $1, MAX_NODE);

	     if (element_identifier == MAX_NODE)
	       {
		 fprintf(stderr, 
			 "Line %3d: Too many nodes defined "
			 "Exceeded symbol table capacity\n",
			 linenum);

		 YYABORT;
	       }
	     $$ = element_identifier;
	   }
	;
 
clauses_list_node
	: /* empty */
	| clauses_list_node clauses_elem_node
	;
 
clauses_elem_node
	: connection_clause
	| parameter_clause
	| demand_clause
	| used_by_clause
	;

connection_clause
	: CONNECTED_TO_TOKEN node_symbolic_identifier THRU_TOKEN
	  link_symbolic_identifier ENDCLAUSE_TOKEN
           {
	     add_connection(local_list, $2, $4);
	   }
	;
 
used_by_clause
	: USED_BY_TOKEN user_symbolic_identifier THRU_TOKEN
	  link_symbolic_identifier ENDCLAUSE_TOKEN
           {
	     add_usage(local_list, $2, $4);
	   }
	;


/*****************
****************** Link rules
*****************/

link_clause
	: LINK_TOKEN link_id WITH_TOKEN clauses_list_link
	  ENDWITH_TOKEN ENDCLAUSE_TOKEN
           {
	     if (!in_list_parse(&global_link_list, local_list))
	       {
		 print_error(local_list, "duplicated link declaration");
		 YYABORT;
	       }
	   }
	;
 
link_id
	: link_symbolic_identifier
           {
	     context = NDC_LINK;
	     local_list = init_parsed_link($1);
	     ++plink_num;
	   }
	;
 
link_symbolic_identifier
	: STRING_TOKEN
           {
	     element_identifier = insert_symbol(link_symbol_table, $1, MAX_LINK);

	     if (element_identifier == MAX_LINK)
	       {
		 fprintf(stderr, 
			 "Line %3d: Too many links defined "
			 "Exceeded symbol table capacity\n",
			 linenum);

		 YYABORT;
	       }
	     $$ = element_identifier;
	   }
	;
 
clauses_list_link
	: /* empty */
	| clauses_list_link clauses_elem_link
	; 
 
clauses_elem_link
	: parameter_clause
	| demand_clause
	;


/*****************
****************** User rules
*****************/

user_clause
	: USER_TOKEN user_id WITH_TOKEN clauses_list_user
	  ENDWITH_TOKEN ENDCLAUSE_TOKEN
           {
	     if (!declared_connection)
	       {
		 print_error(local_list, "user must be connected to a node");
		 YYABORT;
	       }
	     if (!in_list_parse(&global_user_list,local_list))
	       {
		 print_error(local_list, "duplicated user declaration");
		 YYABORT;
	       }
	   }
	;
 
user_id
	: user_symbolic_identifier
           {
	     context = NDC_USER;
	     local_list = init_parsed_user($1);
	     declared_connection = FALSE;
	     ++puser_num;
	   }
	; 
 
user_symbolic_identifier
	: STRING_TOKEN
           {
	     element_identifier = insert_symbol(user_symbol_table, $1, MAX_NET_USER);

	     if (element_identifier == MAX_NET_USER)
	       {
		 fprintf(stderr, 
			 "Line %3d: Too many users defined. "
			 "Exceeded symbol table capacity\n",
			 linenum);

		 YYABORT;
	       }
	     $$ = element_identifier;
	   }
	; 
 
clauses_list_user
	: /* empty */
	| clauses_list_user clauses_elem_user
	; 
 
clauses_elem_user
	: connection_clause
           {
	     if (declared_connection)
	       {
		 fprintf(stderr, 
			 "Line %3d: " 
			 "The User can be connected to one node only\n",
			 local_list->decl_line);

		 YYABORT;
	       }
	     else
	       declared_connection = TRUE;
	   }
	| parameter_clause
	| demand_clause
	;
 

/*****************
****************** Call generator rules
*****************/
 
call_generator_clause 
	: CALL_GENERATOR_TOKEN call_generator_id WITH_TOKEN
	  type_entry clauses_list_call_gen ENDWITH_TOKEN
	  ENDCLAUSE_TOKEN
           {
	     if (!declared_reference)
	       {
		 print_error(local_list, "call generator must referr to a user");
		 YYABORT;
	       }
	     if (!in_list_parse(&global_cgen_list, local_list))
	       {
		 print_error(local_list, "duplicated call generator declaration");
		 YYABORT;
	       }
	   }
	;

call_generator_id
	: call_gen_symbolic_identifier
           {
	     local_list = init_parsed_cgen($1);

	     declared_reference = FALSE;
	     ++pcgen_num;
	   }
	;

call_gen_symbolic_identifier 
	: STRING_TOKEN
           {
	     element_identifier = insert_symbol(callgen_symbol_table, $1, MAX_CALL_GEN);

	     if (element_identifier == MAX_CALL_GEN)
	       {
		 fprintf(stderr,
			 "Line %3d: Too many call generators "
			 "defined. Exceeded symbol table capacity\n",
			 linenum);

		 YYABORT;
	       }
	     $$ = element_identifier;
	   }
	; 

type_entry
	: /* empty */
           {
	     struct nd_instance* instance;

	     /* If no type is specified for this call generator then
		inherit it from the global parameter
		'call_generator_type'. */
	     instance = lookup_instance_by_name(global_instance_list, 
						"call_generator_type");

	     if (instance->status == NDS_ASSIGNED)
	       {
		 local_list->elem.gener->type = instance->value.integer;
		 context = set_cgen_context(instance->value.integer);
		 local_list->instance_list = 
		   create_default_instances(context);
	       }
	     else
	       {
		 print_error(local_list, "missing type definition for call generator");
		 YYABORT;
	       }

	   }
 
	| TYPE_TOKEN scalar_constant ENDCLAUSE_TOKEN
           {
	     if ($2.type == NDT_GENERATOR_TYPE)
	       {
		 local_list->elem.gener->type = $2.value.integer;
		 context = set_cgen_context($2.value.integer);
		 local_list->instance_list = 
		   create_default_instances(context);
	       }
	     else
	       {
		 fprintf(stderr, 
			 "Line %3d: Undefined call generator type '%s'\n",
			 linenum, $2.name); 

		 YYABORT;
	       }
	   }
	; 

clauses_list_call_gen
	: /* empty */
	| clauses_list_call_gen clauses_elem_call_gen
	;

clauses_elem_call_gen
	: referring_clause
	   {
	     if (declared_reference)
	       {
		 fprintf(stderr,
			 "Line %3d: The Call "
			 "Generator can be connected to one user only\n",
			 local_list->decl_line);

		 YYABORT;
	       }
	     else
	       declared_reference = TRUE;
	   }
	| parameter_clause
	| demand_clause
	;
 
referring_clause
	: REFERRING_TO_TOKEN user_symbolic_identifier ENDCLAUSE_TOKEN
	   {
	     add_referring(local_list, $2);
	   }
	; 

/*****************
****************** Route rules
*****************/
/* this part is inserted for explicit routing by Z. Zsoka 2002 */

route_clause
	: ROUTE_TOKEN route_id
	  WITH_TOKEN clauses_list_route ENDWITH_TOKEN
	  ENDCLAUSE_TOKEN
           {
	     if (! in_list_parse(&global_route_list, local_list))
	       {
		 print_error(local_list, "duplicated route declaration");
		 YYABORT;
	       }
	   }
	;
 
route_id
	: route_symbolic_identifier
           {
	     context = NDC_ROUTE;
	     local_list = init_parsed_route($1);
	     ++proute_num;
	   }
	;
 
route_symbolic_identifier
	: STRING_TOKEN
           {
	     element_identifier = insert_symbol(route_symbol_table, $1, MAX_ROUTE);

	     if (element_identifier == MAX_ROUTE)
	       {
		 fprintf(stderr, 
			 "Line %3d: Too many routes defined "
			 "Exceeded symbol table capacity\n",
			 linenum);

		 YYABORT;
	       }
	     $$ = element_identifier;
	   }
	;
 
clauses_list_route
	: /* empty */
	| clauses_list_route clauses_elem_route
	;
 
clauses_elem_route
	: roulinks_clause
	| parameter_clause
	| demand_clause
	| from_node_clause
	| to_node_clause
	;

roulinks_clause
	: LINKLIST_TOKEN  linklist_clause  ENDCLAUSE_TOKEN
	;

linklist_clause
	: link_symbolic_identifier SEPARATOR_TOKEN linklist_clause
           {
	     add_link_to_route(local_list, $1);
	   }
	| link_symbolic_identifier
           {
	     add_link_to_route(local_list, $1);
	   }
	;
 
from_node_clause
	: FROM_NODE_TOKEN node_symbolic_identifier ENDCLAUSE_TOKEN
           {
	     add_fn_to_route(local_list, $2);
	   }
	;

to_node_clause
	: TO_NODE_TOKEN node_symbolic_identifier ENDCLAUSE_TOKEN
           {
	     add_tn_to_route(local_list, $2);
	   }

 
%%
/*
**  C language epilogue: yerror(char *) definition
*/
void yyerror(char *err_msg)
{
 char string[255];

 strcpy(string,err_msg);  /* Copy the error message in a local buffer       */
   /* In fact, there was a problem with the Linux and Sun version of yacc   */
   /* (or with gcc, I really don't know). On these systems, the yyerror     */
   /* argument is a 'static char *': for this reason, probably, the         */
   /* following toupper() instruction failed with a Segmentation Fault.     */

 string[0]=toupper(string[0]);

 

/* The error messages produced by Yacc are controlled here. The Yacc    */
/* error is reported with the line in which it occurs. A very simple    */
/* mechanism to control the error due to wrong parameters type is       */
/* implemented, but it can give 'slighly' wrong messages when the error */
/* is due to a missing '=' sign in the statement.                       */
 
 string[0] = toupper(string[0]);

 fprintf(stderr, "Line %d: %s.\n", linenum, string); 
}
