/*
**  Module    : photo.c
**  Version   : $Revision: 3.2 $
**  Date      : $Date: 1999/01/14 14:59:44 $
**  Authors   : Andrea Francini, Marco Mellia, Claudio Casetti
**              Gabriele Favalessa, Luca Medico
**
**  Synopsis  : this module contains all the functions dealing with the
**              generation of files to be used as input for CLASS simulations.
**
**  Functions : static Boolean cmpinfo(Call *, Call *)
**              static void create_VC_lists(void)
**              static void write_info(FILE *, Call *)
**              static void write_VCs(FILE *, long *)
**              static void write_head(FILE *, char *, int, int *)
**              static void write_global(FILE *, char *)
**              static void write_demand(FILE *)
**		static void write_nodes(FILE *, long *);
**		static void add_tcp_links_to_node(FILE *, Node *);
**		static void write_users(FILE *, long *);
**		static void write_tcp_users(FILE *);
**		static void write_links(FILE *, long *);
**		static void write_tcp_links(FILE *);
**		static void write_tcp_paths(FILE *);
**              void print_state_photo(void)
**
**  History   : 14-MAR-1996 - Creation of the module.   		-AF-
**              01-JUN-1996 - Revision.                 		-AF-
**		20-JUL-1996 - Addition of CLASS users deriving from 
**			      Class U ANCLES calls.			-AF-
**              01-MAR-1997 - Revision			- MM -
**              09-FEB-1998 - Revision                  - MM - GF -  

01-02-2001 Cleaning -ZZ-
*/
 
# ifndef        ANCLES_H
#include        "ancles.h"
# endif

/***************************
**  Constant Definitions  **
***************************/

#ifndef ACC_CL_DEF              /* Default value of 'accuracy'          */
#define ACC_CL_DEF     0.02     /* in CLASS.                            */
#endif

#ifndef CON_CL_DEF              /* Default value of 'confid_lev'        */
#define CON_CL_DEF      0.95    /* in CLASS.                            */
#endif

#ifndef NOB_CL_DEF                      /* Default value of             */
#define NOB_CL_DEF      10000000        /* 'num_observation' in CLASS.  */
#endif
    
#ifndef MTL_CL_DEF              /* Default value of 'min_trans_len'     */
#define MTL_CL_DEF      5000    /* in CLASS.                            */
#endif

#ifndef TBS_CL_DEF              /* Default value of 'trans_batch_size'  */
#define TBS_CL_DEF      30      /* in CLASS.                            */
#endif

#ifndef TSL_CL_DEF              /* Default value of 'trans_seq_len'     */
#define TSL_CL_DEF      3       /* in CLASS.                            */
#endif

#ifndef TRA_CL_DEF              /* Default value of 'trans_accuracy'    */
#define TRA_CL_DEF      0.5     /* in CLASS.                            */
#endif

#ifndef SBS_CL_DEF              /* Default value of 'stat_batch_size'   */
#define SBS_CL_DEF      10000   /* in CLASS.                            */
#endif

#ifndef UTCP_GRAN_DEF           /* Default value of 'utcp_granularity'  */
#define UTCP_GRAN_DEF   50      /* (CLASS parameter).                   */
#endif

#ifndef UTCP_MAX_WIN            /* Default value of 'utcp_granularity'  */
#define UTCP_MAX_WIN   100      /* (CLASS parameter).                   */
#endif

#ifndef ABR_NRM_DEF             /* Default value of 'abr_nrm'           */
#define ABR_NRM_DEF     32      /* (CLASS parameter).                   */
#endif

#ifndef ABR_RIF_DEF             /* Default value of 'abr_rif'           */
#define ABR_RIF_DEF     1       /* (CLASS parameter).                   */
#endif

#ifndef ABR_RDF_DEF             /* Default value of 'abr_rdf'           */
#define ABR_RDF_DEF     1       /* (CLASS parameter).                   */
#endif

#ifndef ERICA_MEAS_CELLS        /* Default value of                     */
#define ERICA_MEAS_CELLS   100  /* 'erica_measure_interval_cells'.      */
#endif                          /* (CLASS parameter).                   */

#ifndef ERICA_MEAS_TIME         /* Default value of                     */
#define ERICA_MEAS_TIME   1.0   /* 'erica_measure_interval_time'.       */
#endif                          /* (CLASS parameter).                   */


#ifndef ERICA_TARGET_DEF        /* Default value of 'erica_target_util' */
#define ERICA_TARGET_DEF 0.98   /* (CLASS parameter).                   */
#endif

                                 
/***********************
**  Global Variables  **
***********************/
 
 
/**************************
**  Function Prototypes  **
**************************/
static Boolean cmpinfo(Call *, Call *);
static void create_VC_lists(void);
static void write_info(FILE *, Call *);
static void write_VCs(FILE *, long *);
 
static void write_head(FILE *, char *, int);
static void write_global(FILE *, char *);
static void write_demand(FILE *);
static void write_nodes(FILE *, long *);
static void add_tcp_links_to_node(FILE *, Node *);
static void write_users(FILE *, long *);
static void write_tcp_users(FILE *);
static void write_links(FILE *, long *);
static void write_tcp_links(FILE *);
static void write_tcp_paths(FILE *);
static void print_state_photo(char *, int);

/***************************
**  Function Definitions  **
***************************/

/* 
**  Function  : static Boolean cmpinfo(Call *c1, Call *c2)
**  Return    : TRUE if the call pointed by 'c1' has the same parameters as
**              the one pointed by 'c2', FALSE otherwise.
**  Remarks   : the 'Call' fields considered to decide for equality are 
**              only the ones involved in the assignment of parameters 
**              qualifying the CLASS users.
*/
static Boolean cmpinfo(Call *c1, Call *c2)
{
 LinkList* ll1, * ll2;

 if(c1->dest_id != c2->dest_id)
        return FALSE;
 if(c1->call_type != c2->call_type)
        return FALSE;

/*
 if(!c1->path_ptr || !c2->path_ptr || !c1->path_ptr->path || !c2->path_ptr->path)
   return FALSE;
*/

/* Check if all the link of the path are the same or not */
 ll1 = c1->path_ptr->path;
 ll2 = c2->path_ptr->path;
 while((ll1 !=NULL) && (ll2 !=NULL))
   {
     if (ll1->link_elem->id != ll2->link_elem->id)
       break;

     ll1 = ll1->next;
     ll2 = ll2->next;
   }
 if ((ll1 != NULL) || (ll2 != NULL))
   return FALSE;

 switch(c1->call_type)
   {
    case CALL_TYPE1 :
        if(((CallType1 *)c1->info)->cbr_value != 
           ((CallType1 *)c2->info)->cbr_value)
                return FALSE;
        return TRUE;

    case CALL_TYPE2 :
        if(((CallType2 *)c1->info)->peak_rate != 
           ((CallType2 *)c2->info)->peak_rate)
                return FALSE;
        if(((CallType2 *)c1->info)->mean_on_time != 
           ((CallType2 *)c2->info)->mean_on_time)
                return FALSE;
        if(((CallType2 *)c1->info)->mean_off_time != 
           ((CallType2 *)c2->info)->mean_off_time)
                return FALSE;
        return TRUE;

    case CALL_TYPE3 :
        if(((CallType3 *)c1->info)->min_vbr_value != 
           ((CallType3 *)c2->info)->min_vbr_value)
                return FALSE;
        if(((CallType3 *)c1->info)->max_vbr_value != 
           ((CallType3 *)c2->info)->max_vbr_value)
                return FALSE;
        return TRUE;

    case CALL_TYPE4 :
        if(((CallType4 *)c1->info)->higher_bit_rate_value != 
           ((CallType4 *)c2->info)->higher_bit_rate_value)
                return FALSE;
        if(((CallType4 *)c1->info)->medium_bit_rate_value != 
           ((CallType4 *)c2->info)->medium_bit_rate_value)
                return FALSE;
        if(((CallType4 *)c1->info)->lower_bit_rate_value != 
           ((CallType4 *)c2->info)->lower_bit_rate_value)
                return FALSE;
        if(((CallType4 *)c1->info)->higher_bit_rate_prob != 
           ((CallType4 *)c2->info)->higher_bit_rate_prob)
                return FALSE;
        if(((CallType4 *)c1->info)->medium_bit_rate_prob != 
           ((CallType4 *)c2->info)->medium_bit_rate_prob)
                return FALSE;
        if(((CallType4 *)c1->info)->lower_bit_rate_prob != 
           ((CallType4 *)c2->info)->lower_bit_rate_prob)
                return FALSE;
        return TRUE;

/* For calls belonging to Class U (coming from CG_TYPE5 call            */
/* generators) the comparison in not needed, because no aggregate       */
/* users are available in CLASS for trunks of UBR connections.          */
/* There is a one-to-one mapping between ANCLES Class U connections and */
/* CLASS TCP users.                                                     */

    default :
        return FALSE;
   }
}      
 

/* 
**  Function  : static void create_VC_lists(void)
**  Return    : none.
**  Remarks   : the information concerning the connections currently 
**              active is collected into the lists associated to the 
**              source_users. 
*/
static void create_VC_lists(void)
{
 Call   *callscan;
 VCset  *vcscan;
 VCset  *vcaux;
 VCset  *newvc;
 Boolean cmpflag;

 callscan = head_of_call_list->next;
 while(callscan != head_of_call_list)
   {
    if(callscan->call_type == CALL_TYPE5)
        {
         Ucall_is_present = TRUE;
         callscan = callscan->next;
         continue;
        }
    cmpflag = FALSE;
    vcaux = head_of_VC_list[callgen_vect[callscan->source_id]->ref_user->id];
    vcscan = vcaux->next;
    while(vcscan != NULL)                                 
        { 
         if((cmpflag = cmpinfo(callscan, vcscan->call_elem)) == TRUE)
           {
            (vcscan->no_vc)++;
            break;
           }
         vcaux = vcscan;
         vcscan = vcscan->next;                          
        }
    if(cmpflag == FALSE)
        {
         newvc = vcset_alloc();
         newvc->next = NULL;
         newvc->no_vc = 1;
         newvc->call_elem = callscan;
         vcaux->next = newvc;
        }
    callscan = callscan->next;
   }                     
}


/* 
**  Function  : static void write_info(FILE *fout, Call *currcall)
**  Return    : none.
**  Remarks   : the information concerning the VCs whose parameters are 
**              referenced by 'currcall' is reported in the file pointed 
**              by 'fout'.
*/
static void write_info(FILE *fout, Call *currcall)
{ 
 double slot_rate;
       
 switch(currcall->call_type)
    {
     case CALL_TYPE1 :
        fprintf(fout, "%f: ", ((CallType1 *)currcall->info)->cbr_value);
        break;

     case CALL_TYPE2 :
        slot_rate = capacity * 1000000 / (53 * 8);
        fprintf(fout, "%f %f %f: ", 
                ((CallType2 *)currcall->info)->peak_rate,
                ((CallType2 *)currcall->info)->mean_on_time * slot_rate,
                ((CallType2 *)currcall->info)->mean_off_time * slot_rate);
        break;

     case CALL_TYPE3 :
     case CALL_TYPE4 :
        fprintf(fout, "NOT_DEFINED: ");

/* Data about Class U calls are not reported in the ACF file.   */

     default :
        break;
    }
}


/* 
**  Function  : static void write_VCs(FILE *fout, long *VCs_per_us)
**  Return    : none.
**  Remarks   : the information concerning the VCs currently open is
**              reported in the file pointed by 'fout'.
**              'VCs_per_us' is filled with the total number of active 
**              calls for each user.
*/
static void write_VCs(FILE *fout, long *VCs_per_us)
{
 int  uu;
 VCset    *vcscan;
 VCset    *vcaux;
 LinkList *llscan;

 for(uu = 0; uu < no_of_users; uu++)
   {
    if((vcscan = head_of_VC_list[uu]->next) == NULL)
        continue;
    fprintf(fout, "@ %s\n", user_name(uu));
    while(vcscan != NULL)
       {
        fprintf(fout, "  %s: ", user_name(
                     callgen_vect[vcscan->call_elem->dest_id]->ref_user->id));
        if(vcscan->call_elem->path_ptr != NULL)
          {
           for(llscan = vcscan->call_elem->path_ptr->path->next;
	       llscan->next != NULL; llscan = llscan->next)
              {
               if(llscan->next->next == NULL)
                  fprintf(fout, "%s", link_name(llscan->link_elem->id));
               else     
                  fprintf(fout, "%s, ", link_name(llscan->link_elem->id));
              }
           fprintf(fout, ": ");
          }
        else
           fprintf(fout, "NO_PATH: ");
        write_info(fout, vcscan->call_elem);
        fprintf(fout, "%ld VC\n", vcscan->no_vc);
        VCs_per_us[uu] += vcscan->no_vc;
        vcaux = vcscan;
        vcscan = vcscan->next;
        vcset_release(vcaux);                         
       }
   }        
 whole_vcset_release();
}


/* 
**  Function  : static void write_head(FILE *fout, char *f_nam, int linkid)
**  Return    : none.
**  Remarks   : general information concerning the network state whose 
**              photograph is currently generated are placed at the 
**              beginning of the RC file.
**              'f_num' is the serial number of the RC file currently 
**              generated; 
**              'f_nam' is the name of the ND file having configured the  
**              current ANCLES simulation;
**              'li' is an array of indices identifying the links 
**              that are considered to decide the creation of the RC file.
*/
static void write_head(FILE *fout, char *f_nam, int linkid)
{
 int ii;

 fprintf(fout, "/*\n**  File :  %s\n", f_nam);
 fprintf(fout, "**  Snapshot of the network state.\n");
 fprintf(fout, "**  ANCLES configuration file : %s\n", descr_file);

 ii = global_stat_vars[0].number;
/* Data about 'total_carried_load' are issued only if stationary        */
/* reports have been already generated by `Statistics'.                 */
 if((curr_accuracy(ii) != 100.0) && (transient_is_over(ii)))
        {       
         fprintf(fout, "**  Current total carried load         : %.2f Mpbs  ", 
                 curr_mean(ii));
         fprintf(fout, "(%ld samples)\n", 
                 curr_nr_of_samples(ii));
        }

 ii = global_stat_vars[1].number;
/* Data about 'total_blocking_probability' are issued only if           */
/* stationary reports have been already generated by `Statistics'.      */
 if((curr_accuracy(ii) != 100.0) && (transient_is_over(ii)))
        {
         fprintf(fout, "**  Current total blocking probability : %.2E  ",
                 curr_mean(ii));
         fprintf(fout, "(%ld samples)\n", 
                 curr_nr_of_samples(ii));
        }

/* Data about 'link_carried_load' are issued independently of the state */
/* of the related statistical analysis.                                 */
 fprintf(fout, "**  Current load on link %s : %.2f\n", 
         link_name(linkid), link_carried_load[linkid][TOTAL_LINK_LOAD]);
 fprintf(fout, "*/\n");
} 


/* 
**  Function  : static void write_global(FILE *fout, char *acf_name)
**  Return    : none.
**  Remarks   : the information concerning the global parameters of the 
**              network is translated into RC statements and added to the 
**              file pointed by 'fout'.
*/
static void write_global(FILE *fout, char *acf_name)
{
 fprintf(fout, "\n\n/*  Global Parameters  */");

/* ACF file name.               */
 fprintf(fout, "\ndefault acf_file = %s;", acf_name);

/* Link Capacity.               */   
 fprintf(fout, "\ndefault capacity = %f;", capacity);

/* Statistical Parameters.      */
 fprintf(fout, "\n\n/*  Statistical Parameters: the default values of CLASS ");
 fprintf(fout, "are assigned here.  */");
 fprintf(fout, "\n/*  You can modify these values if they are not suited ");
 fprintf(fout, "for your needs.      */");
 fprintf(fout, "\ndefault accuracy = %f;", ACC_CL_DEF);
 fprintf(fout, "\ndefault confid_level = %.3f;", CON_CL_DEF);
 fprintf(fout, "\ndefault num_observation = %d;", NOB_CL_DEF);
 fprintf(fout, "\ndefault min_trans_len = %d;", MTL_CL_DEF);
 fprintf(fout, "\ndefault trans_batch_size = %d;", TBS_CL_DEF);
 fprintf(fout, "\ndefault trans_seq_len = %d;", TSL_CL_DEF);
 fprintf(fout, "\ndefault trans_accuracy = %f;", TRA_CL_DEF);
 fprintf(fout, "\ndefault stat_batch_size = %d;", SBS_CL_DEF);

/* Class U parameters.  */
 if(Ucall_is_present == TRUE)
   {
    fprintf(fout, "\n\n/*  Default values for TYPETCP users  */");
    fprintf(fout, "\n default utcp_granularity = %d; ", UTCP_GRAN_DEF);
    fprintf(fout, "\n/* demand tcp_max_segment_size;*/");
    fprintf(fout, "\ndefault tcp_max_segment_size = 1460;");

    fprintf(fout, "\n/* Only one of the following statement must be present */");
    fprintf(fout, "\n/* default utcp_generator_type = SHAPED; */");
    fprintf(fout, "\n default utcp_generator_type = ABR; ");

    fprintf(fout, "\n /* Default ABR definitions */");
    fprintf(fout, "\n default abr_rif = %d; ", ABR_RIF_DEF);
    fprintf(fout, "\n default abr_rdf = %d; ", ABR_RDF_DEF);
    fprintf(fout, "\n default abr_nrm = %d; ", ABR_NRM_DEF);
    fprintf(fout, "\n default control_policy = ERICA; ");
    fprintf(fout, "\n default erica_measure_interval_cells = %d; ",
                    ERICA_MEAS_CELLS);
    fprintf(fout, "\n default erica_measure_interval_time = %f; ",
                    ERICA_MEAS_TIME);
    fprintf(fout, "\n default erica_target_util = %f; ", ERICA_TARGET_DEF);
    fprintf(fout, "\n/* End of the default ABR definitions */\n");
#ifdef CASTELLI
    fprintf(fout, "\ndefault number_of_classes = 3;");      
    fprintf(fout, "\ndefault buffer_class[] = {64, 1024, 1024 };");
    fprintf(fout, "\ndefault link_min_threshold[] = { 0, 0, 0 } ;");
    fprintf(fout, "\ndefault link_max_threshold[] = { 65, 1025, 1025 };");
    fprintf(fout, "\ndefault vc_stat = TRUE;");
#endif
  }

 fprintf(fout, "\n");
}


/* 
**  Function  : static void write_demand(FILE *fout)
**  Return    : none.
**  Remarks   : the demand clauses concerning the global parameters of the 
**              network are printed in the file pointed by 'fout'.
*/
static void write_demand(FILE *fout)
{
 fprintf(fout, "\n\n/*  Demand Clauses  */");

/* User Buffer Size */   
 fprintf(fout, "\ndemand user_buffer;");
#ifndef CASTELLI
 fprintf(fout, "\ndemand buffer_class[];");
#endif
/* Class U Parameters */
 if(Ucall_is_present == TRUE)
   {
    fprintf(fout, "\n");
    fprintf(fout, "\ndemand utcp_max_window;");
    fprintf(fout, "\n/* default utcp_max_window = %d; */",UTCP_MAX_WIN);
    fprintf(fout, "\n\n/* Comment the following lines if you do not want ABR with");
    fprintf(fout, " the class U connections */");
    fprintf(fout, "\n demand abr_pcr; ");
    fprintf(fout, "\n demand abr_tbe; ");
    fprintf(fout, "\n/* End of the demand ABR definitions */\n");
   }
 fprintf(fout, "\n");
}


/* 
**  Function  : static void write_nodes(FILE *fout, long *VCs_per_us)
**  Return    : none.
**  Remarks   : the information concerning the nodes of the network is
**              translated into RC statements and added to the file 
**              pointed by 'fout'.
*/
static void write_nodes(FILE *fout, long *VCs_per_us)
{
 int        ii, pp;


 fprintf(fout, "\n\n/*  Network Nodes  */");
 for(ii = 0; ii < no_of_nodes; ii++)                                         
   {
    fprintf(fout, "\nnode %s with", node_name(ii));

    for(pp = 0; pp < node_vect[ii]->no_of_ports; pp++)
        fprintf(fout, "\n\tconnected_to %s thru %s;", 
                node_name(node_vect[ii]->outqueue[pp]->dest), 
                link_name(node_vect[ii]->outqueue[pp]->id));

    if(Ucall_is_present == TRUE)
        add_tcp_links_to_node(fout, node_vect[ii]);

    for(pp = 0; pp < no_of_cgens; pp++)
      {
        if((callgen_vect[pp]->type == CG_TYPE5) || 
           (callgen_vect[pp]->type == CG_TYPE11))
          continue;
        if((!VCs_per_us[callgen_vect[pp]->ref_user->id]) &&
           !(callgen_vect[pp]->type == CG_TYPE10) ) 
                	/* write only the user with some connection opened */
	continue;
        if (callgen_vect[pp]->ref_user->connected_node == ii)	
          fprintf(fout, "\n\tused_by %s thru %s;", 
               user_name(callgen_vect[pp]->ref_user->id), 
               link_name(callgen_vect[pp]->ref_user->in_link->id));
      }	
    fprintf(fout, "\nendwith;\n");
   }    
}


/* 
**  Function  : static void add_tcp_links_to_node(FILE *fout, Node *obs_node)
**  Return    : none.
**  Remarks   : the information concerning the TYPETCP and TYPETCP_REC 
**              users associated to Class U calls is translated into RC 
**              statements  and added to the file pointed by 'fout'.
*/
static void add_tcp_links_to_node(FILE *fout, Node *obs_node)
{
 CallList       *Ucallscan;
 Node		*curr_node;
 int        ii, nn;
 int	link_id;
 int        su_id;
 int        du_id;
 char           su_name[50];
 char           du_name[50];
 char           su_in_link[50];
 char           du_in_link[50];
 int        us_cnt = 0;

 for(nn = 0; nn < no_of_nodes; nn++)
    {
     curr_node = node_vect[nn];
     for(ii = 0; ii < curr_node->no_of_users; ii++)
        {
         link_id = curr_node->to_user[ii]->id;
         for(Ucallscan = link_Ucall_stat[link_id]->U_calls_on->next;
             Ucallscan != link_Ucall_stat[link_id]->U_calls_on;
	     Ucallscan = Ucallscan->next)
            {
             us_cnt++;
             du_id = callgen_vect[Ucallscan->call_elem->dest_id]->ref_user->id;
	     if(user_vect[du_id]->connected_node == obs_node->id)
		{
/*
             	 sprintf(du_name, "tcp_rec_%s_%d", user_name(du_id), us_cnt);
*/
             	 sprintf(du_name, "TR_%s_%d", user_name(du_id), us_cnt);
             	 sprintf(du_in_link, "%s_%d", 
                        link_name(user_vect[du_id]->in_link->id), us_cnt);
             	 fprintf(fout, "\n\tused_by %s thru %s;", du_name, du_in_link);
		}
             su_id = 
		callgen_vect[Ucallscan->call_elem->source_id]->ref_user->id;
	     if(user_vect[su_id]->connected_node == obs_node->id)
		{
/*
             	 sprintf(su_name, "tcp_%s_%d", user_name(su_id), us_cnt);
*/
             	 sprintf(su_name, "T_%s_%d", user_name(su_id), us_cnt);
             	 sprintf(su_in_link, "%s_%d", 
                        link_name(user_vect[su_id]->in_link->id), us_cnt);
             	 fprintf(fout, "\n\tused_by %s thru %s;", su_name, su_in_link);
		}
            }
        }
    }
}


/* 
**  Function  : static void write_users(FILE *fout, long *VCs_per_us)
**  Return    : none.
**  Remarks   : the information concerning the users of the network is
**              translated into RC statements and added to the file 
**              pointed by 'fout'.
**              'VCs_per_us' contains the total number of active 
**              connections for each user.
*/
static void write_users(FILE *fout, long *VCs_per_us)
{
 int        ii;
 int        usid;

 fprintf(fout, "\n\n/*  Network Users  */");
 for(ii = 0; ii < no_of_cgens; ii++)
   {
/* Type 5 (Class U) and Type 11 (TCP_SINK) Call Generators are managed  */
/* elsewhere.                                                           */
    if((callgen_vect[ii]->type == CG_TYPE5) || 
       (callgen_vect[ii]->type == CG_TYPE11)) 
        continue;

    usid = callgen_vect[ii]->ref_user->id;
    if(!VCs_per_us[usid] && !(callgen_vect[ii]->type == CG_TYPE10))
        /* write only the user with some connection opened */
        continue;

    fprintf(fout, "\nuser %s with\n\ttype ", user_name(usid));
    switch(callgen_vect[ii]->type)
        {
         case CG_TYPE1 : 
                fprintf(fout, "AGGREGATE_CBR;");
                break;
        
         case CG_TYPE2 :
                fprintf(fout, "AGGREGATE_VBR;");
                break;

         case CG_TYPE3 :
                fprintf(fout, "NOT_DEFINED; ");
                break;

         case CG_TYPE4 :
                fprintf(fout, "NOT_DEFINED; ");
                break;

/* Type 5 (Class U) Call Generators are managed elsewhere.      */
         case CG_TYPE5  :
         case CG_TYPE11 :
                break;

         case CG_TYPE10 :
                fprintf(fout, "TYPESINK; ");
                break;

/* Type 11 (TCP_SINK) Call Generators are managed elsewhere.    */

        }
    fprintf(fout, "\n\tconnected_to %s thru %s;", 
                node_name(user_vect[usid]->out_link->dest), 
                link_name(user_vect[usid]->out_link->id));

/* The parameter 'number_of_connections' is not defined for 	*/
/* CG_TYPE10 users.						*/
    if(callgen_vect[ii]->type != CG_TYPE10) 
    	{
    	 fprintf(fout, "\n\tparameter number_of_connections = %ld;", 
                      VCs_per_us[usid]);
#ifdef CASTELLI
	 if(callgen_vect[ii]->type == CG_TYPE1)
	   fprintf(fout, "\n\tparameter priority_class = 0;");
	 else
	   fprintf(fout, "\n\tparameter priority_class = 1;");  
#endif
    	}

    fprintf(fout, "\nendwith;\n");
   }    
}


/* 
**  Function  : static void write_tcp_users(FILE *fout)
**  Return    : none.
**  Remarks   : the information concerning the TYPETCP and TYPETCP_REC 
**              users associated to Class U calls is translated into RC 
**              statements  and added to the file pointed by 'fout'.
*/
static void write_tcp_users(FILE *fout)
{
 CallList       *Ucallscan;
 Node		*curr_node;
 int	ii, nn;
 int        su_id;
 int        du_id;
 int        link_id;
 char           su_name[50];
 char           du_name[50];
 char           su_out_link[50];
 char           du_out_link[50];
 int        us_cnt = 0;

 fprintf(fout, "\n\n/*  TCP Users  */");

 for(nn = 0; nn < no_of_nodes; nn++)
    {
     curr_node = node_vect[nn];
     for(ii = 0; ii < curr_node->no_of_users; ii++)
        {
         link_id = curr_node->to_user[ii]->id;
         for(Ucallscan = link_Ucall_stat[link_id]->U_calls_on->next;
             Ucallscan != link_Ucall_stat[link_id]->U_calls_on;
	     Ucallscan = Ucallscan->next)
            {
             us_cnt++;
             su_id = 
		callgen_vect[Ucallscan->call_elem->source_id]->ref_user->id; 
             du_id = callgen_vect[Ucallscan->call_elem->dest_id]->ref_user->id;

             /* Setting the names of users and links.   */
/*
             sprintf(su_name, "tcp_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "tcp_rec_%s_%d", user_name(du_id), us_cnt);
*/
             sprintf(su_name, "T_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "TR_%s_%d", user_name(du_id), us_cnt);
             sprintf(su_out_link, "%s_%d", 
                        link_name(user_vect[su_id]->out_link->id), us_cnt);
             sprintf(du_out_link, "%s_%d", 
                        link_name(user_vect[du_id]->out_link->id), us_cnt);

             /* Writing the Source User parameters.     */
             fprintf(fout, "\nuser %s with\n\ttype TYPETCP;", su_name);
             fprintf(fout, "\n\tconnected_to %s thru %s;", 
                        node_name(user_vect[su_id]->out_link->dest), 
                        su_out_link);

             fprintf(fout, "\n/* Uncomment the following line if you do not want ABR");
             fprintf(fout, " with the class U connections */");
	     fprintf(fout, "\n/*\tparameter utcp_rate = %f; */", 
			((CallType5 *)Ucallscan->call_elem->info)->mbr_value);
             fprintf(fout, "\n/* End of the TCP definitions */\n");

	     fprintf(fout, "\n/* Comment the following lines if you do not want");
             fprintf(fout, " ABR with the class U connections */");
	     fprintf(fout, "\n\tparameter abr_mcr = %f; ", 
		      ((CallType5 *)Ucallscan->call_elem->info)->mbr_value/10.0);
	     fprintf(fout, "\n\tparameter abr_icr = %.3f; ", 
		      ((CallType5 *)Ucallscan->call_elem->info)->mbr_value/2.0);
             fprintf(fout, "\n/* End of the ABR definitions */\n");

	     fprintf(fout, "\n\tparameter utcp_destination_id = %s;", du_name);
#ifdef CASTELLI
             fprintf(fout, "\n\tparameter priority_class = 2;");
#endif
             fprintf(fout, "\nendwith;\n");

             /* Writing the Destination User parameters.        */
             fprintf(fout, "\nuser %s with\n\ttype TYPETCP_REC;", du_name);
             fprintf(fout, "\n\tconnected_to %s thru %s;", 
                       node_name(user_vect[du_id]->out_link->dest), 
                       du_out_link);
             fprintf(fout, "\nendwith;\n");

             fprintf(fout, "\n");
            }
        }
    }
}


/* 
**  Function  : static void write_links(FILE *fout, long *VCs_per_us)
**  Return    : none.
**  Remarks   : the information concerning the links of the network is
**              translated into RC statements and added to the file 
**              pointed by 'fout'.
*/
static void write_links(FILE *fout, long * VCs_per_us)
{
 int        ii;
 char		marks[MAX_LINK];

 for(ii = 0; ii < no_of_links; ii++)
        marks[ii] = 0;

 for(ii = 0; ii < no_of_cgens; ii++)
   {
    switch(callgen_vect[ii]->type)
        {
/* For Class U and TCP_SINK call generators the user links are defined  */
/* below, in function 'write_tcp_links'.                                */
         case CG_TYPE5  :
         case CG_TYPE11 :
                marks[callgen_vect[ii]->ref_user->in_link->id] = 2;
                marks[callgen_vect[ii]->ref_user->out_link->id] = 2;
                break;

         default :
    		if(!VCs_per_us[callgen_vect[ii]->ref_user->id] && 
    		   !(callgen_vect[ii]->type == CG_TYPE10))
        		/* write only the user with some connection opened */
		  {
                    marks[callgen_vect[ii]->ref_user->in_link->id] = 3;
                    marks[callgen_vect[ii]->ref_user->out_link->id] = 3;
		   }
		else
		  {
                    marks[callgen_vect[ii]->ref_user->in_link->id] = 1;
                    marks[callgen_vect[ii]->ref_user->out_link->id] = 1;
		   }
                break;
        }
   }

 fprintf(fout, "\n\n/*  Node to Node Links  */");
 for(ii = 0; ii < no_of_links; ii++)
   {
    if(marks[ii] == 0)
        {
         fprintf(fout, "\nlink %s with", link_name(ii));               
         fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                        link_vect[ii]->link_len);
         fprintf(fout, "\n\t          link_timescale = %d;", 
                        (int)(link_vect[ii]->timescale));
         fprintf(fout, "\nendwith;\n");
        }    
   }

 fprintf(fout, "\n\n/*  User Links  */");
 for(ii = 0; ii < no_of_links; ii++)
   {
    if(marks[ii] == 1)
        {
         fprintf(fout, "\nlink %s with", link_name(ii));               
         fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                        link_vect[ii]->link_len);
         fprintf(fout, "\n\t          link_timescale = %d;", 
                        (int)(link_vect[ii]->timescale));
         fprintf(fout, "\nendwith;\n");
        }    
   }                                                              
}


/* 
**  Function  : static void write_tcp_links(FILE *fout)
**  Return    : none.
**  Remarks   : the information concerning the links between the TYPETCP 
**              and TYPETCP_REC users and their respective access nodes is 
**              translated into RC statements and added to the file pointed 
**              by 'fout'.
*/
static void write_tcp_links(FILE *fout)
{
 CallList       *Ucallscan;
 Node		*curr_node;
 int	ii, nn;
 int        su_id;
 int        du_id;
 int        link_id;
 char           su_name[50];
 char           du_name[50];
 char           su_in_link[50];
 char           su_out_link[50];
 char           du_in_link[50];
 char           du_out_link[50];
 int        us_cnt = 0;

 fprintf(fout, "\n\n/*  TCP Users' Links  */");

 for(nn = 0; nn < no_of_nodes; nn++)
    {
     curr_node = node_vect[nn];
     for(ii = 0; ii < curr_node->no_of_users; ii++)
        {
         link_id = curr_node->to_user[ii]->id;
         for(Ucallscan = link_Ucall_stat[link_id]->U_calls_on->next;
             Ucallscan != link_Ucall_stat[link_id]->U_calls_on;
	     Ucallscan = Ucallscan->next)
            {
             us_cnt++;
             su_id = 
		callgen_vect[Ucallscan->call_elem->source_id]->ref_user->id; 
             du_id = callgen_vect[Ucallscan->call_elem->dest_id]->ref_user->id;

	     /* Setting the names of users and links.        */
/*
    	     sprintf(su_name, "tcp_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "tcp_rec_%s_%d", user_name(du_id), us_cnt);
*/
    	     sprintf(su_name, "T_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "TR_%s_%d", user_name(du_id), us_cnt);
             sprintf(su_in_link, "%s_%d", 
                     	link_name(user_vect[su_id]->in_link->id), us_cnt);
    	     sprintf(su_out_link, "%s_%d", 
                        link_name(user_vect[su_id]->out_link->id), us_cnt);
             sprintf(du_in_link, "%s_%d", 
                        link_name(user_vect[du_id]->in_link->id), us_cnt);
             sprintf(du_out_link, "%s_%d", 
                        link_name(user_vect[du_id]->out_link->id), us_cnt);

	     /* Writing the Source User's outgoing link parameters. 	*/
    	     fprintf(fout, "\nlink %s with", su_out_link);               
             fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                	user_vect[su_id]->out_link->link_len);
    	     fprintf(fout, "\n\t          link_timescale = %d;", 
                	(int)(user_vect[su_id]->out_link->timescale));
             fprintf(fout, "\nendwith;\n");

	     /* Writing the Source User's incoming link parameters.  	*/
    	     fprintf(fout, "\nlink %s with", su_in_link);               
             fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                	user_vect[su_id]->in_link->link_len);
             fprintf(fout, "\n\t          link_timescale = %d;", 
                	(int)(user_vect[su_id]->in_link->timescale));
       	     fprintf(fout, "\nendwith;\n");

	     /* Writing the Destination User's outgoing link parameters.*/
    	     fprintf(fout, "\nlink %s with", du_out_link);               
             fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                	user_vect[du_id]->out_link->link_len);
    	     fprintf(fout, "\n\t          link_timescale = %d;", 
                	(int)(user_vect[du_id]->out_link->timescale));
    	     fprintf(fout, "\nendwith;\n");

	     /* Writing the Destination User's incoming link parameters.*/
    	     fprintf(fout, "\nlink %s with", du_in_link);               
    	     fprintf(fout, "\n\tparameter link_length    = %.2f,", 
                	user_vect[du_id]->in_link->link_len);
    	     fprintf(fout, "\n\t          link_timescale = %d;", 
                	(int)(user_vect[du_id]->in_link->timescale));
    	     fprintf(fout, "\nendwith;\n");

    	     fprintf(fout, "\n");
	    }
        }
    }    
}


/* 
**  Function  : static void write_tcp_paths(FILE *fout)
**  Return    : none.
**  Remarks   : the information concerning the TYPETCP and TYPETCP_REC 
**              users associated to Class U calls is translated into RC 
**              statements  and added to the file pointed by 'fout'.
*/
static void write_tcp_paths(FILE *fout)
{
 CallList       *Ucallscan;
 Node		*curr_node;
 LinkList       *llscan;
 int	ii, nn;
 int        su_id;
 int        du_id;
 int        link_id;
 char           su_name[50];
 char           du_name[50];
 int        us_cnt = 0;
 

 fprintf(fout, "\n\n/*  Paths connecting TCP Users  */");

 for(nn = 0; nn < no_of_nodes; nn++)
    {
     curr_node = node_vect[nn];
     for(ii = 0; ii < curr_node->no_of_users; ii++)
        {
         link_id = curr_node->to_user[ii]->id;
         for(Ucallscan = link_Ucall_stat[link_id]->U_calls_on->next;
             Ucallscan != link_Ucall_stat[link_id]->U_calls_on;
	     Ucallscan = Ucallscan->next)
            {
             us_cnt++;
             su_id = 
		callgen_vect[Ucallscan->call_elem->source_id]->ref_user->id; 
             du_id = callgen_vect[Ucallscan->call_elem->dest_id]->ref_user->id;

	     /* Setting the names of users and links.        */
/*
    	     sprintf(su_name, "tcp_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "tcp_rec_%s_%d", user_name(du_id), us_cnt);
*/
    	     sprintf(su_name, "T_%s_%d", user_name(su_id), us_cnt);
             sprintf(du_name, "TR_%s_%d", user_name(du_id), us_cnt);
	     /* Writing the Path connecting the Source User 	*/
	     /* and the Destination User.			*/
    	     if(Ucallscan->call_elem->path_ptr != NULL)
          	{
           	 llscan = Ucallscan->call_elem->path_ptr->path->next;
                 if (llscan->next != NULL)
                    /* check if there are node<->node links */
              	    fprintf(fout, "\npath user %s to user %s thru\n\t", 
                         su_name, du_name);
             	 while(llscan->next != NULL)
              	    {
                     if(llscan->next->next == NULL)
                  	fprintf(fout, "%s;", link_name(llscan->link_elem->id));
               	     else     
                  	fprintf(fout, "%s, ", link_name(llscan->link_elem->id));
                     llscan = llscan->next;
              	    }
		}
    	     fprintf(fout, "\n\n");
            }
	}
   }    
}

                     
/*  Function  : void init_photo_struct()
**  Return    : none.
**  Remarks   : each vector element of 'photo_vect' is initialized: a list
**    		of length 'max_state_photos' is created. 'load_level' is
**              initialized to '0'
*/

void init_photo_struct()
{
 register int ii,jj;
 PHattr   *ph_pointer;

 for (ii=0;ii<no_of_links;ii++)
 	{
	 ph_pointer = photo_vect[ii] = (PHattr *)MMmalloc(sizeof(PHattr),
	                                         "init_photo_list");
	 ph_pointer->load_level = DBL_MAX;
         for(jj=0;jj<max_state_photos;jj++)
	 	{
                 ph_pointer->next = (PHattr *)MMmalloc(sizeof(PHattr),
                                                 "init_photo_list");
		 ph_pointer=ph_pointer->next;
		 ph_pointer->load_level=0;
		}
	 ph_pointer->next=photo_vect[ii];
	}
}

/*  Function  : void manage_photo_list(int linkid)
**  Return    : none.
**  Remarks   : This function handles the list of stored network state
** 		snapshots, checking whether the current congestion is to 
** 		be recorded.
*/

void manage_photo_list(int linkid)
{
  register   int while_passed = 0;
  char       filename[50];
  PHattr     *ph_pointer = photo_vect[linkid];


  static int photo_counter=0;

  /* While we have a 'better' snapshot than those already stored */
  /* continue scanning the snapshot list */

  while ((link_carried_load[linkid][TOTAL_LINK_LOAD] >
                             ph_pointer->next->load_level) ||
        ((link_carried_load[linkid][TOTAL_LINK_LOAD] ==
                             ph_pointer->next->load_level) &&
        ((link_vect[linkid]->calls_on->next->call_elem->end_time - 
	  ancles_time) > 
         ph_pointer->next->congestion_duration)))
         {
          while_passed++;
          ph_pointer=ph_pointer->next;
         }
  if (while_passed)
    {
      PHattr *ph_handle = photo_vect[linkid]->next;
      /* if 'while' has cycled at least twice, extract the first */
      /* element of the list and insert it after 'ph_pointer'.   */
      /* Then, overwrite its records.                            */
      if (while_passed > 1)
        {
         photo_vect[linkid]->next=photo_vect[linkid]->next->next;
         ph_handle->next = ph_pointer->next;
         ph_pointer->next = ph_handle;
        }
      if (ph_handle->load_level > 0)
	{
	 sprintf(filename,"%s.rc",ph_handle->filename);
	 if (remove(filename))
           {
            printf("\nError:  The file %s cannot be removed\n", filename);
            exit(1);
           }
	 sprintf(filename,"%s.acf",ph_handle->filename);
	 if (remove(filename))
           {
            printf("\nError:  The file %s cannot be removed\n", filename);
            exit(1);
           }
	}
      ph_handle->load_level = link_carried_load[linkid][TOTAL_LINK_LOAD];
      ph_handle->congestion_duration = 
         link_vect[linkid]->calls_on->next->call_elem->end_time - ancles_time;
      sprintf(ph_handle->filename,
         "%s_link%d_%d",file_radix, linkid, photo_counter++);
      print_state_photo(ph_handle->filename, linkid);
    }
}
                                            
/*  Function  : void rename_photo_files(void)
**  Return    : none.
**  Remarks   : renames snapshot files using 'linkid' and ordering them
**              according to their position in the snapshot list.
*/
void rename_photo_files(void)
{
  register  int ii,jj,cnt;
  char      oldname[50], newname[50], buffer[255], acf_file[50], cmp_string[50];
  PHattr    *ph_pointer;
  FILE      *old_file, *new_file;
  
  for (ii=0;ii<no_of_links;ii++)
    {
     cnt = 0;
     ph_pointer = photo_vect[ii]->next;
     for (jj=0;jj<max_state_photos;jj++)
       {
	  if ((ph_pointer->load_level > 0)&&(cnt < max_state_photos))
	    {
	       cnt++;
	       sprintf(oldname,"%s.rc",ph_pointer->filename); 
	       sprintf(newname,  "%s_l%d_%d.rc", file_radix,ii,cnt);
	       sprintf(acf_file, "%s_l%d_%d.acf",file_radix,ii,cnt);
	       sprintf(cmp_string,"default acf_file = ");
	       
	       if((old_file = fopen(oldname, "r")) == NULL)               
		 {
		    printf("\nError:  The file %s cannot be opened\n", oldname);
		    exit(1);
		 } 
	       if((new_file = fopen(newname, "w")) == NULL)               
		 {
		    printf("\nError:  The file %s cannot be opened\n", newname);
		    exit(1);
		 } 
	       
	       /* start reading the old file, and replace the acf_file statement */
	       while(fgets(buffer, 255, old_file) != NULL)
		 {
		    if(strncmp(buffer,cmp_string,16) == 0)
		      /* got the acf_statement */
		      fprintf(new_file, "default acf_file = %s;\n", acf_file);
		    else
		      /* copy the original string */
		      fprintf(new_file, "%s", buffer); 
		 }
	       fclose(new_file);
	       fclose(old_file);
	       if (remove(oldname))
		 {
		    perror("ERROR: ");
		    printf("\nError:  The file %s cannot be renamed to %s (%d)\n", 
			   oldname, newname,ii);
		    exit(1);
		 }                                                    
	       sprintf(oldname,"%s.acf",ph_pointer->filename); 
	       sprintf(newname,"%s_l%d_%d.acf",file_radix,ii,cnt);
	       if (rename(oldname,newname))
		 {
		    printf("\nError:  The file %s cannot be renamed\n", oldname);
		    exit(1);
		 }
	      }
	  ph_pointer = ph_pointer->next;
       }
    }
}

/*
**  Function  : void print_state_photo(int no_links, int *link_ids)
**  Return    : none.
**  Remarks   : the file whose name is obtained by merging 'file_radix' 
**              and 'file_numb' is created and filled, according to the 
**              current state of the network.
**              The loads carried by the links whose number is given by 
**              'no_links' and whose identifiers are pointed by 'link_ids' 
**              are reported at the beginning of the RC file.
**              The ACF file, collecting data about the active VCs, is 
**              created and filled after the RC file has been processed.
*/
static void print_state_photo(char *filename, int linkid)
{
 char    file_rc[50];
 char    file_acf[50];
 FILE    *lfp;
 int uu;
 long    *VCs_per_user;

 static  Boolean check_write_file = FALSE;

 Ucall_is_present = FALSE;

 sprintf(file_acf, "%s.acf", filename);
 sprintf(file_rc, "%s.rc", filename);

 VCs_per_user = (long *)MMmalloc(no_of_users * sizeof(long),
                                 "print_state_photo");
 for(uu = 0; uu < no_of_users; uu++)
        VCs_per_user[uu] = 0;

 if(!check_write_file)
   {
	check_write_file=TRUE;
        for(uu = 0; uu < no_of_users; uu++)
                {
                 head_of_VC_list[uu] = vcset_alloc(); 
                 head_of_VC_list[uu]->next = NULL;
                }
   }
 else
        for(uu = 0; uu < no_of_users; uu++)
                 head_of_VC_list[uu]->next = NULL;

/* Filling the ACF file.        */
 if((lfp = fopen(file_acf, "w")) == NULL)               
   {
    printf("\nError:  The file %s cannot be opened\n", file_acf);
    exit(1);
   } 
 create_VC_lists();
 write_VCs(lfp, VCs_per_user);                                                
 fclose(lfp);

/* Filling the RC file.         */
 if((lfp = fopen(file_rc, "w")) == NULL)
   {
    printf("\nError:  The file %s cannot be opened\n", file_rc);
    exit(1);
   } 

 write_head(lfp, file_rc, linkid);
 write_global(lfp, file_acf);
 write_demand(lfp);
 write_nodes(lfp, VCs_per_user);
 write_users(lfp, VCs_per_user);
 if(Ucall_is_present == TRUE)
        write_tcp_users(lfp);
 write_links(lfp, VCs_per_user);
 if(Ucall_is_present == TRUE)
        {
         write_tcp_links(lfp);
         write_tcp_paths(lfp);
        }
 fclose(lfp);
 free(VCs_per_user);
}
