/*
**  Module    : routing.c
**  Version   : $Revision: 3.3 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico, Zoltan Zsoka
**		- DLCP, hop by hop and crankback features added by
**		  Filippo Di Natale
**
**  Synopsis  : this module contains the functions dealing with the
**              available routing algorithms.
**
**  Functions : static	int    	eval_max_no_of_hops(void)
**              static 	double 	eval_ln_erlang(double, double)
**              static	void  	evaluate_TR_ths(void)
**              static	void    evaluate_primary_loads(void)
**              static	int    	set_call_type(int    )
**              static  TimeUnit evaluate_call_duration(int    )
**              static 	void    set_call_informations(Call *, double, double,
**						      double)
**              static  void    accepted_call_insert(Path *,int    , int    ,
**		        	 double, double, double)
**              void    init_RA(void)
**              void    exe_RA(int    , int    , double, double, double)
**				Path*	iterate_path_list(Path*, int*, Routing*, Path*)
**				Pair*	iterate_pair_list(Pair*, int*, Routing*, Pair*)
**
**  History   : 12-OCT-1995 - Creation of the module.   -AF-
**              01-JUN-1996 - Revision.                 -AF-
**              01-MAR-1997 - Revision			-MM-
**              14-OCT-1997 - Revision: moved the routing choice at the
**                            CG level                  -MM-
**              08-FEB-1998 - Revision: removed a bug in the erlang evaluation
**                                                      - GF -
**              08-FEB-1998 - Revision: Added the user -> net and net -> user
**                            links to the path_mat. Changed all the routines
**                            involving this.           - GF - MM -
**		20-AUG-1999 - Revision: Added support of discrete update of the
**			      available free bw for Ucalls on the links
**			      MSF and DLCP (made by FN) routing algorithms
**							- ZZ -
**		31-AUG-1999 - Revision: Ucall uses traffic instead of hold_time
**							- ZZ -
29-01-2001 Cleaning...
**              30-APR-2001 - Introduce the support of ASON calls -ZZ-
*/

# ifndef        ANCLES_H
#include        "ancles.h"
# endif

/* Define to get trunking emulation */

/*
#define TRUNKING
*/
#ifdef TRUNKING
#define NO_OF_CHECK_BEFORE_TRUNKING 1
#endif

/* Things used by dlcp routing */
#define PATH_FOUND 1
#define PATH_NOT_FOUND 0
#define CANT_GO_BACK -1

/*************************
**   Global Variables   **
*************************/

/* Array used by the crankback routing function in order to determine if */
/* a node is a deadlock, and by the DLCP function to avoid loops.        */
static Boolean avoid_nodes[MAX_NODE] = { FALSE };

static int last_link; /* identifier of the link who caused the path to be rejected*/

/* For MSF algorithms */
typedef struct ra8_path_unit ra8_Path;  /* this structure is used in the MSF algs */

struct ra8_path_unit {
  double	second;	/* Value corresponding to second criterium of the path */
  double	third;	/* Value corresponding to third criterium of the path */
  double	av_bw;	/* The aviable bw for a U-type call (b_neck / num_of_Ucalls) */
  int		ref_p_no;	/* Pathnumber of the referred path. */
  Path*		ref_path;	/* Refers to the real Path */
  ra8_Path*	next;	/* Next struct in a list */
};

static ra8_Path* ra8_path_array = NULL;  /* All the ra8_path elements are stored as
					    elements of this array for avoid a lot
					    of memory allocations */
static int half_max_path_ra8;     /* Size of the array (static) */
static int num_ra8_path;	  /* Useful elements in the array */

static ra8_Path* ra8_path_array_2 = NULL;  /* All the ra8_path elements are stored as
					      elements of this array for avoid a lot
					      of memory allocations this array is used
					      for the second filtering */
static int half_max_path_ra8_2;   /* Size of the array (static) */
static int num_ra8_path_2;	  /* Useful elements in the array */


/*************************
**  Function Pointers   **
*************************/
extern  void     (*init_CAC)(void);
extern  Boolean  (*exe_CAC)(Link *, double, double, double);


/**************************
**  Function Prototypes  **
**************************/
static  int      eval_max_no_of_hops(void);
static  double   eval_ln_erlang(double, double);
static  void     evaluate_TR_ths(void);
static  void     evaluate_primary_loads(void);
static  int      set_call_type(int    );
static  TimeUnit evaluate_call_duration(int    );
static  void     set_call_informations(Call *, double, double, double);
static  Call*    accepted_call_insert(Path *, int    , int    , int    , double,
				      double, double);
static  int	 branch_search(Path**	, int	, int	, Path*	, Boolean*	);


/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : static int     eval_max_no_of_hops(void)
**  Return    : the maximum number of hops among the allowed paths.
**  Remarks   : all the allowed paths are scanned for finding the needed
**              value; we use the path matrix ordered by lenght (because
**              in each path matrix we have ALL the allowed path).
*/
static int     eval_max_no_of_hops()
{
 register int i, j;
 int      provv_max;
 Path     *path_ptr;

 provv_max = 0;
 for(i = 0; i < no_of_cgens; i++)
    {
     for(j = 0; j < no_of_cgens; j++)
        {
         for(path_ptr = path_mat[IN_ORDER_OF_LENGTH][i][j];
	     path_ptr != NULL; path_ptr = path_ptr->next)
             if(path_ptr->no_of_hops > provv_max)
                provv_max = path_ptr->no_of_hops;
        }
    }
 return provv_max;
}


/*
**  Function  : static double eval_ln_erlang(double cap, double nload)
**  Return    : the system blocking probability, given the normalized
**              capacity of the system ('cap') and the normalized offered
**              load ('nload').
**  Remarks   : the natural logarithm of the Erlang loss probability
**              is obtained by means of a recursive formula.
**              The natural logarithm is used instead of the real value in
**              order to avoid overflow/underflow problems.
**              Real numbers are permitted as input values.
*/
static double eval_ln_erlang(double cap, double nload)
{
 long	j;
 double h, b, n, ylog;

 b = 1.0;
 h = modf(cap, &n);

 for(j = 1; (j <= n) && (b < 1.0E+03) && (b > 1.0E-03); j++)
        b = (b / nload) * j + 1.0;
 /* Check if we can use the LOG approximation or not */
 if(!(j > n))
   {
    double b1log, b2log, blog;

    blog = log(b);

    for(; j <= n; j++)
       blog = - log(nload) + log(j) + blog;

    b1log = - log(nload) + log(n + 1.0) + blog;
    b2log = - log(nload) + log(n + 2.0) + b1log;

    ylog = (1.0 - h) * blog + h * b1log;
    ylog = ylog + (h * (1.0 - h) / 2.0) * (2.0 * b1log - blog - b2log);
   }
 else
   {
   /* No approximation allowed */
    double b1, b2;

    b1 = b * (n + 1.0) / nload + 1.0;
    b2 = b1 * (n + 2.0) / nload + 1.0;

    ylog = (1.0 - h) * log(b) + h * log(b1);
    ylog = ylog + (h * (1.0 - h) / 2.0) * (2.0 * log(b1) - log(b) - log(b2));
   }
 return (- ylog);
}


/*
**  Function  : static void evaluate_TR_ths(void)
**  Return    : none.
**  Remarks   : the state-protection levels required by the Controlled
**              Alternate Routing Algorithm are evaluated here.
**              If the CAC algorithm based on the Equivalent Bandwidth
**              Allocation has been selected, the weighted average over the
**              equivalent bandwidths of the connected call generators is
**              used in order to obtain a single protection level valid
**              for all the traffic classes.
**              If one of the CAC algorithms based on Peak Bandwidth
**              Allocation has been selected, the weighted average over the
**              peak bandwidths of the connected call generators is
**              used in order to obtain a single protection level valid
**              for all the traffic classes.
**              If the Mixed Traffic CAC algorithm has been selected, the
**              weighted average over the Modified Bandwidths of the connected
**              call generators is used in order to obtain a single protection
**              level valid for all the traffic classes.
**              REMARKS: The Class U calls are never taken into account to the
**              evaluation of the protection factor.
*/
static void evaluate_TR_ths(void)
{
 register int i, j;
 int      max_hops;    /* Maximum number of hops among the allowed paths.*/
 double   acc_bw;
 double   acc_weights;
 double   mean, peak, arr_rate, mean_hold_time;
 double   provv_ths, ln_cmp_val;
 double   avg_bw, gen_bw, max_bw, min_bw;

 CGType1 *cg1;
 CGType2 *cg2;
 CGType3 *cg3;
 CGType4 *cg4;
 CGType6 *cg6;
 CGType7 *cg7;
 CGType9 *cg9;

 min_bw = DBL_MAX;
 max_bw = 0.0;
 acc_weights = 0.0;
 acc_bw = 0.0;


/* The maximum number of hops among the allowed paths is evaluated      */
/* here once and for all.                                               */
 max_hops = eval_max_no_of_hops();

/* The values of the parameters used to evaluate the protection levels	*/
/* depend on the type of the call generators.				*/
 for(i = 0; i < no_of_cgens; i++)
    {
     switch(callgen_vect[i]->type)
        {
         case CG_TYPE1 :        /* CBR source.                          */
            cg1 = (CGType1 *)(callgen_vect[i]->info);
	    peak = (cg1->min_bit_rate +cg1->max_bit_rate) / 2.0;
            mean = peak;
            arr_rate = cg1->arrival_rate;
            mean_hold_time = cg1->mean_hold_time;
            break;

         case CG_TYPE2 :        /* On-Off VBR source.                   */
            cg2 = (CGType2 *)(callgen_vect[i]->info);
            peak = (cg2->min_peak_rate + cg2->max_peak_rate) / 2.0;
            mean = peak * cg2->mean_on_time /
                (cg2->mean_on_time + cg2->mean_off_time);
            arr_rate = cg2->arrival_rate;
            mean_hold_time = cg2->mean_hold_time;
            break;

         case CG_TYPE3 :        /* Uniform VBR source.                  */
            cg3 = (CGType3 *)(callgen_vect[i]->info);
            peak = cg3->max_bit_rate_avg;
            mean = (cg3->min_bit_rate_avg + cg3->max_bit_rate_avg) / 2.0;
            arr_rate = cg3->arrival_rate;
            mean_hold_time = cg3->mean_hold_time;
            break;

         case CG_TYPE4 :        /* Video VBR source.                    */
            cg4 = (CGType4 *)(callgen_vect[i]->info);
            peak = cg4->higher_bit_rate_avg;
            mean = cg4->higher_bit_rate_avg * cg4->higher_bit_rate_prob +
                cg4->medium_bit_rate_avg * cg4->medium_bit_rate_prob +
                cg4->lower_bit_rate_avg * cg4->lower_bit_rate_prob;
            arr_rate = cg4->arrival_rate;
            mean_hold_time = cg4->mean_hold_time;
            break;

         case CG_TYPE6 :        /* ASON source.                */
            cg6 = (CGType6 *)(callgen_vect[i]->info);
	        peak = wavelength_capacity *
				(cg6->min_no_of_waves +cg6->max_no_of_waves) / 2.0;
            mean = peak;
            arr_rate = cg6->arrival_rate;
            mean_hold_time = cg6->mean_hold_time;
            break;

         case CG_TYPE7 :        /* On-Off ASON source.         */
            cg7 = (CGType7 *)(callgen_vect[i]->info);
	        peak = wavelength_capacity;
            mean = peak;
            arr_rate = cg7->arrival_rate;
            mean_hold_time = cg7->mean_hold_time;
            break;

         case CG_TYPE9 :        /* Pascal ASON source.         */
            cg9 = (CGType9 *)(callgen_vect[i]->info);
	        peak = wavelength_capacity *
				(cg9->min_no_of_waves +cg9->max_no_of_waves) / 2.0;
            mean = peak;
            arr_rate = cg9->arrival_rate;
            mean_hold_time = cg9->mean_hold_time;
            break;


	/* The UBR, NEVER_CLOSING_ASON, SINK, TCP_SINK and ASON_SINK call */
	/* generators are not taken into account when the primary loads are evaluated.	        */
	case CG_TYPE5:
	case CG_TYPE8:
	case CG_TYPE10:
	case CG_TYPE11:
	case CG_TYPE12:
	  continue;

	default:
	  printf("ERROR: unknown call generator type %d in function evaluate_TR_ths\n",
		 callgen_vect[i]->type);
	  exit(1);
        }

/* The selected CAC algorithm influences the criterion used to define   */
/* the bandwidth requirements of a source.                              */
     switch(selected_cac)
        {
         case CALL_ADM_CTRL1 :  /* Equivalent bandwidth assignment.     */
            gen_bw = eval_equivalent_bw(mean, peak, capacity /
                        callgen_vect[i]->ref_user->out_link->timescale,
                        cell_loss_rate);
            break;

         case CALL_ADM_CTRL2 :
         case CALL_ADM_CTRL3 :  /* Peak bandwidth assignment.           */
            gen_bw = peak;
            break;

	 case CALL_ADM_CTRL4 :  /* Mixed traffic CAC (class 1/2/U calls)*/
	    gen_bw = eval_modified_bw(mean, peak);
	    break;

 	 default:
	  printf("ERROR: unknown CAC %d in function evaluate_TR_ths\n",
		 selected_cac);
	  exit(1);
        }
     acc_bw += gen_bw * arr_rate * mean_hold_time;
     acc_weights += arr_rate * mean_hold_time;
     if(gen_bw < min_bw) min_bw = gen_bw;
     if(gen_bw > max_bw) max_bw = gen_bw;
    }
 avg_bw = acc_bw / acc_weights;

 /* for all the possible threshold */
 for(j = 0; j < no_of_thresholds; j++)
 /* for all the link               */
    for(i = 0; i < no_of_links; i++)
       {
	double link_cap;
 /* if you selected the Sibal DeSimone threshold evaluation method */
 /* change the threshold!                                          */
	if(link_vect[i]->routing_threshold[j] != SIBAL_THRESHOLD )
	   continue;
	if(link_primary_load[i] == 0.0)
           {
            link_vect[i]->routing_threshold[j] = 0.0;
            continue;
           }
	link_cap = capacity / link_vect[i]->timescale;
	ln_cmp_val = eval_ln_erlang((link_cap - avg_bw + min_bw) / avg_bw,
                   link_primary_load[i] / avg_bw) + log(max_hops);
	for(provv_ths = 0.0;
            (provv_ths < (link_cap - avg_bw)) ||
	                (equ_double((link_cap-avg_bw), provv_ths, DBL_CMP_THS));
            provv_ths += min_bw)
                   if(eval_ln_erlang((link_cap - avg_bw + min_bw - provv_ths) /
                                      avg_bw,
                                      link_primary_load[i] /
                                      avg_bw) >
                                      ln_cmp_val)
                           break;
	link_vect[i]->routing_threshold[j] = provv_ths;
       }
/* #if 0
DEBUG - print all the threshold
    for(i=0; i< no_of_links; i++)
       printf(" %s : link_trheshold = {%f , %f, %f, %f }\n", link_name(i),
                   link_vect[i]->routing_threshold[0],
                   link_vect[i]->routing_threshold[1],
                   link_vect[i]->routing_threshold[2],
                   link_vect[i]->routing_threshold[3]);
    for(i=0; i< no_of_cgens; i++)
       printf(" %s : primary_threshold = %d\t alternate_threshold =%d\n",
                    cgen_name(i),
                    callgen_vect[i]->choosen_routing.primary_threshold,
                    callgen_vect[i]->choosen_routing.alternate_threshold);
#endif
*/
}


/*
**  Function  : static void evaluate_primary_loads(void)
**  Return    : none.
**  Remarks   : the primary load on each link is evaluated here, on the
**              basis of the traffic pattern matrix and of the primary
**              path assignments.
**              Primary load is the load that would flow through
**              the links if only primary paths were used to route the
**              connections.
**              REMARKS: the class U call are NOT taken into account
**              in the avaluation of the primary load
*/
static void evaluate_primary_loads(void)
{
  LinkList *list_ptr;
  register int i, ii, jj;
  int ord;

  for(i = 0; i < no_of_links; i++)
    link_primary_load[i] = 0.0;

  for(ii=0; ii < no_of_cgens; ii++)
    {
      if(callgen_vect[ii]->type == CG_TYPE5 ||
		  callgen_vect[ii]->type == CG_TYPE10 ||
		  callgen_vect[ii]->type == CG_TYPE11 ||
		  callgen_vect[ii]->type == CG_TYPE12)
	continue;

      for(jj=0; jj < no_of_cgens; jj++)
        {
	  /* Use the order selected by the source cgen.  */
	  ord = callgen_vect[ii]->choosen_routing.path_order_type;

	  if (path_mat[ord][ii][jj])
	    {
              /* do not consider the first link (user -> network link) */
              /* and the last link (network -> user link)              */
	      for(list_ptr = path_mat[ord][ii][jj]->path->next;
	          list_ptr->next != NULL; list_ptr = list_ptr->next)
		link_primary_load[list_ptr->link_elem->id] +=
		    traff_pattern[callgen_vect[ii]->ref_user->id]
		                 [callgen_vect[jj]->ref_user->id];
	    }
        }
    }

  for(i = 0; i < no_of_links; i++)
    link_primary_load[i] *= (load_conv_fact * load);
}


/*
**  Function  : static int     set_call_type(int     cgen_id)
**  Return    : the type of the call generator indexed by 'cgen_id'.
**  Remarks   : none.
*/
static int     set_call_type(int     cgen_id)
{
 switch(callgen_vect[cgen_id]->type)
   {
    case CG_TYPE1 :
        return CALL_TYPE1;

    case CG_TYPE2 :
        return CALL_TYPE2;

    case CG_TYPE3 :
        return CALL_TYPE3;

    case CG_TYPE4 :
        return CALL_TYPE4;

    case CG_TYPE5 :
        return CALL_TYPE5;

    case CG_TYPE6 :
    case CG_TYPE7 :
    case CG_TYPE9 :
        return CALL_TYPE6;

    case CG_TYPE8 :
        return CALL_TYPE8;

    case CG_TYPE10 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cgen_id));
	printf("\n        being a SINK call generator.");
	exit(1);

    case CG_TYPE11 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cgen_id));
	printf("\n        being a TCP_SINK call generator.");
	exit(1);

    case CG_TYPE12 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cgen_id));
	printf("\n        being a ASON_SINK call generator.");
	exit(1);

    default :
        printf("\nError: call generator %s has an unknown type (%d)\n",
	       cgen_name(cgen_id), callgen_vect[cgen_id]->type);
	exit(1);
   }
}


/*
**  Function  : static TimeUnit evaluate_call_duration(int     cg_id)
**  Return    : the duration of the call whose generator is indexed by
**              'cg_id'.
**  Remarks   : the call duration is evaluated according to the call
**              generator type.
*/
static TimeUnit evaluate_call_duration(int     cg_id)
{
 switch(callgen_vect[cg_id]->type)
   {
    case CG_TYPE1 :
        return ( (TimeUnit)negexp(
                  ((CGType1 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE2 :
         return ( (TimeUnit)negexp(
                  ((CGType2 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE3 :
         return ( (TimeUnit)negexp(
                  ((CGType3 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE4 :
         return ( (TimeUnit)negexp(
                  ((CGType4 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE6 :
        return ( (TimeUnit)negexp(
                  ((CGType6 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE7 :
        return ( (TimeUnit)negexp(
                  ((CGType7 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE8 :
        return max_time - ancles_time + 1 + uniform_0_1(&sh_seed);

    case CG_TYPE9 :
        return ( (TimeUnit)negexp(
                  ((CGType9 *)(callgen_vect[cg_id]->info))->mean_hold_time,
                  &cg_seed_vect[cg_id]) );

    case CG_TYPE10 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cg_id));
	printf("\n        being a SINK call generator.");
	exit(1);

    case CG_TYPE11 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cg_id));
	printf("\n        being a TCP_SINK call generator.");
	exit(1);

    case CG_TYPE12 :
	printf("\nError:  %s cannot generate calls,", cgen_name(cg_id));
	printf("\n        being an ASON_SINK call generator.");
	exit(1);

    default :
        printf("\nError: call generator %s has an unknown type (%d)\n",
	       cgen_name(cg_id), callgen_vect[cg_id]->type);
	exit(1);
   }
}


/*
**  Function  : static void set_call_informations(Call *call_ptr,
**			    double average_bw, double peak_bw, double minim_bw)
**  Return    : none.
**  Remarks   : the information referring to the call are written in
**              the field pointed by 'info', according to the call
**              generator type.
**              The passed parameters are:
**               - '*call_ptr' : pointer to the incoming call;
**               - 'average_bw': average bandwidth requirement of the incoming
**                               calls' generator.
**               - 'peak_bw'   : peak bandwidth requirement of the incoming
**                               call's generator;
**               - 'minim_bw'  : minimum bandwidth requirement of the incoming
**                               calls' generator;
**              The bandwidth requirements are evaluated before opening the
**              call in order to allow the correct development of ROUTING
**              and CAC actions.
*/
static void set_call_informations(Call *call_ptr, double average_bw,
				  double peak_bw, double minim_bw)
{
 CallType1 *ptr1;
 CallType2 *ptr2;
 CallType3 *ptr3;
 CallType4 *ptr4;
 CallType5 *ptr5;
 CallType6 *ptr6;
 CallType8 *ptr8;

 switch(call_ptr->call_type)
   {
    case CALL_TYPE1 :
        ptr1 = call_type1_alloc();
        ptr1->cbr_value = peak_bw;
        call_ptr->info = (void *)ptr1;
        break;

    case CALL_TYPE2 :
        ptr2 = call_type2_alloc();
        ptr2->peak_rate = peak_bw;
        ptr2->mean_on_time =
	  ((CGType2 *)(callgen_vect[call_ptr->source_id]->info))->mean_on_time;
        ptr2->mean_off_time =
          ((CGType2 *)(callgen_vect[call_ptr->source_id]->info))->mean_off_time;
/*  Now I associate the call generator id to the new call; this is used */
/*  as a row index for matrices P_xx                                    */
        ptr2->cgen_id = callgen_vect[call_ptr->source_id]->id;
        ptr2->on_status = TRUE;

        call_ptr->info = (void *)ptr2;
        break;

    case CALL_TYPE3 :
        ptr3 = call_type3_alloc();
        ptr3->min_vbr_value = minim_bw;
        ptr3->max_vbr_value = peak_bw;
        call_ptr->info = (void *)ptr3;
        break;

    case CALL_TYPE4:
        ptr4 = call_type4_alloc();

        ptr4->higher_bit_rate_prob =
	 ((CGType4 *)(callgen_vect[call_ptr->source_id]->info))->
              higher_bit_rate_prob;

        ptr4->medium_bit_rate_prob =
           ((CGType4 *)(callgen_vect[call_ptr->source_id]->info))->
              medium_bit_rate_prob;

        ptr4->lower_bit_rate_prob =
           ((CGType4 *)(callgen_vect[call_ptr->source_id]->info))->
              lower_bit_rate_prob;

        ptr4->higher_bit_rate_value = peak_bw;

/* The lower bit rate value of the call is extracted here.      */
        ptr4->lower_bit_rate_value = minim_bw;

/* The medium bit rate value of the call is evaluated here      */
/* according to the other parameters of the call.               */
        ptr4->medium_bit_rate_value = ( average_bw -
        ( ptr4->higher_bit_rate_prob * peak_bw ) -
        ( ptr4->lower_bit_rate_prob * ptr4->lower_bit_rate_value ) ) /
          ptr4->medium_bit_rate_prob;

        call_ptr->info = (void *)ptr4;
        break;

    case CALL_TYPE5 :
        ptr5 = call_type5_alloc();
        ptr5->mbr_value = peak_bw;
        ptr5->curr_br = peak_bw;
        ptr5->last_load_time = ancles_time;
        ptr5->last_updt_time = ancles_time;
        ptr5->br_integ = 0.0;
        call_ptr->info = (void *)ptr5;
        break;

    case CALL_TYPE6 :
        ptr6 = call_type6_alloc();
        ptr6->nwaves = (int) floor(peak_bw / wavelength_capacity + 0.5);
        call_ptr->info = (void *)ptr6;
        break;

    case CALL_TYPE8 :
        ptr8 = call_type8_alloc();
        ptr8->nwaves = (int) floor(peak_bw / wavelength_capacity + 0.5);
        call_ptr->info = (void *)ptr8;
        break;
   }
}



/*
**  Function  : static double eval_min_free_bw(Path* path)
**  Return    : the bottleneck bw
**  Remarks   : Evaluate the available free band for a new
**              class U call on the bottleneck link of PATH.
*/
static double eval_min_free_bw(Path* path)
{
  LinkList* link;
  double min_free_bw = DBL_MAX;
  double bw_per_Ucall;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {

/* The Ucalls can only see a periodically updated charasteristic of the link. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
      {
       bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->official_free_Uband /
        ((double)link_Ucall_stat[link->link_elem->id]->official_no_of_Ucall
	+ 1.0);
      }

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->free_Uband /
	((double)(link_Ucall_stat[link->link_elem->id]->no_of_Ucall)+1.0);

      if (bw_per_Ucall < min_free_bw
	  && ! equ_double(bw_per_Ucall, min_free_bw, DBL_CMP_THS))
	min_free_bw = bw_per_Ucall;
    }

  return min_free_bw;
}

/*
**  Function  : static double ex_eval_min_free_bw(Path* path)
**  Return    : the bottleneck bw
**  Remarks   : Evaluate the available free band for an old
**              class U call on the bottleneck link of PATH.
*/
static double ex_eval_min_free_bw(Path* path)
{
  LinkList* link;
  double min_free_bw = DBL_MAX;
  double bw_per_Ucall;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {

/* The Ucalls can only see a periodically updated charasteristic of the link.
   The number of U calls has an uncertainty. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
      {
       bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->official_free_Uband /
        ((double)link_Ucall_stat[link->link_elem->id]->official_no_of_Ucall);
      }

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->free_Uband /
	((double)(link_Ucall_stat[link->link_elem->id]->no_of_Ucall));

      if (bw_per_Ucall < min_free_bw
	  && ! equ_double(bw_per_Ucall, min_free_bw, DBL_CMP_THS))
	min_free_bw = bw_per_Ucall;
    }

  return min_free_bw;
}

/*
**  Function  : static double eval_dynamic_path_cost(Path* path)
**  Return    : path cost
**  Remarks   : Evaluate the cost of PATH. The cost is a function of
**              the number of hops, the length and the available bandwidth
**              for class U calls: C = a*H + b*L + c/min_free_bw
*/
static double eval_dynamic_path_cost(Path* path)
{
  double min_free_bw;

  min_free_bw = eval_min_free_bw(path);

  if( equ_double (min_free_bw, 0.0, DBL_CMP_THS))
     return (DBL_MAX);
  else
     return (hop_weight * path->no_of_hops
	     + length_weight * path->length
	     + congestion_weight / min_free_bw);
}


/*	
**  Function  : static double eval_minimum_distance_cost(Path* path)
**  Return    : path cost
**  Remarks   : Evaluate the cost of PATH. The cost is a function of the
**              available bandwidth for class U calls over all
**              links of the path.
*/
static double eval_minimum_distance_cost(Path* path)
{
  LinkList* link;
  double cost = 0.0;
  double bw_per_Ucall;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {

/* The Ucalls can only see a periodly updated charasteristic of the link.
   The number of U calls has an uncertainty. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
      {
       bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->official_free_Uband /
        ((double)link_Ucall_stat[link->link_elem->id]->official_no_of_Ucall
	+ 1.0);
      }

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->free_Uband /
	((double)(link_Ucall_stat[link->link_elem->id]->no_of_Ucall)+1.0);

      if (bw_per_Ucall != 0.0)
	cost += 1.0 / bw_per_Ucall;
      else
	{
	  cost = DBL_MAX;
	  break;
	}
    }
  return cost;
}

/*	
**  Function  : static double ex_eval_minimum_distance_cost(Path* path)
**  Return    : path cost
**  Remarks   : Evaluate the cost of PATH. The cost is a function of the
**              available bandwidth for class U calls over all
**              links of the path. Taking in account only the existing calls!
*/
static double ex_eval_minimum_distance_cost(Path* path)
{
  LinkList* link;
  double cost = 0.0;
  double bw_per_Ucall;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {

/* The Ucalls can only see a periodly updated charasteristic of the link.
   The number of U calls has an uncertainty. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
      {

       if(link_Ucall_stat[link->link_elem->id]->official_no_of_Ucall>0)
       bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->official_free_Uband /
        ((double)link_Ucall_stat[link->link_elem->id]->official_no_of_Ucall);
       else
        bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->official_free_Uband;
      }

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else
       {
        if(link_Ucall_stat[link->link_elem->id]->no_of_Ucall > 0)
	 bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->free_Uband /
	  ((double)(link_Ucall_stat[link->link_elem->id]->no_of_Ucall));
	else
	 bw_per_Ucall = link_Ucall_stat[link->link_elem->id]->free_Uband;
       }

      if (bw_per_Ucall != 0.0)
	cost += 1.0 / bw_per_Ucall;
      else
	{
	  cost = DBL_MAX;
	  break;
	}
    }

  return cost;
}

/*
**  Function  : static double eval_min_whole_bw(Path* path)
**  Return    : the bottleneck bw (for all the U-calls, not divided by
**		its number)
**  Remarks   : Evaluate the available free band for all the
**              class U calls left on the bottleneck link of PATH.
*/
static double eval_min_whole_bw(Path* path)
{
  LinkList* link;
  double min_free_bw = DBL_MAX;
  double whole_bw;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {

/* The Ucalls can only see a periodly updated charasteristic of the link. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
       whole_bw = link_Ucall_stat[link->link_elem->id]->official_free_Uband;

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else
       whole_bw = link_Ucall_stat[link->link_elem->id]->free_Uband;

      if (whole_bw < min_free_bw
	  && ! equ_double(whole_bw, min_free_bw, DBL_CMP_THS))
	min_free_bw = whole_bw;
    }
  return min_free_bw;
}

/*
**  Function  : static Boolean cong_on_path(Path* path)
**  Return    : indicator of congestion
**  Remarks   : TRUE if there is at least one congested link
**		on the path
*/
static Boolean cong_on_path(Path* path)
{
  LinkList* link;

  for (link = path->path->next; link->next != NULL; link = link->next)
    {
/* Only for node-node links */
    if(link_Ucall_stat[link->link_elem->id]->is_node_node)
     {

/* The Ucalls can only see a periodically updated charasteristic of the link. */
      if (link_vect[link->link_elem->id]->bw_update_p > 0.0)
      {
       if(link_Ucall_stat[link->link_elem->id]->official_cong)
        return TRUE;
      }

/* If the period time was 0.0 we use the actual (not the last updated) value. */
      else
      {
       if(link_Ucall_stat[link->link_elem->id]->unused_free_band -
          capacity * cong_thrsh / link_vect[link->link_elem->id]->timescale
	   < DBL_CMP_THS)
        return TRUE;
      }
     }
    }

  return FALSE;
}


/*
**  Function  : Path* iterate_path_list(Path* path, int* curr_path_no,
**			       Routing* cgen_routing, Path* primary_path)
**  Return    : next path
**  Remarks   :	Used to iterate thru all paths of a S-D pair that comply
**              with CG limitations on number of paths and on number of hops
**              and length of paths. Return the next valid path, or NULL
**              when there are no more valid paths.
**
**              The arguments 'path' and 'curr_path_no' hold the status of the
**              iteration.  'path_number' is the CG choice on valid paths.
**              'primary_path' is the first path of the list.
**
**              Example:
**
**              path = primary_path;
**              curr_path_no = 1;
**
**              do
**              {
**
**               ...
**
**              } while ((path = iterate_path_list(path, &curr_path_no,
**                                      cgen_routing,
**                                      primary_path)) != NULL)
*/
Path* iterate_path_list(Path* path, int* curr_path_no,
			       Routing* cgen_routing, Path* primary_path)
{

  int jump_back = 0;
  int response;

  (*curr_path_no)++;

 /* If the hop-by-hop routing is active or the cranckback is enabled
    (used in DLCP routing) we get the next path with the branchsearch
    mechanism */

  if((cgen_routing->use_hop_by_hop)||(cgen_routing->use_crankback))
   {
    if(cgen_routing->use_crankback)
      do
       {
        response = branch_search(&path,last_link,jump_back,primary_path,
            		avoid_nodes);
        switch(response)
         {
          case PATH_FOUND:
            break;
          case PATH_NOT_FOUND:
            jump_back++;
            break;
          case CANT_GO_BACK:
            return NULL;
            break;
         }
       }
      while(response != PATH_FOUND);
	
    else
     if(branch_search(&path,last_link,0,primary_path,avoid_nodes)
          	==PATH_NOT_FOUND)
        return NULL;
   }

 /* In the case of en explicit routing (not hop by hop style) the next path
    is simply taken from the path list */
  else path = path->next;

  if (path)
    {
      switch(cgen_routing->selected_range_type.type)
	{
	case ALL_PATHS_NUMBER:
	  /* Go throught remaining paths*/
	  break;
	case ONLY_PRIMARY_PATH:
	  path = NULL;
	  break;
	case FIXED_PATH_NUMBER:
	  if (*curr_path_no > cgen_routing->selected_range_type.value.ival) /* XXX */
	    path = NULL;
	  break;
	case MAX_HOP_PATH_NUMBER:
	  if (path->no_of_hops >cgen_routing->selected_range_type.value.ival )
	    path = NULL;
	  break;
	case MAX_HOP_MORE_PATH_NUMBER:
	  if (path->no_of_hops - primary_path->no_of_hops >
	       cgen_routing->selected_range_type.value.ival)
	    path = NULL;
	  break;
	case MAX_LENGTH_PATH_NUMBER:
	  if (path->length > cgen_routing->selected_range_type.value.dval)
	    path = NULL;
	  break;
	case MAX_LENGTH_MORE_PATH_NUMBER:
	  if (path->length - primary_path->length >
	       cgen_routing->selected_range_type.value.dval)
	    path = NULL;
	  break;
	default:
	  /* Something went wrong  */
	  printf("ERROR: unknown path number choice %d in function iterate_path_list\n",
		 cgen_routing->selected_range_type.type);
	  exit(1);
	  break;
	}
    }

  return path;
}


/*
**  Function  : Pair* iterate_pair_list(Pair* pair, int* curr_pair_no,
**			       Routing* cgen_routing, Pair* primary_pair)
**  Return    : next pair
**  Remarks   :	Usage is similar to the function iterate_path_list
*/
Pair* iterate_pair_list(Pair* pair, int* curr_pair_no,
			       Routing* cgen_routing, Pair* primary_pair)
{
  (*curr_pair_no)++;

 /* In the case of en explicit routing (not hop by hop style) the next path
    is simply taken from the path list */
  pair = pair->next;

  if (pair)
    {
      switch(cgen_routing->selected_range_type.type)
	{
	case ALL_PATHS_NUMBER:
	  /* Go throught remaining paths*/
	  break;
	case ONLY_PRIMARY_PATH:
	  pair = NULL;
	  break;
	case FIXED_PATH_NUMBER:
	  if (*curr_pair_no > cgen_routing->selected_range_type.value.ival) /* XXX */
	    pair = NULL;
	  break;
	case MAX_HOP_PATH_NUMBER:
	  if (pair->path1->no_of_hops + pair->path2->no_of_hops - 2 >
		    cgen_routing->selected_range_type.value.ival )
	    pair = NULL;
	  break;
	case MAX_HOP_MORE_PATH_NUMBER:
	  if (pair->path1->no_of_hops + pair->path2->no_of_hops
		  - primary_pair->path1->no_of_hops 
		  - primary_pair->path2->no_of_hops >
	       cgen_routing->selected_range_type.value.ival)
	    pair = NULL;
	  break;
	case MAX_LENGTH_PATH_NUMBER:
	  if (pair->path1->length + pair->path2->length > 
		  cgen_routing->selected_range_type.value.dval)
	    pair = NULL;
	  break;
	case MAX_LENGTH_MORE_PATH_NUMBER:
	  if (pair->path1->length + pair->path2->length 
		  - primary_pair->path1->length
		  - primary_pair->path2->length >
	       cgen_routing->selected_range_type.value.dval)
	    pair = NULL;
	  break;
	default:
	  /* Something went wrong  */
	  printf("ERROR: unknown path number choice %d in function iterate_pair_list\n",
		 cgen_routing->selected_range_type.type);
	  exit(1);
	  break;
	}
    }

  return pair;
}


/*
**  Function  : static Boolean validate_path(Path **path)
**  Return    : 1 on success, 0 on fail.
**  Remarks   : scans the given path to determine if it trasverse a node
**		listed in "avoid_nodes" array, if yes changes the given path
**		too and returns 0.
*/
static Boolean validate_path(Path **path)
{
 LinkList *link;

 if((*path) == NULL) return TRUE;
 for(link = (*path)->path; link != NULL; link = link->next)
    if( avoid_nodes[link->link_elem->dest] || (link->link_elem->id == last_link))
      {
       (*path) = (*path)->next;
       return FALSE;
      }
 return TRUE;
}

/*
**  Function  : static int how_long_it_takes_to_destination(int node, int dest,
**				Path *path, int *hops, double *length)
**  Return    : 0 if no paths are available, 1 otherwise.
**  Remarks   : This function stores in the supplied variables "hops_to_go"
**		and "length_to_go" the number of hops and length left to run
**		from "node" to "dest", using minimal path that avoids nodes
**		already trasversed.
*/
static int how_long_it_takes_to_destination(int node, int dest, Path *path,
   			int *hops, double *length)
{
 Boolean path_found = FALSE;
 LinkList *link;
 int skip_hop = 0;
 double skip_length = path->path->link_elem->link_len;
 Path *actpath = path;	/* since in the original version the value of path was
 			** overwritten, I included this variable -ZZ- */

 do
  {
   link = actpath->path;
   do
    {
     if(link->link_elem->dest == node)
      {
        path_found = TRUE;
        break;
       }
      else
       {
        link = link->next;
        if(link != NULL)
         {
          skip_hop++;
          skip_length += link->link_elem->link_len;
         }
        }/*else*/
    }
   while(link->link_elem->dest != dest);

   if(!path_found)
    {
     skip_hop = 0;
     skip_length = actpath->path->link_elem->link_len;
     actpath = actpath->next;
    }
   else
    {
     for(link = actpath->path->next;
         link->link_elem->source != node; link = link->next);
	
     while(link->link_elem->source != dest)
      {
       if(avoid_nodes[link->link_elem->dest])
        {
         path_found = FALSE;
         skip_hop = 0;
         skip_length = 0.0;		/* SUSPICIOUS!!! ZZ*/
         actpath = actpath->next;
         break;
        }
       else
        link = link->next;
      } /* while(link... != dest) */
    } /* else */

  }
 while((!path_found)&&(actpath != NULL));

 if(actpath == NULL) return 0;
 (*hops) = actpath->no_of_hops - skip_hop;	/* SUSPICIOUS!!! ZZ*/
 (*length) = actpath->length - skip_length;
 return 1;
}

/*
**  Function  : static int branch_search(Path **path, int last_link_par,
**				int jump_back, Path *primary_path,
**				Boolean *avoid_nodes)
**  Return    : PATH_FOUND on success, PATH_NOT_FOUND on fail, CANT_GO_BACK
**		if no crankback could be performed.
**  Remarks   : Used by the hop by hop facility;
**
**	This function stores in the supplied variable "path" the first path
**	available that satisfies these two conditions:
**
**   1) from source to destination it shares the same set of link of the one
**	provided, this condition holding until a link with "last_link_par"
**	identificator is encountered;
**   2) no "blind nodes" are trasversed.
*/
static int branch_search(Path **path, int last_link_par, int jump_back,
		Path *primary_path, Boolean *avoid_nodes)
{
  int link_buffer[MAX_LINK];/* Stores the successful */
     		 	     /* trunk of the path    */
  int index = 0;
  int hops_done=0;          /* To compute how many links have been     */
                            /* successfully tested 		       */
  Path *temp_path;
  LinkList *curr_link;


  temp_path = *path; /* blocked path */
  curr_link = temp_path->path;


  if(last_link_par != -1)   /* if we didn't get stuck at the first link */
     do			/* we find good links already trasversed */
       {
        link_buffer[hops_done]=curr_link->link_elem->id;
        curr_link = curr_link->next;
        hops_done++;
       }
     while(curr_link->link_elem->id != last_link_par);
  else	 		/* The user to node link is congested */
       return CANT_GO_BACK;

  hops_done -= jump_back; /*used for crankback */

  if((jump_back > 0)&&(hops_done == 1)) return CANT_GO_BACK;
     					  /* In this case only the user */
     					  /* to node link remains, it's */
                                          /* useless jumping back...    */


  if(jump_back > 0)
     temp_path = primary_path;  /* if called from crankback,the base path */
        			/*  				is needed */
  else
     temp_path = temp_path->next; /* else it suffices to examine next path */

  if (temp_path == NULL)
      {
       avoid_nodes[link_vect[last_link_par]->source]=TRUE;
       return PATH_NOT_FOUND;
      }

  curr_link = temp_path->path;

  /*let's try if we find a path sharing the same set of links */
  /*already trasversed and not containing "blind" nodes       */

  do
    {

      while ((index != hops_done)&&(temp_path != NULL))
         {
          if(curr_link->link_elem->id == link_buffer[index])
            {
             curr_link = curr_link->next;
             index++;
            }

           else /*no way: let's try next path...*/

            {
             index = 0;
             temp_path = temp_path->next;
             if(temp_path != NULL)
             curr_link = temp_path->path;
            }
         }
      index = 0;
    }
  while((!validate_path(&temp_path))&&(temp_path!=NULL));

  if(temp_path == NULL)
                       {
                        avoid_nodes[link_vect[last_link_par]->source]=TRUE;
                        return PATH_NOT_FOUND;
                       }

  else
                       {
                        (*path) = temp_path;
                        return PATH_FOUND;
                       }
}

/*
**  Function  : void print_cgen_to_cgen_path(int     ord, int     i, int     j)
**  Return    : none.
**  Remarks   : function used to check the paths provided by
**              the functions included in getpaths.c.
*/
void print_cgen_to_cgen_path(int ord, int i, int j)
{
 LinkList *ptr;
 Path *path_point;

 if (path_mat[ord][i][j] == NULL)
    return;

 fprintf(stderr, "\nPaths from cgen %s to cgen %s :\n",
   cgen_name(i), cgen_name(j));
 fprintf(stderr, "# hops:\tLength (km):\tPath:\n");

 for(path_point = path_mat[ord][i][j];
     path_point != NULL; path_point = path_point->next)
   {
    fprintf(stderr, "  %d \t %.1f \t",
      path_point->no_of_hops, path_point->length);
    for(ptr = path_point->path; ptr != NULL; ptr = ptr->next)
       fprintf(stderr, "%s ", link_name(ptr->link_elem->id));
    fprintf(stderr, "\n");
   }
}


/*
**  Function  : void init_RA(void)
**  Return    : none.
**  Remarks   : this function initializes all the variables dealing with RA.
*/
void init_routing(void)
{
 int     i, j, ord;

 for(i = 0; i < no_of_cgens; i++)
	 /* Silently force RA1, RA50, RA60 and RA90 to use only the primary path. */
	 if (callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA1 ||
		 callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA50 ||
		 callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA60 /* ||
		 callgen_vect[i]->choosen_routing.selected_routing == ROUTING_RA90 */)	
		 callgen_vect[i]->choosen_routing.selected_range_type.type = ONLY_PRIMARY_PATH;

/* first we alloc the path_mat matrix */
 path_mat = alloc_path_mat(no_of_cgens, NO_OF_PATHS);

 find_all_paths();

/* create the path-pair matrix only if it is needed, i.e., there are
   call generators with routing using protection */
 ord = 0;
 for(i = 0; i < no_of_cgens; i++)
 {
	 if(callgen_vect[i]->choosen_routing.selected_routing >= 70 &&
		 callgen_vect[i]->choosen_routing.selected_routing < 90 || 
		 callgen_vect[i]->choosen_routing.selected_routing >= 120 && 
		 callgen_vect[i]->choosen_routing.selected_routing < 140)
		 ord++;
 }
 if(ord)
 {
	 pair_mat = alloc_pair_mat(no_of_cgens, 2);
	 /* in this version there is only the ordering by hop number and
	    path length are realized */
	 
	 /* if node-node protection is used, section protection
	    is not considered */
	 if(subslinkid != -1)
	 {
	 printf("Error:  Path and section protection can not be \
used together!\n");
			  exit(1);
	 }
 }
 else
	 pair_mat = NULL;

 if(pair_mat != NULL)
	 find_all_pairs();

/* Create the load-balanced path matrices */
#ifndef NO_LOAD_BALANCING

# ifdef PROFILE_LB_ALG
   start_CPU_time = clock();
   balance_paths(BALANCED_IN_ORDER_OF_HOP_NUMBER, load_balancing_algorithm);
   end_CPU_time = clock();
   printf("Charged CPU time during the load balancing algorithm is %f\n",
	     (((double)end_CPU_time) - start_CPU_time) /
	     CLOCKS_PER_SEC);
#else
 balance_paths(BALANCED_IN_ORDER_OF_HOP_NUMBER, load_balancing_algorithm);
#endif
 balance_paths(BALANCED_IN_ORDER_OF_LENGTH, load_balancing_algorithm);
#endif

   /* Evaluate the primary thresholds */
 evaluate_primary_loads();
 evaluate_TR_ths();

 /* If there is a link which has to be substituted do it! */
 if(subspaths != NULL)
	 substitute_link(subslinkid);

 /* print path list if debug is active */

 if(init_debug_level & ROUTING_DEBUG_LEVEL)
   {
    for(ord = 0; ord < NO_OF_PATHS; ord++)
      {
       switch (ord)
         {
          case IN_ORDER_OF_HOP_NUMBER:
             fprintf(stderr,
		     "\nPaths ordered first by number of hops without balancing:\n");
             break;
          case IN_ORDER_OF_LENGTH:
             fprintf(stderr,
		     "\nPaths ordered first by length without balancing :\n");
             break;
          case BALANCED_IN_ORDER_OF_HOP_NUMBER:
             fprintf(stderr,
		     "\nPaths ordered first by number of hops with balancing :\n");
             break;
          case BALANCED_IN_ORDER_OF_LENGTH:
             fprintf(stderr,
		     "\nPaths ordered first by length with balancing:\n");
             break;
         }
        for(i = 0; i < no_of_cgens; i++)
            for(j = 0; j < no_of_cgens; j++)
                 print_cgen_to_cgen_path(ord, i, j);
      }

   }
}

/*
**  Function  : static ra8_Path* def_set_msf(int cgen_src, int cgen_dest,
**			Routing* cgen_routing)
**  Return    : A list of ra8_path-s
**  Remarks   : The first half of the elements of the pathlist
**		will be sorted by the whole bottleneck bandwidth and
**		organized in a descending list.
**		Then the first elements of the this list will be sorted
**		by the aviable bandwidth (for Ucalls) and organized in a
**		descending list.
*/
static ra8_Path* def_set_msf(int cgen_src, int cgen_dest, Routing* cgen_routing)

{
   int i, inserted;
   Path *path_ptr, *primary_path;
   ra8_Path *ra8_path_list, *prev_ra8_path, *ra8_path_ptr, *inp_ra8_ptr;
   int curr_path_no = 1;
   int msf_type = -1;

   primary_path = path_mat[IN_ORDER_OF_HOP_NUMBER][cgen_src][cgen_dest];
   path_ptr = primary_path;
   switch(cgen_routing -> selected_routing)
	{
   	 case ROUTING_RA8:
	  	msf_type = 0;
		break;
	 default:
		printf("ERROR: wrong msf routing algorithm %d in function\
		     def_set_msf\n", cgen_routing->selected_routing);
		exit(1);
	 	break;
	}

   /* If the array doesn't exists, create */
   if (ra8_path_array == NULL)
     {
        if(MSF_MAX_PATH > 2)
	  half_max_path_ra8 = 2 * (int)ceil((double)MSF_MAX_PATH/4);
	else half_max_path_ra8 = 1;
	ra8_path_array = (ra8_Path *)MMmalloc(sizeof(ra8_Path) * half_max_path_ra8,
						"def_set_msf");
     }

   /* Count the aviable paths beetween src and dst, but there is a predefined maximum */
   while((curr_path_no < MSF_MAX_PATH) && ((path_ptr = iterate_path_list(path_ptr,
   				&curr_path_no, cgen_routing, primary_path)) != NULL));

   /* Only the first half of the path-set (sorted by the hop criterium) will be used. */

   if(curr_path_no > 2)
     num_ra8_path = 2 * (int)ceil((double)curr_path_no/4);
   else num_ra8_path = 1;

   path_ptr = primary_path;
   ra8_path_list = ra8_path_array;
   ra8_path_array[0].next = NULL;

   /* Insert the paths in the list, sorted by the second criterium: bottleneck width */
   for(i = 0; i < num_ra8_path; i++)
     {
   /* Evaluate the values of the fields  */
        ra8_path_array[i].ref_path = path_ptr;
        ra8_path_array[i].ref_p_no = i+1;
   /* Secondary criterium depends on the algorithm */
	switch(msf_type){
	  case 0:
	   ra8_path_array[i].second = eval_min_whole_bw(path_ptr);
	   break;
	  default:
	   ra8_path_array[i].second = 1.0;
	   break;}
	ra8_path_array[i].av_bw = eval_min_free_bw(path_ptr);
	inserted = 0;
   /* At the first inserted element inicialize the list */
	if(!i)
	  inserted++;
   /* Insert the element at the right place */
	ra8_path_ptr = ra8_path_list;
        prev_ra8_path = ra8_path_ptr;
	while(!inserted)
	  {
       /* If we are at the end of the actual list */
	    if(ra8_path_ptr == NULL)
	      {
	       prev_ra8_path->next = ra8_path_array + i;
	       ra8_path_array[i].next = NULL;
	       inserted++;
	      }
	    else
	      {
        /* The new element is "better" then the compared one -> insert now */
	       if(ra8_path_ptr->second - ra8_path_array[i].second < widest_bottleneck_range)
		 {
		  if(ra8_path_ptr == ra8_path_list)
		    ra8_path_list = ra8_path_array + i;
		  else
		    prev_ra8_path->next = ra8_path_array + i;
		  ra8_path_array[i].next = ra8_path_ptr;
		  inserted++;
		 }
       /* Take the next element of the list to compare */
	       else
	         {
		   prev_ra8_path = ra8_path_ptr;
		   ra8_path_ptr = ra8_path_ptr->next;
		 }
	      }
	  }
	path_ptr = path_ptr->next;
     }

   /* If the second array doesn't exists, create */
   if (ra8_path_array_2 == NULL)
     {
        if(half_max_path_ra8 > 2)
  	  half_max_path_ra8_2 = 2 * (int)ceil((double)half_max_path_ra8/4);
	else half_max_path_ra8_2 = 1;
	ra8_path_array_2 = (ra8_Path *)MMmalloc(sizeof(ra8_Path) * half_max_path_ra8_2, "def_set_msf");
     }

   if(num_ra8_path > 2)
     num_ra8_path_2 = 2 * (int)ceil((double)num_ra8_path/4);
   else num_ra8_path_2 = 1;


   inp_ra8_ptr = ra8_path_list;
   ra8_path_list = ra8_path_array_2;
   ra8_path_array_2[i].next = NULL;

   /* Insert the ra8_paths in the list, sorted by the third criterium:
      aviable bottleneck width */
   for(i = 0; i < num_ra8_path_2; i++)
     {
   /* Copy the values of the fields  */
        ra8_path_array_2[i].ref_path = inp_ra8_ptr->ref_path;
	ra8_path_array_2[i].second = inp_ra8_ptr->second;
	ra8_path_array_2[i].av_bw = inp_ra8_ptr->av_bw;
	ra8_path_array_2[i].third = inp_ra8_ptr->av_bw;
	inserted = 0;
   /* At the first inserted element inicialize the list */
	if(!i)
	  inserted++;
   /* Insert the element at the right place */
	ra8_path_ptr = ra8_path_list;
        prev_ra8_path = ra8_path_ptr;
	while(!inserted)
	  {
       /* If we are at the end of the actual list */
	    if(ra8_path_ptr == NULL)
	      {
	       prev_ra8_path->next = ra8_path_array_2 + i;
	       ra8_path_array_2[i].next = NULL;
	       inserted++;
	      }
	    else
	      {
        /* The new element is "better" then the compared one -> insert now */
	       if(ra8_path_ptr->third -
		   ra8_path_array_2[i].third < widest_bottleneck_range)
		 {
		  if(ra8_path_ptr == ra8_path_list)
		    ra8_path_list = ra8_path_array_2 + i;
		  else
		    prev_ra8_path->next = ra8_path_array_2 + i;
		  ra8_path_array_2[i].next = ra8_path_ptr;
		  inserted++;
		 }
       /* Take the next element of the list to compare */
	       else
	         {
		   prev_ra8_path = ra8_path_ptr;
		   ra8_path_ptr = ra8_path_ptr->next;
		 }
	      }
	  }
	inp_ra8_ptr = inp_ra8_ptr->next;
     }

   return ra8_path_list;
}

/*
**  Function  : static Path* dlcp_routing(int cgen_src, int cgen_dest,
**			Routing* cg_routing, int* curr_path_no)
**  Return    : The selected path
**  Remarks   :
*/
static Path* dlcp_routing(int cgen_src, int cgen_dest, Routing* cg_routing,
			double mean_bw, double peak_bw, int* curr_path_no)
{
 register int i;
 Path *path_base_ptr;
 int start_node;
 int end_node;
 int curr_node;
 int last_node;
 int first_node;
 int temp_node;
 int primary_no_of_hops;
 LinkList *link_list;
 Link    *start_link = NULL;
 Link    *trasversed_links[MAX_NODE];
 Boolean second_choice_selected = FALSE;
 Boolean primary;
 int curr_hop_number = 0;
 double  curr_length = 0;
 int     hops_to_go;
 double  length_to_go;
 double  primary_length;
 double link_threshold = 0.0;
 double min_dlcp_cost;
 double dlcp_cost;
 double temp_var;
 int ord = cg_routing->path_order_type;
 int state = 0;


 if(ord > IN_ORDER_OF_LENGTH)
   {
        printf("\nError: User %s uses DLCP, unbalanced ordering\
                 is required.\n", user_name(callgen_vect[cgen_src]
                  ->ref_user->id));
	exit(1);
   }
 path_base_ptr = path_mat[ord][cgen_src][cgen_dest];
 link_list = path_base_ptr->path;
 /* Link connecting user with access node is checked for CAC */


 if(!exe_CAC(link_list->link_elem, mean_bw,peak_bw, 0.0))
    return NULL;

 /* initializes trasversed nodes vector   */
 memset((Boolean*)avoid_nodes,FALSE,MAX_NODE);

 /* starts the dlcp routing strategy */
 start_node = link_list->link_elem->dest;
 trasversed_links[0] = link_list->link_elem;
 curr_length = link_list->link_elem->link_len;
 first_node = start_node;
 end_node = callgen_vect[cgen_dest]->ref_user->connected_node;
 avoid_nodes[first_node] = TRUE; /* used to avoid loops */

 if(first_node != end_node)
  {
   /* Set the primary length and number of hops */

   primary_no_of_hops = path_base_ptr->no_of_hops;
   primary_length = path_base_ptr->length;

   while(start_node != end_node)
     {
      last_node = start_node;
      temp_node = start_node;
      min_dlcp_cost = MAX_COST;
      /* try all possible nodes */
      for(i=0; i<node_vect[start_node]->no_of_ports; i++)
        {
         primary = FALSE;
         curr_node = node_vect[start_node]->outqueue[i]->dest;
             /* Evaluates the local cost */
         temp_var = (!equ_double(node_vect[start_node]->
               outqueue[i]->free_band,0.0,DBL_CMP_THS))?
                              1.0/node_vect[start_node]->
                                 outqueue[i]->free_band:
                                               MAX_COST;
         /* Evaluates DLCP cost function */
         dlcp_cost = temp_var + node_vect[curr_node]->aggregate_cost*
            				  node_vect[curr_node]->alpha;

             /* Checks if curr_node is part of a first choice path */
                 if(!avoid_nodes[curr_node])
                 {
                  if(curr_node != end_node)

                    {
                     if(how_long_it_takes_to_destination(curr_node,end_node,
                                                  path_base_ptr,&hops_to_go,
                                                             &length_to_go))
                        /* if this node lead to destination... */

                     switch(ord)
                     {
                      case IN_ORDER_OF_HOP_NUMBER :

                       if( (hops_to_go +
                            curr_hop_number+1) == primary_no_of_hops )
                           primary = TRUE;
                       break;

                      case IN_ORDER_OF_LENGTH :
                       if( (length_to_go + curr_length+
                            node_vect[start_node]->outqueue[i]->link_len)
                                                      == primary_length )

                           primary = TRUE;
                       break;
                     }
                    }/* if(curr_node != end_node) */
                  else
                   if(ord == IN_ORDER_OF_HOP_NUMBER)
                      primary = TRUE;
                   else
                     if(node_vect[start_node]->outqueue[i]->link_len+
                                       curr_length == primary_length)

                         primary = TRUE;
                 }/* if (!avoid_nodes[curr_node]) */

         if(cg_routing->selected_routing == ROUTING_RA200)
           {
            if(second_choice_selected)
              link_threshold = node_vect[start_node]->outqueue[i]->
               	  routing_threshold[cg_routing->primary_threshold];
            else
              link_threshold = node_vect[start_node]->outqueue[i]->
               	  routing_threshold[cg_routing->alternate_threshold];
           }
         else
            link_threshold = 0.0;

         /* elects curr node as new start node only if its associated link */
         /* cost is minimal (computed with the DLCP function) and it's  */
         /* gerarchy is compatible with the algorithm state.		   */

             if(min_dlcp_cost > dlcp_cost)
                switch(state)

                {
                 case 0:
                  if(primary&&exe_CAC(node_vect[start_node]->outqueue[i],
                                          mean_bw,peak_bw,link_threshold))
                     {
                      min_dlcp_cost = dlcp_cost;
                      temp_node = curr_node;
                      start_link = node_vect[start_node]->outqueue[i];
                     }
                  break;
                 case 1:

                  if(!primary&&exe_CAC(node_vect[start_node]->outqueue[i],
                                        mean_bw,peak_bw,link_threshold)&&
                                               !avoid_nodes[curr_node])
                     {
                      min_dlcp_cost = dlcp_cost;
                      temp_node = curr_node;
                      start_link = node_vect[start_node]->outqueue[i];
                     }
                  break;
                 case 2:
                  if(primary&&exe_CAC(node_vect[start_node]->outqueue[i],
                                         mean_bw,peak_bw,link_threshold))
                     {
                      min_dlcp_cost = dlcp_cost;
                      temp_node = curr_node;
                      start_link = node_vect[start_node]->outqueue[i];
                     }
                  break;
                 default:
                  return NULL;
                }



        } /* for(i=0;i<node_vect[start_node]->no_of_ports; i++) */


    if((start_link == NULL)&&(state == 1))
       return NULL;

    if(state == 1)
       second_choice_selected = TRUE;

    if(start_link == NULL)
       state++;

    else

     {
      start_node = temp_node;
      avoid_nodes[start_node] = TRUE;
      trasversed_links[curr_hop_number+1] = start_link;
      curr_hop_number++;
      curr_length += start_link->link_len;
      start_link = NULL;
     }

   if (second_choice_selected&&(state==1))
     {
      how_long_it_takes_to_destination(start_node,end_node, path_base_ptr,
                                                &hops_to_go, &length_to_go);
      primary_no_of_hops = curr_hop_number + hops_to_go;
      primary_length =  length_to_go + curr_length;
      state++;
     }

   }/* while(start_node != end_node) */

   link_list = path_base_ptr->path;
   i=0;

     do
     {
      if(link_list->link_elem == trasversed_links[i])
       {
        link_list = link_list->next;
        i++;
       }
      else
       {
        i = 0;
        path_base_ptr = path_base_ptr->next;
        link_list = path_base_ptr->path;
        (*curr_path_no)++;
       }
     } while(i != curr_hop_number);


   switch (cg_routing->selected_range_type.type)
     {
      case ONLY_PRIMARY_PATH :
           if (second_choice_selected)
               return NULL;
           break;

      case MAX_HOP_PATH_NUMBER :
           if(cg_routing->selected_range_type.value.ival <
                           path_base_ptr->no_of_hops)
              return NULL;
           break;

      case MAX_LENGTH_PATH_NUMBER :
           if(cg_routing->selected_range_type.value.dval <
                               path_base_ptr->length)
              return NULL;
           break;

      case MAX_HOP_MORE_PATH_NUMBER :
           if(cg_routing->selected_range_type.value.ival <
             path_base_ptr->no_of_hops + path_base_ptr->no_of_hops)
                return NULL;
           break;
      case MAX_LENGTH_MORE_PATH_NUMBER :
           if(cg_routing->selected_range_type.value.dval <
             path_base_ptr->length + path_base_ptr->length)
               return NULL;
           break;

      default:
           break;
     }
  } /* if(first_node != end_node) */

   /* Link connecting last node with end user is checked for CAC */
 link_threshold = 0.0;

 if(!exe_CAC(callgen_vect[cgen_dest]->ref_user->in_link,
                         mean_bw,peak_bw,link_threshold))
    return NULL;

 return path_base_ptr;
}


/*
**  Function  : Call* exe_routing(int cgen_src, int cgen_dest,
**                     double mean_bw, double peak_bw, double min_bw)
**  Return    : pointer to the call if successed or NULL
**  Remarks   : this function executes the actions required by RA.
**              The parameters are:
**               - 'cgen_src' : source call generator.
**               - 'cgen_dest': destination call generator.
**               - 'mean_bw'  : Average Bandwidth of the new call request;
**               - 'peak_bw'  : Peak Bandwidth of the new call request.
**               - 'min_bw'   : Minimum Bandwidth of the new call request,
**                             used for CG_TYPE3 and CG_TYPE4 call generators.
**              The last three parameters are indicated by the user to the
**              network at the call opening request instant.
*/
Call* exe_routing(int cgen_src, int cgen_dest, double mean_bw,
             double peak_bw, double min_bw)
{
  Path* path;			/* Current path. */
  Path* primary_path;           /* Primary path. */
  Pair* primary_pair;           /* Primary pathpair. */
  Path* selected_path;          /* Path selected to route the call. */

  Routing* cg_routing;		/* Routing parameters for the source
				   call generator. */
  double min_path_cost;		/* Minimum dynamic cost for all the
				   paths processed. Used by RA4. */
  double widest_bottleneck;     /* Used by RA6. */

  int     curr_path_no;		/* Current number of analized path. */
  int     curr_pair_no;		/* Current number of analized pathpair. */
  int     selected_path_no;	/* Number of the selected path. */
  Boolean stop_checking;       	/* TRUE when a valid path is found or
				   there are no more paths to check. */

  Boolean is_in_primary[MAX_LINK] = { FALSE };
                         /* Array of flags revealing the presence of the*/
                         /* indexed link in the primary path connecting */
                         /* the current pair of source-destination      */
                         /* nodes.                                      */

  LinkList* link;		/* Used to scan the path links.  */
  int	i, num8;	/* Used by the ra8 algorithm */
  ra8_Path* ra8_path_ptr;	/* Pointer on the actual element of the ra8_path list */
  ra8_Path* sel_ra8_path;	/* Pointer on the ra8_path whoose path is selected */
  int min_hop_num;

  CGType9 *ci9;
  Dbllist *newdbl;
  Dbllist *actdbl;
  Dbllist *prevdbl;

  Call* inscall = NULL;

  double percallubw = 0.0;
  CallList *curr_call;

/*  double bottleneck;*/

#ifdef TRUNKING
  int no_of_checked_links = 0;	/* The number of checked links in the */
				/* current path */
#endif

 last_link = -1;                /* Used to keep track of the last link who */
    		                /* caused the path to be rejected (hop-by-hop...) */

 memset((Boolean*)avoid_nodes,FALSE,MAX_NODE); /*Clean the state of */
     						/* the avoid_nodes array */


  /* determines which RA is used by cgen_src */

  cg_routing = &callgen_vect[cgen_src]->choosen_routing;

  /* selects the primary for the S-D pair, according to the selected RA */

  primary_path = path_mat[cg_routing->path_order_type][cgen_src][cgen_dest];

  selected_path = NULL;

  /* as default no protection is used */
  new_prot_chan = FALSE;
  used_prot_path = NULL;

  /* as default no protection channel dedication is used */
  dedicated_protchan = FALSE;

  /* as default the protection was not best effort */
  protection_be = FALSE;

  /* try to route only if there are paths between the src and dst */
  if(primary_path != NULL)
  {

  /* Always test the primary path            */
  min_path_cost = DBL_MAX;
  widest_bottleneck = 0.0;
  min_hop_num = MAX_NODE+1;
  path = primary_path;
  curr_path_no = 1;
  curr_pair_no = 1;
  stop_checking = FALSE;

  if (IS_DYNAMIC_ROUTING(cg_routing->selected_routing))
    {
     if(cg_routing->selected_routing >= ROUTING_RA50) /* special routings */
      {
       if(cg_routing->selected_routing >= ROUTING_RA200) /* DLCP routings */
        selected_path = dlcp_routing(cgen_src, cgen_dest, cg_routing, mean_bw,
	          peak_bw, &curr_path_no);
       else                      /* routing for ASON-type calls */
        {
		/* Reset the path-link-union array */
		   for(i = 0; i < no_of_links; i++)
			   in_path_union[i] = 0;

        /* In this version we can route a call with only one
         wavelength on an unprotected or protected route. */
		   if(cg_routing->selected_routing >= ROUTING_RA70 &&
			   cg_routing->selected_routing < ROUTING_RA90 ||
			   cg_routing->selected_routing >= ROUTING_RA120 && 
			   cg_routing->selected_routing < ROUTING_RA200)
			   /* protection inserted */
		   {
			   primary_pair = pair_mat[cg_routing->path_order_type]
				   [cgen_src][cgen_dest];
			   selected_path = ason_prot_routing(cgen_src, cgen_dest, cg_routing,
				   &curr_pair_no, primary_pair);
			   selected_path_no = curr_pair_no;
		   }
		   else
		   {
			   selected_path = ason_routing(cgen_src, cgen_dest, cg_routing,
				   &curr_path_no, primary_path);
			   selected_path_no = curr_path_no;
		   }
		/* Reset the path-link-union array */
        }
      }
     else
      {
       if((cg_routing->use_hop_by_hop)||(cg_routing->use_crankback))
        {
         printf("\nError:  User %s uses a centralized dynamic routing algorithm,\n",
                             user_name(callgen_vect[cgen_src]->ref_user->id));
         printf("\t hop by hop or crankback are not allowed. \n");
         exit(1);
        }

       selected_path = primary_path;
       selected_path_no = curr_path_no;

    /* If we use the MSF algorithm, we don't have to search among all the paths.
       There is a fix set of avaible paths constructed already in the get_path
       algorithm. */
       if(cg_routing->selected_routing == ROUTING_RA8)	  /* MSF routing */
	{
     /* Sort the first max_ra8_path Paths (ordered by hop number) between the src
	and the dest, the number of sorts and the sorting criterium depends on the
	algorithm */
	  ra8_path_ptr = def_set_msf(cgen_src, cgen_dest, cg_routing);
	  sel_ra8_path = ra8_path_ptr;

     /* In the first half of the list find the element, which has the less
	hop number*/
	 if(num_ra8_path_2 > 1)
	   num8 = num_ra8_path_2 / 2;
	 else num8 = 1;
	 for(i=0; i < num8; i++)
	   {
	     if( ra8_path_ptr->ref_path->no_of_hops < min_hop_num)
	      {
		min_hop_num = ra8_path_ptr->ref_path->no_of_hops;
		selected_path = ra8_path_ptr->ref_path;
		selected_path_no = ra8_path_ptr->ref_p_no;
		sel_ra8_path = ra8_path_ptr;
	      }
	     ra8_path_ptr = ra8_path_ptr->next;
	   }
	}
       else
	{
    /* If the algorithm choosen for this period by the pac method was
       the shortest path */
	 if((cg_routing->selected_routing == ROUTING_RA9) &&
	     pair_routing_info[cgen_src][cgen_dest].sprou)
	  {
	   selected_path = path;
	   selected_path_no = curr_path_no;
	  }
	 else
	  {
	   double cost, bottleneck;
	   Boolean cong;
	   do
	    {
		  /* Searches until all the paths are processed. */
	     switch (cg_routing->selected_routing)
	      {
	       case ROUTING_RA6:	/* Shortest-widest routing */
		{
		  bottleneck = eval_min_free_bw(path);

        	  if (widest_bottleneck - bottleneck < widest_bottleneck_range)
		    {
		      selected_path = path;
		      selected_path_no = curr_path_no;
  	              widest_bottleneck = bottleneck;
		    }
		}
		break;

	       case ROUTING_RA4:		/* Minimum cost routing		*/
	       case ROUTING_RA5:		/* Load-dependent routing	*/
	       case ROUTING_RA7:		/* Minimum distance routing.	*/
	       case ROUTING_RA9:		/* Periodic Alg. Choice routing.*/
	       case ROUTING_RA11:		/* MinDist-CD aposteori routing */
		{
		  if (cg_routing->selected_routing == ROUTING_RA4)
		    cost = eval_dynamic_path_cost(path);
		  else /* RA5, RA7, RA9 or RA11*/
		    cost = eval_minimum_distance_cost(path);

		  if (cost < min_path_cost
		      && ! equ_double(min_path_cost, cost, DBL_CMP_THS))
		    {
		      selected_path = path;
		      selected_path_no = curr_path_no;
		      min_path_cost = cost;
		    }
		  }
		  break;

	       case ROUTING_RA10:	/* MinDist-CD apriori routing */
	       case ROUTING_RA12:	/* MinCost-CD apriori routing */
		 {
		   cong = cong_on_path(path);
		   if(!cong || (curr_path_no == 1))
		    {
	             if(cg_routing->selected_routing == ROUTING_RA10)
		        cost = eval_minimum_distance_cost(path);
		     else
	                cost = eval_dynamic_path_cost(path);
		    }
		   else
	             cost = DBL_MAX;

		   if (cost < min_path_cost
		       && ! equ_double(min_path_cost, cost, DBL_CMP_THS))
		     {
		       selected_path = path;
		       selected_path_no = curr_path_no;
		       min_path_cost = cost;
		     }
		   }
		   break;

	       default:
		  printf("ERROR: wrong dynamic routing algorithm %d in function exe_RA\n",
		       cg_routing->selected_routing);
		  exit(1);
	      }
	    }
	   while ((path = iterate_path_list(path, &curr_path_no,
					    cg_routing, primary_path)) != NULL);
	  }  /* RA9 && SP else */
	}  /* MSF else */

       /* Here is where we actually run the RA5 routing algorithm */
       if (cg_routing->selected_routing == ROUTING_RA5)
	{
	  double eff;
          /* evaluate the efficiency */
          if(callgen_vect[cgen_src]->type == CG_TYPE5)
	     eff = 1/min_path_cost*selected_path->no_of_hops / peak_bw;
          else
   	     eff = 1/min_path_cost*selected_path->no_of_hops / mean_bw;
	  if (eff < ra5_min_efficiency)
	   {
	    selected_path = primary_path;
	    selected_path_no = 1;
	   }
	}

       /* Here is where we actually run the RA11 routing algorithm */
       if (cg_routing->selected_routing == ROUTING_RA11)
        {
	  Boolean cong = cong_on_path(selected_path);
	  if(cong)
	   {
	    selected_path = primary_path;
	    selected_path_no = 1;
	   }
	}

	 /* We have dynamically selected the path.        */
	 /* Now, we have to run the CAC                   */

	 /* The node-to-node links are checked for enough */
	 /* free bandwidth     */
       link = selected_path->path;
       while(link != NULL)
	 {
	    if(!exe_CAC(link->link_elem, mean_bw, peak_bw, 0.0))
	      break;
	    link = link->next;
	 }

       if(link != NULL)  /* In this case at least one link of the     */
	 {               /* path has not been successfully tested.   */
	    selected_path = NULL;
	 }
      } /* dlcp else */

    } /* end of DYNAMIC routing test */
   else
    { /* STATIC routing test */

      /* Mark the links of the primary path.  Skip the first and the last
	 link (user -> network and network -> user links). */
      link = primary_path->path->next;
      while(link->next != NULL)
	 {
	    is_in_primary[link->link_elem->id] = TRUE;
	    link = link->next;
	 }

      selected_path=NULL;
      do
	 {
	  /* RA3 is implemented as RA2 with null thresholds. */
	  /* RA1 is implemented as RA3 without alternate paths. */
#ifdef TRUNKING
	  no_of_checked_links = 0;
#endif
	  /* The node-to-node links are checked for sufficient */
	  /* free bandwidth     */
	  link = path->path;
	  while(link != NULL)
	    {
	      double link_threshold;

#ifdef TRUNKING
	      no_of_checked_links++;
#endif
	      if( cg_routing->selected_routing == ROUTING_RA2)
		{
                  if((cg_routing->use_hop_by_hop)||(cg_routing->use_crankback))
                   {
                    if( is_in_primary[link->link_elem->id] )
		      link_threshold = link->link_elem
		        ->routing_threshold[cg_routing->primary_threshold];
		    else /* use the alternate_threshold */
		      link_threshold = link->link_elem
		        ->routing_threshold[cg_routing->alternate_threshold];
                   }
                  else
                   {
                    if(curr_path_no == 1)
                      link_threshold = link->link_elem
		        ->routing_threshold[cg_routing->primary_threshold];
                    else
                      link_threshold = link->link_elem
		        ->routing_threshold[cg_routing->alternate_threshold];
                   }
		}
	      else
		link_threshold = 0.0;

	      /* Algorithms other than RA2 execute the CAC with a null threshold. */
	      if(!exe_CAC(link->link_elem, mean_bw, peak_bw, link_threshold))
		break; /* exit the cycle: while(link != NULL) */
	      link = link->next;

	/* last_link refers always the last link of the path which is wide enough */
              if(link != NULL)
                 last_link = link->link_elem->id;
	    }

	   if(link == NULL)     /* In this case all the links of the     */
	    {                   /* path have been successfully tested.   */
	      selected_path = path;
	      selected_path_no = curr_path_no;
	      stop_checking = TRUE;
	    }
#ifdef TRUNKING
	    if( (no_of_checked_links > NO_OF_CHECK_BEFORE_TRUNKING) &&
	       (stop_checking == FALSE) )
	      {
		 /* you can not recover the call */
		 selected_path = NULL; /* XXX */
		 stop_checking = TRUE;
	      }
#endif
	 } while (! stop_checking
		  && (path = iterate_path_list(path, &curr_path_no, cg_routing,
					       primary_path)) != NULL);
    } /* END of STATIC routing test */
  } /* if(primary!=NULL) */
  
  if (selected_path == NULL)
  {
	selected_path_no = 0;

	/* if IPoWDM is used, instead of blocking we try to do something else */
	if(ipowdm_ena && callgen_vect[cgen_src]->type < CG_TYPE6 &&
		!open_loglink)
		open_loglink = 1;
	else
	{
		refused_call_stat_update(cgen_src, cgen_dest);
		/* If the ason call generator is ON-OFF, place the next call. */
		if(callgen_vect[cgen_src]->type == CG_TYPE7)
			schedule_next_arrival(cgen_src, cgen_dest);
		/* If a virtual link is opened, we should not set the indicator */
		if(callgen_vect[cgen_src]->type != CG_TYPE8)
			open_loglink = 0;
	}
  }
  else
  {
  /* if IPoWDM is used, instead of interrupt perform
	  we may try to do something else */

	if(ipowdm_ena && !open_loglink)
	{
		/* the number of hops on the selected route is checked */
		if(selected_path->no_of_hops>vlopen_max_hops)
			selected_path_no=0;

		if(callgen_vect[cgen_src]->type == CG_TYPE5)
		{
		/* estimating the bottleneck bandwidth and do not accept route
			if it seems to be too thin (would cause interrupt) */
			for(link = selected_path->path; link && selected_path_no;
			link = link->next)
			{
				if(link->link_elem->is_node_node)
				{
					
		/* check the bandwidth minimum that the new call would receive on the link */
					percallubw = link_Ucall_stat[link->link_elem->id]->
						free_Uband/(link_Ucall_stat[link->link_elem->id]->
						no_of_Ucall+1);
					
					if(percallubw <	peak_bw * ((CGType5 *)(callgen_vect
						[cgen_src]->info))->vlopen_ubw_thrs)
						selected_path_no = 0;
					
					/* grooming EHC checks it for the other calls too */
					if(vl_open_algorithm == EHC_GROOM)
					{
						for(curr_call = link_Ucall_stat[link->link_elem->id]->
							U_calls_on->next;
						curr_call != link_Ucall_stat[link->link_elem->id]->
							U_calls_on;
						curr_call = curr_call->next)
						{
							if(percallubw <	
								((CallType5 *)(curr_call->call_elem->info))->mbr_value * 
								((CGType5 *)(callgen_vect[curr_call->call_elem->
								source_id]->info))->vlopen_ubw_thrs)
								selected_path_no = 0;
						}
					} /* ehc */
				}
			}
		}
	}
	/* it seems to be too thin or too long */
	if(!selected_path_no)
	{
		open_loglink = 1;	
	}
	else
	{
		inscall = accepted_call_insert(selected_path, selected_path_no,
			cgen_src, cgen_dest, mean_bw, peak_bw, min_bw);

		/* If a virtual link is opened, we should not set the indicator */
		if(callgen_vect[cgen_src]->type != CG_TYPE8)
			open_loglink = 0;
		
			/* The intensity is incremented in the case of 
		Pascal distributed arrivals */
		if(callgen_vect[cgen_src]->type == CG_TYPE9)
		{
			ci9 = (CGType9 *)callgen_vect[cgen_src]->info;
			ci9->arrival_rate += ci9->arr_rate_increment;
			
			/* put the close-time to the ordered list */
			newdbl = dbl_list_alloc();
			newdbl->value = inscall->end_time;
			actdbl = ci9 -> call_closures;
			
			while(actdbl->next && actdbl->value < newdbl->value)
			{
				prevdbl = actdbl;
				actdbl = actdbl->next;
			}
			
			if(actdbl->value >= newdbl->value)
			{
				newdbl->next = actdbl;
				prevdbl->next = newdbl;
			}
			else /* last element in the list */
			{
				newdbl->next = actdbl->next;
				actdbl->next = newdbl;
			}
		}
	}
  }
  return inscall;
}

/*
**  Function  : Call* accepted_call_insert(Path *path_base_ptr, int curr_path_no,
**			 int cgen_src, int cgen_dest, double mean_bw,
**                       double peak_bw, double min_bw)
**  Return    : the inserted call
**  Remarks   : This function insert the new accepted call in the call list
**              of the link (in case the call is a class U call, it will
**		be put in the Ucall list), and update all the statistic
**              involved.
**              The parameters are:
**               - 'path_base_prt: pointer to the path of the call in the net
**		 - 'curr_path_no: the "serial" number of the path
**               - 'cgen_src' : source call generator.
**               - 'cgen_dest': destination call generator.
**               - 'mean_bw'  : Average Bandwidth of the new call request;
**               - 'peak_bw'  : Peak Bandwidth of the new call request.
**               - 'min_bw'   : Minimum Bandwidth of the new call request,
**                             used for CG_TYPE3 and CG_TYPE4 call generators.
**              The last three parameters are indicated by the user to the
**              network at the call opening request instant.
*/
static Call* accepted_call_insert(Path *path_base_ptr, int curr_path_no,
				 int cgen_src, int cgen_dest,
				 double mean_bw, double peak_bw, double min_bw)
{
 LinkList *path_ptr;     /* Pointer to a node-to-node path.             */
 Link     *link_ptr;     /* Generic link pointer.                       */
 Call     *new_call;     /* Pointer to the new call.                    */
 CallList *new_call_list;/* Pointer, for each link, to the new          */
                         /* allocated call.                             */
 Event    *end_event;    /* Pointer to the event relating to the end of */
                         /* the new accepted call: it is inserted in    */
                         /* the Event List at the activation time of    */
                         /* the new call.                               */
 TimeUnit call_duration; /* Duration of the new call.                   */
 double   equival_bw;    /* Equivalent Bandwidth of the accepted call;  */
                         /* this variable is used only with Equivalent  */
                         /* Bandwidth Allocation CAC.                   */
 double data_amount;	 /* The U-calls have, instead of duration, an	*/
 			 /* amount of data to transfer.			*/
 double mgt;		 /* mean_gen_traffic   				*/
 double new_link_len = 0.0;
 int i, chl;

/* The connection request has been accepted: the incoming call enters   */
/* the General Call List and the Local Call Lists.                      */
 new_call = call_alloc();
 new_call->call_type = set_call_type(cgen_src);

 new_call->source_id = cgen_src;
 new_call->dest_id = cgen_dest;
 set_call_informations(new_call, mean_bw, peak_bw, min_bw);

 if(new_call->call_type != CALL_TYPE5)
   call_duration = evaluate_call_duration(cgen_src);
/* Ucalls have traffic amount istead of duration */
 else
  {
   mgt = ((CGType5 *)(callgen_vect[cgen_src]->info))->mean_gen_traffic;
   data_amount = (double)negexp( mgt, &cg_seed_vect[cgen_src]);
   ((CallType5 *)(new_call->info))->tr_left = data_amount;
   call_duration = data_amount / peak_bw;
/* Classification of the calls by the amount of transferable data */
   if(data_amount < mgt)
    {
     if(data_amount < mgt * 0.37988549)
       ((CallType5 *)(new_call->info))->class = 0;
     else
       ((CallType5 *)(new_call->info))->class = 1;
    }
   else
    {
     if(data_amount < mgt * 1.69314718)
       ((CallType5 *)(new_call->info))->class = 2;
     else
       ((CallType5 *)(new_call->info))->class = 3;
    }
  }

 new_call->end_time = ancles_time + call_duration;
 new_call->start_time = ancles_time;
 new_call->my_close_event = NULL;
 new_call->path_ptr = path_base_ptr;
 new_call->curr_path_no = curr_path_no;
 new_call->prot_path_ptr = used_prot_path;
 if(dedicated_protchan)
	 new_call->dedic_prot_ch = used_prot_path->prot_channels;
 else
	 new_call->dedic_prot_ch = NULL;
 new_call->repl = FALSE;

 chl = 0;
 for(i = 0; i < no_of_links; i++)
	 if(in_path_union[i])
		 chl++;

 new_call->no_ch_link = chl;

 insert_call(new_call);

/* Updating the statistical variables involved by the call acceptance.  */

 accepted_call_stat_update(new_call, new_prot_chan, protection_be);

/* A pointer to the new call is allocated for all the links over which  */
/* the call is routed.                                                  */
 i = 0;
 for(path_ptr = new_call->path_ptr->path; path_ptr != NULL;
     path_ptr = path_ptr->next)
   {
     link_ptr = path_ptr->link_elem;
     new_call_list = call_list_alloc();
     new_call_list->call_elem = new_call;
     if(new_call->call_type == CALL_TYPE6 || new_call->call_type == CALL_TYPE8)
        new_call_list->used_wave = u_wave[i++];
     else
      {
     switch (selected_cac)
       {
       case CALL_ADM_CTRL1 :
	 equival_bw = eval_equivalent_bw(mean_bw, peak_bw,
					 capacity / link_ptr->timescale, cell_loss_rate);
	 new_call_list->ref_bw = equival_bw;
	 link_ptr->free_band -= equival_bw;
#ifndef NEW_MEASURE
    avail_bw = update_Ucall_open(link_ptr, mean_bw, peak_bw);
	 if(new_call->call_type == CALL_TYPE5)
	   if(((CallType5*)(new_call->info))->curr_br > avail_bw)
	     ((CallType5*)(new_call->info))->curr_br = avail_bw;
#endif
	 break;

       case CALL_ADM_CTRL2 :
	 new_call_list->ref_bw = peak_bw;
	 link_ptr->free_band -= peak_bw;
	 break;

       case CALL_ADM_CTRL3 :
	 new_call_list->ref_bw = peak_bw;
	 link_ptr->free_band -= peak_bw;
	 updt_var_mean(new_call, &link_mean[link_ptr->id],
		       &link_variance[link_ptr->id], TRUE);
	 break;

       case CALL_ADM_CTRL4 :
	 new_call_list->ref_bw = peak_bw;
	 link_ptr->free_band -= peak_bw;
#ifndef NEW_MEASURE
         avail_bw = update_Ucall_open(link_ptr, mean_bw, peak_bw);
	 if(new_call->call_type == CALL_TYPE5)
	   if(((CallType5*)(new_call->info))->curr_br > avail_bw)
	     ((CallType5*)(new_call->info))->curr_br = avail_bw;
#endif
	 break;
       } /* case */
      }/* else */

	 /* if a virtual link was empty before and now a new call
	 arrived, we have to put the closing event to infinite time */
	 if(ipowdm_ena && link_ptr->id >= no_of_links &&
		 vl_close_algorithm == EMPTY_CLOSE &&
		 link_ptr->linked_call->end_time < max_time)
	 {
		 reinsert_call_and_event(link_ptr->linked_call, max_time + 1 +
			 uniform_0_1(&sh_seed));
	 }


     if(new_call->call_type != CALL_TYPE5)
       insert_call_list_on_link(new_call_list, link_ptr);
     else
       insert_Ucall_list_on_link(new_call_list, link_ptr);
   }

/* The call closure event is scheduled in the Event List.       */
 end_event = event_alloc();
 end_event->ev_time = new_call->end_time;
 end_event->ev_type = CLOSING_CALL;
 end_event->ev_dscrpt.deleting_call = new_call;
 new_call->my_close_event = end_event;
 insert_event(end_event);

#ifdef NEW_MEASURE
 update_bottleneck_measure(new_call, mean_bw, peak_bw, OPEN_CALL);
#endif

#ifndef NEW_MEASURE
 if (new_call->call_type == CALL_TYPE5)
   {
     Link_ucall_stat *temp;

     path_ptr = new_call->path_ptr->path;

     while(path_ptr != NULL)
       {
	 temp = link_Ucall_stat[path_ptr->link_elem->id];
	 temp->un_fb_integ += temp->unused_free_band *
	   (ancles_time - temp->last_unus_time);
	 temp->last_unus_time = ancles_time;
	 temp->unused_free_band -= ((CallType5*)(new_call->info))->curr_br;
	 temp->no_of_Ucall++;
	 path_ptr = path_ptr->next;
       }
   }
#endif
 return new_call;
}


/*
**  Function  : void bottleneck_of_MD(int cgen_src, int cgen_dest)
**  Return    : None.
**  Remarks   : this function searches the path wich would choose
**		the MD algorithm between the two given callgenerators
**		and evaluate the (underestimated) bottleneck bandwidth
**		of this path. (The calculation is for an esisting call
**		and not for an arriving one!)
*/
void bottleneck_of_MD(int cgen_src, int cgen_dest)
{
  Path  *primary_path;		/* Shortes path. */
  Path  *selected_md_path;	/* Choosen path by MD. */
  Path  *path;			/* Actual path. */
  Routing* cg_routing;		/* Routing parameters for the source
  				   call generator. */
  int  curr_path_no;
  double cost, tbw;
  double min_path_cost;

  cg_routing = &callgen_vect[cgen_src]->choosen_routing;
 /* The primary path */
  primary_path = path_mat[cg_routing->path_order_type][cgen_src][cgen_dest];

  min_path_cost = DBL_MAX;
  path = primary_path;
  curr_path_no = 1;

 /* Routing with Minimum-Distance */
  do
   {
   /* Searches until all the paths are processed. */
    cost = ex_eval_minimum_distance_cost(path);
    if (cost < min_path_cost
	&& ! equ_double(min_path_cost, cost, DBL_CMP_THS))
      {
	selected_md_path = path;
	min_path_cost = cost;
      }
   }while ((path = iterate_path_list(path, &curr_path_no,
				      cg_routing, primary_path)) != NULL);

   /* for the new LD method we need an estimation of the bottleneck 	*/
   /* bandwidth value of the MD	*/
    tbw = ex_eval_min_free_bw(selected_md_path);
    if(tbw > ((CGType5 *)(callgen_vect[cgen_src]->info))->max_max_rate)
     pair_routing_info[cgen_src][cgen_dest].Bw_MD =
      ((CGType5 *)(callgen_vect[cgen_src]->info))->max_max_rate;
    else
     pair_routing_info[cgen_src][cgen_dest].Bw_MD = tbw;
}
