/*
**  Module    : ancstat.c
**  Version   : $Revision: 3.3 $
**  Date      : $Date: 1999/01/11 14:53:50 $
**  Authors   : Andrea Francini, Marco Mellia, Giorgio Politano
**              Gabriele Favalessa, Luca Medico
**  Synopsis  : this module collects the functions providing the
**              interface between ANCLES and the "statistics" module and
**              the functions dealing with the management of the output
**              files.
**
**  Functions : void    accepted_call_stat_update(Call *, Boolean, Boolean)
**              void    refused_call_stat_update(int    , int    )
**              void    closed_call_stat_update(Call *)
**              void    reset(void)
**              void    stop_simulating(int)
**              void    generate_CBR_report(FILE *)
**              void    generate_CL_report(FILE *)
**              void    generate_OBR_report(FILE *)
**              void    generate_AWL_report(FILE *)
**              void    generate_AWH_report(FILE *)
**              void    generate_ABW_report(FILE *)
**              void    generate_LWH_report(FILE *)
**              void    generate_WLC_report(FILE *)
**              void    generate_ROU_report(FILE *)
**              void    generate_UBR_report(FILE *)
**              void    generate_IBR_report(FILE *)
**              void    generate_ULH_report(FILE *)
**              void    generate_ILH_report(FILE *)
**              void    generate_4CC_report(FILE *)
**              void    generate_UCL_report(FILE *)
**              void    generate_HST_report(FILE *)
**              void    generate_STR_report()
**
**
**  History   : 23-OCT-1995 - Creation of the module.      -AF-
**              01-JUN-1996 - Revision.                    -AF-
**              01-MAR-1997 - Revision			   -MM-
**              06-FEB-1998 - Changed the management of the path
**                            to cgen -> cgen matrix.
**                            Added the support to the user <->user
**                            traffic statistics
**                            Added the support to the primary load
**                            carried by each link statistics
**                            Added the support to the user <-> user
**                            call blocking probabilities statistics
**                            Added the support to the distance class
**                            call blocking probabilities statistics
**                                                        -MM - GF - LM-
**		05-OCT-1999 - Revision
**				Some new statistical variables are put
**				to support the examination of the duration
**				and effectivity of best effort calls - ZZ -

26-01-2001 Cleaning... - ZZ -
30-05-2001 ASON support - ZZ -
*/
#include "ancles.h"
#include <time.h>

static Boolean ctrl_traff_start(int    );
static Boolean ctrl_traff_pattern(int    );



/***************************
**  Variable Definitions  **
***************************/


/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : void link_stat_update(int     link_id, int     pseudo_user,
**                                                           double load_contr)
**  Return    : none.
**  Remarks   : Update the statistics about the carried link load
**              identified by link_id, coming from pseudo_user user (pseudo
**              because we have also the TOTAL_LINK_LOAD and the
**              PRIMARY_LINK_LOAD). The load_contr is the average requested
**              bandwidth bu the call, and may be positive (accepted call) or
**              negative (closed call).
*/
static void link_stat_update(int     link_id, int     pseudo_user,
                                                             double load_contr)
{
  TimeUnit delta_t;            /* elapsed time from the previous update */

  delta_t = ancles_time - prev_load_update_time[link_id][pseudo_user];

  if (pseudo_user == TOTAL_LINK_LOAD)
  /* check if the histogram must be updated */
    {
      if(hst_intervals[link_id] > 0)
	{
	  double unit = capacity /
	    (hst_intervals[link_id] * link_vect[link_id]->timescale);
	  int     index =
	    (int    )floor(link_carried_load[link_id][pseudo_user] / unit);

	  if(index >= hst_intervals[link_id])
	    hst_link_load[link_id][hst_intervals[link_id]] += delta_t;
	  else
	    hst_link_load[link_id][index] += delta_t;
	}
    }
/* update the statistic */
  load_accumulator[link_id][pseudo_user] +=
    link_carried_load[link_id][pseudo_user] * delta_t;

  link_carried_load[link_id][pseudo_user] += load_contr;

  prev_load_update_time[link_id][pseudo_user] = ancles_time;
}

/*
**  Function  : void ason_link_stat_update(int  link_id, int  pseudo_user,
**                                              double load_contr, int peak_wave)
**  Return    : none.
**  Remarks   : Update the statistics similarly to the normal update of links,
**              but no histogram is used since the statistics of the peak
**              wavelength and utilization is calculated.
*/
static void ason_link_stat_update(int  link_id, int  pseudo_user,
                                  double load_contr, double peak_wave)
{
  TimeUnit delta_t;            /* elapsed time from the previous update */
  int j;
/* unsigned char mask;*/

  delta_t = ancles_time - prev_ason_link_update_time[link_id][pseudo_user];

/* update the statistic */
  ason_wave_accumulator[link_id][pseudo_user] +=
    ason_link_carried_waves[link_id][pseudo_user] * delta_t;

  ason_peak_accumulator[link_id][pseudo_user] +=
    ason_link_peak[link_id][pseudo_user] * delta_t;
  if(pseudo_user == TOTAL_ASON_LINK_LOAD)
  {
	  /* update the histograms */
	  histo_ason_wave_accu[link_id]
		  [(int)(ason_link_carried_waves[link_id][TOTAL_ASON_LINK_LOAD])] +=
		  delta_t;
	  
	  histo_ason_peak_accu[link_id]
		  [(int)(ason_link_peak[link_id][TOTAL_ASON_LINK_LOAD])] +=
		  delta_t;

	/* update the link-wavelength statistics */

	  /* time statistics of the mean number of free fibers */

	  for(j = 0; j < link_vect[link_id]->max_waves; j++)
	  {
		  free_asonfibers_accu[link_id][j] += 
			  link_vect[link_id]->free_fibers[j] * delta_t;
		  if(minf_asonfibers[link_id][j] >
				link_vect[link_id]->free_fibers[j])
			  minf_asonfibers[link_id][j] =
				link_vect[link_id]->free_fibers[j];
	  }
  }

/* Don't care about the zero utilization periods. */
  if(ason_link_util[link_id][pseudo_user] > 0.0)
	  ason_util_accumulator[link_id][pseudo_user] +=
         ason_link_util[link_id][pseudo_user] * delta_t;
  else
	  ason_util_zero_time[link_id] += delta_t;

  ason_link_carried_waves[link_id][pseudo_user] += load_contr;
  ason_link_peak[link_id][pseudo_user] = peak_wave;
/* debug */
  if(link_vect[link_id]->is_node_node &&
    ason_link_carried_waves[link_id][pseudo_user]>
	ason_link_peak[link_id][pseudo_user])
	peak_wave = ason_link_peak[link_id][pseudo_user];
  
  if(peak_wave != 0)
    ason_link_util[link_id][pseudo_user] =
      ason_link_carried_waves[link_id][pseudo_user] / peak_wave;
  else
    ason_link_util[link_id][pseudo_user] = 0.0;

  prev_ason_link_update_time[link_id][pseudo_user] = ancles_time;
}

/*
**  Function  : void accepted_call_stat_update(Call *in_call,
**						Boolean newprotch)
**  Return    : none.
**  Remarks   : all the statistical variables involved by the acceptance of
**              'in_call' are updated here.
**              The values of 'tot_carried_load' and 'link_carried_load'
**              are also updated.
**				If newprotch is true the links of the protection path
**				have to be updated too.
**				If prot_be is true, it means that the protection could not
**				be performed, but as best effort we route on one path
*/
void accepted_call_stat_update(Call *in_call, Boolean newprotch,
							   Boolean prot_be)
{
 int      user_id, link_id;
 double   load_contr;
 LinkList *link_ptr;

 CallType2 *ci2;
 CallType3 *ci3;
 CallType4 *ci4;

 user_id = callgen_vect[in_call->source_id]->ref_user->id;

/* The number of requested calls is updated.                            */
 ++req_calls[user_id][callgen_vect[in_call->dest_id]->ref_user->id];

/* The number of protection - best effort calls is updated.             */
 if( prot_be )
	 ++pbe_calls[user_id][callgen_vect[in_call->dest_id]->ref_user->id];

/* Hop, length and number of crossed opaque nodes
   statistics of the paths is updated. */
 path_hops_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
	 in_call->path_ptr->no_of_hops;
 
 path_length_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
	 in_call->path_ptr->length;

 path_crop_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
	 in_call->path_ptr->no_of_opaque_crs;

 if(in_call->prot_path_ptr)
 {
 /* Hop, length and number of crossed opaque nodes
	 statistics of the protection paths is updated. */
	 prot_path_hops_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
		 in_call->prot_path_ptr->no_of_hops;
	 
	 prot_path_length_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
		 in_call->prot_path_ptr->length;
	 
	 prot_path_crop_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
		 in_call->prot_path_ptr->no_of_opaque_crs;
 }

 /* Statistic about the number of links checked during the routing */
 acc_noch_link_accu[user_id][callgen_vect[in_call->dest_id]->ref_user->id]+=
	 in_call->no_ch_link;


 
 /* Update the number of accepted calls. */
 if(in_call->curr_path_no == 1)
   ++primary_calls[user_id];
 else
   ++alternate_calls[user_id];

 /* update the histogram of the number of paths used by the Alternate	*/
 /* routing algorithms.							*/
 if(in_call->call_type != CALL_TYPE5 && !ipowdm_ena)
     no_times_this_path_is_utilized[in_call->curr_path_no-1] ++;

/* If the accepted call belongs to Class U, no other actions are        */
/* required.                                                            */
 if(in_call->call_type == CALL_TYPE5)
    return;

/* Updating the user call blocking probability.                         */
 if(user_stat_vars[user_id].info != NOT_ANALYSED)
     b_statistics(user_stat_vars[user_id].number, 0);

 switch(in_call->call_type)
   {
    case CALL_TYPE1 :
        load_contr = ((CallType1 *)in_call->info)->cbr_value;
        break;

    case CALL_TYPE2 :
	ci2 = (CallType2 *)(in_call->info);
        load_contr = ci2->mean_on_time * ci2->peak_rate /
		     (ci2->mean_on_time + ci2->mean_off_time);
        break;

    case CALL_TYPE3 :
	ci3 = (CallType3 *)(in_call->info);
        load_contr = (ci3->max_vbr_value + ci3->min_vbr_value) * 0.5;
        break;

    case CALL_TYPE4 :
	ci4 = (CallType4 *)(in_call->info);
        load_contr = ci4->higher_bit_rate_value * ci4->higher_bit_rate_prob +
                     ci4->medium_bit_rate_value * ci4->medium_bit_rate_prob +
                     ci4->lower_bit_rate_value * ci4->lower_bit_rate_prob;
        break;

    case CALL_TYPE6 :
        load_contr = (double)((CallType6 *)in_call->info)->nwaves;
        break;

    case CALL_TYPE8 :
        load_contr = (double)((CallType8 *)in_call->info)->nwaves;
        break;

    default:
        fprintf(stderr, "ancstat.c:accepted_call_stat_update()");
        fprintf(stderr, " unknown call generator type\n");
        exit(1);
   }

 if(in_call->call_type != CALL_TYPE6 && in_call->call_type != CALL_TYPE8)
  {
   total_carried_load += load_contr;
   /* Updating the total accepted traffic. */
   if (global_stat_vars[0].info != NOT_ANALYSED)
    c_statistics(global_stat_vars[0].number, total_carried_load, ancles_time);

   for(link_ptr = in_call->path_ptr->path; link_ptr != NULL;
     link_ptr = link_ptr->next)
    {
     link_id = link_ptr->link_elem->id;
	 /* no statistics about virtual links */
	 if(link_id >= no_of_links)
		 continue;

     link_stat_update(link_id, TOTAL_LINK_LOAD, load_contr);

     link_stat_update(link_id, user_id, load_contr);

     if (in_call->curr_path_no == 1)
       link_stat_update(link_id, PRIMARY_LINK_LOAD, load_contr);

     if(link_stat_vars[link_id].info != NOT_ANALYSED)
       c_statistics(link_stat_vars[link_id].number,
		    link_carried_load[link_id][TOTAL_LINK_LOAD], ancles_time);
    }
  }
 else
 {
	 total_carried_waves += load_contr;
	 /* Updating the total accepted traffic. */
	 if (global_stat_vars[2].info != NOT_ANALYSED)
		 c_statistics(global_stat_vars[2].number, total_carried_waves, ancles_time);
	 
	 for(link_ptr = in_call->path_ptr->path; link_ptr != NULL;
     link_ptr = link_ptr->next)
	 {
		 link_id = link_ptr->link_elem->id;
		 
		 ason_link_stat_update(link_id, TOTAL_ASON_LINK_LOAD, load_contr,
			 (double)link_vect[link_id]->peak_wave);
		 
		 /* Not used */
		 /*     ason_link_stat_update(link_id, user_id, load_contr); 
		 
		   if (in_call->curr_path_no == 1)
		   ason_link_stat_update(link_id, PRIMARY_ASON_LINK_LOAD, load_contr,
		 (double)link_vect[link_id]->peak_wave); */
		 
		 if(ason_link_stat_vars[link_id].info != NOT_ANALYSED)
			 c_statistics(ason_link_stat_vars[link_id].number,
			 ason_link_carried_waves[link_id][TOTAL_ASON_LINK_LOAD], ancles_time);
     }
  /* Statistics of the links on the protection path (only the node-node links) */
	 if(newprotch)
	 {
		 for(link_ptr = in_call->prot_path_ptr->path->next; link_ptr->next != NULL;
		 link_ptr = link_ptr->next)
		 {
			 link_id = link_ptr->link_elem->id;
			 ason_link_stat_update(link_id, TOTAL_ASON_LINK_LOAD, load_contr,
				 (double)link_vect[link_id]->peak_wave);
			 if(ason_link_stat_vars[link_id].info != NOT_ANALYSED)
				 c_statistics(ason_link_stat_vars[link_id].number,
				 ason_link_carried_waves[link_id][TOTAL_ASON_LINK_LOAD], ancles_time);
		 }
	 }
 }

/* interarrival time statistics */ 
 if(in_call->call_type == CALL_TYPE8)
 {
	 accu_vlreq_iat[user_id] += ancles_time - last_vlreq_arrival[user_id];
	 sqraccu_vlreq_iat[user_id] += (ancles_time - last_vlreq_arrival[user_id])*
		 (ancles_time - last_vlreq_arrival[user_id]);
	 last_vlreq_arrival[user_id] = ancles_time;
 }

/* Updating the total call blocking probability.                        */
 if (global_stat_vars[1].info != NOT_ANALYSED)
   b_statistics(global_stat_vars[1].number, 0);
}

/*
**  Function  : void refused_call_stat_update(int     cg_id,
**                                           int     callgen_dest)
**  Return    : none.
**  Remarks   : all the statistical variables involved by the refusal of
**              a connection request coming from the generator identified
**              by 'cg_id' and going to the call generator callgen_dest
**              are updated here.
*/
void refused_call_stat_update(int     cg_id, int     callgen_dest)
{
 int      user_id;
 int      user_dest;
 int	i,j,k;
 unsigned char	mask;

 user_id = callgen_vect[cg_id]->ref_user->id;
 user_dest = callgen_vect[callgen_dest]->ref_user->id;

/* The number of both requested and refused calls are updated.          */
 ++req_calls[user_id][user_dest];
 ++ref_calls[user_id][user_dest];

 if(callgen_vect[cg_id]->type == CG_TYPE6 ||
	 callgen_vect[cg_id]->type == CG_TYPE7 ||
	 callgen_vect[cg_id]->type == CG_TYPE8 ||
	 callgen_vect[cg_id]->type == CG_TYPE9)
 {
	 for(i = 0; i < no_of_links; i++)
	 {
/* Take the samples only from the links of the path union which was
   examined in the routing. */
         if(in_path_union[i])
		 {
			 for(j = 0; j < max_wvoct_all_links && 
		    j < ((link_vect[i]->max_waves-1)/8)+1; j++)
			 {
				 mask = link_vect[i]->free_waves[j];
				 for(k = 1; k < 9; k++)
				 {
					 if(!(mask & 0x80))
						 inblock_asonwaves_accu[i][8*j+k]+= 1;
						 /* in_path_union; */
					 mask <<= 1;
				 }
			 }
		 }
	 }
 }

/* Statistic about the number of links checked during the routing */

 j = 0;
 for(i = 0; i < no_of_links; i++)
	 if(in_path_union[i])
		 j++;

 ref_noch_link_accu[user_id][user_dest] += j;

/* For a Class U Call Generator, the blocking probability is updated    */
/* separately.                                                          */
 if(callgen_vect[cg_id]->type == CG_TYPE5)
    return;

/* Updating the Total Call Blocking Probability.        */
 if (global_stat_vars[1].info != NOT_ANALYSED)
   b_statistics(global_stat_vars[1].number, 1);

/* Updating the User Call Blocking Probability.         */
 if(user_stat_vars[user_id].info != NOT_ANALYSED)
      b_statistics(user_stat_vars[user_id].number, 1);

 /* interarrival time statistics */ 
 if(callgen_vect[cg_id]->type == CG_TYPE8)
 {
	 accu_vlreq_iat[user_id] +=  ancles_time - last_vlreq_arrival[user_id];
	 sqraccu_vlreq_iat[user_id] += (ancles_time - last_vlreq_arrival[user_id])*
		 (ancles_time - last_vlreq_arrival[user_id]);
	 last_vlreq_arrival[user_id] = ancles_time;
 }
}


/*
**  Function  : void closed_call_stat_update(Call *out_call,
**							Boolean protchclose)
**  Return    : none.
**  Remarks   : all the statistical variables involved by the closure of
**              'out_call' are updated here.
**				If a protection channel is closed, the links of the 
**				protection path have to be updated.
*/
void closed_call_stat_update(Call *out_call, Boolean protchclose)
{
 int      user_id, dest_id, link_id, len_idx;
 double   load_contr, all_traffic;
 LinkList *link_ptr;

 CallType2 *ci2;
 CallType3 *ci3;
 CallType4 *ci4;

 user_id = callgen_vect[out_call->source_id]->ref_user->id;
 dest_id = callgen_vect[out_call->dest_id]->ref_user->id;

 switch(out_call->call_type)
   {
    case CALL_TYPE1 :
        load_contr = ((CallType1 *)out_call->info)->cbr_value;
        break;

    case CALL_TYPE2 :
        ci2 = ((CallType2 *)out_call->info);
	load_contr = ci2->mean_on_time * ci2->peak_rate /
                     (ci2->mean_on_time + ci2->mean_off_time);
        break;

    case CALL_TYPE3 :
        ci3 = ((CallType3 *)out_call->info);
        load_contr = (ci3->max_vbr_value + ci3->min_vbr_value) * 0.5;
        break;

    case CALL_TYPE4 :
        ci4 = ((CallType4 *)out_call->info);
        load_contr = ci4->higher_bit_rate_value * ci4->higher_bit_rate_prob +
                     ci4->medium_bit_rate_value * ci4->medium_bit_rate_prob +
                     ci4->lower_bit_rate_value * ci4->lower_bit_rate_prob;
        break;

    case CALL_TYPE5 :   /* The variables carrying the performance       */
       {                /* indices for Class U calls are updated here.  */
        CallType5 *this_call = ((CallType5 *)(out_call->info));
		all_traffic = this_call->br_integ + this_call->tr_left;
        this_call->br_integ += this_call->curr_br *
                                (ancles_time - this_call->last_updt_time);
        this_call->tr_left -= this_call->curr_br *
                                (ancles_time - this_call->last_updt_time);
        this_call->last_updt_time = ancles_time;
    /* The histogram index is determined according to the call-length	*/
		len_idx = (int)floor((double)LEN_IDX_MAX * ((out_call->end_time -
			this_call->last_load_time)/((CGType5 *)callgen_vect
			[out_call->source_id]->info)->longest_call));

	if(!interrupt_happened)
	 {
          gen_finUload[out_call->source_id] += (this_call->br_integ /
                                   (ancles_time - this_call->last_load_time));

	  gen_finUtras[out_call->source_id] += this_call->br_integ;
    /* The dilatation of the call-length is registred.	*/
	  fin_calls_dilacc[user_id][dest_id] +=
	      (ancles_time - this_call->last_load_time)/(all_traffic/
	      this_call->mbr_value);

	  usr_finUload[user_id][dest_id] += 
		  this_call->br_integ/(ancles_time - this_call->last_load_time);

    /* The number of finished calls will be updated according to class	*/
	  classed_fin_calls[user_id][dest_id][this_call->class]++;
    /* The histogramm-array of the call-length will be updated */
	  if(len_idx < LEN_IDX_MAX)
	    U_fincall_dur[user_id][len_idx]++;
	  else
	    U_fincall_dur[user_id][LEN_IDX_MAX-1]++;
	 }	
	else
	 {	/* If the call was interrupted	*/
          gen_intUload[out_call->source_id] += (this_call->br_integ /
                                   (ancles_time - this_call->last_load_time));
	  gen_intUtras[out_call->source_id] += this_call->br_integ;

    /* Transported data ratio statistics is updated according to class	*/
	  classed_prop_intUtras[out_call->source_id][this_call->class] +=
	    (this_call->br_integ / all_traffic);
    /* The number of interrupted calls is updated according to class	*/
	  classed_int_calls[user_id][dest_id][this_call->class]++;

    /* The number of the interrupted calls will be incremented and the
       histogramm-array of the call-length will be updated		*/
	  int_calls[user_id][dest_id]++;

	  if(len_idx < LEN_IDX_MAX)
	    U_intcall_dur[user_id][len_idx]++;
	  else
	    U_intcall_dur[user_id][LEN_IDX_MAX-1]++;
	 }

	if (classU_user_stat_vars[user_id].info != NOT_ANALYSED)
	  {
	  c_statistics(classU_user_stat_vars[user_id].number,
		       this_call->br_integ / (ancles_time - this_call->last_load_time),
		       ancles_time);
	  }

/* The performance indices regarding the Class U traffic are not under  */
/* statistical analysis, so that the function execution can be stopped  */
/* here.                                                                */
        return;
       }
    case CALL_TYPE6 :
        load_contr = (double)((CallType6 *)out_call->info)->nwaves;
    case CALL_TYPE8 :
        load_contr = (double)((CallType8 *)out_call->info)->nwaves;
        break;

    default:
        fprintf(stderr, "ancstat.c:closed_call_stat_update()");
        fprintf(stderr, " unknown call generator type\n");
        exit(1);
   }

 if(out_call->call_type != CALL_TYPE6 && out_call->call_type != CALL_TYPE8)
  {
   total_carried_load -= load_contr;
/* Updating the total accepted traffic. */
   if (global_stat_vars[0].info != NOT_ANALYSED)
    c_statistics(global_stat_vars[0].number, total_carried_load, ancles_time);

   for(link_ptr = out_call->path_ptr->path; link_ptr != NULL;
         link_ptr = link_ptr->next)
    {
     link_id = link_ptr->link_elem->id;

	 /* no statistics about virtual links */
	 if(link_id >= no_of_links)
		 continue;
	 
	 link_stat_update(link_id, TOTAL_LINK_LOAD, - load_contr);
	 
	 link_stat_update(link_id, user_id, - load_contr);
	 
	 if (out_call->curr_path_no == 1)
		 link_stat_update(link_id, PRIMARY_LINK_LOAD, - load_contr);
	 
     if(link_stat_vars[link_id].info != NOT_ANALYSED)
              c_statistics(link_stat_vars[link_id].number,
                   link_carried_load[link_id][TOTAL_LINK_LOAD], ancles_time);
    }
  }
 else
  {
   total_carried_waves -= load_contr;
   /* Updating the total accepted traffic. */
   if (global_stat_vars[2].info != NOT_ANALYSED)
    c_statistics(global_stat_vars[2].number, total_carried_waves, ancles_time);

   for(link_ptr = out_call->path_ptr->path; link_ptr != NULL;
     link_ptr = link_ptr->next)
    {
     link_id = link_ptr->link_elem->id;

     ason_link_stat_update(link_id, TOTAL_ASON_LINK_LOAD, -load_contr,
      (double)link_vect[link_id]->peak_wave);

     /* Not used */
/*     ason_link_stat_update(link_id, user_id, -load_contr); 

     if (out_call->curr_path_no == 1)
       ason_link_stat_update(link_id, PRIMARY_ASON_LINK_LOAD, -load_contr,
        (double)link_vect[link_id]->peak_wave); */

     if(ason_link_stat_vars[link_id].info != NOT_ANALYSED)
       c_statistics(ason_link_stat_vars[link_id].number,
		    ason_link_carried_waves[link_id][TOTAL_ASON_LINK_LOAD], ancles_time);
     }
  /* Statistics of the links on the protection path (only the node-node links) */
	 if(protchclose)
	 {
		 for(link_ptr = out_call->prot_path_ptr->path->next; link_ptr->next != NULL;
		 link_ptr = link_ptr->next)
		 {
			 link_id = link_ptr->link_elem->id;
			 ason_link_stat_update(link_id, TOTAL_ASON_LINK_LOAD, -load_contr,
				 (double)link_vect[link_id]->peak_wave);
			 if(ason_link_stat_vars[link_id].info != NOT_ANALYSED)
				 c_statistics(ason_link_stat_vars[link_id].number,
				 ason_link_carried_waves[link_id][TOTAL_ASON_LINK_LOAD], ancles_time);
		 }
	 }
  }
/* accumulate the call durations */
 inon_accumulator[user_id][dest_id] += ancles_time - out_call->start_time;

}

/*
**  Function  : void init_statmem(void)
**  Return    : none.
**  Remarks   : Allocates memory (if needed) and initialize the global variables
**              used in the statistics (and others :) .
*/
void init_statmem(void)
{
	int i, j;

/* Initialization of the stationary phase start time vectors.           */
	sched_events_start_time = 0.0;
	for(i = 0; i < 3; i++)
		global_start_time[i] = 0.0;

	user_start_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_users, "init_statmem");

	classU_user_start_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_users, "init_statmem");

	for(i = 0; i < no_of_users; i++)
	{
		classU_user_start_time[i] = 0.0;
		user_start_time[i] = 0.0;
	}

	link_start_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_links, "init_statmem");
	ason_link_start_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_links, "init_statmem");
	for(i = 0; i < no_of_links; i++)
    {
		link_start_time[i] = 0.0;
		ason_link_start_time[i] = 0.0;
    }


	head_of_VC_list = (VCset **)
		MMmalloc(sizeof(VCset *) * no_of_users, "init_statmem");

	photo_vect = (PHattr **)
		MMmalloc(sizeof(PHattr *) * no_of_links, "init_statmem");

	
	link_carried_load = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	load_accumulator = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");

	ason_link_carried_waves = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	ason_link_peak = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	ason_link_util = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");

	ason_wave_accumulator = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	ason_peak_accumulator = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	ason_util_accumulator = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");

	prev_load_update_time = (TimeUnit **)
		MMmalloc(sizeof(TimeUnit*) * no_of_links, "init_statmem");
	prev_ason_link_update_time = (TimeUnit **)
		MMmalloc(sizeof(TimeUnit*) * no_of_links, "init_statmem");

	ason_util_zero_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_links, "init_statmem");
	
	histo_ason_wave_accu = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	histo_ason_peak_accu = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	free_asonfibers_accu = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");
	minf_asonfibers = (double **)
		MMmalloc(sizeof(double *) * no_of_links, "init_statmem");

	inblock_asonwaves_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_links, "init_statmem");
	calls_on_link_wave = (long **)
		MMmalloc(sizeof(long *) * no_of_links, "init_statmem");

	in_path_union = (int *)
		MMmalloc(sizeof(int) * no_of_links, "init_statmem");

/* Initializing the links carried loads and ason-waves accumulators. */
	total_carried_load = 0.0;
	total_carried_waves = 0.0;

	for(i = 0; i < no_of_links; i++)
    {
		link_carried_load[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_LINK_LOAD + 1),
			"init_statmem");
		load_accumulator[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_LINK_LOAD + 1),
			"init_statmem");
		ason_link_carried_waves[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");
		ason_link_peak[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");
		ason_link_util[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");

		ason_wave_accumulator[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");
		ason_peak_accumulator[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");
		ason_util_accumulator[i] = (double *)
			MMmalloc(sizeof(double) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");

		prev_load_update_time[i] = (TimeUnit *)
			MMmalloc(sizeof(TimeUnit) * (TOTAL_LINK_LOAD + 1),
			"init_statmem");
		prev_ason_link_update_time[i] = (TimeUnit *)
			MMmalloc(sizeof(TimeUnit) * (TOTAL_ASON_LINK_LOAD + 1),
			"init_statmem");
		

		histo_ason_wave_accu[i] = (double *)
			MMmalloc(sizeof(double) * (8*MAX_WVOCTETS + 1),
			"init_statmem");
		histo_ason_peak_accu[i] = (double *)
			MMmalloc(sizeof(double) * (8*MAX_WVOCTETS + 1),
			"init_statmem");
		free_asonfibers_accu[i] = (double *)
			MMmalloc(sizeof(double) * (8*MAX_WVOCTETS + 1),
			"init_statmem");
		minf_asonfibers[i] = (double *)
			MMmalloc(sizeof(double) * (8*MAX_WVOCTETS + 1),
			"init_statmem");

		
		inblock_asonwaves_accu[i] = (long *)
			MMmalloc(sizeof(long) * (8*MAX_WVOCTETS + 1),
			"init_statmem");
		calls_on_link_wave[i] = (long *)
			MMmalloc(sizeof(long) * (8*MAX_WVOCTETS + 1),
			"init_statmem");
		
		link_carried_load[i][TOTAL_LINK_LOAD]=0.0;
		load_accumulator[i][TOTAL_LINK_LOAD] = 0.0;
		prev_load_update_time[i][TOTAL_LINK_LOAD] = 0.0;
		
		link_carried_load[i][PRIMARY_LINK_LOAD]=0.0;
		load_accumulator[i][PRIMARY_LINK_LOAD] = 0.0;
		prev_load_update_time[i][PRIMARY_LINK_LOAD] = 0.0;

		for(j = 0; j< no_of_users; j++)
		{
			link_carried_load[i][j] = 0.0;
			load_accumulator[i][j] = 0.0;
			prev_load_update_time[i][j] = 0.0;
		}

		
		ason_link_carried_waves[i][TOTAL_ASON_LINK_LOAD]=0.0;
		ason_link_peak[i][TOTAL_ASON_LINK_LOAD]=0.0;
		ason_link_util[i][TOTAL_ASON_LINK_LOAD]=0.0;
		ason_wave_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
		ason_peak_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
		ason_util_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
		prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD] = 0.0;

		ason_link_carried_waves[i][PRIMARY_ASON_LINK_LOAD]=0.0;
		ason_link_peak[i][PRIMARY_ASON_LINK_LOAD]=0.0;
		ason_link_util[i][PRIMARY_ASON_LINK_LOAD]=0.0;
		ason_wave_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
		ason_peak_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
		ason_util_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
		prev_ason_link_update_time[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
		
		ason_util_zero_time[i] = 0.0;

		for(j = 0; j < 8*MAX_WVOCTETS + 1; j++)
		{
			histo_ason_wave_accu[i][j]=0.0;
			histo_ason_peak_accu[i][j]=0.0;
			inblock_asonwaves_accu[i][j] = 0;
			calls_on_link_wave[i][j] = 0;
			free_asonfibers_accu[i][j] = 0.0;
			minf_asonfibers[i][j] = MAX_FIBER_NUM;
		}
    }
	
	pair_routing_info = (R_info **)
		MMmalloc(sizeof(R_info *) * no_of_cgens, "init_statmem");

	gen_finUload = (double *)
		MMmalloc(sizeof(double) * no_of_cgens, "init_statmem");
	gen_finUtras = (double *)
		MMmalloc(sizeof(double) * no_of_cgens, "init_statmem");
	gen_intUload = (double *)
		MMmalloc(sizeof(double) * no_of_cgens, "init_statmem");
	gen_intUtras = (double *)
		MMmalloc(sizeof(double) * no_of_cgens, "init_statmem");
	classed_prop_intUtras = (double **)
		MMmalloc(sizeof(double*) * no_of_cgens, "init_statmem");
	gen_Uload_start_time = (TimeUnit *)
		MMmalloc(sizeof(TimeUnit) * no_of_cgens, "init_statmem");

	for(i = 0; i < no_of_cgens; i++)
    {
		classed_prop_intUtras[i] = (double *)
			MMmalloc(sizeof(double) * 4, "init_statmem");

		pair_routing_info[i] = (R_info *)
			MMmalloc(sizeof(R_info) * no_of_cgens, "init_statmem");
	}

/* Initializing the accumulators used to evaluate the users' call       */
/* blocking probabilities and the counters referring to the number      */
/* of accepted calls over primary and alternate routes, user by user.   */

	
	req_calls = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	ref_calls = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	pbe_calls = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	classed_fin_calls = (long ***)
		MMmalloc(sizeof(long **) * no_of_users, "init_statmem");
	int_calls = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	classed_int_calls = (long ***)
		MMmalloc(sizeof(long **) * no_of_users, "init_statmem");
	fin_calls_dilacc = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	usr_finUload = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");

	hopdist = (int **)
		MMmalloc(sizeof(int *) * no_of_users, "init_statmem");
	path_hops_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	path_length_accu = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	path_crop_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	
	prot_path_hops_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	prot_path_length_accu = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	prot_path_crop_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");

	acc_noch_link_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	ref_noch_link_accu = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");

	accu_vlink_lifetime = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	sqraccu_vlink_lifetime = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	closed_vlinks = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	accu_vlink_timehop = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");
	rt_sp_changed = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");

	last_vlreq_arrival = (double *)
		MMmalloc(sizeof(double) * no_of_users, "init_statmem");
	accu_vlreq_iat = (double *)
		MMmalloc(sizeof(double) * no_of_users, "init_statmem");
	sqraccu_vlreq_iat = (double *)
		MMmalloc(sizeof(double) * no_of_users, "init_statmem");
	calls_in_at_reset = (long *)
		MMmalloc(sizeof(long) * no_of_users, "init_statmem");

	inon_accumulator = (double **)
		MMmalloc(sizeof(double *) * no_of_users, "init_statmem");

	primary_calls = (long *)
		MMmalloc(sizeof(long) * no_of_users, "init_statmem");
	alternate_calls = (long *)
		MMmalloc(sizeof(long) * no_of_users, "init_statmem");
	U_fincall_dur = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
	U_intcall_dur = (long **)
		MMmalloc(sizeof(long *) * no_of_users, "init_statmem");

	for(i = 0; i < no_of_users; i++)
    {
		req_calls[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		ref_calls[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
 		pbe_calls[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		classed_fin_calls[i] = (long **)
			MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
		int_calls[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		classed_int_calls[i] = (long **)
			MMmalloc(sizeof(long *) * no_of_users, "init_statmem");
		fin_calls_dilacc[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		usr_finUload[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		
		hopdist[i] = (int *)
			MMmalloc(sizeof(int) * no_of_users, "init_statmem");
		path_hops_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		path_length_accu[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		path_crop_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		
		prot_path_hops_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		prot_path_length_accu[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		prot_path_crop_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");

		acc_noch_link_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		ref_noch_link_accu[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		
		accu_vlink_lifetime[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		sqraccu_vlink_lifetime[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		closed_vlinks[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
		accu_vlink_timehop[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		rt_sp_changed[i] = (long *)
			MMmalloc(sizeof(long) * no_of_users, "init_statmem");
	
		inon_accumulator[i] = (double *)
			MMmalloc(sizeof(double) * no_of_users, "init_statmem");
		
		U_fincall_dur[i] = (long *)
			MMmalloc(sizeof(long) * LEN_IDX_MAX, "init_statmem");
		U_intcall_dur[i] = (long *)
			MMmalloc(sizeof(long) * LEN_IDX_MAX, "init_statmem");
		
		last_vlreq_arrival[i] = 0.0;
		accu_vlreq_iat[i] = 0.0;
		sqraccu_vlreq_iat[i] = 0.0;
		
		for(j = 0; j < no_of_users; j++)
		{
			classed_fin_calls[i][j] = (long *)
				MMmalloc(sizeof(long) * 4, "init_statmem");
			classed_int_calls[i][j] = (long *)
				MMmalloc(sizeof(long) * 4, "init_statmem");

			req_calls[i][j] = 0;
			ref_calls[i][j] = 0;
			pbe_calls[i][j] = 0;
			int_calls[i][j] = 0;
			
			fin_calls_dilacc[i][j] = 0.0;
			usr_finUload[i][j] = 0.0;
			classed_fin_calls[i][j][0] = 0; classed_fin_calls[i][j][1] = 0;
			classed_fin_calls[i][j][2] = 0; classed_fin_calls[i][j][3] = 0;
			classed_int_calls[i][j][0] = 0; classed_int_calls[i][j][1] = 0;
			classed_int_calls[i][j][2] = 0; classed_int_calls[i][j][3] = 0;
			
			inon_accumulator[i][j] = 0.0;

			path_hops_accu[i][j] = 0;
			path_length_accu[i][j] = 0.0;
			path_crop_accu[i][j] = 0;
			prot_path_hops_accu[i][j] = 0;
			prot_path_length_accu[i][j] = 0.0;
			prot_path_crop_accu[i][j] = 0;

			acc_noch_link_accu[i][j] = 0;
			ref_noch_link_accu[i][j] = 0;

			accu_vlink_lifetime[i][j] = 0.0;
			sqraccu_vlink_lifetime[i][j] = 0.0;
			closed_vlinks[i][j] = 0;
			accu_vlink_timehop[i][j] = 0.0;
			rt_sp_changed[i][j] = 0;
		}

		primary_calls[i] = 0;
		alternate_calls[i] = 0;

		for(j = 0; j < LEN_IDX_MAX; j++)
		{
			U_fincall_dur[i][j] = 0;
			U_intcall_dur[i][j] = 0;
		}

	}

	link_primary_load = (double *)
		MMmalloc(sizeof(double) * no_of_links, "init_statmem");

}

/*
**  Function  : void free_statistic_arrays(void)
**  Return    : none.
**  Remarks   : Free the memory allocated for the global variables of statistics.
*/
void free_statistic_arrays(void)
{
	int i, j;

	free(user_start_time);

	free(classU_user_start_time);

	free(link_start_time);
	free(ason_link_start_time);

	for(i = 0; i < no_of_links; i++)
    {
		free(link_carried_load[i]);
		free(load_accumulator[i]);
		free(ason_link_carried_waves[i]);
		free(ason_link_peak[i]);
		free(ason_link_util[i]);

		free(ason_wave_accumulator[i]);
		free(ason_peak_accumulator[i]);
		free(ason_util_accumulator[i]);

		free(prev_load_update_time[i]);
		free(prev_ason_link_update_time[i]);
		

		free(histo_ason_wave_accu[i]);
		free(histo_ason_peak_accu[i]);
		free(free_asonfibers_accu[i]);
		free(minf_asonfibers[i]);

		
		free(inblock_asonwaves_accu[i]);
		free(calls_on_link_wave[i]);

		free(hst_link_load[i]);
    }
	free(head_of_VC_list);
	free(photo_vect);

	free(link_carried_load);
	free(load_accumulator);

	free(ason_link_carried_waves);
	free(ason_link_peak);
	free(ason_link_util);

	free(ason_wave_accumulator);
	free(ason_peak_accumulator);
	free(ason_util_accumulator);

	free(prev_load_update_time);
	free(prev_ason_link_update_time);

	free(ason_util_zero_time);
	
	free(histo_ason_wave_accu);
	free(histo_ason_peak_accu);
	free(free_asonfibers_accu);
	free(minf_asonfibers);

	free(inblock_asonwaves_accu);
	free(calls_on_link_wave);

	free(in_path_union);

	for(i = 0; i < no_of_cgens; i++)
    {
		free(classed_prop_intUtras[i]);
		free(pair_routing_info[i]);
	}
	
	free(pair_routing_info);

	free(gen_finUload);
	free(gen_finUtras);
	free(gen_intUload);
	free(gen_intUtras);
	free(classed_prop_intUtras);
	free(gen_Uload_start_time);

	for(i = 0; i < no_of_users; i++)
    {
		free(req_calls[i]);
		free(ref_calls[i]);
		free(pbe_calls[i]);
		free(classed_fin_calls[i]);
		free(int_calls[i]);
		free(classed_int_calls[i]);
		free(fin_calls_dilacc[i]);
		free(usr_finUload[i]);
		
		free(hopdist[i]);
		free(path_hops_accu[i]);
		free(path_length_accu[i]);
		free(path_crop_accu[i]);
		
		free(prot_path_hops_accu[i]);
		free(prot_path_length_accu[i]);
		free(prot_path_crop_accu[i]);
		
		free(acc_noch_link_accu[i]);
		free(ref_noch_link_accu[i]);

		free(accu_vlink_lifetime[i]);
		free(sqraccu_vlink_lifetime[i]);
		free(closed_vlinks[i]);
		free(accu_vlink_timehop[i]);
		free(rt_sp_changed[i]);

		free(inon_accumulator[i]);
		
		free(U_fincall_dur[i]);
		free(U_intcall_dur[i]);
		
		for(j = 0; j < no_of_users; j++)
		{
			free(classed_fin_calls[i][j]);
			free(classed_int_calls[i][j]);
		}
	}

	free(req_calls);
	free(ref_calls);
	free(pbe_calls);
	free(classed_fin_calls);
	free(int_calls);
	free(classed_int_calls);
	free(fin_calls_dilacc);
	free(usr_finUload);

	free(hopdist);
	free(path_hops_accu);
	free(path_length_accu);
	free(path_crop_accu);
	
	free(prot_path_hops_accu);
	free(prot_path_length_accu);
	free(prot_path_crop_accu);

	free(acc_noch_link_accu);
	free(ref_noch_link_accu);

	free(accu_vlink_lifetime);
	free(sqraccu_vlink_lifetime);
	free(closed_vlinks);
	free(accu_vlink_timehop);
	free(rt_sp_changed);

	free(last_vlreq_arrival);
	free(accu_vlreq_iat);
	free(sqraccu_vlreq_iat);
	free(calls_in_at_reset);

	free(inon_accumulator);

	free(primary_calls);
	free(alternate_calls);
	free(U_fincall_dur);
	free(U_intcall_dur);

	free(link_primary_load);
}

/*
**  Function  : void reset(void)
**  Return    : none.
**  Remarks   : this function is executed at the end of the initial
**              transient for every significant process. The initial
**              transient of a non significant or not statistically analysed
**              process is over when the first transient closure occurs
**              for a significant one.
*/
void reset(void)
{
 register int i, j;
 Call *call_ptr;
 CallList *this_call;
 Link_ucall_stat  *actual_lUs;

/* The counters referring to not significant statistical variables or   */
/* to variables that are not statistically analyzed are reset at the    */
/* first closure of initial transient for a significant process, i.e.   */
/* their initial transient is considered over when the first transient  */
/* closure occurs for a significant variable.                           */
/* Moreover the counter of FES insertion operations and the accumulator */
/* for the number of events seen in the FES by a new insertion are      */
/* reset at the first call of 'reset', i.e. the mean number of events   */
/* in the FES seen by an incoming event is considered as a not          */
/* statistically analysed process whose initial transient ends when     */
/* the first transient closure occurs for a significant variable.       */
 if(first_time)
   {
    first_time = FALSE;

	/* If we want to observe the histeresys behaviour of the network	*/
	/* caused by traffic reduction after the transient phase			*/
	if(gen_load_red_factor < 1.0)
		for(i = 0; i < no_of_cgens; i++)
		{
			switch (callgen_vect[i]->type)
			{
			case CG_TYPE1:
				((CGType1 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			case CG_TYPE2:
				((CGType2 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			case CG_TYPE3:
				((CGType3 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			case CG_TYPE4:
				((CGType4 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			case CG_TYPE5:
				((CGType5 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			case CG_TYPE6:
				((CGType6 *)(callgen_vect[i]->info))->arrival_rate *=
				gen_load_red_factor;
				break;

			default:;
			}
		}

    no_of_FES_operations = 0;
    sched_events_start_time = ancles_time;
    sched_events_acc = 0;

    /* setting the sample_deviation to the given parameter value (NewLD)*/
    sample_deviation = sample_deviation_parameter;

    /* resetting the histogram of the call path distribution over paths */
    for(i=0; i< max_cgen_cgen_no_of_path; i++)
       no_times_this_path_is_utilized[i]=0;

    for(i = 0; i < 3; i++)
        if(global_stat_vars[i].info != SIGNIFICANT)
	  {
              global_start_time[i] = ancles_time;
              global_trans_closed[i] = TRUE;
	  }

    for(i = 0; i < no_of_users; i++)
        if(user_stat_vars[i].info != SIGNIFICANT)
		{
              user_start_time[i] = ancles_time;
              user_trans_closed[i] = TRUE;
              for(j = 0; j < LEN_IDX_MAX; j++)
			  {
				  U_fincall_dur[i][j] = 0;
				  U_intcall_dur[i][j] = 0;
			  }
			  
			  calls_in_at_reset[i] = 0;
			  accu_vlreq_iat[i] = 0.0;
			  sqraccu_vlreq_iat[i] = 0.0;
			  
			  for(j = 0; j < no_of_users; j++)
			  {
				  req_calls[i][j] = 0;
				  ref_calls[i][j] = 0;
				  pbe_calls[i][j] = 0;
				  
				  path_hops_accu[i][j] = 0;
				  path_length_accu[i][j] = 0.0;
				  path_crop_accu[i][j] = 0;
				  
				  prot_path_hops_accu[i][j] = 0;
				  prot_path_length_accu[i][j] = 0.0;
				  prot_path_crop_accu[i][j] = 0;
				  
				  acc_noch_link_accu[i][j] = 0;
				  ref_noch_link_accu[i][j] = 0;

				  accu_vlink_lifetime[i][j] = 0.0;
				  sqraccu_vlink_lifetime[i][j] = 0.0;
				  closed_vlinks[i][j] = 0;
				  accu_vlink_timehop[i][j] = 0.0;
				  rt_sp_changed[i][j] = 0;

				  inon_accumulator[i][j] = 0.0;
				  
				  int_calls[i][j] = 0;
				  fin_calls_dilacc[i][j] = 0.0;
				  usr_finUload[i][j] = 0.0;
				  classed_fin_calls[i][j][0] = 0; classed_fin_calls[i][j][1] = 0;
				  classed_fin_calls[i][j][2] = 0; classed_fin_calls[i][j][3] = 0;
				  classed_int_calls[i][j][0] = 0; classed_int_calls[i][j][1] = 0;
				  classed_int_calls[i][j][2] = 0; classed_int_calls[i][j][3] = 0;
			  }
              primary_calls [i] = 0;
              alternate_calls[i] = 0;
			  /* Reset call-length histogram */
			  for(j = 0; j < LEN_IDX_MAX; j++)
				  U_fincall_dur[i][j] = 0;
			  
				  /* The number of the requested (and accepted) calls in the
			  network is not 0! */
			  for(call_ptr = head_of_call_list->next;
			  call_ptr != head_of_call_list; call_ptr = call_ptr->next)
			  {
				  if(callgen_vect[call_ptr->source_id]->ref_user->id == i)
				  {
					  if(call_ptr->curr_path_no == 1)
						  ++primary_calls[i];
					  else
						  ++alternate_calls[i];
					  
					  req_calls[i][callgen_vect[call_ptr->dest_id]->ref_user->id]++;
					  calls_in_at_reset[i]++;
					  
					  if(callgen_vect[call_ptr->source_id]->choosen_routing
						  .selected_routing >= ROUTING_RA70 &&
						  callgen_vect[call_ptr->source_id]->choosen_routing
						  .selected_routing < ROUTING_RA90 &&
						  call_ptr->prot_path_ptr == NULL)
						  pbe_calls[i][callgen_vect[call_ptr->dest_id]->ref_user->id]++;

					  path_hops_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
						  call_ptr->path_ptr->no_of_hops;
					  path_length_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
						  call_ptr->path_ptr->length;
					  path_crop_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
						  call_ptr->path_ptr->no_of_opaque_crs;
					  
					  acc_noch_link_accu[i]
						  [callgen_vect[call_ptr->dest_id]->ref_user->id]+=
						  call_ptr->no_ch_link;
					  
					  if(call_ptr->prot_path_ptr)
					  {
						  prot_path_hops_accu[i]
							  [callgen_vect[call_ptr->dest_id]->ref_user->id]+=
							  call_ptr->prot_path_ptr->no_of_hops;
						  prot_path_length_accu[i]
							  [callgen_vect[call_ptr->dest_id]->ref_user->id]+=
							  call_ptr->prot_path_ptr->length;
						  prot_path_crop_accu[i]
							  [callgen_vect[call_ptr->dest_id]->ref_user->id]+=
							  call_ptr->prot_path_ptr->no_of_opaque_crs;
					  }

					  /* normal calls seem to be started in this instant */
					  if(call_ptr->call_type != CALL_TYPE8)
						  call_ptr->start_time = user_start_time[i];
				  }
			  }
			  
		}


    for(i = 0; i < no_of_users; i++)
      if(classU_user_stat_vars[i].info != SIGNIFICANT)
	{
	  classU_user_start_time[i] = ancles_time;
	  classU_user_trans_closed[i] = TRUE;
	}

    for(i = 0; i < no_of_links; i++)
        if(link_stat_vars[i].info != SIGNIFICANT)
             {
              link_start_time[i] = ancles_time;
              link_trans_closed[i] = TRUE;
	      for(j = 0; j<no_of_users; j++)
	        {
		 prev_load_update_time[i][j] = ancles_time;
		 load_accumulator[i][j] = 0.0;
		}
              prev_load_update_time[i][TOTAL_LINK_LOAD] = ancles_time;
	      load_accumulator[i][TOTAL_LINK_LOAD] = 0.0;

              prev_load_update_time[i][PRIMARY_LINK_LOAD] = ancles_time;
	      load_accumulator[i][PRIMARY_LINK_LOAD] = 0.0;

	      if(hst_intervals[i] > 0)
                for(j = 0; j <= hst_intervals[i]; j++)
                        hst_link_load[i][j] = 0.0;
             }
      /* Statistics of ASON calls */
    for(i = 0; i < no_of_links; i++)
        if(ason_link_stat_vars[i].info != SIGNIFICANT)
		{
			ason_link_start_time[i] = ancles_time;
			ason_util_zero_time[i] = 0.0;
			ason_link_trans_closed[i] = TRUE;
			prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD] =
				ancles_time;
	           ason_wave_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
			   ason_peak_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
			   ason_util_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
			   for(j = 0; j < 8*MAX_WVOCTETS + 1; j++)
			   {
				   histo_ason_wave_accu[i][j]=0.0;
				   histo_ason_peak_accu[i][j]=0.0;
				   inblock_asonwaves_accu[i][j] = 0;
				   calls_on_link_wave[i][j] = 0;
				   free_asonfibers_accu[i][j] = 0.0;
				   minf_asonfibers[i][j] = MAX_FIBER_NUM;
			   }
			   
			   prev_ason_link_update_time[i][PRIMARY_ASON_LINK_LOAD] =
				   ancles_time;
			   ason_wave_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
			   ason_peak_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
			   ason_util_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
		}
/* Resetting the statistical variables referring to class-U calls.      */
    if( UCall_present )
      {
       for(i = 0; i < no_of_links; i++)
          {
	   actual_lUs = link_Ucall_stat[i];

/* Resetting the variables associated to the network links.     */
           actual_lUs->start_time = ancles_time;
	   actual_lUs->U_PCR_int = 0.0;
	   actual_lUs->no_of_Ucall_integ = 0.0;
           actual_lUs->last_Uupdt_time = ancles_time;
           actual_lUs->fb_integ = 0.0;
           actual_lUs->un_fb_integ = 0.0;
           actual_lUs->last_unus_time = ancles_time;
           actual_lUs->last_updt_time = ancles_time;

/* Resetting the fields of the class-U call records.            */
           for(this_call = actual_lUs->U_calls_on->next;
	       this_call != actual_lUs->U_calls_on;
	       this_call = this_call->next)
             {
              ((CallType5 *)(this_call->call_elem)->info)->br_integ = 0.0;
              ((CallType5 *)(this_call->call_elem)->info)->last_updt_time =
                                ancles_time;
              ((CallType5 *)(this_call->call_elem)->info)->last_load_time =
                                ancles_time;
             }
          }

/* Resetting the accumulators referring to class-U call acceptance.     */
       for(i = 0; i < no_of_cgens; i++)
                if(callgen_vect[i]->type == CG_TYPE5)
		   {
                     gen_finUload[i] = 0.0;
		     gen_finUtras[i] = 0.0;
                     gen_intUload[i] = 0.0;
		     gen_intUtras[i] = 0.0;
		     classed_prop_intUtras[i][0] = 0.0;
		     classed_prop_intUtras[i][1] = 0.0;
		     classed_prop_intUtras[i][2] = 0.0;
		     classed_prop_intUtras[i][3] = 0.0;
		   }
      }
   }


/* The counters of a significant variable are reset only when its       */
/* initial transient is over.                                           */
 for(i = 0; i < 3; i++)
     if( (global_stat_vars[i].info == SIGNIFICANT) &&
         (global_trans_closed[i] == FALSE) &&
         (transient_is_over(global_stat_vars[i].number)) )
           {
            global_trans_closed[i] = TRUE;
            global_start_time[i] = ancles_time;
            return;
           }

 for(i = 0; i < no_of_users; i++)
     if( (user_stat_vars[i].info == SIGNIFICANT) &&
         (user_trans_closed[i] == FALSE) &&
         (transient_is_over(user_stat_vars[i].number) ) )
           {
            user_trans_closed[i] = TRUE;
            user_start_time[i] = ancles_time;
            for(j = 0; j < LEN_IDX_MAX; j++)
              {
				U_fincall_dur[i][j] = 0;
				U_intcall_dur[i][j] = 0;
              }
			calls_in_at_reset[i] = 0;
			accu_vlreq_iat[i] = 0.0;
			sqraccu_vlreq_iat[i] = 0.0;
			
			for(j = 0; j < no_of_users; j++)
            {
	           req_calls[i][j] = 0;
               ref_calls[i][j] = 0;
	           pbe_calls[i][j] = 0;
			   path_hops_accu[i][j] = 0;
			   path_length_accu[i][j] = 0.0;
			   path_crop_accu[i][j] = 0;
			   prot_path_hops_accu[i][j] = 0;
			   prot_path_length_accu[i][j] = 0.0;
			   prot_path_crop_accu[i][j] = 0;
			   
			   acc_noch_link_accu[i][j] = 0;
			   ref_noch_link_accu[i][j] = 0;

			   accu_vlink_lifetime[i][j] = 0.0;
			   sqraccu_vlink_lifetime[i][j] = 0.0;
			   closed_vlinks[i][j] = 0;
			   accu_vlink_timehop[i][j] = 0.0;
			   rt_sp_changed[i][j] = 0;

			   inon_accumulator[i][j] = 0.0;

               int_calls[i][j] = 0;
               fin_calls_dilacc[i][j] = 0.0;
               usr_finUload[i][j] = 0.0;
		       classed_fin_calls[i][j][0] = 0; classed_fin_calls[i][j][1] = 0;
               classed_fin_calls[i][j][2] = 0; classed_fin_calls[i][j][3] = 0;
               classed_int_calls[i][j][0] = 0; classed_int_calls[i][j][1] = 0;
               classed_int_calls[i][j][2] = 0; classed_int_calls[i][j][3] = 0;
			}
            primary_calls[i] = 0;
            alternate_calls[i] = 0;
	/* Reset call-length histogram */
	    for(j = 0; j < LEN_IDX_MAX; j++)
	     U_fincall_dur[i][j] = 0;

	/* The number of the requested (and accepted) calls in the
	   network is not 0! */
	    for(call_ptr = head_of_call_list->next;
	        call_ptr != head_of_call_list; call_ptr = call_ptr->next)
	     {
	      if(callgen_vect[call_ptr->source_id]->ref_user->id == i)
	       {
		if(call_ptr->curr_path_no == 1)
		 ++primary_calls[i];
		else
		 ++alternate_calls[i];

		req_calls[i][callgen_vect[call_ptr->dest_id]->ref_user->id]++;

		calls_in_at_reset[i]++;

		if(callgen_vect[call_ptr->source_id]->choosen_routing
			.selected_routing >= ROUTING_RA70 &&
			callgen_vect[call_ptr->source_id]->choosen_routing
			.selected_routing < ROUTING_RA90 &&
			call_ptr->prot_path_ptr == NULL)
			pbe_calls[i][callgen_vect[call_ptr->dest_id]->ref_user->id]++;
		path_hops_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
			  call_ptr->path_ptr->no_of_hops;
		path_length_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
			  call_ptr->path_ptr->length;
		path_crop_accu[i][callgen_vect[call_ptr->dest_id]->ref_user->id]+=
			  call_ptr->path_ptr->no_of_opaque_crs;

		acc_noch_link_accu[i]
			[callgen_vect[call_ptr->dest_id]->ref_user->id]+=
			call_ptr->no_ch_link;

		if(call_ptr->prot_path_ptr)
		{
			prot_path_hops_accu[i]
				[callgen_vect[call_ptr->dest_id]->ref_user->id]+=
				call_ptr->prot_path_ptr->no_of_hops;
			prot_path_length_accu[i]
				[callgen_vect[call_ptr->dest_id]->ref_user->id]+=
				call_ptr->prot_path_ptr->length;
			prot_path_crop_accu[i]
				[callgen_vect[call_ptr->dest_id]->ref_user->id]+=
				call_ptr->prot_path_ptr->no_of_opaque_crs;
		}
		/* normal calls seem to be started in this instant */
		if(call_ptr->call_type != CALL_TYPE8)
			call_ptr->start_time = user_start_time[i];

		  }
	     }

            return;
           }

 for(i = 0; i < no_of_users; i++)
   if( (classU_user_stat_vars[i].info == SIGNIFICANT) &&
       (classU_user_trans_closed[i] == FALSE) &&
       (transient_is_over(classU_user_stat_vars[i].number) ) )
     {
       classU_user_trans_closed[i] = TRUE;
       classU_user_start_time[i] = ancles_time;
       return;
     }

 for(i = 0; i < no_of_links; i++)
     if( (link_stat_vars[i].info == SIGNIFICANT) &&
         ( (link_trans_closed[i] == FALSE) &&
           transient_is_over(link_stat_vars[i].number) ) )
           {
            link_trans_closed[i] = TRUE;
            link_start_time[i] = ancles_time;
	    for( j = 0; j<no_of_users; j++)
              {
	       prev_load_update_time[i][j] = ancles_time;
	       load_accumulator[i][j] = 0.0;
	      }
            prev_load_update_time[i][TOTAL_LINK_LOAD] = ancles_time;
            load_accumulator[i][TOTAL_LINK_LOAD] = 0.0;

            prev_load_update_time[i][PRIMARY_LINK_LOAD] = ancles_time;
            load_accumulator[i][PRIMARY_LINK_LOAD] = 0.0;

            if(hst_intervals[i] > 0)
                for(j = 0; j <= hst_intervals[i]; j++)
                        hst_link_load[i][j] = 0.0;
            return;
           }
      /* Statistics of ASON calls */
 for(i = 0; i < no_of_links; i++)
     if(ason_link_stat_vars[i].info == SIGNIFICANT &&
         ( (ason_link_trans_closed[i] == FALSE) &&
           transient_is_over(ason_link_stat_vars[i].number) ) )
             {
              ason_link_start_time[i] = ancles_time;
			  ason_util_zero_time[i] = 0.0;
              ason_link_trans_closed[i] = TRUE;
              prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD] =
                  ancles_time;
	           ason_wave_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
	           ason_peak_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;
			   for(j = 0; j < 8*MAX_WVOCTETS + 1; j++)
			   {
				   histo_ason_wave_accu[i][j]=0.0;
				   histo_ason_peak_accu[i][j]=0.0;
				   inblock_asonwaves_accu[i][j] = 0;
				   calls_on_link_wave[i][j] = 0;
				   free_asonfibers_accu[i][j] = 0.0;
				   minf_asonfibers[i][j] = MAX_FIBER_NUM;
			   }
	           ason_util_accumulator[i][TOTAL_ASON_LINK_LOAD] = 0.0;

              prev_ason_link_update_time[i][PRIMARY_ASON_LINK_LOAD] =
                  ancles_time;
	           ason_wave_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
	           ason_peak_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;
	           ason_util_accumulator[i][PRIMARY_ASON_LINK_LOAD] = 0.0;

              return;
             }
}

/*
**  Function  : void stop_simulating()
**  Return    : none.
**  Remarks   : the simulation is stopped, the results are printed and the
**              program execution is closed.
*/
void stop_simulating()
{
 register int   i;
 char           out_file[80];
 FILE           *out_fp;

 end_CPU_time = clock();        /* Getting the CPU time elapsed at the  */
                                /* end of the simulation run.           */
 free(tfree);
 
 /* CBR file.    */
 strcpy(out_file, file_radix);  /* Opening the file collecting data     */
 strcat(out_file, ".cbr");      /* about Call Blocking Ratios.          */
 if((out_fp = fopen(out_file, "w")) == NULL)
   {
    printf("Warning:  The file %s cannot be opened.\n", out_file);
    printf("          Properly formatted data regarding Call ");
    printf("Blocking Probabilities\n          will not be available.\n");
   }
 else
   {
    generate_CBR_report(out_fp);
    fclose(out_fp);
    }

 /* CDL file.    */
 strcpy(out_file, file_radix);  /* Opening the file collecting data     */
 strcat(out_file, ".cdl");      /* about Call Durations and Loads.      */
 if((out_fp = fopen(out_file, "w")) == NULL)
   {
    printf("Warning:  The file %s cannot be opened.\n", out_file);
    printf("          Properly formatted data regarding Call ");
    printf("Durations and Loads\n          will not be available.\n");
   }
 else
   {
    generate_CDL_report(out_fp);
    fclose(out_fp);
    }

 if( nonASON_Call_present )
 {
	 /* CL file.      */
	 strcpy(out_file, file_radix);  /* Opening the file collecting data  */
	 strcat(out_file, ".cl");       /* about Carried Load.                */
	 if((out_fp = fopen(out_file, "w")) == NULL)
	 {
		 printf("Warning:  The file %s cannot be opened.\n", out_file);
		 printf("          Properly formatted data regarding Carried Loads\n");
		 printf("          will not be available.\n");
	 }
	 else
	 {
		 generate_CL_report(out_fp);
		 fclose(out_fp);
	 }
 }

 if(no_of_type6_cg || no_of_type7_cg || no_of_type8_cg || no_of_type9_cg)
   {
/* AWL file.    */
    strcpy(out_file, file_radix);  /* Opening the file collecting data  */
    strcat(out_file, ".awl");      /* about the ASON links (waves...)   */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the utilization\n");
       printf("          of (ASON) optical links will not be available.\n");
      }
    else
      {
       generate_AWL_report(out_fp);
       fclose(out_fp);
      }
/* AWH file.    */
    strcpy(out_file, file_radix);  /* Opening the file with histograms  */
    strcat(out_file, ".awh");      /* about the ASON links (used, peak wave) */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the used and peak\n");
       printf("          histograms of (ASON) optical links will not be available.\n");
      }
    else
      {
       generate_AWH_report(out_fp);
       fclose(out_fp);
      }

/* ABW file.    */
    strcpy(out_file, file_radix);  /* Opening the file with wave occupacy  */
    strcat(out_file, ".abw");      /* at blocking situation */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the occupacy\n");
       printf("          of wavelengths at blocking will not be available.\n");
      }
    else
      {
       generate_ABW_report(out_fp);
       fclose(out_fp);
      }
   }

/* LWH file.    */
    strcpy(out_file, file_radix);  /* Opening the file with histograms  */
    strcat(out_file, ".lwh");      /* about the ASON links (occupancy) */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the occupancy\n");
       printf("          histograms of (ASON) optical links will not be available.\n");
      }
    else
      {
       generate_LWH_report(out_fp);
       fclose(out_fp);
      }

/* WLC file.    */
    strcpy(out_file, file_radix);  /* Opening the file with number of  */
    strcat(out_file, ".wlc");      /* calls on the ASON link-wavelengths */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the call number on\n");
       printf("          (ASON) optical link-wavelengths will not be available.\n");
      }
    else
      {
       generate_WLC_report(out_fp);
       fclose(out_fp);
      }


/* ROU file.    */
    if(list_routes)
	{
		strcpy(out_file, file_radix);  /* Opening the file with the lists of */
		strcat(out_file, ".rou");      /* the available routes between cgens */
		if((out_fp = fopen(out_file, "w")) == NULL)
		{
			printf("Warning:  The file %s cannot be opened.\n", out_file);
			printf("          Properly formatted data regarding the lists of routes \n");
			printf("          will not be available.\n");
		}
		else
		{
			generate_ROU_report(out_fp);
			fclose(out_fp);
		}
	}

/* CG8 file.    */
    if(ipowdm_ena)
	{
		strcpy(out_file, file_radix);  /* Opening the file with info on */
		strcat(out_file, ".cg8");      /* the virtual links */
		if((out_fp = fopen(out_file, "a")) == NULL)
		{
			printf("Warning:  The file %s cannot be opened.\n", out_file);
			printf("          Properly formatted data regarding the virtual IP links\n");
			printf("          will not be available.\n");
		}
		else
		{
			generate_CG8_report(out_fp);
			fclose(out_fp);
		}
	}

/* PAT file.    *
 strcpy(out_file, file_radix);  /* Opening the file collecting data     *
 strcat(out_file, ".pat");      /* about available paths .          *
 if((out_fp = fopen(out_file, "w")) == NULL)
   {
    printf("Warning:  The file %s cannot be opened.\n", out_file);
    printf("          Properly formatted data regarding ");
    printf("Available Paths\n          will not be available.\n");
   }
 else
   {
    generate_PAT_report(out_fp);
    fclose(out_fp);
    }
*/
 if( UCall_present )
   {

/* UCL file.    */
    strcpy(out_file, file_radix);       /* Opening the file collecting  */
    strcat(out_file, ".ucl");           /* data about Carried Load for  */
                                        /* calls belonging to Class U.  */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding Carried Loads");
       printf(" for Class U calls\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_UCL_report(out_fp);
       fclose(out_fp);
      }

/* UBR file.    */
    strcpy(out_file, file_radix);  /* Opening the file collecting data  */
    strcat(out_file, ".ubr");      /* about the UBR generator bit rate  */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding the Bit Rate");
       printf(" for class U generators\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_UBR_report(out_fp);
       fclose(out_fp);
      }

   if(interrupt_statistics)
    {
/* ULH file.    */
    strcpy(out_file, file_radix);  /* Opening the file with data about	*/
    strcat(out_file, ".ulh");      /* the finished UBR call lengths	*/
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding Call Duration");
       printf(" for class U generators\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_ULH_report(out_fp);
       fclose(out_fp);
      }
/* ILH file.    */
    strcpy(out_file, file_radix);  /* Opening the file with data about	*/
    strcat(out_file, ".ilh");      /* the interrupted UBR call lengths	*/
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding interrupted Call ");
       printf("Duration for class U generators\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_ILH_report(out_fp);
       fclose(out_fp);
      }
/* IBR file.    */
    strcpy(out_file, file_radix);  /* Opening the file collecting data  */
    strcat(out_file, ".ibr");      /* about the interrupted UBR calls	*/
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding interrupted Call ");
       printf("Informations for class U generators\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_IBR_report(out_fp);
       fclose(out_fp);
      }
/* 4CC file.    */
    strcpy(out_file, file_radix);  /* Opening the file collecting data  */
    strcat(out_file, ".4cc");      /* about the classified UBR calls	*/
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding Classified Call ");
       printf("Informations for class U generators\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_4CC_report(out_fp);
       fclose(out_fp);
      }
     }
   }

 for(i = 0; (i < no_of_links) && (hst_intervals[i] <= 0); i++);
 if(i < no_of_links)
   {
/* HST file.    */
    strcpy(out_file, file_radix);       /* Opening the file collecting  */
    strcat(out_file, ".hst");           /* data about Load Histograms.  */
    if((out_fp = fopen(out_file, "w")) == NULL)
      {
       printf("Warning:  The file %s cannot be opened.\n", out_file);
       printf("          Properly formatted data regarding Load Histograms\n");
       printf("          will not be available.\n");
      }
    else
      {
       generate_HST_report(out_fp);
       fclose(out_fp);
      }
   }

/* STR file.    */
 generate_STR_report();
 fclose(fp_statistics);

 if (max_state_photos > 0)
   rename_photo_files();

 exit (0);
}


/*
**  Function  : void generate_CBR_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the CBR report.
*/
void generate_CBR_report(FILE * out_fp)
{
 register int   i, j, jj;
 long		sreq_calls;
 long		sref_calls;
 long		spbe_calls;
/* int     	user_src;
 int     	user_dest;*/
 int    	ncol = 8;
 int		M;
 int		maxM;

 long acc_req_calls = 0;
 long acc_ref_calls = 0;
 long acc_pbe_calls = 0;
 long acc_primary = 0;
 long acc_alternate = 0;

 long acc_req_Ucalls = 0;
 long acc_ref_Ucalls = 0;
 long acc_Uprimary = 0;
 long acc_Ualternate = 0;

 Path* path;
 int no_of_paths;

 Pair* pair;
 int no_of_pairs;
 
 double	theores;
 int theoiter;

 time_t time1, time2;
 double utime;

 if(!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }
 fprintf(out_fp, "# Call Blocking Probabilities\n");

 if( UCall_present )
   {
    fprintf(out_fp, "\n# The statistical analysis (confidence interval ");
    fprintf(out_fp, "estimation and initial \n# transient detection) ");
    fprintf(out_fp, "refers exclusively to non-U class users.\n");
    fprintf(out_fp, "# The statistical analysis is not applied to data ");
    fprintf(out_fp, "regarding Class U users.\n");
   }

 if (global_stat_vars[1].info != NOT_ANALYSED)
   {
     report_statistics(out_fp, global_stat_vars[1].number, TRUE);
     if(transient_is_over(global_stat_vars[1].number))
       report_statistics(out_fp, global_stat_vars[1].number, FALSE);
   }

 if(theo_analysis)
 {
	 maxM = 0;
	 j = 8*MAX_WVOCTETS;
	 for(i = 0; i < no_of_links; i++)
	 {
		 if(link_vect[i]->is_node_node &&
			 link_vect[i]->max_waves < j)
		 {
			 j = link_vect[i]->max_waves;
			 M = link_vect[i]->fiber_num;
		 }
		 if(link_vect[i]->is_node_node &&
			 link_vect[i]->fiber_num > maxM)
		 {
			 maxM = link_vect[i]->fiber_num;
		 }
	 } 
	 /* Theoretical results on the blocking */
	 printf("\nTheoretical calculation: Start --- ");
	 createbinom(j*maxM);
	 createfacto(j*maxM);
	 
	 fprintf(out_fp, "\n#Theoretical Blocking probabilties:\n");
	 
/*	 time(&time1);
	 utime = get_usage_utime();
	 fprintf(out_fp, "\n\n# MultiIndep Complex First = %g\n", 
		 theo_multiindep_complex(j, &theores, &theoiter));
	 time(&time2);
	 fprintf(out_fp, "# MultiIndep Complex = %g\n", theores);
	 fprintf(out_fp, "# Iteration in the MultiIndep Complex function = %d\n", theoiter);
	 fprintf(out_fp, "# Time of calculation the MultiIndep Complex function = %g\n",
/*		 difftime(time2, time1)*
	 get_usage_utime() - utime);
	 
	 printf("WIMMstop - ");
	 
	 time(&time1);
	 utime = get_usage_utime();
	 fprintf(out_fp, "\n\n# MultiMost Complex First = %g\n", 
		 theo_multimost_complex(j, &theores, &theoiter));
	 time(&time2);
	 fprintf(out_fp, "# MultiMost Complex = %g\n", theores);
	 fprintf(out_fp, "# Iteration in the MultiMost Complex function = %d\n", theoiter);
	 fprintf(out_fp, "# Time of calculation the MultiMost Complex function = %g\n",
/*		 difftime(time2, time1)*
	 get_usage_utime() - utime);
	 
	 
	 printf("WDMMStop - ");
	 
*/	 time(&time1);/*
	 utime = get_usage_utime();*/
	 fprintf(out_fp, "\n\n# BPP Complex First = %g\n", 
		 theo_bpp_complex(j, &theores, &theoiter));
	 time(&time2);
	 fprintf(out_fp, "# BPP Complex = %g\n", theores);
	 fprintf(out_fp, "# Iteration in the BPP Complex function = %d\n", theoiter);
	 fprintf(out_fp, "# Time of calculation the BPP Complex function = %g\n",
		 difftime(time2, time1)/*
	 get_usage_utime() - utime */);
	 
	 
	 printf("BppStop - ");
/*	 time(&time1);
	 utime = get_usage_utime();
	 fprintf(out_fp, "\n\n# Li Somani MLLC = %g\n", 
		 theo_lisomani_complex(j));
	 time(&time2);
	 fprintf(out_fp, "# Time of calculation the Li Somani function = %g\n",
/*		 difftime(time2, time1)*
	 get_usage_utime() - utime);
	 printf("Lisomani Stop\n");*/
 }

/* Print the total user call blocking statistic */
 for(i = 0; i < no_of_users; i++)
    if(user_stat_vars[i].info != NOT_ANALYSED)
       {
        report_statistics(out_fp, user_stat_vars[i].number, TRUE);
        if(transient_is_over(user_stat_vars[i].number))
           report_statistics(out_fp, user_stat_vars[i].number, FALSE);
       }
 fprintf(out_fp, "\n#  User Id\tReq Calls\t");
 fprintf(out_fp, "Ref Calls\tPrBE Calls\tPrim Calls\tAlt Calls\t");
 fprintf(out_fp, "Block Prob\tUser Name\n\n");

 for(i = 0; i < no_of_users; i++)
     {
     /* sum all the requested call and the refused call to get the total */
     /* blocking probability for each user */
     sreq_calls = 0;
     sref_calls = 0;
     spbe_calls = 0;
     for(j = 0; j < no_of_users; j++)
       {
       sreq_calls += req_calls[i][j];
       sref_calls += ref_calls[i][j];
       spbe_calls += pbe_calls[i][j];
       }
      if(sreq_calls > 0)
             fprintf(out_fp,
   "   %2d\t%-7ld\t%-7ld\t%-7ld\t%-7ld\t%-7ld\t%.2E\t# %s\n",
                   i, sreq_calls, sref_calls, spbe_calls, primary_calls[i],
                   alternate_calls[i],
                   ((double)sref_calls) / sreq_calls, user_name(i));
      else
             fprintf(out_fp,
   "   %2d\t%-7ld\t%-7ld\t%-7ld\t%-7ld\t%-7ld\t#not def\t# %s\n",
                   i, sreq_calls, sref_calls, spbe_calls, primary_calls[i],
                   alternate_calls[i], user_name(i));

/*          if( !UCall_present )
             {
              acc_req_calls += sreq_calls;
              acc_ref_calls += sref_calls;
              acc_primary   += primary_calls[i];
              acc_alternate += alternate_calls[i];
             }
*/        }

/*    if ( UCall_present ) */
/* DEBUG */

/* calculate the non class-U blocking probabilities, the class-U call */
/* blocking probabilities */
      for(i = 0; i < no_of_cgens; i++)
        {
         /* sum all the requested call and the refused call to get the */
         /* total blocking probability for each user */
	 sreq_calls = 0;
	 sref_calls = 0;
	 spbe_calls = 0;
	for(j = 0; j < no_of_users; j++)
          {
	  sreq_calls += req_calls[callgen_vect[i]->ref_user->id][j];
	  sref_calls += ref_calls[callgen_vect[i]->ref_user->id][j];
	  spbe_calls += pbe_calls[callgen_vect[i]->ref_user->id][j];
	  }
         if(callgen_vect[i]->type == CG_TYPE5)
             {
	      acc_req_Ucalls += sreq_calls;
              acc_ref_Ucalls += sref_calls;
              acc_Uprimary +=
                     primary_calls[callgen_vect[i]->ref_user->id];
              acc_Ualternate +=
                     alternate_calls[callgen_vect[i]->ref_user->id];
             }
         else
             {
              acc_req_calls += sreq_calls;
              acc_ref_calls += sref_calls;
              acc_pbe_calls += spbe_calls;
              acc_primary +=
                     primary_calls[callgen_vect[i]->ref_user->id];
              acc_alternate +=
                     alternate_calls[callgen_vect[i]->ref_user->id];
             }
        }

/* check if the transient is over */
 if (global_stat_vars[1].info != NOT_ANALYSED)
   if(!transient_is_over(global_stat_vars[1].number))
       {
        fprintf(out_fp,
            "\nWarning:  The initial transient is not over\n");
        fprintf(out_fp, "          for %s\n",
            names_stat_vars[global_stat_vars[1].number]);
       }
 for(i = 0; i < no_of_users; i++)
       {
        if((user_stat_vars[i].info != NOT_ANALYSED) &&
          (!transient_is_over(user_stat_vars[i].number)))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for %s\n",
                   names_stat_vars[user_stat_vars[i].number]);
           }
        else if((user_stat_vars[i].info == NOT_ANALYSED) &&
                (user_start_time[i] == 0.0))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for call blocking probability ");
            fprintf(out_fp, "for user %s\n",
                   user_name(i));
            fprintf(out_fp, "          (not statistically analysed)\n");
           }
       }

 for(i = 0; i < no_of_users; i++)
   {
     if((classU_user_stat_vars[i].info != NOT_ANALYSED) &&
	(!transient_is_over(classU_user_stat_vars[i].number)))
       {
	 fprintf(out_fp,
		 "\nWarning:  The initial transient is not over\n");
	 fprintf(out_fp, "          for %s\n",
		 names_stat_vars[classU_user_stat_vars[i].number]);
       }
     else if((classU_user_stat_vars[i].info == NOT_ANALYSED) &&
	     (classU_user_start_time[i] == 0.0))
       {
	 fprintf(out_fp,
		 "\nWarning:  The initial transient is not over\n");
	 fprintf(out_fp, "          for call blocking probability ");
	 fprintf(out_fp, "for user %s\n",
		 user_name(i));
	 fprintf(out_fp, "          (not statistically analysed)\n");
       }
   }

/* print the TOTAL call blocking brobabilities */
 fprintf(out_fp, "\n# Total Mean Block Probability       : %.2E",
             ((double)(acc_ref_calls + acc_ref_Ucalls)) /
                      (acc_req_calls + acc_req_Ucalls));
 fprintf(out_fp, "\n# Total Requested Calls              : %ld",
             acc_req_calls + acc_req_Ucalls);
 fprintf(out_fp, "\n# Total Refused Calls                : %ld",
             acc_ref_calls + acc_ref_Ucalls);
 fprintf(out_fp, "\n# Total Protection-Best Effort Calls : %ld",
             acc_pbe_calls);
 fprintf(out_fp, "\n# Total Accepted Primary Calls       : %ld",
             acc_primary + acc_Uprimary);
 fprintf(out_fp, "\n# Total Accepted Alternate Calls     : %ld",
             acc_alternate + acc_Ualternate);
 fprintf(out_fp, "\n");

/* print the class-U call blocking probabilities, if there are any Ucall */
/* generator */
 if( UCall_present )
     {
      fprintf(out_fp, "\n# non-U Class Mean Block Probability   : %.2E",
             ((double)acc_ref_calls) / acc_req_calls);
      fprintf(out_fp, "\n# non-U Class Requested Calls          : %ld",
             acc_req_calls);
      fprintf(out_fp, "\n# non-U Class Refused Calls            : %ld",
             acc_ref_calls);
	  fprintf(out_fp, "\n# non-U Protection-Best Effort Calls   : %ld",
             acc_pbe_calls);
      fprintf(out_fp, "\n# non-U Class Accepted Primary Calls   : %ld",
             acc_primary);
      fprintf(out_fp, "\n# non-U Class Accepted Alternate Calls : %ld",
             acc_alternate);
      fprintf(out_fp, "\n");

      fprintf(out_fp, "\n# Class U Mean Block Probability     : %.2E",
             ((double)acc_ref_Ucalls) / acc_req_Ucalls);
      fprintf(out_fp, "\n# Class U Requested Calls            : %ld",
             acc_req_Ucalls);
      fprintf(out_fp, "\n# Class U Refused Calls              : %ld",
             acc_ref_Ucalls);
      fprintf(out_fp, "\n# Class U Accepted Primary Calls     : %ld",
             acc_Uprimary);
      fprintf(out_fp, "\n# Class U Accepted Alternate Calls   : %ld",
             acc_Ualternate);
      fprintf(out_fp, "\n");
     }

/* Report the unreachable users */
 fprintf(out_fp, "\n# Unreachable Users :");
 for(j = 0; j < no_of_users; j++)
 {
	 if(node_vect[user_vect[j]->connected_node]->unreachable)
	 fprintf(out_fp, "\t%s",user_name(j));
 }
 fprintf(out_fp, "\n");


/* Report the substituted link */
 fprintf(out_fp, "\n# Substituted Link :");
	 fprintf(out_fp, "\t%s\n",link_name(subslinkid));

#define CBPMATRIX
#ifdef CBPMATRIX


/*#define SEPARMATR*/

/* Print the user user call blocking probabilities */
/* requested calls first */
 fprintf(out_fp, "\n\n\n#REQUESTED CALLS\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
 int jjj;
 for(j = 0; j < no_of_users; j+=jj)
 {
     fprintf(out_fp,"# src \t");
 /* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
	     if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[j+jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
					 fprintf(out_fp, "%ld\t", req_calls[i][j + jj]);
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
     fprintf(out_fp, "\n");
 }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
				 fprintf(out_fp, "%ld\t", req_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif

/* refused calls  are printed here */
  fprintf(out_fp, "\n\n\n#REFUSED CALLS\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && (j + jj < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[j+jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
					 fprintf(out_fp, "%ld\t", ref_calls[i][j + jj]);
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
     fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
				 fprintf(out_fp, "%ld\t", ref_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif

/* call blocking probability are printed here */
  fprintf(out_fp, "\n\n\n#CALL BLOCKING PROBABILITY\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == 0)
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.1E\t",
						 ((float)ref_calls[i][j+jj])/req_calls[i][j+jj]);
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == 0)
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.1E\t", 
					 ((float)ref_calls[i][jj])/req_calls[i][jj]);
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif

/* end of the user user call blocing probability report */
#endif

  
/* protection - best effort calls are reported here */
 fprintf(out_fp, "\n\n\n#PROTECTION - BEST EFFORT CALLS\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
 for(j = 0; j < no_of_users; j+=jj)
 {
     fprintf(out_fp,"# src \t");
 /* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
	     if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[j+jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
					 fprintf(out_fp, "%ld\t", pbe_calls[i][j + jj]);
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
     fprintf(out_fp, "\n");
 }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0\t");
				   else
				 fprintf(out_fp, "%ld\t", pbe_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif


/* mean hop number of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN PATH HOP NUMBER\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 ((double)path_hops_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
!						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 ((double)path_hops_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif

/* mean length of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN PATH LENGTH\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 (path_length_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }


#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 (path_length_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif
  
/* mean crossed transit opaque node number of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN NUMBER OF CROSSED TRANSIT OPAQUE NODES\
	  \t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 ((double)path_crop_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 ((double)path_crop_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif


/* Statistics of the protective paths of the calls */

/* mean hop number of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN PROTECTIVE PATH HOP NUMBER\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 ((double)prot_path_hops_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 ((double)prot_path_hops_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif

/* mean length of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN PROTECTIVE PATH LENGTH\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 (prot_path_length_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }


#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 (prot_path_length_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif
  
/* mean crossed transit opaque node number of paths are printed here */
  fprintf(out_fp, "\n\n\n#MEAN NUMBER OF CROSSED TRANSIT OPAQUE NODES IN PROTECTION\
	  \t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_users; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_users; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_users); jj++)
				 if(ctrl_traff_pattern(j + jj))
				 {
					 if(req_calls[i][j+jj] == ref_calls[i][j+jj])
						 fprintf(out_fp, "Not def\t");
					 else
					 {
	                   if(node_vect[user_vect[i]->connected_node]->unreachable ||
					       node_vect[user_vect[j+jj]->connected_node]->unreachable)
					     fprintf(out_fp, "Not imp\t");
				       else
						 fprintf(out_fp, "%.3f\t",
						 ((double)prot_path_crop_accu[i][j+jj])/
						   (req_calls[i][j+jj]-ref_calls[i][j+jj]));
					 }
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
						 fprintf(out_fp, "-1.00\t");
				 else
				 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-2.00\t");
				   else
					 fprintf(out_fp, "%.3f\t",
					 ((double)prot_path_crop_accu[i][jj])/
					  (req_calls[i][jj]-ref_calls[i][jj]));
				 }
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
#endif


#define CLASSDIST
#ifdef CLASSDIST
/* generate a report by a distance aggregate call generators */
{
 long req_calls_dist[MAX_NODE];
 long ref_calls_dist[MAX_NODE];
 double cbp_dist[MAX_NODE];
 int     dist;
 int     max_dist;
 /* vectors initialization */
  for(i = 0; i < MAX_NODE; i++)
	  {
	  req_calls_dist[i] = 0;
	  ref_calls_dist[i] = 0;
	  cbp_dist[i] = -1.0;
	  }

 /* scanning call generator to call generator */
  max_dist = -1;
  for(i = 0; i < no_of_users; i++)
      for(j = 0; j < no_of_users; j++)
	  {
          /* update the stats if there is a path between i and j */
		  if(hopdist[i][j] >= 0)
		  {
			  dist =  hopdist[i][j];
			  if(dist > max_dist)
				  max_dist = dist;
			  req_calls_dist[dist] += req_calls[i][j];
			  ref_calls_dist[dist] += ref_calls[i][j];
		  }
	  }

 /* computation of call blocking probability according to distance i */
  for(i = 0; i <= max_dist; i++)
      if(req_calls_dist[i] > 0)
		  cbp_dist[i] =
		  ((double)ref_calls_dist[i])/(double)(req_calls_dist[i]);
	  else
		  cbp_dist[i] = -1.0;

 /* printing output */
  fprintf(out_fp, "\n\n\n#CALL BLOCKING PROBABILITIES ORDERED ");
  fprintf(out_fp, "BY USERS HOP DISTANCE\n\n");
  fprintf(out_fp, "#distance\tblocking probability\n");
  for(i = 0; i <= max_dist; i++)
      {
      fprintf(out_fp, "%d\t\t", i);
      if(cbp_dist[i] < 0.0)
	  fprintf(out_fp, "not defined\n");
      else
	  fprintf(out_fp, "%E\n", cbp_dist[i]);
      }
}
#endif
/* end of the distance class call blocking probabilities report */

 /* printing th names of the nodes listed by id */
  fprintf(out_fp, "\n\n\n#NODE NAMES\n\n");
  fprintf(out_fp, "#\tid\tname\n");
  
  for(i = 0; i < no_of_nodes; i++)
  {
	  fprintf(out_fp, "\t%d\t%s\n", i, node_name(i));
  }

/*}

*
**  Function  : void generate_PAT_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the PAT report.
*
void generate_PAT_report(FILE * out_fp)
{
 register int   i, j, jj, jjj;

 Path* path;
 int no_of_paths;
*/
  
/* number of paths between source-destination pairs */
  fprintf(out_fp, "\n\n\n#NUMBER OF AVAILABLE PATHS\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_cgens; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_cgens); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_cgens; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_cgens); jj++)
				 if(ctrl_traff_pattern(callgen_vect[j+jj]->ref_user->id))
				 {
					 path = path_mat[callgen_vect[i]->
					  choosen_routing->path_order_type][i][j+jj];
					 no_of_paths = 0;
					 while (path)
					 {
						 path = path->next;
						 no_of_paths ++;
					 }
					 
					 fprintf(out_fp, "%d\t", no_of_paths);
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_cgens; jj++)
	 if(ctrl_traff_pattern(callgen_vect[jj]->ref_user->id))
		 fprintf(out_fp, "%s\t", user_name(callgen_vect[jj]->ref_user->id));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_cgens; i++)
     {
	  if(ctrl_traff_start(callgen_vect[i]->ref_user->id))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_cgens; jj++)
		  {
			  if(ctrl_traff_pattern(callgen_vect[jj]->ref_user->id))
			  {
				  path = path_mat[callgen_vect[i]->
					  choosen_routing.path_order_type][i][jj];
				  no_of_paths = 0;
				  while (path)
				  {
					  path = path->next;
					  no_of_paths ++;
				  }
				  
				  fprintf(out_fp, "%d\t", no_of_paths);
			  }
		  }
		  fprintf(out_fp, "# %s\n", user_name(callgen_vect[i]->ref_user->id));
	  }
	 }
#endif

/* number of pairs between source-destination pairs */
  fprintf(out_fp, "\n\n\n#NUMBER OF AVAILABLE PAIRS\t\tFrom-To Matrix\n\n");

#ifdef SEPARMATR
  for(j = 0; j < no_of_cgens; j+=jj)
     {
     fprintf(out_fp,"#dest\t\t");
/* this is done to use only "ncol" coloumns in the report */
     for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_cgens); jj++)
             if(ctrl_traff_pattern(j + jj))
		 {
		 fprintf(out_fp, "%d\t", j + jj);
		 jjj++;
		 }
     fprintf(out_fp, "\n");
     fprintf(out_fp, "#src\t\t");
     for(i = 0; i < 63; i++)
	 fprintf(out_fp, "-");
     fprintf(out_fp,"\n");
     for(i = 0; i < no_of_cgens; i++)
	 {
		 if(ctrl_traff_start(i))
		 {
			 fprintf(out_fp, "%d\t|\t", i);
			 for(jj = 0, jjj = 0; (jjj < ncol) && ((j + jj) < no_of_cgens); jj++)
				 if(ctrl_traff_pattern(callgen_vect[j+jj]->ref_user->id))
				 {
					 if(pair_mat)
						 pair = pair_mat[callgen_vect[i]->
						 choosen_routing.path_order_type][i][j+jj];
					 else
						 pair = NULL;
					 no_of_pairs = 0;
					 while (pair)
					 {
						 pair = pair->next;
						 no_of_pairs ++;
					 }
					 
					 fprintf(out_fp, "%d\t", no_of_pairs);
					 jjj++;
				 }
				 fprintf(out_fp, "\n");
		 }
	 }
	 fprintf(out_fp, "\n");
     }

#else
	 
  fprintf(out_fp,"#");
  for(jj = 0; jj < no_of_cgens; jj++)
	 if(ctrl_traff_pattern(callgen_vect[jj]->ref_user->id))
		 fprintf(out_fp, "%s\t", user_name(callgen_vect[jj]->ref_user->id));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_cgens; i++)
     {
	  if(ctrl_traff_start(callgen_vect[i]->ref_user->id))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_cgens; jj++)
		  {
			  if(ctrl_traff_pattern(callgen_vect[jj]->ref_user->id))
			  {
				  if(pair_mat)
					  pair = pair_mat[callgen_vect[i]->
					  choosen_routing.path_order_type][i][jj];
				  else
					  pair = NULL;
				  no_of_pairs = 0;
				  while (pair)
				  {
					  pair = pair->next;
					  no_of_pairs ++;
				  }
				  
				  fprintf(out_fp, "%d\t", no_of_pairs);
			  }
		  }
		  fprintf(out_fp, "# %s\n", user_name(callgen_vect[i]->ref_user->id));
	  }
	 }
#endif
}

/*
**  Function  : void generate_CDL_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the CDL report->
**				Calls Duration and Load
**				This file contains the statistics on the
**				number of checked links during the routing
*/
void generate_CDL_report(FILE * out_fp)
{
  register int   i, j, jj;
  double	sats[MAX_NET_USER];
  CallGen	*cgen_ptr;
  Call		*call_ptr;

  fprintf(out_fp,"# Duration of calls / Total time   and traffic relation\n");
  fprintf(out_fp,"# Number of checked links durign the routing.\n");

  for(call_ptr = head_of_call_list->next; call_ptr != 
	  head_of_call_list; call_ptr = call_ptr->next)
  {
	  inon_accumulator[callgen_vect[call_ptr->source_id]->ref_user->id]
		  [callgen_vect[call_ptr->dest_id]->ref_user->id] += 
		  ancles_time - call_ptr->start_time;
  }
  
  
  fprintf(out_fp,"\n#DURATION/TIME\n\n#");
  for(i = 0; i < no_of_users; i++)
  {
	  /* reset saturation vector */
	  sats[i] = 0.0;
	  
	  if(ctrl_traff_pattern(i))
		  fprintf(out_fp, "%s\t", user_name(i));
  }
  fprintf(out_fp,"# From\tOO-Sat\n");

  for(j = 0; j < no_of_cgens; j++)
  {
	  cgen_ptr = callgen_vect[j];
	  if(cgen_ptr->type == CG_TYPE7)
		  sats[cgen_ptr->ref_user->id] =
		  ((CGType7 *)cgen_ptr->info)->saturation;
  }

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
	               if(node_vect[user_vect[i]->connected_node]->unreachable ||
					   node_vect[user_vect[jj]->connected_node]->unreachable)
					 fprintf(out_fp, "-0.0\t");
				   else
				 fprintf(out_fp, "%f\t", inon_accumulator[i][jj]/(ancles_time - 
				 user_start_time[i]));
			 }
		  }
		  fprintf(out_fp, "# %s\t%f\n", user_name(i), sats[i]);
	  }
	 }

  fprintf(out_fp,"\n#TRAFFIC RELATION\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 fprintf(out_fp, "%f\t",traff_pattern[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  
  fprintf(out_fp,"\n\n\n#MEAN NUM OF CHECKED LINKS -> ACCEPTED CALLS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(req_calls[i][jj] == ref_calls[i][jj])
					 fprintf(out_fp, "-1.00\t");
				 else
					 fprintf(out_fp, "%f\t", (double)acc_noch_link_accu[i][jj] / 
					 ((double) req_calls[i][jj] - (double) ref_calls[i][jj]));
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#MEAN NUM OF CHECKED LINKS -> REFUSED CALLS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(ctrl_traff_pattern(jj))
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(ctrl_traff_start(i))
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(ctrl_traff_pattern(jj))
			 {
				 if(!ref_calls[i][jj])
					 fprintf(out_fp, "-1.00\t");
				 else
					 fprintf(out_fp, "%f\t", (double)ref_noch_link_accu[i][jj] / 
					 (double) ref_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
}

/*
**  Function  : void generate_CL_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the CL report.
*/
void generate_CL_report(FILE * out_fp)
{
 int   i, j, tot;
 int    	ncol = 8;
 double         link_load;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# Carried Loads\n");

 if( UCall_present )
     {
      fprintf(out_fp, "\n# The statistical analysis (confidence interval ");
      fprintf(out_fp, "estimation and initial \n# transient detection) ");
      fprintf(out_fp, "refers exclusively to the load deriving ");
      fprintf(out_fp, "from calls\n# generated by non-U class users.\n");
      fprintf(out_fp, "# The statistical analysis is not applied to data ");
      fprintf(out_fp, "regarding calls generated \n# by Class U users.\n");
      fprintf(out_fp, "\n# The Mean Loads carried by the links include ");
      fprintf(out_fp, "contributions \n# coming only from calls of ");
      fprintf(out_fp, "non-U Class: contributions from calls of Class U ");
      fprintf(out_fp, "\n# are reported in the file %s.ucl\n", file_radix);
     }

 if (global_stat_vars[0].info != NOT_ANALYSED)
   {
     report_statistics(out_fp, global_stat_vars[0].number, TRUE);
     if(transient_is_over(global_stat_vars[0].number))
       report_statistics(out_fp, global_stat_vars[0].number, FALSE);
   }

 for(i = 0; i < no_of_links; i++)
    if(link_stat_vars[i].info != NOT_ANALYSED)
       {
        report_statistics(out_fp, link_stat_vars[i].number, TRUE);
        if(transient_is_over(link_stat_vars[i].number))
           report_statistics(out_fp, link_stat_vars[i].number, FALSE);
       }

 fprintf(out_fp, "\n#Link Id\tCarried Load\tPrimary Load\t");
 fprintf(out_fp, "Link Name\tCapacity\n\n");
 for(i = 0; i < no_of_links; i++)
    {
      double primary_load;

     link_load = load_accumulator[i][TOTAL_LINK_LOAD] /
	                               (ancles_time - link_start_time[i]);

     primary_load = load_accumulator[i][PRIMARY_LINK_LOAD] /
	                               (ancles_time - link_start_time[i]);

     fprintf(out_fp, "\t%d\t%f\t%f\t# %s\t%.1f\n",
 i, link_load, primary_load, link_name(i), capacity/link_vect[i]->timescale);
    }

 fprintf(out_fp, "\n\n# Histogram of the distribution of call over the path");

 fprintf(out_fp, "\n#path number\t:\t routed calls \t percentage");
 tot = 0;
 j = 0;
 for(i=0; i< max_cgen_cgen_no_of_path; i++)
 {
    tot += no_times_this_path_is_utilized[i];
    if (no_times_this_path_is_utilized[i] != 0)
       j = i;
 }

 for(i=0; i<= j; i++)
    fprintf(out_fp, "\n\t%3d\t:\t %8d\t%8.4f",i+1, no_times_this_path_is_utilized[i],
                    (double)no_times_this_path_is_utilized[i]/(double)tot*100.0);


/* Now printing the user contribution on the link load */
fprintf(out_fp, "\n# The following table report the link carried  ");
fprintf(out_fp, "\n# load, divided for each call generator. ");

 for(j=0; j < no_of_users; j += ncol)
   {
    int jj;
    fprintf(out_fp, "\n\n #gen\t");
    for( jj=j ; (jj < no_of_users) && ( (jj - j) < ncol) ; jj++)
       fprintf(out_fp, "\t%d",jj);
    fprintf(out_fp, "\n # link\t|");

    for(i = 0; i < no_of_links; i++)
      {
       fprintf(out_fp, "\n   %d\t|",i);
       for(jj = j; (jj < no_of_users) && ( (jj - j) < ncol); jj++)
	 {
          link_load = load_accumulator[i][jj] /
	                                  (ancles_time - link_start_time[i]);
          fprintf(out_fp, "\t%.3f", link_load);
         }
      }
    fprintf(out_fp, "\n");
  }

 if (global_stat_vars[0].info != NOT_ANALYSED)
   if(!transient_is_over(global_stat_vars[0].number))
         {
           fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
           fprintf(out_fp, "          for %s\n",
               names_stat_vars[global_stat_vars[0].number]);
         }
 for(i = 0; i < no_of_links; i++)
       {
        if((link_stat_vars[i].info != NOT_ANALYSED) &&
          (!transient_is_over(link_stat_vars[i].number)))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for %s\n",
                   names_stat_vars[link_stat_vars[i].number]);
           }
        else if((link_stat_vars[i].info == NOT_ANALYSED) &&
                (link_start_time[i] == 0.0))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for carried load for link ");
            fprintf(out_fp, "%s\n",
                   link_name(i));
            fprintf(out_fp, "          (not statistically analysed)\n");
           }
       }
}

/*
**  Function  : void generate_AWL_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the AWL report.
*/
void generate_AWL_report(FILE * out_fp)
{
 int   i;
 int    	ncol = 8;
 double     used;
 double     peak;
 double     util;
 double		delta_t;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# ASON: Used Wavelengths\n");

 if (global_stat_vars[2].info != NOT_ANALYSED)
   {
     report_statistics(out_fp, global_stat_vars[2].number, TRUE);
     if(transient_is_over(global_stat_vars[2].number))
       report_statistics(out_fp, global_stat_vars[2].number, FALSE);
   }

 for(i = 0; i < no_of_links; i++)
    if(ason_link_stat_vars[i].info != NOT_ANALYSED)
       {
        report_statistics(out_fp, ason_link_stat_vars[i].number, TRUE);
        if(transient_is_over(ason_link_stat_vars[i].number))
           report_statistics(out_fp, ason_link_stat_vars[i].number, FALSE);
       }

 fprintf(out_fp, "\n#Link Id\tUsed Waves\t");
 fprintf(out_fp, "Peak Wave\t");
 fprintf(out_fp, "Utilization\t");
 fprintf(out_fp, "Link Name\tMax Waves\tFromId\tToId\tFrom\tTo\tReachable\tFibers\n\n");
 for(i = 0; i < no_of_links; i++)
    {
/* In this version print not only the node-node	links! */
	 if(link_vect[i]->is_node_node)
	 {
		 delta_t = ancles_time - 
			 prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD];

/* update the statistic */
		 ason_wave_accumulator[i][TOTAL_ASON_LINK_LOAD] +=
			 ason_link_carried_waves[i][TOTAL_ASON_LINK_LOAD] * delta_t;

		 ason_peak_accumulator[i][TOTAL_ASON_LINK_LOAD] +=
			 ason_link_peak[i][TOTAL_ASON_LINK_LOAD] * delta_t;

/* Don't care about the zero utilization periods. */
		 if(ason_link_util[i][TOTAL_ASON_LINK_LOAD] > 0.0)
			 ason_util_accumulator[i][TOTAL_ASON_LINK_LOAD] +=
			 ason_link_util[i][TOTAL_ASON_LINK_LOAD] * delta_t;
		 else
			 ason_util_zero_time[i] += delta_t;


		 used = ason_wave_accumulator[i][TOTAL_ASON_LINK_LOAD] /
	                               (ancles_time - ason_link_start_time[i]);

/*     prim_used = ason_wave_accumulator[i][PRIMARY_ASON_LINK_LOAD] /
	                               (ancles_time - ason_link_start_time[i]);*/

		 peak = ason_peak_accumulator[i][TOTAL_ASON_LINK_LOAD] /
	                               (ancles_time - ason_link_start_time[i]);

/*     prim_peak = ason_peak_accumulator[i][PRIMARY_ASON_LINK_LOAD] /
	                               (ancles_time - ason_link_start_time[i]);*/

/*	 if(link_vect[i]->is_node_node) */
		 if(ancles_time - ason_link_start_time[i] - ason_util_zero_time[i])
			 util = ason_util_accumulator[i][TOTAL_ASON_LINK_LOAD] /
			 (ancles_time - ason_link_start_time[i] - ason_util_zero_time[i]);
		 else
			 util = -1.0;
/*	 else
		 util = 0.0;*/

/*     prim_util = ason_util_accumulator[i][PRIMARY_ASON_LINK_LOAD] /
	      (ancles_time - ason_link_start_time[i] - ason_util_zero_time[i]);*/

/*     if(util) *print only the used node-node links */
		 fprintf(out_fp, "\t%d\t%f\t%f\t%f\t# %s\t%d\t%d\t%d\t%s\t%s",
		 i, used, peak, util, link_name(i), 
		 link_vect[i]->max_waves, link_vect[i]->source,
		 link_vect[i]->dest, node_name(link_vect[i]->source),
		 node_name(link_vect[i]->dest));

	/* print the reachability information about the endpoints of link */	 
		 if(node_vect[link_vect[i]->source]->unreachable ||
			 node_vect[link_vect[i]->dest]->unreachable)
			 fprintf(out_fp, "\t-");
		 else
			 fprintf(out_fp, "\t+");

		 fprintf(out_fp, "\t%d\n", link_vect[i]->fiber_num);
	 }
    }

 if (global_stat_vars[2].info != NOT_ANALYSED)
   if(!transient_is_over(global_stat_vars[2].number))
         {
           fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
           fprintf(out_fp, "          for %s\n",
               names_stat_vars[global_stat_vars[2].number]);
         }
 for(i = 0; i < no_of_links; i++)
       {
        if((ason_link_stat_vars[i].info != NOT_ANALYSED) &&
          (!transient_is_over(ason_link_stat_vars[i].number)))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for %s\n",
                   names_stat_vars[ason_link_stat_vars[i].number]);
           }
        else if((ason_link_stat_vars[i].info == NOT_ANALYSED) &&
                (ason_link_start_time[i] == 0.0))
           {
            fprintf(out_fp,
              "\nWarning:  The initial transient is not over\n");
            fprintf(out_fp, "          for ason wavelengths for link ");
            fprintf(out_fp, "%s\n",
                   link_name(i));
            fprintf(out_fp, "          (not statistically analysed)\n");
           }
       }
}

/*
**  Function  : void generate_AWH_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the AWH report -> 
**				histogram about link ason loads and peaks
*/
void generate_AWH_report(FILE * out_fp)
{
 int   i, j;
 double	delta_t;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# ASON: Histogram of Used and Peak Wavelengths on the Links\n");

 fprintf(out_fp, "\n#Used Wavelengths\n");
 fprintf(out_fp, "\n#Link Id\tLink Name\tCapacity\tLength");
 for(j=0; j < floor(max_link_capa/wavelength_capacity) + 1; j++)
	 fprintf(out_fp, "\tu%d", j);

 for(i = 0; i < no_of_links; i++)
    {

/* In this version print only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
	 delta_t = ancles_time - prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD];

	 /* update the histograms */
	 histo_ason_wave_accu[i]
		 [(int)(ason_link_carried_waves[i][TOTAL_ASON_LINK_LOAD])] +=
		 delta_t;
	  
	 fprintf(out_fp, "\n\t%d\t%s\t%d\t%.2f",
		 i,link_name(i), link_vect[i]->max_waves *
		 link_vect[i]->fiber_num,
		 link_vect[i]->link_len);
 
	 for(j=0; j < (link_vect[i]->max_waves *
		 link_vect[i]->fiber_num)+1; j++)
		 fprintf(out_fp, "\t%8f", histo_ason_wave_accu[i][j]/
		 (ancles_time-ason_link_start_time[i]));
	 }     

    }

 fprintf(out_fp, "\n\n#Peak Wavelength\n");
 fprintf(out_fp, "\n#Link Id\tLink Name\tMax Waves\tLength");
 for(j=0; j < 8*max_wvoct_all_links + 1; j++)
	 fprintf(out_fp, "\tp%d", j);

 for(i = 0; i < no_of_links; i++)
    {
/* In this version print only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
	 delta_t = ancles_time - prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD];
/* update the histogram */
	  histo_ason_peak_accu[i]
		  [(int)(ason_link_peak[i][TOTAL_ASON_LINK_LOAD])] +=
		  delta_t;
	 fprintf(out_fp, "\n\t%d\t%s\t%d\t%.2f",
		 i,link_name(i), link_vect[i]->max_waves,
		 link_vect[i]->link_len);
 
	 for(j=0; j < link_vect[i]->max_waves+1; j++)
		 fprintf(out_fp, "\t%8f", histo_ason_peak_accu[i][j]/
		 (ancles_time-ason_link_start_time[i]));
	 }     

    }
 fprintf(out_fp, "\n");

}

/*
**  Function  : void generate_ABW_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the ABW report -> 
**				occupacy of wavelengths on the links at
**				call blocking events
*/
void generate_ABW_report(FILE * out_fp)
{
 int   i, j;
 long	all_ref = 0;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# ASON: Occupacy of Wavelengths on the Links at");
 fprintf(out_fp, " Call Blocking\n");

 for(i = 0; i < no_of_users; i++)
	 for(j = 0; j < no_of_users; j++)
		 all_ref += ref_calls[i][j];

 if(!all_ref)
 {
	 fprintf(out_fp, "NO BLOCKING OCCURRED!\n");
	 return;
 }

 fprintf(out_fp, "\n#Link Id\tLink Name\tMax Waves\tLength");
 for(j=1; j < 8*max_wvoct_all_links + 1; j++)
	 fprintf(out_fp, "\tw%d", j);
 fprintf(out_fp, "\n");
 
 for(i = 0; i < no_of_links; i++)
    {
/* In this version print not only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
	 fprintf(out_fp, "\t%d\t%s\t%d\t%.2f",
		 i,link_name(i), link_vect[i]->max_waves,
		 link_vect[i]->link_len);
 
	 for(j = 1; j < link_vect[i]->max_waves + 1; j++)
		 fprintf(out_fp, "\t%8f", (double)inblock_asonwaves_accu[i][j]/
		 (double)all_ref);

/*	 for(j = link_vect[i]->max_waves + 1; j < 8*MAX_WVOCTETS + 1; j++)
		 fprintf(out_fp, "\t-1.0");*/

	 fprintf(out_fp, "\n");
	 }     

    }
}

/*
**  Function  : void generate_LWH_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the LWH report -> 
**				histogram about link ason occupancy of waves
*/
void generate_LWH_report(FILE * out_fp)
{
 int   i, j;
 double delta_t;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# ASON: Mean and Maximum Number of Used Fibers\
	 per Wavelength on the Links\n");

 fprintf(out_fp, "\n#Mean Used Fiber\n");
 fprintf(out_fp, "\n#Link Id\tLink Name\tMax Waves\tLength");
 for(j=1; j < 8*max_wvoct_all_links + 1; j++)
	 fprintf(out_fp, "\tw%d", j);
 fprintf(out_fp, "\n");
 
 for(i = 0; i < no_of_links; i++)
    {
/* In this version print not only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
		 delta_t = ancles_time - 
			 prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD];
		 
		 
		 fprintf(out_fp, "\t%d\t%s\t%d\t%.2f",
			 i,link_name(i), link_vect[i]->max_waves,
			 link_vect[i]->link_len);
		 
		 for(j=0; j < link_vect[i]->max_waves; j++)
		 {
			 /* update the histograms */
			 free_asonfibers_accu[i][j] += 
				 link_vect[i]->free_fibers[j] * delta_t;
			 
			 fprintf(out_fp, "\t%8f", link_vect[i]->fiber_num -
				 (free_asonfibers_accu[i][j]/
				 (ancles_time-ason_link_start_time[i])) );
		 }
		 
		 fprintf(out_fp, "\n");
	 }     

    }

 fprintf(out_fp, "\n\n#Max Used Fiber\n");
 fprintf(out_fp, "\n#Link Id\tLink Name\tMax Waves\tLength");
 for(j=1; j < 8*max_wvoct_all_links + 1; j++)
	 fprintf(out_fp, "\tw%d", j);
 fprintf(out_fp, "\n");
 
 for(i = 0; i < no_of_links; i++)
    {
/* In this version print not only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
		 delta_t = ancles_time - 
			 prev_ason_link_update_time[i][TOTAL_ASON_LINK_LOAD];

		 fprintf(out_fp, "\t%d\t%s\t%d\t%.2f",
			 i,link_name(i), link_vect[i]->max_waves,
			 link_vect[i]->link_len);
		 
		 for(j=0; j < link_vect[i]->max_waves; j++)
		 {
			 if(minf_asonfibers[i][j] >
				 link_vect[i]->free_fibers[j])
				 minf_asonfibers[i][j] =
				 link_vect[i]->free_fibers[j];

			 fprintf(out_fp, "\t%8f", link_vect[i]->fiber_num -
			 minf_asonfibers[i][j] );
		 }
		 
		 fprintf(out_fp, "\n");
	 }     

    }


}

/*
**  Function  : void generate_WLC_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the WLC report -> 
**				number of calls using the waves
*/
void generate_WLC_report(FILE * out_fp)
{
 int   i, j;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# ASON: Number of Calls using the Wavelength on the Links\n");

 fprintf(out_fp, "\n#Link Id\tLink Name\tMax Waves\tLength");
 for(j=1; j < 8*max_wvoct_all_links + 1; j++)
	 fprintf(out_fp, "\tw%d", j);
 fprintf(out_fp, "\n");
 
 for(i = 0; i < no_of_links; i++)
    {
/* In this version print not only the node-node	links */
	 if(link_vect[i]->is_node_node)
	 {
		 fprintf(out_fp, "\t%d\t%s\t%d\t%.2f",
			 i,link_name(i), link_vect[i]->max_waves,
			 link_vect[i]->link_len);
		 
		 for(j=1; j < link_vect[i]->max_waves + 1; j++)
			 fprintf(out_fp, "\t%ld", calls_on_link_wave[i][j]);
		 
		 fprintf(out_fp, "\n");
	 }     

    }
}


/*
**  Function  : void generate_ROU_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the ROU report -> 
**				lists of routes
*/
void generate_ROU_report(FILE * out_fp)
{
	int i, jj, pathno;
	Path* path;
	LinkList* link;

  fprintf(out_fp,"# Lists of routes between call generators\n");
  fprintf(out_fp,"# order number, list of links\n\n");

  for(i = 0; i < no_of_cgens; i++)
     {
	  if(ctrl_traff_start(callgen_vect[i]->ref_user->id))
	  {
		  for(jj = 0; jj < no_of_cgens; jj++)
		  {
			  if(ctrl_traff_pattern(callgen_vect[jj]->ref_user->id))
			  {
				  fprintf(out_fp, "\n# From user\tTo user\tFrom node\tTo node\n");
				  fprintf(out_fp, "%s\t%s\t%s\t%s\n\n",
					  user_name(callgen_vect[i]->ref_user->id), 
					  user_name(callgen_vect[jj]->ref_user->id),
					  node_name(callgen_vect[i]->ref_user->connected_node),
					  node_name(callgen_vect[jj]->ref_user->connected_node));
				  pathno = 1;
				  path = path_mat[callgen_vect[i]->
					  choosen_routing.path_order_type][i][jj];
				  while (path)
				  {
					  fprintf(out_fp, "%d\t", pathno);
					  for (link = path->path; link; link = link->next)
						  if(link->link_elem->is_node_node)
							  fprintf(out_fp, "%s  ", link_name(link->link_elem->id));
					  fprintf(out_fp, "\n");
					  pathno++;
					  path = path->next;
				  }
				  
			  }
		  }
	  }
  }
}

/*
**  Function  : void generate_CG8_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the CG8 report->
**				Number of (opened and) closed virtual links,
**				their mean and second order holding time statistics
**				Interarrival time statistics of the virtual link reqs
*/
void generate_CG8_report(FILE * out_fp)
{
  register int   i, jj;
  long	sreq_calls;
  Call  *call_ptr;
  double htime;
  int srcuser, dstuser;

  fprintf(out_fp,"\n\n\n#NUMBER OF CLOSED VIRTUAL LINKS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE12)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE12)
				 fprintf(out_fp, "%ld\t", closed_vlinks[i][jj]);
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#MEAN LIFETIME OF VLINKS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE12)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE12)
			 {
				 if(closed_vlinks[i][jj] == 0)
					 fprintf(out_fp, "-1.00\t");
				 else
					 fprintf(out_fp, "%.2f\t", accu_vlink_lifetime[i][jj] / 
					 (double) closed_vlinks[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#LIFETIME RELATIVE DEVIATION (C2) OF VLINKS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE12)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE12)
			 {
				 if(closed_vlinks[i][jj] == 0)
					 fprintf(out_fp, "-1.00\t");
				 else
					 fprintf(out_fp, "%.2f\t", sqraccu_vlink_lifetime[i][jj] * 
					 (double) closed_vlinks[i][jj]/
					 (accu_vlink_lifetime[i][jj]*accu_vlink_lifetime[i][jj])-1.0);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#VLINK INTERARRIVAL TIME STATISTICS\n\n#");
  fprintf(out_fp,"# From\tNumReqs\tIatMean\tC2\tNumVlinks_at_reset\n");

  for(i = 0; i < no_of_users; i++)
  {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
		  fprintf(out_fp, "%s\t", user_name(i));
		  
		  /* sum all the requested call of a generator minus those opened at the reset */
		  sreq_calls = 0 - calls_in_at_reset[i];
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			  sreq_calls += req_calls[i][jj];
		  }

		  if(sreq_calls == 0)
			  fprintf(out_fp, "0\t-1.00\t-1.00\t%ld\n",calls_in_at_reset[i]);
		  else
			  fprintf(out_fp, "%ld\t%.2f\t%.2f\t%ld\n", sreq_calls, 
				accu_vlreq_iat[i]/(double)sreq_calls,
				sqraccu_vlreq_iat[i]*(double)sreq_calls/
				(accu_vlreq_iat[i]*accu_vlreq_iat[i])-1.0,
				calls_in_at_reset[i]);
	  }
  }

/* collecting data from the remained calls too */
  for(call_ptr = head_of_call_list->next; call_ptr != head_of_call_list;
       call_ptr = call_ptr->next)
	if(call_ptr->call_type == CALL_TYPE8)
	{
		htime = ancles_time - call_ptr->start_time;
		srcuser = callgen_vect[call_ptr->source_id]->ref_user->id;
		dstuser = callgen_vect[call_ptr->dest_id]->ref_user->id;
		
		accu_vlink_lifetime[srcuser][dstuser] += htime;
		accu_vlink_timehop[srcuser][dstuser] += 
			htime * call_ptr->path_ptr->no_of_hops;
	}

  fprintf(out_fp,"\n\n\n#MEAN TIMEWEIGHTED HOPS OF PATHS FOR VLINKS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE12)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE12)
			 {
				 if(accu_vlink_lifetime[i][jj] == 0)
					 fprintf(out_fp, "-1.00\t");
				 else
					 fprintf(out_fp, "%.2f\t", accu_vlink_timehop[i][jj] /
					 accu_vlink_lifetime[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#TIMEWEIGHTS FOR VLINKS\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE12)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE8)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE12)
			 {
				 fprintf(out_fp, "%.2f\t", accu_vlink_lifetime[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
}

/*
**  Function  : void generate_UCL_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the UCL report.
*/
void generate_UCL_report(FILE * out_fp)
{
 register int   i;
 double         link_load;
 Link_ucall_stat  *actual_lUs;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }

 fprintf(out_fp, "# Load coming from calls of Class U.\n\n");
 fprintf(out_fp, "# Link Id : Link Identifier.\n");
 fprintf(out_fp, "# FBW     : Free BandWidth, available for Class U ");
 fprintf(out_fp, "calls.\n");
 fprintf(out_fp, "# ABW     : BandWidth Assigned to Class U calls.\n");
 fprintf(out_fp, "# UFBW    : Unemployed Free BandWidth, available ");
 fprintf(out_fp, "for Class U calls,\n");
 fprintf(out_fp, "#\t    but not directly assigned.\n");
 fprintf(out_fp, "# TPCR    : Total Peak Cell Rate of Class U calls.");
 fprintf(out_fp, "\n\n");
 fprintf(out_fp, "# Link Id\tFBW\tABW\tUFBW\tTPCR\t");
 fprintf(out_fp, "Act calls\tCapacity\tLink Name\n\n");
 for(i = 0; i < no_of_links; i++)
   {
    actual_lUs = link_Ucall_stat[i];
  /* First, the performance indices are updated.          */
    actual_lUs->fb_integ += actual_lUs->free_Uband *
                         (ancles_time - actual_lUs->last_updt_time);
    actual_lUs->last_updt_time = ancles_time;

    actual_lUs->un_fb_integ += actual_lUs->unused_free_band *
                         (ancles_time - actual_lUs->last_unus_time);
    actual_lUs->last_unus_time = ancles_time;

    actual_lUs->U_PCR_int += actual_lUs->acc_PCR_U *
                         (ancles_time - actual_lUs->last_Uupdt_time);
    actual_lUs->no_of_Ucall_integ += actual_lUs->no_of_Ucall *
                         (ancles_time - actual_lUs->last_Uupdt_time);
    actual_lUs->last_Uupdt_time = ancles_time;

  /* Then, the results are printed in the UCL file.       */
    link_load = actual_lUs->fb_integ /
                 (ancles_time - actual_lUs->start_time);
    fprintf(out_fp, "   %d\t%.3f\t", i, link_load);

    link_load =  actual_lUs->fb_integ /
                 (ancles_time - actual_lUs->start_time) -
                  actual_lUs->un_fb_integ /
                 (ancles_time - actual_lUs->start_time);
    fprintf(out_fp, "%.3f\t", link_load);

    link_load = actual_lUs->un_fb_integ /
                 (ancles_time - actual_lUs->start_time);
    fprintf(out_fp, "%.3f\t", link_load);

    link_load = actual_lUs->U_PCR_int /
                 (ancles_time - actual_lUs->start_time);
    fprintf(out_fp, "%.3f\t", link_load);

    fprintf(out_fp, "%.3f\t",
	    actual_lUs->no_of_Ucall_integ
	    / (ancles_time - actual_lUs->start_time));

    link_load = capacity / link_vect[i]->timescale;
    fprintf(out_fp, "%.3f\t# %s\n", link_load, link_name(i));
   }
}


/*
**  Function  : void generate_UBR_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the UBR report.
*/
void generate_UBR_report(FILE * out_fp)
{
 register int   i, j, jj;
 int 		len_idx;
 long		sreq_calls;
 long		sref_calls;
 long		sint_calls;
 CallType5      *call_info;
 Call           *call_ptr;
 double		sdilat, mdilat;
 double		load, trans;

 switch (normal)
   {
#ifdef unix
       case SIGINT:
       fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
       fprintf(out_fp,"SIGINT SIGNAL.");
       fprintf(out_fp,"\n          All the output file are written using");
       fprintf(out_fp,"\n          current acheived precision.\n\n");
       break;
    case SIGUSR2:
       fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
       fprintf(out_fp,"SIGUSR2 SIGNAL.");
       fprintf(out_fp,"\n          the average bandwidth value (Awg bw)");
       fprintf(out_fp,"\n          may be inaccurate.\n\n");
       break;
#endif
    default:
       break;
   }


 for(i = 0; i < no_of_cgens; i++)
  {
   if((callgen_vect[i]->type == CG_TYPE5) &&
      (user_start_time[callgen_vect[i]->ref_user->id] == 0.0))
      {
       fprintf(out_fp,
         "\nWarning:  The initial transient is not over\n");
       fprintf(out_fp, "          for call blocking probability ");
       fprintf(out_fp, "regarding user %s\n",
                user_name(callgen_vect[i]->ref_user->id));
      }
  }

/* first update the statistic variable, ONLY if we are at the end of the */
/* simulation */
#ifdef unix
 if(normal != SIGUSR2)
#endif
 for(call_ptr = head_of_call_list->next; call_ptr != head_of_call_list;
       call_ptr = call_ptr->next)
  if(call_ptr->call_type == CALL_TYPE5)
   {
    call_info = ((CallType5 *)(call_ptr->info));

  /* If the call is not closed yet, update its statistics */
    if(ancles_time != call_info->last_load_time)
      {
       call_info->br_integ += call_info->curr_br *
                         (ancles_time - call_info->last_updt_time);
       call_info->last_updt_time = ancles_time;
       len_idx = (int)floor((double)LEN_IDX_MAX * ((ancles_time -
	 call_info->last_load_time)/((CGType5 *)callgen_vect
	 [call_ptr->source_id]->info)->longest_call));

       gen_finUload[call_ptr->source_id] += (call_info->br_integ /
                         (ancles_time - call_info->last_load_time));
       fin_calls_dilacc[callgen_vect[call_ptr->source_id]->ref_user->id]
	 [callgen_vect[call_ptr->dest_id]->ref_user->id] +=
	  (ancles_time - call_info->last_load_time)/(call_info->br_integ/
	    call_info->mbr_value);
       gen_finUtras[call_ptr->source_id] += call_info->br_integ;

       usr_finUload[callgen_vect[call_ptr->source_id]->ref_user->id]
	 [callgen_vect[call_ptr->dest_id]->ref_user->id]+=call_info->br_integ /
                         (ancles_time - call_info->last_load_time);

       classed_fin_calls[callgen_vect[call_ptr->source_id]->ref_user->id]
	 [callgen_vect[call_ptr->dest_id]->ref_user->id][call_info->class]++;
       if(len_idx < LEN_IDX_MAX)
	 U_fincall_dur[callgen_vect[call_ptr->source_id]
	   ->ref_user->id][len_idx]++;
       else
	 U_fincall_dur[callgen_vect[call_ptr->source_id]
	   ->ref_user->id][LEN_IDX_MAX-1]++;

       call_info->last_load_time = ancles_time;
      }
   }


 fprintf(out_fp, "# Avg Bw: Average Bandwidth used by the completed calls ");
 fprintf(out_fp, "generated by Class U users.\n");
 fprintf(out_fp, "# Req Bw: Average value of the Maximum Bandwidth ");
 fprintf(out_fp, "requested by Class U users.\n");
 fprintf(out_fp, "# Avg Tr: Average Amount of data transfered by the completed calls ");
 fprintf(out_fp, "generated by Class U users.\n");
 fprintf(out_fp, "# Acc Calls: Number of Accepted Calls (only ");
 fprintf(out_fp, "stationary phase).\n");
 fprintf(out_fp, "# Req Calls: Number of Requested Calls (only ");
 fprintf(out_fp, "stationary phase).\n");
 fprintf(out_fp, "# Int Calls: Number of Interrupted Calls (only ");
 fprintf(out_fp, "stationary phase).\n");
 fprintf(out_fp, "# Dilat: Time Dilatation of Completed Calls (only ");
 fprintf(out_fp, "stationary phase).\n");

 fprintf(out_fp, "\n# User\tAvg Bw\tReq Bw\tAvg Tr\tAcc Calls");
 fprintf(out_fp, "\tReq Calls\tInt Calls\tDilat\tUser Name\n\n");

 for(i = 0; i < no_of_cgens; i++)
   {
    if(callgen_vect[i]->type == CG_TYPE5
       && source_prob[callgen_vect[i]->ref_user->id] > 0.0)
      {
      /* sum all the requested, refused and interrupted calls to get	*/
      /* the average bandwidth & transported data amount for each user	*/
       sreq_calls = 0;
       sref_calls = 0;
       sint_calls = 0;
       sdilat = 0.0;
       for( j = 0; j < no_of_users; j++)
	  {
	  sreq_calls += req_calls[callgen_vect[i]->ref_user->id][j];
	  sref_calls += ref_calls[callgen_vect[i]->ref_user->id][j];
	  sint_calls += int_calls[callgen_vect[i]->ref_user->id][j];
	  sdilat += fin_calls_dilacc[callgen_vect[i]->ref_user->id][j];
	  }

       load = gen_finUload[i] / (sreq_calls - sref_calls - sint_calls);
       trans = gen_finUtras[i] / (sreq_calls - sref_calls - sint_calls);
       mdilat = sdilat / (sreq_calls - sref_calls - sint_calls);

       fprintf(out_fp,"   %d\t%.2f\t%.2f\t%.2f\t%ld\t%ld\t%ld\t%.2f\t# %s\n",
                  callgen_vect[i]->ref_user->id, load,
		  (((CGType5 *)callgen_vect[i]->info)->min_max_rate +
		  ((CGType5 *)callgen_vect[i]->info)->max_max_rate) /
		  2.0,
		  trans,
                  primary_calls[callgen_vect[i]->ref_user->id] +
                  alternate_calls[callgen_vect[i]->ref_user->id],
                  sreq_calls,
		  sint_calls,
		  mdilat,
                  user_name(callgen_vect[i]->ref_user->id));
      }
   }

/* We report also the more detailed from-to matrices for the
	staitistics about: number of generated, refused and interrupted calls
	and the mean transport bandwidths */
  fprintf(out_fp,"\n\n#########################################\n\n");
  fprintf(out_fp,"\n\n\n#PAIR DETAILS FOR AVAILABLE BANDWIDTH\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE11)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE5)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE11)
			 {
				 if(req_calls[i][jj]-ref_calls[i][jj]-int_calls[i][jj])
					 fprintf(out_fp, "%.2f\t", usr_finUload[i][jj]/
					 (double)(req_calls[i][jj]-ref_calls[i][jj]-int_calls[i][jj]));
				 else
					 fprintf(out_fp, "-1.00\t");
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }

  fprintf(out_fp,"\n\n\n#PAIR DETAILS FOR ACCEPTED CALL NUMBER\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE11)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE5)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE11)
			 {
				 fprintf(out_fp,"%ld\t",req_calls[i][jj]-ref_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
  
  fprintf(out_fp,"\n\n\n#PAIR DETAILS FOR INTERRUPTED CALL NUMBER\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE11)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE5)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE11)
			 {
				 fprintf(out_fp,"%ld\t",int_calls[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
  
  fprintf(out_fp,"\n\n\n#NUMBER OF CHANGES OF THE ROUTING TABLE SP\n\n#");
  for(jj = 0; jj < no_of_users; jj++)
	 if(user_vect[jj]->generator->type == CG_TYPE11)
		 fprintf(out_fp, "%s\t", user_name(jj));
  fprintf(out_fp,"# From\n");

  for(i = 0; i < no_of_users; i++)
     {
	  if(user_vect[i]->generator->type == CG_TYPE5)
	  {
/* every coloumn is reported */
		  for(jj = 0; jj < no_of_users; jj++)
		  {
			 if(user_vect[jj]->generator->type == CG_TYPE11)
			 {
				 fprintf(out_fp,"%ld\t",rt_sp_changed[i][jj]);
			 }
		  }
		  fprintf(out_fp, "# %s\n", user_name(i));
	  }
	 }
  
}

/*
**  Function  : void generate_IBR_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the IBR report.
*/
void generate_IBR_report(FILE * out_fp)
{
 register int   i, j;
 long		sint_calls;
 double		load, trans;

 fprintf(out_fp, "# Avg Bw: Average Bandwidth used by the interrupted calls ");
 fprintf(out_fp, "generated by Class U users.\n");
 fprintf(out_fp, "# Req Bw: Average value of the Maximum Bandwidth ");
 fprintf(out_fp, "requested by Class U users.\n");
 fprintf(out_fp, "# Avg Tr: Average Amount of data transfered by interrupted calls ");
 fprintf(out_fp, "generated by Class U users.\n");

 fprintf(out_fp, "\n# User\tAvg Bw\tReq Bw\tAvg Tr\tUser Name\n\n");
 for(i = 0; i < no_of_cgens; i++)
   {
    load = 0.0;
    trans = 0.0;
    if(callgen_vect[i]->type == CG_TYPE5
       && source_prob[callgen_vect[i]->ref_user->id] > 0.0)
      {
      /* sum all the interrupted calls to get the average	*/
      /* bandwidth and transported data amount for each user	*/
       sint_calls = 0;
       for( j = 0; j < no_of_users; j++)
	 sint_calls += int_calls[callgen_vect[i]->ref_user->id][j];

       if(sint_calls > 0)
        {
         load = gen_intUload[i] / sint_calls;
         trans = gen_intUtras[i] / sint_calls;
	}

       fprintf(out_fp,"   %d\t%.2f\t%.2f\t%.2f\t# %s\n",
                  callgen_vect[i]->ref_user->id, load,
		  (((CGType5 *)callgen_vect[i]->info)->min_max_rate +
		  ((CGType5 *)callgen_vect[i]->info)->max_max_rate) /
		  2.0,
		  trans,
                  user_name(callgen_vect[i]->ref_user->id));
      }
   }
}

/*
**  Function  : void generate_4CC_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the 4CC report.
*/
void generate_4CC_report(FILE * out_fp)
{
 register int   i, j, k;
 long		sint_calls;
 long		sfin_calls;
 double		prtrans;

 fprintf(out_fp, "# ACx: Accepted calls of amount class x ");
 fprintf(out_fp, "generated by Class U users.\n");
 fprintf(out_fp, "# ICx: Interrupted calls of amount class x ");
 fprintf(out_fp, "generated by Class U users.\n");
 fprintf(out_fp, "# AvITPx: Average rate of transported/total data by ");
 fprintf(out_fp, " the Interrupted calls of amount class x\n");

 fprintf(out_fp, "\n# User\tAC0\tIC0\tAvITP0\tAC1\tIC1\tAvITP1");
 fprintf(out_fp, "\tAC2\tIC2\tAvITP2\tAC3\tIC3\tAvITP3\tUser Name\n\n");
 for(i = 0; i < no_of_cgens; i++)
   {
    prtrans = 0.0;
    if(callgen_vect[i]->type == CG_TYPE5
       && source_prob[callgen_vect[i]->ref_user->id] > 0.0)
      {
       fprintf(out_fp,"   %d", callgen_vect[i]->ref_user->id);
      /* for all the 4 classes sum all the interrupted calls to get the	*/
      /* average bandwidth and transported data amount for each user	*/
       for(k=0; k<4; k++)
        {
         sfin_calls = 0;
         sint_calls = 0;
         for( j = 0; j < no_of_users; j++)
	   {
	    sfin_calls += classed_fin_calls[callgen_vect[i]->
	    	ref_user->id][j][k];
	    sint_calls += classed_int_calls[callgen_vect[i]->
	    	ref_user->id][j][k];
	   }
         if(sint_calls > 0)
          {
           prtrans = classed_prop_intUtras[i][k] / sint_calls;
	  }

         fprintf(out_fp,"\t%ld\t%ld\t%.3f", sfin_calls + sint_calls,
	     sint_calls, prtrans);
	}
       fprintf(out_fp,"\t# %s\n", user_name(callgen_vect[i]->ref_user->id));
      }
   }
}

/*
**  Function  : void generate_ULH_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the ULH report.
*/
void generate_ULH_report(FILE * out_fp)
{
 register int   i, j;
 long		sacc_calls;

 fprintf(out_fp, "# Best effort calls duration histogram (completed calls only!)\n");
 fprintf(out_fp, "# Step: Histogram step given in seconds.");
 fprintf(out_fp, "# <number>: Histogram value of call lengths (only ");
 fprintf(out_fp, "stationary phase).\n");
 fprintf(out_fp, "\n# User\tStep");
 for(j = 0; j < LEN_IDX_MAX; j++)
    fprintf(out_fp, "\t%d",j);
 fprintf(out_fp, "\tName\n\n");
 for(i = 0; i < no_of_cgens; i++)
   {
    if(callgen_vect[i]->type == CG_TYPE5
       && source_prob[callgen_vect[i]->ref_user->id] > 0.0)
      {
      /* for the computation of probablities sum all the accepted calls minus
         the  interrupted ones */
       sacc_calls = 0;
       for( j = 0; j < no_of_users; j++)
	  {
	  sacc_calls += req_calls[callgen_vect[i]->ref_user->id][j];
	  sacc_calls -= ref_calls[callgen_vect[i]->ref_user->id][j];
	  sacc_calls -= int_calls[callgen_vect[i]->ref_user->id][j];
	  }

       fprintf(out_fp,"   %d\t%.2f", callgen_vect[i]->ref_user->id,
       		  ((CGType5 *)callgen_vect[i]->info)->longest_call/
		  LEN_IDX_MAX);
		
       if(sacc_calls > 0 )
        {
	 for(j = 0; j < LEN_IDX_MAX; j++)
           fprintf(out_fp,"\t%.4f",
	     (double)U_fincall_dur[callgen_vect[i]->ref_user->id][j]/
	     (double)sacc_calls);
	}
       else
        {
         for(j = 0; j < LEN_IDX_MAX; j++)
           fprintf(out_fp,"\t0.000");
	}

       fprintf(out_fp,"\t# %s\n",user_name(callgen_vect[i]->ref_user->id));
      }
   }
}

/*
**  Function  : void generate_ILH_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the ILH report.
*/
void generate_ILH_report(FILE * out_fp)
{
 register int   i, j;
 long		sint_calls;

 fprintf(out_fp, "# Best effort calls duration histogram (interrupted calls only!)\n");
 fprintf(out_fp, "# Step: Histogram step given in seconds.");
 fprintf(out_fp, "# <number>: Histogram value of call lengths (only ");
 fprintf(out_fp, "stationary phase).\n");
 fprintf(out_fp, "\n# User\tStep");
 for(j = 0; j < LEN_IDX_MAX; j++)
    fprintf(out_fp, "\t%d",j);
 fprintf(out_fp, "\tName\n\n");
 for(i = 0; i < no_of_cgens; i++)
   {
    if(callgen_vect[i]->type == CG_TYPE5
       && source_prob[callgen_vect[i]->ref_user->id] > 0.0)
      {
      /* sum the interrupted calls to compute the probablity */
       sint_calls = 0;
       for( j = 0; j < no_of_users; j++)
	  {
	  sint_calls += int_calls[callgen_vect[i]->ref_user->id][j];
	  }

       fprintf(out_fp,"   %d\t%.2f", callgen_vect[i]->ref_user->id,
       		  ((CGType5 *)callgen_vect[i]->info)->longest_call/
		  LEN_IDX_MAX);
		
       if(sint_calls > 0 )
        {
         for(j = 0; j < LEN_IDX_MAX; j++)
           fprintf(out_fp,"\t%.4f",
	     (double)U_intcall_dur[callgen_vect[i]->ref_user->id][j]/
	     (double)sint_calls);
	}
       else
        {
         for(j = 0; j < LEN_IDX_MAX; j++)
           fprintf(out_fp,"\t0.000");
	}

       fprintf(out_fp,"\t# %s\n",user_name(callgen_vect[i]->ref_user->id));
      }
   }
}

/*
**  Function  : void generate_HST_report(FILE * out_fp)
**  Return    : none
**  Remarks   : Fills file_fp with the HST report.
*/
void generate_HST_report(FILE * out_fp)
{
 register int   i, j;

 if (!normal)
   {
    fprintf(out_fp,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(out_fp,"SIGINT SIGNAL.");
    fprintf(out_fp,"\n          All the output file are written using");
    fprintf(out_fp,"\n          current acheived precision.\n\n");
   }
 for(i = 0; i < no_of_links; i++)
    {
     if(hst_intervals[i] > 0)
      {
       double acc_time = 0.0;
       fprintf(out_fp, "\n# Load Distribution in Link %d (%s)",
           i, link_name(i));
       for(j = 0; j <= hst_intervals[i]; j++)
           acc_time += hst_link_load[i][j];
       for(j = 0; j <= hst_intervals[i]; j++)
           fprintf(out_fp, "\n\t%d\t%.2f\t%f", j + 1, (capacity * j) /
                    (link_vect[i]->timescale * hst_intervals[i]),
                    hst_link_load[i][j] / acc_time);
       fprintf(out_fp, "\t# Link Load exceeding Capacity\n");
      }
    }
}

/*
**  Function  : void generate_STR_report()
**  Return    : none
**  Remarks   : Fills file_fp with the HST report.
*/
void generate_STR_report()
{
 register int   i;
 double charged_CPU_time;

 fprintf(fp_statistics, "\n# --- Final Statistical Report ---\n");

 for(i = 0; i < no_stat_vars; i++)
        {
         report_statistics(fp_statistics, i, TRUE);
         if(transient_is_over(i))
             report_statistics(fp_statistics, i, FALSE);
        }

 for(i = 0; i < 2; i++)
   if (global_stat_vars[i].info != NOT_ANALYSED)
     {
       if(transient_is_over(global_stat_vars[i].number))
         {
          fprintf(fp_statistics, "\n# Transient Phase Duration for ");
          fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                  names_stat_vars[global_stat_vars[i].number]);
          fprintf(fp_statistics, "\t\t\t\t      : %f",
                  global_start_time[i]);
          fprintf(fp_statistics, "\n# Stationary Phase Duration for ");
          fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                  names_stat_vars[global_stat_vars[i].number]);
          fprintf(fp_statistics, "\t\t\t\t      : %f",
                  (ancles_time - global_start_time[i]) );
         }
       else
          fprintf(fp_statistics, "\n# Transient Phase not over for : %s\n",
                        names_stat_vars[global_stat_vars[i].number]);
     }

 for(i = 0; i < no_of_links; i++)
    if(link_stat_vars[i].info != NOT_ANALYSED)
         {
          if(transient_is_over(link_stat_vars[i].number))
              {
               fprintf(fp_statistics, "\n# Transient Phase Duration for ");
               fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                       names_stat_vars[link_stat_vars[i].number]);
               fprintf(fp_statistics, "\t\t\t\t      : %f",
                       link_start_time[i]);
               fprintf(fp_statistics, "\n# Stationary Phase Duration for ");
               fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                       names_stat_vars[link_stat_vars[i].number]);
               fprintf(fp_statistics, "\t\t\t\t      : %f",
                       (ancles_time - link_start_time[i]) );
              }
          else
              fprintf(fp_statistics, "\n# Transient Phase not over for : %s\n",
                        names_stat_vars[link_stat_vars[i].number]);
         }

 for(i = 0; i < no_of_users; i++)
    if(user_stat_vars[i].info != NOT_ANALYSED)
         {
          if(transient_is_over(user_stat_vars[i].number))
              {
               fprintf(fp_statistics, "\n# Transient Phase Duration for ");
               fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                       names_stat_vars[user_stat_vars[i].number]);
               fprintf(fp_statistics, "\t\t\t\t      : %f",
                       user_start_time[i]);
               fprintf(fp_statistics, "\n# Stationary Phase Duration for ");
               fprintf(fp_statistics, "%s\n# (absolute simulation time)",
                       names_stat_vars[user_stat_vars[i].number]);
               fprintf(fp_statistics, "\t\t\t\t      : %f",
                       (ancles_time - user_start_time[i]) );
              }
          else
              fprintf(fp_statistics, "\n# Transient Phase not over for : %s\n",
                        names_stat_vars[user_stat_vars[i].number]);
         }

 for(i = 0; i < no_of_users; i++)
   if(classU_user_stat_vars[i].info != NOT_ANALYSED)
     {
       if(transient_is_over(classU_user_stat_vars[i].number))
	 {
	   fprintf(fp_statistics, "\n# Transient Phase Duration for ");
	   fprintf(fp_statistics, "%s\n# (absolute simulation time)",
		   names_stat_vars[classU_user_stat_vars[i].number]);
	   fprintf(fp_statistics, "\t\t\t\t      : %f",
		   classU_user_start_time[i]);
	   fprintf(fp_statistics, "\n# Stationary Phase Duration for ");
	   fprintf(fp_statistics, "%s\n# (absolute simulation time)",
		   names_stat_vars[classU_user_stat_vars[i].number]);
	   fprintf(fp_statistics, "\t\t\t\t      : %f",
		   (ancles_time - classU_user_start_time[i]) );
	 }
       else
	 fprintf(fp_statistics, "\n# Transient Phase not over for : %s\n",
		 names_stat_vars[classU_user_stat_vars[i].number]);
     }

 if(sched_events_start_time != 0.0)
     {
      fprintf(fp_statistics, "\n# Mean Number of Events seen in FES by a new ");
      fprintf(fp_statistics, "arrival\n# during the Stationary Phase");
      fprintf(fp_statistics, "\t\t\t\t      : %f",
                            (double)sched_events_acc / no_of_FES_operations);
     }
 else
     {
      fprintf(fp_statistics, "\n# Mean number of Events seen in FES by a new ");
      fprintf(fp_statistics, "arrival\n# during the Transient Phase");
      fprintf(fp_statistics, "\t\t\t\t      : %f",
                            (double)sched_events_acc / no_of_FES_operations);
     }

 fprintf(fp_statistics, "\n# Number of Events realized during the whole ");
 fprintf(fp_statistics, "simulation run   : %ld", real_events_cnt);

 charged_CPU_time = ((double)end_CPU_time - (double)start_CPU_time) /
                    CLOCKS_PER_SEC;
 fprintf(fp_statistics, "\n# CPU time required by the whole simulation ");
 fprintf(fp_statistics, "run (seconds)     : %.2f", charged_CPU_time);
 fprintf(fp_statistics, "\n# Number of realized events per charged CPU ");
 fprintf(fp_statistics, "second            : %.2f",
                        ((double)real_events_cnt) / charged_CPU_time);
 fprintf(fp_statistics, "\n# Internal Ancles time at the end of the sim");
 fprintf(fp_statistics, "ulation           : %.2f\n", ancles_time);

 if (!normal)
   {
    fprintf(fp_statistics,"\n\n\n WARNING: THE SIMULATION WAS STOPPED BY A ");
    fprintf(fp_statistics,"SIGINT SIGNAL.");
    fprintf(fp_statistics,"\n          All the output file are written using");
    fprintf(fp_statistics,"\n          current acheived precision.\n\n");
   }
}

/*
**  Function  : Boolean ctrl_traff_start(int     user_src)
**  Return    : TRUE if the traffic starts from user_src is greater than zero.
**  Remarks   : none
*/
static Boolean ctrl_traff_start(int     user_src)
{
register int i;
for(i = 0; i < no_of_users; i++)
    if(destination_prob[user_src][i] > 0.0 ||
/* Type8 calls can be generated even if there is no traffic directed to it! */
		req_calls[user_src][i] > 0)
	return TRUE;
return FALSE;
}

/*
**  Function  : Boolean ctrl_traff_pattern(int     user_dest)
**  Return    : TRUE if the traffic going to user_dest is greater than zero.
**  Remarks   : none
*/
static Boolean ctrl_traff_pattern(int     user_dest)
{
register int i;
if(user_dest > 0)
    for(i = 0; i < no_of_users; i++)
	{
	    if(destination_prob[i][user_dest] - destination_prob[i][user_dest - 1] > 0.0
/* Type8 calls can be generated even if there is no traffic directed to it! */
			|| req_calls[i][user_dest] > 0)
		return TRUE;
	}	
else
    for(i = 0; i < no_of_users; i++)
	if(destination_prob[i][user_dest] > 0.0
/* Type8 calls can be generated even if there is no traffic directed to it! */
			|| req_calls[i][user_dest] > 0)
	    return TRUE;
return FALSE;
}

/* returns system usage in sec LINUX only! *
double get_usage_stime() {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    return usage.ru_stime.tv_sec +
((double)usage.ru_stime.tv_usec)/1000000.0;
}*/

/* returns user usage in sec LINUX only! *
double get_usage_utime() {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    return usage.ru_utime.tv_sec +
((double)usage.ru_utime.tv_usec)/1000000.0;
}
*/